/*  $Header$
 *
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "identification/identification.h"
#include "identification/bootrec.h"
#include "component_interface.h"


/*
 * This file contains the boot record areas (BRAs) for the AP Firmware
 */


/*
 * Exposed Interfaces
 */
#define BR_EXP_IF_SIZE		((sizeof(struct s_exposed_if_list) - \
				  sizeof(struct s_bootrec_hdr)) / 4)

#define BR_NOF_EXP_IF		5
struct s_exposed_if_list {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_exp_if exposed_if[BR_NOF_EXP_IF];
};

/*
 * Dependent Interfaces. Each dependent interface record
 * specifies that at least one of the interfaces should be
 * present (logical OR).
 * If a component depends on more intefaces, this is specified
 * by defining a dependent interface record for each dependent
 * interface.
 */
#define BR_DEP_IF_SIZE(x)	((sizeof(struct s_dependent_if_list_##x) - \
				  sizeof(struct s_bootrec_hdr)) / 4)

struct s_dependent_if_list_1 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_dependent_if_list_2 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_dependent_if_list_3 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_dependent_if_list_4 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_dependent_if_list_5 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_dependent_if_list_6 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

/*
 * System Bootrecord Definition
 */
const struct s_system_boot_record {
	struct s_bootrec_comp_id  component_id;
	struct s_bootrec_comp_ver component_version;
	struct s_dependent_if_list_1 dependent_if_list_1;
	struct s_dependent_if_list_2 dependent_if_list_2;
	struct s_dependent_if_list_3 dependent_if_list_3;
	struct s_dependent_if_list_4 dependent_if_list_4;
        struct s_dependent_if_list_5 dependent_if_list_5;
        struct s_dependent_if_list_6 dependent_if_list_6;
        struct s_exposed_if_list exposed_if_list;
	struct s_bootrec_end  bootrecord_end;
} system_boot_record = {
	component_id:
	{
		{ BR_CODE_COMPONENT_ID, 1 },
                BR_COMP_ID_DAP1
	},
	component_version:
	{
		{ BR_CODE_COMPONENT_VERSION, VER_STRING_SIZE },
		AP_FW_VERSION_STRING
	},
	dependent_if_list_1:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(1) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_ISL39300A, AP_FW_NEEDED_HW_IF},
		}
	},
	dependent_if_list_2:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(2) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_DBO1, AP_FW_NEEDED_DBO1_IF},
		}
	},
	dependent_if_list_3:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(3) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_PRIMARY_PRODUCT_IF, AP_FW_NEEDED_PRIMARY_PRODUCT_IF},
		}
	},
	dependent_if_list_4:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(4) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_PRIMARY_OEM_IF, AP_FW_NEEDED_PRIMARY_OEM_IF},
		}
	},
        dependent_if_list_5:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(5) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_OEM_IF, AP_FW_NEEDED_OEM_IF},
		}
	},
        dependent_if_list_6:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(6) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_PRODUCT_IF, AP_FW_NEEDED_PRODUCT_IF},
		}
	},
        exposed_if_list:
	{
		{ BR_CODE_EXPOSED_IF, BR_EXP_IF_SIZE },
		{	/* role, interface ID, variant, bottom compat, top compat */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_SNMP_IF, AP_FW_EXPOSED_SNMP_IF,
			  AP_FW_EXPOSED_SNMP_IF_BTM, AP_FW_EXPOSED_SNMP_IF_TOP},
                        { BR_IF_ROLE_SERVER, BR_IF_ID_HTTP_IF, AP_FW_EXPOSED_HTTP_IF,
                          AP_FW_EXPOSED_HTTP_IF_BTM, AP_FW_EXPOSED_HTTP_IF_TOP},
                        { BR_IF_ROLE_SERVER, BR_IF_ID_TFTP, AP_FW_EXPOSED_TFTP_IF,
                          AP_FW_EXPOSED_TFTP_IF_BTM, AP_FW_EXPOSED_TFTP_IF_TOP},
                        { BR_IF_ROLE_SERVER, BR_IF_ID_DAP1_MAP, AP_FW_EXPOSED_DAP1_MAP,
                          AP_FW_EXPOSED_DAP1_MAP_BTM, AP_FW_EXPOSED_DAP1_MAP_TOP},
                        { BR_IF_ROLE_SERVER, BR_IF_ID_DAP1_FS, AP_FW_EXPOSED_DAP1_FS,
                          AP_FW_EXPOSED_DAP1_FS_BTM, AP_FW_EXPOSED_DAP1_FS_TOP},
		}
	},
	bootrecord_end:
	{
		{ BR_CODE_END_OF_BRA, 5 },	/* bootrec code, length */
		AP_FW_CRC16,			/* CRC-16 */
		0,				/* padding */
		AP_FW_MD5,			/* MD5 */
	}
};
