/*
 *      IMPORTANT--READ CAREFULLY: The use, reproduction, modification, and
 *      creation of derivative works of this Source Code ("Document") is
 *      governed by a TECHNOLOGY LICENSE AGREEMENT("Agreement"), a legal
 *      agreement between you (either an individual or a single entity) and
 *      No Wires Needed B.V. ("NWN").  By reproducing, modifying, creating
 *      derivative works, accepting, installing, viewing, accessing,
 *      transforming or otherwise using this Document, you agree to be bound
 *      by the terms of the Agreement.  If you do not agree to the terms of
 *      the Agreement, NWN is unwilling to license the rights to use,
 *      reproduce, modify, and create derivative works of this Document to you.
 *      In such event, you may not use, reproduce, modify, create derivative
 *      works, install, view, access, transport, transform or circulate this
 *      Document, and you should promptly contact NWN or the licensor from
 *      which you acquired this Document for instructions on returning or
 *      destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:  This Source
 *      Code software contains confidential, proprietary and trade secret
 *      information of No Wires Needed B.V. and should be treated in a
 *      confidential manner.  Furthermore, this Source Code software is
 *      protected by United States copyright laws and international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Copyright (C) 1994-2000 by No Wires Needed B.V. (NWN), The Netherlands,
 *      ALL RIGHTS RESERVED
 *
 */

#ifndef __BLOBV2_H__
#define __BLOBV2_H__

#include "bloboid.h"

/*
 * Blob Information Structure
 */
#define BIS_CME                 0x0000
#define BIS_DEVS                0x0001
#define BIS_DEVDESCR            0x0002
#define BIS_END                 0xFFFF

/*
 * BIS Configuration Element size
 */
#define BIS_CME_SIZE    128

/*
 * BIS Device type definitions
 */
#define BIS_DT_DEBUG            0
#define BIS_DT_802DOT11         1
#define BIS_DT_802DOT3          2
#define BIS_DT_PCMCIA           3
#define BIS_DT_PCI_CB           4
#define BIS_DT_BLUETOOTH        5
#define BIS_DT_UART             6

#ifndef __ASSEMBLER__

#ifdef CONF_3877
#include "blob3877.h"
#endif

/*
 * BIS structures
 */

struct bis_dev_element
{
    int devs;
    char *irq_stats;

    short mtu;
    short header;
    short trailer;
};

struct device;

struct bis_dev_descr_element
{
    int type;
    int long( **funcs )( unsigned long msg, struct device *device );
};

#endif /* __ASSEMBLER__ */

/*
 * Definitions
 */

/* Recommended dynamic device numbers */
#define DEVICE_DEBUG            0
#define DEVICE_WLAN             1
#define DEVICE_HOSTDS           2
#define DEVICE_WAN		3

/* Fixed device numbers */
#define DEVICE_BLOB             14
#define DEVICE_GENERIC		15 // device can be found in msg_blob.device
#define DEVICE_MAX_COUNT	16

/* Message Identifiers */

/* These are the CRm values */

/* MCR/MRC */

#define MSG_START		0	// others
#define MSG_INIT		0	// blob
#define	MSG_ABORT		1
#define MSG_STOP		1	// blob
#define MSG_RUN                 2
#define MSG_RESET		2	// blob
#define MSG_HALT                3
#define MSG_SLEEP		3	// blob
#define MSG_STATE		4
#define MSG_REQS		5	// all
#define MSG_SERVICE		6	// all
#define MSG_TRAP		7	// all
#define MSG_FRAME_RETURN        8	// all
#define MSG_FRAME_RX            9
#define MSG_READ                9
#define MSG_CONF                10	// all
#define MSG_SETUP               11
#define MSG_FRAME_ADD           11
#define MSG_FRAME_TX            12
#define MSG_WRITE               12
#define MSG_WATCHDOG            13	// all



/*
 * assembler offsets
 */

/*
 * struct msg_blob, size 0x8 (8)
 */

#define cMSG_BLOB_sizeof		 0x8	// sizeof( struct msg_blob )
#define cMSG_BLOB_ERROR			 0	// int
#define cMSG_BLOB_DEVICE		 0x4	// char
#define cMSG_BLOB_RESERVED		 0x5	// array[3] of char

/*
 * struct msg_conf, size 0x18 (24)
 */

#define cMSG_CONF_sizeof		 0x18	// sizeof( struct msg_conf )
#define cMSG_CONF_BLOB			 0	// msg_blob
#define cMSG_CONF_BLOB_ERROR		 0	// int
#define cMSG_CONF_BLOB_DEVICE		 0x4	// char
#define cMSG_CONF_BLOB_RESERVED		 0x5	// array[3] of char
#define cMSG_CONF_OPERATION		 0x8	// int
#define cMSG_CONF_OID			 0xc	// int
#define cMSG_CONF_DATA			 0x10	// pointer to int
#define cMSG_CONF_SIZE			 0x14	// int

/*
 * struct msg_data, size 0x14 (20)
 */

#define cMSG_DATA_sizeof		 0x14	// sizeof( struct msg_data )
#define cMSG_DATA_BLOB			 0	// msg_blob
#define cMSG_DATA_BLOB_ERROR		 0	// int
#define cMSG_DATA_BLOB_DEVICE		 0x4	// char
#define cMSG_DATA_BLOB_RESERVED		 0x5	// array[3] of char
#define cMSG_DATA_DATA			 0x8	// pointer to unsigned char
#define cMSG_DATA_LENGTH		 0xc	// int
#define cMSG_DATA_MAX			 0x10	// int

/*
 * struct msg_frame, size 0x38 (56)
 */

#define cMSG_FRAME_sizeof		 0x38	// sizeof( struct msg_frame )
#define cMSG_FRAME_BLOB			 0	// msg_blob
#define cMSG_FRAME_BLOB_ERROR		 0	// int
#define cMSG_FRAME_BLOB_DEVICE		 0x4	// char
#define cMSG_FRAME_BLOB_RESERVED	 0x5	// array[3] of char
#define cMSG_FRAME_NEXT			 0x8	// pointer to msg_frame
#define cMSG_FRAME_HANDLE		 0xc	// pointer to void
#define cMSG_FRAME_MPDU			 0x10	// pointer to unsigned char
#define cMSG_FRAME_SIZE			 0x14	// int
#define cMSG_FRAME_PRIORITY		 0x18	// int
#define cMSG_FRAME_STATUS		 0x1c	// int
#define cMSG_FRAME_FLAGS		 0x20	// unsigned short
#define cMSG_FRAME_DEST			 0x22	// array[6] of unsigned char
#define cMSG_FRAME_SRC			 0x28	// array[6] of unsigned char
#define cMSG_FRAME_WDS			 0x2e	// array[6] of unsigned char
#define cMSG_FRAME_SERVICESET		 0x34	// char
#define cMSG_FRAME_ANNEX		 0x38	// pointer to frame_annex

/*
 * struct msg_init, size 0x1c (28)
 */

#define cMSG_INIT_sizeof		 0x1c	// sizeof( struct msg_init )
#define cMSG_INIT_BLOB			 0	// msg_blob
#define cMSG_INIT_BLOB_ERROR		 0	// int
#define cMSG_INIT_BLOB_DEVICE		 0x4	// char
#define cMSG_INIT_BLOB_RESERVED		 0x5	// array[3] of char
#define cMSG_INIT_UND_HANDLER		 0x8	// pointer to function
#define cMSG_INIT_SWI_HANDLER		 0xc	// pointer to function
#define cMSG_INIT_PRF_HANDLER		 0x10	// pointer to function
#define cMSG_INIT_DAT_HANDLER		 0x14	// pointer to function
#define cMSG_INIT_IRQ_HANDLER		 0x18	// pointer to function

/*
 * struct msg_setup, size 0x18 (24)
 */

#define cMSG_SETUP_sizeof		 0x18	// sizeof( struct msg_setup )
#define cMSG_SETUP_BLOB			 0	// msg_blob
#define cMSG_SETUP_BLOB_ERROR		 0	// int
#define cMSG_SETUP_BLOB_DEVICE		 0x4	// char
#define cMSG_SETUP_BLOB_RESERVED	 0x5	// array[3] of char
#define cMSG_SETUP_TX_BUF		 0x8	// pointer to char
#define cMSG_SETUP_TX_BUF_SIZE		 0xc	// int
#define cMSG_SETUP_RX_BUF		 0x10	// pointer to char
#define cMSG_SETUP_RX_BUF_SIZE		 0x14	// int

/*
 * struct msg_start, size 0x10 (16)
 */

#define cMSG_START_sizeof		 0x10	// sizeof( struct msg_start )
#define cMSG_START_BLOB			 0	// msg_blob
#define cMSG_START_BLOB_ERROR		 0	// int
#define cMSG_START_BLOB_DEVICE		 0x4	// char
#define cMSG_START_BLOB_RESERVED	 0x5	// array[3] of char
#define cMSG_START_NR			 0x8	// int
#define cMSG_START_MODE			 0xc	// int

#ifndef __ASSEMBLER__

/*
 * Message structures
 */

/*
 * The blob message appears first in any type of message.
 */

typedef struct msg_blob
{
    int error;          /* Error status returned by blob */
    char device;        /* Device number */
    char reserved[3];   /* Reserved */
} msg_blob;

typedef struct msg_init
{
    msg_blob blob;

    void(*und_handler)(void);
    void(*swi_handler)(void);
    void(*prf_handler)(void);
    void(*dat_handler)(void);
    void(*irq_handler)(void);
} msg_init;

typedef struct msg_start
{
    msg_blob blob;

    int nr;
    int mode;
} msg_start;

typedef struct msg_conf
{
    msg_blob blob;

    int operation;
    int oid;
    long *data;
    long size;
} msg_conf;

/* Annex types */
#define ANNEX_TGENERIC          -1
#define ANNEX_TRX               0
#define ANNEX_TTX               1

/* Annex flags */
#define ANNEX_RXF_CRC           0x0001
#define ANNEX_RXF_RUNT          0x0002
#define ANNEX_RXF_OVERSIZE      0x0004
#define ANNEX_RXF_ALIGN         0x0008

typedef union frame_annex
{
    struct annex
    {
        short type;
        short length;
    } generic;

    struct annex_rx
    {
        short type;
        short length;

        unsigned long clock;
        unsigned short flags;

        unsigned short rate;
        unsigned short frequency;
        unsigned char antenna;
        unsigned char preamble;
        signed char rssi;
        unsigned char quality;
    } rx;

    struct annex_tx
    {
        short type;
        short length;

        unsigned short rate;
        unsigned char antenna;
        unsigned char preamble;

        unsigned short lifetime;
        unsigned short fragment;
        unsigned char retries;
    } tx;

} frame_annex;

typedef struct msg_frame
{
    msg_blob blob;
    struct msg_frame *next;

    void *handle;

    unsigned char *mpdu;
    int size;
    int priority;
    int status;

    unsigned short flags;
    unsigned char dest[6];
    unsigned char src[6];
    unsigned char wds[6];

    unsigned char serviceset;
    unsigned char pad[3];

    frame_annex *annex;
} msg_frame;

typedef struct msg_data
{
    msg_blob blob;

    unsigned char *data;
    int length;
    int max;
} msg_data;

typedef struct msg_setup
{
    msg_blob blob;

    char *tx_buf;
    int tx_buf_size;
    char *rx_buf;
    int rx_buf_size;
} msg_setup;

#endif /* __ASSEMBLER__ */

#endif /* __BLOBV2_H__ */
