/*
 *      IMPORTANT--READ CAREFULLY: The use, reproduction, modification, and
 *      creation of derivative works of this Source Code ("Document") is
 *      governed by a TECHNOLOGY LICENSE AGREEMENT("Agreement"), a legal
 *      agreement between you (either an individual or a single entity) and
 *      No Wires Needed B.V. ("NWN").  By reproducing, modifying, creating
 *      derivative works, accepting, installing, viewing, accessing,
 *      transforming or otherwise using this Document, you agree to be bound
 *      by the terms of the Agreement.  If you do not agree to the terms of
 *      the Agreement, NWN is unwilling to license the rights to use,
 *      reproduce, modify, and create derivative works of this Document to you.
 *      In such event, you may not use, reproduce, modify, create derivative
 *      works, install, view, access, transport, transform or circulate this
 *      Document, and you should promptly contact NWN or the licensor from
 *      which you acquired this Document for instructions on returning or
 *      destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:  This Source
 *      Code software contains confidential, proprietary and trade secret
 *      information of No Wires Needed B.V. and should be treated in a
 *      confidential manner.  Furthermore, this Source Code software is
 *      protected by United States copyright laws and international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Copyright (C) 1994-2000 by No Wires Needed B.V. (NWN), The Netherlands,
 *      ALL RIGHTS RESERVED
 *
 */

#ifndef __BOOTDBG_H__
#define __BOOTDBG_H__

#define BOOTMAGIC_OKAY  0x59414B4F
#define BOOTMAGIC_DBUG  0x47554244
#define BOOTMAGIC_BRAM  0x4D415242
#define BOOTMAGIC_xMVC	0x43564D78
#define BOOTMAGIC_BSv2	0x32765342
#define BOOTMAGIC_PARM	0x4d524150
#define BOOTMAGIC_BOOT  0x544F4F42

#ifdef CONF_3893

// BOOT PARAMETER BLOCK	(3893 BootROM, see doc 553183)

#define BP_MAGIC	 0
#define BP_MESSAGE	 4
#define BP_COMMAND	 8
#define BP_CONT	        12
#define BP_PBUF_COUNT	16
#define BP_PBUF_INFO	20
#define BP_MAC_ADDR	24
#define BP_DBG_ADDR	32
#define BP_BAUDRATE	40
#define BP_DBG_ETHER	44
#define BP_CRASH_INFO	48
#define BP_BOOT_INFO	52

#define BP_SIZE		56
#define BP_END_OFFSET   0x00040000
#ifndef __ASSEMBLER__

struct printf_buf
{
    long head;
    long tail;
    long queue[1];
};

struct printf_info
{
    long size;
    struct printf_buf *buf;
};

struct mac_address 
{
    unsigned long lo;
    unsigned long hi;
};

struct memory_area
{
    void *start;
    int size;
};

struct crash_info 
{
    char *log_file;
    char *message;
    long registers;
    long printf_queues;
    char *dump_file;
    long area_count;
    struct memory_area *areas;
};

struct boot_info
{
    int delay;
    char *filename;
};

struct boot_parms 
{
    long magic;
    char *message;
    int (*command)(int argc, int *hex, const char **argv);
    void (*cont)(void);
    long printf_buf_count;
    struct printf_info *printf_info;
    struct mac_address mac_address;
    struct mac_address debug_addr;
    unsigned baudrate;
    int debug_ether;
    struct crash_info *crash_info;
    struct boot_info *boot_info;
};

#endif

// BOOT STRUCT

#define BS_OFFSET       0x0003F000 /* Offset relative to start of RAM */
#define BS_MAGIC	0x00
#define BS_MVC_RESET	0x04
#define BS_MVC_BIS	0x08
#define	BS_SRAM_BASE	0x0c
#define BS_SRAM_SIZE	0x10

#define BS_SIZE		0x14

#ifndef __ASSEMBLER__

struct boot_struct {
    unsigned long magic;
    void (*reset)( void );
    unsigned long *bis;
    unsigned long sram_base;
    unsigned long sram_size;
};
 
#endif

#else // ---------------------- !3893 -------------------------

#define BOOT_OFFSET     0x0003F000 /* Offset relative to start of RAM */
#define BOOT_MAGIC	0
#define BOOT_RAMBASE	4
#define BOOT_RAMSIZE	8

#define	DEBUG_REGDUMP	16

#ifndef __ASSEMBLER__

struct boot_struct
{
	unsigned long	boot_magic;
	unsigned long	boot_rambase;
	unsigned long	boot_ramsize;
};

#endif

#define DEBUG_OFFSET		16

#define DEBUG_MAX_QUEUE		128	/* Queue of debug messages */
#define DEBUG_MAX_EVENTS	32	/* Queue of event messages */

#ifndef __ASSEMBLER__

struct debug_struct
{
	unsigned long regdump[37];

	unsigned long head;
	unsigned long tail;
	unsigned long queue[DEBUG_MAX_QUEUE*4];
	unsigned long ehead;
	unsigned long etail;
	unsigned long events[DEBUG_MAX_EVENTS*4];
};

#endif

#endif

#endif /*__BOOTDBG_H__*/
