
/*
 * arp.c
 *
 * some parts stolen from: dhcpcd - DHCP client daemon
 * by Yoichi Hariguchi <yoichi@fore.com>
 */

#include <stdio.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "arp.h"
#include "debug.h"

/* local prototypes */
void mkArpMsg(int opcode, in_addr_t tInaddr, u_char *tHaddr, in_addr_t sInaddr, u_char *sHaddr, struct arpMsg *msg);
int openRawSocket (int *s, u_short type);


int send_arp(in_addr_t inaddr, struct ifinfo *ifbuf)
{
	int		s;			/* socket */
	int		rv;			/* return value */
	struct sockaddr addr;			/* for interface name */
	struct arpMsg	arp;

	rv = ARP_FREE;
	openRawSocket(&s, ETH_P_ARP);

	/* send arp request */
	mkArpMsg(ARPOP_REQUEST, inaddr, NULL, ifbuf->addr, ifbuf->haddr, &arp);
	bzero(&addr, sizeof(addr));
	strcpy(addr.sa_data, ifbuf->ifname);
        if ( sendto(s, &arp, sizeof(arp), 0, &addr, sizeof(addr)) < 0 ) {
		printf("sendto (arpCheck)");
		rv = ARP_ERR;
	}
	
	rv = ARP_OK;
	
	close(s);
	return(rv);
}

int recv_arp(in_addr_t inaddr, struct ifinfo *ifbuf, long timeout, int send_new_arp)
{
	int		s;			/* socket */
	int		rv;			/* return value */
	struct arpMsg	arp;
	fd_set		fdset;
	struct timeval	tm;
	time_t		prevTime;

	rv = ARP_FREE;

	openRawSocket(&s, ETH_P_ARP);

	/* wait for arp reply, and check it */
	
	tm.tv_usec = 0;
	time(&prevTime);
	
	while ( timeout > 0 ) {

		FD_ZERO(&fdset);
		FD_SET(s, &fdset);
		tm.tv_sec  = timeout;

                if ( send_new_arp ) {
                        send_arp(inaddr, ifbuf);
                        send_new_arp = 0;
                }
                
                if ( select(s+1, &fdset, NULL, NULL, &tm) < 0 ) {
			printf("select (arpCheck)");
			rv = ARP_ERR;
		}

		if ( FD_ISSET(s, &fdset) ) {
			if (recv(s, &arp, sizeof(arp), 0) < 0 ) {
				printf("recv (arpCheck)");
				rv = ARP_ERR;
			}

			if(arp.operation == htons(ARPOP_REPLY)) {
#if 0
				struct in_addr tmp;
				tmp.s_addr = arp.sInaddr;
				printf("ARP reply sip : %s  ", inet_ntoa(tmp));
				printf("shdr : %02x-%02x-%02x-%02x-%02x-%02x  ",
					arp.tHaddr[0],arp.tHaddr[1],arp.tHaddr[2],arp.tHaddr[3],arp.tHaddr[4],arp.tHaddr[5]);
				tmp.s_addr = arp.tInaddr;
				printf("tip : %s  ", inet_ntoa(tmp));
				printf("thdr : %02x-%02x-%02x-%02x-%02x-%02x\n",
				arp.tHaddr[0],arp.tHaddr[1],arp.tHaddr[2],arp.tHaddr[3],arp.tHaddr[4],arp.tHaddr[5]);
#endif	

				// If we receive an arp-reply with the requested
				// inet-address, the address is in use

				if (arp.sInaddr == inaddr) {
					rv = ARP_USED;
					break;
				}
                        }
                        
			if(arp.operation == htons(ARPOP_REQUEST)) {
				// If an arp-reply is received with *our* inet address
				// as source, somebody else is using it. Collision !
#if 0
				struct in_addr tmp;
				tmp.s_addr = arp.sInaddr;
				printf("ARP dup sip : %s  ", inet_ntoa(tmp));
				printf("shdr : %02x-%02x-%02x-%02x-%02x-%02x  ",
					arp.tHaddr[0],arp.tHaddr[1],arp.tHaddr[2],arp.tHaddr[3],arp.tHaddr[4],arp.tHaddr[5]);
				tmp.s_addr = arp.tInaddr;
				printf("tip : %s  ", inet_ntoa(tmp));
				printf("thdr : %02x-%02x-%02x-%02x-%02x-%02x\n",
				arp.tHaddr[0],arp.tHaddr[1],arp.tHaddr[2],arp.tHaddr[3],arp.tHaddr[4],arp.tHaddr[5]);
#endif	

				if(arp.sInaddr == ifbuf->addr) {
					rv = ARP_COLL;
					break;
				}
			}
		}

		timeout -= time(NULL) - prevTime;
		time(&prevTime);
	}
	close(s);
	return rv;
}


void mkArpMsg(int opcode, in_addr_t tInaddr, u_char *tHaddr,
		 	  in_addr_t sInaddr, u_char *sHaddr, struct arpMsg *msg) {
		 	  
	bzero(msg, sizeof(*msg));

	// fill in ethernet headers

	bcopy(MAC_BCAST_ADDR, msg->ethhdr.ether_dhost, 6); 	/* MAC DA */
	bcopy(sHaddr, msg->ethhdr.ether_shost, 6);		/* MAC SA */
	msg->ethhdr.ether_type = htons(ETH_P_ARP);		/* protocol type (Ethernet) */

	// fill in ARP fields

	msg->htype = htons(ARPHRD_ETHER);		/* hardware type */
	msg->ptype = htons(ETH_P_IP);			/* protocol type (ARP message) */
	msg->hlen = 6;					/* hardware address length */
	msg->plen = 4;					/* protocol address length */
	msg->operation = htons(opcode);			/* ARP op code */
	msg->sInaddr = sInaddr;				/* source IP address */
	bcopy(sHaddr, msg->sHaddr, 6);			/* source hardware address */
	msg->tInaddr = tInaddr;				/* target IP address */
	if ( opcode == ARPOP_REPLY ) {
		bcopy(tHaddr, msg->tHaddr, 6);		/* target hardware address */
	}
}


int openRawSocket (int *s, u_short type) {
	int optval = 1;

	if((*s = socket (AF_PACKET, SOCK_PACKET, htons (type))) == -1) {
#if 1
		perror("socket");
		printf("socket err\n");
#endif	
		return -1;
	}
	
	if(setsockopt (*s, SOL_SOCKET, SO_BROADCAST, &optval, sizeof (optval)) == -1) {
#if 1
		perror("setsockopt");
		printf("setsockopt err\n");
#endif	
		return -1;
    }
    
    return 0;
}


// end 
