/* 
 *
 * $Id: //depot/Software/firmware/uClinux/DuetteAPDK/main/apps-gpl/autoip/autoip.c#2 $
 * $Header: //depot/Software/firmware/uClinux/DuetteAPDK/main/apps-gpl/autoip/autoip.c#2 $
 * $Date: 2003/03/11 $
 * $DateTime: 2003/03/11 13:43:02 $
 * $Change: 14432 $
 * $File: //depot/Software/firmware/uClinux/DuetteAPDK/main/apps-gpl/autoip/autoip.c $
 * $Revision: #2 $
 * $Author: jerne $
 *
 */

#include <syslog.h>
#include <net/if.h>
#include <linux/if_ether.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>

#include "version.h"
#include "debug.h"
#include "if.h"
#include "arp.h"
#include "filenames.h"

extern struct ifinfo ifbuf;


void cleanup(int a);
int gen_ip_address(struct in_addr *inaddr, unsigned char *haddr);
void print_ifinfo(void);
void help(char **argv);
void version(void);

	

int main(int argc, char **argv) {
	char destifname[32] = "eth0";
	char probeifname[32] = "eth0";
	char initial_address[64] = "";
	struct in_addr testip;
	int rv;
	int tries, successfull_tries;
        int send_new_arp = 0;
        int update_ip_addr = 0;
        int delay;
	time_t starttime;
	signed char c;
	char *p = NULL;
	char cmd[128];
	
	while((c = getopt(argc, argv, "vha:i:")) != EOF) {
		switch(c) {
			case 'i' :
				strncpy(destifname, optarg, sizeof(destifname)-1);
				break;
			case 'a' :
				strncpy(initial_address, optarg, sizeof(initial_address)-1);
				break;
			case 'h' :
				help(argv);
				break;
			case 'v' :
				version();
				break;
		}
	}

	if( (argc - optind) > 0) {
		help(argv);
	}
	
	// If we are running on an alias interface, do all probing via
	// the parent interface... remove the ':' from the interface name
	// if required
	
	strncpy(probeifname, destifname, sizeof(probeifname));
	if( (p=strchr(probeifname, ':')) != NULL) {
		*p=0;
	}

	openlog("autoip", 0, 0);
	syslog(LOG_INFO, "Intersil autoip %s started", VERSION);
	syslog(LOG_INFO, "Destination interface '%s'", destifname);
	syslog(LOG_INFO, "Probing via interface '%s'", probeifname);		

	// Get our interface hardware address
	if_gethwaddr(probeifname);
	
	// If an initial addres is given on the command line, use this, 
 	// else make up something else.
	
	if(strlen(initial_address) > 0) {
		rv = inet_aton(initial_address, &testip);
		
		if(rv == 0) {
			syslog(LOG_INFO, "Illegal address : %s\n", initial_address);
			exit(1);
		}

	} else {
		gen_ip_address(&testip, ifbuf.haddr);
	}
	
	tries=0;
	successfull_tries=0;

        while(1) {

		// Start the selection process	
	
		while(successfull_tries < 4) {
                        // successfull tries < 4 indicates we are in the 
                        // process of trying out an ip address.
                        
                        // Start with 2 sec. interval and move to 60
                        // 60 sec. after 10 total attempts.
			delay = (tries >= 10) ? 60 : 2;
			starttime = time(NULL);		
                        
                        // Sending the arp is done in the recv_arp function
                        // since we otherwise might be too late for the 
                        // response. The send_new_arp parameter is set.
                        // source address is 0 for the new arp.
                        send_new_arp = 1;
                        ifbuf.addr=0;
                        
                        rv = recv_arp(testip.s_addr, &ifbuf, delay, send_new_arp);
		
		
			syslog(LOG_INFO, "Try %d : address %s is %s (%d).",
				tries,
				inet_ntoa(testip),
				(rv == ARP_FREE) ? "free" : "in use",
				rv);
		
			if(rv != ARP_FREE) {
				// This address is in use. Select another and restart process

				gen_ip_address(&testip, NULL);
				successfull_tries = 0;
			} else {
				successfull_tries ++;
			}
		
			while(time(NULL) < starttime + delay) {
				sleep(1);
			}
		        update_ip_addr = 1;
			tries++;
		}
                
		//
		// Update the ip address if we have a new one. Fill in the
		// address, mask and bcast fields for initilizing the interface
		//
                send_new_arp = 0; // No need to send an arp probe.
                if ( update_ip_addr ) {
                        // if_config must be called with the destifname.
                        
                        strncpy(ifbuf.ifname, destifname, sizeof(ifbuf.ifname));
                        
        	        ifbuf.addr = testip.s_addr;
	                ifbuf.mask = inet_addr("255.255.0.0");
	                ifbuf.bcast = ifbuf.addr;
		        ifbuf.bcast |= htonl(0x0000ffff);
		        ifbuf.flags = IFF_UP;

		        // Set interface to new address
	
	 	        syslog(LOG_INFO, "Configuring interface %s", destifname);
		        print_ifinfo();
                        if_config(&ifbuf);
		        
                        //
		        // send two gratuitous ARP's with our new address. 
		        //
	
		        syslog(LOG_INFO, "Sending gratuitous ARP's");
                        
                        // send_arp must be called with the generic interface name.
                        strncpy(ifbuf.ifname, probeifname, sizeof(ifbuf.ifname));
                        
		        send_arp(testip.s_addr, &ifbuf);	
		        sleep(2);
		        send_arp(testip.s_addr, &ifbuf);	


		        snprintf(cmd, sizeof(cmd), "%s %s %s 255.255.0.0",
			         IPUP_SCRIPT,
				 ifbuf.ifname,
				 inet_ntoa(testip));
		
		        system(cmd);
               	} //end of update_ip_addr.

                signal(SIGTERM, cleanup);
		signal(SIGINT,  cleanup);
		signal(SIGKILL, cleanup);
					
		//
		// Wait until a ARP reply with our inaddr may occure,
		// this means we have a collission and should retry
		// the address selection
		//

		rv = recv_arp(testip.s_addr, &ifbuf, 10000000, send_new_arp);	

                if(rv != ARP_FREE) {
			// Someone is using our selected IP address.
                        // Check if another ARP is seen in the next
                        // 10 seconds.

                        rv = recv_arp(testip.s_addr, &ifbuf, 10, send_new_arp);
                        if ( rv != ARP_FREE ) {
                           // We backoff on the selected IP address
                           // and let the other one have it.
                           // Generate a new one and tries this by 
                           // jumping to the begin of the while(1).
                           gen_ip_address(&testip, NULL);
                           successfull_tries = 0;
		        } else {
                           // Keep the one we had.
                           update_ip_addr = 0;
		        }
	       }
	}
	
	closelog();
	return 0;
}


/*
 * Removes the IP address from the interface at exit
 */
 
void cleanup(int a)
{
	ifbuf.flags = 0;
	if_config(&ifbuf);
	
	exit(0);
}



void print_ifinfo(void)
{
	struct in_addr tmp;
	
	syslog(LOG_INFO, "ifname    : %s\n", ifbuf.ifname);
	tmp.s_addr = ifbuf.addr;
	syslog(LOG_INFO, "addr      : %s\n", inet_ntoa(tmp));
	tmp.s_addr = ifbuf.mask;
	syslog(LOG_INFO, "mask      : %s\n", inet_ntoa(tmp));
	tmp.s_addr = ifbuf.bcast;
	syslog(LOG_INFO, "bcast     : %s\n", inet_ntoa(tmp));
	syslog(LOG_INFO, "hwaddr    : %02x-%02x-%02x-%02x-%02x-%02x\n", 
		ifbuf.haddr[0],
		ifbuf.haddr[1],
		ifbuf.haddr[2],
		ifbuf.haddr[3],
		ifbuf.haddr[4],
		ifbuf.haddr[5]);
}

void help(char **argv) 
{
	printf("Usage : %s [-v] [-i interface] [-a IP-address]\n", argv[0]);
	exit(0);
}

void version(void) 
{
	printf("autoip %s\n", VERSION);
	exit(0);
}

/* Generate a IP address in the range 169.254.0.0/16. If a mac address
 * is given, tries to create a address based on this address, if
 * hwaddr==NULL. generate random IP's. If it's not possible to generate
 * a valid address out of the MAC, returns random address..
 */

int gen_ip_address(struct in_addr *inaddr, unsigned char *haddr)
{
	static int firstcall=1;
	char astr[256];

	if(haddr != NULL) {
		// check if MAC address is ok to create a valid IP address.
		if( (haddr[5] != 0) && (haddr[5] != 255) ) {
			syslog(LOG_INFO, "Generating IP address from MAC address");
			snprintf(astr, sizeof(astr), "169.254.%d.%d", haddr[4], haddr[5]);
			inet_aton(astr, inaddr);
			return(0);
		} else {
			syslog(LOG_INFO, "Can't generate IP address from MAC address");
			return(gen_ip_address(inaddr, NULL));
		}
	} else {
		syslog(LOG_INFO, "Generating random IP address");
		snprintf(astr, sizeof(astr), "169.254.%d.%d",
			(rand() % 253)+1, (rand() % 253)+1);
		inet_aton(astr, inaddr);
		return(0);
	}
	// shouldnt get here...
	return 1;
}
