/*
 * Submits an event to the aplogd daemon. 
 */
 
#include <stdio.h>
#include <errno.h>
#include <aplog.h>
#include <string.h>

#define LOGDB "/etc/aplogdb"


unsigned long db_event_lookup(char *symbol);
int split_whitespace(char *str, char **part, int maxparts);
unsigned long hex2long(char *orgstr);


int aplog_main(int argc, char **argv) 
{
	unsigned long evid;

	if(argc==1) {
		fprintf(stderr, "buh\n");
		return(1);
	}

	evid = db_event_lookup(argv[1]);

	if(evid == 0) {
		fprintf(stderr, "Event %s not found in database\n", argv[1]);
		return(1);
	}

	switch(argc) {
		case 2: 
			aplog(evid, NULL);
			break;
		case 3:
			aplog(evid, argv[2], NULL);
			break;
		case 4:
			aplog(evid, argv[2], argv[3], NULL);
			break;
		case 5:
			aplog(evid, argv[2], argv[3], argv[4], NULL);
			break;
		case 6:
			aplog(evid, argv[2], argv[3], argv[4], argv[5], NULL);
			break;
		default:
			fprintf(stderr, "too many arguments.\n");
			return(1);
			break;
	}
	
	return(0);
}




/*
 * Find the given event symbol in the database
 * and return the eventid.
 */

unsigned long db_event_lookup(char *symbol)
{
	FILE *f;
	char buf[64];
	char *part[4];
	int parts;
	
	f=fopen(LOGDB, "r");

	if(f==NULL) {
		fprintf(stderr, "Can't open database %s : %s\n",
				LOGDB, strerror(errno));
		exit(1);
	}

	while(fgets(buf, sizeof(buf), f)) {
		parts = split_whitespace(buf, part, 4);
	
		if(parts>3) {
			if( (strcmp(part[0], "ev") == 0) &&
			    (strcmp(part[2], symbol) == 0) ) {
				return(hex2long(part[1]));
			}
		}
	}

	return(0);
}



/*
 * Split a string, seperated by one ore more whitespace, in 0-term parts
 */
 
int split_whitespace(char *str, char **part, int maxparts)
{
	char *p = str;
	int parts=0;

	while(*p) {
		while( (*p == ' ') || (*p == '\t')) p++;
		part[parts]=p;
		parts++;
		if(parts>=maxparts) return(parts);
		while( (*p != ' ') && (*p != '\t') && (*p != 0)) p++;
		*p=0;
		p++;
	}

	return(parts);
}



/*
 * hex2long.... ??
 */

unsigned long hex2long(char *orgstr)
{
	unsigned long t=0;
	char *p;
	char str[9];

	p=strncpy(str, orgstr, sizeof(str));

	while(*p) {
		*p=*p|32;
		if(*p>='a') *p-=39;
		*p -= 48;
		if(((*p)<0) || ((*p)>15)) return(0);
		t=(t<<4)|(*p++);
	}
	
	return(t);
}



// eof..
