#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdio.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <aplog.h>
#include <netinet/if_ether.h>
#include <arpa/inet.h>
#include <net/if_arp.h>
#include <net/if.h>
#include <netinet/in.h>
#include <errno.h>

#define MAC_BCAST_ADDR	"\xff\xff\xff\xff\xff\xff"

struct arpMsg {
    struct ethhdr ethhdr;	/* Ethernet header */
    u_short htype;		/* hardware type (must be ARPHRD_ETHER) */
    u_short ptype;		/* protocol type (must be ETH_P_IP) */
    u_char hlen;		/* hardware address length (must be 6) */
    u_char plen;		/* protocol address length (must be 4) */
    u_short operation;		/* ARP opcode */
    u_char sHaddr[6];		/* sender's hardware address */
    u_char sInaddr[4];		/* sender's IP address */
    u_char tHaddr[6];		/* target's hardware address */         
    u_char tInaddr[4];		/* target's IP address */
    u_char pad[18];		/* pad for min. Ethernet payload (60 bytes) */
} __attribute__((packed));

struct ifinfo {
    char ifname[IFNAMSIZ];
    u_long addr;		/* network byte order */
    u_long mask;		/* network byte order */
    u_long bcast;		/* network byte order */
    u_char haddr[6];
    short flags;
} __attribute__((packed));


int arpping(u_int32_t yiaddr, u_int32_t ip, char *arp, char *interface);
void usage(void);


extern int ifup_main(int argc, char **argv)
{
	char *ifname;
	char *address;
	char *netmask;
	char *gateway;
	char *hostname;
	char cmd[256];
	int r;
	char ifname2[32];
	char *p;

	if(argc >= 2) {
		ifname = argv[1];
	} else {
		fprintf(stderr, "missing interface and method parameters\n");
		usage();
		exit(1);
	}
	
	switch(argv[2][0]) {

		// static

		case 's':	
			if((argc != 5 ) && (argc != 6)) {
				fprintf(stderr, "Wrong number of parameters\n");
				usage();
				exit(1);
			}
	
			address = argv[3];
			netmask = argv[4];
		
			strncpy(ifname2, ifname, sizeof(ifname2));
			if((p=strchr(ifname2, ':'))) *p=0;
			
			/* bring interface up */
			snprintf(cmd, sizeof(cmd), "ifconfig %s up", ifname2);
			system(cmd);
			
			r=arpping((u_int32_t)inet_addr(address), (u_int32_t)0, "\xff\xff\xff\xff\xff\xff", ifname2);
	
			if(r == 0) {
				aplog(EV_NET_IPADDR_INUSE, address, NULL);
			}
	
			snprintf(cmd, sizeof(cmd), "ifconfig %s %s netmask %s",	
				ifname, address, netmask);
			system(cmd);
	
			if(argc == 6) {
				gateway = argv[5];

				snprintf(cmd, sizeof(cmd), "route add default gw %s",
					gateway);
				system(cmd);
			}

			break;

		// dhcp

		case 'd':
			if(argc == 4) {
				hostname = argv[3];
				
				snprintf(cmd, sizeof(cmd), "dhcpcd -H %s -h %s &",
						ifname, hostname);
			}
			else {
				snprintf(cmd, sizeof(cmd), "dhcpcd -H %s &", 
					ifname);
			}
			system(cmd);
	
			break;
			
			
		// autoip
	
		case 'a':
			snprintf(cmd, sizeof(cmd), "autoip -i %s &",
				ifname);
			system(cmd);
	
			break;
		case 'n':
			snprintf(cmd, sizeof(cmd), "ifconfig %s 0.0.0.0", ifname);
			system(cmd);
			break;
		default:
			fprintf(stderr, "Illegal method %s\n", argv[2]);
			usage();
			exit(1);	
	}


	return(0);
}


#define LOG fprintf(stderr, "Error at line %d : %s\n", __LINE__, strerror(errno))

void usage(void)
{
	fprintf(stderr, "ifup <interface> <method> [arguments]\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "   methods: static <address> <netmask> [gateway]\n");
	fprintf(stderr, "            dhcp [hostname]\n");
	fprintf(stderr, "            autoip\n");
	fprintf(stderr, "            none (removes ip address)\n");
}



////////////////////////////////////////////////////////////////////////
///////////////// arpping code, stolen from udhcpc /////////////////////
////////////////////////////////////////////////////////////////////////


void mkArpMsg(int opcode, u_long tInaddr, u_char * tHaddr,
	      u_long sInaddr, u_char * sHaddr, struct arpMsg *msg)
{
    bzero(msg, sizeof(*msg));
    bcopy(MAC_BCAST_ADDR, msg->ethhdr.h_dest, 6);	/* MAC DA */
    bcopy(sHaddr, msg->ethhdr.h_source, 6);	/* MAC SA */
    msg->ethhdr.h_proto = htons(ETH_P_ARP);	/* protocol type (Ethernet) */
    msg->htype = htons(ARPHRD_ETHER);	/* hardware type */
    msg->ptype = htons(ETH_P_IP);	/* protocol type (ARP message) */
    msg->hlen = 6;		/* hardware address length */
    msg->plen = 4;		/* protocol address length */
    msg->operation = htons(opcode);	/* ARP op code */
    memcpy(msg->sInaddr, &sInaddr, 4);  /* source IP address */ 
    bcopy(sHaddr, msg->sHaddr, 6);	/* source hardware address */
    memcpy(msg->tInaddr, &tInaddr, 4);    /* target IP address */
    if (opcode == ARPOP_REPLY)
	bcopy(tHaddr, msg->tHaddr, 6);	/* target hardware address */       
     
}


int openRawSocket(int *s, u_short type)
{
    int optval = 1;

    if ((*s = socket(AF_INET, SOCK_PACKET, htons(type))) == -1) {
    	LOG;
	return -1;
    }

    if (setsockopt(*s, SOL_SOCKET, SO_BROADCAST, &optval, sizeof(optval))
	== -1) {
	LOG;
	return -1;
    }
    return 0;
}


int arpCheck(u_long inaddr, struct ifinfo *ifbuf, long timeout)
{
    int s;			/* socket */
    int rv;			/* return value */
    struct sockaddr addr;	/* for interface name */
    struct arpMsg arp;
    fd_set fdset;
    struct timeval tm;
    time_t prevTime;

    rv = 1;
    if (openRawSocket(&s, ETH_P_ARP) < 0) {
    	LOG;
	return -1;
    }

    /* send arp request */
    mkArpMsg(ARPOP_REQUEST, inaddr, NULL, ifbuf->addr, ifbuf->haddr, &arp);
    bzero(&addr, sizeof(addr));

    strcpy(addr.sa_data, ifbuf->ifname);

    if ((rv = sendto(s, &arp, sizeof(arp), 0, &addr, sizeof(addr))) < 0) {
    	LOG;
	return rv;
    }

    /* wait arp reply, and check it */
    tm.tv_usec = 0;
    time(&prevTime);
    while (timeout > 0) {
	FD_ZERO(&fdset);
	FD_SET(s, &fdset);
	tm.tv_sec = timeout;
	if (select(s + 1, &fdset, (fd_set *) NULL, (fd_set *) NULL, &tm) <
	    0) {
	    if (errno != EINTR) {
		rv = 0;

	    }
	} else if (FD_ISSET(s, &fdset)) {
	    if (recv(s, &arp, sizeof(arp), 0) < 0) {
		rv = 0;
	    }
	    if (arp.operation == htons(ARPOP_REPLY) &&
		*((u_int *) arp.sInaddr) == inaddr) {
		rv = 0;
		break;
	    }
	}
	timeout -= time(NULL) - prevTime;
	time(&prevTime);
    }
    close(s);
    return rv;
}

/* args:	yiaddr - what IP to ping (eg. on the NETtel cb189701)
 * retn: 	1 addr free
 *		0 addr used
 *		-1 error
 */
int arpping(u_int32_t yiaddr, u_int32_t ip, char *arp, char *interface)
{
    struct ifinfo ifbuf;

    strcpy(ifbuf.ifname, interface);
    ifbuf.addr = ip;
    ifbuf.mask = 0x0;
    ifbuf.bcast = 0x0;

    memcpy(ifbuf.haddr, arp, 6);
    ifbuf.flags = 0;

    return arpCheck(yiaddr, &ifbuf, 2);
}

