/*
ebt_natwatcher.c
	ebtables extension for MAC NAT watcher
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <netinet/ether.h>

#include "../include/ebtables_u.h"
#include <linux/netfilter_bridge/ebt_macnat.h>

#define EBT_NATWATCHER_NAME	"natwatcher"

#define NAT_RECIPMACPAIR	'1'
#define NAT_BRIF	'2'
static struct option opts[] = {
	{ "rec-ipmac-pair", required_argument, 0, NAT_RECIPMACPAIR},
	{0}
};

static void help(){
//	printf("No extra parameter is needed, but should be used on PREROUTING chain");
	printf("--rec-ipmac-pair gateway_mac_addr : record IP and MAC address pair and configure gateway MAC address\n");
}

static void init(struct ebt_entry_watcher *w){
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)w->data;
	memset(natinfo->gwmac, 0xFF, 6);
	return ;
}

#define OPT_RECIPMACPAIR	0x01
#define OPT_GWMAC		0x02
static int parse(int c, char **argv, int argc, const struct ebt_u_entry *entry, 
    unsigned int *flags, struct ebt_entry_watcher **watcher){
    struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)(*watcher)->data;
    struct ether_addr *addr;
	switch(c){
	 case '1':
             check_option(flags, OPT_RECIPMACPAIR);
             if((addr = ether_aton(optarg)))
                 memcpy(natinfo->gwmac, addr, 6);
                 else
                     print_error("MAC address format error!\n");
		break;
	 default:
	 	return 0; // no known options
	}
	return 1;
}

static void final_check(const struct ebt_u_entry *entry, const struct ebt_entry_watcher *watcher,
	const char *name, unsigned int hookmask, unsigned int time){
	
}

static void print(const struct ebt_u_entry *entry, const struct ebt_entry_watcher *watcher){
	struct ebt_macnat_info *natinfo = (struct ebt_macnat_info *)watcher->data;
	printf("--rec-ipmac-pair %s ", ether_ntoa(natinfo->gwmac));
}

static int compare(const struct ebt_entry_watcher *w1, const struct ebt_entry_watcher *w2){
	if(strcmp(w1->u.name, w2->u.name)==0)
		return 1;
	return 0;
	
}

static struct ebt_u_watcher natwatcher = {
	EBT_NATWATCHER_NAME,
	sizeof(struct ebt_macnat_info),
	help,
	init,
	parse,
	final_check,
	print,
	compare,
	opts
};

//static void _init(void) __attribute__((constructor));
//static void _init(void){
void _init_natwatcher(void){
	register_watcher(&natwatcher);
}

