/*
 * files.c - file manipulation
 *
 * Copyright (C) 2002 by Intersil Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>
#include <string.h>
#include <sys/wait.h>
#include <inl_debug.h>
#include <sockutil.h>

#include "files.h"
#include "imginstall.h"

/* global declarations */
extern char *sector_buf;

static int file_install_image(int rfd, int size, char *destpath, int status_offset)
{
   int wfd;
   int ret = -EINVAL;
   int totalread = 0;

   wfd = open(destpath, O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR);
   if(wfd < 0) {
      error("Error opening image file %s : %m", destpath);
      return ret;
   }

   while(1) {
      int nread;

      nread = readn(rfd, sector_buf, SECTOR_SIZE);
      if(nread < 0) {
         error("Error reading image file fd %d : %m", rfd);
         goto err_out;
      }

      if(nread == 0)
         break;

      ret = writen(wfd, sector_buf, nread);
      if(ret < 0) {
         ret = -errno;
         error("Error writing image file %s : %m", destpath);
         goto err_out;
      }

      totalread += nread;
      imginstall_write_status(status_offset + (100 * totalread) / size);
   }

   /* Make sure all data is written to flash */
   fsync(wfd);
   close(wfd);
   return 0;

err_out:
   close(wfd);
   unlink(destpath);
   return ret;
}

int file_install(int fd, int type, int size)
{
   switch (type) {
      case image_uapfirmware:
         return file_install_image(fd, size, FILE_UAP_FIRMWARE,
                             STATUS_INSTALLING_UAPFIRMWARE);
      case image_webfileset:
         return file_install_image(fd, size, FILE_WEBFILESET,
                             STATUS_INSTALLING_WEBFILESET);
      default:
         warn ("file_install: invalid image type %d", type);
         return -EINVAL;
   }
}

int open_image(char *fname, int *size)
{
   int fd = open( fname, O_RDONLY);
   int ret;
   struct stat st;

   if(fd == -1) {
      warn("Can't open file %s : %m", fname);
      return -1;
   }

   ret = fstat(fd, &st);
   if (ret < 0) {
      warn("Can't fstat file %s : %m", fname);
      close(fd);
      return -1;
   }

   *size = st.st_size;
   return fd;
}

/* eof */
