/* Shared library add-on to iptables to add an user-defined index to a rule. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#if defined(__GLIBC__) && __GLIBC__ == 2
#include <net/ethernet.h>
#else
#include <linux/if_ether.h>
#endif
#include <iptables.h>
#include <linux/netfilter_ipv4/ipt_userindex.h>

/* Function which prints out usage message. */
static void
help(void)
{
#ifndef EMBED	
	printf(
"userindex v%s options:\n"
" --userindex <int>"
"				Add index to rule."
"\n", NETFILTER_VERSION);
#endif	
}

static struct option opts[] = {
	{ "userindex", 1, 0, '1' },
	{0}
};

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	/* Can't cache this */
	*nfcache |= NFC_UNKNOWN;
}


/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_userindex_info *userindexinfo = (struct ipt_userindex_info *)(*match)->data;

	switch (c) {
	case '1':
		userindexinfo->userindex = atoi(argv[optind - 1]);
		*flags = 1;
		break;

	default:
		return 0;
	}

	return 1;
}


/* Final check; must have specified --mac. */
static void final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
			   "You must specify `--userindex'");
}

/* Prints out the matchinfo. */
static void print(	const struct ipt_ip *ip,
		  	const struct ipt_entry_match *match,
      			int numeric)
{	
	printf("userindex %d ", 
		((struct ipt_userindex_info *)match->data)->userindex
		);
}

/* Saves the union ipt_matchinfo in parsable form to stdout. */

static void save(	const struct ipt_ip *ip, 
			const struct ipt_entry_match *match)
{

	printf("--userindex %d ", 
		((struct ipt_userindex_info *)match->data)->userindex
		);
}

static
struct iptables_match userindex
= { NULL,
    "userindex",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ipt_userindex_info)),
    IPT_ALIGN(sizeof(struct ipt_userindex_info)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void ipt_userindex_init(void)
{
	register_match(&userindex);
}
