/*  $Header$
 *  
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "identification/identification.h"
#include "image.h"
#include "pda.h"


/*=======================================================================================*
 | NAME:
 |    pda_Print
 |
 | DESCRIPTION:
 |    Print the complete PDA in readable format
 |    Serves purely as a debugging function
 |
 | PARAMETERS:
 |    None
 |
 | RETURN:
 |    None
 *=======================================================================================*/
void pda_Print (void)
{
   u_char *StrCode;
   u_short pdaSize   = 0;
   u_short pdaOffset = 0;
   struct  Pdr_Struct *pdrPtr;
   u_char *pdaPtr;
   u_char *Pda;
   char buf[80];

   Pda = pda_Get_PdaPtr();
   
   /*------------------------------------------------------*
    | Walk through all the PDR's in the PDA and print them |
    *------------------------------------------------------*/
   pdaPtr = (u_char *) Pda;
   pdrPtr = (struct Pdr_Struct *)Pda;
   while (pdrPtr->PdrCode != PDR_END) {
      switch (pdrPtr->PdrCode) {
         case PDR_MF_PART_NUMBER:
            StrCode = "PDR_MF_PART_NUMBER";
            break;
         case PDR_VERSION:
            StrCode = "PDR_VERSION";
            break;
         case PDR_NIC_SERIAL:
            StrCode = "PDR_NIC_SERIAL";
            break;
         case PDR_NIC_RAM_SIZE:
            StrCode = "PDR_NIC_RAM_SIZE";
            break;
         case PDR_RFMODEM_SUP_RANGE:
            StrCode = "PDR_RFMODEM_SUP_RANGE";
            break;
         case PDR_PRISM_MAC_SUP_RANGE:
            StrCode = "PDR_PRISM_MAC_SUP_RANGE";
            break;
         case PDR_NIC_ID:
            StrCode = "PDR_NIC_ID";
            break;
         case PDR_MAC_ADDRESS:
            StrCode = "PDR_MAC_ADDRESS";
            break;
         case PDR_REGULATORY_DOMAIN_LIST:
            StrCode = "PDR_REGULATORY_DOMAIN_LIST";
            break;
         case PDR_ALLOWED_CHAN_SET:
            StrCode = "PDR_ALLOWED_CHAN_SET";
            break;
         case PDR_DEFAULT_CHAN:
            StrCode = "PDR_DEFAULT_CHAN";
            break;
         case PDR_TEMPERATURE_TYPE:
            StrCode = "PDR_TEMPERATURE_TYPE";
            break;
         case PDR_IFR_SETTING:
            StrCode = "PDR_IFR_SETTING";
            break;
         case PDR_RFR_SETTING:
            StrCode = "PDR_RFR_SETTING";
            break;
         case PDR_3861_BASELINE_REG_SETTINGS:
            StrCode = "PDR_3861_BASELINE_REG_SETTINGS";
            break;
         case PDR_3861_SHADOW_REG_SETTINGS:
            StrCode = "PDR_3861_SHADOW_REG_SETTINGS";
            break;
         case PDR_3861_IFRF_REG_SETTINGS:
            StrCode = "PDR_3861_IFRF_REG_SETTINGS";
            break;
         case PDR_3861_CHAN_CALIB_SET_POINTS:
            StrCode = "PDR_3861_CHAN_CALIB_SET_POINTS";
            break;
         case PDR_3861_CHAN_CALIB_INTEGRATOR:
            StrCode = "PDR_3861_CHAN_CALIB_INTEGRATOR";
            break;
         case PDR_3842_PRISM_II_NIC_CONFIG:
            StrCode = "PDR_3842_PRISM_II_NIC_CONFIG";
            break;
         case PDR_PRISM_USB_ID:
            StrCode = "PDR_PRISM_USB_ID";
            break;
         case PDR_PRISM_PCI_ID:
            StrCode = "PDR_PRISM_PCI_ID";
            break;
         case PDR_PRISM_PCI_IF_CONFIG:
            StrCode = "PDR_PRISM_PCI_IF_CONFIG";
            break;
         case PDR_PRISM_PCI_PM_CONFIG:
            StrCode = "PDR_PRISM_PCI_PM_CONFIG";
            break;
         case PDR_3861_MF_TEST_CHAN_SET_POINTS:
            StrCode = "PDR_3861_MF_TEST_CHAN_SET_POINTS";
            break;
         case PDR_3861_MF_TEST_CHAN_INTEGRATORS:
            StrCode = "PDR_3861_MF_TEST_CHAN_INTEGRATORS";
            break;
         case PDR_COUNTRY_INFO:
            StrCode = "PDR_COUNTRY_INFO";
            break;
         case PDR_INTERFACE_LIST:
            StrCode = "PDR_INTERFACE_LIST";
            break;
         case PDR_HW_PLATFORM_COMPONENT_ID:
            StrCode = "PDR_HW_PLATFORM_COMPONENT_ID";
            break;
         case PDR_OEM_NAME:
            StrCode = "PDR_OEM_NAME";
            break;
         case PDR_PRODUCT_NAME:
            StrCode = "PDR_PRODUCT_NAME";
            break;
         case PDR_UTF8_OEM_NAME:
            StrCode = "PDR_UTF8_OEM_NAME";
            break;
         case PDR_UTF8_PRODUCT_NAME:
            StrCode = "PDR_UTF8_PRODUCT_NAME";
            break;
         case PDR_PRISM_2_PA_CALIBRATION_DATA:
            StrCode = "PDR_PRISM_2_PA_CALIBRATION_DATA";
            break;
         case PDR_PRISM_INDIGO_PA_CALIBR_DATA:
            StrCode = "PDR_PRISM_INDIGO_PA_CALIBR_DATA";
            break;
         case PDR_RSSI_LINEAR_APPROXIMATION:
            StrCode = "PDR_RSSI_LINEAR_APPROXIMATION";
            break;
         case PDR_COUNTRY_LIST:
            StrCode = "PDR_COUNTRY_LIST";
            break;
         case PDR_DEFAULT_COUNTRY:
            StrCode = "PDR_DEFAULT_COUNTRY";
            break;
	 case PDR_ANTENNA_GAIN:
	    StrCode = "PDR_ANTENNE_GAIN";
	    break;
	 case PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS:
	    StrCode = "PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS";
	    break;
	 case PDR_PRISM_PA_CAL_CURVE_DATA:
	    StrCode = "PDR_PRISM_PA_CAL_CURVE_DATA";
	    break;
	 case PDR_RSSI_LINEAR_APPROXIMATION_DB:
	    StrCode = "PDR_RSSI_LINEAR_APPROXIMATION_DB";
	    break;
	 case PDR_ZIF_TX_IQ_CALIBR_DATA:
	    StrCode = "PDR_ZIF_TX_IQ_CALIBR_DATA";
	    break;
         default:
	    snprintf(buf, sizeof(buf), "<<< UNKNOWN PDR : %4X >>>", pdrPtr->PdrCode);
	    StrCode = buf ;
            break;
      }
      printf(" PDA-> PDR.Length=%03d, PDR.Code=%s (%04X) \n", pdrPtr->PdrLength, StrCode, pdrPtr->PdrCode);
      pdaSize = pdaSize + pdrPtr->PdrLength + 1;
      pdaOffset = pdaOffset + ((pdrPtr->PdrLength + 1) * sizeof(pdrPtr->PdrLength)); /* in bytes */
      pdaPtr = (u_char *)Pda + pdaOffset;
      pdrPtr = (struct Pdr_Struct *)pdaPtr;
   }
   printf(" PDA-> PDR.Length=%02d, PDR.Code=%s\n", pdrPtr->PdrLength, "PDR_END"); /* must be PDR_END */
   pdaSize = pdaSize + pdrPtr->PdrLength + 1;  /* for PDR_END */
   printf(" PDA Total Size = %d bytes\n", pdaSize);
}


static void print_Country_List() {
	struct  Pdr_Struct *pdrPtr;
	u_char* p;
	int i;
	
	pdrPtr = pda_Find_Pdr(PDR_COUNTRY_LIST, NULL);
	if(!pdrPtr) {
		printf("PDR_COUNTRY_LIST not found in PDA\n");
		exit(-1);
	}
	p = &(pdrPtr->PdrData[0]);
	
	for(i=0; i < (pdrPtr->PdrLength/2); i++) {
		printf("Regul.Domain  : 0x%02X  ", p[0 + (i*4)]);
		printf("Country string: [%c%c%c]\n", p[1 + (i*4) ], p[2 + (i*4) ], p[3 + (i*4)]);
	}
}


static void print_Interface_List() {
	struct  Pdr_Struct *pdrPtr;
	u_char* p; 
	int i;
	unsigned int temp, temp2;
	unsigned int ifaceID;
	
	
	pdrPtr = pda_Find_Pdr(PDR_INTERFACE_LIST, NULL);
	if(!pdrPtr) {
		printf("PDR_INTERFACE_LIST not found in PDA\n");
		exit(-1);
	}
	p = &(pdrPtr->PdrData[0]);
	
	for(i=0; i < (pdrPtr->PdrLength/5); i++) {
	/*	
		printf("%02X ", *(p++));
		printf("%02X ", *(p++));
		printf("%02X ", *(p++));
		printf("%02X ", *(p++));
		printf("%02X ", *(p++));
		printf("%02X ", *(p++));
		printf("%02X ", *(p++));
		printf("%02X ", *(p++));								
		printf("%02X ", *(p++));
		printf("%02X ", *(p++));								
		printf("\n");
		continue;
	*/
		
		temp = (*p);
		p++;
		temp |= (*p << 8);
		
		printf("Interface %i\n", i);
		printf("Role        : ");
		switch(temp) {
			case 0x0000:
				printf("Server or producer of service\n");
				break;
			case 0x0001:
				printf("Client or consumer of service\n");
				break;
			case 0xFFFF:
				printf("Invalid\n");
				break;
			default:
				printf("ERROR. Undefined (%04X)\n", temp);
				break;
		}
		
		p++;
		temp = (*p);
		p++;
		temp |= (*p << 8);
		
		ifaceID = temp;

		printf("Interface ID: %04X ", temp);
		switch(temp) {
			case BR_IF_ID_ISL39300A:
				printf("( = ISL39300A-IF)");
			default:
				printf("\n");
		}
		 
		p++;
		temp = (*p);
		p++;
		temp |= (*p << 8);
		
		printf("Variant     : %04X\n", temp);
		switch(ifaceID) {
			case BR_IF_ID_ISL39300A:
				switch(temp & BR_IF_VAR_ISL39000_SYNTH_MASK) {
					case BR_IF_VAR_ISL39000_SYNTH_1:
						printf(" - SYNT used for all 32/33/34 prescaler 3690 ZIF synthesizers\n");
						break;
					case BR_IF_VAR_ISL39000_SYNTH_2:
						printf(" - SYNT used for all 32/33 prescaler 3690 ZIF synthesizers\n");
						break;
					case BR_IF_VAR_ISL39000_SYNTH_3:
						printf(" - SYNT used for the ISL3684 Prism 3 Freedom ZIF on Frisbee hardware (does not support 5GHz).\n");
						break;
					default:
						printf(" - SYNT UNSUPPORTED !!\n");
						break;
				}						
				
				switch(temp & BR_IF_VAR_ISL39000_IQ_CAL_MASK) {
					case BR_IF_VAR_ISL39000_IQ_CAL_PA_DETECTOR:
						printf(" - online TX IQ calibration uses PA detector\n");
						break;
					case BR_IF_VAR_ISL39000_IQ_CAL_DISABLED:
						printf(" - online TX IQ calibration is disabled.\n");
						break;
					case BR_IF_VAR_ISL39000_IQ_CAL_ZIF:
						printf(" - online TX IQ calibration uses ZIF detector.\n");
						break;
					default:
						printf(" - UNSUPPORTED IQ CAL ! \n");
						break;
				}

				switch(temp & BR_IF_VAR_ISL39000_FAA_SWITCH_MASK) {
					case BR_IF_VAR_ISL39000_FAA_SWITCH_ENABLED:
						printf(" - FAA switch enabled\n");
						break;
					default:
						printf(" - FAA switch is not present or disabled\n");
						break;
				}

				switch(temp & BR_IF_VAR_ISL39000_24_GHZ_MASK) {
					case BR_IF_VAR_ISL39000_24_GHZ_DISABLED:
						printf(" - Operation in the 2.4 GHz band is disabled or not supported.\n");
						break;
					default:
						printf(" - Operation in the 2.4 GHz band is enabled.\n");
						break;
				}
				
				switch(temp & BR_IF_VAR_ISL39000_5_GHZ_MASK) {
					case BR_IF_VAR_ISL39000_5_GHZ_DISABLED:
						printf(" - Operation in the 5 GHz band is disabled or not supported.\n");
						break;
					default:
						printf(" - Operation in the 5 GHz band is enabled.\n");
						break;
				}

				switch(temp & BR_IF_VAR_ISL39000_RX_DIV_MASK) {
					case BR_IF_VAR_ISL39000_RX_DIV_SUPPORTED:
						printf(" - Hardware supports Rx Antenna Diversity..\n");
						break;
					default:
						printf(" - Hardware does not support Rx Antenna Diversity..\n");
						break;
				}
				
				switch(temp & BR_IF_VAR_ISL39000_TX_DIV_MASK) {
					case BR_IF_VAR_ISL39000_TX_DIV_SUPPORTED:
						printf(" - Hardware supports Tx Antenna Diversity..\n");
						break;
					default:
						printf(" - Hardware does not support Tx Antenna Diversity..\n");
						break;
				}

				break;
				

				
			default:
				/* no variant decoding */
				break;
		}
		
		p = p + 5;
		
		
	}

}

static void print_Default_Country() {
	
	struct  Pdr_Struct *pdrPtr;
	u_char* p;
	
	pdrPtr = pda_Find_Pdr(PDR_DEFAULT_COUNTRY, NULL);
	if(!pdrPtr) {
		printf("PDR_COUNTRY_LIST not found in PDA\n");
		exit(-1);
	}
	
	p = &(pdrPtr->PdrData[0]);
	printf("Regul.Domain  : 0x%02X  ", p[0]);
	printf("Country string: [%c%c%c]\n", p[1], p[2], p[3]);
	
}

static void print_MAC_Address() {
	
	struct  Pdr_Struct *pdrPtr;
	u_char* p;
	
	pdrPtr = pda_Find_Pdr(PDR_MAC_ADDRESS, NULL);
	if(!pdrPtr) {
		printf("PDR_MAC_ADDRESS not found in PDA\n");
		exit(-1);
	}
	
	p = &(pdrPtr->PdrData[0]);
	printf("MAC : %02X:%02X:%02X:%02X:%02X:%02X\n", p[0], p[1], p[2], p[3], p[4], p[5]);
}

static void print_Antenna_Gain() {
	struct Pdr_Struct *pdrPtr;
	u_char* p;
	int i;
	
	pdrPtr = pda_Find_Pdr(PDR_ANTENNA_GAIN, NULL);
	if(!pdrPtr) {
		printf("PDR_ANTENNA_GAIN not found in PDA\n");
		exit(1);
	}
	
	for(i=0; i < pdrPtr->PdrLength - 1; i++) {
		printf("Antenna %i\n", i + 1);
		printf("Gain 5   Ghz: %03u / 0.25 dBi\n", pdrPtr->PdrData[i * 2]);
		printf("Gain 2.5 Ghz: %03u / 0.25 dBi\n", pdrPtr->PdrData[(i *2) + 1]);
	}			
}

void print_OutputPowerLimits() {
	struct Pdr_Struct *pdrPtr;
	u_char* p;
	int i;
	int no_channels;
	
	pdrPtr = pda_Find_Pdr(PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS, NULL);
	if(!pdrPtr) {
		printf("PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS not found in PDA\n");
		exit(1);
	}
	
	p = pdrPtr->PdrData;
	
	
	printf("Calibration method revision: %02X\n", *(p++));
	no_channels = *p;
	printf("n_channels                 : %i\n", no_channels);
	p++;
	
	while(no_channels > 0) {
		printf("Frequency: %04i / MHz\n", *p + ((*(p+1)) << 8));
		printf("Limits:\n");
		p = p + 2;
		printf("OFDM BPSK   : %03i / 0.25 dBm\n", (char) *(p++));
		printf("OFDM QPSK   : %03i / 0.25 dBm\n", (char) *(p++));
		printf("OFDM 16-QAM : %03i / 0.25 dBm\n", (char) *(p++));
		printf("OFDM 64-QAM : %03i / 0.25 dBm\n", (char) *(p++));
		p++;
		printf("Barker / CCK: %03i / 0.25 dBm\n", (char) *(p++));		
		printf("\n");
		no_channels--;
	}
	
	
	

}

void print_CurveData() {
	struct Pdr_Struct *pdrPtr;
	u_char* p, p1;
	int i;
	int no_channels, no_points;
	
	pdrPtr = pda_Find_Pdr(PDR_PRISM_PA_CAL_CURVE_DATA, NULL);
	if(!pdrPtr) {
		printf("PDR_PRISM_PA_CAL_CURVE_DATA not found in PDA\n");
		exit(1);
	}
	
	p = pdrPtr->PdrData;
	
	
	printf("Calibration method revision: %02X\n", *(p++));
	no_channels = *p;
	printf("n_channels                 : %i\n", no_channels);
	p++;
	no_points = *p;
	printf("n_points per channel       : %i\n", no_points);
	p++;
	printf("PA detector interpretation : %X\n", *(p++));

	while(no_channels > 0) {
		p1 = p;
		printf("Frequency: %04i / MHz\n", *p + ((*(p+1)) << 8));
		p = p + 2;
		for(i = 1 ;  i <= no_points; i++) {
			signed char det;
			int deti;
			printf("  Point %i\n", i);
			printf("    RF Power            : %i / 0.25 dBm\n", *((unsigned char*)(p++)));
			det = *(p++);
			deti = det;
			printf("    PA detector         : %i\n",  deti);
			printf("    Power Control Value : %i\n", (char) *(p++));
		}
		/* TODO padding if n_points not even */
		no_channels--;
		
	}
	
	
	

}


int main(int argc, char* argv[]) {

	pda_Cache( "/dev/mtd0" );
	
	if(argc == 1) {
		pda_Print();
		exit(0);
	}
	
	if(strcmp(argv[1], "PDR_DEFAULT_COUNTRY") == 0) {
		print_Default_Country();
	}
	else if (strcmp(argv[1], "PDR_COUNTRY_LIST") == 0) {
		print_Country_List();
	}
	else if (strcmp(argv[1], "PDR_MAC_ADDRESS") == 0) {
		print_MAC_Address();
	}
	else if (strcmp(argv[1], "PDR_ANTENNA_GAIN") == 0) {
		print_Antenna_Gain();
	}
	else if (strcmp(argv[1], "PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS") == 0) {
		print_OutputPowerLimits();
	}
	else if (strcmp(argv[1], "PDR_PRISM_PA_CAL_CURVE_DATA") == 0) {
		print_CurveData();
	}	
	else if (strcmp(argv[1], "PDR_INTERFACE_LIST") == 0) {
		print_Interface_List();
	}

	else {
		printf("Getting detailed information for %s not (yet) supported.\n", argv[1]);
	}

}
