/*  $Header$
 *  
 *  sdhcp - simple dhcp server
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Heavily based on the work of the ISC.
 */

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <syslog.h>

#include "debug.h"
#include "sdhcpd.h"

#if DEBUG > DBL_NONE
  static void do_percentm (char *obuf, char *ibuf);

  static char mbuf [1024];
  static char fbuf [1024];
#endif /* DEBUG > 0 */

/* Log an error message, then exit... */
void error (char *fmt, ...)
{
#if DEBUG >= DBL_ERROR
    va_list list;

    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_ERR, mbuf);
    syslog (LOG_CRIT, "exiting.");
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_ERROR */
    cleanup();

    exit (1);
}

/* Log an warning message */
void warn (char *fmt, ...)
{
#if DEBUG >= DBL_WARNING
    va_list list;

    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_WARNING, mbuf);
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_WARNING */
}

/* Log an notice message */
void note (char *fmt, ...)
{
#if DEBUG >= DBL_NOTICE
    va_list list;

    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_NOTICE, mbuf);
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_NOTICE */
}

/* Log an debug message */
void debug (char *fmt, ...)
{
#if DEBUG >= DBL_DEBUG
    va_list list;

    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_DEBUG, mbuf);
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_DEBUG */
}

/* Find %m in the input string and substitute an error message string. */
#if DEBUG > DBL_NONE
static void do_percentm (char *obuf, char *ibuf)
{
    char *s = ibuf;
    char *p = obuf;
    int infmt = 0;
    char *m;

    while (*s)
    {
	if (infmt)
	{
	    if (*s == 'm')
	    {
		m = strerror (errno);
		if (!m)
		    m = "<unknown error>";
		strcpy (p - 1, m);
		p += strlen (p);
		++s;
	    }
	    else
		*p++ = *s++;
	    infmt = 0;
	}
	else
	{
	    if (*s == '%')
		infmt = 1;
	    *p++ = *s++;
	}
    }
    *p = 0;
}
#endif /* DEBUG > 0 */

char
*print_hw_addr( const unsigned char *data, int len)
{
    static char habuf [49];
    char *s;
    int i;

#if DEBUG > DBL_NONE
    if (len == 0) {
	strcpy (habuf, "<null>");
    } else {
	s = habuf;
	for (i = 0; i < len; i++) {
	    sprintf (s, "%02x", data [i]);
	    s += strlen (s);
	    *s++ = ':';
	}
	*--s = 0;
    }
#endif /* DEBUG > DBL_NONE */
    return habuf;
}

char
*print_client_state( const client_state_t state )
{
#if DEBUG >= DBL_DEBUG
    switch( state ) {
    case cs_uninitialized_e:
	return "UNINITIALIZED";
    case cs_init_e:
	return "INIT";
    case cs_selecting_e:
	return "SELECTING";
    case cs_requesting_e:
	return "REQUESTING";
    case cs_init_reboot_e:
	return "INIT-REBOOT";
    case cs_rebooting_e:
	return "REBOOTING";
    case cs_bound_e:
	return "BOUND";
    case cs_renewing_e:
	return "RENEWING";
    case cs_rebinding_e:
	return "REBINDING";
    default:
	return "UNKNOWN";
    }
#endif /* DEBUG > DBL_NONE */
    return NULL;
}

/*
 *
 */
void
print_leases()
{
#if DEBUG >= DBL_DEBUG
    struct lease_t *tmp = NULL;

    tmp = pending_lease_list;
    debug("PENDING LEASE LIST");
    if(!tmp){
	debug( "No leases present" );
    }else{
	do {
	    debug( "lease = %08X", tmp );
	    debug( "lease->next = %08X", tmp->next );
	    debug( "lease->ipaddr = %s", inet_ntoa( tmp->ip_addr ));
	    debug( "lease->chaddr = %s",
		   print_hw_addr( tmp->hw_addr.u.chaddr, tmp->hw_addr.len));
	    debug( "lease->state = %s", print_client_state( tmp->state ) );
	} while( tmp = tmp->next );
    }

    tmp = own_lease_list;
    debug("OWN LEASE LIST");
    if(!tmp){
	debug( "No leases present" );
    }else{
	do {
	    debug( "lease = %08X", tmp );
	    debug( "lease->next = %08X", tmp->next );
	    debug( "lease->ipaddr = %s", inet_ntoa( tmp->ip_addr ));
	    debug( "lease->chaddr = %s",
		   print_hw_addr( tmp->hw_addr.u.chaddr, tmp->hw_addr.len));
	    debug( "lease->state = %s", print_client_state( tmp->state ) );
	} while( tmp = tmp->next );
    }
    tmp = unavailable_lease_list;
    debug("UNAVAILABLE LEASE LIST");
    if(!tmp){
	debug( "No leases present" );
    }else{
	do {
	    debug( "lease = %08X", tmp );
	    debug( "lease->next = %08X", tmp->next );
	    debug( "lease->ipaddr = %s", inet_ntoa( tmp->ip_addr ));
	    debug( "lease->chaddr = %s",
		   print_hw_addr( tmp->hw_addr.u.chaddr, tmp->hw_addr.len));
	    debug( "lease->state = %s", print_client_state( tmp->state ) );
	} while( tmp = tmp->next );
    }
    tmp = reserved_lease_list;
    debug("RESERVED LEASE LIST");
    if(!tmp){
	debug( "No leases present" );
    }else{
	do {
	    debug( "lease = %08X", tmp );
	    debug( "lease->next = %08X", tmp->next );
	    debug( "lease->ipaddr = %s", inet_ntoa( tmp->ip_addr ));
	    debug( "lease->chaddr = %s",
		   print_hw_addr( tmp->hw_addr.u.chaddr, tmp->hw_addr.len));
	    debug( "lease->state = %s", print_client_state( tmp->state ) );
	} while( tmp = tmp->next );
    }
#endif /* DEBUG >= DBL_DEBUG */
}


