/*  $Header$
 *  
 *  sdhcp - simple dhcp server
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef _DHCPD_H
#define _DHCPD_H

/* dhcpd.h */
#include <sys/types.h>
#include <net/if.h>		/* for IFHWADDRLEN */
#include <netinet/in.h>		/* for in_addr */

#include "dhcp.h"

/* where to find the DHCP server files */
#define DHCPD_CONF_FILE         "/etc/config/dhcpd.conf"
#define DHCPD_LEASES_FILE       "/etc/config/dhcpd.leases"
#define RESOLV_FILENAME		"/etc/resolv.conf"

/* Token definitions of the configuration file */
#define T_SUBNET	"SUBNET"
#define T_NETMASK	"NETMASK"
#define T_ROUTER	"ROUTER"

/*
In order for dhcpd to work correctly with picky DHCP clients (e.g.,
Windows 95), it must be able to send packets with an IP destination
address of 255.255.255.255.  Unfortunately, Linux changes an IP
destination of 255.255.255.255 into the local subnet broadcast address
(here, that's 192.5.5.223).  This isn't a problem on Linux 2.2 and
later kernels, since we completely bypass the Linux IP stack, but on
old versions of Linux 2.1 and all versions of Linux prior to 2.1, it
is a problem
*/
//#define ADD_ROUTE
#define HAVE_SO_BINDTODEVICE

#define PING_TIMEOUT		1	/* wait max. 1 second */
#define MAX_DNS_ENTRIES		3

  /* type length value structure */
struct tlv_t {
    unsigned char type;
    unsigned char len;
    unsigned char *val;
};

  /* hardware information structure */
struct hardware_t {
    unsigned char type;
    unsigned char len;
    union{
	u_int8_t haddr [IFHWADDRLEN];	/* MAC address */
	u_int8_t chaddr [16];		/* client hardware address */
    } u;
};

  /* Interface information structure */
struct if_info_t {
    int main_fd;			/* fd to primary address */
    int alias_fd;			/* fd to aliased address */
    char name[IFNAMSIZ];		/* Interface name */
    struct hardware_t hw_addr;		/* Its physical address. */
    struct in_addr ip_addr;		/* interface address. */
    //struct in_addr alias_ip_addr;	/* Alias address */
};

typedef enum{
    cs_uninitialized_e = 0,		/* state not initialized */
    cs_init_e,
    cs_selecting_e,
    cs_requesting_e,
    cs_init_reboot_e,
    cs_rebooting_e,
    cs_bound_e,
    cs_renewing_e,
    cs_rebinding_e
} client_state_t;

  /* lease information structure */
struct lease_t {
    struct lease_t *next;		/* next lease */
    struct in_addr ip_addr;		/* IP address of lease */
    struct hardware_t hw_addr;		/* hardware address of lease */
    client_state_t state;		/* state of associated client */
};

  /* parsed packet structured */
struct packet_t {
    struct dhcp_packet * raw;		/* link to raw packet */
    int length;				/* raw packet length */
    unsigned char msg_type;		/* dhcp message type */
};

/*****************************************************************************
 * Global Variables
 *****************************************************************************/
extern char *conf_file;
extern char *lease_file;
extern char *interface_name;
extern unsigned char interface_hwaddr[6];
extern struct if_info_t server_ifinfo;
extern int listen_port;
extern int debug_level;
extern struct in_addr subnet;
extern struct in_addr netmask;
extern struct in_addr router;
extern struct lease_t *own_lease_list;
extern struct lease_t *pending_lease_list;
extern struct lease_t *reserved_lease_list;
extern struct lease_t *unavailable_lease_list;
extern struct in_addr dns[MAX_DNS_ENTRIES];
extern int dns_entries;

/*****************************************************************************
 * Function Prototypes
 *****************************************************************************/
/* from ping.c */
extern int ping_check( struct in_addr addr, unsigned long timeout );

/* from socket.c */
extern int open_interface(struct if_info_t *if_info);
extern int open_icmp_socket();

/* from leases.c */
extern struct lease_t *find_offered_lease_by_hwaddr( struct hardware_t * );
extern struct lease_t *find_lease_by_hwaddr( struct lease_t *lease_list,
					     struct hardware_t *hw_addr );
extern struct lease_t *find_lease_by_ipaddr( struct lease_t *lease_list,
					     struct in_addr ipaddr );
extern struct lease_t *add_pending_lease( struct packet_t *packet,
					  struct in_addr address );
extern void remove_lease( struct lease_t **lease_list, struct lease_t *lease );
extern int move_lease( struct lease_t *lease,
		       struct lease_t **dst_list,
		       struct lease_t **src_list );
extern void remove_list( struct lease_t **lease_list );
extern unsigned char is_known_ipaddr( struct in_addr address );
extern int get_free_ip_address( struct in_addr *address );
extern unsigned char match_hw_addr( struct hardware_t *addr1,
				    struct hardware_t *addr2 );

/* from dhcp.c */
extern void dhcp( struct packet_t *packet );
extern int get_dhcp_option( struct dhcp_packet *packet,
			    int packet_length,
			    unsigned char option,
			    unsigned char *option_store,
			    int *store_size );

/* from shdpcd.c */
void cleanup();

#endif /* !_DHCPD_H */

