//#define DEBUG(x,y) fprintf(stderr, "%s:%d ", __FILE__, __LINE__) ; fprintf(stderr, x, y)
#define DEBUG(x,y) 

/* This should always be included first before anything else */
#include <config.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>

#ifdef IN_UCD_SNMP_SOURCE
/* If we're compiling this file inside the ucd-snmp source tree */

/* minimal include directives */
#include "mibincl.h"
#include "util_funcs.h" 

#else				/* !IN_UCD_SNMP_SOURCE */

#include <ucd-snmp/ucd-snmp-config.h>
#include <ucd-snmp/ucd-snmp-includes.h>
#include <ucd-snmp/ucd-snmp-agent-includes.h>

#endif				/* !IN_UCD_SNMP_SOURCE */

#include "datastore_if.h"
#include "datastore.h"
#include "snmp_mib.h"
/*
 * datastore_variables_oid:
 *   this is the top level oid that we want to register under.  This
 *   is essentially a prefix, with the suffix appearing in the
 *   variable below.
 */

/**********************************************************
 * Define the MIB-parts that are handled by the datastore *
 **********************************************************/
 
/* ieee802dot11 MIB */
oid datastore_variables_oid_1_2_840[] = { 1, 2, 840 }; 

/* MIB - 2 */
oid datastore_variables_oid_1_3_6_1_2[] = { 1, 3, 6, 1, 2 }; 

 /* Intersil Private MIB */ 
oid datastore_variables_oid_1_3_6_1_4_1[] = { 1, 3, 6, 1, 4, 1 };


	/*********************
	 *
	 *  Kernel & interface information,
	 *   and internal forward declarations
	 *
	 *********************/

/* ieee802dot11 MIB */
struct variable2 datastore_variables_1_2_840[] = {
/*  magic number        , variable type , ro/rw , callback fn  , L, oidsuffix */
	{1, ASN_INTEGER, RWRITE, var_datastore, 1, {10036}},
};

/* MIB - 2 */	
struct variable2 datastore_variables_1_3_6_1_2[] = {
	{1, ASN_INTEGER, RWRITE, var_datastore, 0, {1}},
};

/* Intersil Private MIB */ 		
struct variable2 datastore_variables_1_3_6_1_4_1[] = {
	{1, ASN_INTEGER, RWRITE, var_datastore, 0, {937}},
};


/*
 * init_datastore():
 *   Initialization routine.  This is called when the agent starts up.
 *   At a minimum, registration of your variables should take place here.
 */
void
init_datastore (void)
{
	/* register ourselves with the agent to handle our mib tree */
	REGISTER_MIB ("ieee802dot11", datastore_variables_1_2_840, variable2,
		      datastore_variables_oid_1_2_840);
	REGISTER_MIB ("MIB-II", datastore_variables_1_3_6_1_2, variable2,
		      datastore_variables_oid_1_3_6_1_2);
	REGISTER_MIB ("Intersil", datastore_variables_1_3_6_1_4_1, variable2,
		      datastore_variables_oid_1_3_6_1_4_1);
}

static void
oid2string (oid * name, size_t length, char *oid_str, size_t oid_str_size)
{

	size_t i, part_length;
	char* max_address = oid_str + oid_str_size - 1;
	
	for (i = 0; (i < length) && (oid_str < max_address) ; i++) {
		part_length = snprintf (oid_str, 
				        (size_t) (max_address - oid_str), 
				        ".%lu", 
				        name[i]);
		oid_str = oid_str + part_length;
	}
}

static void
string2oid (char *oid_str, oid * name, size_t * length)
{

	char *tok;

	*length = 0;
	tok = strtok (oid_str, ".");

	if (tok) {
		name[*length] = atol (tok);
		(*length)++;

	}

	while ((tok = strtok (NULL, "."))) {
		name[*length] = atol (tok);
		(*length)++;
	}

}

static void
ipstr2iplong (unsigned char *iplong, char *ipstr)
{
	char *part[4];
	int parts;
	int i;

	parts = split (ipstr, '.', part, 4);

	if (parts != 4) {
		fprintf (stderr, "Error parsing IP address (%i)\n", parts);
		return;
	}

	for (i = 0; i < 4; i++) {
		iplong[i] = atoi (part[i]);
	}

}


static void write_stats() {
	FILE* fp;
	static time_t last_write = 0;
	if ((time(NULL) - 3) >  last_write) {
		last_write = time(NULL);
		
		fp = fopen("/var/log/snmp_stats", "w");
		if(!fp) {
			fprintf(stderr, "Error, couldn't open /var/log/snmp_stats for writing\n");
			return;
		}
		
		fprintf(fp,"SNMPINPKTS %i\n", snmp_get_statistic(SNMPINPKTS + STAT_SNMP_STATS_START - 1));
		fprintf(fp,"SNMPINBADVERSIONS %i\n", snmp_get_statistic(SNMPINBADVERSIONS + STAT_SNMP_STATS_START - 1));
		//fprintf(fp,"SNMPINBADCOMMUNITYNAMES %i\n", snmp_get_statistic(SNMPINBADCOMMUNITYNAMES + STAT_SNMP_STATS_START - 1));
		//fprintf(fp,"SNMPINBADCOMMUNITYUSES %i\n", snmp_get_statistic(SNMPINBADCOMMUNITYUSES + STAT_SNMP_STATS_START - 1));
		fprintf(fp,"SNMPINASNPARSEERRORS %i\n", snmp_get_statistic(SNMPINASNPARSEERRORS + STAT_SNMP_STATS_START - 1));				
		fprintf(fp,"SNMPSILENTDROPS %i\n", snmp_get_statistic(SNMPSILENTDROPS + STAT_SNMP_STATS_START - 1));
		fprintf(fp,"SNMPPROXYDROPS %i\n", snmp_get_statistic( SNMPPROXYDROPS  + STAT_SNMP_STATS_START - 1));				
		fclose(fp);
	}
}

/*
 * var_datastore():
 *   This function is called every time the agent gets a request for
 *   a scalar variable that might be found within your mib section
 *   registered above.  It is up to you to do the right thing and
 *   return the correct value.
 *     You should also correct the value of "var_len" if necessary.
 *
 *   Please see the documentation for more information about writing
 *   module extensions, and check out the examples in the examples
 *   and mibII directories.
 */
unsigned char *
var_datastore (struct variable *vp,
	       oid * name,
	       size_t * length,
	       int exact, size_t * var_len, WriteMethod ** write_method)
{

	/* variables we may use later */
	static long long_ret;
	static struct counter64 c64;
	static unsigned char oid_str[MAX_OID_LENGTH];
	static unsigned char data[MAX_DATA_LENGTH];
	int data_len;
	static unsigned char snmp_data[MAX_DATA_LENGTH];
	static oid oid_data[MAX_OID_LENGTH];
	int type;
	int ret_val;
	u_char* community_p;
	size_t community_len;
	unsigned long long ull;

	if(vp->asp->pdu->community_len == 0) {
		/* we don't support empty communities */
		return NULL;
	}
	
	/* set the generic write callback */
	*write_method = write_datastore;
	
	write_stats();

	if (vp->asp->rw == WRITE) {
		/* 
		 * SNMPSET 
		 * 
		 * This is a SNMPSET command. If this is the beginning of 
		 * the creation of a new Row in a dynamic table, the 
		 * GET of the OID will fail, because it simply does not
		 * exist yet. The datastore will report NoSuchOid. 
		 * Because we don't know anything about the MIB-structure
		 * (We are just a dump proxy), we return something
		 * valid here. The write-function will handle all the 
		 * type-checking, OID validition etc.
		 */
		*var_len = 4;
		long_ret = 0;
		return (unsigned char *) &long_ret;
	}

	oid2string (name, *length, oid_str, sizeof(oid_str) - 1);
	
	DEBUG("%s\n", oid_str);	

	community_p = vp->asp->pdu->community;
	community_len = vp->asp->pdu->community_len;
	
	if (exact) {
		ret_val = getoid (oid_str, community_p, community_len, data, &data_len, &type);
	} else {
		ret_val = getnextoid(oid_str, community_p, community_len, data, &data_len, &type);
	}

	DEBUG("retval = %i\n", ret_val);
	
	switch (ret_val) {
	case R_NOERR:
		break;
	case R_NOSUCHOID:
		/* OID does not exist */
		return NULL;
		break;
	case R_ENDOFTREE:
		/* END OF TREE */
		return NULL;
		break;
	case R_NOACCESS:
		/* NO ACCESS */
		return NULL;
	default:
		fprintf (stderr, "Unexpected returnvalue from datastore: %i\n",
			 ret_val);
		return NULL;
		break;
	}

	/* replace requested oid with returned oid (getnext) */
	if (!exact) {
		string2oid (oid_str, name, length);
	}

	switch (type) {

	case T_UNKNOWN:
		fprintf (stderr, "Unknown type\n");
		return NULL;
		break;

	case T_INTEGER:
		*var_len = 4;
		long_ret = atol (data);
		vp->type = ASN_INTEGER;
		return (unsigned char *) &long_ret;
		break;
		
	case T_ROWSTAT:
		*var_len = 4;
		long_ret = atol (data);
			
		/* convert internal (illegal for SNMP) states 
		 * datastore doesn't convert these for us, because
		 * it has to remember the two different CREATE states
		 */
		switch(long_ret) {
			case SNMP_ROW_CREATEANDGO:
      			case SNMP_ROW_CREATEANDWAIT:
				long_ret = SNMP_ROW_NOTREADY;
				break;
			default:
				break;
		}
		vp->type = ASN_INTEGER;
		return (unsigned char *) &long_ret;
		break;
	
	case T_UNSIGNED:
		*var_len = 4;
		long_ret = atol(data);
		vp->type = ASN_UNSIGNED;
		return (unsigned char*) &long_ret;
		break;

	case T_BITS:		
	case T_OCTETSTR:
		*var_len = data_len;
		/* TODO buff-overflow check */
		memcpy(snmp_data, data, data_len);
		vp->type = ASN_OCTET_STR;
		return (unsigned char *) snmp_data;
	
	case T_IPADDR:
		*var_len = 4;
		switch(strlen(data)) {
			case 0:
				memset(snmp_data, 0, sizeof(char) * 4);
				break;
			default:
				ipstr2iplong (snmp_data, data);
				break;
		}
		
		vp->type = ASN_IPADDRESS;
		return (unsigned char *) snmp_data;

	case T_COUNTER:
		*var_len = 4;
		long_ret = atol(data);
		vp->type = ASN_COUNTER;
		return (unsigned char*) &long_ret;
		break;
		
	case T_COUNTER64:
		*var_len = sizeof(struct counter64);
		ull = atoll(data);
		c64.high = ull >> 32;
		c64.low = 0x00000000FFFFFFFF & ull;
		vp->type = ASN_COUNTER64;
		return (unsigned char*) &c64;
		break;
		
	case T_GAUGE:
		*var_len = 4;
		long_ret = atol(data);
		vp->type = ASN_GAUGE;
		return (unsigned char*) &long_ret;
		break;
		
	case T_OBJECTID:
		
		string2oid (data, oid_data, var_len);
		vp->type = ASN_OBJECT_ID;
		
		/* length is length in octets */
		*var_len = (*var_len) * 4;

		return (unsigned char *) oid_data;
		break;
		
	case T_OPAQUE:
		*var_len = data_len;
		/* TODO buff-overflow check */
		memcpy(snmp_data, data, data_len);
		vp->type = ASN_OPAQUE;
		return (unsigned char *) snmp_data;
		break;
						
	case T_TICKS:
		*var_len = 4;
		if (data) {
			long_ret = atol(data);
		} else {
			long_ret = 0;
		}
		vp->type = ASN_TIMETICKS;
		return (unsigned char*) &long_ret;
		break;
		
	case T_NOTIFY:
		fprintf (stderr, "NOTIFY not implemented !\n");
		return NULL;
		break;
		
	default:
		fprintf (stderr, "datastore type (%d) not yet implemented\n", type);
		return NULL;
		break;
	}

}

int
write_datastore (int action,
		 u_char * var_val,
		 u_char var_val_type,
		 size_t var_val_len,
		 u_char * statP, oid * name, size_t name_len,
		 struct agent_snmp_session* asp)
{

	static int pdu_counter = 0; /* transaction counter */
	static int set_counter = 0; /* counts the number of SETS */
	
	/* next two are static to save stackspace */
	static unsigned char data[SPRINT_MAX_LEN + 1];
	int data_len;
	static unsigned char oid_str[SPRINT_MAX_LEN + 1] ;
	u_char* community_p;
	size_t community_len;
	int i;
	size_t length;
	unsigned int ui;
	unsigned long long ull;
	struct counter64 *c64;

	int type, ret_val;
	DEBUG ("action=%i\n", action);
	DEBUG ("var_val_type=%i\n", var_val_type);
	DEBUG ("var_val_len=%li\n", var_val_len);
	DEBUG ("action = %i\n", action);
	
	switch (action) {
	case RESERVE1:
		if(pdu_counter == 0) {
			/* this is the start of a new PDU */
			if( begin_transaction() == -1)
			{
				return SNMP_ERR_GENERR;
			}
		}
		pdu_counter++;
		break;

	case RESERVE2:

		break;

	case FREE:

		break;

	case ACTION:
	
		switch (var_val_type) {
		case ASN_INTEGER:
			DEBUG ("%s\n", "ASN_INTEGER");
			
			/* check length */
			if(var_val_len != 4) {
				return SNMP_ERR_BADVALUE;
			}
			
			i = *((long*) var_val);
			snprintf (data, sizeof(data), "%i", i);
			data_len = strnlen(data, sizeof(data));
			type = T_INTEGER;
			break;
		case ASN_UNSIGNED:
			DEBUG ("%s\n", "ASN_UNSIGNED");
			
			/* check length */
			if(var_val_len != 4) {
				return SNMP_ERR_BADVALUE;
			}
			
			ui = *((long*) var_val);
			snprintf (data, sizeof(data), "%u", ui);
			data_len= strnlen(data, sizeof(data));
			type = T_UNSIGNED;
			break;
		case ASN_IPADDRESS:
			DEBUG ("%s\n", "ASN_IPADDRESS");
			
			/* check length */
			if(var_val_len != 4) {
				return SNMP_ERR_BADVALUE;
			}
			
			snprintf (data, sizeof(data), "%d.%d.%d.%d",
				 var_val[0], var_val[1], var_val[2],
				 var_val[3]);
			data_len = strnlen(data, sizeof(data));
			type = T_IPADDR;
			break;
		case ASN_OCTET_STR:
			DEBUG ("%s\n", "ASN_OCTET_STRING");
			
			/* calculate max length. Use sizeof(data) -1 
			 * to keep space left for a \0 */
			length = SNMP_MIN(var_val_len, sizeof(data) - 1);
			
			memcpy(data, var_val, length);
			
			/* Make 'data' NULL-terminated */
			data[length] = '\0';
			data_len=length;
			
			type = T_OCTETSTR;
			break;
			
		case ASN_OPAQUE:
			DEBUG ("%s\n", "ASN_OPAQUE");
			
			/* calculate max length. Use sizeof(data) -1 
			 * to keep space left for a \0 */
			length = SNMP_MIN(var_val_len, sizeof(data) - 1);
			
			memcpy(data, var_val, length);
			
			/* Make 'data' NULL-terminated */
			data[length] = '\0';
			data_len=length;
			
			type = T_OPAQUE;
			break;

		case ASN_COUNTER:
			DEBUG ("%s\n", "ASN_COUNTER");
			
			/* check length */
			if(var_val_len != 4) {
				return SNMP_ERR_BADVALUE;
			}
			
			ui = *((long*) var_val);
			sprintf(data, "%u", ui);
			data_len = strlen(data);
			type = T_COUNTER;
			break;

		case ASN_COUNTER64:
			DEBUG ("%s\n", "ASN_COUNTER64");
			
			/* check length */
			if(var_val_len != 8) {
				return SNMP_ERR_BADVALUE;
			}
			
			c64 = (struct counter64 *) var_val;
			
			ull = (long long) c64->high;
			ull = ull << 32;
			ull = ull | c64->low;
			snprintf(data, sizeof(data), "%llu", ull);
			data_len = strnlen(data, sizeof(data));
			type = T_COUNTER64;
			break;
/*			
		case ASN_GAUGE:
			DEBUG ("%s\n", "ASN_GAUGE");
			
			ui = *((long*) var_val);
			sprintf(data, "%u", ui);
			data_len = strlen(data);
			type = T_GAUGE;
			break;
*/		
		case ASN_OBJECT_ID:
			DEBUG ("%s\n", "ASN_OBJECT_ID");
			
			/* var_val_len contains the lenght in octets,
			 * but the actual OID is a array of integers
			 * the length of the OID is therefore 
			 * var_val_len / 4 
			 */
			oid2string ((oid*) var_val, var_val_len / 4, data, sizeof(data) - 1);
			data_len = strlen(data);
			type = T_OBJECTID;
			break;
			
	
		case ASN_TIMETICKS:
			DEBUG ("%s\n", "ASN_TIMETICKS");

			/* check length */
			if(var_val_len != 4) {
				return SNMP_ERR_BADVALUE;
			}
			
			ui = *((long*) var_val);
			sprintf(data, "%u", ui);
			data_len = strlen(data);
			type = T_TICKS;
			break;
/*
		case ASN_NOTIFY:
			fprintf (stderr, "NOTIFY not implemented !\n");
			return SNMP_ERR_WRONGTYPE;
			break;
*/					
		default:
			fprintf (stderr, "ASN type (%d) not yet implemented\n",
				 var_val_type);
			return SNMP_ERR_WRONGTYPE;
			break;

		}
	
		oid2string (name, name_len, oid_str, sizeof(oid_str) - 1);
		
		DEBUG ("data_len=%i\n", data_len);
		DEBUG ("data=%s\n", data);
		DEBUG ("oid_str=%s\n", oid_str);
		DEBUG ("asp->pdu->community_len=%i\n", asp->pdu->community_len);
		
		community_p = asp->pdu->community;
		community_len = asp->pdu->community_len;
	
		ret_val = setoid(oid_str, community_p, community_len, data, &data_len, &type);
		set_counter++;

		DEBUG ("ret_val from SETOID = %i\n", ret_val);

		switch(ret_val) {
		case R_NOERR:	
			break;
		case R_OUTOFRANGE:
			return SNMP_ERR_WRONGVALUE;
			break;
		case R_NOACCESS:
			return SNMP_ERR_NOTWRITABLE;
			break;
		case R_WRONGTYPE:
			return SNMP_ERR_WRONGTYPE;
			break;
		case R_NOSUCHOID:
			return SNMP_ERR_NOCREATION;
			break;
		default:
			return SNMP_ERR_GENERR;
			break;
	  	}

		/* check if this is the last set in this pdu */
		if (set_counter == pdu_counter) {

			/* try to commit sets */
			if (commit_transaction() != R_NOERR ) {
				return SNMP_ERR_INCONSISTENTVALUE;
			}
			
		}
		
		break;

	case UNDO:
		/* check pdu counter to prevent double rollback */
		if (pdu_counter != 0) {
			set_counter = 0;
			pdu_counter = 0;
			rollback_transaction();
		}
		
		break;

	case COMMIT:

		set_counter = 0;
		pdu_counter = 0;

		break;
	}

	return SNMP_ERR_NOERROR;
}
