/*
 *  System MIB group implementation - system.c
 *
 */

#include <config.h>
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if HAVE_WINSOCK_H
#include <winsock.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#include <ctype.h>
#if HAVE_UTSNAME_H
#include <utsname.h>
#else
#if HAVE_SYS_UTSNAME_H
#include <sys/utsname.h>
#endif
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#if HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#include <pda.h>

#include "mibincl.h"
#include "system_mib.h"
#include "../struct.h"
#include "../util_funcs.h"
#include "read_config.h"
#include "agent_read_config.h"
#include "system.h"
#include "sysORTable.h"


	/*********************
	 *
	 *  Kernel & interface information,
	 *   and internal forward declarations
	 *
	 *********************/

#define SYS_STRING_LEN	256
char version_descr[ SYS_STRING_LEN ] = VERS_DESC;
char sysContact[    SYS_STRING_LEN ] = SYS_CONTACT;
char sysName[       SYS_STRING_LEN ] = SYS_NAME;
char sysLocation[   SYS_STRING_LEN ] = SYS_LOC;

char oldversion_descr[ SYS_STRING_LEN ];
char oldsysContact[    SYS_STRING_LEN ];
char oldsysName[       SYS_STRING_LEN ];
char oldsysLocation[   SYS_STRING_LEN ];

int sysServices=72;
int sysServicesConfiged=0;

/* SNMP allows more, but we only have 92 bytes for storing it in flash */
#define MAX_VERSION_ID_OID_LEN 20
oid version_id[MAX_VERSION_ID_OID_LEN];
int version_id_len;

extern struct timeval starttime;

WriteMethod writeSystem;
int header_system(struct variable *,oid *, size_t *, int, size_t *, WriteMethod **);

/* snmpd.conf config parsing */

void system_parse_config_sysnam(const char *token,
				char *cptr)
{
  char tmpbuf[1024];

  if (strlen(cptr) < sizeof(sysName)) {
    strcpy(sysName,cptr);
  } else {
    sprintf(tmpbuf, "sysname token too long (must be < %d):\n\t%s",
		 sizeof(sysName), cptr);
    config_perror(tmpbuf);
  }
}

void system_parse_config_sysloc(const char *token,
				char *cptr)
{
  char tmpbuf[1024];

  if (strlen(cptr) < sizeof(sysLocation)) {
    strcpy(sysLocation,cptr);
  } else {
    sprintf(tmpbuf, "syslocation token too long (must be < %d):\n\t%s",
		 sizeof(sysLocation), cptr);
    config_perror(tmpbuf);
  }
}

void system_parse_config_sysServices(const char *token, char *cptr)
{
  sysServices = atoi(cptr);
  sysServicesConfiged = 1;
}

void system_parse_config_syscon(const char *token,
				char *cptr)
{
  char tmpbuf[1024];

  if (strlen(cptr) < sizeof(sysContact)) {
    strcpy(sysContact,cptr);
  } else {
    sprintf(tmpbuf, "syscontact token too long (must be < %d):\n\t%s",
                 sizeof(sysContact), cptr);
    config_perror(tmpbuf);
  }
}


	/*********************
	 *
	 *  Initialisation & common implementation functions
	 *
	 *********************/

/* define the structure we're going to ask the agent to register our
   information at */
struct variable2 system_variables[] = {
    {VERSION_DESCR, ASN_OCTET_STR, RONLY, var_system, 1, {1}},
    {VERSIONID, ASN_OBJECT_ID, RONLY, var_system, 1, {2}},
    {UPTIME, ASN_TIMETICKS, RONLY, var_system, 1, {3}},
    {SYSCONTACT, ASN_OCTET_STR, RWRITE, var_system, 1, {4}},
    {SYSTEMNAME, ASN_OCTET_STR, RWRITE, var_system, 1, {5}},
    {SYSLOCATION, ASN_OCTET_STR, RWRITE, var_system, 1, {6}},
    {SYSSERVICES, ASN_INTEGER, RONLY, var_system, 1, {7}},
    {SYSORLASTCHANGE, ASN_TIMETICKS, RONLY, var_system, 1, {8}}
};
/* Define the OID pointer to the top of the mib tree that we're
   registering underneath */
oid system_variables_oid[] = { SNMP_OID_MIB2,1 };
oid system_module_oid[]    = { SNMP_OID_SNMPMODULES,1 };
int system_module_oid_len  = sizeof( system_module_oid ) / sizeof( oid );
int system_module_count    = 0;

int string_to_sysobjectid (unsigned char *input, unsigned long *output)
{
	unsigned int i=0;
	unsigned int j=0;
	unsigned char *p;

	output[0] = 1;
	output[1] = 3;
	output[2] = 6;
	output[3] = 1;
	output[4] = 4;
	output[5] = 1;
	j= 6;
	while ( input[i] != '\0' && i < 92)
	{
		if ( isspace(input[i]) )
		{
			input[i]='.';
		}
		i++;
	}
	if (i > 92)
	{
		fprintf(stderr, "string_to_sysobjectid: PDA part of SysObjectId to big\n");
		return -1;
	}
	i=0;
	for ( p = strtok(input, " .-:"); p; p = strtok(NULL, " .-:") )
	{
		output[j] = strtol(p, (char**)NULL, 10);
		/*fprintf(stderr, "string_to_long_octet: octetnr %d parsed to (longnr %d) %lu\n", i, j, output[j]);*/
		i++;j++;
		if (j >= MAX_VERSION_ID_OID_LEN)
		{
			fprintf(stderr, "string_to_sysobjectid: Too many longs in string, stoped parsing!\n");
			return (j);
		}
	}
	return(j);
}

void init_system_mib(void)
{
  int i=0;
#ifdef HAVE_UNAME
  struct utsname utsName;

  uname(&utsName);
  sprintf(version_descr, "%s %s %s %s %s", utsName.sysname, utsName.nodename,
          utsName.release, utsName.version, utsName.machine);
#else
#if HAVE_EXECV
  struct extensible extmp;

  /* set default values of system stuff */
  sprintf(extmp.command,"%s -a",UNAMEPROG);
  /* setup defaults */
  extmp.type = EXECPROC;
  extmp.next = NULL;
  exec_command(&extmp);
  strncpy(version_descr,extmp.output, sizeof(version_descr));
  version_descr[strlen(version_descr)-1] = 0; /* chomp new line */
#else
  strcpy(version_descr, "unknown" );
#endif
#endif

  /* Make sure the PDA is cached in RAM */
  pda_Cache( "/dev/mtd0" );
  /* Fill the len with amount of octets parsed */
  version_id_len = string_to_sysobjectid(GetHcSysObjId(), version_id);
  fprintf(stderr, "init_system_mib: Parsed SysObjectId to: ");
  while (i < version_id_len)
  {
	  fprintf(stderr, "%lu.", version_id[i++]);
  }
  fprintf(stderr, "\b\n");

#ifdef HAVE_GETHOSTNAME
  gethostname(sysName,sizeof(sysName));
  fprintf(stderr,"Hostname: by gethostname: %s\n", sysName);
#else
#ifdef HAVE_UNAME
  strncpy(sysName,utsName.nodename,sizeof(sysName));
  fprintf(stderr,"Hostname: by uname: %s\n", sysName);
#else
#ifdef HAVE_EXECV
  {
  struct extensible extmp;

  sprintf(extmp.command,"%s -n",UNAMEPROG);
  /* setup defaults */
  extmp.type = EXECPROC;
  extmp.next = NULL;
  exec_command(&extmp);
  strncpy(sysName,extmp.output, sizeof(sysName));
  sysName[strlen(sysName)-1] = 0; /* chomp new line */
  }
#else
  strcpy(sysName, "unknown" );
#endif /* HAVE_EXECV */
#endif /* HAVE_UNAME */
#endif /* HAVE_GETHOSTNAME */

  /* register ourselves with the agent to handle our mib tree */
  REGISTER_MIB("mibII/system", system_variables, variable2, \
               system_variables_oid);

  if ( ++system_module_count == 3 )
	REGISTER_SYSOR_ENTRY( system_module_oid,
		"The MIB module for SNMPv2 entities");

  /* register our config handlers */
  snmpd_register_config_handler("sysname", system_parse_config_sysnam,
				NULL,
				system_write_config_sysnam,
                                "system-name");
  snmpd_register_config_handler("syslocation", system_parse_config_sysloc,
				NULL,
				system_write_config_sysloc,
				"location");
  snmpd_register_config_handler("syscontact", system_parse_config_syscon,
				NULL,
				system_write_config_syscon,
                                "contact-name");
}

/*
  header_system(...
  Arguments:
  vp	  IN      - pointer to variable entry that points here
  name    IN/OUT  - IN/name requested, OUT/name found
  length  IN/OUT  - length of IN/OUT oid's
  exact   IN      - TRUE if an exact match was requested
  var_len OUT     - length of variable or 0 if function returned
  write_method

*/

int
header_system(struct variable *vp,
	      oid *name,
	      size_t *length,
	      int exact,
	      size_t *var_len,
	      WriteMethod **write_method)
{
#define SYSTEM_NAME_LENGTH	8
    oid newname[MAX_OID_LEN];
    int result;

    DEBUGMSGTL(("mibII/system", "var_system: "));
    DEBUGMSGOID(("mibII/system", name, *length));
    DEBUGMSG(("mibII/system"," %d\n", exact));

    memcpy((char *)newname, (char *)vp->name, vp->namelen * sizeof(oid));
    newname[SYSTEM_NAME_LENGTH] = 0;
    result = snmp_oid_compare(name, *length, newname, vp->namelen + 1);
    if ((exact && (result != 0)) || (!exact && (result >= 0)))
        return(MATCH_FAILED);
    memcpy( (char *)name,(char *)newname, (vp->namelen + 1) * sizeof(oid));
    *length = vp->namelen + 1;

    *write_method = 0;
    *var_len = sizeof(long);	/* default to 'long' results */
    return(MATCH_SUCCEEDED);
}

	/*********************
	 *
	 *  System specific implementation functions
	 *	(actually common!)
	 *
	 *********************/

#ifdef USING_MIBII_SYSORTABLE_MODULE
extern struct timeval sysOR_lastchange;
#endif

u_char	*
var_system(struct variable *vp,
	   oid *name,
	   size_t *length,
	   int exact,
	   size_t *var_len,
	   WriteMethod **write_method)
{

  struct timeval now, diff;

    if (header_system(vp, name, length, exact, var_len, write_method) == MATCH_FAILED )
	return NULL;

    switch (vp->magic){
        case VERSION_DESCR:
            *var_len = strlen(version_descr);
            *write_method = writeSystem;
            return (u_char *)version_descr;
        case VERSIONID:
            *var_len = version_id_len*sizeof(version_id[0]);
            return (u_char *)version_id;
        case UPTIME:
            gettimeofday(&now, NULL);
            now.tv_sec--;
            now.tv_usec += 1000000L;
            diff.tv_sec = now.tv_sec - starttime.tv_sec;
            diff.tv_usec = now.tv_usec - starttime.tv_usec;
            if (diff.tv_usec > 1000000L){
                diff.tv_usec -= 1000000L;
                diff.tv_sec++;
            }
            long_return = ((diff.tv_sec * 100) + (diff.tv_usec / 10000));
            return ((u_char *) &long_return);
        case SYSCONTACT:
            *var_len = strlen(sysContact);
            *write_method = writeSystem;
            return (u_char *)sysContact;
        case SYSTEMNAME:
            *var_len = strlen(sysName);
            *write_method = writeSystem;
            return (u_char *)sysName;
        case SYSLOCATION:
            *var_len = strlen(sysLocation);
            *write_method = writeSystem;
            return (u_char *)sysLocation;
        case SYSSERVICES:
#if NO_DUMMY_VALUES
            if (!sysServicesConfiged)
                return NULL;
#endif
            long_return = sysServices;
            return (u_char *)&long_return;

#ifdef USING_MIBII_SYSORTABLE_MODULE
        case SYSORLASTCHANGE:
              diff.tv_sec = sysOR_lastchange.tv_sec - 1 - starttime.tv_sec;
              diff.tv_usec =
                sysOR_lastchange.tv_usec + 1000000L - starttime.tv_usec;
              if (diff.tv_usec > 1000000L){
                diff.tv_usec -= 1000000L;
                diff.tv_sec++;
              }
              if ((diff.tv_sec * 100) + (diff.tv_usec / 10000) < 0)
                long_return = 0;
              else
                long_return = ((diff.tv_sec * 100) + (diff.tv_usec / 10000));
              return ((u_char *) &long_return);
#endif

	default:
	    DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_system\n", vp->magic));
    }
    return NULL;
}



int
writeSystem(int action,
	    u_char *var_val,
	    u_char var_val_type,
	    size_t var_val_len,
	    u_char *statP,
	    oid *name,
	    size_t name_len)
{
    u_char *cp;
	/*
	 * dougal - there used to be no static before these so set's wouldn't
	 * be undo-able. A bug that needs feeding back into the official SNMP
	 * src
	 */
    static char *buf = NULL, *oldbuf = NULL;
    int count;
    char needs_writing = FALSE;

    switch((char)name[7]){
      case VERSION_DESCR:
        buf    = version_descr;
        oldbuf = oldversion_descr;
        break;
      case SYSCONTACT:
        buf    = sysContact;
        oldbuf = oldsysContact;
	needs_writing = TRUE;
        break;
      case SYSTEMNAME:
        buf    = sysName;
        oldbuf = oldsysName;
	needs_writing = TRUE;
        break;
      case SYSLOCATION:
        buf    = sysLocation;
        oldbuf = oldsysLocation;
	needs_writing = TRUE;
        break;
      default:
	return SNMP_ERR_GENERR;		/* ??? */
    }

    switch ( action ) {
	case RESERVE1:		/* Check values for acceptability */
	    if (var_val_type != ASN_OCTET_STR){
                snmp_log(LOG_ERR, "not string\n");
		return SNMP_ERR_WRONGTYPE;
	    }
	    if (var_val_len > sizeof(version_descr)-1){
                snmp_log(LOG_ERR, "bad length\n");
		return SNMP_ERR_WRONGLENGTH;
	    }

	    for(cp = var_val, count = 0; count < (int)var_val_len; count++, cp++){
		if (!isprint(*cp)){
                    snmp_log(LOG_ERR, "not print %x\n", *cp);
		    return SNMP_ERR_WRONGVALUE;
		}
	    }
	    break;

	case RESERVE2:		/* Allocate memory and similar resources */

		/* Using static strings, so nothing needs to be done */
	    break;

	case ACTION:		/* Perform the SET action (if reversible) */
	    /*
	     * For purity's sake we ought to be looking for the config_token
	     * in the list of config_lines hanging under the config_file
	     * However there doesn't seem to be a way of mapping from the
	     * (char)name[7] defines above into the strings required.
	     * This is fine
	     */
	    if (needs_writing) {
		extern struct config_files *config_files;
		struct config_files *ctmp = config_files;
		char found_it = FALSE;

		while (ctmp != NULL ) {
		    if (!strcmp(ctmp->fileHeader, "snmpd")) {
			ctmp->write_flag = TRUE;
			found_it = TRUE;
			set_next_write_time();
			break;
		    }
		    ctmp = ctmp->next;
		}
		if (!found_it) {
		    /*
		     * not fatal to SNMP but needs to be fixed if it occurs
		     */
		    DEBUGMSG(("writeSystem", "Don't know file to write %i to",
								(char)name[7]));
		}
	    }

		/* Save the old value, in case of UNDO */
	    strcpy( oldbuf, buf);
	    memcpy( buf, var_val, var_val_len);
	    buf[var_val_len] = 0;
	    break;

	case UNDO:		/* Reverse the SET action and free resources */
	    revert_next_write_time();
	    strcpy( buf, oldbuf);
	    oldbuf[0] = 0;
	    break;

	case COMMIT:		/* Confirm the SET, performing any irreversible actions,
					and free resources */
	case FREE:		/* Free any resources allocated */

		/* No resources have been allocated, but "empty" the 'oldbuf' */
	    oldbuf[0] = 0;
	    break;
    }
    return SNMP_ERR_NOERROR;
} /* end of writeSystem */

	/*********************
	 *
	 *  Internal implementation functions - These generate output lines
	 *  for saving out to file
	 *
	 *********************/

/* Enough for 51.2 lines of 80 chars, should be enough */
#define MAX_CONFIG_LINE_LENGTH	4096
static char line[MAX_CONFIG_LINE_LENGTH];

char *
system_write_config_sysloc(const char *token)
{
    strcpy(line, token);
    strcat(line, " ");
    strcat(line, sysLocation);
    strcat(line, "\n");

    return line;
}

char *
system_write_config_syscon(const char *token)
{
    strcpy(line, token);
    strcat(line, " ");
    strcat(line, sysContact);
    strcat(line, "\n");

    return line;
}

char *
system_write_config_sysnam(const char *token)
{
    strcpy(line, token);
    strcat(line, " ");
    strcat(line, sysName);
    strcat(line, "\n");

    return line;
}

