#include <config.h>

#include <stdio.h>
#include <ctype.h>
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#if HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#if TIME_WITH_SYS_TIME
# ifdef WIN32
#  include <sys/timeb.h>
# else
#  include <sys/time.h>
# endif
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#if HAVE_WINSOCK_H
#include <winsock.h>
#endif
#if HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#if HAVE_NETDB_H
#include <netdb.h>
#endif
#include <errno.h>

#if HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#include "asn1.h"
#include "mib.h"
#include "parse.h"
#include "system.h"
#include "snmp_api.h"
#include "snmp_debug.h"
#include "snmp_logging.h"
#include "snmp_impl.h"
#include "default_store.h"
#include "callback.h"
#include "read_config.h"
#include "write_config.h"

extern struct config_files *config_files;
extern char sysLocation[];
extern char sysContact[];
extern char sysName[];

static void adjust_next_write_time(int mode);

#define REVERT  1
#define SET 2

/*
 * Keep it all nice and object orientated and tightly scoped in keeping with
 * the spirit of the rest of the code
 */
void
revert_next_write_time(void)
{
    adjust_next_write_time(REVERT);
}

void
set_next_write_time(void)
{
    adjust_next_write_time(SET);
}

static void
adjust_next_write_time(int mode)
{
    static struct timeval old_next_write_time;

    if (mode == REVERT) {
  next_write_time.tv_sec = old_next_write_time.tv_sec;
  next_write_time.tv_usec = old_next_write_time.tv_usec;
    }
    else {  /* SET */
  old_next_write_time.tv_sec = next_write_time.tv_sec;
  old_next_write_time.tv_usec = next_write_time.tv_usec;

  gettimeofday(&next_write_time, (struct timezone *)NULL);
  next_write_time.tv_sec++;
    }
}

void
write_config_store(const char *file, const char *line)
{
    DEBUGMSGTL(("write_config_store", "file: %s line: %s\n",
              file, line));
    read_config_store(file, line);
}

/*
 * we've been called because the 1 second timeout signalling that someone has
 * something to write has expired. Scan all files to find out which need writing
 */

void
write_config(void)
{
    extern struct timeval next_write_time;
    struct config_files *ctmp = config_files;
    struct config_line *ltmp;
    char *ret_line;

    next_write_time.tv_sec = 0; next_write_time.tv_usec = 0;
    /*
     * Scan the list of file types to find which need writing
     */

    while (ctmp != NULL ) {
	if (ctmp->write_flag) {
	    ctmp->write_flag = FALSE;

	    snmp_save_persistent(ctmp->fileHeader);

	    read_config_store(ctmp->fileHeader, "\
# Copyright (C) 2001 Intersil Netherlands BV\n\
#\n\
# This file is auto-generated, kill snmpd and any associated application\n\
# daemon and reboot if you want to make changes by hand. Only changes to\n\
# comments after this one (down to the \"End of user comments\" comment) will\n\
# be preserved.\n\n");

            /* write any user added comments */

            /* dougal, this isn't getting init'd yet...
             * read_config_store(ctmp->fileHeader, ctmp->comment);
             */

            read_config_store(ctmp->fileHeader, "# End of user comments\n\n");

            if (!strcmp("snmpd", ctmp->fileHeader)) {
		 read_config_store(ctmp->fileHeader, "\
# This file contains variables related to the SNMP protocol itself such as\n\
# the system MIB.\n\n");
 	    }
	    else if (!strcmp("private", ctmp->fileHeader)) {
		read_config_store(ctmp->fileHeader, "\
# This file contains secret variables that should be readable only by root\n\
# such as community names, encryption keys etc\n\n");
            }

            ltmp = ctmp->start;

	    while (ltmp != NULL) {
		if (ltmp->write_line == 0 || ltmp->config_token == NULL) {
		    DEBUGMSGTL(("write_config",
		    		"no write func for token: %s\n",
		    		ltmp->config_token));
		}
		else {
		    if ((ret_line = (*(ltmp->write_line))(ltmp->config_token))
			    == NULL) {
			DEBUGMSGTL(("write_config",
				    "no text returned for token: %s\n",
				    ltmp->config_token));
		    }
		    else {
			read_config_store(ctmp->fileHeader, ret_line);
		    }
		}

		ltmp = ltmp->next;
	    }

	    snmp_clean_persistent(ctmp->fileHeader);
	    
	    /* Check if a callbackfunction has been defined, if so -> call it */
	    if (ctmp->callback != NULL) {
		(*ctmp->callback)();
	    }

	}
	ctmp = ctmp->next;
    }
}
