/*
 *	$Header$
 *
 *	IMPORTANT--READ CAREFULLY:
 *	Right, including without limitation, use, reproduction, modification,
 *	and creation of derivative works, pertaining to this Source Code
 *	("Document") are governed by a legal Agreement ("Agreement") between
 *	you (either an individual or a single entity) and Intersil Americas
 *	Inc ("Intersil") and/or its licensee (both hereinafter called
 *	Licensor).
 *	By reproducing, modifying, creating derivative works, accepting,
 *	installing, viewing, accessing, transforming or otherwise using this
 *	Document, you agree to be bound by the terms of the Agreement. If you
 *	do not agree to the terms of the Agreement, Licensor is unwilling to
 *	license rights to this Document to you. In such event, you may not
 *	use, reproduce, modify, create derivative works, install, view,
 *	access, transport, transform or circulate this Document, and you
 *	should promptly contact Licensor from which you acquired this Document
 *	for instructions on returning or destroying this Document.
 *
 *	CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *	This Document contains confidential, proprietary and trade secret
 *	information of Intersil Americas Inc. and should be treated in a
 *	confidential manner.
 *	Furthermore, this Document is protected by national copyright laws and
 *	international treaties.
 *	Unauthorized use, reproduction, modification, creation of derivative
 *	works or circulation is strictly prohibited.
 *
 *	Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *	Copyright  2001-2003 Intersil Americas Inc. All Rights Reserved.
 */

//#define DEBUG

#ifdef DEBUG
#include <errno.h>
#endif

#include <sys/socket.h>

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "datastore_if.h"

#define TRUE (1==1)
#define FALSE (!TRUE)

static char snmpReadCommunity[65];
static int snmpReadCommunity_len;

#define SNMP_PASSWD_FILE "/var/etc/passwd.conf"

static char* last_oid;

/**********************************************************************
 ** read_comm_strings
 *
 *  DESCRIPTION: Read the communitystrings from the SNMP config file
 *
 *  PARAMETERS:
 *
 *  RETURN:
 *
***********************************************************************/
void wb_read_comm_strings (void)
{
  FILE* snmp_conf_file;
  char temps[80];

  if ( (snmp_conf_file = fopen( SNMP_PASSWD_FILE, "r")) == NULL )
  {
    fprintf(stderr, "Error, couldn't open SNMP passwd file\n");
    return;
  }

  if (fgets(temps,80, snmp_conf_file)) {
	  /* chomp */
	  temps[strlen(temps)-1] = '\0';
  }
  else {
	  /* hmm. empty file */
	  temps[0] = '\0';
  }
  
  strcpy(snmpReadCommunity, temps);
  snmpReadCommunity_len = strlen(snmpReadCommunity);

  fclose (snmp_conf_file);
}

/**********************************************************************
 ** wb_datastore_init
 *
 *  DESCRIPTION:
 *
 *  PARAMETERS:
 *
 *  RETURN:
 *
***********************************************************************/
int wb_datastore_init (void)
{
    static  int		initialized = FALSE;

    if (!(initialized))
    {

	/* for now set static IP address. */
	/* FIXME: we should use something like gethostbyname(gethostname()) */
	/* Open the socket used for connecting to SNMP */

  	initialized = TRUE;
    }

    /* Read community strings */
    wb_read_comm_strings();

    return 0;
}

#if 0
/**********************************************************************
 ** datastore_oid2ssi_oid
 *
 *  DESCRIPTION: Converts a datastore oid (.X.X.X.X) to a ssi oid
 *               (X_X_X_X).
 *
 *  PARAMETERS:
 *
 *  RETURN:
 *
***********************************************************************/
static void datastore_oid2ssi_oid(char* out, char* in) {

	char* cp;
	
	strcpy(out, in + 1); /* skip first . */
	
   	 /* convert any '.' characters in TableOID to _ characters */
    	for (cp = out; (cp = strchr (cp, '.')) != NULL; cp++) {
		*cp = '_';
	}

}
#endif

/**********************************************************************
 ** ssi_oid2datastore_oid
 *
 *  DESCRIPTION: Converts a ssi oid ([i]X_X_X_X_X) to a datastore oid
 *               (.X.X.X.X). If the ssi_oid was prefixed with an 
 *               conversion character. This character is return in
 *               conv_char. Otherwise the returned conv_char is 0x00;
 *
 *  PARAMETERS:
 *
 *  RETURN:
 *
***********************************************************************/
static void ssi_oid2datastore_oid(char* out, char* in, char* conv_char) {
	 
	char* i = in;
	char* o = out;

	/* prefix output with . */
	*o = '.';
	o++;
	  
	switch(*i) {
		case 'i':  
		case 's':
		case 'd':  
		case 'n':  
		case 'o':  
		case 't':  
		case 'a':  
		case 'm':  
		case 'u':
		case 'x':  
			*conv_char = *i;
			i++;
			break;
		default:
			*conv_char = 0x00;
			break;
	}

	while(*i != 0x00) {
		
		if(*i != '_') {
			*o = *i;
		}
		else {
			*o = '.';
		}
		
		i++;
		o++;
	}
	
	/* make out NULL-terminated */
	*o = 0x00;
	  
}

static char* retval2string(int retval){
	static char buf[40];

	switch (retval){
		case R_NOSUCHOID:
			return "No such OID";
		case R_NOACCESS:
			return "No access to OID";
		case R_ENDOFTREE:
			return "End of MIB";
		default:
			snprintf(buf, sizeof(buf), "Generic Datastore error (%i)", retval);
			return buf;
	}
			
}



static char* typeid2typename(int typeid) {
	
	switch(typeid) {
		case T_UNKNOWN:
			return "Unknown";
		case T_INTEGER:
			return "INTEGER";
		case T_UNSIGNED:
			return "UNSIGNED";
		case T_IPADDR:
			return "IPADDR";
		case T_TABLE:
			return "TABLE";
		case T_ROWSTAT:
			return "ROWSTAT";
		case T_OCTETSTR:
			return "OCTETSTR";
		case T_COUNTER:
			return "COUNTER";
		case T_COUNTER64:
			return "COUNTER64";
		case T_GAUGE:
			return "GAUGE";
		case T_OBJECTID:
			return "OBJECTID";
		case T_OPAQUE:
			return "OPAGUE";
		case T_TICKS:
			return "TICKS";
		case T_BITS:
			return "BITS";
		default:
			return "Type Error";
	}		
}

static void data_convert(char* conv_data, char* data, int data_len, int type, char conv_char) {

	int i;
	char* p;

	/* be sure to start with empty data */
	*conv_data = '\0';
	
	switch (conv_char) {
		case 'm':
			if((data_len == 6) && (type == T_OCTETSTR)) {
				/* MAC should be 6, type should be T_OCTETSTR */
				sprintf(conv_data, "%02X:%02X:%02X:%02X:%02X:%02X", data[0], data[1], data[2], data[3], data[4], data[5]);
				return;
			}
			break;
		case 'x':
			if(   (type == T_OCTETSTR)
			   || (type == T_BITS)
					 
				){
				p = conv_data;
				for(i=0; i< data_len; i++) {
					sprintf(p, "%02X", data[i]);
					p+=2;
				}
				/* make it NULL-terminated */
				*p = '\0';
				return;
			}
			if(   (type == T_INTEGER)
			   || (type == T_UNSIGNED)
			   || (type == T_ROWSTAT)
			   || (type == T_COUNTER)
			   || (type == T_COUNTER64)
			   || (type == T_GAUGE)
			   || (type == T_TICKS)
				){
				sprintf(conv_data, "%X", atoi(data));
				return;
			}
			break;
		case 'i':
		case 'u':
		case 't':
			if(   (type == T_INTEGER)
			   || (type == T_UNSIGNED)
			   || (type == T_ROWSTAT)
			   || (type == T_COUNTER)
			   || (type == T_COUNTER64)
			   || (type == T_GAUGE)
			   || (type == T_TICKS)
				){
				strcpy(conv_data, data);
				return;
			}
			break;
		case 'a':
			if(type == T_IPADDR) {
				strcpy(conv_data, data);
			}
			return;
		case 's':
			if(    (type == T_OCTETSTR)
			    ){
				strncpy(conv_data, data, data_len);
			}
			return;
		case 'o':
			if(    (type == T_OBJECTID)
			    ){
				strncpy(conv_data, data, data_len);
			}
		default:
			break;
			
	}
	
	sprintf(conv_data, "Conversion Error (from %s to %c, length = %i)", typeid2typename(type), conv_char, data_len);	
	
}



/**********************************************************************
 ** wb_datastore_get
 *
 *  DESCRIPTION: Generates a datastore get message, send it to the
                 datastore and process the result.
 *
 *  PARAMETERS:
 *
 *  RETURN:
 *
***********************************************************************/
char *wb_datastore_get (char *var_name_str) {
	static char oidbuf[160];
	static char data[160];
	static char conv_data[160];
	int data_len = sizeof(data);
	char conv_char;
	int type;
	int retval;
	char* p;
#ifdef DEBUG
        printf("wb_snmp_get in : (%s)\n", var_name_str);
#endif		
	/* convert oid to datastore oid */
	ssi_oid2datastore_oid(oidbuf, var_name_str, &conv_char);
	
	/* get the value from datastore */
	retval = datastore_getoid(oidbuf, snmpReadCommunity, snmpReadCommunity_len, data, &data_len, &type);

	last_oid = oidbuf + 1;
	
	if(retval != R_NOERR) {
		/* error encountered, return the corresponding string */
		p = retval2string(retval);
#ifdef DEBUG		
		printf("wb_snmp_get out: (%s)=(%s)\n", var_name_str, p);
#endif
		return(p);
	}	
	else {
		if(conv_char == 0x00) {
			/* no conversion needed */
#ifdef DEBUG			
			printf("wb_snmp_get out: (%s)=(%s)\n", var_name_str, data);
#endif
			return data;
		}
		else {
			data_convert(conv_data, data, data_len, type, conv_char);
#ifdef DEBUG			
			printf("wb_snmp_get out: (%s)=(conv_data)\n", var_name_str, p);
#endif			
			return conv_data;
		}
	}
	
}


/**********************************************************************
 ** wb_snmp_getnext
 *
 *  DESCRIPTION: Generates a snmp getnext message to our self
 *
 *  PARAMETERS:
 *
 *  RETURN:
 *
***********************************************************************/
char *wb_datastore_getnext (char *var_name_str, int *error)
{
	static char oidbuf[160];
	static char data[160];
	static char conv_data[160];
	int data_len = sizeof(data);
	char conv_char;
	int type;
	char* p;
#ifdef DEBUG	   
    	printf("wb_snmp_getnext in : (%s)\n", var_name_str);
	printf("wb_snmp_getnext in :e[%i]\n", *error);
#endif    
	
	/* convert oid to datastore oid */
	ssi_oid2datastore_oid(oidbuf, var_name_str, &conv_char);
	
	/* get the value from datastore */
	*error = datastore_getnextoid(oidbuf, snmpReadCommunity, snmpReadCommunity_len, data, &data_len, &type);

	//strcpy(last_oid, oidbuf + 1);	
	last_oid = oidbuf + 1;

	if(*error != R_NOERR) {
		/* error encountered, return the corresponding string */
		p = retval2string(*error);
#ifdef DEBUG		
		printf("wb_snmp_getnext out : (%s)=(%s)\n", var_name_str, p);
		printf("wb_snmp_getnext out :e[%i]\n", *error);
#endif		
		return(p);
		
	}	
	else {
		if(conv_char == 0x00) {
			/* no conversion needed */
#ifdef DEBUG			
			printf("wb_snmp_getnext out : (%s)=(%s)\n", var_name_str, data);
			printf("wb_snmp_getnext out :e[%i]\n", *error);
#endif			

			return data;
		}
		else {
			data_convert(conv_data, data, data_len, type, conv_char);
#ifdef DEBUG			
			printf("wb_snmp_getnext out : (%s)=(%s)\n", var_name_str, conv_data);
			printf("wb_snmp_getnext out :e[%i]\n", *error);
#endif			
			return conv_data;
		}
	}
	
}


/**********************************************************************
 ** wb_datatore_last_oid
 *
 *  DESCRIPTION: Returns the oid of the last get(next) response converted
 *		 to a string.
 *
 *  PARAMETERS:
 *
 *  RETURN:
 *
***********************************************************************/
char *wb_datastore_last_oid (void)
{
#ifdef DEBUG	
	printf("wb_snmp_last_oid = (%s)\n", last_oid);
#endif    
	return last_oid;
}

/**********************************************************************
 ** wb_snmp_read_community
 *
 *  DESCRIPTION: Returns the current read community string for the
 *		 Webserver/SNMP agent interface.
 *
 *  PARAMETERS:
 *
 *  RETURN:
 *
***********************************************************************/
char *wb_snmp_read_community (void)
{
    return snmpReadCommunity;
}

