/*
 *	$Header$
 *
 *	IMPORTANT--READ CAREFULLY:
 *	Right, including without limitation, use, reproduction, modification,
 *	and creation of derivative works, pertaining to this Source Code
 *	("Document") are governed by a legal Agreement ("Agreement") between
 *	you (either an individual or a single entity) and Intersil Americas
 *	Inc ("Intersil") and/or its licensee (both hereinafter called
 *	Licensor).
 *	By reproducing, modifying, creating derivative works, accepting,
 *	installing, viewing, accessing, transforming or otherwise using this
 *	Document, you agree to be bound by the terms of the Agreement. If you
 *	do not agree to the terms of the Agreement, Licensor is unwilling to
 *	license rights to this Document to you. In such event, you may not
 *	use, reproduce, modify, create derivative works, install, view,
 *	access, transport, transform or circulate this Document, and you
 *	should promptly contact Licensor from which you acquired this Document
 *	for instructions on returning or destroying this Document.
 *
 *	CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *	This Document contains confidential, proprietary and trade secret
 *	information of Intersil Americas Inc. and should be treated in a
 *	confidential manner.
 *	Furthermore, this Document is protected by national copyright laws and
 *	international treaties.
 *	Unauthorized use, reproduction, modification, creation of derivative
 *	works or circulation is strictly prohibited.
 *
 *	Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *	Copyright  2001-2003 Intersil Americas Inc. All Rights Reserved.
 */

// #define DEBUG

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <arpa/inet.h>			/* inet_ntoa */
#include <sys/types.h>			/* socket, bind, accept */
#include <sys/socket.h>			/* socket, bind, accept, setsockopt, */
#include <sys/un.h>				/* sockaddr_un */

#include <signal.h>
#include <aplog.h>
#include "wb_datastore.h"

static struct sockaddr_un servaddr;

static char *wb_get_rootdir (void);
static int *wb_dedot (char *out, char *in);

#define MAX_PATH 255
#define UNIX_DOMAIN_SOCKET "/var/run/ssi.sock"

#define INCSTART    "<!--#" /* start of include sequence */
#define INCEND	    "-->"   /* end of include sequence	 */
#define CMTSTART    "<!--"  /* start of comment sequence */
#define CMTEND	    INCEND  /* end of comment sequence */

#define FILETAG	    "file"

#define WB_SSI_ERROR	    -2
#define WB_TAB_NO_TABLE	    -3
#define WB_TAB_OUT_TABLE    -4
#define WB_TAB_BAD_WHERE    -5

#define MIN_KEYW_ID	256
#define MAX_KEYWORD_LEN 32
#define MAX_VALUE_SIZE	256

/* Command identifiers */
#define ILLCMD_ID	-1
#define GETCMD_ID	300
#define NEXTCMD_ID	301
#define INCCMD_ID	302
#define TABLECMD_ID	303

#define ENDTABLECMD_ID	304
#define COLVALCMD_ID	305
#define COLOIDCMD_ID	306

#define INCSTART_ID	310
#define INCEND_ID	311
#define CMTSTART_ID	312
#define CMTEND_ID	313

#define OIDTAG_ID	320
#define TYPETAG_ID	321
#define CASETAG_ID	322
#define AMOUNTTAG_ID	323
#define BEGINTAG_ID	324
#define ENDTAG_ID	325
#define WHERETAG_ID	326

#define INTEGERTYPE_ID	330
#define STRINGTYPE_ID	331
#define IPTYPE_ID	332
#define MACTYPE_ID	333
#define OIDTYPE_ID	334
#define TIMETYPE_ID	335
#define HEXSTRTYPE_ID	336

/* structure to describe a keyword */
typedef struct
{
    char    *name;	/* name of the keyword */
    int	    namelen;	/* length of the name string */
    int	    id;		/* identifier for this keyword */
}
KeyWord;


/* The supported server side include commands (alphabetic order!!) */
static KeyWord ssi_cmds[] =
{
    {"col_oid", 7,  COLOIDCMD_ID},
    {"col_val", 7,  COLVALCMD_ID},
    {"endtable",8,  ENDTABLECMD_ID},
    {"get",	3,  GETCMD_ID},
    {"include", 7,  INCCMD_ID},
    {"next",	4,  NEXTCMD_ID},
    {"table",	5,  TABLECMD_ID},
    {NULL,	0,  ILLCMD_ID}
};


/* The start server side include keyword */
static KeyWord start_keyw[] =
{
    {INCSTART,	5,  INCSTART_ID},
    {NULL,	0,  ILLCMD_ID}
};


/* The end server side include keyword */
static KeyWord end_keyw[] =
{
    {INCEND,	    3,	INCEND_ID},
    {NULL,	0,  ILLCMD_ID}
};


/* The start and end comment keywords (alphabetic order!!) */
static KeyWord comment_keyw[] =
{
    {CMTEND,	3,  CMTEND_ID},
    {CMTSTART,	4,  CMTSTART_ID},
    {NULL,	0,  ILLCMD_ID}
};


/* Tags supported by the "get" and "next" command (alphabetic order!!) */
static KeyWord get_tags[] =
{
    {INCEND,	3,  INCEND_ID},
    {"case",	4,  CASETAG_ID},
    {"oid",	3,  OIDTAG_ID},
    {"type",	4,  TYPETAG_ID},
    {NULL,	0,  ILLCMD_ID}
};

/* Tags supported by the "table" command (alphabetic order!!) */
static KeyWord table_tags[] =
{
    {INCEND,	3,  INCEND_ID},
    {"amount",	6,  AMOUNTTAG_ID},
    {"begin",	5,  BEGINTAG_ID},
    {"end",	3,  ENDTAG_ID},
    {"oid",	3,  OIDTAG_ID},
    {"where",	5,  WHERETAG_ID},
    {NULL,	0,  ILLCMD_ID}
};

/* the skip table commands */
static KeyWord skiptable_cmds[] =
{
    {"<!--#endtable",	13, ENDTABLECMD_ID},
    {"<!--#table",	10, TABLECMD_ID},
    {NULL,		0,  ILLCMD_ID}
};

/* type values supported (alphabetic order!!) */
static KeyWord types[] =
{
    {"hex",	3,  HEXSTRTYPE_ID},
    {"integer", 7,  INTEGERTYPE_ID},
    {"ip",	2,  IPTYPE_ID},
    {"mac",	3,  MACTYPE_ID},
    {"oid",	3,  OIDTYPE_ID},
    {"string",	6,  STRINGTYPE_ID},
    {"time",	    4,	TIMETYPE_ID},
    {NULL,	0,  ILLCMD_ID}
};


/* full path to the file name that was requested by the browser */
static char *url_filepath;

static int wb_parse_cmd (FILE *fp_in, FILE *fp_out);
static int wb_parse_file (FILE *fp_in, FILE *fp_out);

void sig_exit_handler(int signal){
    fprintf(stderr, "SSI receiving signal %i.\n", signal);
    abort;
}

static void sig_hup_handler(int signal)
{
  aplog(EV_SNMP_INFO,"SSI re-reading SNMP passwords",NULL);
  wb_read_comm_strings();
}


ssize_t readline(int fd, void *vptr, size_t maxlen)
{
	ssize_t n, rc;
	char c, *ptr;

	ptr = vptr;

	for (n = 1; n < maxlen ; n++ )
	{
		again:
			if ( ( rc = read(fd, &c, 1)) == 1) {
				*ptr++ = c;
				if ((c == '\n') || (c == '\r')) {
					ptr--;
					break; /* cr and/or newline are NOT stored, unlike fgets() */
				}
			} else if (rc == 0) {
					if (n == 1)
						return (0); /*EOF no data read */
					else
						break; /* EOF, some data read */
			} else {
				if (errno == EINTR)
					goto again;
				return (-1);
			}
	}

	*ptr = 0;
	return (n);
}


/**********************************************************************
 ** wb_scan_nonspace
 *
 *  DESCRIPTION: Scan the file 'fp' for the first non space character.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 The first non space character, or EOF on EOF.
 *
***********************************************************************/
static int wb_scan_nonspace (FILE *fp)
{
    int c;

    do
    {
	c = fgetc (fp);
    }
    while (isspace (c));

    return c;
}


/**********************************************************************
 ** wb_scan_string
 *
 *  DESCRIPTION: Scan the file 'fp' and store all characters read in
 *		 buffer 'buffer'. Stop scanning if a character in
 *		 'terminators' was read, the buffer is full or the end
 *		 of file was reached.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 The character that did terminate the scanning, EOF
 *		 on EOF, or WB_SSI_ERROR on error.
 *
***********************************************************************/
static int wb_scan_string (FILE *fp, char *terminators, char *buffer,
			    int maxlen)
{
    int len;
    int c = 0;

    for (len = 0; len < maxlen; len++)
    {
	c = fgetc (fp);
	if (c == EOF)
	    break;
	if (strchr (terminators, c) != NULL)
	{
	    if (len == 0 || buffer[len-1] != '\\')
		break;
	}
	buffer[len] = c;
    }

    buffer[len] = '\0';
    if (len == maxlen)
	c = WB_SSI_ERROR;

    return c;
}

/**********************************************************************
 ** strcncmp
 *
 *  DESCRIPTION: A case insensetive compare of two strings for at
 *		 most n characters.
 *
 *  PARAMETERS:	 cs and ct are the pointers to the strings with need
 *		 to be comared. n is the number of maximal characters
 *		 to compare
 *
 *  RETURN:	 <0 if cs<ct, 0 if cs==ct or >0 if cs>ct.
 *
***********************************************************************/
int strcncmp( const char *cs, const char *ct, int n )
{
    unsigned char chars, chart;
    int cnt;
    int res=0;

    for( cnt=0; cnt<n && !res; cnt++, cs++, ct++ )
    {
      if( (*cs>='A') && (*cs<='Z'))
	chars = tolower( *cs );
      else
	chars = *cs;

      if( (*ct>='A') && (*ct<='Z'))
	chart = tolower( *ct );
      else
	chart = *ct;

      res = (int)chars - (int)chart;
    }

    return res;
}


/**********************************************************************
 ** keyword_term_scan
 *
 *  DESCRIPTION: Scan the file 'fp_in' for keywords and/or terminators.
 *		 If the file pointer 'fp_out' is not NULL, all characters
 *		 scanned until a keyword or a terminator is found are
 *		 copied to this file.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 The character that did terminate the scanning or
 *		 the identifier of the keyword found or EOF if the
 *		 end of file 'fp_in' was reached.
 *
***********************************************************************/
static int keyword_term_scan (FILE *fp_in, FILE *fp_out,
				KeyWord *keywords, char *terminators)
{
    char line[MAX_KEYWORD_LEN + 1], *cp;
    int linelen, cmplen, res, c;
    KeyWord *curkey;

    res = 0;
    linelen = 0;
    cmplen = 1;
    cp = line;
    curkey = keywords;
    while (1)
    {
	/* if we want to compare more chars than currently in the buffer.. */
	if (cmplen > linelen)
	{
	    /* then read the next char */
	    if ((c = fgetc (fp_in)) == EOF)
		return EOF;

	    /* and store it in the buffer */
	    *cp++ = c;
	    linelen++;
	}

	/* if the current match length is 1 then check if it's a terminator */
	if (terminators && cmplen == 1 && strchr (terminators, *line) != NULL)
	    return *line;

	/* search for first keyword that still matches with what we scanned */
	while ((curkey->name) &&
	      (res = strcncmp (curkey->name, line, cmplen)) < 0)
	    curkey++;

	/* if no keyword found... */
	if (curkey->name == NULL || res > 0)
	{
	    /* and there is an output file, then write the first char */
	    if (fp_out)
		fputc (*line, fp_out); /* don't return EOF on EOF!! */

	    /* remove this char from the buffer */
	    linelen--;
	    strncpy (line, line + 1, linelen);
	    cp--;

	    /* start comparison again with first char of first keyword */
	    curkey = keywords;
	    cmplen = 1;
	}
	else
	{
	    if (cmplen == curkey->namelen)
	    {
#ifdef DEBUG
		printf ("keyword_scan: found '%s'\n", curkey->name);
#endif
		/* return the identifier of the keyword found */

		return curkey->id;
	    }
	    else
		/* if we still match, try next time with one more character */
		cmplen++;
	}
    }
    return 0;
}


/**********************************************************************
 ** wb_scan_value
 *
 *  DESCRIPTION: Scan file 'fp_in' for a value and write this value to
 *		 file 'fp_out'. The value is terminated by one of the
 *		 characters found in 'terminators'. If a SSI command is
 *		 found, this command is executed and the result is used
 *		 as value.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 The character that did terminate the value or an error code.
 *
***********************************************************************/
static int wb_scan_value (FILE *fp_in, FILE *fp_out, char *terminators)
{
    int c;
    int error;

    do
    {
	/* scan until a terminator char or the <!--# sequence is found */
	c = keyword_term_scan (fp_in, fp_out, start_keyw, terminators);

	/* if we did find the start SSI sequence.. */
	if (c >= MIN_KEYW_ID)
	{
	    /* then start to parse this new command */
	    error = wb_parse_cmd (fp_in, fp_out);

	    if (error != 0)
		c = error;
	}
    }
    /* keep on scanning until we found a terminator or reached EOF */
    while (c >= MIN_KEYW_ID);

    return c;
}


/**********************************************************************
 ** wb_skip_value
 *
 *  DESCRIPTION: Read from file 'fp_in' until the end of a value is
 *		 reached. The value is terminated by one of the
 *		 characters found in 'terminators'. If SSI commands where
 *		 found, these commands are also skipped.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 The character that did terminate the scanning or
 *		 WB_SSI_ERROR on error.
 *
***********************************************************************/
static int wb_skip_value (FILE *fp_in, char *terminators)
{
    int nested_cmds;
    int c;
    char *scan_terminators;

    /* we currently didn't find a command or comment */
    nested_cmds = 0;
    do
    {
	/* if we are inside a command, we don't stop on terminators */
	scan_terminators = (nested_cmds == 0) ? terminators : NULL;

	/* scan for the next start or end command sequence (or terminator) */
	c = keyword_term_scan (fp_in, NULL, comment_keyw,
	scan_terminators);

	/* increment the number of times we are inside a command or comment */
	if (c == CMTSTART_ID)
	    nested_cmds++;
	else
	{
	    /* decrement the number of times we are inside a comment */
	    if (c == CMTEND_ID)
	    {
		nested_cmds--;
		/* return an error if we found more end than start sequnces */
		if (nested_cmds < 0)
		    return WB_SSI_ERROR;
	    }
	}
    }
    /* keep on scanning until we found a terminator or reached EOF */
    while (c >= MIN_KEYW_ID);

    return c;
}


/**********************************************************************
 ** wb_parse_case
 *
 *  DESCRIPTION: Parse the value behind a "case" tag.
 *
 *  EXAMPLE:	  case="valueA=valueX|valueB=valueY|=valueZ"
 *		 fp_in--^
 *
 *		 Scan the 'valueA=valueX' combinations and find which
 *		 valueA matches with the given value 'value'. In that
 *		 case write the valueX to file 'fp_out'.
 *
 *  NOTE:	 valueX can contain another SSI commands.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_case (FILE *fp_in, FILE *fp_out, char *value,
			int cvt_type)
{
    char buffer[MAX_VALUE_SIZE+1];
    int c;
    unsigned int size;
    int value_found = 0;

    size = strlen (value) + 1;
    do
    {
	/* if we didn't found value 'value' yet... */
	if (value_found == 0)
	{
	    /* scan for the next value uptill the = */
	    if ((c = wb_scan_string (fp_in, "=", buffer,
				    MAX_VALUE_SIZE)) < 0)
	    {
		fprintf (fp_out, "case: no '<value>=' found");
		return c;
	    }

	    /* if this is the default case.. */
	    if (buffer[0] == '\0')
		value_found = 1;

	    else
	    {
		/* or the value we are looking for..  */
		if (cvt_type == 'm' || cvt_type == 'x')
		    value_found = strcncmp (buffer, value, size) == 0;
		else
		    value_found = strcmp (buffer, value) == 0;
	    }
	    if (value_found)
		/* .. then scan valueX (behind the =) till a | or " is found */
		c = wb_scan_value (fp_in, fp_out, "|\"");
	    else
		/* otherwise skip valueX so we can start with next valueA */
		c = wb_skip_value (fp_in, "|\"");
	}
	else
	    /* if we al ready have our value found, skip al other values */
	    c = wb_skip_value (fp_in, "|\"");
    }
    /* keep scanning as long as we find | terminators */
    while (c == '|');

    if (c < 0)
    {
	fprintf (fp_out, "case: no '|' or '\"' found");
	return c;
    }

    /* if we didn't find the value in the list... */
    if (value_found == 0)
    {
	/* write our value untranslated to file 'fp_out' */
	size = strlen (value);
	if (fwrite (value, 1, size, fp_out) < size)
	    return WB_SSI_ERROR;
    }

    return 0;
}



/**********************************************************************
 ** get_column_value
 *
 *  DESCRIPTION: Get, with a SNMP getnext command, the value of the
 *		 object in a table with columnar oid
 *		 OidBuffer[1..TableOIDlen+1]+'column_oid_ext' and the row
 *		 following 'prev_row_index'.
 *		 If 'cvt_type' not equals -1, then the value is converted
 *		 with the specified type.
 *		 If 'row_index' equals "", then the index for the object
 *		 is copied in the buffer 'row_index' is pointing at.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 A pointer to the value, or NULL if we walked out of the
 *		 column.
 *
***********************************************************************/
static char *get_column_value (char *OidBuffer, int TableOIDlen,
				char *column_oid_ext, char *prev_row_index,
				int cvt_type, char *row_index)
{
    int	    offset;
    int	    column_oid_len;
    char    *value;
    char    *lastoid;
    int	    error;

    /* set the conversion type in OidBuffer[0], if any */
    if (cvt_type != -1)
    {
	OidBuffer[0] = cvt_type;
	offset = 0;
    }
    else
	offset = 1;

    column_oid_len = TableOIDlen + strlen (column_oid_ext);

    /* copy the columnar oid plus the index to OidBuffer */
    strcpy (&OidBuffer[1+TableOIDlen], column_oid_ext);
    strcpy (&OidBuffer[1+column_oid_len], prev_row_index);

    /* do a SNMP getnext command, get the value and oid */
    value = wb_datastore_getnext (&OidBuffer[offset], &error);
    lastoid = wb_datastore_last_oid ();

    /* test if the returned oid is still in the same column */
    if (error || strncmp (&OidBuffer[1], lastoid, column_oid_len) != 0)
	return NULL;

    /* copy the index for this object to 'row_index', if it's still empty */
    if (*row_index == '\0')
	strncpy (row_index, &lastoid[column_oid_len], MAX_PATH);

    return value;
}


/**********************************************************************
 ** get_columnar_oid_end
 *
 *  DESCRIPTION: Search for the dot behind the columnar OID.
 *
 *  PARAMETERS:	 oid	       - oid of object in the table (in ASCII).
 *		 table_oid_len - length of table oid prefix in oid.
 *
 *  RETURN:	 A pointer to the dot behind the columnar OID or NULL
 *		 if 'oid' is not part of a table.
 *
***********************************************************************/
static char *get_columnar_oid_end (char *oid, int table_oid_len)
{
    char *cp;

    /* cp points at dot after table oid */
    cp = oid + table_oid_len;

    if (*cp != '.')
	return NULL;

    /* search for dot behind entry OID */
    cp++;
    if ((cp = strchr (cp, '.')) == NULL)
	return NULL;

    /* search for dot behind columnar OID */
    cp++;
    return strchr (cp, '.');
}


/**********************************************************************
 ** get_prev_row_index
 *
 *  DESCRIPTION: Search in SNMP table 'TableOID' for the index of the
 *		 row before row 'BeginRow'.
 *		 Copy this index to buffer 'PrevRowIndex' with length
 *		 'MaxIndexLen'.
 *
 *  PARAMETERS:	 TableOid      - oid of the table (in ASCII).
 *
 *  RETURN:	 WB_TAB_NO_TABLE:  TableOID is not the OID of a table.
 *		 WB_TAB_OUT_TABLE: There is now row 'BeginRow' in table.
 *
***********************************************************************/
static int get_prev_row_index (char *TableOID, int BeginRow,
				char *PrevRowIndex, int MaxIndexLen)
{
    char    OidBuffer[MAX_PATH+1];
    int	    row;
    char    *valp;
    char    *lastoid;
    char    *cp;
    int	    TableOIDlen;
    int	    column_oid_len = 0;
    int	    test_oid;
    int	    error;

    /* for the first row, make the index empty (we start with the colum OID) */
    if (BeginRow == 0)
    {
	PrevRowIndex[0] = '\0';

	/* we want to test if TableOID is a valid table oid */
	test_oid = 1;
    }
    else
	test_oid = 0;

    TableOIDlen = strlen (TableOID);
    lastoid = TableOID;
    for (row = 0; test_oid || row < BeginRow; row++)
    {
	/* do SNMP getnext request to get the OID of next object */
	valp = wb_datastore_getnext (lastoid, &error);
	lastoid = wb_datastore_last_oid ();

	if (row == 0)
	{
		
	    if (error) {
		/* disabled WB_TAB_NO_TABLE because this could be the
		 * end of the MIB 
		return WB_TAB_NO_TABLE;
		*/
		return WB_TAB_OUT_TABLE;
	    }

	    /* see if the OID is still in the requested table */
	    if (strncmp (TableOID, lastoid, TableOIDlen) != 0
		|| (cp = get_columnar_oid_end (lastoid, TableOIDlen)) == NULL)
		return WB_TAB_OUT_TABLE;

	    /* if we where only testing the TableOID, return it's OK */
	    if (test_oid)
		return 0;

	    /* remember the columnar oid and its length for the first row */
	    column_oid_len = cp - lastoid;
	    strncpy (OidBuffer, lastoid, column_oid_len);
	}
	else
	{
	    /* test if the returned OID is still of the first column */
	    if (error || strncmp (OidBuffer, lastoid, column_oid_len) != 0)
		return WB_TAB_OUT_TABLE;
	}
    }

    /* store index part of the OID as the index of the row before BeginRow */
    strncpy (PrevRowIndex, &lastoid[column_oid_len], MaxIndexLen);

    return 0;
}


/**********************************************************************
 ** wb_test_row
 *
 *  DESCRIPTION: Test if the columns specified in 'where_str' of the row
 *		 following 'prev_row_index' of SNMP table
 *		 OidBuffer[1..'TableOIDlen'+1] have the specified values.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 0: one or columns didn't pass the test
 *		 1: all columns did pass the test.
 *		 error_code: there was an error.
 *
***********************************************************************/
static int wb_test_row (char *where_str, char *OidBuffer, int TableOIDlen,
			char *prev_row_index, char *row_index)
{
    int cvt_type;
    char buffer[MAX_PATH];
    char *cp;
    char *value, *test_value;
    int res, len;

    while (*where_str)
    {
	/* see if conversion type is specified */
	if (!isdigit(*where_str))
	    cvt_type = *where_str++;
	else
	    cvt_type = -1;

	/* copy column OID extention to buffer (terminated by '=') */
	cp = buffer;
	while (*where_str && *where_str != '=')
	    *cp++ = *where_str++;
	*cp = '\0';

	/* return error if there is no column OID specified */
	if (buffer[0] == '\0')
	    return WB_TAB_BAD_WHERE;

	/* return error if we didn't find the '=' */
	if (*where_str != '=')
	    return WB_TAB_BAD_WHERE;
	where_str++;

	/* get the value of the object following
	    OidBuffer[1..TableOIDlen+1]+buffer+prev_row_index */
	value = get_column_value (OidBuffer, TableOIDlen, buffer,
	prev_row_index,
				cvt_type, row_index);
	if (value == NULL)
	    return WB_TAB_OUT_TABLE;

	/* save start and determine length of test value (terminated by ' ') */
	test_value = where_str;
	for (len = 0; *where_str && !isspace (*where_str); where_str++)
	    len++;

	/* compare the returned value with the test value */
	if (cvt_type == 'm' || cvt_type == 'x')
	    res = strcncmp (test_value, value, len);
	else
	    res = strncmp (test_value, value, len);

	if (res != 0)
	    return 0;

	/* skip "spaces AND spaces" from 'where_str', if any */
	while (*where_str && isspace (*where_str))
	    where_str++;

	if (strcncmp (where_str, "and", 3) == 0)
	{
	    where_str += 3;
	    while (*where_str && isspace (*where_str))
		where_str++;
	}
    }
    return 1;
}

/* variables to store parameters of current table command */
static char OidBuffer[MAX_PATH+1];
static char PrevRowIndex[MAX_PATH+1];
static char RowIndex[MAX_PATH+1];
static int  TableOIDlen = 0;

static int wb_table_output (FILE *fp_in, FILE *fp_out, char *TableOID,
			    int BeginRow, int EndRow, int nRows,
			    char *where_str)
{
    char    cOidBuffer[MAX_PATH+1];
    char    cPrevRowIndex[MAX_PATH+1];
    char    cRowIndex[MAX_PATH+1];
    int	    cTableOIDlen;
    int	    error;
    int	    test_passed;
    int	    row, rows;
    long    start_pos;
    int	    keywordID;
    int	    intables;

    /* save parameters of current table command, if any */
    strcpy (cPrevRowIndex, PrevRowIndex);
    strcpy (cRowIndex, RowIndex);
    strcpy (cOidBuffer, OidBuffer);
    cTableOIDlen = TableOIDlen;

    /* seach for the index of the row before BeginRow */
    error = get_prev_row_index (TableOID, BeginRow, PrevRowIndex,
				sizeof (PrevRowIndex));
    
    /* only test for NO_TABLE, handle OUT_TABLE later on */
    if (error == WB_TAB_NO_TABLE)
    {
	fprintf (fp_out, "table: not a table OID");
	strcpy (PrevRowIndex, cPrevRowIndex);
	return error;
    }

    /* copy TableOID in global buffer (leave room for any conversion char) */
    strcpy (&OidBuffer[1], TableOID);
    TableOIDlen = strlen (TableOID);

    /* add index for table entry after TableOID in the global buffer */
    strcpy (&OidBuffer[1+TableOIDlen], ".1.");
    TableOIDlen += 3;

    /* remember the position in the file after the table command */
    start_pos = ftell (fp_in);

    /* walk through the table from BeginRow till EndRow, an error occurs, or
       the requested number of rows is reached */
    for (rows = 0, row = BeginRow; error == 0 && row <= EndRow
	    && rows < nRows; row++)
    {
	/* indicate we don't know the index of this row yet */
	RowIndex[0] = '\0';

	/* if we have to test each row, then test this one */
	if (where_str[0])
	{
	    test_passed = wb_test_row (where_str, OidBuffer, TableOIDlen,
			PrevRowIndex, RowIndex);

	    if (test_passed < 0)
	    {
		if (test_passed == WB_TAB_BAD_WHERE)
		    fprintf (fp_out, "table: bad where string");
		error = test_passed;
	    }
	}
	else
	    test_passed = 1;

	if (test_passed == 1)
	{
	    /* increment the number of rows handled */
	    rows++;

	    /* set file pointer back on row behind table command */
	    fseek (fp_in, start_pos, SEEK_SET);

	    /* parse file till endtable command for the current row */
	    error = wb_parse_file (fp_in, fp_out);
	}
	strcpy (PrevRowIndex, RowIndex);
    }

    /* if no rows are handled, then skip till after <!--#endtable --> tag */
    if (rows == 0 && (error == 0 || error == WB_TAB_OUT_TABLE))
    {
	error = 0;
	intables = 1;
	while (intables && error == 0)
	{
	    /* search for table and endtable commands */
	    keywordID = keyword_term_scan (fp_in, NULL, skiptable_cmds,
					    NULL);

	    /* if we found an endtable command... */
	    if (keywordID == ENDTABLECMD_ID)
	    {
		/* ..and the closing '-->' tag... */
		keywordID = keyword_term_scan (fp_in, NULL, end_keyw,
						NULL);
		if (keywordID == INCEND_ID)
		    intables--;	    /* ..then decrement the number of times we
					are inside a table */
		else
		    error = WB_SSI_ERROR;
	    }
	    else
	    {
		/* if found a table command... */
		if (keywordID == TABLECMD_ID)
		    intables++;	    /* ..then increment the number of times we
					are inside a table */
		else
		    error = WB_SSI_ERROR;
	    }
	}
    }

    /* restore parameters of any previous table command */
    strcpy (PrevRowIndex, cPrevRowIndex);
    strcpy (RowIndex, cRowIndex);
    strcpy (OidBuffer, cOidBuffer);
    TableOIDlen = cTableOIDlen;

    return (error == WB_TAB_OUT_TABLE) ? 0 : error;
}


/**********************************************************************
 ** wb_parse_endtable_cmd
 *
 *  DESCRIPTION: Parse the '-->' tag behind the endtable command.
 *  PARAMETERS:
 *
 *  RETURN:	 EOF if the '-->' tag was found, or an error code.
 *
***********************************************************************/
static int wb_parse_endtable_cmd (FILE *fp_in, FILE *fp_out)
{
    int	    keywordID;
    int	    OID_len;
    int	    row_index_len;
    int	    column_oid_len;
    char    *valp;
    char    *lastoid;
    char    *cp;

    /* scan '-->' tag from the input file */
    keywordID = keyword_term_scan (fp_in, NULL, end_keyw, NULL);
    if (keywordID != INCEND_ID)
	return WB_SSI_ERROR;

    /* see if we are processing a table, return an error if not */
    if (TableOIDlen == 0)
    {
	fprintf (fp_out, "Found endtable without any table command");
	return WB_SSI_ERROR;
    }

    /* see if we still operate in the same colomn of the same table */
    OID_len = strlen (&OidBuffer[1]);
    row_index_len = strlen (PrevRowIndex);
    column_oid_len = OID_len - row_index_len;

    /* if we didn't do a get next from the table (no where_str, col_oid or
	col_val).. */
    if (RowIndex[0] == '\0')
    {
	/* and we didn't get a value from this table before.. */
	if (OID_len == TableOIDlen)
	{
	    /* then there is no column index in OidBuffer.
	       Get the first value from the table */
	    valp = get_column_value (OidBuffer, TableOIDlen, "", "", -1,
				    "notempty");
	    if (valp == NULL)
		return WB_TAB_NO_TABLE;

	    /* store the OID found in OidBuffer */
	    lastoid = wb_datastore_last_oid ();
	    strcpy (&OidBuffer[1], lastoid);

	    /* if PrevRowIndex is empty, then assume a one number column OID */
	    if (row_index_len == 0)
	    {
		cp = &OidBuffer[1+TableOIDlen];

		/* search for dot behind columnar OID */
		if ((cp = strchr (cp, '.')) == NULL)
		    return WB_TAB_OUT_TABLE;

		column_oid_len = (cp - &OidBuffer[1]);
	    }
	    else
	    {
		OID_len = strlen (&OidBuffer[1]);
		column_oid_len = OID_len - row_index_len;
	    }
	}

	/* get the value following PrevRowIndex to get its RowIndex */
	valp = get_column_value (OidBuffer, column_oid_len, "", PrevRowIndex,
				    -1, RowIndex);
	if (valp == NULL)
	    return WB_TAB_OUT_TABLE;
    }

    /* see if the RowIndex found is still in the last requested column */
    valp = get_column_value (OidBuffer, column_oid_len, "", RowIndex, -1,
			    "notempty");
    return (valp == NULL) ? WB_TAB_OUT_TABLE : EOF;
}


/**********************************************************************
 ** wb_parse_table_cmd
 *
 *  DESCRIPTION: Parse the tags behind a "table" command.
 *		 Scan the 'oid="a.b.c.d"' and the optional 'begin="rownr"',
 *		 'end="rownr"', 'amount="number"', and/or 'where="condition"'
 *		 tag and value pairs.
 *		 Determine via SNMP getnext messages the geometry of the
 *		 table and return the values and/or OIDs of the colums value
 *		 of the MIB variable, optionaly convert this value with 'type'
 *		 and/or 'case', and write this value to file 'fp_out'.
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_table_cmd (FILE *fp_in, FILE *fp_out)
{
    int	    keywordID;
    char    TableOID[MAX_PATH+1];
    char    where_str[MAX_PATH+1];
    char    tempbuffer[20];
    int	    BeginRow, EndRow, nRows;
    int	    c;
    char    *cp;

    TableOID[0] = where_str[0] = '\0';
    BeginRow = 0;
    EndRow = nRows = 0x7FFFFFF;

    do
    {
	/* scan for the next 'oid', 'amount', 'begin', 'end', 'where',
	    or '-->' tag */
	keywordID = keyword_term_scan (fp_in, NULL, table_tags, NULL);
	if (keywordID < MIN_KEYW_ID)
	{
	    fprintf (fp_out, "table: bad tag scan");
	    return (keywordID == EOF) ? EOF : WB_SSI_ERROR;
	}

	/* if we found an 'oid','amount','begin','end' or 'where' tag ... */
	if (keywordID != INCEND_ID)
	{
	    /* then scan for the = and " behind these tags */
	    c = wb_scan_nonspace (fp_in);
	    if (c != '=')
	    {
		fprintf (fp_out, "table: expected =, found '%c'!", c);
		return WB_SSI_ERROR;
	    }

	    c = wb_scan_nonspace (fp_in);
	    if (c != '"')
	    {
		fprintf (fp_out, "table: expected \", found '%c'!", c);
		return WB_SSI_ERROR;
	    }
	    switch (keywordID)
	    {
		case OIDTAG_ID:
		    /* scan the oid uptill the ", and store it in TableOID */
		    if (wb_scan_string (fp_in, "\"", TableOID,
					MAX_PATH) < 0)
		    {
			fprintf (fp_out, "table: bad OID scan");
			return WB_SSI_ERROR;
		    }
		    break;
		case WHERETAG_ID:
		    /* scan the where till the ", and store it in where_str */
		    if (wb_scan_string (fp_in, "\"", where_str,
					MAX_PATH) < 0)
		    {
			fprintf (fp_out, "table: bad where scan");
			return WB_SSI_ERROR;
		    }
		    break;
		case BEGINTAG_ID:
		case ENDTAG_ID:
		case AMOUNTTAG_ID:
		    /* scan the number till the ", store it in tempbuffer */
		    if (wb_scan_string (fp_in, "\"", tempbuffer,
			sizeof (tempbuffer) - 1) < 0)
		    {
			fprintf (fp_out, "table: bad number scan");
			return WB_SSI_ERROR;
		    }
		    if (keywordID == BEGINTAG_ID)
			BeginRow = atoi (tempbuffer);
		    else
		    {
			if (keywordID == ENDTAG_ID)
			    EndRow = atoi (tempbuffer);
			else
			    nRows = atoi (tempbuffer);
		    }
		    break;
		default:
		    fprintf (fp_out, "table: unsupported tag ID: %d",
		    keywordID);
		    return WB_SSI_ERROR;
	    }
	}
    }
    /* keep on scanning until we found the --> sequence */
    while (keywordID != INCEND_ID);

    /* complain if we didn't find an oid tag */
    if (TableOID[0] == '\0')
    {
	fprintf (fp_out, "table: no OID specified");
	return WB_SSI_ERROR;
    }

    /* convert any _ characters in TableOID to . characters */
    for (cp = TableOID; (cp = strchr (cp, '_')) != NULL; cp++)
	*cp = '.';

    return wb_table_output (fp_in, fp_out, TableOID, BeginRow, EndRow, nRows,
			    where_str);
}


/**********************************************************************
 ** wb_parse_get_next_cmd
 *
 *  DESCRIPTION: Parse the tags behind a "get" or "next" command.
 *		 Scan the 'oid="a.b.c.d"' and the optional
 *		 'type="typename"' and/or 'case="value"' tags and values.
 *		 Get via a SNMP get or getnext message the value of the
 *		 MIB variable, optionaly convert this value with 'type'
 *		 and/or 'case', and write this value to file 'fp_out'.
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_get_next_cmd (FILE *fp_in, FILE *fp_out, int cmd)
{
    int	    keywordID;
    char    oidbuffer[MAX_PATH+1];
    int	    c;
    int	    do_get, did_get;
    char    *valp;
    char    *cp;
    unsigned int size;
    int	    error;
    int	    cvt_type;
    int	    offset;

    did_get = 0;
    valp = NULL;
    oidbuffer[1] = '\0';
    cvt_type = -1;
    do
    {
	do_get = 0;

	/* scan for the next 'oid', 'type', 'case', or '-->' tag */
	keywordID = keyword_term_scan (fp_in, NULL, get_tags, NULL);
	if (keywordID < MIN_KEYW_ID)
	{
	    fprintf (fp_out, "get: bad tag scan");
	    return (keywordID == EOF) ? EOF : WB_SSI_ERROR;
	}

	/* if we found an 'oid', 'type', or 'case' tag... */
	if (keywordID != INCEND_ID)
	{
	    /* then scan for the = and " behind these tags */
	    c = wb_scan_nonspace (fp_in);
	    if (c != '=')
	    {
		fprintf (fp_out, "get: expected =, found '%c'!", c);
		return WB_SSI_ERROR;
	    }

	    c = wb_scan_nonspace (fp_in);
	    if (c != '"')
	    {
		fprintf (fp_out, "get: expected \", found '%c'!", c);
		return WB_SSI_ERROR;
	    }
	}

	/* complain if we find a 'type','case', or '-->' tag before 'oid' */
	if (oidbuffer[1] == '\0' && keywordID != OIDTAG_ID)
	{
	    fprintf (fp_out, "get: first pass OID");
	    return WB_SSI_ERROR;
	}

	switch (keywordID)
	{
	    case OIDTAG_ID:
		/* scan the oid uptill the ", and store it in oidbuffer */
		if (wb_scan_string (fp_in, "\"", &oidbuffer[1],
				    MAX_PATH) < 0)
		{
		    fprintf (fp_out, "get: bad OID scan");
		    return WB_SSI_ERROR;
		}
		break;
	    case TYPETAG_ID:
		/* scan the type, complain if we didn't find one */
		keywordID = keyword_term_scan (fp_in, NULL, types, NULL);
		if (keywordID < MIN_KEYW_ID)
		{
		    fprintf (fp_out, "get: bad type value scan");
		    return (keywordID == EOF) ? EOF : WB_SSI_ERROR;
		}

		switch (keywordID)
		{
		    case INTEGERTYPE_ID:    cvt_type = 'i'; break;
		    case IPTYPE_ID:	    cvt_type = 'a'; break;
		    case MACTYPE_ID:	    cvt_type = 'm'; break;
		    case OIDTYPE_ID:	    cvt_type = 'o'; break;
		    case STRINGTYPE_ID:	    cvt_type = 's'; break;
		    case TIMETYPE_ID:	    cvt_type = 't'; break;
		    case HEXSTRTYPE_ID:	    cvt_type = 'x'; break;
		    default:
			fprintf (fp_out, "get: unsupported type ID: %d",
				    keywordID);
			return WB_SSI_ERROR;
		}

		/* scan for the " behind the type */
		c = wb_scan_nonspace (fp_in);
		if (c != '"')
		{
		    fprintf (fp_out, "get: \" expected!");
		    return WB_SSI_ERROR;
		}
		break;
	    case CASETAG_ID:
		do_get = 1;
		break;
	    case INCEND_ID:
		do_get = 1;
		break;
	    default:
		fprintf (fp_out, "get: unsupported tag ID: %d", keywordID);
		return WB_SSI_ERROR;
	}

	/* if we need to do a SNMP command, and we didn't do it yet.. */
	if (do_get == 1 && did_get == 0)
	{
	    /* then set the conversion type in the oid buffer.. */
	    if (cvt_type != -1 && cmd != COLVALCMD_ID)
	    {
		oidbuffer[0] = cvt_type;
		offset = 0;
	    }
	    else
		offset = 1;

	    /* and do a SNMP get or getnext command */
	    switch (cmd)
	    {
		case GETCMD_ID:
		    valp = wb_datastore_get (&oidbuffer[offset]);
		    break;
		case NEXTCMD_ID:
		    valp = wb_datastore_getnext (&oidbuffer[offset], &error);
		    break;
		case COLVALCMD_ID:
		case COLOIDCMD_ID:
		    valp = get_column_value (OidBuffer, TableOIDlen,
					    &oidbuffer[1], PrevRowIndex,
					    cvt_type, RowIndex);
		    if (valp == NULL)
			return WB_TAB_OUT_TABLE;
		    if (cmd == COLOIDCMD_ID)
		    {
			valp = wb_datastore_last_oid ();

			/* convert any '.' to '_' character */
			cp = oidbuffer;
			while (*valp)
			{
			    if (*valp == '.')
				*cp = '_';
			    else
				*cp = *valp;
			    valp++;
			    cp++;
			}
			*cp = '\0';
			valp = oidbuffer;
		    }
		    break;
	    }
	    did_get = 1;


	    /* now we have a value we can do the evaluation of a 'case' tag */
	    if (keywordID == CASETAG_ID)
	    {
		if ((error = wb_parse_case (fp_in, fp_out, valp,
					    cvt_type)) != 0)
		    return error;
	    }
	    else
	    {
		/* otherwise write the value directly to file 'fp_out' */
		size = strlen (valp);

		/* check if we write enough bytes */
		if (fwrite (valp, 1, size, fp_out) < size)
		    return WB_SSI_ERROR;
	    }
	}
    }
    /* keep on scanning until we found the --> sequence */
    while (keywordID != INCEND_ID);

    return 0;
}


/**********************************************************************
 ** wb_parse_getcmd
 *
 *  DESCRIPTION: Parse a "get" command.
 *		 Scan the 'oid="a.b.c.d"' tag+value fields and the
 *		 optional 'type="typename"' and/or 'case="value"' tags
 *		 and values.
 *		 Get via a SNMP get message the value of the MIB variable,
 *		 optionaly convert this value with 'type' and/or 'case',
 *		 and write this value to file 'fp_out'.
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_getcmd (FILE *fp_in, FILE *fp_out)
{
    return wb_parse_get_next_cmd (fp_in, fp_out, GETCMD_ID);
}

/**********************************************************************
 ** wb_parse_nextcmd
 *
 *  DESCRIPTION: Parse a "next" command.
 *		 Scan the 'oid="a.b.c.d"' tag+value fields and the
 *		 optional 'type="typename"' and/or 'case="value"' tags
 *		 and values.
 *		 Get via a SNMP getnext message the value of the MIB variable,
 *		 optionaly convert this value with 'type' and/or 'case',
 *		 and write this value to file 'fp_out'.
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_nextcmd (FILE *fp_in, FILE *fp_out)
{
    return wb_parse_get_next_cmd (fp_in, fp_out, NEXTCMD_ID);
}


/**********************************************************************
 ** wb_parse_col_val_cmd
 *
 *  DESCRIPTION: Parse a "col_val" command.
 *		 Scan the 'oid="a.b.c.d"' tag+value fields and the
 *		 optional 'type="typename"' and/or 'case="value"' tags
 *		 and values.
 *		 Get via a SNMP get message the value of the MIB variable,
 *		 optionaly convert this value with 'type' and/or 'case',
 *		 and write this value to file 'fp_out'.
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_col_val_cmd (FILE *fp_in, FILE *fp_out)
{
    if (TableOIDlen)
	return wb_parse_get_next_cmd (fp_in, fp_out, COLVALCMD_ID);

    fprintf (fp_out, "col_val: should use col_val only in table");
    return WB_SSI_ERROR;
}


/**********************************************************************
 ** wb_parse_col_oid_cmd
 *
 *  DESCRIPTION: Parse a "col_oid" command.
 *		 Scan the 'oid="a.b.c.d"' tag+value fields and the
 *		 optional 'type="typename"' and/or 'case="value"' tags
 *		 and values.
 *		 Get via a SNMP get message the value of the MIB variable,
 *		 optionaly convert this value with 'type' and/or 'case',
 *		 and write this value to file 'fp_out'.
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_col_oid_cmd (FILE *fp_in, FILE *fp_out)
{
    if (TableOIDlen)
	return wb_parse_get_next_cmd (fp_in, fp_out, COLOIDCMD_ID);

    fprintf (fp_out, "col_oid: should use col_oid only in table");
    return WB_SSI_ERROR;
}


/**********************************************************************
 ** wb_parse_inccmd
 *
 *  DESCRIPTION: Parse an "include" command.
 *		 Scan the 'file="filename"' tag and value and include
 *		 the file "filename" in the file 'fp_out'.
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_inccmd (FILE *fp_in, FILE *fp_out)
{
    int		c;
    FILE	*fp_in2;
    char	buffer[MAX_PATH+1];
    char	fname[MAX_PATH+1];
    char	*cp;
    int		error;
    int		cmdID;
    int		len;


    /* skip any leading blank chars */
    buffer[0] = wb_scan_nonspace (fp_in);

    if ((c = wb_scan_string (fp_in, "= \t\r\n", &buffer[1],
			    MAX_PATH-1)) < 0)
    {
	fprintf (fp_out, "include: no 'file=' found");
	return WB_SSI_ERROR;
    }

    if (strcmp (buffer, FILETAG) != 0)
    {
	fprintf (fp_out, "include: Unsupported tag '%s'", buffer);
	return WB_SSI_ERROR;
    }

    if (isspace (c))
	c = wb_scan_nonspace (fp_in);

    if (c != '=')
    {
	fprintf (fp_out, "include: = expected!");
	return WB_SSI_ERROR;
    }

    c = wb_scan_nonspace (fp_in);
    if (c != '"')
    {
	fprintf (fp_out, "include: \" expected!");
	return WB_SSI_ERROR;
    }

    if (wb_scan_string (fp_in, "\"", fname, MAX_PATH) < 0)
    {
	fprintf (fp_out, "include: Illegal filename");
	return WB_SSI_ERROR;
    }

    if (fname[0] == '/')
	strcpy (buffer, wb_get_rootdir ());
    else
    {
	if ((cp = strrchr (url_filepath, '/')) != NULL)
	{
	    cp++;
	    len = cp - url_filepath;
	    strncpy (buffer, url_filepath, len);
	    buffer[len] = '\0';
	}
	else
	    buffer[0] = '\0';
    }
    strcat (buffer, fname);

    while (wb_dedot (NULL, buffer))
	;

    if ((fp_in2 = fopen (buffer, "rb")) == NULL)
    {
		fprintf (fp_out, "include: can't open file '%s'", buffer);

		return WB_SSI_ERROR;
    }

    error = wb_parse_file (fp_in2, fp_out);
    fclose (fp_in2);

    if (error == 0)
    {
	cmdID = keyword_term_scan (fp_in, NULL, end_keyw, NULL);
	if (cmdID != INCEND_ID)
	{
	    fprintf (fp_out, "SSI: no end sequnce found");
	    error = (cmdID == EOF) ? EOF : WB_SSI_ERROR;
	}
    }
    return error;
}


/**********************************************************************
 ** wb_parse_cmd
 *
 *  DESCRIPTION: Parse a command from file 'fp_in' and write the output
 *		 to file 'fp_out'.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 0 if everything was OK, EOF if EOF was reached, or
 *		 an error code.
 *
***********************************************************************/
static int wb_parse_cmd (FILE *fp_in, FILE *fp_out)
{
    int	    cmdID;
    int	    error;

    cmdID = keyword_term_scan (fp_in, NULL, ssi_cmds, " \t\r\n");
    if (cmdID < MIN_KEYW_ID)
    {
	fprintf (fp_out, "SSI: no or illegal command found");
	return (cmdID == EOF) ? EOF : WB_SSI_ERROR;
    }

    switch (cmdID)
    {
	case GETCMD_ID:
	    error = wb_parse_getcmd (fp_in, fp_out);
	    break;
	case NEXTCMD_ID:
	    error = wb_parse_nextcmd (fp_in, fp_out);
	    break;
	case INCCMD_ID:
	    error = wb_parse_inccmd (fp_in, fp_out);
	    break;
	case COLVALCMD_ID:
	    error = wb_parse_col_val_cmd (fp_in, fp_out);
	    break;
	case COLOIDCMD_ID:
	    error = wb_parse_col_oid_cmd (fp_in, fp_out);
	    break;
	case TABLECMD_ID:
	    error = wb_parse_table_cmd (fp_in, fp_out);
	    break;
	case ENDTABLECMD_ID:
	    error = wb_parse_endtable_cmd (fp_in, fp_out);
	    break;
	default:
	    fprintf (fp_out, "SSI: Unsupported command ID: %d", cmdID);
	    error = WB_SSI_ERROR;
	    break;
    }

    return error;
}


/**********************************************************************
 ** wb_parse_file
 *
 *  DESCRIPTION: Copy the file 'fp_in' to 'fp_out'. If the file 'fp_in'
 *		 contains start and end include seqences, the string
 *		 inbetween is expanded in 'fp_out'.
 *  PARAMETERS:
 *
 *  RETURN:	 0 if there where no error, otherwise an error code.
 *
***********************************************************************/
static int wb_parse_file (FILE *fp_in, FILE *fp_out)
{
    int rc;

	rc = wb_scan_value (fp_in, fp_out, NULL);
    return (rc == EOF) ? 0 : rc;
}


/**********************************************************************
 ** server_side_include
 *
 *  DESCRIPTION: Expand the reqested file with server side include.
 *
 *  PARAMETERS:
 *
 *  RETURN:	 Nothing
 *
***********************************************************************/
void server_side_include (char *fname, FILE* fp_out)
{
    FILE *fp_in;

    if ((fp_in = fopen (fname, "r")) == NULL)
    {
		fprintf (stderr, "Can't open file %s\n", fname);
		return;
    }
    url_filepath = fname;

    if (wb_parse_file (fp_in, fp_out) != 0) {
		printf("wb_parse_file returned error\n");
	}

    fclose(fp_in);
}

/* Patches */
static char *wb_get_rootdir (void)
{
	char *doc_root;

	doc_root = getenv("DOCUMENT_ROOT");

	if (doc_root != NULL) {
		return doc_root;
	}
	else
	{
		return ".";
	}
}

static int *wb_dedot (char *out, char *in)
{
    return 0;
}



int main (int argc, char *argv[])
{
    FILE *fp_sock;
    int listenfd, connfd;
    char filename[MAX_PATH];
    int addr_len;

    if (argc > 1) {
	fprintf(stderr, "Usage: %s (no arguments)\n", argv[0]);
	exit(1);
    }
    
    /* initialise datastore module */
    wb_datastore_init();

    /* Ignore SIGPIPE. These can occur when the receiving process closes
     * the socket before we are finished
     */
    signal(SIGPIPE, sig_exit_handler);

    /* Catch INT to print some nice messages when SSI stops */
    signal(SIGINT, sig_exit_handler);

    /* Catch HUP to reread configuration (snmp comm. strings) */
    signal(SIGHUP, sig_hup_handler);

    aplog(EV_SNMP_INFO,"SSI module starting.",NULL);

    if ((listenfd = socket(AF_UNIX, SOCK_STREAM, 0)) == -1 ) {
	    perror("socket:");
	    exit(1);
    }

    servaddr.sun_family = AF_UNIX;
    strcpy(servaddr.sun_path, UNIX_DOMAIN_SOCKET);
    addr_len = sizeof(servaddr.sun_family) + strlen(servaddr.sun_path);

    /* Clean up stuff from previous runs */
    unlink(UNIX_DOMAIN_SOCKET);

    if (bind(listenfd, (struct sockaddr*) &servaddr, addr_len) == -1 ) {
	perror("bind:");
	exit(1);
    }

    if (listen(listenfd, 10) == -1 ) {
	perror("listen:");
	exit(1);
    }

    for ( ; ; ) {

	if ((connfd = accept(listenfd, (struct sockaddr*) &servaddr, &addr_len)) == -1 ) {

	    if (errno == EINTR) {
		/* Got a SIGHUP, continue */
		continue;
	    }

	    perror("accept:");
	    exit(1);
	}

	if (readline(connfd, &filename, MAX_PATH) == -1 ) {
	    perror("readline:");
	    exit(1);
	}

	/* Create a File stream using the connection socket */
	fp_sock = fdopen( connfd, "w" );

	server_side_include (filename, fp_sock);

	/* Close everything */
	fclose(fp_sock);
	close(connfd);

    }

    /* We should never reach this */
    fprintf(stderr, "Unexpected end of for-loop \n");
    exit(1);
}
