/*
 * socket.c -- DHCP server client/server socket creation
 *
 * Moreton Bay DHCP Server
 * Copyright (C) 1999 Matthew Ramsay <matthewr@moreton.com.au>
 *			Chris Trew <ctrew@moreton.com.au>
 *
 * Rewrite by Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>
#include <net/if.h>
#include <errno.h>
#include <features.h>
#if __GLIBC__ >=2 && __GLIBC_MINOR >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>
#endif


#include "debug.h"

int listen_socket(unsigned int ip, int port, char *inf)
{
    struct ifreq interface;
    char localname[IFNAMSIZ];
    char *primary_name;
    int fd;
    struct sockaddr_in addr;
    int n = 1;

    if ((fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
	DEBUG(LOG_ERR, "socket call failed: %s", sys_errlist[errno]);
	return -1;
    }

    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = ip;

    if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char *) &n, sizeof(n)) ==
	-1) {
	close(fd);
	return -1;
    }
    if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST, (char *) &n, sizeof(n)) ==
	-1) {
	close(fd);
	return -1;
    }

    /* check if it's an aliased interface */
    strncpy(localname, inf, sizeof(localname));
    primary_name = strtok(localname, ":");

    strncpy(interface.ifr_ifrn.ifrn_name, primary_name, IFNAMSIZ);
    if (setsockopt
	(fd, SOL_SOCKET, SO_BINDTODEVICE, (char *) &interface,
	 sizeof(interface)) < 0) {
	close(fd);
	return -1;
    }

    if (bind(fd, (struct sockaddr *) &addr, sizeof(struct sockaddr)) == -1) {
	close(fd);
	return -1;
    }

    return fd;
}


int raw_socket(int ifindex)
{
    int fd;
    struct sockaddr_ll sock;

    if ((fd = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP))) < 0) {
	DEBUG(LOG_ERR, "socket call failed: %s", sys_errlist[errno]);
	return -1;
    }

    sock.sll_family = AF_PACKET;
    sock.sll_protocol = htons(ETH_P_IP);
    sock.sll_ifindex = ifindex;
    if (bind(fd, (struct sockaddr *) &sock, sizeof(sock)) < 0) {
	DEBUG(LOG_ERR, "bind call failed: %s", sys_errlist[errno]);
	close(fd);
	return -1;
    }

    return fd;

}
