#include <stdio.h>
#include <time.h>

#ifdef WIN32

#include <winsock2.h>
#include <ws2tcpip.h>
#include <wsipx.h>

#define close(s)	closesocket(s)

#else

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#ifdef IPX
#include <netipx/ipx.h>
#endif /* IPX */
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h> /* close */

/* Winsock uses different structure field names */
#define sa_family sipx_family
#define sa_netnum sipx_network
#define sa_socket sipx_port
#define NSPROTO_SPXII	AF_IPX

#endif /* WIN32 */

#define WALL_VERSION	"2.1.0.0"

#define SUCCESS 0
#define WALL_ERROR   1

#define DEFAULT_PORT	  1500
#define DEFAULT_SUMTIME   5

#define MEBI		0x100000
#define MEGA		1000000

#define ETH_HEAD_SIZE	14
#define IP_HEAD_SIZE	20
#define TCP_HEAD_SIZE	20
#define UDP_HEAD_SIZE	8
#define LLC_HEAD_SIZE	8
#define IPX_HEAD_SIZE	30
#define SPX_HEAD_SIZE	14
#define MTU_SIZE	1500
#define IPX_DATA_SIZE	(MTU_SIZE-LLC_HEAD_SIZE-IPX_HEAD_SIZE-SPX_HEAD_SIZE) /* 1448 */
#define TCP_DATA_SIZE	(MTU_SIZE-IP_HEAD_SIZE-TCP_HEAD_SIZE) /* 1460 */
#define UDP_DATA_SIZE	(MTU_SIZE-IP_HEAD_SIZE-UDP_HEAD_SIZE) /* 1472 */

#define MCAST_INL_DISCOVERY  "224.0.1.43" // Intersil discover multicast address

int   fTime     = 0;
int   fUDP      = 0;
int   fIPX      = 0;
int   fDatagram = 0;
int   fEcho     = 0;
int   fReceiveIP = 0;
int   fPER      = 0;
int   fBurst	= 0;
int   fHeader	= 1;
int   BlockSize = 0;
int   SumTime   = DEFAULT_SUMTIME;
short PortNum   = DEFAULT_PORT;

char *buffer;
#ifdef IPV6
struct addrinfo *host_addresses = NULL;
#endif

print_usage(char *progname)
{
    printf ("%s, version: %s\n\n", progname, WALL_VERSION);
    printf ("Wait for TCP, UDP or IPX packets from a sending client and calculate the\n");
    printf ("troughput for the data received from this client\n");
    printf ("Usage: %s [options]\n", progname);
    printf ("Options:\n");
    printf ("    -b <size> : use block size <size> [%d TCP, %d UDP, %d IPX]\n", TCP_DATA_SIZE, UDP_DATA_SIZE, IPX_DATA_SIZE);
    printf ("    -e        : echo each packet back to originator\n");
    printf ("    -h        : ignore headers in throughput calculation\n");
    printf ("    -i        : use streaming IPX (=SPX) communication\n");
    printf ("    -p <port> : use port number <port> [%d]\n", DEFAULT_PORT);
    printf ("    -r        : print for each packet the address it was received on\n");
    printf ("    -s <secs> : calculate and print throughput each <secs> seconds [%d]\n", DEFAULT_SUMTIME);
    printf ("    -t        : print time stamp with each throughput value\n");
    printf ("    -u        : use UDP communication\n");
    printf ("    -x        : calculate the packet error rate.\n");
    printf ("    -y        : print info for each packet error burst\n");
}


/**********************************************************************
 *  handle_connection
 *
 *  DESCRIPTION: Receive packets from socket 'newSd' and depending on
 *               the command line options specified, echo the data back
 *               and print messages.
 *
 *  PARAMETERS:  newSd:    socket descriptor identifying a socket
 *
 *  RETURN:      WALL_ERROR in case of an error, otherwise it doesn't
 *               return.
 *
 **********************************************************************/
int handle_connection(int newSd)
{
    time_t cur_time, prev_time, diff_time;
    char *timestr;
    unsigned long nbytes;
    int head_size, data_size;
    int n, npackets, cliLen, recv_bytes;
    struct sockaddr cliAddr;
    unsigned long packet_nr, prev_packet_nr, lost_packets;

    if (fIPX)
    {
        head_size = ETH_HEAD_SIZE + LLC_HEAD_SIZE + IPX_HEAD_SIZE + SPX_HEAD_SIZE;
        data_size = IPX_DATA_SIZE;
    }
    else
    {
	if (fUDP)
	{
	    head_size = ETH_HEAD_SIZE + IP_HEAD_SIZE + UDP_HEAD_SIZE;
	    data_size = UDP_DATA_SIZE;
	}
	else
	{
	    head_size = ETH_HEAD_SIZE + IP_HEAD_SIZE + TCP_HEAD_SIZE;
	    data_size = TCP_DATA_SIZE;
	}
    }

    nbytes = 0;
    prev_packet_nr = 0;
    lost_packets = 0;
    prev_time = time (NULL);
    while(1)
    {
        /* read data from socket */
        if (fDatagram)
        {
            cliLen = sizeof(cliAddr);
            n = recv_bytes = recvfrom(newSd, buffer, BlockSize, 0, &cliAddr, &cliLen);
        }
        else
	{
	    recv_bytes = 0;
	    do
	    {
		n = recv(newSd, (buffer+recv_bytes), (BlockSize-recv_bytes), 0);
		recv_bytes += n;
	    }
	    while( n > 0 && recv_bytes < BlockSize);
	}

        if (n<0)
        {
             perror(" cannot receive data ");
             return WALL_ERROR;
        }
        if (n==0)
        {
            printf(" connection closed by client\n");
            close(newSd);
            return WALL_ERROR;
        }

        if (fEcho)
        {
	    if (fDatagram)
		n = sendto (newSd, buffer, recv_bytes, 0, &cliAddr, cliLen);
	    else
		n = send (newSd, buffer, recv_bytes, 0);
	}


        if (fReceiveIP)
            printf ("Received %d bytes on %s\n", n, buffer+4);
        else
        {
	    /* calculate the number of lost packets */
	    if (fPER)
	    {
		packet_nr = *((unsigned long*) buffer);
		if (packet_nr > (prev_packet_nr+1))
		{
		    lost_packets += packet_nr - prev_packet_nr +1;
		    if (fBurst)
		        printf("packet: %d, prev. packet: %d, lost packets: %d\n",
				packet_nr, prev_packet_nr, lost_packets);
		}
		prev_packet_nr = packet_nr;
	    }

            /* calculate the number of bytes for headers */
            if (fHeader)
            {
		npackets = recv_bytes / data_size;
		if (recv_bytes % data_size)
		    npackets++;
		recv_bytes += npackets * head_size;
	    }

            if (fEcho)
            	recv_bytes <<= 1;

            nbytes += recv_bytes;
            cur_time = time (NULL);
            diff_time = cur_time - prev_time;
            if (diff_time >= SumTime)
            {
                if (fTime)
                {
                    if ((timestr = ctime (&cur_time)) != NULL)
                        timestr[24] = '\0';
                    else
                        timestr = "*unknown*";
                    printf ("%s ", timestr);
                }
		if( !fPER )
                    printf ("%3.4f Mbits/sec\n", (nbytes * 8) / (double) diff_time / MEGA);
		else
                    printf ("%3.4f Mbits/sec packer error rate %d\n", (nbytes * 8) / (double) diff_time / MEGA, lost_packets);
                fflush (stdout);
                prev_time = cur_time;
                nbytes = 0;
		lost_packets=0;
            }
        }
    }
}


/**********************************************************************
 *  parse_options
 *
 *  DESCRIPTION: Parse the command line arguments and set correponding
 *               global variables.
 *
 *  PARAMETERS:  argc: number of arguments.
 *               argv: array with pointers to each argument.
 *
 *  RETURN:      Nothing, but will exit program in case of an error.
 *
 **********************************************************************/
void parse_options (int argc, char *argv[])
{
    int i, fIllOption;

    for (i = 1; i < argc; i++)
    {
        fIllOption = 0;
        if (argv[i][0] == '-')
        {
            switch (argv[i][1])
            {
                case 'b':
                    if (++i < argc)
                        BlockSize = atoi (argv[i]);
                    break;
                case 'e':
                    fEcho = 1;
                    break;
                case 'h':
                    fHeader = 0;
                    break;
                case 'i':
#ifdef IPX
                    fIPX = 1;
                    break;
#else
                    printf("This version of wallrx is not compiled for IPX !!!\n");
                    exit (1);
#endif /* IPX */
                case 'p':
                    if (++i < argc)
                        PortNum = atoi (argv[i]);
                    break;
                case 'r':
                    fReceiveIP = 1;
                    break;
                case 's':
                    if (++i < argc)
                        SumTime = atoi (argv[i]);
                    break;
                case 't':
                    fTime = 1;
                    break;
                case 'u':
                    fUDP = 1;
                    fDatagram = 1;
                    break;
		case 'x':
		    fPER = 1;
		    break;
		case 'y':
		    fBurst = 1;
		    break;
                case '?':
                    print_usage (argv[0]);
                    exit (0);
                    break;
                default:
                    fIllOption = 1;
                    break;
            }
        }
        if (fIllOption)
        {
            print_usage (argv[0]);
            printf ("\nIllegal option: '%s'\n", argv[i]);
            exit (1);
        }
    }
}


/**********************************************************************
 *  gethostaddresses
 *
 *  DESCRIPTION: Get all IP addresses of this host and store them in
 *               global link list 'host_addresses'
 *
 *  RETURN:      Nothing.
 *
 **********************************************************************/
void gethostaddresses()
{
#ifdef IPV6
    struct addrinfo hints;
    char hostname[256];

    gethostname(hostname, sizeof(hostname));
    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = PF_INET;
    hints.ai_socktype = SOCK_DGRAM;

    getaddrinfo(hostname, NULL, &hints, &host_addresses);
#endif /* IPV6 */
}


void wall_init ()
{
#ifdef WIN32
    WSADATA wsaData;
    int rc;

    rc = WSAStartup(MAKEWORD(2, 2), &wsaData);
    if (rc)
    {
        printf ("WSAStartup() failed, err: %d\n", rc);
        exit(2);
    }
#endif

    if (BlockSize <= 0)
    	BlockSize = fIPX ? IPX_DATA_SIZE : (fUDP ? UDP_DATA_SIZE : TCP_DATA_SIZE);

    if ((buffer = malloc (BlockSize)) == NULL)
    {
        printf ("Out of memory\n");
        exit (1);
    }
    gethostaddresses ();
}


int main (int argc, char *argv[])
{
    int i=1, sd, newSd, cliLen, AddrSize;
    struct sockaddr *cliAddr, *servAddr;
    struct sockaddr_in cliInAddr, servInAddr;
#ifdef IPX
    struct sockaddr_ipx cliIpxAddr, servIpxAddr;
#endif /* IPX */
    struct ip_mreq mreq;

    parse_options (argc, argv);
    wall_init ();

    /* create socket */
    if (!fIPX)
    {
	if (fUDP)
	    sd = socket(AF_INET, SOCK_DGRAM, 0);
	else
	    sd = socket(AF_INET, SOCK_STREAM, 0);
    }
#ifdef IPX
    else
        sd = socket(AF_IPX, SOCK_STREAM, NSPROTO_SPXII);
#endif /* IPX */

    if(sd<0)
    {
        perror("cannot open socket ");
        return WALL_ERROR;
    }

    setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (void *)&i, sizeof(i));

    /* bind server port to socket */
    if (fIPX)
    {
#ifdef IPX
	memset (&servIpxAddr, 0, sizeof(servIpxAddr));
	servIpxAddr.sa_family = AF_IPX;
	servIpxAddr.sa_socket = htons(PortNum);
	servAddr = (struct sockaddr *) &servIpxAddr;
	cliAddr = (struct sockaddr *) &cliIpxAddr;
	AddrSize = sizeof(servIpxAddr);
#endif
    }
    else
    {
	servInAddr.sin_family = AF_INET;
	servInAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	servInAddr.sin_port = htons(PortNum);
	servAddr = (struct sockaddr *) &servInAddr;
	cliAddr = (struct sockaddr *) &cliInAddr;
	AddrSize = sizeof(servInAddr);
    }

    if(bind(sd, servAddr, AddrSize)<0)
    {
        perror("cannot bind port ");
        return WALL_ERROR;
    }

    if (fUDP)
    {
	/*** For UDP join the Intersil discovery multicast group ***/
#ifdef IPV6
	struct addrinfo *host_addr;

	for (host_addr = host_addresses; host_addr != NULL;
	    host_addr = host_addr->ai_next)
	{
	    if (host_addr->ai_family == PF_INET)
	    {
		mreq.imr_interface.s_addr =
		  ((struct sockaddr_in *)host_addr->ai_addr)->sin_addr.s_addr;
#else
	{
	    {
		mreq.imr_interface.s_addr = INADDR_ANY;
#endif
		mreq.imr_multiaddr.s_addr = inet_addr(MCAST_INL_DISCOVERY);
		if (setsockopt (sd, IPPROTO_IP, IP_ADD_MEMBERSHIP,
				(char *) &mreq, sizeof(mreq)) < 0)
		    perror ("can't join multicast group ");
	    }
	}
    }

    if (!fDatagram)
	listen(sd,5);

    while(1)
    {
        printf("%s: waiting for data on port %s %u\n",argv[0],
                fIPX ? "IPX" : (fUDP ? "UDP" : "TCP"), PortNum);

        if (fDatagram)
            newSd = sd;
        else
        {
            cliLen = AddrSize;
            newSd = accept(sd, cliAddr, &cliLen);
            if(newSd<0)
            {
                perror("cannot accept connection ");
                return WALL_ERROR;
            }
        }
	handle_connection(newSd);
    }
}
