#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>


#ifdef WIN32

#include <winsock2.h>
#include <ws2tcpip.h>
#include <wsipx.h>

#define close(s)  closesocket(s)

#else

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#ifdef IPX
#include <netipx/ipx.h>
#endif /* IPX */
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h> /* close */

#define sa_family sipx_family
#define sa_netnum sipx_network
#define sa_nodenum sipx_node
#define sa_socket sipx_port
#define NSPROTO_SPXII	0
#define IPPROTO_TCP		SOL_TCP

#endif /* WIN32 */

#define WALL_VERSION           "2.1.0.0"
#define DEFAULT_PORT	  1500
#define DEFAULT_WAIT	  1000
#define DEFAULT_SUMTIME   5
#define MEBI              0x100000
#define MEGA              1000000
#define DATA_MSG          "When you can read this, it is NOT encrypted!!"

#define ETH_HEAD_SIZE	14
#define IP_HEAD_SIZE	20
#define TCP_HEAD_SIZE	20
#define UDP_HEAD_SIZE	8
#define LLC_HEAD_SIZE	8
#define IPX_HEAD_SIZE	30
#define SPX_HEAD_SIZE	14
#define MTU_SIZE	1500
#define IPX_DATA_SIZE	(MTU_SIZE-LLC_HEAD_SIZE-IPX_HEAD_SIZE-SPX_HEAD_SIZE) /* 1448 */
#define TCP_DATA_SIZE	(MTU_SIZE-IP_HEAD_SIZE-TCP_HEAD_SIZE) /* 1460 */
#define UDP_DATA_SIZE	(MTU_SIZE-IP_HEAD_SIZE-UDP_HEAD_SIZE) /* 1472 */

int   fTime     = 0;
int   fUDP      = 0;
int   fIPX      = 0;
int   fDatagram = 0;
int   fEcho     = 0;
long  tWait     = 0;
long  nBurst	= 0;
int   fAllNIC   = 0;
int   fReceiveAddr = 0;
int   fHeader	= 1;
int   BlockSize = 0;
long  nPackets  = -1;
int   SumTime   = DEFAULT_SUMTIME;
short PortNum   = DEFAULT_PORT;
char  *hostname = NULL;

char *sendbuffer, *recvbuffer;
#ifdef IPV6
struct addrinfo *host_addresses;
#endif

print_usage(char *progname)
{
    printf ("%s, version: %s\n\n", progname, WALL_VERSION);
    printf ("Send TCP, UDP or IPX packets to server <hostname>.\n");
    printf ("Usage: %s <hostname> [options]\n", progname);
    printf ("Options:\n");
    printf ("    -a        : send out multicast traffic on all NICs\n");
    printf ("    -b <size> : use block size <size> [%d TCP, %d UDP, %d IPX]\n", TCP_DATA_SIZE, UDP_DATA_SIZE, IPX_DATA_SIZE);
    printf ("    -e        : echo mode. Read back packets from server\n");
    printf ("    -h        : ignore headers in throughput calculation\n");
    printf ("    -i        : use streaming IPX (=SPX) communication\n");
    printf ("    -n <num>  : send <num> packets [continuous]\n");
    printf ("    -p <port> : use port number <port> [%d]\n", DEFAULT_PORT);
    printf ("    -r        : print for each received packet remote address\n");
    printf ("    -s <secs> : calculate and print throughput each <secs> seconds [%d]\n", DEFAULT_SUMTIME);
    printf ("    -t        : print time stamp with each throughput value\n");
    printf ("    -u        : use UDP communication\n");
    printf ("    -w <msecs>: wait <msecs> milli seconds between packets\n");
    printf ("    -z <num>  : number of packets per burst\n");
}


#ifdef IPX
/**********************************************************************
 *  ctoh
 *
 *  DESCRIPTION: Convert a hexadecimal character to its value.
 *
 *  PARAMETERS:  ch: Character to convert
 *
 *  RETURN:	 The hex value of character, or -1 if 'ch' isn't a hex
 *               character.
 *
 **********************************************************************/
int ctoh(char ch)
{
    if ( ch >= '0' && ch <= '9' )
	return ( ch - '0' );

    if ( ch >= 'A' && ch <= 'F' )
	return ( ch - 'A' + 0xA );

    if (ch >= 'a' && ch <= 'f' )
	return ( ch - 'a' + 0xA );

    return -1;
}


/**********************************************************************
 *  atoh
 *
 *  DESCRIPTION: Convert a sting to its hexadecimal value. Store the
 *               result in a buffer of 'cnt' bytes long.
 *
 *  PARAMETERS:  hex: The buffer to store the converted bytes.
 *               str: The string to convert.
 *               cnt: maximum number of bytes to convert.
 *
 *  RETURN:	 A pointer to the character in 'str' that did terminate
 *               the conversion.
 *
 **********************************************************************/
char *atoh(char *hex, char *str, int cnt)
{
    int i, lownibble, highnibble;
    char *cvtend;

    hex += cnt-1;
    for (i = 0; i < cnt*2 && ctoh (*(str+i)) != -1; i++)
    	;
    cvtend = str+i;
    if (i)
    {
	str += i-1;
	while (i-- > 0)
	{
	    lownibble = ctoh (*str--);
	    if (i--)
	    	highnibble = ctoh (*str--);
	    else
	    	highnibble = 0;
	    *hex-- = (highnibble << 4) + lownibble;
	}
    }
    return cvtend;
}

/**********************************************************************
 *  ipx_addr
 *
 *  DESCRIPTION: Convert an ASCII IPX address to its binary socket
 *               library specific representation. The IPX address must
 *               have the format 00112233.AABBCCDDEEFF
 *
 *  PARAMETERS:  sipx:     pointer to IPX socket address to store result.
 *               hostname: pointer to ASCII IPX address.
 *
 *  RETURN:      0 if the address could be correctly converted,
 *               1 otherwise.
 *
 **********************************************************************/
int ipx_addr (struct sockaddr_ipx *sipx, char *hostname)
{
    char *endhost, *seperator;

#ifdef WIN32
    seperator = atoh (sipx->sa_netnum, hostname, 4);
#else
    sipx->sa_netnum = strtol (hostname, &seperator, 16);
#endif
    if (*seperator == '.')
	endhost = atoh (sipx->sa_nodenum, seperator+1, 6);

    return (*seperator != '.' || *endhost != '\0');
}


/**********************************************************************
 *  ipx_ntoa
 *
 *  DESCRIPTION: Convert an IPX network address into its string
 *               representation (00112233.AABBCCDDEEFF)
 *
 *  PARAMETERS:  sipx: pointer to structure that represents an IPX
 *                     host address.
 *
 *  RETURN:      Pointer to a static buffer containing the text address.
 *
 **********************************************************************/
char *ipx_ntoa (struct sockaddr_ipx *sipx)
{
    static char str_addr[22];
    char *nonzero;

    sprintf (str_addr, "%02X%02X%02X%02X.%02X%02X%02X%02X%02X%02X",
#ifdef WIN32
	sipx->sa_netnum[0], sipx->sa_netnum[1], sipx->sa_netnum[2], sipx->sa_netnum[3],
#else
	(sipx->sa_netnum >> 24) & 0xFF, (sipx->sa_netnum >> 16) & 0xFF,
	(sipx->sa_netnum >> 8) & 0xFF, sipx->sa_netnum & 0xFF,
#endif
	((unsigned int) sipx->sa_nodenum[0]) & 0xFF,
	((unsigned int) sipx->sa_nodenum[1]) & 0xFF,
	((unsigned int) sipx->sa_nodenum[2]) & 0xFF,
	((unsigned int) sipx->sa_nodenum[3]) & 0xFF,
	((unsigned int) sipx->sa_nodenum[4]) & 0xFF,
	((unsigned int) sipx->sa_nodenum[5]) & 0xFF);
    for (nonzero = str_addr; *nonzero == '0'; nonzero++)
    	;
    if (*nonzero == '.')
    	nonzero--;
    return nonzero;
}
#endif /* IPX */

/**********************************************************************
 *  parse_options
 *
 *  DESCRIPTION: Parse the command line arguments and set correponding
 *               global variables.
 *
 *  PARAMETERS:  argc: number of arguments.
 *               argv: array with pointers to each argument.
 *
 *  RETURN:      Nothing, but will exit program in case of an error.
 *
 **********************************************************************/
void parse_options (int argc, char *argv[])
{
    int i, fIllOption;

    if(argc < 2)
    {
        print_usage (argv[0]);
        exit(1);
    }

    hostname = NULL;
    for (i = 1; i < argc; i++)
    {
        fIllOption = 0;
        if (argv[i][0] == '-')
        {
            switch (argv[i][1])
            {
                case 'a':
#ifdef IPV6
                    fAllNIC = 1;
                    break;
#else
		    printf("This version of walltx doesn't support all NICs option !!\n");
		    exit (1);
#endif /* IPV6 */
                case 'b':
                    if (++i < argc)
                        BlockSize = atoi (argv[i]);
                    break;
                case 'e':
                    fEcho = 1;
                    break;
                case 'h':
                    fHeader = 0;
                    break;
                case 'i':
#ifdef IPX
                    fIPX = 1;
                    break;
#else
		    printf("This version of walltx is not compiled for IPX !!!\n");
		    exit (1);
#endif
                case 'n':
                    if (++i < argc)
                        nPackets = atoi (argv[i]);
                    break;
                case 'p':
                    if (++i < argc)
                        PortNum = atoi (argv[i]);
                    break;
                case 'r':
                    fReceiveAddr = 1;
                    break;
                case 's':
                    if (++i < argc)
                        SumTime = atoi (argv[i]);
                    break;
                case 't':
                    fTime = 1;
                    break;
                case 'u':
                    fUDP = 1;
                    fDatagram = 1;
                    break;
                case 'w':
                    if (++i < argc)
                        tWait = atoi (argv[i]);
                    break;
		case 'z':
		    if (++i < argc)
			nBurst = atoi (argv[i]);
		    break;
                case '?':
                    print_usage (argv[0]);
                    exit (0);
                    break;
                default:
                    fIllOption = 1;
                    break;
            }
        }
        else
        {
            if (hostname)
                fIllOption = 1;
            else
                hostname = argv[i];
        }
        if (fIllOption)
        {
            print_usage (argv[0]);
            printf ("\nIllegal option: '%s'\n", argv[i]);
            exit (1);
        }
    }
    if (hostname == NULL)
    {
        print_usage (argv[0]);
        printf ("\nNo hostname specified!!\n");
        exit (1);
    }
}


/**********************************************************************
 *  gethostaddresses
 *
 *  DESCRIPTION: Get all IP addresses of this host and store them in
 *               global link list 'host_addresses'
 *
 *  RETURN:      Nothing.
 *
 **********************************************************************/
void gethostaddresses()
{
#ifdef IPV6
    struct addrinfo hints;
    char hostname[256];

    gethostname(hostname, sizeof(hostname));
    memset(&hints, 0, sizeof(hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = PF_INET;
    hints.ai_socktype = SOCK_DGRAM;

    getaddrinfo(hostname, NULL, &hints, &host_addresses);
#endif /* IPV6 */
}


void wall_init ()
{
#ifdef WIN32
    WSADATA wsaData;
    int rc;

    rc = WSAStartup(MAKEWORD(2, 2), &wsaData);
    if (rc)
    {
        printf ("WSAStartup() failed, err: %d\n", rc);
        exit(2);
    }
#endif

    if (BlockSize <= 0)
    	BlockSize = fIPX ? IPX_DATA_SIZE : (fUDP ? UDP_DATA_SIZE : TCP_DATA_SIZE);

    if ((sendbuffer = malloc (BlockSize)) == NULL ||
        (recvbuffer = malloc (BlockSize)) == NULL)
    {
        printf ("Out of memory\n");
        exit (1);
    }
    if (fReceiveAddr)
    	fEcho = 1;

    if ((fEcho || nBurst) && tWait == 0)
    	tWait = DEFAULT_WAIT;
    if (fAllNIC)
    	gethostaddresses ();
}


/**********************************************************************
 *  sendto_all_nics
 *
 *  DESCRIPTION: Send a (multicast) datagram out through all interfaces
 *               of this host.
 *
 *  PARAMETERS:  sd:    socket descriptor identifying a socket
 *               buf:   buffer containing the data to be transmitted.
 *               len:   the size of the data in buf.
 *               to:    pointer to the address of the target socket.
 *               tolen: size of the address in to.
 *
 *  RETURN:      The total number of bytes sent, or -1 in case of an
 *               error.
 *
 **********************************************************************/
int sendto_all_nics (int sd, char *buf, int len, struct sockaddr *to, int tolen)
{
#ifdef IPV6
    struct addrinfo *host_addr;
    int rc;

    for (rc = 1, host_addr = host_addresses; rc > 0 && host_addr != NULL;
	host_addr = host_addr->ai_next)
    {
        if (host_addr->ai_family == PF_INET)
	{
	    setsockopt (sd, IPPROTO_IP, IP_MULTICAST_IF,
	                (char *) &((struct sockaddr_in *)host_addr->ai_addr)->sin_addr,
	                sizeof(struct in_addr));
	    rc = sendto (sd, buf, len, 0, to, tolen);
	}
    }
    return rc;
#endif /* IPV6 */
}


/**********************************************************************
 *  handle_connection
 *
 *  DESCRIPTION: Send packets through socket 'sd' to target address
 *               'sendAddr'. Depending on the command line options
 *               specified, wait, read data echoed back and print
 *               messages.
 *
 *  PARAMETERS:  sd:       socket descriptor identifying a socket
 *               sendAddr: pointer to the address of the target socket.
 *               AddrSize: size of the address in sendAddr.
 *
 *  RETURN:      Nothing.
 *
 **********************************************************************/
void handle_connection (int sd, struct sockaddr *sendAddr, int AddrSize)
{
    struct sockaddr *servAddr;
    struct sockaddr_in servInAddr;
#ifdef IPX
    struct sockaddr_ipx servIpxAddr;
#endif
    struct timeval tv;
    fd_set socks;
    time_t cur_time, prev_time, diff_time;
    char *timestr;
    unsigned long nbytes;
    int head_size, data_size;
    int nfragments;
    int recv_packets;
    int PacketsSize;
    int i, n, rc, rbytes;
	unsigned long burst_nr;
    long *packet_nr;
    int WaitNow;


    if (fIPX)
    {
        head_size = ETH_HEAD_SIZE + LLC_HEAD_SIZE + IPX_HEAD_SIZE + SPX_HEAD_SIZE;
        data_size = IPX_DATA_SIZE;
    }
    else
    {
	if (fUDP)
	{
	    head_size = ETH_HEAD_SIZE + IP_HEAD_SIZE + UDP_HEAD_SIZE;
	    data_size = UDP_DATA_SIZE;
	}
	else
	{
	    head_size = ETH_HEAD_SIZE + IP_HEAD_SIZE + TCP_HEAD_SIZE;
	    data_size = TCP_DATA_SIZE;
	}
    }

    /* calculate the number of bytes for headers */
    if (fHeader)
    {
	nfragments = BlockSize / data_size;
	if (BlockSize % data_size)
	    nfragments++;
	PacketsSize = BlockSize + nfragments * head_size;
    }
    else
    	PacketsSize = BlockSize;

    if (fDatagram)
    {
#ifdef IPX
	if(fIPX)
	    servAddr = (struct sockaddr *) &servIpxAddr;
	else
#endif /* IPX */
	    servAddr = (struct sockaddr *) &servInAddr;
    }
    else
	servAddr = sendAddr;

    packet_nr = (long*)sendbuffer;
    *packet_nr = 0;
    burst_nr = nBurst;
    nbytes = 0;
    prev_time = time (NULL);
    for (i = 0; i != nPackets; i++)
    {
	(*packet_nr)++ ;
        if (fDatagram)
        {
	    if (fAllNIC)
		rc = sendto_all_nics (sd, sendbuffer, BlockSize, sendAddr, AddrSize);
            else
		rc = sendto (sd, sendbuffer, BlockSize, 0, sendAddr, AddrSize);
	}
        else
            rc = send (sd, sendbuffer, BlockSize, 0);

        if(rc<0)
        {
            perror("cannot send data ");
            close(sd);
            exit(1);
        }


	if (nBurst)
	{
	    burst_nr--;
	    if (burst_nr == 0)
	    {
		burst_nr = nBurst;
		WaitNow = 1;
	    }
	    else
		WaitNow = 0;
	}
	else
	    WaitNow = tWait;

        if (WaitNow || fEcho)
    	{
            tv.tv_sec = 0;
	    tv.tv_usec = tWait * 1000;

	    n = -1;
	    rbytes = 0;
	    recv_packets = 0;
            FD_ZERO (&socks);
	    FD_SET ((unsigned) sd, &socks);
            do
            {
		rc = select (1, &socks, NULL, NULL, &tv);
		if (rc == 1)
		{
		    if (fDatagram)
			n = recvfrom (sd, recvbuffer+rbytes, BlockSize-rbytes, 0,
			              servAddr, &AddrSize);
		    else
			n = recv (sd, recvbuffer+rbytes, BlockSize-rbytes, 0);
		    if (n > 0)
		    {
		    	if (fReceiveAddr)
#ifdef IPX
			if(fIPX)
			    printf ("Received %d bytes from %s\n", n,
				    ipx_ntoa ((struct sockaddr_ipx *) servAddr));
			else
#endif /* IPX */
			    printf ("Received %d bytes from %s\n", n,
				    inet_ntoa (((struct sockaddr_in *) servAddr)->sin_addr));
			rbytes += n;
			if (rbytes == BlockSize)
			{
			    recv_packets++;
			    rbytes = 0;
			}
		    }
		}
		else
		{
		    if (fEcho && recv_packets == 0)
		    	printf ("Timeout\n");
		}
	    }
	    while (rc == 1 && ((recv_packets == 0) || tWait));
	}

	/* be sure it will never end (if requested) */
        if (nPackets == -1)
            i = 0;

	nbytes += PacketsSize;
	if (fEcho && n != -1)
	    nbytes += PacketsSize;
	cur_time = time (NULL);
	diff_time = cur_time - prev_time;
	if (diff_time >= SumTime)
	{
	    if (fTime)
	    {
		if ((timestr = ctime (&cur_time)) != NULL)
		    timestr[24] = '\0';
		else
		    timestr = "*unknown*";
		printf ("%s ", timestr);
	    }
	    printf ("%3.4f Mbits/sec\n", (nbytes * 8) / (double) diff_time / MEGA);
	    fflush (stdout);
	    prev_time = cur_time;
	    nbytes = 0;
	}
    }
    close(sd);
}


int main (int argc, char *argv[])
{
    int sd;
    struct sockaddr *localAddr, *sendAddr;
    struct sockaddr_in localInAddr, sendInAddr;
#ifdef IPX
    struct sockaddr_ipx localIpxAddr, sendIpxAddr;
#endif /* IPX */
    struct hostent *h;
    int i, n, AddrSize;

    parse_options (argc, argv);
    wall_init ();

   /* set up the socket address we are going to send to */
    if (!fIPX)
    {
	h = gethostbyname(hostname);
	if(h==NULL)
	{
	    printf("%s: unknown host '%s'\n",argv[0],hostname);
	    exit(1);
	}
	sendInAddr.sin_family = h->h_addrtype;
	memcpy((char *) &sendInAddr.sin_addr.s_addr, h->h_addr_list[0], h->h_length);
	sendInAddr.sin_port = htons(PortNum);
	sendAddr = (struct sockaddr *) &sendInAddr;
    }
#ifdef IPX
    else
    {
	memset (&sendIpxAddr, 0, sizeof(sendIpxAddr));
	sendIpxAddr.sa_family = AF_IPX;
	sendIpxAddr.sa_socket = htons(PortNum);
	if (ipx_addr (&sendIpxAddr, hostname) != 0)
	{
	    printf("%s: illegal IPX hostname '%s'\n",argv[0],hostname);
	    exit(1);
	}
	sendAddr = (struct sockaddr *) &sendIpxAddr;
    }

    /* create socket */
    if (fIPX)
        sd = socket(AF_IPX, SOCK_STREAM, NSPROTO_SPXII);
    else
#endif /* IPX */
    {
	if (fUDP)
	    sd = socket(AF_INET, SOCK_DGRAM, 0);
	else
	{
	    if ((sd = socket(AF_INET, SOCK_STREAM, 0)) > 0)
	    {
		i = 1;
		setsockopt(sd, IPPROTO_TCP, TCP_NODELAY, (char *) &i, sizeof(i));
	    }
	}
    }
    if(sd<0)
    {
        perror("cannot open socket ");
        exit(1);
    }

    /* bind any local port number to the socket */
    if (!fIPX)
    {
	localInAddr.sin_family = AF_INET;
	localInAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	localInAddr.sin_port = htons(0);
	localAddr = (struct sockaddr *) &localInAddr;
	AddrSize = sizeof(localInAddr);
    }
#ifdef IPX
    else
    {
	memset (&localIpxAddr, 0, sizeof(localIpxAddr));
	localIpxAddr.sa_family = AF_IPX;
	localIpxAddr.sa_socket = htons(0);
	localAddr = (struct sockaddr *) &localIpxAddr;
	AddrSize = sizeof(localIpxAddr);
    }
#endif

    if (bind (sd, localAddr, AddrSize) < 0)
    {
        printf("%s: cannot bind socket to local %s port\n",argv[0],
        	fIPX ? "IPX" : (fUDP ? "UDP" : "TCP"));
        perror("error ");
        exit(1);
    }

    /* for connection oriented operation, make a connection with the server */
    if (!(fDatagram))
    {
        if (connect (sd, sendAddr, AddrSize) < 0)
        {
            perror("cannot connect ");
            exit(1);
        }
    }

    /* put the address we send to and a readable message in the buffer */
    strncpy (sendbuffer+4, fIPX ? hostname : inet_ntoa(sendInAddr.sin_addr), BlockSize-4);
    sendbuffer[BlockSize-1]='\0';
    n = strlen (DATA_MSG) + 1;
    for (i = strlen (sendbuffer+4) + 5; i < BlockSize; i += n)
	strncpy (&sendbuffer[i], DATA_MSG, BlockSize - i);
    sendbuffer[BlockSize-1]='\0';

    /* go for it */
    handle_connection (sd, sendAddr, AddrSize);
    return 0;
}
