#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>

#include <prismoids.h>

//#define DEBUG(x) x


void usage(void)
{
    fprintf(stderr,"\n");
    fprintf(stderr,"Usage: WDSConf [-a xx-xx-xx-xx-xx-xx] [-d xx-xx-xx-xx-xx-xx]\n");
    fprintf(stderr,"   Options:\n");
    fprintf(stderr,"     -a   : Add a WDS link to an other AP\n");
    fprintf(stderr,"     -d   : Delete a WDS link from this AP\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"   No arguments will list the currently availabe WDS links\n");
    fprintf(stderr,"\n");
}

/* Parse a MACaddress out of the arguments list */
int parse_network_address(u_char* out, char* in)
{
    int n;

    for ( n=0; n<6; n++ )
    {
	if ( !isxdigit(in[n*3]) || !isxdigit(in[n*3+1]) || (n<5 && in[n*3+2]!='-') ||
		 (n==5 && in[n*3+2]!='\0')
	   )
	{
	    fprintf(stderr,"syntax error in network address, use xx-xx-xx-xx-xx-xx\n");
	    return -2;
	}
	out[n] = (u_char)strtol(&in[n*3], NULL, 16);
    }

    return 0;
}

struct wds_link_msg {
    unsigned char mac[6];
    char name[32];
} __attribute__ ((packed));


/* Add a WDS link to the MVC */
int add_wds_link_for_address (u_char* MACAddr)
{
    struct wds_link_msg wlm;
    char cmd[64];
    
    memcpy(wlm.mac, MACAddr, 6);
    wlm.name[0] = 0;
    
    if ( set_prismoid("eth1", DOT11_OID_WDSLINKADD, &wlm, sizeof (struct wds_link_msg)) < 0 )
    {
        fprintf(stderr, "add_wds_link_for_address ioctl DOT11_OID_WDSLINKADD (%d) to wireless interface driver failed\n", DOT11_OID_WDSLINKADD);
        return(-1);
    } else {
        snprintf( cmd, 64, "brctl addif br0 %s", wlm.name );
        system( cmd ); 
        return 0;
    }
}

/* Del a WDS link from the MVC */
int del_wds_link_for_address (u_char* MACAddr)
{
    struct wds_link_msg wlm;
    char cmd[64];
    
    memcpy(wlm.mac, MACAddr, 6);
    wlm.name[0] = 0;
    
    if ( set_prismoid("eth1", DOT11_OID_WDSLINKREMOVE, &wlm, sizeof (struct wds_link_msg)) < 0 )
    {
	    fprintf(stderr, "del_wds_link_for_address ioctl DOT11_OID_WDSLINKREMOVE (%d) to wireless interface driver failed\n", DOT11_OID_WDSLINKREMOVE);
        return(-1);
    } else {
        snprintf( cmd, 64, "brctl delif br0 %s", wlm.name );
        system( cmd ); 
        return 0;
    }
}


int list_current_wdslinks(void)
{
    static unsigned int nofthislink = 1;
    unsigned long currentnofclients = 0;
    unsigned int i;
    struct obj_sta sta;

    fprintf (stderr, "Currently available WDS links:\n");

    for (i=0; i < (DOT11_OID_CLIENT2007 - DOT11_OID_CLIENT1); i++)
    {
    	if ( get_prismoid("eth1", (DOT11_OID_CLIENT1 + i), &sta, sizeof(struct obj_sta)) < 0)
    	{
    	    return -1;
    	}
    	
        if (sta.state == DOT11_STATE_WDS)
    	{
    	    fprintf (stderr, "LinkNr. %d: MACAddress %02X-%02X-%02X-%02X-%02X-%02X rssi: %d\n",
    			nofthislink, sta.address[0], sta.address[1], sta.address[2],
    			sta.address[3], sta.address[4], sta.address[5], sta.rssi);
    	    nofthislink++;
    	}
    }

    fprintf (stderr, "\n\n");
    return 0;
}

/*
 * Main function navigates through all arguments supplied
 * and execute the opperations for the wdsconfig
 * Returns: Zero is succesfull, <0 on error.
 */
int main(int argc,char *argv[])
{
    unsigned int i, iError=0;
    unsigned char MACAddress[6];

    fprintf(stderr,"WDSConf version 0.1 Copyrights Intersil, 2001\n\n");

    /* If no aguments supplied list the WDS link status currently in the MVC */
    if (argc <= 1)
    {
	iError = list_current_wdslinks();
	return iError;
    }

    for ( i=1;i<argc&&!iError;i++ )
    {
	if ( argv[i][0] == '-' && argc>i+1 )
	{
	    switch ( argv[i][1] )
	    {
	    /* Add a WDS link supplied with a MAC addess. */
	    case 'a':
		iError = parse_network_address(MACAddress, argv[++i]);

		fprintf(stderr, "- Add WDSLink for MAC address: %02X-%02X-%02X-%02X-%02X-%02X\n",
				MACAddress[0], MACAddress[1], MACAddress[2],
				MACAddress[3], MACAddress[4], MACAddress[5]
			   );
		iError = add_wds_link_for_address(MACAddress);
		break;

	    /* Delete a WDS link supplied with a MAC addess. */
	    case 'd':
		iError = parse_network_address(MACAddress, argv[++i]);

		fprintf(stderr, "- Delete WDSLink for MAC address: %02X-%02X-%02X-%02X-%02X-%02X\n",
				MACAddress[0], MACAddress[1], MACAddress[2],
				MACAddress[3], MACAddress[4], MACAddress[5]
			   );
		iError = del_wds_link_for_address(MACAddress);
		break;

	    default:
		fprintf(stderr,"Unknown arguments supplied: -%s\n", argv[i]);
		usage();
		iError = -1;
	    }
	}
    }
    return iError;
}

