/*  $Header$
 *
 *  webupgrade - CGI to support upload of firmware
 *  Copyright (C) 2001, 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "aplog.h"
#include "inl_misc.h"

#define BUFFER_SIZE 80
#define LOCAL_BUF_SIZE 128
#define ERROR_FILE "/_standardfiles/cgi/webupgrade/uploaderror.shtml"
#define READY_FILE "/_standardfiles/cgi/webupgrade/uploadcomplete.shtml"
#define IMAGE_FILE "/tftp/upload.tar.gz"

//#define DEBUG

/*****************************************************************************
 * NAME:
 *      start_imginstall
 *
 * DESCRIPTION:
 *
 *      Starts the imginstall process with the webserver document-root as
 *      arguments. imginstall needs this info to know where to find the
 *      status messages it has to copy to /tmp
 *
 * PARAMETERS:
 *      none
 *
 * RETURN:
 *      none
 *****************************************************************************/
static void start_imginstall(void)
{
    pid_t pid;
    char *cmd[10];

    cmd[0] = "/bin/imginstall";
    cmd[1] = IMAGE_FILE;
    cmd[2] = NULL;

    if ( (pid = vfork()) == 0 )
    {
         fclose(stdout);
         fclose(stdin);
         execv(cmd[0], cmd);
    }

}


/*****************************************************************************
 * NAME:
 *      print_file
 *
 * DESCRIPTION:
 *
 *      Prints out file with the givin input filename (relative to doc-root)
 *
 * PARAMETERS:
 *      filename   - IN  Name of the file (relative to doc-root) to print
 *
 * RETURN:
 *      -1 if the file does not exist
 *****************************************************************************/
static int print_file(char* filename)
{
    char f_name[255];
    FILE* in_file;
    char line[(BUFFER_SIZE*2) + 1];
    char* doc_root;
    
    doc_root = getenv("DOCUMENT_ROOT");
    if (!doc_root) {
	    doc_root = "/usr/www/html";
    }
    
    strcpy(f_name, doc_root);
    strcat(f_name, "/");
    strcat(f_name, filename);

    in_file = fopen(f_name, "r");

    if( in_file == NULL ) {
        return -1;
    }

    while( fgets(line, BUFFER_SIZE*2, in_file) != NULL ) {
        printf("%s", line);
    }

    fclose(in_file);

    return 0;

}

int main(int argc, char * argv[])
{
    FILE* out_file;
    char buffer[BUFFER_SIZE];
    char mime_tag[BUFFER_SIZE+2];
    int mime_tag_length;
    int total_matched;

    int read_p;
    int write_p;
    int fd;
    int image_fd;
    int c;
    long no_bytes;

    /* print cgi-header */
    printf("Content-type: text/html%c%c",10,10);

    if (snmp_get_aplockstate() != 0) {
	    
	if( print_file(ERROR_FILE) == -1 ) {
            printf("<HTML>Upload failed, AP is locked</HTML>");
        }
	
        exit(0);
    }

#ifdef DEBUG
    if ((out_file = fopen(IMAGE_FILE, "w")) == NULL ) {
        printf("Upload failed, error opening output file\n");
        exit(1);
    }

    while( (c = getc(stdin)) != EOF) {
        putc(c, out_file);
    }

    fclose(out_file);
    exit(1);
#endif

    /* Read mime_tag line and store it */
    if (fgets(buffer, BUFFER_SIZE - 1 , stdin) == NULL) {
        if( print_file(ERROR_FILE) == -1 ) {
            printf("<HTML>Upload failed, couldn't read mime-tag</HTML>");
        }
        exit(0);
    }

    /* cut trailing end-of-lines */
    while((strlen(buffer) > 0)
    		&& ((buffer[strlen(buffer)-1] == 0x0D) || (buffer[strlen(buffer)-1] == 0x0A))) {
	buffer[strlen(buffer)-1] = 0x00;
    }

    sprintf(mime_tag, "%c%c", 13, 10);
    strcat(mime_tag, buffer);

    mime_tag_length = strlen(mime_tag);

    /*
     * Read lines until we find the 'name="filename"' somewhere. This indicates
     * the start of our upgrade image
     */
    while(strstr(buffer, "name=\"filename\"") == NULL) {
        if (fgets(buffer, BUFFER_SIZE-1, stdin) == NULL) {
            if( print_file(ERROR_FILE) == -1 ) {
                printf("<HTML>Upload failed, name=\"filename\" not found</HTML>");
            }
            exit(1);
        }
    }

    /* Read until we find HEX 0D0A0D0A */

    /* The first two bytes of the sequence could already be in 'buffer' */
    if ((buffer[strlen(buffer)-2] == 0x0D) && (buffer[strlen(buffer)-1] == 0x0A)) {
    	total_matched = 2;
    }
    else {
    	total_matched = 0;
    }

    while( (total_matched < 4 ) && ((c = getc(stdin)) != EOF) ) {
        switch(c) {
        case 0x0D:
            switch(total_matched) {
                case 0:
                case 2:
                    total_matched++;
                    break;
                case 1:
                    /* found 0D0D */
                    break;
                case 3:
                    /* found 0D0A0D0D */
                    total_matched = 1;
                    break;
                default:
                    printf("<HTML>Internal error in 'webupgrade'</HTML>");
                    break;
            }
            break;
        case 0x0A:
            switch(total_matched) {
                case 0:
                    /* found 0A without 0D in front of it */
                case 1:
                case 3:
                    total_matched++;
                    break;
                case 2:
                    /* found 0D0A0A */
                    total_matched=0;
                    break;
                default:
                    printf("<HTML>Internal error in 'webupgrade'</HTML>");
                    break;
            }
            break;
        default:
            total_matched = 0;
            break;
        }
    }

    if (total_matched != 4) {
	printf("ERROR !!, didn't find 0D0A0D0A\n");
	exit(0);
    }

    /* open the outfile exclusive. This prevents concurrent uploads */
    image_fd = open( IMAGE_FILE , O_WRONLY | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
    if (image_fd < 0) {
	if (print_file(ERROR_FILE) == -1 ) {
		printf("<HTML>Conflicting upgrades, upgrade already in progress</HTML>");
	}
	exit(0);
    }
    
    /*
     * Open our output file
     */
    fd = open(getenv("CGI_POST_TMPFILE") , O_WRONLY );
    if( fd < 0 ) {
        if( print_file(ERROR_FILE) == -1 ) {
		printf("<HTML>Couldn't open POST temp file</HTML>");
	}
	exit(0);
    }

    out_file = fdopen( fd, "w" );

    if (out_file == NULL ) {
        if( print_file(ERROR_FILE) == -1 ) {
            printf("<HTML>Upload failed, couldn't open output file.</HTML>");
        }
        exit(0);
    }

    /* Let the fun begin */
    read_p = 0;
    write_p = 0;
    total_matched = 0;
    no_bytes = 0;

    while( (c = getc(stdin)) != EOF) {

        /* store byte */
        buffer[read_p] = c;

        /* Try to match with mime_tag */
        if ( buffer[read_p] == mime_tag[total_matched] ) {

                total_matched++;

                if (total_matched == mime_tag_length) {
                    /* OK, we're finished */
                    break; /* while */
                }
        }
        else {
            if (total_matched > 0) {
                /* The current read byte doesn't match with mime_tag, write out
                 * first byte of buffer and look if there is anything left to
                 * match in the buffer
                 */
                int match_p;

                /* reset total_matched */
                total_matched = 0;

                while ( (write_p != read_p) && (total_matched == 0) ) {
		    no_bytes++;
                    putc(buffer[write_p], out_file);
                    write_p = (write_p + 1) % BUFFER_SIZE;

                    match_p = write_p;

                    while( (buffer[match_p] == mime_tag[total_matched])
                            && (match_p != read_p) ) {

                        total_matched++;

                        match_p = (match_p+1) % BUFFER_SIZE;

                    }

                    if ( match_p != read_p ) {
                        /* matched failed, reset total_matched */
                        total_matched=0;
                    }
                }

            }
        }


        read_p = (read_p + 1) % BUFFER_SIZE;

        if (total_matched == 0) {
            /* no match of (part of) mime_tag, write bytes to destination file */
            while ( write_p != read_p ) {
		no_bytes++;
                putc(buffer[write_p], out_file);
                write_p = (write_p + 1) % BUFFER_SIZE;
            }

        }
    }

    if (no_bytes == 0) {
	    /* no file received */
	    fclose(out_file);
	    unlink(IMAGE_FILE);
	    
	    if( print_file(ERROR_FILE) == -1 ) {
		    printf("<HTML>Upload failed, empty or non-existing file uploaded.</HTML>");
	    }
	    exit(0);
    }
		  
    /* Truncate the output file */
    if (ftruncate(fileno(out_file), no_bytes) == -1) {
	printf("Error, couldn't truncate file (errno=%i)\n", -1);
	exit(0);
    };

    /* OK, finished, close output file */
    fclose(out_file);
    
    /* Close lock file and unlink it */
    close(image_fd);
    unlink(IMAGE_FILE);

    /* rename this tempfile to a imginstall file */
    if (rename(getenv("CGI_POST_TMPFILE"), IMAGE_FILE) == -1) {
	printf("Error, couldn't rename temporary file\n");
	exit(0);
    }

    /* print file if it exists, print some info otherwise */
    if( print_file(READY_FILE) == -1 ) {
        printf("<HTML><HEAD><meta http-equiv=\"Refresh\" content=\"5; url=/index.html\"></HEAD><BODY>File upload succeeded, starting flash installer</BODY></HTML>");
    }

    /* Start image install */
    start_imginstall();

    exit(0);

}
