/*  $Header: /projects/cvsroot/Broadcom/intersil-1.0.1.0/debug/acld.c,v 1.1 2003/06/11 09:11:17 maxyin Exp $
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*  Modified from catchoids.c to catch WLAN trap to do ACL/MAC filtering 
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>
#include <unistd.h>
#include <netinet/in.h>

#include "prismoids.h"

#define max(v1, v2) (v1 > v2) ? v1 : v2

struct element{
 unsigned char id;
 unsigned char length;
 unsigned char info[1]; 
};

struct assoc_req{
 unsigned short capability;
 unsigned short listen_interval;
 struct element	elements[1];
};

struct macaddr{
 unsigned char addr[6];
};

#define FILE_ACLLST	"/tmp/acl.lst"
/*
	acl.lst's file structure:
	first line:
		number of MAC address entry
	then:
		each line lists one MAC address
*/
static struct macaddr *acl_array=NULL;
static int acl_entries=0;

int check_acllist(struct macaddr *sta_addr){
 int i;
 struct macaddr *mac_p;
 int isAllowed=0;
 
 	if(acl_entries == 0)	return 1; // always allowed
 	mac_p = acl_array;
 	for(i=0 ; i<acl_entries ; i++, mac_p++){
		if(memcmp(sta_addr, mac_p, sizeof(struct macaddr))==0){
			isAllowed = 1;
			break;
		}
	}
	return isAllowed;
}

void read_acllist(void){
 FILE *file_acllst = fopen(FILE_ACLLST, "r");
 int i=0;
 int int_addr[6];
 char buf[64];
 	if(file_acllst == NULL){
		goto ALL_PASS;
	}
/*	if(fscanf(file_acllst, "%d", &acl_entries)==EOF){
		goto ALL_PASS;
	}
*/
	if(fgets(buf, 64, file_acllst)==NULL){
		goto ALL_PASS;
	}
	acl_entries = atoi(buf);
	/*debug code*/
	printf("acld: %d entry(ies) in %s\n", acl_entries, FILE_ACLLST);
	/* debug code */
	if(acl_array != NULL)	free(acl_array);
	acl_array = (struct macaddr*)malloc(sizeof(struct macaddr)*acl_entries);
	if(acl_array == NULL){
		printf("acld: cannot allocate memory to hold ACL list\n");
		goto ALL_PASS;
	}
	
	for(i=0 ; i<acl_entries ; i++){
		if(fgets(buf, 64, file_acllst)==NULL){
			acl_entries = i;
			break;
		}
		sscanf(buf, "%X:%X:%X:%X:%X:%X\n",
			&int_addr[0],&int_addr[1],&int_addr[2],
			&int_addr[3],&int_addr[4],&int_addr[5] );
		acl_array[i].addr[0] = (unsigned char)(int_addr[0]&0xFF);
		acl_array[i].addr[1] = (unsigned char)(int_addr[1]&0xFF);
		acl_array[i].addr[2] = (unsigned char)(int_addr[2]&0xFF);
		acl_array[i].addr[3] = (unsigned char)(int_addr[3]&0xFF);
		acl_array[i].addr[4] = (unsigned char)(int_addr[4]&0xFF);
		acl_array[i].addr[5] = (unsigned char)(int_addr[5]&0xFF);
		/* debug code */
		printf("acld: %02X:%02X:%02X:%02X:%02X:%02X\n",
			acl_array[i].addr[0],acl_array[i].addr[1],acl_array[i].addr[2],
			acl_array[i].addr[3],acl_array[i].addr[4],acl_array[i].addr[5]);
		/* debug code */
		
	}
	
	return;
ALL_PASS:
	acl_entries = 0;
	return;
}

int main(int argc, char *argv[])
{
	fd_set fdset;
    //__u32 trapgroup = TRAPGRP_ETH_TRAPS | TRAPGRP_WLAN_TRAPS | TRAPGRP_BLOBDEV_TRAPS;
    __u32 trapgroup = TRAPGRP_WLAN_TRAPS;
	struct timeval tv;
	int trapfd, maxfd;
	char *ifname;
	long oid_val;
	struct obj_mlmeex *mlme_data;
    	int retval = 0;
	int deny_any=0;
	struct islnetlink_handle islnetlink_h;
	struct assoc_req *assoc_data;
	struct macaddr sta_addr;
	
	if(argc<2){
		printf("Usage:acld ifname [(0|1) for deny_any]\n");
		return -1;
	}
	else{
		ifname = argv[1];
		if(argc == 3){
			deny_any = atoi(argv[2]);
		}
	}
	
	read_acllist();
	
	if ( ( trapfd = prismtraps_open(&islnetlink_h, trapgroup ) ) < 0 ) {
        	printf("prismtraps_open failed\n");
	        return -1;
	}
	// Enter MLME extended level
	oid_val = DOT11_MLME_EXTENDED;
	if(set_prismoid(ifname, DOT11_OID_MLMEAUTOLEVEL, &oid_val, 4)!=0){
		printf("acld: cannot go into extended MLME level!!!\n");
		return -1;
	}
	
	while(1) {
		maxfd = 0;
		FD_ZERO(&fdset);

		/* Timeout 1 second */
		tv.tv_sec = 1;
		tv.tv_usec = 0;

		/* monitor the trapfd descriptor */
		FD_SET(trapfd, &fdset);
		maxfd = max(maxfd, trapfd);

		retval = select(maxfd+1, &fdset, NULL, NULL, &tv);

		if(retval < 0) {
			printf("select failed\n");
			continue;
		}
		
		if(FD_ISSET(trapfd, &fdset)) {
		 unsigned int oid;
		 unsigned int len;
		 unsigned int devid;
		 int i;
		 char *string;
            
            		retval = 
			  prismtraps_decode(&islnetlink_h, &oid, &devid, 
			  	(char **)&mlme_data, &len);
			if(retval < 0) {
				printf("prismtraps_decode failed\n");
				continue;
			}
			/* debug code 
			printf("acld: received trap:\n");
			printf(" oid      : 0x%x\n", oid);
			printf(" devid    : %d\n", devid);
			printf(" MLME data:\n");
			printf("    MAC   : %02x:%02x:%02x:%02x:%02x:%02x\n",
				mlme_data->address[0]&0xff,
				mlme_data->address[1]&0xff,
				mlme_data->address[2]&0xff,
				mlme_data->address[3]&0xff,
				mlme_data->address[4]&0xff,
				mlme_data->address[5]&0xff);
			printf("    id    : %d\n", mlme_data->id);
			printf("    state : %d\n", mlme_data->state);
			printf("    code  : %d\n", mlme_data->code);
			printf("    extended data size: %d\n", mlme_data->size);
			printf("    data:\n");
			for(i=0 ; i<mlme_data->size ; i++){
				printf(" %02x ", mlme_data->data[i]&0xFF);
			}
			printf("\n");
			if(oid == DOT11_OID_ASSOCIATEEX && 
				mlme_data->state==DOT11_STATE_ASSOCING){
				assoc_data = (struct assoc_req*)mlme_data->data;
				string = (char*) malloc(assoc_data->elements[0].length + 1);
				memset(string, 0, assoc_data->elements[0].length + 1);
				memcpy(string, assoc_data->elements[0].info, 
					assoc_data->elements[0].length);
				printf("    element id: %d\n", assoc_data->elements[0].id);
				printf("    element len: %d\n", assoc_data->elements[0].length);
				printf("    SSID  : %s\n", string);
				free(string);
			}
			debug code */
			
			/*check MAC address list*/
			memcpy(sta_addr.addr, mlme_data->address, 6);
			if(!check_acllist(&sta_addr)){
/*				mlme_data->state = DOT11_STATE_NONE;
				mlme_data->code = DOT11_SC_UNSPECIFIED;
				set_prismoid(ifname, oid-7, mlme_data, sizeof(struct obj_mlme));
*/
				goto NEXT;
			}
			
			/*  */
			switch(oid){
			 case DOT11_OID_AUTHENTICATEEX:
			 	if(mlme_data->state==DOT11_STATE_AUTHING){
					mlme_data->state = DOT11_STATE_NONE;
					mlme_data->code = DOT11_SC_SUCCESSFUL;
					set_prismoid(ifname, DOT11_OID_AUTHENTICATE, mlme_data, sizeof(struct obj_mlme));
				}
				break;
			 case DOT11_OID_ASSOCIATEEX:
				if(mlme_data->state==DOT11_STATE_ASSOCING){
				  	assoc_data = (struct assoc_req*)mlme_data->data;
					if(deny_any){//if deny any is ON
					    char buf[128];
					    	memset(buf, 0, 128);
					    	get_prismoid(ifname, DOT11_OID_SSID, buf, 127);
						
						if(memcmp(assoc_data->elements[0].info,
							buf+1, buf[0])!=0){
							printf("acld: STA doesn't specify my SSID\n");
							
							/* send failure back */
							mlme_data->state = DOT11_STATE_NONE;
							mlme_data->code = DOT11_SC_CANNOT_HANDLE_ASSOCS;
							set_prismoid(ifname, DOT11_OID_ASSOCIATE, mlme_data, 
								sizeof(struct obj_mlme));
							break; // not allowed
						}
					}
					mlme_data->state = DOT11_STATE_NONE;
					mlme_data->code = DOT11_SC_SUCCESSFUL;
					set_prismoid(ifname, DOT11_OID_ASSOCIATE, mlme_data, 
						sizeof(struct obj_mlme));
				}
				break;
			 case DOT11_OID_REASSOCIATEEX:
			 	break;
			}
NEXT:
			free(mlme_data);
		}
	}
out:
	prismtraps_close(&islnetlink_h);
	return retval;
}

