/*  Modified from acld.c to catch WLAN trap to do determine if this device is allowed
*   to associate to a certain AP.
*
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>
#include <unistd.h>
#include <netinet/in.h>
#include <signal.h>

#include "prismoids.h"
#include "datastore_if.h"
#include "assocd.h"

#define max(v1, v2) (v1 > v2) ? v1 : v2
#define DEBUG 1

struct element{
	unsigned char id;
	unsigned char length;
	unsigned char info[1]; 
}__attribute__((packed));;

struct assoc_req{
	unsigned short capability;
	unsigned short listen_interval;
	struct element	elements[1];
}__attribute__((packed));;

struct macaddr{
	unsigned char addr[6];
}__attribute__((packed));;

struct probe_resp{
	unsigned long long timestamp;
	unsigned short binterval;
	unsigned short capability;
	struct element elements[1];
}__attribute__((packed));

char probed_ssid[33];
char probed_channel;
//unsigned char probed_bssid[6];
char probed_country[4];
unsigned long authtype=0;

char allowed_ssid[33];
int allowed_channels[] = ASSOCD_ALLOWED_CHANNELS ;
int allowed_bands[2] = ASSOCD_ALLOWED_BANDS;

#define MAX_PROBE_RESULTS   6
struct probe_result{
	char ssid[33];
	unsigned char channel;
	unsigned char bssid[6];
	char country[4];
};
struct probe_result probe_results[MAX_PROBE_RESULTS];
int probe_index = 0;

static assocd_insert_probe_result(struct probe_result *pr)
{int i;
	for(i=0 ; i<MAX_PROBE_RESULTS ; i++){
		if(memcmp(&probe_results[i], pr, sizeof(struct probe_result))==0){
			// found the same one, return
			return;
		}
	}
	memcpy(&probe_results[probe_index++], pr, sizeof(struct probe_result));
	probe_index %= MAX_PROBE_RESULTS;
}

static void	assocd_check_datastore();
//#define DUMP_PROBE 1
static void assocd_parse_probe(const char *probed_bssid, void *data, int len)
{
	struct probe_resp *probe = (struct probe_resp*)data;
	struct element *e;
	int  i;
	struct probe_result pr;
#ifdef DUMP_PROBE	
	printf("Probe response dump:(from %02x:%02x:%02x:%02x:%02x:%02x)\n",
		   probed_bssid[0],probed_bssid[1],probed_bssid[2],
		   probed_bssid[3],probed_bssid[4],probed_bssid[5]);
	printf("  Timestamp: %lu\n", probe->timestamp);
	printf("  Beacon interval: %f\n", (float)(probe->binterval)/1000.0);
	printf("  Capability: 0x%04x\n", probe->capability);
	printf("  Information element:\n");
#endif
	//memset(probe_results, 0, sizeof(struct probe_result)*MAX_PROBE_RESULTS);
	//probe_index = 0;
	memcpy(pr.bssid, probed_bssid, 6);
	//probed_channel = 0;
	pr.channel = 0;
	len -= 12;
	e = probe->elements;
	while(len > 0){
		if(e->id == 0){ // SSID
			memset(pr.ssid/*probed_ssid*/, 0, 33);
			memcpy(pr.ssid/*probed_ssid*/, e->info, e->length);
		}
		else if(e->id == 3){//Channel
			//probed_channel = e->info[0];
			pr.channel = e->info[0];
		}
		else if(e->id == 7){
			/*
			probed_country[0] = e->info[0];
			probed_country[1] = e->info[1];
			probed_country[2] = e->info[2];
			probed_country[3] = '\0';
			 */
			pr.country[0] = e->info[0];
			pr.country[1] = e->info[1];
			pr.country[2] = e->info[2];
			pr.country[3] = '\0';
		}
		
#ifdef DUMP_PROBE
		printf("    Tag number: %d %s\n", e->id, 
			   (e->id==0)?"(SSID)":((e->id==3)?"(Channel)":((e->id==7)?"(Country info)":"")) );
		printf("    Tag length: %d\n", e->length);
		if(e->id == 0){
			printf("    SSID: %s\n\n", /*probed_ssid*/pr.ssid);
		}
		else if(e->id == 3){
			printf("    Channel: %d\n\n", e->info[0]);
		}
		else if(e->id == 7){
			printf("    Country: %s\n", /*probed_country*/pr.country);
			printf("    Channels: %d - %d\n", e->info[3], e->info[4]);
			printf("    Power: 0x%02x\n\n", e->info[5]);
		}
		else{
			printf("    Tag data:");
			for(i=0 ; i<e->length ; i++){
				printf(" %02x", e->info[i]);
			}
			printf("\n\n");
		}
#endif
		len -= (e->length+2);
		e = (struct element*)(((unsigned char*)e) + e->length+2);
	}
	
	// Save into probe_results array
	assocd_insert_probe_result(&pr);
	/*
	memcpy(&probe_results[probe_index], &pr, sizeof(struct probe_result));
	probe_index++;
	probe_index%=MAX_PROBE_RESULTS;
	*/
}

static int assocd_check_allow(const char *probed_bssid){
	int rc = 0, i, pindex=0;
	// OK, parse over, let's determine the primitives
	
	assocd_check_datastore();
	for(pindex=0 ; pindex<MAX_PROBE_RESULTS ; pindex ++){
		rc = 0;
		// first, check BSSID
#ifdef DEBUG
		printf("Auth BSSID: %02x:%02x:%02x:%02x:%02x:%02x, desired BSSID: %02x:%02x:%02x:%02x:%02x:%02x\n",
			   probed_bssid[0],probed_bssid[1],probed_bssid[2],
			   probed_bssid[3],probed_bssid[4],probed_bssid[5],
			   probe_results[pindex].bssid[0],probe_results[pindex].bssid[1],
			   probe_results[pindex].bssid[2],probe_results[pindex].bssid[3],
			   probe_results[pindex].bssid[4],probe_results[pindex].bssid[5]);
#endif
		if(memcmp(probed_bssid, probe_results[pindex].bssid, 6)==0){
			rc ++;
		}
		else{
#ifdef DEBUG
			printf(" BSSID not matched\n");
#endif
			continue;
		}
		// check if SSID matched
#ifdef DEBUG
		printf("%s(): probed ssid: %s, allowed_ssid: %s\n", __PRETTY_FUNCTION__,
			   /*probed_ssid*/probe_results[pindex].ssid, allowed_ssid);
#endif
		if(strncmp(/*probed_ssid*/probe_results[pindex].ssid, allowed_ssid, 
			   ( (strlen(/*probed_ssid*/probe_results[pindex].ssid)>33)?33:
				 strlen(/*probed_ssid*/probe_results[pindex].ssid) ) )==0)
			rc ++;
		else{
#ifdef DEBUG
			printf(" SSID not matched\n");
#endif
			continue;
		}
		// then, channel
		for(i=0 ; i<sizeof(allowed_channels)/4 ; i++){
			if(/*probed_channel*/probe_results[pindex].channel == allowed_channels[i]){
				rc ++;
#ifdef DEBUG
				printf(" Matched channel : %d\n", /*probed_channel*/probe_results[pindex].channel);
#endif
				break;
			}
		}
		// second, G-only or A-only
		if(/*probed_channel*/probe_results[pindex].channel>14 && allowed_bands[1]){
#ifdef DEBUG
			printf(" Matched A band\n");
#endif
			rc++;
		}
		else if(/*probed_channel*/probe_results[pindex].channel<=14 && allowed_bands[0]){
#ifdef DEBUG
			printf(" Matched G band\n");
#endif
			rc++;
		}
		printf("%s(): rc: %d\n", __PRETTY_FUNCTION__, rc);
		if(rc==4)   return 1; // four rules are passed
	}
	return 0;
}

#define OID_ROAMPREFERENCE ".1.3.6.1.4.1.937.2.2.2.1.1.27.3"
#define OID_SSID ".1.2.840.10036.1.1.1.9.3"
#define OID_AUTHTYPE_OS ".1.2.840.10036.1.2.1.3.3.1"
#define OID_AUTHTYPE_SK ".1.2.840.10036.1.2.1.3.3.2"
static void assocd_check_datastore()
{
	static char oid[64];
    static char data[128];
    int data_len = sizeof(data);
    int type, retval, i;
	static char community[128];
	unsigned int profile;
	
	strcpy(oid, OID_ROAMPREFERENCE) ;
    memset(data, 0, (data_len=sizeof(data)));
	strcpy(community, "public"); // only for reading
    type = T_OCTETSTR;
    retval = datastore_getoid(oid, community, strlen(community), data, &data_len, &type);
    if(retval == R_NOERR){
		i = atoi(data);
#ifdef DEBUG
		printf("%s(): Roamming preference: ", __PRETTY_FUNCTION__);
		switch(i){
			case 0: // none
				printf("none\n");
				break;
			case 1: // prefer G
				printf("prefer G\n");
				break;
			case 2: // prefer A
				printf("prefer A\n");
				break;
			case 3: // G only
				printf("G only\n");
				break;
			case 4: // A only
				printf("A only\n");
				break;
			default:
				printf("%d, unknown ?!\n", i);
				break;
		}
#endif
		switch(i){
			default:
			case 0: // none
			case 1: // prefer G
			case 2: // prefer A
				allowed_bands[0]=1;
				allowed_bands[1]=1;
				get_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
#if 0
				if(profile!=1){
					profile=2; // mixed
					set_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
				}
#endif
				break;
			case 3: // G only
				allowed_bands[0]=1;
				allowed_bands[1]=0;
				get_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
#if 1
				if(profile!=3){
					profile = 3; // G only
					set_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
				}
#endif
				break;
			case 4: // A only
				allowed_bands[0]=0;
				allowed_bands[1]=1;
				get_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
#if 1
				if(profile!=6){
					profile = 6; // A only
					set_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
				}
#endif
				break;
		}
    }
	
	strcpy(oid, OID_SSID) ;
    memset(data, 0, (data_len=sizeof(data)));
	memset(allowed_ssid, 0, sizeof(allowed_ssid));
	strcpy(community, "public"); // only for reading
    type = T_OCTETSTR;
    retval = datastore_getoid(oid, community, strlen(community), data, &data_len, &type);
    if(retval == R_NOERR){
		memcpy(allowed_ssid, data, 33);
#if DEBUG
		printf("%s(): allowed SSID: %s\n", __PRETTY_FUNCTION__, allowed_ssid);
#endif
	}
}

static void assocd_datastore_authtype(){
	static char oid[64];
    static char data[128];
    int data_len = sizeof(data);
    int type, retval, i;
	static char community[128];
	
	strcpy(oid, OID_AUTHTYPE_OS) ;
    memset(data, 0, (data_len=sizeof(data)));
	strcpy(community, "public"); // only for reading
    type = T_OCTETSTR;
    retval = datastore_getoid(oid, community, strlen(community), data, &data_len, &type);
    if(retval == R_NOERR){
		authtype = atoi(data);
	}
	strcpy(oid, OID_AUTHTYPE_SK) ;
    memset(data, 0, (data_len=sizeof(data)));
	strcpy(community, "public"); // only for reading
    type = T_OCTETSTR;
    retval = datastore_getoid(oid, community, strlen(community), data, &data_len, &type);
    if(retval == R_NOERR){
		authtype |= atoi(data)<<1;
	}
#ifdef DEBUG
	printf("%s(): authtype: %d\n", __PRETTY_FUNCTION__, authtype);
#endif
}

static char *ifname;
void sighandler_alarm(int signo)
{
	long oid_val = DOT11_MLME_EXTENDED;
	if(set_prismoid(ifname, DOT11_OID_MLMEAUTOLEVEL, &oid_val, 4)!=0){
		printf("assocd: cannot go into extended MLME level!!!\n");
	}
}

int main(int argc, char *argv[])
{
	fd_set fdset;
    //__u32 trapgroup = TRAPGRP_ETH_TRAPS | TRAPGRP_WLAN_TRAPS | TRAPGRP_BLOBDEV_TRAPS;
    __u32 trapgroup = TRAPGRP_WLAN_TRAPS;
	struct timeval tv;
	int trapfd, maxfd, i;
	long oid_val;
	struct obj_mlmeex *mlme_data;
	int retval = 0;
	struct islnetlink_handle islnetlink_h;
	int value = 0;
	struct assoc_req *assoc_data;
	char *string;
	int fail_count=0, assoc_count = 0;
	struct obj_attachment attach;
	int privacy;
		
	if(argc<2){
		printf("Usage:assocd ifname\n");
		return -1;
	}
	else{
		ifname = argv[1];
	}
	
	signal(SIGALRM, sighandler_alarm);
	
	// clear probe_results
	memset(probe_results, 0, sizeof(struct probe_result)*MAX_PROBE_RESULTS);
	
	if ( ( trapfd = prismtraps_open(&islnetlink_h, trapgroup ) ) < 0 ) {
		printf("prismtraps_open failed\n");
		return -1;
	}
	
	assocd_check_datastore(); // collect datastore settings
	assocd_datastore_authtype(); 
	
#if 1
	// Enter MLME extended level
	oid_val = DOT11_MLME_EXTENDED;
	if(set_prismoid(ifname, DOT11_OID_MLMEAUTOLEVEL, &oid_val, 4)!=0){
		printf("assocd: cannot go into extended MLME level!!!\n");
		return -1;
	}
#endif
	while(1) {
		maxfd = 0;
		FD_ZERO(&fdset);
		
		/* Timeout 1 second */
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		
		/* monitor the trapfd descriptor */
		FD_SET(trapfd, &fdset);
		maxfd = max(maxfd, trapfd);
		
		retval = select(maxfd+1, &fdset, NULL, NULL, &tv);
		
		if(retval < 0) {
			printf("select failed\n");
			continue;
		}
		
		if(FD_ISSET(trapfd, &fdset)) {
			unsigned int oid;
			unsigned int len;
			unsigned int devid;
            
			retval = 
				prismtraps_decode(&islnetlink_h, &oid, &devid, 
								  (char **)&mlme_data, &len);
			if(retval < 0) {
				printf("prismtraps_decode failed\n");
				continue;
			}
#ifdef DEBUG_1
			/* debug code */
			printf("assocd: received trap:\n");
			printf(" oid      : 0x%x\n", oid);
			printf(" devid    : %d\n", devid);
			printf(" MLME data:\n");
			printf("    MAC   : %02x:%02x:%02x:%02x:%02x:%02x\n",
				   mlme_data->address[0]&0xff,
				   mlme_data->address[1]&0xff,
				   mlme_data->address[2]&0xff,
				   mlme_data->address[3]&0xff,
				   mlme_data->address[4]&0xff,
				   mlme_data->address[5]&0xff);
			printf("    id    : %d\n", mlme_data->id);
			printf("    state : %d\n", mlme_data->state);
			printf("    code  : %d\n", mlme_data->code);
			printf("    extended data size: %d\n", mlme_data->size);
			printf("    data:\n");
			for(i=0 ; i<mlme_data->size ; i++){
				printf(" %02x ", mlme_data->data[i]&0xFF);
			}
			printf("\n");
			if(oid == DOT11_OID_ASSOCIATEEX && 
			   mlme_data->state==DOT11_STATE_ASSOCING){
				assoc_data = (struct assoc_req*)mlme_data->data;
				string = (char*) malloc(assoc_data->elements[0].length + 1);
				memset(string, 0, assoc_data->elements[0].length + 1);
				memcpy(string, assoc_data->elements[0].info, 
					   assoc_data->elements[0].length);
				printf("    element id: %d\n", assoc_data->elements[0].id);
				printf("    element len: %d\n", assoc_data->elements[0].length);
				printf("    SSID  : %s\n", string);
				free(string);
			}
			/*debug code */
#endif			
			/*  */
			switch(oid){
				case DOT11_OID_PROBE:
					//memcpy(probed_bssid, mlme_data->address, 6);
					assocd_parse_probe(mlme_data->address, mlme_data->data, mlme_data->size);
					break;
				case DOT11_OID_AUTHENTICATEEX:
					//system("killall -STOP cybersupplicant");
#ifdef DEBUG
					printf("Received DOT11_OID_AUTHENTICATEEX, id: %d state: %d code: %d\n",
						   mlme_data->id, mlme_data->state, mlme_data->code);
#endif
					assocd_datastore_authtype();
					if(mlme_data->code !=0){
						if((mlme_data->code == 13 || mlme_data->code == 15) && authtype==3){
							get_prismoid(ifname, DOT11_OID_AUTHENABLE, &privacy, 4);
							if(privacy == 3 || privacy == 2){
								privacy = 1;
								set_prismoid(ifname, DOT11_OID_AUTHENABLE, &privacy, 4);
							}
							else if(privacy == 1){
								privacy = 2;
								set_prismoid(ifname, DOT11_OID_AUTHENABLE, &privacy, 4);
							}
						}
						break;
					}
					if(	assocd_check_allow(mlme_data->address)>0){
						// OK, this AP is allowed
						printf("  AP is allowed, send ASSOC...\n");
						mlme_data->state = DOT11_STATE_ASSOCING;
						mlme_data->code = DOT11_SC_SUCCESSFUL;
						set_prismoid(ifname, DOT11_OID_ASSOCIATEEX, mlme_data, sizeof(struct obj_mlme));
					//	system("killall -CONT cybersupplicant");
					}
					else{
#ifdef DEBUG
						printf("BSSID or conditions not match, deauthenticate it \n");
#endif
						fail_count ++;
						mlme_data->state = DOT11_STATE_AUTH;
						mlme_data->code = DOT11_RC_PREV_AUTH_NOT_VALID;
						//set_prismoid(ifname, DOT11_OID_DEAUTHENTICATEEX, mlme_data, sizeof(struct obj_mlme));
						get_prismoid(ifname, GEN_OID_LINKSTATE, &value, 4);
						if(value || fail_count>3){// well, MVC still on, turn it down
							system("/bin/ifconfig eth1 down");
							sleep(1);
							system("/bin/ifconfig eth1 up");
							fail_count = 0;
						}
						if(fail_count == 2)
							system("/bin/killall -USR1 sysd");
					}
					memset(probe_results, 0, sizeof(struct probe_result)*MAX_PROBE_RESULTS);
					break;
				case DOT11_OID_REASSOCIATEEX:
#ifdef DEBUG
					printf("Received DOT11_OID_REASSOCIATEEX, id: %d state: %d code: %d\n",
						   mlme_data->id, mlme_data->state, mlme_data->code);
#endif
					break;
				case DOT11_OID_ASSOCIATEEX:
#ifdef DEBUG
					printf("Received DOT11_OID_ASSOCIATEEX,bssid: %02x:%02x:%02x:%02x:%02x:%02x, id: %d state: %d code: %d\n",
						   mlme_data->address[0],mlme_data->address[1],mlme_data->address[2],
						   mlme_data->address[3],mlme_data->address[4],mlme_data->address[5],
						   mlme_data->id, mlme_data->state, mlme_data->code);
#endif
					if(assoc_count>3 && mlme_data->state == 2 && mlme_data->code==18){
						if(allowed_bands[0] == 1 && allowed_bands[1] == 1){
							// both bands are allowed
							int profile;
							get_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
							if(profile == DOT11_PROFILE_A_ONLY){
								profile = 2; // set to G-mixed
								set_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
							}
							else{
								profile = 6; // set to A-only
								set_prismoid("eth1", DOT11_OID_PROFILES, &profile, 4);
							}
						}
						assoc_count = 0;
					}
						assoc_count ++;
					break;
				case GEN_OID_LINKSTATE:
					get_prismoid(ifname, GEN_OID_LINKSTATE, &value, 4);
					printf("Link state changed to %s\n", value==0?"off":"on");
#if 0
					if(value == 0 ){
				//		system("killall -STOP cybersupplicant");
						// check if currently this device is under adhoc mode
						get_prismoid(ifname, DOT11_OID_BSSTYPE, &value, 4);
						if(value == DOT11_BSSTYPE_IBSS) // not to control IBSS mode
							break;
						// Link state is down, trig MLME auto level to 2
						value = DOT11_MLME_EXTENDED;
						set_prismoid(ifname, DOT11_OID_MLMEAUTOLEVEL, &value, 4);
					}
#endif
					break;
			}
NEXT:
				free(mlme_data);
		}
	}
out:
		prismtraps_close(&islnetlink_h);
	return retval;
}

