/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <time.h>
#include <stdlib.h>
#include "alarm.h"

struct alarm_t {
	void            (*func)(void);
	time_t          expiretime;
	struct alarm_t* next;
};


static struct alarm_t* alarmlist = NULL;


void alarm_register(void (*func)(void), time_t expire) {
	
	struct alarm_t* new_alarm;
	
	new_alarm = (struct alarm_t*) malloc(sizeof(struct alarm_t));
	
	new_alarm->func = func;
	new_alarm->expiretime = expire;
	new_alarm->next = alarmlist;
	
	alarmlist = new_alarm;
	
}

void alarm_scheduler_check() {
	
	struct alarm_t* alarm, *prev_alarm, *next_alarm;
	
	alarm = alarmlist;
	prev_alarm = NULL;
	while(alarm) {
		next_alarm = alarm->next;
		if(alarm->expiretime <= 0) {
			/* call function */
			alarm->func();
			
			/* remove alarm from list */
			if (prev_alarm) {
				prev_alarm->next = alarm->next;
			}
			else {
				alarmlist = alarm->next;
			}
				
			/* destroy entry */
			free(alarm);
		}		
		else {
			alarm->expiretime--;
			prev_alarm = alarm;
		}
		
		
		
		alarm = next_alarm;
	}

}
