/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
#include <string.h>
#include <stdio.h>

#include "aplock.h"
#include "object.h"
#include "objdata.h"
#include "assert.h"
#include "handler.h"
#include "configfile.h"
#include "objflag.h"

#define DEFAULT_READ_COMMUNITY  "public"
#define DEFAULT_WRITE_COMMUNITY "private"

#define APLOCK_OID       ".1.3.6.1.4.1.937.2.1.2.5.0"
#define READ_COMM_OID    ".1.3.6.1.4.1.937.2.1.2.1.0"
#define WRITE_COMM_OID   ".1.3.6.1.4.1.937.2.1.2.2.0"

extern int snmp_lock_state;

const char aplock_cfg[] = "/var/etc/aplock.conf";

static int write_aplock_file(int locked)
{
	FILE *fp;

	fp = fopen(aplock_cfg,"w");
	if(!fp)
		return -1;
	
	fprintf(fp,"%d\n", locked);
	fclose(fp);

	return 0;
}

void aplock_nolock(void)
{
	write_aplock_file(0);
}

void aplock(void)
{
	snmp_lock_state = SNMP_LOCKSTATE_LOCKED;
	write_aplock_file(1);
}

void apunlock(void)
{
	struct object_t *obj;

	snmp_lock_state = SNMP_LOCKSTATE_UNLOCKED;
	
	/* reset aplock OID */
	obj = object_get(APLOCK_OID);
	assert(obj);
	object_set_int(obj, 2); /* unlocked */
	/* mark object dirty */
	object_flag_set(obj, OF_DIRTY);
	/* trigger set handler */
	handler_call(APLOCK_OID, H_SET, 3);
#if 0
	/* Stardate 10312.17 Max Yin
		It shouldn't change the community/password
	*/		
	/* reset read community string */
	obj = object_get(READ_COMM_OID);
	assert(obj);
	object_set_string(obj, DEFAULT_READ_COMMUNITY, strlen(DEFAULT_READ_COMMUNITY));
	/* mark object dirty */
	object_flag_set(obj, OF_DIRTY);
	/* trigger set handler */
	handler_call(READ_COMM_OID, H_SET, 3);
	
	/* reset write community string */
	obj = object_get(WRITE_COMM_OID);
	assert(obj);
	object_set_string(obj, DEFAULT_WRITE_COMMUNITY, strlen(DEFAULT_WRITE_COMMUNITY));
	/* mark object dirty */
	object_flag_set(obj, OF_DIRTY);
	/* trigger set handler */
	handler_call(WRITE_COMM_OID, H_SET, 3);
#endif	
	/* schedule writing of configfile */
	configfile_schedule_write();

	write_aplock_file(0);
}

