/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "debug.h"

#define color_default             "\x1b[0m"
#define color_black               "\x1b[30m"
#define color_red                 "\x1b[31m"
#define color_green               "\x1b[32m"
#define color_yellow              "\x1b[33m"
#define color_blue                "\x1b[34m"
#define color_purple              "\x1b[35m"
#define color_cyan                "\x1b[36m"
#define color_white               "\x1b[37m"
#define color_brightblack         "\x1b[01;30m"
#define color_brightred           "\x1b[01;31m"
#define color_brightgreen         "\x1b[01;32m"
#define color_brightyellow        "\x1b[01;33m"
#define color_brightblue          "\x1b[01;34m"
#define color_brightmagenta       "\x1b[01;35m"
#define color_brightcyan          "\x1b[01;36m"
#define color_brightwhite         "\x1b[01;37m"

struct dbginfo_t {
	long flag;
	char opt;
	char *label;
	char *color;
};

struct dbginfo_t dbginfo[] =  {
	{DBG_NETWORK, 	'n',	"net",	color_brightblack},
	{DBG_OBJECT,	'o',	"obj",	color_brightblue},
	{DBG_MODULE,	'm',	"mod",	color_white},
	{DBG_TABLE,	't',	"tbl",	color_brightgreen},
	{DBG_CFGFILE,	'c',	"cfg",	color_green},
	{DBG_JOURNAL,	'j',	"jrn",	color_brightmagenta},
	{DBG_RANGE,	'r',	"rng",	color_brightgreen},
	{DBG_HANDLER,	'h',	"hnd",	color_cyan},
	{DBG_SCHEMA,	's',	"sch",	color_purple},
	{0,		0,	NULL,	NULL},
};


long debugflags = 0;
int debugcolor = 0;

void dbg_initflags(char *flagstr)
{
	char *p;
	int i;
	int negate=0;
	
	printf("DEBUGFLAGS: %s\n", flagstr);

	for(p=flagstr; *p; p++) {
		if(*p=='+') negate=0;
		if(*p=='-') negate=1;

		for(i=0; dbginfo[i].flag; i++) {
			if((*p == dbginfo[i].opt) || (*p == 'A')) {
				if(!negate) 
					debugflags |= dbginfo[i].flag;
				else
					debugflags &= ~dbginfo[i].flag;
			}
		}
	}

}

void dbg_showflags(void)
{
	int i;

	printf("Possible debug flags :\n\n");
	for(i=0; dbginfo[i].flag; i++) {
		printf("      %c : %s\n", dbginfo[i].opt, dbginfo[i].label);
	}
	printf("      A : Enable all of the above\n\n");
}


void dbg(int flag, char *fmt, ...)
{
	va_list ap;
	char buf[256];
	int i;
	
	if(debugflags & flag) {
	
		va_start(ap, fmt);
		vsnprintf(buf, sizeof(buf), fmt, ap);
		va_end(ap);

		for(i=0; dbginfo[i].flag; i++) {
			if(flag == dbginfo[i].flag) {
				if(debugcolor) fprintf(stderr, "%s", dbginfo[i].color);
				fprintf(stderr, "%s: %s\n", dbginfo[i].label, buf);
				if(debugcolor) fprintf(stderr, "%s", color_default);
				break;
			}
		}
	}
}



