/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <ctype.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>

#include "types.h"
#include "schema.h"
#include "object.h"
#include "socket.h"
#include "handler.h"
#include "modules.h"
#include "journal.h"
#include "configfile.h"
#include "misc.h"
#include "debug.h"
#include "table.h"
#include "objdata.h"
#include "alarm.h"
#include "aplock.h"

extern void prism_init(void);
int stdin_handle(void);
int read_cmdline_args(int argc, char **argv);

        
int udpport = 1122;
int interactive=0;
int running = 1;
int unlock = 0;

char schema_file[] = "/etc/schema";
char default_cfgfile[] = "/etc/defaults/datastore.dat";
char user_cfgfile[] = "/usr/etc/datastore.dat";

static void datastore_shutdown(int a)
{
	running = 0;
}

static void datastore_unlock(int a)
{
	unlock = 1;
}

static void datastore_debug_on(int a)
{
	static int debug_on = B_FALSE;
	FILE* fp;
	char buf[80];
	
	fp = fopen("/tmp/ds_debug", "r");
	if(!fp) {
		if(!debug_on) {
			dbg_initflags("A");
			debug_on = B_TRUE;
		}
		else {
			dbg_initflags("-A");
			debug_on = B_FALSE;
		}
	}
	else {
		if(fgets(buf, sizeof(buf), fp)) {
			dbg_initflags(buf);
		}
		else {
			dbg_initflags("-A");
		}
		fclose(fp);
	}
}

int main(int argc, char **argv)
{
	int sockfd;
	struct timeval tv;
	fd_set fds;
	int r;
	FILE * tmp_fd;

	set_malloc_mmap_size(32000);

	read_cmdline_args(argc, argv);

	sockfd = socket_open(udpport);

	schema_inittree();
	object_inittree();

	if(schema_read(schema_file) < 0) {
		exit(1);
	}

	if(configfile_read(default_cfgfile, user_cfgfile) < 0) {
		exit(1);
	}
	
	prism_init();
	
	modules_init();

	printf("\n");
	printf("Accepting connections on port %d\n", udpport);
	if(interactive) {
		printf("Accepting commands on console\n");
	}
	printf("\n");

	signal(SIGTERM, datastore_shutdown);
	signal(SIGINT, datastore_shutdown);
	signal(SIGUSR1, datastore_unlock);
	signal(SIGUSR2, datastore_debug_on);

	while(running) {
		
		FD_ZERO(&fds);

		if(interactive) {
			FD_SET(0, &fds);
		}
	
		FD_SET(sockfd, &fds);
	
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		
		r = select(sockfd+1, &fds, NULL, NULL, &tv);
	
		if(r > 0) {
			if(FD_ISSET(0, &fds)) {
				stdin_handle();
			}

			if(FD_ISSET(sockfd, &fds)) {
				socket_handle(sockfd);
			}
		}
	
		handler_scheduler_check();
		
		configfile_scheduler_check();
		
		alarm_scheduler_check();
		
		if (unlock) {
			apunlock();
			unlock = 0;
		}
	}
	
	printf("Datastore shutdown.\n");
	
	shutdown(sockfd, 2);
	
	close(sockfd);

	return(0);
}


void usage(void)
{
	printf( "\n"
		"usage: datstore2 [options]\n"
		"\n"
		"  -c          Use colored logging\n"
		"  -d <flags>  Enable debugging flags\n"
		"  -i          Enable interactive console\n"
		"  -h          Display this help\n"
	        "  -p <port>   Listen on UDP port <port>\n"
		"\n"
	);

	dbg_showflags();
}


int read_cmdline_args(int argc, char **argv)
{
	int c;

	while(( c = getopt(argc, argv, "cd:hip:")) != EOF) {

		switch(c) {
			case 'c':
				debugcolor = 1;
				break;
			case 'd':
				dbg_initflags(optarg);
				break;
			case 'h':
				usage();
				exit(0);
			case 'i':
				interactive = 1;
				break;
			case 'p':
				udpport = atoi(optarg);
				break;
		}
	}
	
	return(0);
}			
        

int stdin_handle(void)
{
	char buf[64];
	char outbuf[64];
	int r;
	char *p;

	printf("\x1b[1m");
	
	fgets(buf, 64, stdin);
	
	p = strchr(buf, '\n');
	if(p) *p=0;


	if(buf[0] == 0) {
		printf(	"\n"
			"Enter command or :\n"
			"\n"
			"/b Begin journal\n" 
			"/c Commit journal\n"
			"/h Dump scheduled handlers\n"
			"/j Journal dump\n"
			"/o Object dump\n"
			"/q Quit\n"
			"/r Rollback journal\n"
			"/s Schema dump\n"
			"/w Write configfile\n"
		);
	}


	else if(buf[0] != '/') {
		printf("\x1b[0m\n");
		r = parse_command(buf, outbuf);
		printf("%d: %s\n", r, outbuf);
	}

	else {
		switch(tolower(buf[1])) {
		case 'b':
			printf("Activating\n");
			journal_begin();
			break;
		case 'c':
			printf("Committing\n");
			journal_commit();
			break;
		case 'h':
			printf("------------ Scheduled handlers ----------\n");
			handler_scheduler_dump();
			printf("------------------------------------------\n");
			break;
		case 'j':
			printf("----------------- Journal ----------------\n");
			journal_dump();
			printf("------------------------------------------\n");
			break;
		case 'o':
			printf("--------------- Object dump --------------\n");
			object_dumpall();
			printf("------------------------------------------\n");
			break;
		case 'r':
			printf("Rolling back\n");
			journal_rollback();
			break;
		case 's':
			printf("--------------- Schema dump --------------\n");
			schema_dump();
			printf("------------------------------------------\n");
			break;
		case 'q':
			printf("Quit\n");
			exit(0);
			break;
		case 'w':
			configfile_write("/tmp/ap.cfg");
			break;
		default:
			printf("Unknown command %s", buf);
		}
	}

	printf("\x1b[0m\n");

	return(0);
}

