/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>

#include <aplog.h>
#include "debug.h"
#include "types.h"
#include "misc.h"
#include "schema.h"
#include "object.h"

#define OBJCMP_ASSERT
#define OBJCMP_DEBUG

int oidcmp(char *pa, char *pb)
{
	unsigned long va,vb;

	va = vb = 0;

	while(*pa && *pa == *pb) {
	    pa++;
	    pb++;
	}

	/* 'b' is equal or longer than 'a' */
	if(*pa == 0)
		return ((*pb == 0) ? 0 : -1);

	/* 'a' is equal or longer than 'b' */
	if(*pb == 0)
		return ((*pa == 0) ? 0 : 1);

	/* compare the first nonequal oids in 'a' and 'b' */
	while(*pa && *pa != '.') va = (va<<4) + *pa++;
	while(*pb && *pb != '.') vb = (vb<<4) + *pb++;

	return( (va > vb) ? 1 : -1 );
}

#ifdef OBJCMP_DEBUG
static int objcmp_debug = B_FALSE;

void objcmp_debug_on(void) {
	objcmp_debug = B_TRUE;
}

void objcmp_debug_off(void) {
	objcmp_debug = B_FALSE;
}

#endif

int objcmp(struct object_t* obj_a, struct object_t* obj_b)
{
	unsigned long va, vb;
	char* pa, *pb;
	int fa, fb;
	int ret_val;
	char *oidsuff = ".0";

	fa = fb = 0;
	va = vb = 0;

#ifdef OBJCMP_ASSERT		
	assert(obj_a);
	assert(obj_b);

	if(obj_a->sch && obj_a->row) {
		if(obj_a->row->oidpart[0] != '.') {
			printf("screwed\n");
			object_dump(obj_a);
			assert(B_FALSE);
		}
	}
	
	if(obj_b->sch && obj_b->row) {
		if(obj_b->row->oidpart[0] != '.') {
			printf("screwed\n");
			object_dump(obj_b);
			assert(B_FALSE);
		}
	}
#endif	

#ifdef OBJCMP_DEBUG	
	if(objcmp_debug) {
		if(obj_a->sch) {
			dbg(DBG_OBJECT, "objcmp: obj_a->sch->oid   = %s", obj_a->sch->oid);
			dbg(DBG_OBJECT, "objcmp: obj_a suffix      = %s", obj_a->row ? obj_a->row->oidpart : ".0");
		}
		else {
			dbg(DBG_OBJECT, "objcmp: obj_a OID         = %s", obj_a->row->oidpart);
		}
	
		if(obj_b->sch) {
			dbg(DBG_OBJECT, "objcmp: obj_b->sch->oid   = %s", obj_b->sch->oid);	
			dbg(DBG_OBJECT, "objcmp: obj_b suffix      = %s", obj_b->row ? obj_b->row->oidpart : ".0");
		}
		else {
			dbg(DBG_OBJECT, "objcmp: obj_b OID         = %s", obj_b->row->oidpart);
		}
	}
#endif
		
	/* temp. objects are tricky. They have the complete OID inside, in 
	   the row->oidpart */
	if(obj_a->sch) {
		pa = obj_a->sch->oid;
	}
	else {
#ifdef OBJCMP_ASSERT		
		assert(obj_a->row);
		assert(obj_a->row->oidpart);
#endif		
		pa = obj_a->row->oidpart;
		fa = 1;
	}
	
	if(obj_b->sch) {
		pb = obj_b->sch->oid;
	}
	else {
#ifdef OBJCMP_ASSERT		
		assert(obj_b->row);
		assert(obj_b->row->oidpart);
#endif		
		pb = obj_b->row->oidpart;
		fb = 1;
	}
	
	while(*pa && *pa == *pb) {
		pa++;
		pb++;
		
		/* if where at the end of the string, but we are not
		 * reading the oid_suffix yet, adjust 'pa' */
		if((*pa == 0x00) && (fa == 0)) {
			if(obj_a->row) {
#ifdef OBJCMP_ASSERT				
				assert(obj_a->row->oidpart);
#endif				
				pa = obj_a->row->oidpart;
			}
			else {
				pa = oidsuff;
			}
			fa = 1;
		}
		
		/* same for 'b' */
		if((*pb == 0x00) && (fb == 0)) {
			if(obj_b->row) {
#ifdef OBJCMP_ASSERT				
				assert(obj_b->row->oidpart);
#endif				
				pb = obj_b->row->oidpart;
			}
			else {
				pb = oidsuff;
			}
			fb = 1;
		}
		
    	}
	
	/* We assume that the oid is split exactly at the dot. The oid_suffix
	   starts with a dot (.) Therefore we don't need to switch strings
	   anymore from here */
	
	/* 'b' is equal or longer than 'a' */
	if(*pa == 0) {
		ret_val = ((*pb == 0) ? 0 : -1);
#ifdef OBJCMP_DEBUG		
		if(objcmp_debug) {
			dbg(DBG_OBJECT, "objcmp: returning %i", ret_val);
		}
#endif		
		
		return ret_val;
	}

	/* 'a' is equal or longer than 'b' */
	if(*pb == 0) {
		ret_val = ((*pa == 0) ? 0 : 1);
#ifdef OBJCMP_DEBUG		
		if(objcmp_debug) {
			dbg(DBG_OBJECT, "objcmp: returning %i", ret_val);
		}
#endif		
		return ret_val;
	}

	/* compare the first nonequal oids in 'a' and 'b' */
	while(*pa && *pa != '.') va = (va<<4) + *pa++;
	while(*pb && *pb != '.') vb = (vb<<4) + *pb++;


	ret_val = ( (va > vb) ? 1 : -1 );
#ifdef OBJCMP_DEBUG		
	if(objcmp_debug) {
		dbg(DBG_OBJECT, "objcmp: returning %i", ret_val);
	}
#endif		
	return ret_val;
}

int split (char *str, char sep, char **part, int maxparts)
{
	int parts = 0;
	char *p1 = str;
	char *p2;

	assert(part);
	assert(str);
	assert(maxparts > 0);

	part[parts++] = p1;

	while ((p2 = strchr(p1, sep)) && (parts < maxparts)) {
		*p2 = 0;
		p1 = p2 + 1;
		part[parts++] = p1;
	}

	return parts;
}

int splitws(char *buf, char **part, int maxparts)
{
	int parts=0;
	char *p;

	assert(buf);
	assert(part);
	assert(maxparts > 0);

//	/* remove trailing whitespace */
//	p = buf;
//	while(*p) p++;
//	p--;
//	while(isws(*p)) { *p=0; p--; }

	/* split the string on whitespace */
	p = buf;

	while(*p) {
		while(isws(*p)) p++;
		part[parts++] = p;

		if(parts>=maxparts) {
			return(parts);
		}

		while(*p && !isws(*p)) {
			if(*p == '\\') {
				memmove(p, p+1, strlen(p));
				if(*p == 't') *p = '\t';
				if(*p == 'n') *p = '\n';
			}
			p++;
		}
		if(*p != 0) {
			*p++ = 0;
		} else {
			return(parts);
		}
	}

	return(parts);
}



/*
 * Count the number of dots in an oid
 */

int dotcount(char *oid)
{
	int c = 0;
	char *p;

	assert(oid);

	p = oid;
	while(*p) {
		if(*p == '.') c++;
		p++;
	}

	return(c);
}

char *memdup(char *p, int len)
{
	char *new_mem;

	new_mem = malloc(len + 1);
	assert(new_mem);

	memcpy(new_mem, p, len);
	*(new_mem + len) = 0;

	return(new_mem);
}

char *obj_strerror(int e)
{
	switch(e) {
		case R_NOERR:		return("No error");
		case R_OUTOFRANGE:	return("Value out of range");
		case R_NOSUCHOID:	return("No such oid");
		case R_NOACCESS:	return("No access for requested action");
		default:		return("Unknown error");
	}
}

string *mkstring(char *data, int len)
{
	string *str;

	assert(len >= 0);

	str = (string *) malloc(sizeof(int) + len + 1);
	assert(str);
	str->len = len;

	if(data != NULL)
		memcpy(str->data, data, len);
	else
		if(len > 0)
			str->data[0] = 0;

	str->data[len] = 0;

	return(str);
}


void exec(char *const cmd[])
{
  pid_t pid, pid2;

  if ( (pid = vfork()) == 0 )
  {
    if ( ( pid2 = vfork()) == 0)
    {
       execv(cmd[0], cmd);

       /* Should never reach this line */
       aplog(EV_SNMP_GENERAL_ERROR,"Failed to start application",NULL);
       exit(-1);
    }
    exit(0);
  }
  else
  {
    /* wait until child 1 ends */
    waitpid(pid, NULL, 0);
  }
}


// eot
