/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdlib.h>
#include <assert.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "types.h"
#include "object.h"
#include "range.h"
#include "misc.h"
#include "objflag.h"

/* Some macro's for smooth handling of various types of data */

#define checkobjtype(obj, chktype)  	do { assert(obj); \
				 	     assert(obj->sch); \
				 	     assert(obj->sch->type == chktype); \
					} while(0);


int object_get_max_bit(struct object_t *obj)
{
	char *p;
	int max = 0;
	int i;
	checkobjtype(obj, T_BITS);
		
	p = obj->sch->range;
	
	/* BITS range is defined as "1,3,4", which means that only
	   BIT 1, 3 and 4 are defined */
	while(*p) {
		i = atol(p);
		if (i > max) {
			max = i;
		}
		
		while(isdigit(*p)) {
			p++;
		}	
		
		p++;	
	}
	
	return (max);
}

int object_get_int(struct object_t *obj)
{
	checkobjtype(obj, T_INTEGER);
	if(!(obj->flags & OF_DATAPRESENT)){	
		printf("OF_DATAPRESENT failed: oid: %s\n", obj->sch->oid);
	}
	assert(obj->flags & OF_DATAPRESENT);
	return((int) obj->value);
}

unsigned int object_get_uint(struct object_t *obj)
{
	checkobjtype(obj, T_UNSIGNED);
	assert(obj->flags & OF_DATAPRESENT);
	return((unsigned int) obj->value);
}

unsigned int object_get_rowstat(struct object_t *obj)
{
	checkobjtype(obj, T_ROWSTAT);
	assert(obj->flags & OF_DATAPRESENT);
	return((unsigned int) obj->value);
}

struct in_addr* object_get_ipaddr(struct object_t *obj)
{
	checkobjtype(obj, T_IPADDR);
	assert(obj->flags & OF_DATAPRESENT);
	return((struct in_addr*) &(obj->value));
}

string *object_get_string(struct object_t *obj)
{
	checkobjtype(obj, T_OCTETSTR);
	assert(obj->flags & OF_DATAPRESENT);
        return((string *)(obj->value));
}

char *object_get_bits(struct object_t *obj)
{
	checkobjtype(obj, T_BITS);
	assert(obj->flags & OF_DATAPRESENT);
	return((char*) obj->value);	
}

unsigned int object_get_bit(struct object_t *obj, int bit) 
{
	int bytenr, bit_in_byte;
	unsigned char mask;
	char *bytes;
	
	checkobjtype(obj, T_BITS);
	assert(obj->flags & OF_DATAPRESENT);
	assert(bit <= object_get_max_bit(obj));
	
	bytenr = bit / 8;
	bit_in_byte = bit % 8;
	mask = 1 << (7 - bit_in_byte);
	bytes = (char *) obj->value;
	
	return(((bytes[bytenr] & mask) > 0) ? 1 : 0);
}

unsigned int object_get_counter(struct object_t *obj)
{
	checkobjtype(obj, T_COUNTER);
	assert(obj->flags & OF_DATAPRESENT);
	return((unsigned int) obj->value);
}

unsigned long long object_get_counter64(struct object_t *obj)
{
	checkobjtype(obj, T_COUNTER64);
	assert(obj->flags & OF_DATAPRESENT);
	return(*((unsigned long long *) obj->value));
}

unsigned int object_get_gauge(struct object_t *obj)
{
	checkobjtype(obj, T_GAUGE);
	assert(obj->flags & OF_DATAPRESENT);
	return((unsigned int) obj->value );
}

string *object_get_objectid(struct object_t* obj) 
{
	checkobjtype(obj, T_OBJECTID);
	assert(obj->flags & OF_DATAPRESENT);
        return((string *)(obj->value));
}

string *object_get_opaque(struct object_t* obj)
{
	checkobjtype(obj, T_OPAQUE);
	assert(obj->flags & OF_DATAPRESENT);
        return((string *)(obj->value));
}

unsigned int object_get_ticks(struct object_t* obj)
{
	checkobjtype(obj, T_TICKS);
	assert(obj->flags & OF_DATAPRESENT);
	return((unsigned int) obj->value );
}

string *object_get_notify(struct object_t* obj)
{
	checkobjtype(obj, T_NOTIFY);
	assert(obj->flags & OF_DATAPRESENT);
        return((string *)(obj->value));
}

int object_set_int(struct object_t *obj, int val)
{
	checkobjtype(obj, T_INTEGER);

	if(range_check(obj, (long long) val) == 0) {
		obj->value = (unsigned long) val;
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_uint(struct object_t *obj, unsigned int val)
{
	checkobjtype(obj, T_UNSIGNED);

	if(range_check(obj,(long long) val) == 0) {
		obj->value = (unsigned long) val;
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_rowstat(struct object_t *obj, int val)
{
	checkobjtype(obj, T_ROWSTAT);

	if(range_check(obj, (long long) val) == 0) {
		obj->value = (unsigned long) val;
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_ipaddr(struct object_t *obj, struct in_addr val)
{
	checkobjtype(obj, T_IPADDR);

	/* there's no rangecheck on IP addresses possible */
	
	memcpy(&(obj->value), &val, 4);

	object_flag_set(obj, OF_DATAPRESENT);
	
	return R_NOERR;
}


int object_set_string(struct object_t *obj, char *data, int len)
{
	checkobjtype(obj, T_OCTETSTR);

	if(range_check(obj, (long long) len) == 0) {
		if(object_flag_isset(obj, OF_DATAPRESENT)) {
			free((string*) obj->value);
		}
		obj->value = (unsigned long) mkstring(data, len);
		assert(obj->value);
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}


int object_set_bit(struct object_t *obj, int bit, int value)
{
	int byte;
	int bit_in_byte; 
	unsigned char mask;
	char *data;

	checkobjtype(obj, T_BITS);
	assert((value == 1) || (value == 0));
	
	if(range_check(obj, (long long) bit) == 0) {
		if(!object_flag_isset(obj, OF_DATAPRESENT)) {
			obj->value = (unsigned long) calloc(1, (object_get_max_bit(obj) / 8) + 1);
			assert(obj->value);
			object_flag_set(obj, OF_DATAPRESENT);
		}
		
		byte = bit / 8;
 		bit_in_byte = bit % 8;
 		mask = 1 << (7-bit_in_byte); /* switch bit on */
 		mask = mask ^ 0xFF; /* invert mask */
		
		data = (char*) obj->value;
		
		data[byte] = data[byte] & mask; /* switch bit off */
		
		if (value == 1) {
			/* invert mask */
			mask = mask ^ 0xFF;
	 
	 		/* switch bit on */
	 		data[byte] = data[byte] | mask;
		}
										
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}


int object_set_counter(struct object_t *obj, unsigned int val)
{
	checkobjtype(obj, T_COUNTER);
	
	if(range_check(obj, (long long) val) == 0) {
		obj->value = (unsigned long) val;
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_counter64(struct object_t *obj, unsigned long long val)
{
	checkobjtype(obj, T_COUNTER64);
	
	if(range_check(obj, (long long) val) == 0) {
		if(!object_flag_isset(obj, OF_DATAPRESENT)) {
			obj->value = (unsigned long) malloc(sizeof(unsigned long long));
			assert(obj->value);
			object_flag_set(obj, OF_DATAPRESENT);
		}
		*((unsigned long long*) obj->value) = val;
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_gauge(struct object_t *obj, unsigned int val)
{
	checkobjtype(obj, T_GAUGE);
	
	if(range_check(obj,(long long) val) == 0) {
		obj->value = (unsigned long) val;
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_objectid(struct object_t *obj, char *data)
{
	int len;
	
	checkobjtype(obj, T_OBJECTID);
	
	len = strlen(data);

	if(range_check(obj, (long long) len) == 0) {
		if(object_flag_isset(obj, OF_DATAPRESENT)) {
			free((string*) obj->value);
		}
		obj->value = (unsigned long) mkstring(data, len);
		assert(obj->value);
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_opaque(struct object_t *obj, char *data, int len)
{
	checkobjtype(obj, T_OPAQUE);
	
	if(range_check(obj, (long long) len) == 0) {
		if(object_flag_isset(obj, OF_DATAPRESENT)) {
			free((string*) obj->value);
		}
		obj->value = (unsigned long) mkstring(data, len);
		assert(obj->value);
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_ticks(struct object_t *obj, unsigned int val)
{
	checkobjtype(obj, T_TICKS);
	
	if(range_check(obj,(long long) val) == 0) {
		obj->value = (unsigned long) val;
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}

int object_set_notify(struct object_t *obj, char *data, int len)
{
	checkobjtype(obj, T_NOTIFY);
	
	if(range_check(obj, (long long) len) == 0) {
		if(object_flag_isset(obj, OF_DATAPRESENT)) {
			free((string*) obj->value);
		}
		obj->value = (unsigned long) mkstring(data, len);
		assert(obj->value);
		object_flag_set(obj, OF_DATAPRESENT);
		return R_NOERR;
	}
	else {
		return R_OUTOFRANGE;
	}
}


