/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>



#include "types.h"
#include "misc.h"
#include "gAVLTree.h"
#include "object.h"
#include "schema.h"
#include "journal.h"
#include "handler.h"
#include "range.h"
#include "debug.h"
#include "serialize.h"
#include "objflag.h"
#include "objdata.h"

gAVLTree *objtree;

int obj_errno;

static gAVLKey object_getavlkey(void *item)
{
	
	return((gAVLKey)item);
}

int object_inittree(void)
{
        objtree = gAVLAllocTree((gAVLKey) object_getavlkey, (gAVLCompare) objcmp);
        obj_errno = R_NOERR;

	return(0);
}

#define OIDCHECK(oid)  if(!oid || (oid[0] != '.')) \
			printf("Invalid oid %s at %s %i\n", oid, __FILE__, __LINE__);
/*
 * Get an object from the datastore. Check expiretimes etc...
 */

struct object_t *object_get(char *oid)
{
	struct object_t *obj;
	static struct object_t tmp_obj;
	static struct row_t tmp_row;

	OIDCHECK(oid);
	
	tmp_obj.sch = NULL;
	tmp_obj.row = &tmp_row;
	tmp_row.oidpart = oid;

	obj = (struct object_t *)gAVLSearch(objtree, (gAVLKey)&tmp_obj);
	dbg(DBG_OBJECT, "object_get %s", oid);

	if(obj == NULL) {
		obj_errno = R_NOSUCHOID;
	} else {
		obj_errno = R_NOERR;
	}

	return (obj);
}



struct object_t *object_get_next(char *oid)
{
	gAVLCursor cursor;
	struct object_t *obj;
	static struct object_t tmp_obj;
	static struct row_t tmp_row;

	OIDCHECK(oid);
	
	tmp_obj.sch = NULL;
	tmp_obj.row = &tmp_row;
	tmp_row.oidpart = oid;

	/* See if we can find the given object */
	obj = gAVLSearchNext(objtree, (gAVLKey)&tmp_obj, &cursor);
	dbg(DBG_OBJECT, "object_get_next %s", oid);

	while (obj && (obj->sch->type == T_TABLE)) {
		handler_call(obj->sch->oid, H_GET, 0);
		obj = gAVLNext(&cursor);
	} 

	if(obj) {
		obj_errno = R_NOERR;
		dbg(DBG_OBJECT, "object_get_next returns: obj->sch->oid = %s, obj->oid_suffix = %s", obj->sch->oid, obj->row ? obj->row->oidpart : ".0");					
	}
	else {
		obj_errno = R_ENDOFTREE;
		dbg(DBG_OBJECT, "object_get_next: END OF MIB");
	}
	return(obj);		
}


/*
 * Create an object struct. if the argument <oid> is 
 * set to NULL the object is created, but *NOT* put in the objtree,
 * and it's schent_t pointer is set to NULL;
 */
struct object_t *object_create_in_row(char *oid, struct object_t* row_obj)
{
	struct object_t *obj = NULL;
	struct schent_t *sch = NULL;
	int l;
	char *p;
	static struct object_t tmp_obj;
	static struct row_t tmp_row;

	dbg(DBG_OBJECT, "object_create %s", oid);

	if(oid) {
		OIDCHECK(oid);
	
		tmp_obj.sch = NULL;
		tmp_obj.row = &tmp_row;
		tmp_row.oidpart = oid;
		obj = (struct object_t *)gAVLSearch(objtree, (gAVLKey)&tmp_obj);

		if(obj) {
			printf("object_create: object is already in tree\n");
			return(NULL);
		}
	
		sch = schema_get(oid);
		if(sch == NULL) {
			sch = schema_find_schent_for_oid(oid);
		}
		if(sch == NULL) {
			printf("object_create: no parent schema entry\n");
			return(NULL);
		}

		/*
		 * Check if the schema entry is a table member. If it's not,
		 * the oid should be the sch's oid + ".0"; Only T_TABLE oid's
		 * can exist (ghost object);
		 */
		if((sch->prev == NULL) && (sch->type != T_TABLE)){
			l = strlen(oid) - strlen(sch->oid);
	
			if(l != 2 ) {
				printf("Oid of non-table should be sch->oid + '.0' a");
				printf("oid=%s type=%d\n", sch->oid, sch->type);
				return(NULL);
			}
			p = oid + strlen(oid) - 2;
	
			if((*p++ != '.') && (*p != '0')) {
				printf("Oid of non-table should be sch->oid + '.0' b");
				printf("oid=%s type=%d\n", sch->oid, sch->type);
				return(NULL);
			}
		}		
	}
	
	if(!row_obj) {
		obj = calloc(1, sizeof(struct object_t));
		assert(obj);
	}
	else {
		obj = row_obj;
	}
	
	if(sch) {
		obj->sch = sch;
		/* Set default value, if applicable */
		if(obj->sch->defval) {
			dbg(DBG_OBJECT, "Setting default value %s", obj->sch->defval);
			if (deserialize(obj, obj->sch->defval) != R_NOERR) {
				printf("Default value for object conflicts with range\n");
			}
		}
		
		gAVLInsert(objtree, (void *)obj);
	}
	else {
		/* mark this object as floating (= not in AVL tree) */
		object_flag_set(obj, OF_FLOATING);
	}

	return(obj);
}
	
struct object_t *object_create(char *oid) {
	return(object_create_in_row(oid, NULL));
}

void object_delete(struct object_t *obj)
{
	assert(obj);

	if(!object_flag_isset(obj, OF_FLOATING)) {
		dbg(DBG_OBJECT, "object delete %s%s", obj->sch->oid, obj->row ? obj->row->oidpart : ".0");
		gAVLDelete(objtree, (gAVLKey)obj);
	} else {
		dbg(DBG_OBJECT, "object delete unnamed object");
	}

	if((obj->flags & OF_DATAPRESENT) && (object_has_malloced_data(obj))) {
		free((void*) obj->value);
	}
	
	/* if this object is not part of a row, free it */
	if(!obj->row) {
		free(obj);
	}
}

int object_get_oid(char* oid, size_t max, struct object_t* obj)
{
	assert(obj);
	assert(obj->sch);
	return(snprintf(oid, max, "%s%s", obj->sch->oid, obj->row ? obj->row->oidpart : ".0"));
}


void object_dumpall(void)
{
	struct object_t *obj;
	gAVLCursor avlcursor;
	char buf[256];
	int count=0;

	obj = (struct object_t *) gAVLFirst(&avlcursor, objtree);

	while(obj) {
		serialize(obj, buf, sizeof(buf));
	
		printf("%s%s (%d) = %s\n", obj->sch->oid, obj->row ? obj->row->oidpart : ".0", obj->sch->type, buf);
	
		count++;
		obj = (struct object_t *) gAVLNext(&avlcursor);
	}

	printf("\nTotal: %d\n", count);
}

void object_dump(struct object_t *obj) {
	
	printf("==== START OBJECT DUMP =====\n");
	printf("oidend     = %s\n", obj->row ? obj->row->oidpart : ".0");
	printf("data       = %x\n", (int) obj->value);
	printf("flags      = %x\n", obj->flags);
	printf("row        = %x\n", (int) obj->row);
	printf("sch        = %x\n", (int) obj->sch);
	if (obj->sch) {
		printf("sch->oid   = %s\n", obj->sch->oid);
	}
	printf("==== END OBJECT DUMP =====\n");	
}


// EOT
