/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <assert.h>

#include "types.h"
#include "misc.h"
#include "debug.h"


#define isdigit2(a) (isdigit(a) || (a == '-') || (a == '+'))


/*
 * Check if the given object matches the required datatype and range
 */
int range_check(struct object_t *obj, long long value)
{
	char *p1;
	long long from, to;
	int retval = 1;
	struct schent_t *sch;

	assert(obj);
	sch = obj->sch;
	assert(sch);

	/* If no range defined (NULL), it's always OK */
	if(sch->range == NULL) {
		return(0);
	}
	
	dbg(DBG_RANGE, "range_check %s for range %s : val=%lli", sch->oid, sch->range, value);

	if((sch->type == T_UNSIGNED) && (value<0)) 
		return(-1);

	// Go split up the range string and check if the value matches
	
	p1 = sch->range;
	
	while(*p1) {
		from = to = atoll(p1);
			
		while(isdigit2(*p1)) p1++;
	
		if(*p1 == '/') {
			p1++;
			if(isdigit2(*p1)) {
				to = atoll(p1);
				while(isdigit2(*p1)) p1++;
			}
		}
	
		if((value >= from) && (value <= to)) retval = 0;
		
		if(*p1 == '\x00') {
			return(retval);
		}
		p1++;
	}


	
	return(retval);
}

int range_fixed_length(struct schent_t* sch)
{
	char *p;
	
	assert(sch);
	
	/* Check type */
	if (sch->type != T_OCTETSTR) {
		return 0;
	}
	
	/* Check for empty range */
	if (sch->range == NULL) {
		return -1;
	}
	
	p = strchr(sch->range, '/');
	
	if(!p) {
		die(); // This indicates corrupt data
		return -1;
	}
	
	/* Place p1 after '/' */
	p++; 
	
	/*
	 * |-|-|-|-|-|--|
	 * |2|0|/|2|0|\0|
	 * |-|-|-|-|-|--|
	 *  ^     ^
	 *  |     + - p
	 *  |
	 *  + - sch->range
	 */
	if (strncmp(sch->range, p, strlen(p)) == 0) {
		return atoi(p);
	}
	else {
		return -1;
	}
}

