/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */


#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <arpa/inet.h>

#include "types.h"
#include "object.h"
#include "schema.h"
#include "misc.h"
#include "objdata.h"
#include "debug.h"
#include "objflag.h"


static int index_size(struct schent_t *sch) 
{
	switch(sch->type) {
		case T_UNSIGNED:
			return 1;
			break;
		case T_INTEGER:
			return 1;
			break;
		case T_IPADDR:
			return 4;
			break;
		default:
			printf("Unsupported type in index (%i)\n", sch->type);
			die();
			break;
	}
}


static struct index_t *index_create(struct schent_t *sch) 
{
	struct index_t *idx;
	struct indexoid_t *ioid;
	int indices;

	assert(sch);

	idx = malloc(sizeof(struct index_t));
	assert(idx);
	idx->sch = sch;

	ioid = sch->indexlist;

	indices = 0;
	while(ioid) {
		indices++;
		ioid = ioid->next;
	}
	assert(indices > 0);

	idx->obj = malloc(sizeof(struct object_t *) * indices);
	assert(idx->obj);
	memset(idx->obj, sizeof(struct object_t *) * indices, 0);
	idx->num = indices;
	
	return(idx);
}
	
struct index_t *index_new(struct schent_t *sch)
{
	struct index_t* idx;
	struct indexoid_t *ioid;
	int i;
	
	idx = index_create(sch);
	
	i=0;
	ioid = idx->sch->indexlist;
	while(ioid) {
		idx->obj[i] = object_create(NULL);
		assert(idx->obj[i]);
		idx->obj[i]->sch = schema_get(ioid->oid);
		assert(idx->obj[i]->sch);
		i++;
		ioid = ioid->next;
	}

	return(idx);
}

static void object_fill_by_index(struct object_t *obj, char* indexoid[], int offset) {
	
	char ipbuf[16];
	struct in_addr ipaddr;

	assert(obj);
	assert(obj->sch);
	
	switch(obj->sch->type) {
		case T_UNSIGNED:
			object_set_uint(obj, atoi(indexoid[offset]));
			break;
		case T_INTEGER:
			object_set_int(obj, atoi(indexoid[offset]));
			break;
		case T_IPADDR:
			snprintf(ipbuf, sizeof(ipbuf), "%s.%s.%s.%s", indexoid[offset], indexoid[offset+1], indexoid[offset+2], indexoid[offset+3]);
			inet_aton(ipbuf, &ipaddr);
			object_set_ipaddr(obj, ipaddr);
			break;
		default:
			printf("Unsupported type in index (%i)\n", obj->sch->type);
			die();
			break;
	}
}

struct index_t *index_get(struct row_t *row)
{
	struct index_t *idx;
	struct object_t *obj;
	char *indexoid;
	char *indexoid_p[MAX_INDEX_LENGTH];
	int parts;
	int i, j;
	struct indexoid_t *ioid;
	
	assert(row);
	assert(row->sch);
	
	/* Create an index with dummy objects */
	idx = index_new(row->sch);
	
	indexoid = strdup(row->oidpart);
	assert(indexoid);
	
	parts = split(indexoid+1, '.', indexoid_p, MAX_INDEX_LENGTH);
	
	ioid = idx->sch->indexlist;
	i = 0;
	j = 0;
	while(ioid) {
		obj = idx->obj[i];
		assert(obj);
		
		object_fill_by_index(obj, indexoid_p, j);
	
		j = j + index_size(obj->sch);
		assert(j <= parts);
		
		i++;		
		ioid = ioid->next;
	}
	
	free(indexoid);
	return(idx);
	
}

struct index_t *index_get_old(struct row_t *row)
{	
	struct index_t *idx;
	struct indexoid_t *ioid;
	char *indexoid;
	char *indexoid_p[MAX_INDEX_LENGTH];
	char oid[MAX_OID_LENGTH];
	int parts, idx_dig;
	int n;
	int i;
	int idx_size;
	char* oid_p;

	// Create a new index object

	assert(row);
	assert(row->sch);

	indexoid = strdup(row->oidpart);
	assert(indexoid);
	
	parts = split(indexoid+1, '.' , indexoid_p, MAX_INDEX_LENGTH);

	idx = index_create(row->sch);

	ioid = row->sch->indexlist;
	i=0;
	idx_dig = 0;

	// Fill in the objects in the index

	while(ioid) {
		if (strncmp(ioid->oid, row->sch->oid, strlen(row->sch->oid)) == 0 ) {
			/* internal index */
			sprintf(oid, "%s%s", ioid->oid, row->oidpart);
		}		
		else {
			/* external index */
			strcpy(oid, ioid->oid);
			idx_size = index_size(schema_get(ioid->oid));
			for(n= idx_dig; n < (idx_dig + idx_size); n++) {
				oid_p = oid + strlen(oid);
				sprintf(oid_p, ".%s", indexoid_p[n]);
			}
						
		}
		
		idx->obj[i] = object_get(oid);
		assert(idx->obj[i]);
		
		ioid = ioid->next;
		
		idx_dig += index_size(idx->obj[i]->sch);				
		i++;
	}
	free(indexoid);
	return(idx);
}

void index_done(struct index_t *idx)
{
	int i;
	assert(idx);

	for(i=0; i<idx->num; i++) {
		/* only delete 'floating' objects. During the creating of a 
		 * row, a index with emtpy objects has been created (with
		 * index_new(). These objects should be removed. If an index
		 * was created with index_get(). The objects in the AVLtree
		 * are used in the index. These objects should NOT be 
		 * deleted 
		 */
		if(object_flag_isset(idx->obj[i], OF_FLOATING)) {
			object_delete(idx->obj[i]);
		}
	}
	free(idx->obj);	
	free(idx);
}




struct object_t *index_get_obj(struct index_t *idx, int num)
{
	assert(idx->obj[num]);
	return(idx->obj[num]);
}


struct row_t *row_create_by_oid(char *oid)
{
	struct schent_t* sch, *tbl_sch;
	int columns;
	char* oidpart;
	int i;
	char new_oid[MAX_OID_LENGTH];
	struct row_t* row;
	
	sch = schema_find_schent_for_oid(oid);
	assert(sch);
	
	oidpart = oid + strlen(sch->oid);
	
	/* find table sch entry */
	while(sch->prev) {
		sch = sch->prev;
	}
	
	tbl_sch = sch;
	
	columns = 0;
	/* count columns */
	sch = sch->next;  // first column
	while(sch) {
		columns++;
		sch = sch->next;
	}
	
	/* create row with objects */
	row = calloc(1, sizeof(struct row_t) +(sizeof(struct object_t) *(columns - 1)));
	assert(row);
	row->columns = columns;
	row->oidpart = strdup(oidpart);
	assert(row->oidpart);
	
	sch = tbl_sch->next;
	for(i=0; i<columns; i++) {
		snprintf(new_oid, sizeof(new_oid), "%s%s", sch->oid, oidpart);
		row->objs[i].row = row;
		object_create_in_row(new_oid, &(row->objs[i]));
		sch=sch->next;
	}
	
	// Insert the row in the sch's rowlist 
	row->sch= tbl_sch;
	sch = row->sch;
	
	if(sch->rowlist == NULL) {
		row->prev = NULL;
		row->next = NULL;
		sch->rowlist = row;
	}
	else {
		row->next = sch->rowlist;
		row->prev = NULL;
		row->next->prev = row;
		sch->rowlist = row;
	}
	
	sch->rowcount++;
	
	return(row);	
	
}

int row_remove(struct row_t *row) {

	struct schent_t* sch;
	struct row_t *tmp_row;
	
	dbg(DBG_TABLE, "Removing row from schema: %s.X%s", row->sch->oid, row->oidpart);
	sch = row->sch;
	assert(sch);
	
	tmp_row = sch->rowlist;
	
	if(tmp_row == row) {
		sch->rowlist = row->next;
		if(row->next) {
			row->next->prev = row->prev; /* = NULL */
		}

		row->next = NULL; /* just to be clean */

		sch->rowcount--;
		
		return(1);
	}	
	else {
		while((tmp_row) && (tmp_row->next != row)) {
			tmp_row = tmp_row->next;
		}

		if (tmp_row) {
			tmp_row->next = tmp_row->next->next;
			if(tmp_row->next) {
				tmp_row->next->prev = tmp_row;
			}
			
			row->next = NULL; 
			row->prev = NULL;
			
			sch->rowcount--;
			
			return(1);
		}
		else {
			return(0); /* row not found */
		}
	}

}

int row_delete(struct row_t *row)
{
	int i;
	
	assert(row);
	
	dbg(DBG_TABLE, "Deleting Row: %s.X%s", row->sch->oid, row->oidpart);
	
	/* Delete all objects in this row. Because these objects are part
	 * of a row, the actual object is not freed. */
	for(i=0; i < row->columns; i++){
		object_delete(&(row->objs[i]));
	}
	
	/* Remove row from rowlist */
	row_remove(row);				
	
	free(row->oidpart);
	free(row);
		
	return(0);
}



static char *gen_oidpart_from_index(struct index_t *idx)
{
	static char oidpart[MAX_OID_LENGTH];
	char *p;
	int i,n;
	struct object_t *obj;
	struct in_addr *ad;
	struct string_t *str;

	p = oidpart;
	*p=0;

	for(i=0; i<idx->num; i++) {

		obj = index_get_obj(idx, i);
	
		switch(obj->sch->type) {
	
			case T_INTEGER:
			case T_ROWSTAT:
	
				p += sprintf(p, ".%d", object_get_int(obj));
				break;
		
			case T_UNSIGNED:
		
				p += sprintf(p, ".%d", object_get_uint(obj));
				break;
		
			case T_IPADDR:
	
				ad = object_get_ipaddr(obj);
				p += sprintf(p, ".%s", inet_ntoa(*ad));
				break;
		
			case T_OCTETSTR:
	
				str = object_get_string(obj);
		
				for(n=0; n<str->len; n++) {
					p+= sprintf(p, ".%d", str->data[n]);
				}
		
				break;
			default:
				die();
		}
			
	}

	return(oidpart);
}



struct row_t *row_create_by_index(struct index_t *idx)
{
	char oid[MAX_OID_LENGTH];
	char *oidpart;
	int i;
	struct row_t *row;
	struct schent_t *sch;
	int columns;

	oidpart = gen_oidpart_from_index(idx);

	sch = idx->sch;	

	sch = sch->next;
	columns=0;
	while(sch) {
		columns++;
		sch = sch->next;
	}
	
	row = calloc(1, sizeof(struct row_t) +(sizeof(struct object_t) *(columns - 1)));
	assert(row);
	row->columns = columns;
	row->oidpart = strdup(oidpart);
	assert(row->oidpart);

	sch = idx->sch->next;

	for(i=0; i<columns; i++) {
		snprintf(oid, sizeof(oid), "%s%s", sch->oid, oidpart);
		row->objs[i].row = row;
		object_create_in_row(oid, &(row->objs[i]));
		sch = sch->next;
	}

	/* Insert the row in the sch's rowlist */
	sch = idx->sch;
	row->sch = idx->sch;

	if(sch->rowlist == NULL) {
		row->prev = NULL;
		row->next = NULL;
		sch->rowlist = row;
	} else {
		row->next = sch->rowlist;
		row->prev = NULL;
		row->next->prev = row;
		sch->rowlist = row;
	}

	sch->rowcount++;
		
	return(row);
}




struct row_t *row_get_by_index(struct index_t *idx)
{
	char *oidpart;
	struct row_t *row;

	assert(idx);
	oidpart = gen_oidpart_from_index(idx);

	row = idx->sch->rowlist;

	while(row) {
		if(strcmp(row->oidpart, oidpart) == 0) return(row);
		row = row->next;
	}

	return(NULL);
}


struct object_t *row_get_obj(struct row_t *row, int column)
{
	assert(column < row->columns);
	return(&(row->objs[column]));
}

int table_count(struct schent_t* sch)
{
	assert(sch);
	return(sch->rowcount);
}
