/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
#ifndef _DATASTORE_MISC_H_
#define _DATASTORE_MISC_H_
/** @file misc.h
 *
 * Various helper functions
 */

#include <netinet/in.h>
#include <arpa/inet.h>

#include "schema.h"


#define TODO printf("TODO at %s %d\n", __FILE__, __LINE__)
#define isws(a) (((a)==' ')||((a)=='\t')||((a)=='\n'))

/** Print a message containing file, function and line number, and exit with errorcode */
#define die() {printf("Died at %s, %s line %d\n", __FUNCTION__, __FILE__, __LINE__); exit(1);}

/**
 * Compare two OIDS. This function is performance critical for the whole
 * program since it's used extensively by all gAVL..() functions
 * Return -1 if a<b, 0 if a==b, +1 of a>b
 *
 * Fastest version to date by Michiel. 
 */
int oidcmp(char *a, char *b);

int objcmp(struct object_t* obj_a, struct object_t* obj_b);

/** Split the given <str> into maximum <maxparts> pieces seperated by <sep>. 
 *  Returns the actual number of pieces */
int split(char *str, char sep, char **part, int maxparts);

/** Split the given <str> into <maxparts> pieces seperated by one or more whitespaces 
 *
 * parts = splitws("  een en\ twee    drie \\vier vijf zes", part, 4) 
 *
 * results in :
 *
 * parts   = 4
 *
 * part[0] = "een"
 * part[1] = "en twee"
 * part[2] = "drie"
 * part[3] = "\vier vijf zes"
 */
int splitws(char *buf, char **part, int maxparts);

/** Return the number of dots in the given oid */
int dotcount(char *oid);

/** Duplicate the memory pointed to by <p> and the given length, and return the pointer */
char *memdup(char *p, int len);

/** Return a pointer to a human-readable string describing the obj_error <e> */
char *obj_strerror(int e);

struct string_t *mkstring(char *data, int len);

void exec(char* const cmd[]);

#endif


