/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
#ifndef _DATASTORE_OBJDATA_H_
#define _DATASTORE_OBJDATA_H_
/** @file objdata.h
 *
 * Functions for setting and getting data from objects. 
 */
 
#include <arpa/inet.h>

#include "types.h"


int object_get_int(struct object_t *obj);
unsigned int object_get_uint(struct object_t *obj);
unsigned int object_get_rowstat(struct object_t *obj);
struct in_addr* object_get_ipaddr(struct object_t *obj);
string *object_get_string(struct object_t *obj);
unsigned int object_get_counter(struct object_t *obj);
unsigned long long object_get_counter64(struct object_t *obj);
unsigned int object_get_gauge(struct object_t *obj);
string *object_get_objectid(struct object_t *obj);
string *object_get_opaque(struct object_t *obj);
unsigned int object_get_ticks(struct object_t *obj);
string *object_get_notify(struct object_t *obj);

int object_set_int(struct object_t *obj, int val);
int object_set_uint(struct object_t *obj, unsigned int val);
int object_set_rowstat(struct object_t *obj, int val);
int object_set_ipaddr(struct object_t *obj, struct in_addr val);
int object_set_string(struct object_t *obj, char *data, int len);
int object_set_counter(struct object_t *obj, unsigned int val);
int object_set_counter64(struct object_t *obj, unsigned long long val);
int object_set_gauge(struct object_t *obj, unsigned int val);
int object_set_objectid(struct object_t *obj, char *data);
int object_set_opaque(struct object_t *obj, char *data, int len);
int object_set_ticks(struct object_t *obj, unsigned int val);
int object_set_notify(struct object_t *obj, char *data, int len);

#define object_has_malloced_data(obj) (   (obj->sch->type == T_OCTETSTR) \
				        || (obj->sch->type == T_BITS) \
		                        || (obj->sch->type == T_OBJECTID) \
 					|| (obj->sch->type == T_COUNTER64) \
					|| (obj->sch->type == T_OPAQUE) \
					|| (obj->sch->type == T_NOTIFY) \
				       ) 

/************************/
/* Functions for T_BITS */
/************************/

/** Gets a single bit from a T_BITS object. bit is the bit to be
 *  retrieved. bit should be defined in the range of the object */
unsigned int object_get_bit(struct object_t *obj, unsigned int bit);

/** Gets all the bits from a T_BITS object. The length of the returned 
 *  pointer can be derived from object_get_max_bit() 
 *  (length = (object_get_max_bits() % 8) + 1) */
unsigned char* object_get_bits(struct object_t *obj);

/** Sets a single bit in a T_BITS object. value should be 0 or 1, bit
 *  should be defined in the range of the object. Returns R_NOERR if
 *  OK or R_OUTOFRANGE if the value is .... Out Of Range */
int object_set_bit(struct object_t *obj, unsigned int bit, unsigned int value);

/** Returns the maximum bit defined for this object */
int object_get_max_bit(struct object_t *obj);
   

#endif
