/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
#ifndef _DATASTORE_OBJECT_H_
#define _DATASTORE_OBJECT_H_
/** @file object.h
 *
 * @brief Functions for creation, retrieval and destroying of objects
 */


#include "gAVLTree.h"
#include "table.h"

extern int obj_errno;

int object_inittree(void);

struct object_t *object_get(char *oid);

struct object_t *object_get_next(char *oid);

/**
 * Create an object struct. if the argument <oid> is 
 * set to NULL the object is created, but *NOT* put in the objtree,
 * and it's schent_t pointer is set to NULL;
 */
struct object_t *object_create(char *oid);
 
/**
 * Create an object struct which is part of a row. The given object
 * is already malloc as part of the row.
 */
struct object_t *object_create_in_row(char* oid, struct object_t* row_obj);

void object_delete(struct object_t *obj);

void object_dumpall(void);

void object_dump(struct object_t *obj);

/** 
 * Retrieves the OID from a object. This OID stored in oid. max is max
 * size of the oid buffer
 */
int object_get_oid(char* oid, size_t max, struct object_t* obj);

#endif
