/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
#ifndef _DATASTORE_TABLE_H_
#define _DATASTORE_TABLE_H_
/** @file table.h
 *
 * Functions for handling tables (row_t objects)
 */

#include "types.h"

/** create a new index_t for a row. */
struct index_t *index_new(struct schent_t *sch);

/** get index_t object corresponding to the row */
struct index_t *index_get(struct row_t *row);

/** free's the index object */
void index_done(struct index_t *idx);

/** Get the num-th object from the index */
struct object_t *index_get_obj(struct index_t *idx, int num);

/** Create a row using a oid of one of its field */
struct row_t *row_create_by_oid(char *oid);

/** Create a row using an index_t */
struct row_t *row_create_by_index(struct index_t *idx);

/** Gets a row by its index */
struct row_t *row_get_by_index(struct index_t *idx);

/** Gets the columns-th object from a row */
struct object_t *row_get_obj(struct row_t *row, int column);

/** Delete all the objects from a row and removes the row from 
 *  the rowlist in its schema */
int row_delete(struct row_t *row);

/** Removes the row from its schema. The row and objects in the row
  * will not be freed. Returns non-zero if the row was found, zero (0) if
    the row was not found */
int row_remove(struct row_t *row);

/** Returns the number of rows in a table */
int table_count(struct schent_t *sch);
    
#endif
