/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2003 Intersil Americas Inc. All Rights Reserved.
 */

#include <ctype.h>
#include <fcntl.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <arpa/inet.h>
#include <arpa/tftp.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <asm/socket.h>


#include "socket.h"
#include "object.h"
#include "misc.h"
#include "schema.h"
#include "objdata.h"
#include "types.h"
#include "journal.h"
#include "handler.h"
#include "table.h"
#include "serialize.h"
#include "debug.h"
#include "range.h"
#include "objflag.h"
#include "aplock.h"

#include "iface_snmp_hexstring.h"
#include "iface_snmp_community.h"
#include "configfile.h"

#define MAX_INDEXLENGTH 20
#define MAX_INDEX 10
#define MAX_STRINGLENGTH_IN_INDEX 32

extern char *typestr[];
struct sockaddr_in server_sockaddr;
int snmp_lock_state = SNMP_LOCKSTATE_UNLOCKED;

int socket_open(int port)
{
	int sd;
	int setreuse = 1;

	if(( sd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP )) == -1 ) {
		fprintf(stderr, "Can't open socket : %s\n", strerror(errno));
		return(-1);
	}

	if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, &setreuse,
     		sizeof(setreuse)) == -1) {
		fprintf(stderr, "Can't set socketoption: %s\n", strerror(errno));
	}

	
	server_sockaddr.sin_family = AF_INET;
	server_sockaddr.sin_port = htons(port);
	inet_pton(AF_INET, "127.0.0.1", &server_sockaddr.sin_addr);
	
	bzero( &server_sockaddr.sin_zero, sizeof( server_sockaddr.sin_zero ) );

	if( bind( sd, (struct sockaddr *)&server_sockaddr, sizeof( server_sockaddr ) ) == -1 ) {
		close(sd);
	
		fprintf(stderr, "Can't bind to socket : %s\n", strerror(errno));
		return(-1);
	}

	return(sd);	
}

int socket_handle(int sd)
{
	char buf[256];
	char outbuf[256];

	struct sockaddr_in from;
	int address_len;
	int recv_length;
	int r;
	char *p;

	from.sin_family = AF_INET;
	from.sin_port = htons(0);
	from.sin_addr.s_addr = htonl(0);
	bzero( &from.sin_zero, sizeof( from.sin_zero ) );

	address_len = sizeof( struct sockaddr_in );
	recv_length = recvfrom(sd, buf, sizeof(buf), 0, (struct sockaddr *)&from, &address_len );

	if( recv_length == -1 ) {
		fprintf(stderr, "Recvfrom failed : %s\n", strerror(errno));
		return(-1);
	}

	p = strchr(buf, '\n');
	if(p)* p=0;
	p = strchr(buf, '\r');
	if(p)* p=0;

	dbg(DBG_NETWORK, "Rx: '%s'", buf);

	r = parse_command(buf, outbuf);

	if(r != 0) {
		strncpy(outbuf, "Error\n", sizeof(outbuf));
	}
	
	sendto(sd, outbuf, strlen(outbuf), 0, (struct sockaddr *)&from, address_len);

	p = strchr(outbuf, '\n');
	if(p) *p=0;

	dbg(DBG_NETWORK, "Tx: '%s'", outbuf);

	return(0);
}


int set_object_by_index(struct object_t* obj, 
		         char **index_arr, 
		         int start, 
		         int length){
	char stringbuf[MAX_STRINGLENGTH_IN_INDEX];
	int i;
	int ret = R_NOERR;
	
	assert(length <= MAX_STRINGLENGTH_IN_INDEX);
	
	switch(obj->sch->type) {
		case T_INTEGER:
		case T_UNSIGNED:
		case T_GAUGE:
			assert(length == 1);
			ret = deserialize(obj, index_arr[start]);
			break;
		case T_IPADDR:
			assert(length == 4);
			snprintf(stringbuf, sizeof(stringbuf), "%s.%s.%s.%s", 
					index_arr[start], 
					index_arr[start+1], 
					index_arr[start+2], 
					index_arr[start+3]);
			
			ret = deserialize(obj, stringbuf);
			break;
		case T_OCTETSTR:
			for(i=0; i < length; i++) {
				stringbuf[i] = atoi(index_arr[start+i]);
			}			
			stringbuf[i] = '\0';
			ret = deserialize(obj, stringbuf);
			break;

		default:
			fprintf(stderr, "type unsupported in index\n");
			die();
			break;
		
	}	
	
	return (ret);
}

/* returns the indexlength of a sch(type). 
 * Returns 0 if the type has a variable length
 * Returns -1 if the type is not supported in a index
 */
int indexlength(struct schent_t *sch){
	int length;

	assert(sch);
	
	switch(sch->type) {
		case T_UNKNOWN:
		case T_TABLE:
		case T_ROWSTAT:
		case T_OBJECTID:
		case T_OPAQUE:
		case T_NOTIFY:
			// These types are not supported (yet) in a index
			return (-1); // error
			break;
		case T_INTEGER:
		case T_UNSIGNED:
		case T_COUNTER:
		case T_COUNTER64:
		case T_GAUGE:
		case T_TICKS:
			return 1;
			break;
		case T_IPADDR:
			return 4;
			break;
		case T_OCTETSTR:
			length = range_fixed_length(sch);
			if (length != -1) {
				return length;
			}
			else {
				return 0; // variable
			}
			break;
		default:
			assert(1==0);
			return (1);
			break;
	
	}
}

/* created a new row and fill in the indexfields using the index in oid */
int create_row(struct schent_t* schent, char* oid){
	
	struct row_t* row;
	struct indexoid_t *indexlist;
	char oidbuf[256];

	struct object_t* obj;
	struct schent_t* sch;
	
	int i;
	char* index_arr[MAX_INDEXLENGTH];
	char* index;
	int parts;
	int index_i;
	int total_length;
	int oid_i;
	
	int length_arr[MAX_INDEX];
	
	int variable_field = -1;
	
	dbg(DBG_TABLE, "Creating Row with rowstatus oid: %s", oid);
	
	row = row_create_by_oid(oid);
	assert(row);
	
	index = strdup(row->oidpart + 1); 
	parts = split(index, '.', index_arr, MAX_INDEXLENGTH);	
	
	indexlist = schent->indexlist;
	assert(indexlist);

	index_i = 0;
	variable_field = -1;
	total_length = 0;
	while(indexlist) {
		sch = schema_get(indexlist->oid);
		assert(sch);
		
		length_arr[index_i] = indexlength(sch);
		
		assert(length_arr[index_i] != -1);
		
		if(length_arr[index_i] == 0) {
			assert(variable_field == -1);
			variable_field = index_i;
		}
		else {
			total_length += length_arr[index_i];
		}

		indexlist = indexlist->next;
		index_i++;
	}
	if(variable_field != -1) {
		length_arr[variable_field] = parts - total_length;
	}
	else {
		if(parts != total_length){
			dbg(DBG_TABLE, "Index for row create is incorrect");
			row_delete(row);
			return (0);
		}
	}
	

	oid_i = 0;
	indexlist = schent->indexlist;
	for(i = 0; i < index_i; i++) {
		sch = schema_get(indexlist->oid);
		
		if(schema_are_related(sch, row->sch)) {
			snprintf(oidbuf, sizeof(oidbuf), "%s%s", sch->oid, row->oidpart);
			obj = object_get(oidbuf);
			assert(obj);
			if (set_object_by_index(obj, index_arr, oid_i, length_arr[i]) != R_NOERR) {
				row_delete(row);
				return (0);
			}
		}
		else {
			// external index
			// TODO check for existance
		}				
		
		oid_i += length_arr[i];	
		
		indexlist = indexlist->next;
	}
	
	journal_rowadd(row);
	
	return(1);
}

int parse_command(char *inbuf, char *outbuf)
{

	char *cmd;
	char *args;
	char *oid = NULL;
	int tid;	
	int type;
	char *data = NULL;
	int len;
	char buf[256];
	u_char community[256];
	size_t  community_len;
	int tmp;
	int i;
	int j;
	int hasempty;
	char *part[5];
	int parts;

	struct object_t *obj;
	struct schent_t *sch;
	struct row_t *row;

	parts = split(inbuf, ' ', part, 2);

	obj = NULL;

	if(parts != 2) return(-1);

	cmd = part[0];
	args = part[1];
					
	if(strcmp(cmd, "GETOID") == 0) {
		parts = split(args, ' ', part, 3);

		if(parts != 3) return(-1);
	
		tid = atoi(part[0]);
		oid = part[1];

		// KICKSTART HACK 
		if (strncmp(oid, ".1.2.840.10036.2.1.1.1.1", 24) == 0) {
			oid[23] = '3';
			
		} 
		else if (strncmp(oid, ".1.2.840.10036.1.1.1.9.1", 24) == 0) {
			oid[23] = '3';
		}
		
		community_len = string_decode(part[2], community, 255);
		
		if (iface_snmp_permission_granted(oid, 
				                  community,
				                  community_len,
				                  A_READ)) {
			handler_call(oid, H_GET, 0);
			obj = object_get(oid);
			
			if (obj) {
				/* check data */
				/* Show only objects that have the correct access type. 
				 * Object with access A_WRITE only should be visible to 
				 * the outside as being a known OID, but the data inside 
				 * should be hidden. (The hiding is done at the end of 
				 * this function)
				 * Only objects with data will be return. Otherwise
				 * OID not found will be returned
				 */
				if ((!(obj->flags & OF_DATAPRESENT))
				     	|| (!(obj->sch->access & (A_READ | A_CREATE | A_WRITE)))
					) {
					obj = NULL;
					obj_errno = R_NOSUCHOID;
				}
			}
		}
		else {
			obj = NULL;
			obj_errno = R_NOACCESS;
		}
		
	}
	
	else if(strcmp(cmd, "GETNEXTOID") == 0) {
		char oid_buf[MAX_OID_LENGTH];
		parts = split(args, ' ', part, 3);

		if(parts != 3) return(-1);
	
		tid = atoi(part[0]);
		oid = part[1];

		community_len = string_decode(part[2], community, 255);
		handler_call(oid, H_GET, 0);
				
		obj = object_get_next(oid);
		
		while(obj 
			&& object_get_oid(oid_buf, sizeof(oid_buf), obj)
			&& ((!iface_snmp_permission_granted(oid_buf, 
						           community,
						           community_len,
						           A_READ))
				/* Show only objects that have the correct access type. 
				 * Object with access A_WRITE only should be visible to 
				 * the outside as being a known OID, but the data inside 
				 * should be hidden. (The hiding is done at the end of 
				 * this function)
				 */
				|| (!(obj->sch->access & (A_READ | A_CREATE | A_WRITE)))
				|| (!(obj->flags & OF_DATAPRESENT))
			      ) ) {
			obj = object_get_next(oid_buf);
		}
	}
	
	else if(strcmp(cmd, "SETOID") == 0) {

		parts = split(args, ' ', part, 5);

		if(parts != 5) return(-1);
	
		tid  = atoi(part[0]);
		oid  = part[1];
		
		community_len = string_decode(part[2], community, 255);
		
		if ((snmp_lock_state == SNMP_LOCKSTATE_UNLOCKED) &&
				iface_snmp_permission_granted(oid, 
				                  community,
				                  community_len,
				                  A_WRITE)) {
			type = atoi(part[3]);
			data = part[4];
			len  = strlen(data);

			/* Call the get handler to be sure the object is up-to-date 
			 * It seems a bit strange to call the GET handler, the object
			 * will be overwritten by the SET anyway. 
			 * However, calling the GET could CREATE the object we're
			 * setting.
			 */
			handler_call(oid, H_GET, 0);

			// Get the object, create row is this is a rowstatus object
			obj = object_get(oid);

			if(obj) {
				assert(obj->sch);

				/* check type */
				if ((type != obj->sch->type) && !(
						((type == T_INTEGER) && (obj->sch->type == T_ROWSTAT))
						|| ((type == T_OCTETSTR) && (obj->sch->type == T_BITS))
						)) {
					/* wrong type */
					obj_errno = R_WRONGTYPE;
					obj = NULL;
				}
				else if (((obj->sch->access & A_WRITE) != A_WRITE) 
						&& ((obj->sch->access & A_CREATE) != A_CREATE)){
					/* no write/create access to this obj */
					obj_errno = R_NOACCESS;
					obj = NULL;
				}
				/* check for valid rowstatus state changes */
				else if(obj->sch->type == T_ROWSTAT) {
					int curr_rowstat, new_rowstat;
					assert(obj->flags & OF_DATAPRESENT);
					curr_rowstat = object_get_rowstat(obj);	
					new_rowstat = atoi(data);
					
					switch(new_rowstat) {
					case RS_NOTREADY:
					case RS_CREATEGO:
					case RS_CREATEWAIT:
						/* This RS is not 'settable' on a existing object */
						obj_errno = R_OUTOFRANGE;
						obj = NULL;
						break;
					case RS_ACTIVE:
					case RS_NOTINSERVICE:
						switch(curr_rowstat) {
						case RS_ACTIVE:
						case RS_NOTINSERVICE:
							break;
						default:
							obj_errno = R_OUTOFRANGE;
							obj = NULL;
					        }
					case RS_DESTROY:
						/* always correct */
						break;
					default:
						/* incorrect rowstatus */
						obj_errno = R_OUTOFRANGE;
						obj = NULL;
					}									
								
				}
			
			}
			else {
				/* object not found, check for rowstatus */
				sch = schema_find_schent_for_oid(oid);
				
				if(sch && (sch->type == T_ROWSTAT)) {

					// It's a set on a rowstatus. Peek in the 
					// value, and if it's 4 or 5 create the row
					assert(data);
					tmp = atoi(data);

					if((tmp == RS_CREATEGO) || (tmp == RS_CREATEWAIT)) {
						
						/* find parent */
						while(sch->prev) {
							sch = sch->prev;
						}
						
						if (create_row(sch, oid)) {
							obj = object_get(oid);
						}
						else {
							obj = NULL;
							obj_errno = R_OUTOFRANGE;
						}

					}
				}
			}
			
			// At this point 'obj' should exist, otherwise there was 
			// an error in the SNMP request (i.e. incorrect OID, no access)
			if(obj) {
				/* store object in journal */
				journal_objset(obj);
				
				/* try to store the data into the object */
				obj_errno = deserialize(obj, data);
				
				if (obj_errno == R_NOERR) {
							
  					object_flag_set(obj, OF_DIRTY);
//					handler_call(oid, H_SET, 3);
//					configfile_schedule_write();

					// Check if this set is on a table row that is not
					// ready yet. If it completes the row, change the
					// rowstatus
					row = obj->row;

					if(row) {
					
						/* check for delete */
						if((obj->sch->type == T_ROWSTAT) && (object_get_rowstat(obj) == RS_DESTROY)) {
							dbg(DBG_TABLE, "Removing row with rowstatus oid %s%s", obj->sch->oid, obj->row ? obj->row->oidpart : ".0" );
							journal_rowdel(row);
							row_remove(row);											
						}
						else {
					
							hasempty = 0;

							/* Check for empty, writable fields */
							for(i=0; i<row->columns; i++) {
								if((!(row->objs[i].flags & OF_DATAPRESENT)) 
										&& (row->objs[i].sch->access & A_CREATE)) {
									hasempty = 1;
								}
							}
						
							if(hasempty != 1) {
								for(i=0; i < row->columns; i++) {
									if(row->objs[i].sch->type == T_ROWSTAT) {
	
										j = object_get_rowstat(&(row->objs[i]));

										if(j == RS_CREATEGO) {
											dbg(DBG_TABLE, "Setting rowstatus %s%s to RS_ACTIVE", row->objs[i].sch->oid, row->oidpart);
											object_set_rowstat(&(row->objs[i]), RS_ACTIVE);
										}
	
										if(j == RS_CREATEWAIT) {
											dbg(DBG_TABLE, "Setting rowstatus %s%s to RS_NOTINSERVICE", row->objs[i].sch->oid, row->oidpart);
											object_set_rowstat(&(row->objs[i]), RS_NOTINSERVICE);
										}
	
									}
								}
							}
						}
	
					}
				}

			}
		}
		else {
			obj = NULL;
			obj_errno = R_NOACCESS;
		}
		
	}
	else if(strcmp(cmd, "BEGIN") == 0) {
		parts = split(args, ' ', part, 1);
		if(parts != 1) return(-1);
		tid = atoi(part[0]);
	
		journal_begin();
	
		obj_errno = R_NOERR;
	}
	else if(strcmp(cmd, "COMMIT") == 0) {
		struct object_t *obj;
		struct jent_t *journal;
		char oid_buf[MAX_OID_LENGTH];
	
		parts = split(args, ' ', part, 1);
	
		if(parts != 1) 
		{
			return(-1);
		}
		tid = atoi(part[0]);
		journal = journal_getfirstentry();
		obj_errno = R_NOERR;
		while(journal)
		{
			if (journal->action == J_OBJSET) {
				int ret_val;
				
				obj = journal->obj;
				
				object_get_oid(oid_buf, sizeof(oid_buf), obj);
				
				ret_val = handler_call(oid_buf, H_CHK, 0);
				
				if (ret_val != R_NOERR) {
					obj_errno = ret_val;
					break;
				}
			}
			
			journal = journal->next;
		}
		
		if(obj_errno == R_NOERR) {
			
			/* schedule all set handlers */
			journal = journal_getfirstentry();
			while(journal)
			{
				if (journal->action == J_OBJSET) {
					obj = journal->obj;
					
					object_get_oid(oid_buf, sizeof(oid_buf), obj);
					
					handler_call(oid_buf, H_SET, 3);
				}
				journal = journal->next;
			}
			
			journal_commit();
			configfile_schedule_write();
			
			obj = NULL; /* to prevent oid & data in return message */
		}
		
	}
	else if(strcmp(cmd, "ROLLBACK") == 0) {
		parts = split(args, ' ', part, 1);
		if(parts != 1) return(-1);
		tid = atoi(part[0]);
	
		journal_rollback();
	
		obj_errno = R_NOERR;
	}

	else {
		return(-1);
	}

	if(obj) {
		/* Only show data when there is data and the A_READ or A_CREATE bit is set */
		if((obj->flags & OF_DATAPRESENT) && (obj->sch->access & (A_READ | A_CREATE))) {
			serialize(obj, buf, sizeof(buf));
			sprintf(outbuf, "%d %d %s%s %d %s", obj_errno, tid, obj->sch->oid, obj->row ? obj->row->oidpart : ".0" , obj->sch->type, buf);
		} else {
			sprintf(outbuf, "%d %d %s%s %d", obj_errno, tid, obj->sch->oid, obj->row ? obj->row->oidpart : ".0", obj->sch->type);
		}
	} 
	else 
	{
		sprintf(outbuf, "%d %d", obj_errno, tid);
	}
	
	return(0);
	
}


//eot
