/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2003 Intersil Americas Inc. All Rights Reserved.
 */

#include <string.h>
#include "iface_snmp_community.h"
#include <stdarg.h>
#include "object.h"
#include "objdata.h"
#include "types.h"
#include <assert.h>

#define PUBLIC   1
#define PRIVATE  2
#define DISCOVER 3

static int community2view(unsigned char* community, int community_len);

struct community_entry_t {
	unsigned char view;
	char   oid[255];
	unsigned char access;
};

static struct community_entry_t accesslist[] = {
	{ PUBLIC,   ".1.3.6.1.4.1.937.2.1.2.2", A_NOACCESS },
	{ PUBLIC,   ".1"                      , A_READ },
	{ PRIVATE,  ".1"                      , A_READ | A_WRITE },
	{ DISCOVER, ".1.3.6.1.2.1.1.5"        , A_READ },
	{ DISCOVER, ".1.3.6.1.2.1.1.2"        , A_READ },
	{ DISCOVER, ".1.3.6.1.2.1.1.6"        , A_READ },
	{ DISCOVER, ".1.3.6.1.2.1.1.4"        , A_READ },
	{ DISCOVER, ".1.2.840.10036.1.1.1.9"  , A_READ },
	{ DISCOVER, ".1.2.840.10036.2.1.1.1"  , A_READ },
	{ DISCOVER, ".1.3.6.1.4.1.937.2.1.2.5", A_READ },
	{ DISCOVER, ".1.3.6.1.4.1.937.2.1.2.7", A_READ },
	{ DISCOVER, ".1.3.6.1.4.1.937.2.1.11.2.1.0", A_READ },
	{ 0,        ""                        , 0                      }			
};


int iface_snmp_permission_granted(char* oid, 
                                  unsigned char* community, 
				  int community_len, 
				  unsigned char access){
	unsigned char view;
	int index = 0;
	
	view = community2view(community, community_len);
	
	if (view == 0) {
		return (IFACE_SNMP_PERM_DENIED);
	}
	
	while(accesslist[index].view != 0) {
		
		/* TODO make strlen static, because it is fixed during compiletime */
		if ( (accesslist[index].view == view)
				&& (strncmp(accesslist[index].oid, oid, strlen(accesslist[index].oid)) == 0)) {
			/* check access */
			if ((access & accesslist[index].access) == access) {
				return IFACE_SNMP_PERM_GRANTED;
			}
			else {
				return IFACE_SNMP_PERM_DENIED;
			}
		}
		
	
		
		index++;
	}		
	
	
	/* no match */
	return IFACE_SNMP_PERM_DENIED;	
}



static int community2view(unsigned char* community, int community_len) 
{
	static struct object_t *rc_obj = NULL;
	static struct object_t *wc_obj = NULL;
	struct string_t *rc_str = NULL;
	struct string_t *wc_str = NULL;
	
	char* read_comm_str = NULL;
	char* write_comm_str = NULL;
	
	
	if(rc_obj == NULL) rc_obj = object_get(".1.3.6.1.4.1.937.2.1.2.1.0");
	if(wc_obj == NULL) wc_obj = object_get(".1.3.6.1.4.1.937.2.1.2.2.0");

	assert(rc_obj);
	assert(wc_obj);
	
	rc_str = object_get_string(rc_obj);
	wc_str = object_get_string(wc_obj);

	assert(rc_str);
	assert(wc_str);
			
	read_comm_str  = rc_str->data;
	write_comm_str = wc_str->data;
	
	if ((strncmp(community, write_comm_str, community_len) == 0) 
			&& (community_len == strlen(write_comm_str))) {
		return PRIVATE;
	}			
	else if ((strncmp(community, read_comm_str, community_len) == 0)
			&& (community_len == strlen(read_comm_str))) {
		return PUBLIC;
	}
	else if ((strncmp(community, "discovery", community_len) == 0)
			&& (community_len == 9)){
		return DISCOVER;
	}
	else {
		return 0;
	}
}
