/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2003 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "aplog.h"
#include "pda.h"
#include "types.h"
#include "regulatory.h"
#include "misc.h"
#include "easyprism.h"
#include "calibration.h"

//#define DEBUG_PDAREAD
//#define DEBUG_OUTPUTTABLE
//#define DEBUG_CHANNELLIST

struct dot11d_channel 
{
    unsigned short 	freq;
    signed char 	power;
    char		fill0;
};

struct sys_point
{
    signed char pow;
    signed char det;		// might be unsigned...
    unsigned char ctr;
} __attribute__ ((packed));

struct sys_channel
{
    unsigned short frequency;
    signed char max_cck;
    signed char max_bpsk;
    signed char max_qpsk;
    signed char max_qam16;
    signed char max_qam64;
    struct sys_point point[1];
};

struct pdr_1903
{
    unsigned short f;
    signed char max[6];
};

struct sys_power
{
    char *table;		// power table
    char size;			// size of table entries
    char channels;		// number of channels
    char points;		// curve calibration points
    char olm;			// open loop mask
};

static struct sys_power sys_power3; /* power table for interface 3 */
static struct sys_power sys_power5; /* power table for interface 5 */

static int user_power3 = -1; /* User configured max output power in mWatt */
static int user_power5 = -1; /* User configured max output power in mWatt */


void calibration_set_tx_powerlevel(int iface, int power) {
	
	switch(iface) {
		case 3:
			user_power3 = power;
			calibration_update_outputpowertable(iface);
			break;
		case 5:
			user_power5 = power;
			/* TODO outputpower to uAP */
			break;
		default:
			die();
			break;
	}
	
}


/*
 * load the powertable into sys_power.
 */
int calibration_load_powertable(int ifnum)
{
	int curve, size, i;
	char *buf, *ptr;
	struct pdr_1903 *pdr_ptr;
	struct sys_channel *sys_ptr;
	struct Pdr_Struct *pdr_str_Ptr, *pda_start_ptr = NULL;
	struct sys_power* sys_power_p;

	switch(ifnum) {
		case 3:
			pda_start_ptr = NULL;
			sys_power_p = &(sys_power3);
			break;
		case 5:
			/* TODO set pointer to start of cached PDA of micro AP */
			sys_power_p = &(sys_power5);
			break;
		default:
			die();
			break;
	}
			
	/* get length of output limit PDR */
	pdr_str_Ptr = pda_Find_Pdr(PDR_PRISM_PA_CAL_CURVE_DATA, pda_start_ptr);
	if(!pdr_str_Ptr) {
		sys_power_p->channels = 0;
		return R_NONEXIST;
	}


	sys_power_p->size     = (8 + 3 * pdr_str_Ptr->PdrData[2]) & ~1;
	sys_power_p->channels = pdr_str_Ptr->PdrData[1];
	sys_power_p->points   = pdr_str_Ptr->PdrData[2];
	sys_power_p->olm      = pdr_str_Ptr->PdrData[3];
	
#ifdef DEBUG_PDAREAD
	fprintf(stderr, "sys_power_p->size     = %i\n", sys_power_p->size);
	fprintf(stderr, "sys_power_p->channels = %i\n", sys_power_p->channels);
	fprintf(stderr, "sys_power_p->points   = %i\n", sys_power_p->points);
	fprintf(stderr, "sys_power_p->olm      = %i\n", sys_power_p->olm);
#endif	

	curve = (3 + 3 * pdr_str_Ptr->PdrData[2]) & ~1; // size of curve data per channel

#ifdef DEBUG_PDAREAD
	fprintf(stderr, "curve                 = %i\n", curve);	
#endif
	/* allocate the powertable */
	sys_power_p->table = (char *)malloc( sys_power_p->channels * sys_power_p->size );
	if( !sys_power_p->table ) {
		goto mem_error;
	}

	if( (buf = (char *)malloc( sys_power_p->channels * sys_power_p->size )) == NULL ) {
		free( sys_power_p->table );
mem_error:
		sys_power_p->channels = 0;
		return R_MEMORYFAIL;
	}

	/* load power limits */
	size = sys_power_p->channels * 8 + 2;

	pdr_str_Ptr = pda_Find_Pdr(PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS, pda_start_ptr);
	if(!pdr_str_Ptr || pdr_str_Ptr->PdrData[0]) {
		goto error;
	}

	pdr_ptr = (struct pdr_1903 *)(pdr_str_Ptr->PdrData + 2);
	for( i=0; i<sys_power_p->channels; i++ ) {

		sys_ptr = (struct sys_channel *)(sys_power_p->table + i * sys_power_p->size);

		sys_ptr->frequency = pdr_ptr->f;
		sys_ptr->max_cck   = pdr_ptr->max[5];
		sys_ptr->max_bpsk  = pdr_ptr->max[0];
		sys_ptr->max_qpsk  = pdr_ptr->max[1];
		sys_ptr->max_qam16 = pdr_ptr->max[2];
		sys_ptr->max_qam64 = pdr_ptr->max[3];
#ifdef DEBUG_PDAREAD
		fprintf(stderr, "          i = %i\n", i);
		fprintf(stderr, "->frequency = %i\n", sys_ptr->frequency);	
		fprintf(stderr, "->max_cck   = %i\n", sys_ptr->max_cck);	
		fprintf(stderr, "->max_bpsk  = %i\n", sys_ptr->max_bpsk);	
		fprintf(stderr, "->max_qpsk  = %i\n", sys_ptr->max_qpsk);	
		fprintf(stderr, "->max_qam16 = %i\n", sys_ptr->max_qam16);	
		fprintf(stderr, "->max_qam64 = %i\n", sys_ptr->max_qam64);	
#endif				
		pdr_ptr++;
		
	}

	/* load curve data */
	size = sys_power_p->channels * curve + 4;
	pdr_str_Ptr = pda_Find_Pdr(PDR_PRISM_PA_CAL_CURVE_DATA, pda_start_ptr);

	if(!pdr_str_Ptr || pdr_str_Ptr->PdrData[0]) {
		goto error;
	}

	ptr = pdr_str_Ptr->PdrData + 4;
	for( i=0; i<sys_power_p->channels; i++ ) {
				
		sys_ptr = (struct sys_channel *)(sys_power_p->table + i * sys_power_p->size);
		memcpy( &sys_ptr->point, ptr + 2, 3 * sys_power_p->points );
#ifdef DEBUG_PDAREAD
		fprintf(stderr, "CHANNEL = %i\n", *ptr + ((*(ptr+1)) << 8));
		{
			int j;
			for (j = 0; j < sys_power_p->points; j++) {
				fprintf(stderr, "sys_ptr->point[j].pow = %i\n", sys_ptr->point[j].pow);
				fprintf(stderr, "sys_ptr->point[j].det = %i\n", sys_ptr->point[j].det);
				fprintf(stderr, "sys_ptr->point[j].ctr = %i\n", sys_ptr->point[j].ctr);								
			}
		}
#endif		
		ptr += curve;
	}

	free( buf );
	return R_NOERR;

error:
	free( buf );
	free( sys_power_p->table );
	sys_power_p->channels = 0;
	return R_CHECKFAILED;
}

/* TODO move this */
#define min(x, y) (x < y ? x : y )

/* accumulator backoff to compensate for detector curve cal @ 54Mbps. */
static unsigned char acc_backoff[5] = { 50, 36, 24, 12, 0 };

/*
 * power_interpolate: do the interpolation for max power
 */
static void power_interpolate( struct sys_channel *cal, int pts,
			       signed char max, struct obj_powertable *table,
			       struct sys_power *sys_power_p )
{
	struct obj_power *result = &table->cck;
	signed char *limit = &cal->max_cck;
	int mod;
    
	for( mod=0; mod<5; mod++ ) {
		struct sys_point *this = cal->point, *last = NULL;
		signed char power = min( max, limit[mod] );
		signed char *set_point = &result[mod].set_point;
		unsigned char *accumulator = &result[mod].accumulator;
		int pt;
		
		/* This function is only called for the internal radio so we
		 * know which user power should be used */
		power = min(power, mW2dBm(user_power3) * 4);
		
		for( pt=0; pt<pts; pt++ ) {
			if( power >= this->pow ) {
				if( !last || (this->det == sys_power_p->olm) ) {
					*set_point   = this->det;
					*accumulator = this->ctr;
				}
				else
				{
					/* set_point may be signed or unsigned. Do computations 
					 * treating them as unsigned after adding 128, and subtract
					 * these 128 before writing the result. accumulator is 
					 * defined to be unsigned. */
					int ldiff = power     - this->pow;
					int pdiff = last->pow - this->pow;

					int this_sp = this->det + 128;
					int last_sp = last->det + 128;
					int my_sp = this->det + 128 
							+ (ldiff * (last_sp - this_sp)) / pdiff;

					*set_point   = my_sp - 128;
					*accumulator = this->ctr + 
						(ldiff * (last->ctr - this->ctr)) / pdiff;
				}
				break;
	    		}
			last = this;
			this++;
		}
		if( pt == pts )	{
			*set_point   = last->det;	// clip to bottom
			*accumulator = last->ctr;
		}
		
		if( acc_backoff[mod] > *accumulator ) {
			*accumulator = 0;
		}
		else {
			*accumulator -= acc_backoff[mod];
		}
    	}
}

/*
 * sys_adjust_powertable: adjust the powertable based on the dot11d channel
 * list and calibration data. Configure the blob.
 */
static int sys_adjust_powertable(int ifnum, struct dot11d_channel *ch, int max )
{
	struct obj_powertable *bbuf;
	int i, j;
	static char errbuf[160];
	static struct sys_power* sys_power_p = NULL;

	switch(ifnum) {
		case 3:
			sys_power_p = &(sys_power3);
			break;
		case 5:
			sys_power_p = &(sys_power5);
			break;
		default:
			die();
			break;
	}
	
	/* do nothing if no calibration data present. */
	if( !sys_power_p || !sys_power_p->channels ) {
		return R_NONEXIST;
	}

	/* alloc buffer */
	bbuf = (struct obj_powertable *) malloc( max * sizeof(struct obj_powertable) );
    	if( !bbuf )
		return R_MEMORYFAIL;

	/* setup the table */
    	for( i=0; i<max; i++ ) {
		bbuf[i].frequency = ch[i].freq;

		/* look for the channel's power calibration */
		for( j=0; j<sys_power_p->channels; j++ ) {
	    		struct sys_channel *sys_ptr = 
				(struct sys_channel *)(sys_power_p->table + j * sys_power_p->size);
	    
			if( sys_ptr->frequency == bbuf[i].frequency ) {
				power_interpolate( sys_ptr, sys_power_p->points, ch[i].power, bbuf + i , sys_power_p);
				break;
	    		}
		}
	
		/* if no calibration data is found, use open loop mode */
		if( j == sys_power_p->channels ) {
			struct obj_power *pow = &bbuf[i].cck;

			for( j=0; j<5; j++ ) {
				pow[j].set_point = sys_power_p->olm;	// open loop
				pow[j].accumulator = 0x40;		// '25%' power
	    		}
			
			snprintf(errbuf, sizeof(errbuf), "WARNING: no calibration data for frequency %d", bbuf[i].frequency );
			aplog(EV_SNMP_GENERAL_ERROR, errbuf, NULL);
		}
    	}

    	/* configure the blob */
	/* FIX THIS. Interface should be configurable */

#ifdef DEBUG_OUTPUTTABLE
	{
		int i;
		fprintf(stderr, "START DUMP OUTPUTPOWERTABLE\n");
		for(i = 0; i < max; i++) {
			fprintf(stderr, "               i  = %i\n", i);
			fprintf(stderr, "frequency         = %i\n", bbuf[i].frequency);
			fprintf(stderr, "cck.set_point     = %i\n", bbuf[i].cck.set_point);
			fprintf(stderr, "cck.accumulator   = %i\n", bbuf[i].cck.accumulator);			
			fprintf(stderr, "bpsk.set_point    = %i\n", bbuf[i].bpsk.set_point);
			fprintf(stderr, "bpsk.accumulator  = %i\n", bbuf[i].bpsk.accumulator);			
			fprintf(stderr, "qpsk.set_point    = %i\n", bbuf[i].qpsk.set_point);
			fprintf(stderr, "qpsk.accumulator  = %i\n", bbuf[i].qpsk.accumulator);			
			fprintf(stderr, "qam16.set_point   = %i\n", bbuf[i].qam16.set_point);
			fprintf(stderr, "qam16.accumulator = %i\n", bbuf[i].qam16.accumulator);			
			fprintf(stderr, "qam64.set_point   = %i\n", bbuf[i].qam64.set_point);
			fprintf(stderr, "qam64.accumulator = %i\n", bbuf[i].qam64.accumulator);			
		} 
		
		fprintf(stderr, "END DUMP OUTPUTPOWERTABLE\n");	}

	{
		int i;
		u_char *p;
		fprintf(stderr, "START DUMP HEX OUTPUTPOWERTABLE \n");
		p = (u_char*) bbuf;
		for(i = 0; i < max * (sizeof(struct obj_powertable)); i++ ) {
			fprintf(stderr, "%02X", p[i]);
		}
		fprintf(stderr, "\n");
		fprintf(stderr, "END DUMP HEX OUTPUTPOWERTABLE\n");
	}		
#endif		
		
    	set_prismoid("eth1", DOT11_OID_OUTPUTPOWERTABLE, 
		    (long *)bbuf, max * sizeof( struct obj_powertable ) );
    	free( bbuf );
    	return R_NOERR;
}

static int create_dot11d_channel_list(int ifnum, struct dot11d_channel *ch, int max) {
	
	int size;
	int freq;
	struct country_domain_t* cd;
	struct regulatory_group_list_t* rgl;
	struct regulatory_group_t* rg;
	struct power_group_t* pg;
	
	cd = regulatory_get_current_country_domain(ifnum);
	assert(cd);

	rgl = cd->regul_groups;
	
	size = 0;
	
	while(rgl) {
		rg = rgl->regul_grp;
		assert(rg);
		pg = rg->power_groups;
		
		while(pg) {
			int i;
			/* walk all frequencies for this power group */
			for(i=0; i < pg->no_freqs; i++) {
				
				freq = pg->frequencies[i];
				
				ch[size].freq = freq;
				/* Power in the channelist is in 1/4 dBm */
				ch[size].power = mW2dBm(pg->max_power) * 4;
				size++;

				if (size == max) {
					return(size);
				}
			}		
			pg=pg->next;
		}				
		rgl = rgl->next;
	}
	
	return(size);
}

#ifdef DEBUG_CHANNELLIST
static void dump_channel_list(struct dot11d_channel* ch, int size) {
	int i;
	
	printf("Channelist dump:\n");
	printf("================\n");
	for(i=0; i < size; i++) {
		printf("[%02i] [f:%04i] [p:%04i]\n", i, ch[i].freq, ch[i].power);
	}
	printf("================\n");
}
#endif

#define MAX_CHANNELLIST_SIZE 64
void calibration_update_outputpowertable(int ifnum) {
	int size;
	struct dot11d_channel channels[MAX_CHANNELLIST_SIZE];
	
	size = create_dot11d_channel_list(ifnum, channels, MAX_CHANNELLIST_SIZE);

#ifdef DEBUG_CHANNELLIST
	dump_channel_list(channels, size);
#endif	
	
	sys_adjust_powertable(ifnum, channels,  size);
		
}
