/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <arpa/inet.h>
#include <assert.h>
#include <sys/types.h>
#include <signal.h>
#include <dirent.h>
#include <ctype.h>
#include <stdlib.h>
#include <aplog.h>
#include "schema.h"
#include "object.h"
#include "objdata.h"
#include "table.h"
#include "types.h"
#include "cfg_file_lib.h"

#define CFG_FILE_DIR "/var/etc/"
#define CFG_FILE_SUFFIX ".conf"
#define READ_BUF_SIZE 64

FILE* create_cfg_file(char* name)
{
	FILE* ret;
	char temp[256];

	snprintf(temp, sizeof(temp), "%s%s%s", CFG_FILE_DIR, name, CFG_FILE_SUFFIX);
	ret = fopen(temp, "w");

	if(ret)
	{
		fprintf(ret, "# %s \n", name);
	}
	else
	{
	   char tmp[256];
	   snprintf(tmp, sizeof(tmp), "Error opening configfile for writing: %s\n", temp);
	   aplog(EV_SNMP_GENERAL_ERROR,tmp,NULL);
	}

	return ret;
}

FILE* open_cfg_file(char* name)
{
	FILE* ret;
	char temp[256];
	snprintf(temp, sizeof(temp), "%s%s%s", CFG_FILE_DIR, name, CFG_FILE_SUFFIX);
	ret = fopen(temp, "r");

	if(!ret) {
		char tmp[256];
		snprintf(tmp, sizeof(tmp), "Error opening configfile for reading: %s\n", temp);
	   	aplog(EV_SNMP_GENERAL_ERROR,tmp,NULL);
	}

	return ret;
}

int close_cfg_file(FILE * fp)
{
	return (fclose(fp));
}

int write_cfg_line_hex(FILE* fp, char* token, struct string_t* valuep)
{
	int ret, i;

	ret = fprintf(fp, "%s ", token);
	/* Convert to HEX */
	ret |= fprintf(fp, "HEX:");
	for(i=0; i < valuep->len; i++) {
		ret |= fprintf(fp, "%02X", valuep->data[i]);
	}
	ret |= fprintf(fp, "\n");

	return ret;
}

int write_cfg_line_str(FILE* fp, char* token, char* value)
{
	int ret;

	ret = fprintf(fp, "%s %s\n", token, value);

	return ret;
}

int write_cfg_line_str_tab(FILE* fp, char* token, char* value)
{
	int ret;

	ret = fprintf(fp, "%s\t%s\n", token, value);

	return ret;
}

int write_cfg_line_int(FILE* fp, char* token, int value)
{
	int ret;

	ret = fprintf(fp, "%s %i\n", token, value);

	return ret;
}

int write_cfg_line_int_tab(FILE* fp, char* token, int value)
{
	int ret;

	ret = fprintf(fp, "%s\t%i\n", token, value);

	return ret;
}

int write_cfg_token(FILE* fp, char* token)
{
	int ret;

	ret = fprintf(fp, "%s ", token);

	return ret;
}

int write_cfg_value(FILE* fp, char* value)
{
	int ret;

	ret = fprintf(fp,"%s ", value);

	return ret;
}

int write_cfg_value_tab(FILE* fp, char* value)
{
	int ret;

	ret = fprintf(fp,"%s ", value);

	return ret;
}

/* find_pid_by_name()
 *
 *  This finds the pid of the specified process.
 *
 *  If more processes exist with the same name, only the first found
 *  pid is returned
 */
static pid_t find_pid_by_name( char* pidName)
{
	DIR *dir;
	struct dirent *next;

	dir = opendir("/proc");
	assert(dir);

	while ((next = readdir(dir)) != NULL) {
		FILE *status;
		char filename[READ_BUF_SIZE];
		char buffer[READ_BUF_SIZE];
		char name[READ_BUF_SIZE];

		/* If it isn't a number, we don't want it */
		if (!isdigit(*next->d_name))
			continue;

		sprintf(filename, "/proc/%s/status", next->d_name);
		if (! (status = fopen(filename, "r")) ) {
			continue;
		}
		if (fgets(buffer, READ_BUF_SIZE-1, status) == NULL) {
			fclose(status);
			continue;
		}
		fclose(status);

		/* Buffer should contain a string like "Name:   binary_name" */
		sscanf(buffer, "%*s %s", name);
		if (strcmp(name, pidName) == 0) {
			return(strtol(next->d_name, NULL, 0));
		}
	}

	return (0);
}

static int SigHup(char* procname) {
	pid_t pid;

	pid = find_pid_by_name(procname);

	return (kill(pid, SIGHUP));
}

#define CONFIGFILE_APLOG "aplog"
int write_config_aplog(void) {

	struct schent_t *sch;
	struct object_t *obj;
	struct row_t *row;
	int prio,  rs, traptype;
	char buf[80];
	struct in_addr *ipaddr;
	FILE* fp;
	int facility;
	u_char* reporting;

	fp = create_cfg_file(CONFIGFILE_APLOG);

	/* EventTable Max Length */
	obj = object_get(".1.3.6.1.4.1.937.2.3.3.3.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	fprintf(fp, "MaxTableLength %u\n", object_get_uint(obj));

	/* EventControl table */
	sch = schema_get(".1.3.6.1.4.1.937.2.3.3.2.1");
	assert(sch);
	row = sch->rowlist;
	while(row) {
		obj = row_get_obj(row,0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		prio = object_get_int(obj);

		obj = row_get_obj(row,1);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);

		reporting = object_get_bits(obj);

		snprintf(buf, sizeof(buf), "%i %02X", prio, reporting[0]);
		write_cfg_line_str(fp, "EventControl", buf);
		row = row->next;
	}

	/* Trap destinations */
	sch = schema_get(".1.3.6.1.4.1.937.2.1.2.4.1");
	assert(sch);

	row = sch->rowlist;
	while(row)
	{
		obj = row_get_obj(row, 3);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);

		rs = object_get_rowstat(obj);

		if (rs == 1) { /* RowStatus == Active */

			obj = row_get_obj(row,1);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			ipaddr = object_get_ipaddr(obj);

			obj = row_get_obj(row,2);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			traptype = object_get_int(obj);

			snprintf(buf, sizeof(buf), "%s %i", inet_ntoa(*ipaddr), traptype);

			write_cfg_line_str(fp, "TrapDestination", buf);

		}

		row = row->next;
	}

	/* Event facilities */
	sch = schema_get(".1.3.6.1.4.1.937.2.3.3.5.1");
	assert(sch);

	row = sch->rowlist;
	while(row)
	{
		obj = row_get_obj(row, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		facility = object_get_int(obj);
		obj = row_get_obj(row, 1);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		reporting = object_get_bits(obj);

		/* Assumption: only 8 bits */
		snprintf(buf, sizeof(buf), "%i %02X", facility, reporting[0]);
		write_cfg_line_str(fp, "Facility", buf);

		row = row->next;
	}



	close_cfg_file(fp);

	SigHup("aplogd");

	return(1);
}
