/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
 
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
 
#include "regulatory.h"
#include "types.h"
#include "schema.h"
#include "objdata.h"
#include "table.h"
#include "assert.h"
#include "object.h"
#include "prismoids.h"

#include "mod_mibII_interfaces.h"


/* TODO define a good value for this */
#define MAX_IE_SIZE 256
static u_char ie[MAX_IE_SIZE];

#define COUNTRY_INFO_ID 0x07

//#define DEBUG 

static void set_ie(struct country_domain_t* cd, char* country_code, int ifnum) {
	char* p;
	struct regulatory_group_list_t* rgl;
	struct power_group_t* pg;
	int no_blocks;
	u_char length;
	char oidbuf[MAX_OID_LENGTH];
	struct object_t* obj;
	int phytype;
	struct obj_attachment* oa;
#ifdef DEBUG	
	int i;
#endif	
	
	assert(cd);
	assert(country_code);
	
	/* get phy_type for this inteface */
	snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.18.%u", ifnum);
	obj = object_get(oidbuf);
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	
	phytype = object_get_int(obj);	

	/* set IE id */
	ie[0] = COUNTRY_INFO_ID;
	
	/* set COUNTRY_CODE */

	ie[2] = country_code[0];
	ie[3] = country_code[1];
	ie[4] = country_code[2];
	
	p = &(ie[5]); /* start of first power/channel group */
	
	no_blocks = 0;
	
	/* create a triplet for each power_group */
	rgl = cd->regul_groups;
	
	while(rgl) {

		pg = rgl->regul_grp->power_groups;
		
		while(pg) {
			/* loop all frequencies in this group */
			
			/* TODO. For now, asume that all frequencies in a power group are 
			   'op-een-volgend'. This should be changed to be more flexible. */
			
			/* only add channels for the correct band */
			if( ((phytype == 6) && (pg->frequencies[0] < 3000))
					||
			    ((phytype == 4) && (pg->frequencies[0] > 3000)) 
				) {
				no_blocks++;
				*(p++) = (u_char) regulatory_freq2channel(pg->frequencies[0]);
				*(p++) = (u_char) pg->no_freqs;
				/* fill in max_power in dBm. Add 0.5 for correct rounding */
				*(p++) = (u_char) (mW2dBm(pg->max_power) + 0.5);  
			}
			
			pg = pg->next;
		}
		rgl = rgl->next;
	}
	
	/* calculate length */
	length = 3 + (no_blocks * 3);
	
	/* check for padding */
	if (length != (length & 0xFE)) {
		/* length is odd, pad with one byte */
		/* set pad field to 0 */
		*(p++) = (u_char) 0x00;
		length++;
	}
	
	ie[1] = length;

#ifdef DEBUG
	printf("IE element for interface %i: \n", ifnum);
	
	for(i=0; i < length + 2; i++) {
		printf("%02X %02i\n", ie[i], ie[i]);
	}
	
	printf("END of IE \n");
#endif
	
	oa = malloc(sizeof(struct obj_attachment) + (length + 1));
	assert(oa);
	
	oa->reserved = 0;
	oa->id = -1;
	oa->size = length + 2; 
	memcpy(oa->data, ie, length + 2);
	
	oa->type = 0x80; // beacon
	set_prismoid(ifindex2ifname(ifnum), DOT11_OID_ATTACHMENT, 
			oa, 
			(sizeof(struct obj_attachment) + length + 1)); 
	oa->type = 0x50; // probe respons
	set_prismoid(ifindex2ifname(ifnum), DOT11_OID_ATTACHMENT, 
			oa, 
			(sizeof(struct obj_attachment) + length + 1)); 
	
	
}


/* This functions sets/updates the current country information element on
 * the given wireless interface, depending on the current configuration */
void country_information_set_ie(int ifnum) {
	
	struct object_t* obj;
	int regdom;
	char* regdom_id;
	struct string_t* country;
	char oidbuf[MAX_OID_LENGTH];
	char* country_code;
	struct country_domain_t* cd;


	/* get desired country for this interface */
	snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.15.%u", ifnum);
	obj = object_get(oidbuf);
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	country = object_get_string(obj);
	
	/* get current regdomain for this interface */
	snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.4.1.1.2.%u", ifnum);
	obj = object_get(oidbuf);
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	regdom = object_get_int(obj);
	/* check if we have a regdom */
	if (regdom != 0x00) {
			
		regdom_id = regulatory_value2domainid(regdom);
			
		cd = regulatory_get_domain(regdom_id);
		assert(cd);
			
		/* check if the desired country is part of this domain */
		if(regulatory_country_in_domain(cd, country->data) == R_NOERR) {
			/* ok, use desired country */
			country_code = country->data;
		}
		else {
			/* use the default country for this domain */
			country_code = cd->default_country;
		}
	}
	else {
		/* use the desired country */
		cd = regulatory_get_country(country->data);
		assert(cd);
			
		/* set the country_code that should be used */
		country_code = cd->id;	
	}
		
	/* country countrystring in dot11StationConfig table */
	snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.23.%u", ifnum);
	obj = object_get(oidbuf);
	assert(obj);
	object_set_string(obj, country_code, 3);
		
	//set_ie(cd, country_code, ifnum);
		
}
