/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>
#include <aplog.h>


#include "misc.h"
#include "easyprism.h"

//#define DEBUG

static char *ifname;
static char oidbuf[9]; /* HHHHHHHH\0 */
static int UAP_present = B_FALSE;

void prism_init(void) {
	
	int fd;
	struct ifreq ifr;
	
	fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	assert(fd != -1);	
	
	strncpy(ifr.ifr_name, "eth3", IF_NAMESIZE);
	
	if (ioctl(fd, SIOCGIFMTU, &ifr) == 0) {
		UAP_present = B_TRUE;
	}
	else {
		UAP_present = B_FALSE;
	}
	
	close(fd);
	
}

void prism_set_ifname(char *name)
{
	if((strcmp("eth3", name) == 0) && (!UAP_present)) {
		ifname = NULL;
	}
	else {
		ifname = name;
	}
}


long prism_get_long(int prismoid)
{
	long data = 0;
	int r;

	if(!ifname) {
		return 0;
	}
	
	r = get_prismoid(ifname, prismoid, &data, sizeof(data));
	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
		//aplog(EV_NET_IFACE_FAILED, ifname, oidbuf, NULL);
		return(0);
	}
	return(data);
}


int prism_set_long(int prismoid, int v)
{
	long data = v;
	int r;
	
	if(!ifname) {
		return 0;
	}

#ifdef DEBUG
	fprintf(stderr, "prism_set_long(%s, %08X, %i)\n", ifname, prismoid, v);	
#endif	
	
	r = set_prismoid(ifname, prismoid, &data, sizeof(data));


	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
		//aplog(EV_NET_IFACE_FAILED, ifname, oidbuf, NULL);
	}

	return(r);
}

int prism_set_ssid(int prismoid, struct string_t *str)
{
	int r;
	struct obj_ssid data;

	if(!ifname) {
		return 0;
	}
	
	memcpy(data.octets, str->data, str->len);
	data.length = str->len;

	r = set_prismoid(ifname, prismoid, &data, sizeof(data));

	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
		//aplog(EV_NET_IFACE_FAILED, ifname, oidbuf, NULL);
	}

	return (r);
}

int prism_set_defaultkey(int prismoid, struct string_t *str)
{
	int r;
	struct obj_key data;

	if(!ifname) {
		return 0;
	}
	
   	data.type = DOT11_PRIV_WEP;
    	data.length = str->len;
#ifdef MVC1_3
        data.domain = 0;
#endif
	memcpy(data.key, str->data, str->len);

	r = set_prismoid(ifname, prismoid, &data, sizeof(data));

	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
	}

	return (r);

}

int prism_set_string(int prismoid, struct string_t *str)
{
	int r;
	char* tmp;

	if(!ifname) {
		return 0;
	}
	
	/* strings in the MVC should be NULL-terminated */
	tmp = calloc(1, str->len + 1);
	memcpy(tmp, str->data, str->len);

#ifdef DEBUG
	fprintf(stderr, "prism_set_string(%s, %08X, %X)\n", ifname, prismoid, tmp);	
#endif	
		
	r = set_prismoid(ifname, prismoid, tmp, str->len + 1);


	free(tmp);

	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
		//aplog(EV_NET_IFACE_FAILED, ifname, oidbuf, NULL);
	}

	return(r);
}



struct string_t *prism_get_mlme_addr(int prismoid)
{
	int r;
	struct obj_mlme data;
	struct string_t *str;

        memset(&data, 0, sizeof(data));
	
	if(!ifname) {
		str = mkstring(data.address, 6);
		return(str);
	}

        r = get_prismoid(ifname, prismoid, &data, sizeof(data));
	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
		//aplog(EV_NET_IFACE_FAILED, ifname, oidbuf, NULL);
	}

	str = mkstring(data.address, 6);

	return(str);
}


long prism_get_mlme_code(int prismoid)
{
	int r;
	struct obj_mlme data;

	data.code = 0;
	
	if(!ifname) {
		return (data.code);
	}

        r = get_prismoid(ifname, prismoid, &data, sizeof(data));
	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
		//aplog(EV_NET_IFACE_FAILED, ifname, oidbuf, NULL);
	}
	return(data.code);
}



struct string_t *prism_get_string(int prismoid, int len)
{
	int r;
	struct string_t *str;

	str = mkstring("", len);
	
	if(!ifname) {
		return(str);
	}

	r = get_prismoid(ifname, prismoid, str->data, str->len);
	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
		//aplog(EV_NET_IFACE_FAILED, ifname, oidbuf, NULL);
		str->len = 0;
	}

	return(str);
}

struct string_t *prism_get_ssid(int prismoid)
{
	int r;
	struct obj_ssid data;
	struct string_t *str;

	memset(&data, 0, sizeof(data));
	if(!ifname) {
		str = mkstring(data.octets, data.length);
		return(str);
	}

        r = get_prismoid(ifname, prismoid, &data, sizeof(data));
	if(r != 0) {
		snprintf(oidbuf, sizeof(oidbuf), "%08X", prismoid);
		//aplog(EV_NET_IFACE_FAILED, ifname, oidbuf, NULL);
	}

	str = mkstring(data.octets, data.length);

	return(str);
}


//eot
