/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include "lib_calls.h"
#include "assert.h"

int IfIoctl(int ioctl_num, struct ifreq *ifr)
{
	int fd, ret;
	
	fd = socket(PF_INET, SOCK_DGRAM, 0);
	assert(fd > 0);

	ret = ioctl(fd, ioctl_num, ifr);
	close(fd);

	return ret;
}


int RunAndRead(char* cmd[]) {
	int pid, pid2;
	int p[2];
	
	if(pipe(p) == -1) {
		fprintf(stderr, "%s: pipe failed\n", __FUNCTION__);
		return -1;
	}
	
	if ( (pid = vfork()) == 0 )
  	{
    		if ( ( pid2 = vfork()) == 0)
    		{
			if (dup2(p[1], STDOUT_FILENO) == -1) {
				fprintf(stderr, "%s: dup2 failed\n", __FUNCTION__);
				return -1;
			}
      			execv(cmd[0], cmd);
			fprintf(stderr, "%s: execv failed\n", __FUNCTION__);
    		}
		_exit(0);
  	}
  	else
  	{
    		/* wait until child 1 ends */
    		waitpid(pid, NULL, 0);
  	}	

	close(p[1]);	
	
	return(p[0]);
}

int StartAppl(char* opt0, char* opt1, char* opt2,char* opt3,char* opt4,char* opt5, char* opt6,char* opt7, char* opt8)
{
	pid_t pid, pid2;
	char *cmd[10];

	cmd[0] = opt0;
	cmd[1] = opt1;
	cmd[2] = opt2;
	cmd[3] = opt3;
	cmd[4] = opt4;
	cmd[5] = opt5;
	cmd[6] = opt6;    
	cmd[7] = opt7;
	cmd[8] = opt8;    
	cmd[9] = NULL;  

	if ((pid = vfork()) == 0) {
		if ((pid2 = vfork()) == 0) {
			int err = execv(cmd[0], cmd);
			exit(err);
		}
		exit(0);
	} else {
		/* wait until child 1 ends */
		waitpid(pid, NULL, 0);
	}
	return(0);
}

int IfUp(char* opt1,char* opt2,char* opt3,char* opt4,char* opt5,char* opt6)
{
  return StartAppl("/bin/ifup", opt1, opt2, opt3, opt4, opt5, opt6, NULL, NULL);
}

int IpTables(char* opt1, char* opt2,char* opt3,char* opt4,char* opt5, char* opt6,char* opt7, char* opt8)
{
  return StartAppl("/bin/iptables", opt1, opt2, opt3, opt4, opt5, opt6, opt7, opt8);
}	

int Kill(char* opt1, char* opt2)
{
  return StartAppl("/bin/kill", opt1, opt2, NULL, NULL, NULL, NULL, NULL, NULL);
} 

int Killall(char* opt1, char* opt2)
{
  return StartAppl("/bin/killall", opt1, opt2, NULL, NULL, NULL, NULL, NULL, NULL);
} 

int Route(char* opt1,char* opt2,char* opt3,char* opt4,char* opt5,char* opt6)
{
  return StartAppl("/bin/route", opt1, opt2, opt3, opt4, opt5, opt6, NULL, NULL);
}

int IfConfig(char* opt1, char* opt2,char* opt3, char* opt4)
{
  return StartAppl("/bin/ifconfig", opt1, opt2, opt3, opt4, NULL, NULL, NULL, NULL);
}
