/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <aplog_defs.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "object.h"
#include "easyprism.h"
#include <sys/fcntl.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include "hardconfig.h"
#include <linux/if_bridge.h>
#include <sys/time.h>
#include <asm/param.h>
#include "mod_bridge.h"

int port_id_to_num(u_int16_t port_id)
{
	return (int)(port_id & 0xff);
}

int port_id_to_prio(u_int16_t port_id)
{
	return (int)((port_id >> 8) & 0xff);
}


char *
idxtoname (unsigned int ifindex, char *ifname)
{
  struct ifreq ifr;
  int fd;

  fd = socket (AF_INET, SOCK_DGRAM, 0);

  if (fd < 0)
	return NULL;

   ifr.ifr_ifindex = ifindex;
   ioctl (fd, SIOCGIFNAME, &ifr);
   close (fd);
   
   return strncpy (ifname, ifr.ifr_name, IFNAMSIZ);
   
}


static  void jiffies_to_timespec( struct timespec *value,unsigned long jiffies)
{
	value->tv_nsec = (jiffies % HZ) * (1000000000L / HZ);
	value->tv_sec = jiffies / HZ;
}

static void cp_prt_inf(struct port_info *info, struct __port_info *i)
{
	memcpy(&info->designated_root, &i->designated_root, 8);
	memcpy(&info->designated_bridge, &i->designated_bridge, 8);
	info->port_id = i->port_id;
	info->designated_port = i->designated_port;
	info->path_cost = i->path_cost;
	info->designated_cost = i->designated_cost;
	info->state = i->state;
	info->top_change_ack = i->top_change_ack;
	info->config_pending = i->config_pending;
	jiffies_to_timespec(&info->message_age_timer_value, i->message_age_timer_value);
	jiffies_to_timespec(&info->forward_delay_timer_value, i->forward_delay_timer_value);
	jiffies_to_timespec(&info->hold_timer_value, i->hold_timer_value);
}

static void cp_br_inf(struct bridge_info *info, struct __bridge_info *i)
{
	memcpy(&info->designated_root, &i->designated_root, 8);
	memcpy(&info->bridge_id, &i->bridge_id, 8);
	info->root_path_cost = i->root_path_cost;
	info->topology_change = i->topology_change;
	info->topology_change_detected = i->topology_change_detected;
	info->root_port = i->root_port;
	info->stp_enabled = i->stp_enabled;
	jiffies_to_timespec(&info->max_age, i->max_age);
	jiffies_to_timespec(&info->hello_time, i->hello_time);
	jiffies_to_timespec(&info->forward_delay, i->forward_delay);
	jiffies_to_timespec(&info->bridge_max_age, i->bridge_max_age);
	jiffies_to_timespec(&info->bridge_hello_time, i->bridge_hello_time);
	jiffies_to_timespec(&info->bridge_forward_delay, i->bridge_forward_delay);
	jiffies_to_timespec(&info->ageing_time, i->ageing_time);
	jiffies_to_timespec(&info->gc_interval, i->gc_interval);
	jiffies_to_timespec(&info->hello_timer_value, i->hello_timer_value);
	jiffies_to_timespec(&info->tcn_timer_value, i->tcn_timer_value);
	jiffies_to_timespec(&info->topology_change_timer_value, i->topology_change_timer_value);
	jiffies_to_timespec(&info->gc_timer_value, i->gc_timer_value);
}

static int collect_all_ports(int bridge_fd, struct bridge *br)
{
	int err = -1;
	int i;
	int ifindices[256];
	unsigned long args[4];
	struct ifreq ifr;

	args[0] = BRCTL_GET_PORT_LIST;
	args[1] = (unsigned long)ifindices;
	args[2] = 0;
	args[3] = 0;

	memcpy(ifr.ifr_name, br->ifname, IFNAMSIZ);

	ifr.ifr_data = (void *)args;

	if(ioctl(bridge_fd, SIOCDEVPRIVATE, &ifr)<0)
	{
		return errno;
	}

	for (i=255;i>=0;i--) 
	{
		struct port *prt;
		struct __port_info prtinf;
		
		if (!ifindices[i])
			continue;

		prt = malloc(sizeof(struct port));
		prt->index = i;
		prt->ifindex = ifindices[i];
		prt->parent = br;
		br->ports[i] = prt;
		prt->next = br->firstport;
		br->firstport = prt;

		args[0] = BRCTL_GET_PORT_INFO;
		args[1] = (unsigned long)&prtinf;
		args[2] = prt->index;
		args[3] = 0;
		memcpy(ifr.ifr_name, prt->parent->ifname, IFNAMSIZ);
		ifr.ifr_data = (void *)args;

		if(ioctl(bridge_fd, SIOCDEVPRIVATE, &ifr) <0)
		{
			err = errno;
		}		
		else
		{
			cp_prt_inf(&prt->info, &prtinf);
			err = 0;
		}
		
		if (err  != 0)
		{
		 	while (++i < 256)
			{
				free(br->ports[i]);
			}
			return err;
		}
	}
	return 0;

}

int port_index_find(struct bridge *br, int port_id)
{
	while (br != NULL) {
		struct port *prt;

		prt = br->firstport;
		while (prt != NULL) {
			if((prt->info.port_id & 0xff) == port_id)
				return prt->index;

			prt = prt->next;
		}
		br = br->next;
	}
	return -1;
}

int bridge_ioctl(struct bridge *br, int bridge_fd, int arg0, int arg1, int arg2, int arg3)
{
	int args_dev[4];
	struct ifreq ifr;

	args_dev[0] = arg0;
	args_dev[1] = arg1;
	args_dev[2] = arg2;
	args_dev[3] = arg3;
	memcpy(ifr.ifr_name, br->ifname, IFNAMSIZ);
	ifr.ifr_data = (void *)args_dev;

	return ioctl(bridge_fd, SIOCDEVPRIVATE, &ifr);
}

int free_bridges(struct bridge *bridges)
{
	struct bridge *bridge = bridges;

	while (bridge != NULL) {
		struct bridge *next_bridge;
		struct port *prt;

		next_bridge = bridge->next;
		prt = bridge->firstport;
		while (prt != NULL) {
			struct port *pnext;
			pnext = prt->next;
			free(prt);
			prt = pnext;
		}

		free(bridge);		
		bridge = next_bridge;
	}

	return 0;
}

struct bridge *collect_bridges(int bridge_fd)
{
	int err = -1;
	int index;
	int ifindices[32];
	int num;
	struct bridge *all_bridges;

	unsigned long ioctl_arg[3];
	ioctl_arg[0] = BRCTL_GET_BRIDGES;
	ioctl_arg[1] = (unsigned long)ifindices;
	ioctl_arg[2] = 32;

	if ((num = ioctl(bridge_fd, SIOCGIFBR, ioctl_arg)) < 0) {
		return NULL;
	}

	all_bridges = NULL;
	for (index=0;index<num;index++) {
		struct bridge *br;
		struct __bridge_info brinf;
		
		br = malloc(sizeof(struct bridge));
		memset(br, 0, sizeof(struct bridge));
		br->ifindex = ifindices[index];
		br->firstport = NULL;
		br->next = all_bridges;
		all_bridges = br;
		
		if (idxtoname(br->ifindex, br->ifname) == NULL) {
			free_bridges(all_bridges);
			return NULL;
		}

		if ( bridge_ioctl(br, bridge_fd, BRCTL_GET_BRIDGE_INFO, (int)&brinf, 0, 0) < 0) {
			free_bridges(all_bridges);
			return NULL;
		}
		cp_br_inf(&br->info, &brinf);
		err = 0;
		
		if ((err = collect_all_ports(bridge_fd, br)) != 0) {
			free_bridges(all_bridges);
			return NULL;
		}
	}

	return all_bridges;
}

