/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2003 Intersil Americas Inc. All Rights Reserved.
 */

#ifndef MOD_BRIDGE_H
#define MOD_BRIDGE_H
#include <net/if.h>
struct bridge;
struct bridge_info;
struct fdb_entry;
struct port;
struct port_info;

struct bridge *collect_bridges(int bridge_fd);
int free_bridges(struct bridge *bridges);
int bridge_ioctl(struct bridge *br, int bridge_fd, int arg0, int arg1, int arg2, int arg3);
int port_index_find(struct bridge *br, int port_id);

int port_id_to_num(u_int16_t port_id);
int port_id_to_prio(u_int16_t port_id);

/* Bridge MIB STP values are in 1/100 sec */
#define BRMIB_TIME2JIF(x)	(HZ * x / 100)

struct bridge_id
{
	unsigned char prio[2];
	unsigned char addr[6];
};

struct bridge_info
{
	struct bridge_id designated_root;
	struct bridge_id bridge_id;
	int root_path_cost;
	struct timespec max_age;
	struct timespec hello_time;
	struct timespec forward_delay;
	struct timespec bridge_max_age;
	struct timespec bridge_hello_time;
	struct timespec bridge_forward_delay;
	unsigned topology_change:1;
	unsigned topology_change_detected:1;
	int root_port;
	unsigned stp_enabled:1;
	struct timespec ageing_time;
	struct timespec gc_interval;
	struct timespec hello_timer_value;
	struct timespec tcn_timer_value;
	struct timespec topology_change_timer_value;
	struct timespec gc_timer_value;
};

struct bridge
{
	struct bridge *next;

	int ifindex;
	char ifname[IFNAMSIZ];
	struct port *firstport;
	struct port *ports[256];
	struct bridge_info info;
};

struct port_info
{
	struct bridge_id designated_root;
	struct bridge_id designated_bridge;
	u_int16_t port_id;
	u_int16_t designated_port;
	int path_cost;
	int designated_cost;
	int state;
	unsigned top_change_ack:1;
	unsigned config_pending:1;
	struct timespec message_age_timer_value;
	struct timespec forward_delay_timer_value;
	struct timespec hold_timer_value;
};

struct port
{
	struct port *next;

	int index;
	int ifindex;
	struct bridge *parent;
	struct port_info info;
};
#endif
