/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <aplog.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <aplog_defs.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "object.h"
#include "easyprism.h"
#include <sys/fcntl.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include "hardconfig.h"
#include <linux/if_bridge.h>
#include <sys/time.h>
#include <asm/param.h>
#include "mod_bridge.h"
#include "mod_mibII_interfaces.h"

static int bridge_fd;

static int kernel_ifindex2ifindex(int ifindex) {
	struct ifreq ifr;
	
	ifr.ifr_ifindex = ifindex;
      	if (ioctl (bridge_fd, SIOCGIFNAME, &ifr) < 0) {
		aplog(EV_SNMP_GENERAL_ERROR, "mod_bridge_base couldn't get interface name", NULL);
	  	return(-1);
	}
	
	return(ifname2ifindex(ifr.ifr_name));
}

static int init(void)
{
	bridge_fd = socket(AF_INET, SOCK_STREAM, 0);
	assert(bridge_fd > 0);

	return (R_NOERR);
}

static int get(void)
{
	struct row_t *row;
	struct row_t *nextrow;
	int i;
	struct schent_t *sch;
	struct object_t *obj;
	struct index_t * idx;
	struct port *br_port;

	struct bridge *br;
	unsigned long arg[3];
	arg[0] = BRCTL_GET_VERSION;
	arg[1] = 0;
	arg[2] = 0;

	if(ioctl(bridge_fd, SIOCGIFBR, arg) == BRCTL_VERSION) {
		br = collect_bridges(bridge_fd);
		if (!br)
			return R_NOSUCHOID;
	} else {
		return R_NOSUCHOID;
	}
	
	// MAC Address
	obj=object_get(".1.3.6.1.2.1.17.1.1.0");
	assert(obj);
	object_set_string(obj,br->info.bridge_id.addr, 6);

	// Number of Ports
	br_port = br->firstport;
	for(i = 0; br_port; br_port = br_port->next, i++);
	obj=object_get(".1.3.6.1.2.1.17.1.2.0");
	assert(obj);
	object_set_int(obj, i);

	// Type
	//obj=object_get(".1.3.6.1.2.1.17.1.3.0");
	//assert(obj);
	//object_set_int(obj,2);					

	sch = schema_get(".1.3.6.1.2.1.17.1.4.1");
	assert(sch);
	
	row = sch->rowlist;

	// First delete table
	while (row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}

	br_port = br->firstport;
	while(br_port) {
		int port_num;
		int ifindex;

		idx= index_new(sch);
		assert(idx);
		obj = index_get_obj(idx, 0);	
		assert(obj);
		port_num = port_id_to_num(br_port->info.port_id);
		object_set_int(obj, port_num);
		row = row_create_by_index(idx);
		assert(row);
		index_done(idx);

		ifindex = kernel_ifindex2ifindex(br_port->ifindex);
		
		object_set_int(row_get_obj(row, 0), port_num);		
		object_set_int(row_get_obj(row, 1), ifindex); 
		object_set_objectid(row_get_obj(row, 2), "0.0"); // all indices are unique
		object_set_counter(row_get_obj(row, 3), 0); // no counters 
		object_set_counter(row_get_obj(row, 4), 0); // no counters

		br_port = br_port->next;
	}

	free_bridges(br);
	return(3); /* cache for 3 seconds */
}




struct moddescr_t mod_bridge_base = {

	name:		__FILE__,
	fn_ini:		init,
	fn_get:		get,
	oidlist:	".1.3.6.1.2.1.17.1.1 .1.3.6.1.2.1.17.1.2 .1.3.6.1.2.1.17.1.3 .1.3.6.1.2.1.17.1.4.1"

};

	


