/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <linux/if_bridge.h>
#include <string.h>
#include <assert.h>
#include <sys/time.h>
#include <asm/param.h>
#include <aplog.h>

#include "types.h"
#include "object.h"
#include "objdata.h"
#include "objflag.h"
#include "schema.h"
#include "table.h"

#define BRIDGE "br0"
#define MAX_FDB_ENTRIES 32

static int br_fd = 0;
static unsigned long args[4];
static struct __fdb_entry fdb[MAX_FDB_ENTRIES];
static struct ifreq ifr;

static unsigned long secs2jiffies(int secs)
{
	return (HZ*(1000000ULL * secs)/1000000);
}



static int init() {
	
	struct object_t* obj;
	
	if ((br_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		aplog(EV_SNMP_GENERAL_ERROR, "mod_bridge_dot1dTp couldn't create socket", NULL);
	}

	/* dot1dTpLearnedEntryDiscards cannot be retrieved from the kernel.
	   just make it static 0 */
	obj = object_create(".1.3.6.1.2.1.17.4.1.0");
	assert(obj);
	object_set_counter(obj, 0);
	
	return(0);
}

static int get() {
	
	int num,i;
	int is_local;
	int self_created;
	struct schent_t* sch;
	struct object_t* obj;
	struct index_t* idx;
	struct row_t* row;
	struct row_t* nextrow;
	
	
	

	
	
	
	sch = schema_get(".1.3.6.1.2.1.17.4.3.1");
	assert(sch);

	/* remove all existing rows */	
	row = sch->rowlist;
	while(row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}
	
	/* get the FDB entries */
	args[0] = BRCTL_GET_FDB_ENTRIES;
	args[1] = (unsigned long) fdb;
	args[2] = MAX_FDB_ENTRIES; 
	args[3] = 0; /* offset */
		
	ifr.ifr_data = (void*)args;
	memcpy(ifr.ifr_name, BRIDGE, IFNAMSIZ);
	
	num = ioctl(br_fd, SIOCDEVPRIVATE, &ifr);
	
	if(num < 0) {
		aplog(EV_SNMP_GENERAL_ERROR, "mod_bridge_dot1dTp couldn't get FDB list", NULL);
		goto end;
	}
	
	self_created = B_FALSE;
	for(i = 0; i < num ; i++) {
		is_local = fdb[i].is_local;
		if(is_local && self_created) {
			/* This is one of the 'local' entries. Only one entrie
			   should be created for the local MAC and this has
			   already been created.*/
			continue;
		}
		
		idx = index_new(sch);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		object_set_string(obj, fdb[i].mac_addr, 6);
		row = row_create_by_index(idx);
		assert(row);
		index_done(idx);
		
		/* get & set mac address object */
		obj = row_get_obj(row, 0);
		assert(obj);
		object_set_string(obj, fdb[i].mac_addr, 6);
		
		/* get portnr object */
		obj = row_get_obj(row, 1);
		assert(obj);
		
		if(is_local) {
			object_set_int(obj, 0);
			self_created = B_TRUE;
			/* get status object */
			obj = row_get_obj(row, 2);
			assert(obj);
			object_set_int(obj, 4); /* 4 == self */
		}
		else {
			object_set_int(obj, fdb[i].port_no);
			
			/* get status object */
			obj = row_get_obj(row, 2);
			assert(obj);
			object_set_int(obj, 3); /* 4 == learned */
		}
		
	}
end:	
	return(3); /* cache for 3 seconds */
}


static int set() {
	
	struct object_t* obj;
	int aging;
	
	obj = object_get(".1.3.6.1.2.1.17.4.2.0");
	assert(obj);
	if (object_flag_isset(obj, OF_DIRTY)) {
		assert(obj->flags & OF_DATAPRESENT);
		aging = object_get_int(obj);
		
		args[0] = BRCTL_SET_AGEING_TIME;
		args[1] = secs2jiffies(aging);
		args[2] = 0;
		args[3] = 0;
		
		ifr.ifr_data = (void*)args;
		memcpy(ifr.ifr_name, BRIDGE, IFNAMSIZ);
	
		if(ioctl(br_fd, SIOCDEVPRIVATE, &ifr) < 0) {
			aplog(EV_SNMP_GENERAL_ERROR, "mod_bridge_dot1dTp couldn't set AgingTime", NULL);
		}
		
		object_flag_clear(obj, OF_DIRTY);
	}
	
			
	
	return R_NOERR;
}

struct moddescr_t mod_bridge_dot1dTp = {

	name:		__FILE__,
	fn_ini:		init,
	fn_get:		get,
	fn_set:		set,
	oidlist:	".1.3.6.1.2.1.17.4.2 .1.3.6.1.2.1.17.4.3.1"
};


/*

O .1.3.6.1.2.1.17.4.1
L dot1dTpLearnedEntryDiscards
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.17.4.2
L dot1dTpAgingTime
T INTEGER
A RW
R 10/1000000

O .1.3.6.1.2.1.17.4.3.1
L dot1dTpFdbEntry
T TABLE
I .1.3.6.1.2.1.17.4.3.1.1

0	O .1.3.6.1.2.1.17.4.3.1.1
	L dot1dTpFdbAddress
	T OCTETSTR
	A R
	R 6

1	O .1.3.6.1.2.1.17.4.3.1.2
	L dot1dTpFdbPort
	T INTEGER
	A R

2	O .1.3.6.1.2.1.17.4.3.1.3
	L dot1dTpFdbStatus
	T INTEGER
	A R
	R 1,2,3,4,5


*/
