/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <aplog_defs.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "object.h"
#include "objflag.h"
#include <linux/if_bridge.h>
#include <sys/time.h>
#include <asm/param.h>
#include "easyprism.h"
#include "mod_bridge.h"
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include "cfg_file_lib.h"
#include <net/if.h>
#include <sys/fcntl.h>
#include "mod_bridge.h"
#include <aplog.h>

static int bridge_fd;

#define BRMIB_STATE_DISABLED	1	
#define BRMIB_STATE_BLOCKING	2
#define BRMIB_STATE_LISTENING	3
#define BRMIB_STATE_LEARNING	4
#define BRMIB_STATE_FORWARDING	5
#define BRMIB_STATE_BROKEN	6

static int mib_convert_port_state(int port_state)
{
	switch(port_state) {
		case BR_STATE_DISABLED:
			return BRMIB_STATE_DISABLED;
		case BR_STATE_LISTENING:
			return BRMIB_STATE_LISTENING;
		case BR_STATE_LEARNING:
			return BRMIB_STATE_LEARNING;
		case BR_STATE_FORWARDING:
			return BRMIB_STATE_FORWARDING;
		case BR_STATE_BLOCKING:
			return BRMIB_STATE_BLOCKING;
		default:
			return BRMIB_STATE_BROKEN;
	}
}

static int get_dot1dStpPort_table(struct bridge *bridge)
{
	struct row_t *row;
	struct schent_t *sch;
	struct index_t *idx;
	struct object_t *obj;
	struct port *br_port;

	assert(bridge);
	sch = schema_get(".1.3.6.1.2.1.17.2.15.1");
	assert(sch);
	row = sch->rowlist;

	while (row) {
		int port_num;

		obj = row_get_obj(row, 0);
		port_num = object_get_int(obj);

		br_port = bridge->firstport;
		while(br_port) {
			if(port_num == port_id_to_num(br_port->info.port_id))
				break;

			br_port = br_port->next;
		}
		if(!br_port) {
			row_delete(row);
		}
		row = row->next;
	}

	br_port = bridge->firstport;
	while(br_port) {
		row = sch->rowlist;
		while(row) {
			int port_num;

			obj = row_get_obj(row, 0);
			port_num = object_get_int(obj);
			if(port_num == port_id_to_num(br_port->info.port_id))
				break;

			row = row->next;
		}

		if(!row) {
			/* If someone else added a port to the bridge, we
			 * have to sync up here. This means add a row
			 * and set the RW objects once.
			 */
			idx= index_new(sch);
			assert(idx);
			obj = index_get_obj(idx, 0);
			assert(obj);
			object_set_int(obj, port_id_to_num(br_port->info.port_id));
			row = row_create_by_index(idx);
			assert(row);
			index_done(idx);

			/* FIXME: BUG in datastore agent
			 * the agent should set the index object in the row, but
			 * it doesn't so we do that here for now.
			 */
			object_set_int(row_get_obj(row, 0), port_id_to_num(br_port->info.port_id));

			// dot1dStpPortPriority
			object_set_int(row_get_obj(row, 1), port_id_to_prio(br_port->info.port_id));

			// dot1dStpPortEnable
			object_set_int(row_get_obj(row, 3),
				(br_port->info.state == BR_STATE_DISABLED) ? 2 : 1);

			// dot1dStpPortPathCost32
			object_set_int(row_get_obj(row, 9), br_port->info.path_cost);
		}

		// dot1dStpPortState
		object_set_int(row_get_obj(row, 2), mib_convert_port_state(br_port->info.state));

		// dot1dStpPortDesignatedRoot
		object_set_string(row_get_obj(row, 4), &(br_port->info.designated_root.prio[0]),
				sizeof(struct bridge_id));

		// dot1dStpPortDesignatedCost
		object_set_int(row_get_obj(row, 5), br_port->info.designated_cost);

		// dot1dStpPortDesignatedBridge
		object_set_string(row_get_obj(row, 6), &(br_port->info.designated_bridge.prio[0]),
				sizeof(struct bridge_id));

		// dot1dStpPortDesignatedPort
		object_set_string(row_get_obj(row, 7), (char *)&(br_port->info.designated_port), 2);

		// dot1dStpPortForwardTransitions
		object_set_counter(row_get_obj(row, 8), 0);	// NOT SUPPORTED BY THE KERNEL

		br_port = br_port->next;
	}

	return 0;
}

static int init(void)
{
	bridge_fd = socket(AF_INET, SOCK_STREAM, 0);
	assert(bridge_fd > 0);

	return (R_NOERR);
}

static int get(void)
{
	struct object_t *obj;
	int priority, timechange;	
	struct timespec tv;
	struct bridge *br;

	unsigned long arg[3];
	arg[0] = BRCTL_GET_VERSION;
	arg[1] = 0;
	arg[2] = 0;

	if(ioctl(bridge_fd, SIOCGIFBR, arg) == BRCTL_VERSION) {
		br = collect_bridges(bridge_fd);
		if (!br)
			return R_NOSUCHOID;
	} else {
		return R_NOSUCHOID;
	}

	// Protocol Spec
	obj=object_get(".1.3.6.1.2.1.17.2.1.0");
	assert(obj);
	object_set_int(obj,3);		

	// Priority
	obj=object_get(".1.3.6.1.2.1.17.2.2.0");
	assert(obj);
	priority =(int) br->info.bridge_id.prio;
	object_set_int(obj,priority);	

	// TimeSinceTopologyChange
	obj=object_get(".1.3.6.1.2.1.17.2.3.0");
	assert(obj);
	tv = br->info.topology_change_timer_value;
	timechange = 1000000ULL * tv.tv_sec + tv.tv_nsec;
	object_set_ticks(obj,(HZ*timechange)/1000000);					

	// TopChanges
	obj=object_get(".1.3.6.1.2.1.17.2.4.0");
	assert(obj);
	object_set_counter(obj,br->info.topology_change);					

	// Designated Root
	obj=object_get(".1.3.6.1.2.1.17.2.5.0");
	assert(obj);
	object_set_string(obj,&(br->info.designated_root.prio[0]), sizeof(struct bridge_id));					

	// dot1dStpPort Table
	get_dot1dStpPort_table(br);

	free_bridges(br);
	return(3);
}


static int set(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct object_t *obj;
	struct bridge *br;
	
	unsigned long arg[3];
	arg[0] = BRCTL_GET_VERSION;
	arg[1] = 0;
	arg[2] = 0;

	if(ioctl(bridge_fd, SIOCGIFBR, arg) == BRCTL_VERSION) {
		br = collect_bridges(bridge_fd);
		if (!br)
			return R_NOSUCHOID;
	} else {
		return R_NOSUCHOID;
	}

	// Priority
	obj=object_get(".1.3.6.1.2.1.17.2.2.0");
	assert(obj);
	bridge_ioctl(br, bridge_fd, BRCTL_SET_BRIDGE_PRIORITY,
			object_get_int(obj), 0, 0);
	
	// Max Age
	obj=object_get(".1.3.6.1.2.1.17.2.12.0");
	assert(obj);
	bridge_ioctl(br, bridge_fd, BRCTL_SET_BRIDGE_MAX_AGE,
			BRMIB_TIME2JIF(object_get_int(obj)), 0, 0);

	// Hello Time
	obj=object_get(".1.3.6.1.2.1.17.2.13.0");
	assert(obj);
	bridge_ioctl(br, bridge_fd, BRCTL_SET_BRIDGE_HELLO_TIME,
			BRMIB_TIME2JIF(object_get_int(obj)), 0, 0);

	// Forward Delay
	obj=object_get(".1.3.6.1.2.1.17.2.14.0");
	assert(obj);
	bridge_ioctl(br, bridge_fd, BRCTL_SET_BRIDGE_FORWARD_DELAY,
			BRMIB_TIME2JIF(object_get_int(obj)), 0, 0);

	sch = schema_get(".1.3.6.1.2.1.17.2.15.1");
	assert(sch);

	row = sch->rowlist;
	while(row)
	{
		int port_index;

		// dot1dStpPort
		obj = row_get_obj(row, 0);
		port_index = port_index_find(br, object_get_int(obj));

		if(port_index < 0)
			goto next_row;

		// dot1dStpPortPriority
		obj = row_get_obj(row, 1);
		if(object_flag_isset(obj, OF_DIRTY)) {
			bridge_ioctl(br, bridge_fd, BRCTL_SET_PORT_PRIORITY, port_index,
				object_get_int(obj), 0);
			object_flag_clear(obj, OF_DIRTY);
		}

		// dot1dStpPortEnable
		obj = row_get_obj(row, 3);
		if(object_flag_isset(obj, OF_DIRTY)) {
			// TODO enable/disable port
			object_flag_clear(obj, OF_DIRTY);
		}

		// dot1dStpPortPathCost32
		obj = row_get_obj(row, 9);
		if(object_flag_isset(obj, OF_DIRTY)) {
			bridge_ioctl(br, bridge_fd, BRCTL_SET_PATH_COST, port_index,
				object_get_int(obj), 0);
			object_flag_clear(obj, OF_DIRTY);
		}
next_row:
		row = row->next;
	}

        /*
         * Now, disable STP
         */
        bridge_ioctl(br, bridge_fd, BRCTL_SET_BRIDGE_STP_STATE, 0, 0, 0);

	free_bridges(br);
	return(0);
}


struct moddescr_t mod_bridge_stp = {

	name:		__FILE__,
	fn_ini:		init,
	fn_get:		get,
	fn_set:		set,
	oidlist:	".1.3.6.1.2.1.17.2.1 .1.3.6.1.2.1.17.2.2 .1.3.6.1.2.1.17.2.3 .1.3.6.1.2.1.17.2.4 .1.3.6.1.2.1.17.2.5 .1.3.6.1.2.1.17.2.6 .1.3.6.1.2.1.17.2.7 .1.3.6.1.2.1.17.2.8 .1.3.6.1.2.1.17.2.9 .1.3.6.1.2.1.17.2.10 .1.3.6.1.2.1.17.2.11 .1.3.6.1.2.1.17.2.12 .1.3.6.1.2.1.17.2.13 .1.3.6.1.2.1.17.2.14 .1.3.6.1.2.1.17.2.15.1"

};

	


