/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <aplog.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include <arpa/inet.h>

#include "table.h"
#include "objdata.h"
#include "schema.h"

#define UDHCPD_PID_FILE "/var/run/udhcpd.pid"

int signal_dhcpd()
{
    FILE *pid_fd;
    int  pid;
    
    /* Get pid of dhcpServer */
    if (!(pid_fd = fopen(UDHCPD_PID_FILE, "r"))) {
    	     return 0;
    }
    
    fscanf(pid_fd, "%i", &pid);
    
    if (kill(pid, SIGUSR1))
    {
	 aplog(EV_SNMP_GENERAL_ERROR,"Dhcp client: Unable to send SIGUSR2 to udhcpd",NULL);
    }

    fclose(pid_fd);
    return 1;
}

/* copied from dumpleases.c in udhcpd package */
struct lease_t {
    unsigned char chaddr[6];
    u_int32_t yiaddr;
    u_int32_t expires;
    u_int8_t  state;
} __attribute__ ((packed));

static int get(void)
{
	
	struct schent_t *sch;
	struct row_t *row, *nextrow;
	struct index_t *idx;	
	struct object_t *obj;
	FILE* f;
	struct lease_t lease;
	unsigned int max_server_index=0;	
	int server_index;	
	int row_created = B_FALSE;
	struct in_addr ipaddr;

	sch = schema_get(".1.3.6.1.4.1.937.2.1.11.3.1.1");
	assert(sch);
	row = sch->rowlist;

	f = fopen("/var/run/udhcpd.leases", "r");
	if(f == NULL)
	{
		return 1;
	}
	
	while( fread(&lease, sizeof(lease), 1, f) != 0 ) 
	{
		if(!row) {
			/* there are more entries as rows, create a new row */
			max_server_index++;
			
			idx = index_new(sch);
			assert(idx);
			obj = index_get_obj(idx, 0);	
			assert(obj);
			object_set_uint(obj,max_server_index);
		
			row = row_create_by_index(idx);
			assert(row);
			index_done(idx);
			
			/* FIXME, this object should be set during create_by_index */
			obj = row_get_obj(row,0);
			assert(obj);
			object_set_uint(obj, max_server_index);
			
			row_created = B_TRUE;
		} 
		else {
			/* reusing an existing row, check index */
			obj = row_get_obj(row,0);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			server_index = object_get_uint(obj);
			
			if(server_index > max_server_index) {
				max_server_index = server_index;
			}
			
			row_created = B_FALSE;
		}
		
		/* set MAC address */
		obj = row_get_obj(row, 1);
		assert(obj);
		object_set_string(obj,lease.chaddr,6);
		
		/* set IP address */
		obj = row_get_obj(row, 2);
		assert(obj);
		ipaddr.s_addr = lease.yiaddr;
		object_set_ipaddr(obj,ipaddr);		
		
		/* set state */
		obj = row_get_obj(row, 3);
		assert(obj);
		object_set_int(obj, lease.state);
			
		/* set expire time */
		obj = row_get_obj(row, 4);
		assert(obj);
		object_set_uint(obj,ntohl(lease.expires));		
		
		
		/* This is a bit tricky. We only go to the next row if 
		 * we didn't create one. The new row is stored at the
		 * beginning of the schema's rowlist, so we can't just
		 * do a row = row->next without checking row_created */
		if(!row_created) {
			row = row->next;
		}
		else {
			row = NULL;
		}
	}
	
	fclose(f);
	
	/* delete remaining rows */
	while(row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}
	
	
	return(5);
}


struct moddescr_t mod_dhcp_client = {

	name:		__FILE__,
	fn_get:		get,
	oidlist:	".1.3.6.1.4.1.937.2.1.11.3.1.1"

};



/*

The following schema entries are provided by this module :
O .1.3.6.1.4.1.937.2.1.11.3.1.1
L clientEntry
T TABLE
I .1.3.6.1.4.1.937.2.1.11.3.1.1.1

	O .1.3.6.1.4.1.937.2.1.11.3.1.1.1
	L clientIndex
	T UNSIGNED32
	A -

	O .1.3.6.1.4.1.937.2.1.11.3.1.1.2
	L clientMacAddress
	T OCTETSTR
	A R
	R 6

	O .1.3.6.1.4.1.937.2.1.11.3.1.1.3
	L clientIpAddress
	T IPADDR
	A R
	R 4

	O .1.3.6.1.4.1.937.2.1.11.3.1.1.4
	L clientState
	T INTEGER
	A R
	R 1,2,3,4

	O .1.3.6.1.4.1.937.2.1.11.3.1.1.5
	L clientExpiration
	T UNSIGNED32
	A R

*/
