/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>

#include <linux_mvc_oids.h>
#include "table.h"
#include "misc.h"
#include "object.h"
#include "objdata.h"
#include "cfg_file_lib.h"
#include "mod_paed_cfg.h"

static int set(void)
{
	struct row_t* row;
	struct schent_t* sch;
	struct object_t *obj, *objentry;
	int activeACLRows;
	int sec_state;
	struct string_t* addressp;
	char line[128];
	char oidbuf[MAX_OID_LENGTH];
	FILE* fp;

	/* check if regdomain is supported */
	sch = schema_get(".1.3.6.1.4.1.937.3.1.3.2.2.1");
	assert(sch);

	fp = create_cfg_file("acl");

	activeACLRows = 0;

	/* Order of rows is important, so we use getnext */
	objentry = object_get_next(".1.3.6.1.4.1.937.3.1.3.2.2.1.2");
	while(objentry && (strcmp(objentry->sch->oid, ".1.3.6.1.4.1.937.3.1.3.2.2.1.2") == 0)) {

		row = objentry->row;
		
		obj = row_get_obj(row, 3) ; /* RowStatus */
		assert(obj);
		/* Check if row is Active */
		if(object_get_rowstat(obj) == 1)
		{
			/* Mac address */
			obj = row_get_obj(row, 1);
			assert(obj);
		    	addressp = object_get_string(obj);

		    	/* secMACAclAllowed */
			obj = row_get_obj(row, 2);
			assert(obj);
			sec_state = object_get_int(obj);

			// TOKEN INDEX MACADDR
			// Acl 0 ff:ff:ff:ff:ff:ff 1
			sprintf(line, "%d %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x %d",
				activeACLRows,
				addressp->data[0], addressp->data[1], addressp->data[2],
				addressp->data[3], addressp->data[4], addressp->data[5],
				sec_state);

			write_cfg_line_str(fp, "Acl", line);
			activeACLRows++;
		}

		
		object_get_oid(oidbuf, sizeof(oidbuf), objentry);
		objentry = object_get_next(oidbuf);
	}


	close_cfg_file(fp);

	restart_paed();	
	
	return(R_NOERR);
}


static int check(void) {
	struct object_t* obj;
	struct schent_t* sch;
	
	sch = schema_get(".1.3.6.1.4.1.937.3.1.3.2.2.1");
	assert(sch);
	
	/* check max */
	obj = object_get(".1.3.6.1.4.1.937.3.1.3.2.1.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	if(table_count(sch) > object_get_int(obj)) {
		return R_CHECKFAILED;
	}

	/* check if the first row still exist and that the value
	 * of the MAC address still if FF:FF:FF:FF:FF:FF */
	obj = object_get(".1.3.6.1.4.1.937.3.1.3.2.2.1.2.1");
	if(!obj || (memcmp(object_get_string(obj)->data, "\xFF\xFF\xFF\xFF\xFF\xFF" , 6) != 0)) {
		return R_CHECKFAILED;
	}
	
	/* Check if the rowstatus of the first row still is '1' */
	obj = object_get(".1.3.6.1.4.1.937.3.1.3.2.2.1.4.1");
	if(!obj || (object_get_rowstat(obj) != 1)) {
		return R_CHECKFAILED;
	}
		
	
	return R_NOERR;
}


struct moddescr_t mod_dot11ExtAPObjs_security = {

	name:		__FILE__,
	fn_set:		set,
	fn_chk:		check,
	oidlist:	".1.3.6.1.4.1.937.3.1.3.2.2.1"

};

/*

The following schema entries are provided by this module :

O .1.3.6.1.4.1.937.3.1.3.2.1
L secMACAclMaxTableLength
T INTEGER
A R

O .1.3.6.1.4.1.937.3.1.3.2.2.1
L secMACAclEntry
T TABLE
I .1.3.6.1.4.1.937.3.1.3.2.2.1.1

	O .1.3.6.1.4.1.937.3.1.3.2.2.1.1
	L secMACAclIndex
	T INTEGER
	A -
	R 1/2007

	O .1.3.6.1.4.1.937.3.1.3.2.2.1.2
	L secMACAclAddress
	T OCTETSTR
	A C
	R 6

	O .1.3.6.1.4.1.937.3.1.3.2.2.1.3
	L secMACAclAllowed
	T INTEGER
	A C
	R 1,2

	O .1.3.6.1.4.1.937.3.1.3.2.2.1.4
	L secMACAclRowStatus
	T ROWSTAT
	A C
	R 1,2,3,4,5,6

O .1.3.6.1.4.1.937.3.1.3.2.3
L secLastError
T OCTETSTR
A R
R 0/80

O .1.3.6.1.4.1.937.3.1.3.2.4
L secLastErrorAddress
T OCTETSTR
A R
R 6

O .1.3.6.1.4.1.937.3.1.3.2.5
L secWEPKeySize
T INTEGER
A R

*/
