/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <assert.h>
#include <string.h>
#include <stdio.h>

#include "misc.h"

#include "schema.h"
#include "table.h"
#include "objdata.h"
#include "objflag.h"
#include "object.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"
#include "mod_dot11smt_sc.h"
#include "calibration.h"
#include "regulatory.h"
#include "pda.h"

static struct Pdr_Struct *get_default_country_PDR(ifindex) {
	
	switch(ifindex) {
		case 3:
			 return(pda_Find_Pdr(PDR_DEFAULT_COUNTRY, NULL));
			 break;
		case 5:
			/* TODO */
			return(pda_Find_Pdr(PDR_DEFAULT_COUNTRY, NULL));
			break;
		default:
			return (NULL);
			break;
	}
	
	return NULL;
}

#define MACADDRESS_LEN 6

static int get(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;
	struct object_t *obj;
	struct object_t *cap_obj;
	struct string_t *macaddr;
	struct string_t *str;
	char oid_buf[MAX_OID_LENGTH];
	unsigned long ul;
	
	char *ifname;
	
	int ifnum = 0;
			
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.1.1");
	assert(sch);

	row = sch->rowlist;
	while(row) {
		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		index = index_get(row);

		ifnum = object_get_int(index_get_obj(index, 0));
		ifname = ifindex2ifname(ifnum);

		index_done(index);

		// Put the ifname in the global...
		prism_set_ifname(ifname);
	

		/* smtCapabilityInfo dynamic BITS */
		cap_obj = row_get_obj(row,1);
		assert(cap_obj);
		
		/* Privacy = 0 or 1 (0 if dot11PrivacyInvoked = false, 1 if true) */
		sprintf(oid_buf, ".1.2.840.10036.1.5.1.1.%i", ifnum);
		obj = object_get(oid_buf);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		object_set_bit(cap_obj, 4, (object_get_int(obj) == 1) ? 1 : 0);
		
		/* Short Preamble = 0 or 1 (0 if smtShortPreambleInvoked = false, 1 if true) */
		obj = row_get_obj(row,12); /* smtShortPreableInvoked */
		assert(obj);
		object_set_bit(cap_obj, 5, (object_get_int(obj) == 1) ? 1 : 0);
		
		/* smtCurrentBSSID ==  dot11MACAddress for an AP */
		/*
		sprintf(oid_buf, ".1.2.840.10036.2.1.1.1.%i", ifnum);
		obj = object_get(oid_buf);
		assert(obj);
		macaddr = object_get_string(obj);
		object_set_string(row_get_obj(row, 6), macaddr->data, macaddr->len);
		*/// No, it has to be retrieved from MVC
		macaddr = prism_get_string(DOT11_OID_BSSID, 6);
		object_set_string(row_get_obj(row, 6), macaddr->data, macaddr->len);
		/* smtCurrentSSID = dot11DesiredSSID for an AP */
		sprintf(oid_buf, ".1.2.840.10036.1.1.1.9.%i", ifnum);
		obj = object_get(oid_buf);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		str = object_get_string(obj);
		object_set_string(row_get_obj(row,7), str->data, str->len);
		
		/* smtSlotTime */
		ul = prism_get_long(DOT11_OID_SLOTTIME);
		obj = row_get_obj(row, 13);
		assert(obj);
		object_set_int(obj, ul);
		
		switch(ifnum) {
			case 3: /* FULL MAC INTERFACES */
		
				/* Desired countrycode. If NULL -> get default country */
				obj = row_get_obj(row, 14);
				if(!(obj->flags & OF_DATAPRESENT)) {
					struct Pdr_Struct *pdr;
					struct country_domain_t* cd;

					pdr = get_default_country_PDR(ifnum);
					if(pdr) {
						if(pdr->PdrData[0] != 0x00) {
							/* default 'country' in PDA is reg.domain */

							/* get the PDA domain */
							cd = regulatory_get_domain(regulatory_value2domainid(pdr->PdrData[0]));
							assert(cd);

							/* get the default country for this domain */
							object_set_string(obj, cd->default_country, 3);
						}
						else {
							/* default country is in pdrData[1-3] */
							object_set_string(obj, &(pdr->PdrData[1]), 3);
						}
					}
					else {
						cd = regulatory_get_default_country_domain();
						if(cd->type == REGULATORY_TYPE_DOMAIN) { 
							object_set_string(obj, cd->default_country, 3);
						}
						else {
							object_set_string(obj, cd->id, 3);
						}

					}
				}
				break;
			case 5: /* SOFTMAC INTERFACES */
				/* The desired countrycode is initialised by the post init handler */
				break;
			default:
				assert(0==1);
				break;
		}
		
				
		row = row->next;
	}

	return(3);
}


static int set(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct object_t *obj;
	struct index_t *index;
	int dbtimeout, cwmin, cwmax,bsstimeout, phytype, fixedrate;
	int longval, mvcval = 0;			
	char *ifname;
	int ifnum = 0;	
	char oidbuf[MAX_OID_LENGTH];		
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.1.1");
	assert(sch);

	row = sch->rowlist;
	while(row) 
	{
		/* determine ifname and store it */
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
		ifname = ifindex2ifname(ifnum);
		index_done(index);
		prism_set_ifname(ifname);
		
		obj = row_get_obj(row, 9);
		if(object_flag_isset(obj, OF_DIRTY)) {
			dbtimeout = object_get_int(obj);
			prism_set_long(DOT11_OID_STATIMEOUT, dbtimeout);
			object_flag_clear(obj, OF_DIRTY);
		}
		obj = row_get_obj(row, 10);
		if(object_flag_isset(obj, OF_DIRTY)) {
			cwmin = object_get_int(obj);
			prism_set_long(DOT11_OID_CWMIN, cwmin);
			object_flag_clear(obj, OF_DIRTY);
		}
		obj = row_get_obj(row, 11);
		if(object_flag_isset(obj, OF_DIRTY)) {
			cwmax = object_get_int(obj);
			prism_set_long(DOT11_OID_CWMAX, cwmax);
			object_flag_clear(obj, OF_DIRTY);
		}

		/* smtPreambleSetting */
		obj = row_get_obj(row, 12);
		if(object_flag_isset(obj, OF_DIRTY)) {

			longval = object_get_int(obj);
			switch(longval) {
				case 1:
					mvcval = DOT11_PREAMBLESETTING_SHORT;
					break;
				case 2:
					mvcval = DOT11_PREAMBLESETTING_LONG;
					break;
				case 3:
					mvcval = DOT11_PREAMBLESETTING_DYNAMIC;
					break;
				default:
					die();
					break;
			}
			prism_set_long(DOT11_OID_PREAMBLESETTINGS, mvcval);
			object_flag_clear(obj, OF_DIRTY);
		}
		
		/* smtDesiredCountry */
		obj = row_get_obj(row, 14);
		if(object_flag_isset(obj, OF_DIRTY)) {
			
			/* do nothing, this is handled in the dot11phy_op module */
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 15);
		if(object_flag_isset(obj, OF_DIRTY)) {
			bsstimeout = object_get_uint(obj);
			prism_set_long(DOT11_OID_BSSTIMEOUT, bsstimeout);
			object_flag_clear(obj, OF_DIRTY);
		}
		
		obj = row_get_obj(row, 16);
		if(object_flag_isset(obj, OF_DIRTY)) {
			phytype = object_get_int(obj);
			
			/* The dsss and ofdm module will be triggered by this change 
			 * The actual switch of band will be done there by selecting a
			 * new channel/frequency */
			
			object_flag_clear(obj, OF_DIRTY);
		}
		
		/* smtNonERPProtection */
		obj = row_get_obj(row, 17);
		if(object_flag_isset(obj, OF_DIRTY)) {
			longval = object_get_int(obj);
			
			switch(longval) {
				case 1:
					mvcval = DOT11_NONERP_ALWAYS;
					break;
				case 2:
					mvcval = DOT11_NONERP_DYNAMIC;
					break;
				case 3:
					mvcval = DOT11_NONERP_NEVER;
					break;
				default:
					die();
					break;
			}
			prism_set_long(DOT11_OID_NONERPPROTECTION, mvcval);
			object_flag_clear(obj, OF_DIRTY);
		}
		
		/* smtShortSlotPolicy */
		obj = row_get_obj(row, 18);
		if(object_flag_isset(obj, OF_DIRTY)) {
			longval = object_get_int(obj);
			
			switch(longval) {
				case 1:
					mvcval = DOT11_SLOTSETTING_LONG;
					break;
				case 2:
					mvcval = DOT11_SLOTSETTING_SHORT;
					break;
				case 3:
					mvcval = DOT11_SLOTSETTING_DYNAMIC;
					break;
				default:
					die();
					break;
			}
	
			prism_set_long(DOT11_OID_SLOTSETTINGS, mvcval);
			object_flag_clear(obj, OF_DIRTY);
		}
		
		/*  smtExtendedRateSet */
		obj = row_get_obj(row,19);
		if(object_flag_isset(obj, OF_DIRTY)) {
			mod_dot11smt_sc_SetRates(ifnum);
			object_flag_clear(obj, OF_DIRTY);
		}
		
		/* smtMaxFrameBurst */
		obj = row_get_obj(row, 20);
		if(object_flag_isset(obj, OF_DIRTY)) {
			longval = object_get_uint(obj);
			prism_set_long(DOT11_OID_MAXFRAMEBURST, longval);
			object_flag_clear(obj, OF_DIRTY);
		}
		
		/* smtShroudedSSIDEnabled */
		obj = row_get_obj(row, 21);
		if(object_flag_isset(obj, OF_DIRTY)) {
			int ShroudedSSID;
			struct string_t *ssid;
			ShroudedSSID = object_get_int(obj);
			if(ShroudedSSID == 1) {
                                char null_ssid[] = "\00";
                                ssid = mkstring(null_ssid, 1);
                                assert(ssid);
                                prism_set_ssid(DOT11_OID_SSIDOVERRIDE,ssid);
			}
			else {
				/* Write the SSID in SSIDOVERRIDE */
				snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.9.%d", ifnum);
				obj = object_get(oidbuf);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				ssid = object_get_string(obj);
				prism_set_ssid(DOT11_OID_SSIDOVERRIDE, ssid);
			}
			object_flag_clear(obj, OF_DIRTY);
		}
		
		/* smtERPPolicy */
		obj = row_get_obj(row, 22);
		if(object_flag_isset(obj, OF_DIRTY)) {
			int ERPPolicy = object_get_int(obj);
			if(ERPPolicy == 1 || ERPPolicy == 2)     // MIXED or B-ONLY
			{	
			    longval = prism_get_long(DOT11_OID_MLMEPOLICIES);
			    longval &= ~DOT11_MLMEPOLICY_EXCLUDEDOT11B;		// B-Allowed
			    prism_set_long(DOT11_OID_MLMEPOLICIES, longval);
			}
			else if (ERPPolicy == 3)   // G-ONLY
			{
			    longval = prism_get_long(DOT11_OID_MLMEPOLICIES);
			    longval |= DOT11_MLMEPOLICY_EXCLUDEDOT11B;		// B-NOT-Allowed
			    prism_set_long(DOT11_OID_MLMEPOLICIES, longval);
			}
			object_flag_clear(obj, OF_DIRTY);
		}
		
		/* smtFixedRate */
        obj = row_get_obj(row, 23);
        if(object_flag_isset(obj, OF_DIRTY)) {
        	fixedrate = object_get_int(obj);
#ifdef DEBUG
            printf("Datastore: fixedrate: %d\n", fixedrate);
#endif
            prism_set_long(DOT11_OID_ALOFT_FIXEDRATE, fixedrate);
            object_flag_clear(obj, OF_DIRTY);
		}
		
		/* smtRoamPreference */
		obj = row_get_obj(row, 24);
		if(obj!=NULL){
			if(object_flag_isset(obj, OF_DIRTY)){
				int roam = object_get_int(obj);
				if(roam == 3) roam = 1;//Gonly, set to prefer G
					if(roam == 4) roam = 2;//Aonly, set to prefer A, 
										   // and another program to restrict band
						prism_set_long(DOT11_OID_ROAMPREFERENCE, roam);
						object_flag_clear(obj, OF_DIRTY);
						system("/bin/killall -ALRM assocd");
			}
		}
		
		/* smtProfiles */
		obj = row_get_obj(row, 25);
		if(obj!=NULL){
			if(object_flag_isset(obj, OF_DIRTY)){
				int profile = object_get_int(obj);
				if(profile>=0 && profile <= 7){
					prism_set_long(DOT11_OID_PROFILES, profile);
					object_flag_clear(obj, OF_DIRTY);
				}
			}
		}
		
		row = row->next;
	}

	return(0);
}

static int check(void) {
	
	struct schent_t* sch, *sc_sch;
	struct row_t* row, *sc_row;
	struct object_t* obj;
	struct string_t* country, *sc_country;
	int ifnum, sc_ifnum;
	struct index_t* idx;
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.1.1");
	assert(sch);
	
	/* get schema for Supported Countries table */
	sc_sch = schema_get(".1.3.6.1.4.1.937.2.2.2.6.1");
	assert(sc_sch);
	
	row = sch->rowlist;
	while(row) {
		
		/*************************************/
		/* CHECK DESIRED COUNTRY             */
		/*************************************/
		
		/* get desired country */
		obj = row_get_obj(row, 14);
		assert(obj);
		
		/* only check 'dirty' objects */
		if(object_flag_isset(obj, OF_DIRTY)) {
			assert(obj->flags & OF_DATAPRESENT);
			country = object_get_string(obj);

			/* get interfacenr */
			idx = index_get(row);
			assert(idx);
			obj = index_get_obj(idx,0);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			ifnum = object_get_int(obj);
			index_done(idx);

			sc_row = sc_sch->rowlist;
			while(sc_row) {
				/* get interfacenr */
				idx = index_get(sc_row);
				assert(idx);
				obj = index_get_obj(idx, 0);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				sc_ifnum = object_get_int(obj);
				index_done(idx);

				/* get country */
				obj = row_get_obj(sc_row, 1);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				sc_country = object_get_string(obj);

				if((sc_ifnum == ifnum) &&
					(strncmp(country->data, sc_country->data, country->len) == 0)
						) {
					break;
				}

				sc_row = sc_row->next;
			}
			
			if(!sc_row) {
				return R_CHECKFAILED;
			}

		}
		
		
		
		
			
		
		row = row->next;
	}
	return R_NOERR;
}

#define DOT11_OID_COUNTRY               0x1f000002
static int post_init(void) {
	
	struct schent_t *sch;
	struct row_t *row;
	struct object_t *obj;
	struct index_t *index;
	struct string_t *str;
	char *ifname;
	int ifnum = 0;	
	char buf[4];
	char oidbuf[MAX_OID_LENGTH];
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.1.1");
	assert(sch);
	row = sch->rowlist;
	while(row) 
	{
		/* determine ifname */
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
		ifname = ifindex2ifname(ifnum);
		index_done(index);

		
		/* If the country code is not yet initialized, get it from the
		 * Softmac Interfaces, otherwise, set it to the SoftMac interfaces */
		
		switch(ifnum) {
			case 5: /* SOFTMAC */
				obj = row_get_obj(row, 14);
				assert(obj);
				
				if(obj->flags & OF_DATAPRESENT) {
					str = object_get_string(obj);
					set_prismoid(ifname, DOT11_OID_COUNTRY, str->data, 4);
				}
				else {
					get_prismoid(ifname, DOT11_OID_COUNTRY, buf, sizeof(buf));
					object_set_string(obj, buf, 3);
					
					/* And initialize the regdom to 0x00 */
					snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.4.1.1.2.%d", ifnum);
					obj = object_get(oidbuf);
					assert(obj);
					object_set_int(obj, 0x00);
					
				}
				
				break;
				
			default:
				/* ignore other interfaces */
				break;
			
		}
		
		row = row->next;	
	}
	
	return R_NOERR;
}

struct moddescr_t mod_dot11extsmt = {

	name:		__FILE__,
	fn_get:		get,
	fn_set:		set,  
	fn_chk:		check,
	fn_pin:		post_init,
	oidlist:	".1.3.6.1.4.1.937.2.2.2.1.1"

};

/*

O .1.3.6.1.4.1.937.2.2.2.1.1
L smtEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1

0	O .1.3.6.1.4.1.937.2.2.2.1.1.1
	L smtAssociationID
	T INTEGER
	A R
	R 0/2007

1	O .1.3.6.1.4.1.937.2.2.2.1.1.2
	L smtCapabilityInfo
	T BITS
	A R
	R 0,1,2,3,4,5,6,7

2	O .1.3.6.1.4.1.937.2.2.2.1.1.3
	L smtPowerSaveInterval
	T INTEGER
	A R

3	O .1.3.6.1.4.1.937.2.2.2.1.1.4
	L smtListenInterval
	T INTEGER
	A R

4	O .1.3.6.1.4.1.937.2.2.2.1.1.5
	L smtATIMWindow
	T INTEGER
	A R

5	O .1.3.6.1.4.1.937.2.2.2.1.1.6
	L smtOperationalChannels
	T OCTETSTR
	A R
	R 0/255

6	O .1.3.6.1.4.1.937.2.2.2.1.1.7
	L smtCurrentBSSID
	T OCTETSTR
	A R
	R 6

7	O .1.3.6.1.4.1.937.2.2.2.1.1.8
	L smtCurrentSSID
	T OCTETSTR
	A R
	R 0/32

8	O .1.3.6.1.4.1.937.2.2.2.1.1.9
	L smtCurrentBSSType
	T INTEGER
	A R
	R 1,2

9	O .1.3.6.1.4.1.937.2.2.2.1.1.10
	L smtStationDBTimeout
	T INTEGER
	A RW

10	O .1.3.6.1.4.1.937.2.2.2.1.1.11
	L smtCWMin
	T INTEGER
	A RW
	R 0/65535

11	O .1.3.6.1.4.1.937.2.2.2.1.1.12
	L smtCWMax
	T INTEGER
	A RW
	R 0/65535

12	O .1.3.6.1.4.1.937.2.2.2.1.1.13
	L smtPreambleSetting
	T INTEGER
	A RW
	R 1,2,3

13	O .1.3.6.1.4.1.937.2.2.2.1.1.14
	L smtSlotTime
	T INTEGER
	A RW

14	O .1.3.6.1.4.1.937.2.2.2.1.1.15
	L smtDesiredCountry
	T OCTETSTR
	A RW
	R 3

15	O .1.3.6.1.4.1.937.2.2.2.1.1.16
	L smtBSSDBTimeout
	T UNSIGNED32
	A RW

16	O .1.3.6.1.4.1.937.2.2.2.1.1.18
	L smtSelectedPHYType
	T INTEGER
	A RW
	R 1,2,3,4,5,6

17	O .1.3.6.1.4.1.937.2.2.2.1.1.20
	L smtNonERPProtection
	T INTEGER
	A RW
	R 1,2,3
	
18	O .1.3.6.1.4.1.937.2.2.2.1.1.21
	L smtShortSlotPolicy
	T INTEGER
	A RW
	R 1,2,3
	D 3

19	O .1.3.6.1.4.1.937.2.2.2.1.1.22
	L smtExtendedRateSet
	T OCTETSTR
	A RW
	R 0/255	
	
20      O .1.3.6.1.4.1.937.2.2.2.1.1.23
        L smtMaxFrameBurst
        T UNSIGNED32
        A RW

21      O .1.3.6.1.4.1.937.2.2.2.1.1.24
        L smtShroudedSSIDEnabled
        T INTEGER
        A RW
        R 1,2
	
22	O .1.3.6.1.4.1.937.2.2.2.1.1.25
	L smtERPPolicy
	T INTEGER
	A RW
	R 1,2,3
	D 1

23      O .1.3.6.1.4.1.937.2.2.2.1.1.26
        L smtFixedRate
        T INTEGER
        A RW

23	O .1.3.6.1.4.1.937.2.2.2.1.1.26
	L smtChannelSelectionMode
	T INTEGER
	A RW
	R 1,2,3,4
	D 1

24	O .1.3.6.1.4.1.937.2.2.2.1.1.27
	L smtAlternativeChannel
	T INTEGER
	A RW
	R 0/255

25	O .1.3.6.1.4.1.937.2.2.2.1.1.28
	L smtWWRMode
	T INTEGER
	A RW
	R 1,2
	D 1

*/
