/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <assert.h>

#include "schema.h"
#include "table.h"
#include "objdata.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"

static int init(void)
{
	struct schent_t *sch, *ifsch;
	struct index_t *index;
	struct row_t* row;
	int ifnum;
	
	
	/* This is a readonly table. All the values are retrieved from the
	 * MVC. The rows a created during initialisation 
         */
	sch = schema_get(".1.2.840.10036.2.2.1");
	assert(sch);
	
	/* There should be NO rows */
	assert(!sch->rowlist);
	
	ifsch = schema_get(".1.3.6.1.2.1.2.2.1");
	assert(ifsch);
	row = ifsch->rowlist;
	
	/* loop iftable to get all wireless interfaces */
	while(row) {
		
		/* check for wireless interface (ieee80211) and its operational 
		   state. */
		if( (object_get_int(row_get_obj(row, 2)) == 71) /* 71 = ieee80211 */
				&& (object_get_int(row_get_obj(row,7)) != 6) /* 6 = Not Present */){

			ifnum = object_get_int(row_get_obj(row,0));
			
			/* create new row */
			index = index_new(sch);
			object_set_int(index_get_obj(index,0), ifnum);
			row_create_by_index(index);
			
			index_done(index);			
		}
				
		row = row->next;
	}
	
	return (R_NOERR);			
}

static int get(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;


	char* ifname;
	
	int mpdutxsuc;
	int msutxmult;
	int mpdutxfail;
	int mpduoneretry;
	int mpdumultretries;
	int mpdurxdups;
	int rtssuc;
	int rtsfail;
	int ackfail;				
	int mpdurxsuc;
	int msdurxmult;
	int frameerr;
	int msdutxsuc;
	int privrxnokey;				
	
	int ifnum = 0;
			
	sch = schema_get(".1.2.840.10036.2.2.1");
	assert(sch);

	row = sch->rowlist;
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
			
		ifname = ifindex2ifname(ifnum);

		index_done(index);

		// Put the ifname in the global...
	
		prism_set_ifname(ifname);

		mpdutxsuc = prism_get_long(DOT11_OID_MPDUTXSUCCESSFUL);
		object_set_counter(row_get_obj(row, 0), mpdutxsuc);
		
		msutxmult= prism_get_long(GEN_OID_MSDUTXMULTICAST);
		object_set_counter(row_get_obj(row, 1), msutxmult);
		
		mpdutxfail= prism_get_long(DOT11_OID_MPDUTXFAILED);
		object_set_counter(row_get_obj(row, 2), mpdutxfail);
		
		mpduoneretry= prism_get_long(DOT11_OID_MPDUTXONERETRY);
		object_set_counter(row_get_obj(row, 3), mpduoneretry);
		
		mpdumultretries= prism_get_long(DOT11_OID_MPDUTXMULTIPLERETRIES);
		object_set_counter(row_get_obj(row, 4), mpdumultretries);
		
		mpdurxdups= prism_get_long(DOT11_OID_MPDURXDUPS);
		object_set_counter(row_get_obj(row, 5), mpdurxdups);
		
		rtssuc= prism_get_long(DOT11_OID_RTSSUCCESSFUL);
		object_set_counter(row_get_obj(row, 6), rtssuc);
		
		rtsfail = prism_get_long(DOT11_OID_RTSFAILED);
		object_set_counter(row_get_obj(row, 7), rtsfail);
		
		ackfail= prism_get_long(DOT11_OID_ACKFAILED);
		object_set_counter(row_get_obj(row, 8),ackfail);
		
		mpdurxsuc = prism_get_long(DOT11_OID_MPDURXSUCCESSFUL);
		object_set_counter(row_get_obj(row, 9), mpdurxsuc);
		
		msdurxmult = prism_get_long(GEN_OID_MSDURXMULTICAST);
		object_set_counter(row_get_obj(row, 10),msdurxmult);
		
		frameerr = prism_get_long(DOT11_OID_FRAMEERRORS);
		object_set_counter(row_get_obj(row, 11), frameerr);
		
		msdutxsuc= prism_get_long(GEN_OID_MSDUTXSUCCESSFUL);
		object_set_counter(row_get_obj(row, 12),msdutxsuc);
		
		privrxnokey = prism_get_long(DOT11_OID_PRIVRXNOKEY);
		object_set_counter(row_get_obj(row, 13), privrxnokey);
		
		row = row->next;
	}

	return(3);
}


struct moddescr_t mod_dot11mac_count = {

	name:		__FILE__,
	fn_ini:		init,
	fn_get:		get,
	oidlist:	".1.2.840.10036.2.2.1"

};




/*

The following schema entries are provided by this module :
O .1.2.840.10036.2.2.1
L dot11CountersEntry
T TABLE

	O .1.2.840.10036.2.2.1.1
	L dot11TransmittedFragmentCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.2
	L dot11MulticastTransmittedFrameCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.3
	L dot11FailedCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.4
	L dot11RetryCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.5
	L dot11MultipleRetryCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.6
	L dot11FrameDuplicateCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.7
	L dot11RTSSuccessCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.8
	L dot11RTSFailureCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.9
	L dot11ACKFailureCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.10
	L dot11ReceivedFragmentCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.11
	L dot11MulticastReceivedFrameCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.12
	L dot11FCSErrorCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.13
	L dot11TransmittedFrameCount
	T COUNTER
	A R

	O .1.2.840.10036.2.2.1.14
	L dot11WEPUndecryptableCount
	T COUNTER
	A R


*/
