/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <string.h>
#include <assert.h>

#include "schema.h"
#include "table.h"
#include "objdata.h"
#include "easyprism.h"
#include "objflag.h"
#include "mod_mibII_interfaces.h"
#include "pda.h"

/* PostInit because the MAC address is only available in post_init */
static int post_init(void)
{
	struct schent_t *sch, *ifsch;
	struct row_t *row, *ifrow;
	struct index_t *index;
	struct string_t *dot11MAC;
        struct Pdr_Struct *pdrPtr;
	
	struct object_t *obj;

	int ifnum = 0;
			
	sch = schema_get(".1.2.840.10036.2.1.1");
	assert(sch);

	row = sch->rowlist;
	while(row) {
		
		/* get dot11MACAddress object */
		obj = row_get_obj(row,0);
		
		/* check dot11MACAddress for empty value */
		dot11MAC = object_get_string(obj);
		if(memcmp(dot11MAC->data, "\x00\x00\x00\x00\x00\x00", 6) == 0) {
			
			/* get index from this row to retrieve ifnum */
			index = index_get(row);
			ifnum = object_get_int(index_get_obj(index, 0));
			index_done(index);
			
			/* find entry in iftable to get MAC address */
			ifsch = schema_get(".1.3.6.1.2.1.2.2.1");
			assert(ifsch);
			
			index = index_new(ifsch);
			object_set_int(index_get_obj(index, 0), ifnum);
			ifrow = row_get_by_index(index);
			index_done(index);
			
			dot11MAC = object_get_string(row_get_obj(ifrow, 5));
			object_set_string(obj, dot11MAC->data, dot11MAC->len);
			
		}
		
		/* get dot11ManufacturerID */
		obj = row_get_obj(row, 7);		
                pdrPtr = pda_Find_Pdr(PDR_OEM_NAME, PDR_START);
                if (pdrPtr != 0) {
                    object_set_string(obj, pdrPtr->PdrData, strnlen(pdrPtr->PdrData,31));
                }
			
                /* get dot11ProductID obj */
                obj = row_get_obj(row, 8);
                pdrPtr = pda_Find_Pdr(PDR_PRODUCT_NAME, PDR_START);
                if (pdrPtr != 0) {
                   object_set_string(obj, pdrPtr->PdrData, strnlen(pdrPtr->PdrData,31));
                }
                
		row = row->next;
	}

	return(0);


}


static int set(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;

	char *ifname;

	int rtsthresh;
	int shortretries;
	int longretries;
	int fragthresh;
	int maxtxlt;
	int maxrxlt;		
	

	struct object_t *obj;

	int ifnum = 0;
			
	sch = schema_get(".1.2.840.10036.2.1.1");
	assert(sch);

	row = sch->rowlist;
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
		ifname= ifindex2ifname(ifnum);
		index_done(index);

		// Put the ifname in the global...
		prism_set_ifname(ifname);
							
		obj = row_get_obj(row, 1);
		if(object_flag_isset(obj, OF_DIRTY)) {
			rtsthresh = object_get_int(obj);
			prism_set_long(DOT11_OID_RTSTHRESH, rtsthresh);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 2);
		if(object_flag_isset(obj, OF_DIRTY)) {
			shortretries = object_get_int(obj);
			prism_set_long(DOT11_OID_SHORTRETRIES, shortretries);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 3);
		if(object_flag_isset(obj, OF_DIRTY)) {
			longretries = object_get_int(obj);
			prism_set_long(DOT11_OID_LONGRETRIES, longretries);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 4);
		if(object_flag_isset(obj, OF_DIRTY)) {
			fragthresh = object_get_int(obj);
			prism_set_long(DOT11_OID_FRAGTHRESH, fragthresh);
			object_flag_clear(obj, OF_DIRTY);
		}
		obj = row_get_obj(row,5);
		if(object_flag_isset(obj, OF_DIRTY)) {
			maxtxlt = object_get_uint(obj);
			prism_set_long(DOT11_OID_MAXTXLIFETIME, maxtxlt);
			object_flag_clear(obj, OF_DIRTY);
		}	
		obj = row_get_obj(row, 6);
		if(object_flag_isset(obj, OF_DIRTY)) {
			maxrxlt = object_get_uint(obj);
			prism_set_long(DOT11_OID_MAXRXLIFETIME, maxrxlt);
			object_flag_clear(obj, OF_DIRTY);
		}

		row = row->next;
	}

	return(0);

}


struct moddescr_t mod_dot11mac_op = {

	name:		__FILE__,
	fn_pin:		post_init,
	fn_set:		set,
	oidlist:	".1.2.840.10036.2.1.1"

};


/*

The following schema entries are provided by this module :

O .1.2.840.10036.2.1.1
L dot11OperationEntry
T TABLE

	O .1.2.840.10036.2.1.1.1
	L dot11MACAddress
	T OCTETSTR
	A R
	R 6/6

	O .1.2.840.10036.2.1.1.2
	L dot11RTSThreshold
	T INTEGER
	A RW
	R 0/2347

	O .1.2.840.10036.2.1.1.3
	L dot11ShortRetryLimit
	T INTEGER
	A RW
	R 1/255

	O .1.2.840.10036.2.1.1.4
	L dot11LongRetryLimit
	T INTEGER
	A RW
	R 1/255

	O .1.2.840.10036.2.1.1.5
	L dot11FragmentationThreshold
	T INTEGER
	A RW
	R 256/2346

	O .1.2.840.10036.2.1.1.6
	L dot11MaxTransmitMSDULifetime
	T INTEGER
	A RW
	R 1/2147483647

	O .1.2.840.10036.2.1.1.7
	L dot11MaxReceiveLifetime
	T INTEGER
	A RW
	R 1/2147483647

	O .1.2.840.10036.2.1.1.8
	L dot11ManufacturerID
	T OCTETSTR
	A R
	R 0/128

	O .1.2.840.10036.2.1.1.9
	L dot11ProductID
	T OCTETSTR
	A R
	R 0/128

*/
