/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>

#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "objflag.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"

static int set(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;
	char *ifname;
	
	struct object_t *obj;
	int antennarx;
	int antennatx;
	int ifnum = 0;
			
	sch = schema_get(".1.2.840.10036.4.2.1");
	assert(sch);

	row = sch->rowlist;
	
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
			
		ifname = ifindex2ifname(ifnum);

		index_done(index);

		// Put the ifname in the global...
	
		prism_set_ifname(ifname);
/* Stardate 10308.7 Max Yin
                    Follow the defaults/datastore.dat's setting to disable or enable antenna diversity
                   Stardate 10308.8 Max Yin
                    Fix antenna diversity and Tx/Rx selection to antenna 1
		   Stardate 10309.19 Max Yin
		    Change antenna to antenna 0
                */
                /*
                prism_set_long(DOT11_OID_ANTENNADIVERSITY, 0);
                prism_set_long(DOT11_OID_ANTENNATX, 0);
                prism_set_long(DOT11_OID_ANTENNARX, 0);
                */
#if 1
		obj = row_get_obj(row, 0);
		if(object_flag_isset(obj, OF_DIRTY)) {
			antennatx = object_get_int(obj) - 1; /* MVC starts at 0 */
			prism_set_long(DOT11_OID_ANTENNATX, antennatx);
			object_flag_clear(obj, OF_DIRTY);
		}
		
		obj = row_get_obj(row, 2);
		if(object_flag_isset(obj, OF_DIRTY)) {
			antennarx = object_get_int(obj) - 1; /* MVC starts at 0 */
			prism_set_long(DOT11_OID_ANTENNARX, antennarx);
			object_flag_clear(obj, OF_DIRTY);
		}
#endif
		row = row->next;
	}

	return(0);

}

static int check(void) {
	
	struct schent_t *sch;
	struct schent_t *sch_antlist; /* schema of antenna list */
	struct row_t* row;
	struct row_t* row_antlist;
	struct index_t* index;
	struct index_t* index_antlist;
	int ifindex;
	int rx_match;
	int tx_match;
	
	sch = schema_get(".1.2.840.10036.4.2.1");
	assert(sch);
	
	sch_antlist = schema_get(".1.2.840.10036.4.8.1");
	assert(sch_antlist);
	
	row = sch->rowlist;
	
	while(row) {
		rx_match = -1;
		tx_match = -1;
		
		index = index_get(row);
		
		ifindex = object_get_int(index_get_obj(index,0));
		
		/* loop antennalist */
		row_antlist = sch_antlist->rowlist;
		while(row_antlist) {
			
			index_antlist = index_get(row_antlist);
			
			/* check if this is an entry for this interface */
			if(ifindex == object_get_int(index_get_obj(index_antlist,0))) {
				
				/* Check RX antenna */
				if (object_get_int(row_get_obj(row,2)) == 
						object_get_int(index_get_obj(index_antlist,1))) {
					/* Check if this antenna is RX capable */
					if (object_get_int(row_get_obj(row_antlist,2)) != 1) {
						/* FAIL */
						return R_CHECKFAILED;						
					}
					else {
						/* RX matched */
						rx_match = 1;
					}
				}
				
				/* Check TX antenna */
				if (object_get_int(row_get_obj(row,0)) == 
						object_get_int(index_get_obj(index_antlist,1))) {
					/* Check if this antenna is TX capable */
					if (object_get_int(row_get_obj(row_antlist,1)) != 1) {
						/* FAIL */
						return R_CHECKFAILED;						
					}
					else {
						/* TX matched */
						tx_match = 1;
					}
				}
				
			}
			
			index_done(index_antlist);
			row_antlist = row_antlist->next;
		}	
		
		
		if ((tx_match == -1) || (rx_match == -1)) {
			/* selected antenna was not found */
			return R_CHECKFAILED;
		}
		
		index_done(index);
		row = row->next;
	}
	
	return R_NOERR;	
	
}

struct moddescr_t mod_dot11phy_antenna = {

	name:		__FILE__,
	fn_set:		set,
	fn_chk:		check,			
	oidlist:	".1.2.840.10036.4.2.1"

};





/*

The following schema entries are provided by this module :
O .1.2.840.10036.4.2.1
L dot11PhyAntennaEntry
T TABLE

	O .1.2.840.10036.4.2.1.1
	L dot11CurrentTxAntenna
	T INTEGER32
	A RW
	R 1/255

	O .1.2.840.10036.4.2.1.2
	L dot11DiversitySupport
	T INTEGER
	A R
	R 1,2,3

	O .1.2.840.10036.4.2.1.3
	L dot11CurrentRxAntenna
	T INTEGER32
	A RW
	R 1/255


*/
