/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>

#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"


static int check(void)
{
	struct schent_t* sch;
	struct row_t* row;
	
	sch = schema_get(".1.2.840.10036.4.8.1");
	assert(sch);
	
	row = sch->rowlist;
	while(row) {
		
		/* dot11DiversitySelectionRx can only be true if 
		   dot11SupportedRxAntenna is true */
		if (object_get_int(row_get_obj(row, 3)) == 1) {
			if (object_get_int(row_get_obj(row,2)) != 1) {
				return R_CHECKFAILED;
			}
		}
		

		row = row->next;
	}
	
	return R_NOERR;

}

static int set(void) 
{
	struct schent_t* sch;
	struct row_t* row;
	struct object_t* obj;
	struct index_t* index;
	int count3 = 0;
	int count5 = 0;
	int ifnum;
	char *ifname;
	
	sch = schema_get(".1.2.840.10036.4.8.1");
	assert(sch);
	
	/* Walk the antennalist table. When the dot11DiversitySelectionRx
	 * is true (== 1) for more than one antenna for an interface the
	 * antenna diversity (TX AND RX) will be switched on */
	
	row = sch->rowlist;
	while(row) {
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
		index_done(index);				
		ifname = ifindex2ifname(ifnum);
		
		obj = row_get_obj(row, 3);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		
		if(object_get_int(obj) == 1) {
			switch(ifnum) {
				case 3:
					count3++;
					break;
				case 5:
					count5++;
					break;
				default:
					assert(1==0);
					break;
			}
		} 
		row = row->next;
	}

	prism_set_ifname("eth1");
	if(count3 > 1) {
		prism_set_long(DOT11_OID_ANTENNADIVERSITY, DOT11_DIVERSITY_BOTH);
	}
	else {
		prism_set_long(DOT11_OID_ANTENNADIVERSITY, DOT11_DIVERSITY_NONE);
	}

	prism_set_ifname("eth3");
	if(count5 > 1) {
		prism_set_long(DOT11_OID_ANTENNADIVERSITY, DOT11_DIVERSITY_BOTH);
	}
	else {
		prism_set_long(DOT11_OID_ANTENNADIVERSITY, DOT11_DIVERSITY_NONE);
	}
	
	return R_NOERR;

}

struct moddescr_t mod_dot11phy_antennalist = {

	name:		__FILE__,
	fn_chk:		check,
	fn_set:		set,
	oidlist:	".1.2.840.10036.4.8.1"

};

/*


The following schema entries are provided by this module :

O .1.2.840.10036.4.8.1
L dot11AntennasListEntry
T TABLE

	O .1.2.840.10036.4.8.1.1
	L dot11AntennaListIndex
	T INTEGER32
	R 1/255
	I 2

	O .1.2.840.10036.4.8.1.2
	L dot11SupportedTxAntenna
	T INTEGER
	A RW
	R 1,2

	O .1.2.840.10036.4.8.1.3
	L dot11SupportedRxAntenna
	T INTEGER
	A RW
	R 1,2

	O .1.2.840.10036.4.8.1.4
	L dot11DiversitySelectionRx
	T INTEGER
	A RW
	R 1,2

*/
