/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <string.h>
#include <assert.h>
#include <stdio.h>

#include <aplog.h>

#include "schema.h"
#include "table.h"
#include "objdata.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"
#include "objflag.h"
#include "object.h"
#include "handler.h"
#include "regulatory.h"
#include "serialize.h"
#include "ds_defaults.h"
#include "country_information.h"

//#define DEBUG

/* function returns the default channel, depending on the current country
 * /domain */
static int get_default_channel(int ifnum) {
	
	struct object_t* obj;
	char oidbuf[MAX_OID_LENGTH];
	int regdomain;
	struct string_t *country;
	struct country_domain_t* cd;
	struct regulatory_group_list_t *rgl;
	struct regulatory_group_t *rg;
	
	/* TEMP for X_BOW */
	if(ifnum == 5) {
		return 6;
	}
		
	/* get the current regdomain for this interface */
	snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.4.1.1.2.%u", ifnum);
	obj = object_get(oidbuf);
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	regdomain = object_get_int(obj);
	
	if(regdomain != 0x00) {
		cd = regulatory_get_domain(regulatory_value2domainid(regdomain));
	}	
	else {
		/* no regdomain, lookup current country */
		snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.23.%u", ifnum);
		obj = object_get(oidbuf);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		country = object_get_string(obj);
		assert(country);
		cd = regulatory_get_country(country->data);
	}
	
	/* at this point we should have a country_domain struct */
	assert(cd);
	
	/* walk all regulatory_groups to get the default frequency */
	rgl = cd->regul_groups;
	while(rgl) {
		rg = rgl->regul_grp;
		if(rg->default_freq2 != 0) {
			
			/* found the default frequency for the 2.4Ghz band */
			return(regulatory_freq2channel(rg->default_freq2));
		}
		
		rgl = rgl->next;
	}
	
	
	/* hmm. we should get here */
	return(0);

}

/* checks if the given channel is valid for this interface */
static int channel_is_valid(int iface, int channel, int phytype) {
	
	struct schent_t* sch; 
	struct row_t* row; 
	struct index_t* idx;
	struct object_t* obj;
	struct string_t* country;
	struct string_t* country_list;
	char oidbuf[MAX_OID_LENGTH];
	int country_index = -1;
	int cif;
	int cs_phytype;
	int start_channel;
	int channel_gap;
	int no_channels;
	
	/* TODO REMOVE THE NEXT LINES WHEN THE MICROAP IS FUNCTIONAL */
	if(iface == 5) { 
		return R_NOERR;
	}
	
	
	/* get the current CountryCode for this interface */
	snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.23.%u", iface);
	obj = object_get(oidbuf);
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	country = object_get_string(obj);
#ifdef DEBUG	
	printf("Country found: (%s)\n", country->data);
#endif
	/* phytype = ERP == 6 (fixed in DSSS module */
	
	/* Find the index for this country in the supported countries
	 * table 
	 */
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.6.1");
	assert(sch);
	row = sch->rowlist;
	while(row) {
		
		/* get the interface of this row */
		idx = index_get(row);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		cif = object_get_int(obj);
		index_done(idx);
		
		if(iface != cif) {
			goto next1;
		}		
		
		obj = row_get_obj(row, 1);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		
		if(strcmp(country->data, (object_get_string(obj))->data) == 0) {
			/* found */
			obj = row_get_obj(row, 0);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			country_index = object_get_int(obj);
#ifdef DEBUG			
			printf("country index = %i\n", country_index);
#endif			
			break;
		}
next1:		
		row = row->next;
	}
	assert(country_index != -1); /* we should find it */
	
	/* loop channelset table */
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.8.1");
	assert(sch);
	row = sch->rowlist;
	while(row) {
		/* get the interface of this row */
		idx = index_get(row);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		cif = object_get_int(obj);
		index_done(idx);
		
		if(iface != cif) {
			goto next2;
		}

		/* check the phytype for this row */
		obj = row_get_obj(row, 1);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		cs_phytype = object_get_int(obj);
		
		if(cs_phytype != phytype) {
			goto next2;
		}
				
		/* get the countrylist for this row */
		obj = row_get_obj(row, 5);
		assert(obj);
		country_list = object_get_string(obj);
		
		/* check if the found country_index is part of this list */
		if (memchr(country_list->data, country_index, country_list->len) == NULL) {
			goto next2;
		}
		
		obj = row_get_obj(row, 2);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		start_channel = object_get_int(obj);

		obj = row_get_obj(row, 3);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		no_channels = object_get_int(obj);
		
		obj = row_get_obj(row, 4);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		channel_gap = object_get_int(obj);
		
#ifdef DEBUG
		fprintf(stderr, "channel       = %i\n", channel);
		fprintf(stderr, "start_channel = %i\n", start_channel);
		fprintf(stderr, "channel_gap   = %i\n", channel_gap);
		fprintf(stderr, "no_channels   = %i\n", no_channels);
#endif
				
		/* check start */
		if(channel < start_channel) {
			goto next2;
		}
		
		/* check gap */
		if(((channel - start_channel) % channel_gap) != 0) {
			goto next2;
		}
		
		/* check no_channels (if no_channels <= 2 then gap can be 0) */
		if( no_channels <= 1 ) {
			if (channel != start_channel) {
				goto next2;
			}
		}
		else {
			if(((channel - start_channel) / channel_gap) >= no_channels) {
				goto next2;
			}
		}
		
		/* channel is OK */
		return R_NOERR;
		
next2:		
		row = row->next;
	}
	
	return R_CHECKFAILED;
}

int mod_dot11phy_dsss_channel_valid(int iface, int channel, int phytype) {
	return(channel_is_valid(iface, channel, phytype));
}
	
static int set(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;
	char *ifname;
	
	struct object_t *obj;
	int ifnum = 0;
	int channel;
	int ccamode;
	int edthresh;
			
	sch = schema_get(".1.2.840.10036.4.5.1");
	assert(sch);
	row = schema_get_first_row(sch);
	
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		index = index_get(row);
		obj = index_get_obj(index,0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
			
		ifname = ifindex2ifname(ifnum);

		index_done(index);

		// Put the ifname in the global...
	
		prism_set_ifname(ifname);

		obj = row_get_obj(row, 0);
		if(object_flag_isset(obj, OF_DIRTY)) {
			assert(obj->flags & OF_DATAPRESENT);
			channel = object_get_int(obj);
			prism_set_long(DOT11_OID_CHANNEL, channel);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 2);
		if(object_flag_isset(obj, OF_DIRTY)) {
			assert(obj->flags & OF_DATAPRESENT);
                        ccamode = object_get_int(obj);
                        prism_set_long(DOT11_OID_CCAMODE, ccamode);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 3);
		if(object_flag_isset(obj, OF_DIRTY)) {
			assert(obj->flags & OF_DATAPRESENT);
			edthresh = object_get_int(obj);
			prism_set_long(DOT11_OID_EDTHRESHOLD, edthresh);
			object_flag_clear(obj, OF_DIRTY);
		}

		row = row->next;
	}

	return(0);

}

static int get(void) {
	
	struct schent_t *sch, *smt_sch;
	struct row_t * row, *smt_row;
	struct object_t* obj;
	struct index_t* idx;
	int ifnum, phytype, channel;
	char oidbuf[MAX_OID_LENGTH];
	
	sch = schema_get(".1.2.840.10036.4.5.1");
	assert(sch);
	
	smt_sch = schema_get(".1.3.6.1.4.1.937.2.2.2.1.1"); /* dot11ext smttable */
	assert(smt_sch);
	smt_row = schema_get_first_row(smt_sch);
	while(smt_row) {
		/* get ifnum for this row */
		idx = index_get(smt_row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		/* get phytype */
		obj = row_get_obj(smt_row, 16);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		phytype = object_get_int(obj);
		
		/* try to get the current channel to see if a row for this interface
		 * exists */
		snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.4.5.1.1.%u", ifnum);
		obj = object_get(oidbuf); /* try to get current channel */
		
		if(phytype == 6 /*erp*/) {
			/* a row should exist for this interface, if not create it */
			if(!obj) {
				/* row does not exist, create it */
				idx = index_new(sch);
				obj = index_get_obj(idx, 0);
				assert(obj);
				object_set_int(obj, ifnum);
				row = row_create_by_index(idx);
				assert(row);
				index_done(idx);
				
				obj = row_get_obj(row, 0); /* current channel */
				assert(obj);
				/* get the default channel for the current country/domain */
				channel = get_default_channel(ifnum);
				object_set_int(obj, channel);
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				/* 3893 only support CCA mode ED ...*/
				obj = row_get_obj(row, 1); /* dot11CCAModeSupported */
				assert(obj);
				object_set_int(obj, 1); /* CCA_ED */
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				/* Current CCA mode is therefore fixed */
				obj = row_get_obj(row, 2); /* dot11CurrentCCAMode */
				assert(obj);
				object_set_int(obj, 1); /* CCA_ED */
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				/* EDT threshold */
				obj = row_get_obj(row, 3); /* dot11EDThreshold */
				assert(obj);
				object_set_int(obj, 0x44); /* TODO: Get a value for this */
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				/* The dot11OperationRates should be set for this interface */
				snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.11.%u", ifnum);	
				obj = object_get(oidbuf);
				assert(obj);
				deserialize(obj, DEFAULT_OPERATIONAL_RATES_802_11G);
				object_flag_set(obj, OF_DIRTY);
				
				/* And the extended rates */
				snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.22.%u", ifnum);	
				obj = object_get(oidbuf);
				assert(obj);
				deserialize(obj, DEFAULT_EXTENDED_RATES_802_11G);
				object_flag_set(obj, OF_DIRTY);
			}
		}
		else {
			/* no row should exist for this interface. if so, remove it */
			if(obj) {
				/* get the row for this object */
				row = obj->row;
				assert(row);
				row_delete(row);
			}
		}
		
		smt_row = smt_row->next;	
	}
	
	return 1;
}

/* The DSSS table only shows 802.11g wireless interfaces. When the selected
   phytype changes, rows should be removed or created. This function checks
   the validity of the rows and takes nececary actions */
static int phytype_updated(void) {

	struct schent_t *sch, *smt_sch;
	struct row_t * row, *smt_row;
	struct object_t* obj;
	struct index_t* idx;
	int ifnum, phytype, channel;
	char oidbuf[MAX_OID_LENGTH];
	
	sch = schema_get(".1.2.840.10036.4.5.1");
	assert(sch);
	
	smt_sch = schema_get(".1.3.6.1.4.1.937.2.2.2.1.1"); /* dot11ext smttable */
	assert(smt_sch);
	smt_row = schema_get_first_row(smt_sch);
	while(smt_row) {
		/* get ifnum for this row */
		idx = index_get(smt_row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		prism_set_ifname(ifindex2ifname(ifnum));
		
		/* get phytype */
		obj = row_get_obj(smt_row, 16);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		phytype = object_get_int(obj);
		
		/* try to get the current channel to see if a row for this interface
		 * exists */
		snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.4.5.1.1.%u", ifnum);
		obj = object_get(oidbuf); /* try to get current channel */
		
		if(phytype == 6 /*erp*/) {
			/* a row should exist for this interface, if not create it */
			if(!obj) {
				/* row does not exist, create it */
				idx = index_new(sch);
				obj = index_get_obj(idx, 0);
				assert(obj);
				object_set_int(obj, ifnum);
				row = row_create_by_index(idx);
				assert(row);
				index_done(idx);
				
				obj = row_get_obj(row, 0); /* current channel */
				assert(obj);
				/* get the default channel for the current country/domain */
				channel = get_default_channel(ifnum);
				object_set_int(obj, channel);
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				/* 3893 only support CCA mode ED ...*/
				obj = row_get_obj(row, 1); /* dot11CCAModeSupported */
				assert(obj);
				object_set_int(obj, 1); /* CCA_ED */
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				/* Current CCA mode is therefore fixed */
				obj = row_get_obj(row, 2); /* dot11CurrentCCAMode */
				assert(obj);
				object_set_int(obj, 1); /* CCA_ED */
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				/* EDT threshold */
				obj = row_get_obj(row, 3); /* dot11EDThreshold */
				assert(obj);
				object_set_int(obj, 0x44); /* TODO: Get a value for this */
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				/* The dot11OperationRates should be set for this interface */
				snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.11.%u", ifnum);	
				obj = object_get(oidbuf);
				assert(obj);
				deserialize(obj, DEFAULT_OPERATIONAL_RATES_802_11G);
				object_flag_set(obj, OF_DIRTY);
				/* schedule a handler for this oid*/
				handler_call(oidbuf, H_SET, 3);
				
				/* And the extended rates */
				snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.22.%u", ifnum);	
				obj = object_get(oidbuf);
				assert(obj);
				deserialize(obj, DEFAULT_EXTENDED_RATES_802_11G);
				object_flag_set(obj, OF_DIRTY);
				/* schedule a handler for this oid*/
				handler_call(oidbuf, H_SET, 3);
				
				/* set the Country Information element for FULLMAC */
				if(ifnum == 3) {
					country_information_set_ie(ifnum);
				}
			}
		}
		else {
			/* no row should exist for this interface. if so, remove it */
			if(obj) {
				/* get the row for this object */
				row = obj->row;
				assert(row);
				row_delete(row);
			}
		}
		
		smt_row = smt_row->next;	
	}
	
	/* call the set handler to submit all the (new) settings to the MVC */
	set();
	
	
	return R_NOERR;
	
}

/* trigger function for updated regdomain or desired country 
 * Check validity of channels */
static int country_domain_updated(void) {
	
	struct schent_t* sch;
	struct row_t* row; 
	struct index_t* idx;
	struct object_t* obj;
	int ifnum;
	int channel;
	
	
	sch = schema_get(".1.2.840.10036.4.5.1");
	assert(sch);
	
	row = schema_get_first_row(sch);
	while(row) {
		
		/* get ifnum */
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		/* get channel */
		obj = row_get_obj(row, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		channel = object_get_int(obj);
		
		if( channel_is_valid(ifnum, channel, 6) != R_NOERR) {
			/* invalid channel, set to default */
			channel = get_default_channel(ifnum);
			object_set_int(obj, channel);
		}		
			
		row = row->next;
	}

	return R_NOERR;
}

static int init(void) {
	struct handler_t* hnd;
	
	/* register our update handler to the oid's we depend on and from
	 * whom we should know when they are set */
	hnd = handler_add(H_SET, phytype_updated);
	/* selected PhyType in dot11ext2 smt table */
	handler_connect(hnd, ".1.3.6.1.4.1.937.2.2.2.1.1");
	

	hnd = handler_add(H_SET, country_domain_updated);	
	/* current Regdomain in dot11PhyOperationTable */
	handler_connect(hnd, ".1.2.840.10036.4.1.1");
	/* Desired country */
	/* TODO */
	
	
	
	return R_NOERR;
	
}

static int post_init(void) {

	/* call our extra handlers for initialisation */
	phytype_updated();
	country_domain_updated();
	
	return R_NOERR;
}



static int check(void) {
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *idx;
	struct object_t* obj;
	int channel;
	int ifnum = 0;
	
	/* get dot11phy_dss schema */		
	sch = schema_get(".1.2.840.10036.4.5.1");
	assert(sch);
	
	row = schema_get_first_row(sch);
	
	while(row) {
		
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		/*********************************/
		/* check validity of channel     */
		/*********************************/
		obj = row_get_obj(row,0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		channel = object_get_int(obj);
		
		if(channel_is_valid(ifnum, channel, 6) != R_NOERR) {
			return R_CHECKFAILED;
		}
		
		row = row->next;
	}
	
	return R_NOERR;

}


struct moddescr_t mod_dot11phy_dsss = {

	name:		__FILE__,
	fn_ini:		init,
	fn_set:		set,
	fn_chk:		check,
	fn_pin:		post_init,
	fn_get:		get,
	oidlist:	".1.2.840.10036.4.5.1"

};




/*

The following schema entries are provided by this module :
O .1.2.840.10036.4.5.1
L dot11PhyDSSSEntry
T TABLE

0	O .1.2.840.10036.4.5.1.1
	L dot11CurrentChannel
	T INTEGER
	A RW
	R 1/14

1	O .1.2.840.10036.4.5.1.2
	L dot11CCAModeSupported
	T INTEGER
	A R
	R 1/7

2	O .1.2.840.10036.4.5.1.3
	L dot11CurrentCCAMode
	T INTEGER
	A RW
	R 1,16,2,4,8

3	O .1.2.840.10036.4.5.1.4
	L dot11EDThreshold
	T INTEGER32
	A RW



*/
