/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <string.h>
#include <assert.h>
#include <stdio.h>

#include "schema.h"
#include "table.h"
#include "objdata.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"
#include "objflag.h"
#include "object.h"
#include "handler.h"
#include "regulatory.h"
#include "serialize.h"
#include "ds_defaults.h"
#include "mod_dot11phy_dsss.h"
#include "country_information.h"

#define DOT11_OID_COUNTRY               0x1f000002
struct obj_frequency
{
    uint16_t mhz;
    uint16_t reserved;
};

struct obj_country
{
    uint8_t string[4];
    struct obj_frequency frequencies[1];
};
		
static int get_default_softmac_channel(ifnum) {
	
	char* ifname;
	char buf[1024];
	
	ifname = ifindex2ifname(ifnum);
	
	/* clear first 4 bytes of buf */
	memset(buf, 0x00, 4);
	
	/* get current country from SoftMac */
	if(get_prismoid(ifname, DOT11_OID_COUNTRY, buf, sizeof(buf)) == 0) {
		if(get_prismoid(ifname, DOT11_OID_COUNTRY, buf, sizeof(buf)) == 0) {
			int i = 0;
			struct obj_country *country = (struct obj_country*) buf;
						
			while(country->frequencies[i].mhz != 0) {
					
				if(country->frequencies[i].mhz > 3000) {
					return regulatory_freq2channel(country->frequencies[i].mhz);
				}				
				i++;
			}
		}
	}
	/* fallback */
	return 0;
}


/* function returns the default frequency, depending on the current country
 * /domain */
static int get_default_channel(int ifnum) {
	
	struct object_t* obj;
	char oidbuf[MAX_OID_LENGTH];
	int regdomain;
	struct string_t *country;
	struct country_domain_t* cd;
	struct regulatory_group_list_t *rgl;
	struct regulatory_group_t *rg;
	
	/* temp XBOW workaround */
	if(ifnum == 5) {
		return(get_default_softmac_channel(ifnum));
	}
		
	/* get the current regdomain for this interface */
	snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.4.1.1.2.%u", ifnum);
	obj = object_get(oidbuf);
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	regdomain = object_get_int(obj);
	
	if(regdomain != 0x00) {
		cd = regulatory_get_domain(regulatory_value2domainid(regdomain));
	}	
	else {
		/* no regdomain, lookup current country */
		snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.23.%u", ifnum);
		obj = object_get(oidbuf);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		country = object_get_string(obj);
		assert(country);
		cd = regulatory_get_country(country->data);
	}
	
	/* at this point we should have a country_domain struct */
	assert(cd);
	
	/* walk all regulatory_groups to get the default frequency */
	rgl = cd->regul_groups;
	while(rgl) {
		rg = rgl->regul_grp;
		if(rg->default_freq5 != 0) {
			
			/* found the default frequency for the 5Ghz band */
			return(regulatory_freq2channel(rg->default_freq5));
		}
		
		rgl = rgl->next;
	}
	
	
	/* hmm. we should get here */
	return(0);

}


static int set(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;
	char *ifname;
	
	struct object_t *obj;
	int ifnum = 0;
	int channel;
			
	sch = schema_get(".1.2.840.10036.4.11.1");
	assert(sch);

	row = schema_get_first_row(sch);
	
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		index = index_get(row);
		obj = index_get_obj(index, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
			
		ifname = ifindex2ifname(ifnum);

		index_done(index);

		// Put the ifname in the global...
	
		prism_set_ifname(ifname);

		obj = row_get_obj(row, 0);
		if(object_flag_isset(obj, OF_DIRTY) && (obj->flags & OF_DATAPRESENT)) {
			assert(obj->flags & OF_DATAPRESENT);
			channel = object_get_int(obj);
			if(channel == 0) {
				channel = get_default_channel(ifnum);
				object_set_int(obj, channel);
			}
			prism_set_long(DOT11_OID_CHANNEL, channel);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 1); /* dot11TIThreshold */
		if(object_flag_isset(obj, OF_DIRTY)) {
			/* TODO */
			object_flag_clear(obj, OF_DIRTY);
		}

		row = row->next;
	}

	return(0);

}


/* The ODFM table only shows 802.11 wireless interfaces. When the selected
   phytype changes, rows should be removed or created. This function checks
   the validity of the rows and takes nececary actions */
static int phytype_updated(void) {

	struct schent_t *sch, *smt_sch;
	struct row_t * row, *smt_row;
	struct object_t* obj;
	struct index_t* idx;
	int ifnum, phytype, channel;
	char oidbuf[MAX_OID_LENGTH];
	
	sch = schema_get(".1.2.840.10036.4.11.1");
	assert(sch);
	
	smt_sch = schema_get(".1.3.6.1.4.1.937.2.2.2.1.1"); /* dot11ext smttable */
	assert(smt_sch);
	smt_row = schema_get_first_row(smt_sch);
	while(smt_row) {
		/* get ifnum for this row */
		idx = index_get(smt_row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		prism_set_ifname(ifindex2ifname(ifnum));
		
		/* get phytype */
		obj = row_get_obj(smt_row, 16);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		phytype = object_get_int(obj);
		
		/* try to get the current channel to see if a row for this interface
		 * exists */
		snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.4.11.1.1.%u", ifnum);
		obj = object_get(oidbuf); /* try to get current channel */
		
		if(phytype == 4 /*ofdm*/) {
			/* a row should exist for this interface, if not create it */
			if(!obj) {
				/* row does not exist, create it */
				idx = index_new(sch);
				obj = index_get_obj(idx, 0);
				assert(obj);
				object_set_int(obj, ifnum);
				row = row_create_by_index(idx);
				assert(row);
				index_done(idx);
				
				obj = row_get_obj(row, 0); /* current channel */
				assert(obj);
				/* get the default channel for the current country/domain */
				channel = get_default_channel(ifnum);
				object_set_int(obj, channel);
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				obj = row_get_obj(row, 1); /* dot11TIThreshold */
				assert(obj);
				object_set_int(obj, 0); /* TODO */
				
				obj = row_get_obj(row, 2); /* dot11FrequencyBandsSupported  */
				assert(obj);
				object_set_int(obj, 3); /* TODO */
				
				/* The dot11OperationRates should be set for this interface */
				snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.11.%u", ifnum);	
				obj = object_get(oidbuf);
				assert(obj);
				deserialize(obj, DEFAULT_OPERATIONAL_RATES_802_11A);
				object_flag_set(obj, OF_DIRTY);
				/* schedule a handler for this oid*/
				handler_call(oidbuf, H_SET, 3);
				
				/* And the extended rates should be empty for 802_11A */
				snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.22.%u", ifnum);	
				obj = object_get(oidbuf);
				assert(obj);
				deserialize(obj, DEFAULT_EXTENDED_RATES_802_11A);
				object_flag_set(obj, OF_DIRTY);
				/* schedule a handler for this oid*/
				handler_call(oidbuf, H_SET, 3);
				
				/* set the Country Information element for FULLMAC */
				if(ifnum == 3) {
					country_information_set_ie(ifnum);
				}
			
			}
		}
		else {
			/* no row should exist for this interface. if so, remove it */
			if(obj) {
				/* get the row for this object */
				row = obj->row;
				assert(row);
				row_delete(row);
			}
		}
		
		smt_row = smt_row->next;	
	}
	
	/* call the set handler to submit all the (new) settings to the MVC */
	set();
	return R_NOERR;
	
}

/* FIX */
static int get(void) {

	struct schent_t *sch, *smt_sch;
	struct row_t * row, *smt_row;
	struct object_t* obj;
	struct index_t* idx;
	int ifnum, phytype, channel;
	char oidbuf[MAX_OID_LENGTH];
	
	sch = schema_get(".1.2.840.10036.4.11.1");
	assert(sch);
	
	smt_sch = schema_get(".1.3.6.1.4.1.937.2.2.2.1.1"); /* dot11ext smttable */
	assert(smt_sch);
	smt_row = schema_get_first_row(smt_sch);
	while(smt_row) {
		/* get ifnum for this row */
		idx = index_get(smt_row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		/* get phytype */
		obj = row_get_obj(smt_row, 16);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		phytype = object_get_int(obj);
		
		/* try to get the current channel to see if a row for this interface
		 * exists */
		snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.4.11.1.1.%u", ifnum);
		obj = object_get(oidbuf); /* try to get current channel */
		
		if(phytype == 4 /*ofdm*/) {
			/* a row should exist for this interface, if not create it */
			if(!obj) {
				/* row does not exist, create it */
				idx = index_new(sch);
				obj = index_get_obj(idx, 0);
				assert(obj);
				object_set_int(obj, ifnum);
				row = row_create_by_index(idx);
				assert(row);
				index_done(idx);
				
				obj = row_get_obj(row, 0); /* current channel */
				assert(obj);
				/* get the default channel for the current country/domain */
				channel = get_default_channel(ifnum);
				object_set_int(obj, channel);
				/* mark this object as 'changed' */
				object_flag_set(obj, OF_DIRTY);
				
				obj = row_get_obj(row, 1); /* dot11TIThreshold */
				assert(obj);
				object_set_int(obj, 0); /* TODO */
				
				obj = row_get_obj(row, 2); /* dot11FrequencyBandsSupported  */
				assert(obj);
				object_set_int(obj, 3); /* TODO */
				
				/* The dot11OperationRates should be set for this interface */
				snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.11.%u", ifnum);	
				obj = object_get(oidbuf);
				assert(obj);
				deserialize(obj, DEFAULT_OPERATIONAL_RATES_802_11A);
				object_flag_set(obj, OF_DIRTY);
				
				/* And the extended rates should be empty for 802_11A */
				snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.22.%u", ifnum);	
				obj = object_get(oidbuf);
				assert(obj);
				deserialize(obj, DEFAULT_EXTENDED_RATES_802_11A);
				object_flag_set(obj, OF_DIRTY);
			}
		}
		else {
			/* no row should exist for this interface. if so, remove it */
			if(obj) {
				/* get the row for this object */
				row = obj->row;
				assert(row);
				row_delete(row);
			}
		}
		
		smt_row = smt_row->next;	
	}
	
	return 1;
	
}




/* trigger function for updated regdomain or desired country 
 * Check validity of channels */
static int country_domain_updated(void) {
	
	struct schent_t* sch;
	struct row_t* row; 
	struct index_t* idx;
	struct object_t* obj;
	int ifnum;
	int channel;
	
	
	sch = schema_get(".1.2.840.10036.4.11.1");
	assert(sch);
	
	row = schema_get_first_row(sch);
	while(row) {
		
		/* get ifnum */
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		/* get channel */
		obj = row_get_obj(row, 0);
		assert(obj);
		if(obj->flags & OF_DATAPRESENT) {
			channel = object_get_int(obj);
		}
		else {
			channel = 0;
		}
		
		if((!(obj->flags & OF_DATAPRESENT)) || (mod_dot11phy_dsss_channel_valid(ifnum, channel, 4) != R_NOERR)) {
			/* invalid channel, set to default */
			channel = get_default_channel(ifnum);
			object_set_int(obj, channel);
		}		
			
		row = row->next;
	}

	return R_NOERR;
}

static int init(void) {
	struct handler_t* hnd;
	
	/* register our update handler to the oid's we depend on and from
	 * whom we should know when they are set */
	hnd = handler_add(H_SET, phytype_updated);
	/* selected PhyType in dot11ext2 smt table */
	handler_connect(hnd, ".1.3.6.1.4.1.937.2.2.2.1.1");
	

	hnd = handler_add(H_SET, country_domain_updated);	
	/* current Regdomain in dot11PhyOperationTable */
	handler_connect(hnd, ".1.2.840.10036.4.1.1");
	/* Desired country */
	/* TODO */
	
	
	
	
	return R_NOERR;
	
}


static int check(void) {
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *idx;
	struct object_t* obj;
	int channel;
	int ifnum = 0;
	
	/* get dot11phy_dss schema */		
	sch = schema_get(".1.2.840.10036.4.11.1");
	assert(sch);
	
	row = schema_get_first_row(sch);
	
	while(row) {
		
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		/*********************************/
		/* check validity of channel     */
		/*********************************/
		obj = row_get_obj(row,0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		channel = object_get_int(obj);
		
		if(mod_dot11phy_dsss_channel_valid(ifnum, channel, 4) != R_NOERR) {
			return R_CHECKFAILED;
		}
		
		row = row->next;
	}
	
	return R_NOERR;

}

int post_init(void) {
	
	/* call our extra handlers for initialisation */
	phytype_updated();
	country_domain_updated();
	
	return R_NOERR;
}

struct moddescr_t mod_dot11phy_ofdm = {

	name:		__FILE__,
	fn_ini:		init,
	fn_set:		set,
	fn_chk:		check,
	fn_pin:		post_init,
	fn_get:		get,
	oidlist:	".1.2.840.10036.4.11.1"

};


/*

The following schema entries are provided by this module :

O .1.2.840.10036.4.11.1
L dot11PhyOFDMEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1

	O .1.2.840.10036.4.11.1.1
	L dot11CurrentFrequency
	T INTEGER
	A RW
	R 0/99

	O .1.2.840.10036.4.11.1.2
	L dot11TIThreshold
	T INTEGER
	A RW

	O .1.2.840.10036.4.11.1.3
	L dot11FrequencyBandsSupported
	T INTEGER
	A R
	R 1/7


*/
