/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "table.h"
#include "object.h"
#include "objdata.h"
#include "objflag.h"
#include "schema.h"
#include "handler.h"
#include "regulatory.h"
#include "country_information.h"
#include "calibration.h"
#include "pda.h"
#include "prismoids.h"

static struct Pdr_Struct *get_default_country_PDR(ifindex) {
	
	switch(ifindex) {
		case 3:
			return(pda_Find_Pdr(PDR_DEFAULT_COUNTRY, NULL));
			break;
		default:
			return (NULL);
			break;
	}
	
	return NULL;
}

/* this function is called if the dot11ext2 smt table is updated */
static int desired_country_updated(void) {
	struct schent_t* sch;
	struct row_t* row;
	struct object_t* obj;
	struct index_t* idx;
	char oidbuf[MAX_OID_LENGTH];
	int regdom, ifnum;
	char* regdom_id;
	struct string_t* country;
	struct country_domain_t* cd;
	
	sch = schema_get(".1.2.840.10036.4.1.1");
	assert(sch);
	
	row = schema_get_first_row(sch);
	while(row) {
		
		/* get the interfacenr for this row out of the index */
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx,0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		/* get the desired country for this interface */
		snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.15.%u", ifnum);
		obj = object_get(oidbuf);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		country = object_get_string(obj);
		
		switch(ifnum) {
			case 3: /* FULL MAC */
				/* get the current regdomain from this row */
				obj = row_get_obj(row, 1);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				regdom = object_get_int(obj);
		
				regdom_id = regulatory_value2domainid(regdom);
				/* get the corresponding country_domain struct for this domain */
				cd = regulatory_get_domain(regdom_id);
				
				/* check if the current desired country is part of this domain */
				if((!cd) || (regulatory_country_in_domain(cd, country->data) != R_NOERR)) {
				/* country is not in this domain, update domain */
			
					/* try to find a domain in which this country is valid */
					cd = regulatory_get_country_domains();
					while(cd) {
						
						if(cd->type == REGULATORY_TYPE_DOMAIN) {
					
							if(regulatory_country_in_domain(cd, country->data) == R_NOERR) {
								/* found one, adapt regdomain */
								regdom = regulatory_domainid2value(cd->id);
								object_set_int(obj, regdom);
						
								/* update the 'current' country/domain */
								regulatory_set_current_country_domain(ifnum, cd);
								
								break;
							}
						}				
				
						cd = cd->next;
					}
			
					if(!cd) {
						/* we didn't find a valid regdom, set regdom to 'other' */
						object_set_int(obj, 0x00);
				
						/* get country_domain struct for this country */
						cd = regulatory_get_country(country->data);
						assert(cd);
				
						regulatory_set_current_country_domain(ifnum, cd);
					}
			
					/* update/set the country information element for this interface */
					country_information_set_ie(ifnum);
			
					/* update the outputpower table for this interface */
					calibration_update_outputpowertable(ifnum);

				}
				break;
			case 5: /* softmac */
#define DOT11_OID_COUNTRY               0x1f000002
				
				set_prismoid("eth3", DOT11_OID_COUNTRY, country->data, 4);
				
				/* country countrystring in dot11StationConfig table */
				snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.23.%u", ifnum);
				obj = object_get(oidbuf);
				assert(obj);
				object_set_string(obj, country->data, 3);
				
				/* TODO SET OR CHECK REGDOMAIN */
				
				/* TODO SET/UPDATE CHANNEL */
				break;
			default:
				assert(0==1);
				break;
		}
		
		row = row->next;
	}
	
	return R_NOERR;
}


static int get(void) {
	
	struct row_t* row;
	struct schent_t* sch;
	struct object_t* obj, *selphy_obj;
	struct index_t* idx;
	int ifindex;
	char oidbuf[MAX_OID_LENGTH];
	
	sch = schema_get(".1.2.840.10036.4.1.1");
	
	row = sch->rowlist;
	while(row) {
		
		/* copy dot11PHYType from smtSelectedPhyType */

		/* get ifindex */
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifindex = object_get_int(obj);
		index_done(idx);
		
		snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.18.%u", ifindex);
		selphy_obj = object_get(oidbuf);
		assert(selphy_obj);
		assert(selphy_obj->flags & OF_DATAPRESENT);
		
		obj = row_get_obj(row, 0);
		assert(obj);
		object_set_int(obj, object_get_int(selphy_obj));
		
		row = row->next;
	}
	
	
	return(5);
}

static int init(void) {
	
	struct row_t* row;
	struct schent_t* sch;
	struct object_t* obj;
	int regdom;
	struct country_domain_t* cd;
	struct handler_t* hnd;
	struct index_t* idx;
	int ifnum;
	static struct Pdr_Struct *pdr;
	
	/* Check if the currentRegdomain is initialised. If not, get default. */
	sch = schema_get(".1.2.840.10036.4.1.1");
	assert(sch);
	
	row = sch->rowlist;
	while(row) {
		/* get the interfacenr for this row out of the index */
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx,0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		switch(ifnum) {
			case 3: /* FULL MAC INTERFACES */
						
				/* load powertable (from PDA) for this interface */
				calibration_load_powertable(ifnum);
		
				/* determine default regdomain */
				/* First try PDA */
				pdr = get_default_country_PDR(ifnum);
				if(!pdr) {
					cd = regulatory_get_default_country_domain();
					assert(cd);
					if (cd->type == REGULATORY_TYPE_DOMAIN) {
						regdom = regulatory_domainid2value(cd->id);
					}
					else {
						regdom = 0x00; /* no default regulatory domain*/
					}
				}
				else {
					/* First octet defines reg domain. */
					regdom = *(pdr->PdrData);
					if(regdom == 0x00) {
						/* There is a default country defined in the PDR,
						 * check if this country is part of a domain */
				
						char idbuf[4] = { 0x00, 0x00, 0x00, 0x00 };
						memcpy(idbuf, &(pdr->PdrData[1]), 3);
				
						cd = regulatory_get_country_domain_by_country_code(idbuf);
						/* this country SHOULD be present, as country or as domain */
						assert(cd); 
						if(cd->type == REGULATORY_TYPE_DOMAIN) {
							regdom = regulatory_domainid2value(cd->id);
						}	
					}
				}
		
				obj = row_get_obj(row, 1);
				assert(obj);
		
				if(!(obj->flags & OF_DATAPRESENT)) {
					object_set_int(obj, regdom);
				}
				break;

			case 5: /* SOFTMAC init of regulatory domain is set during
			         * the initialisation of the desired country */

				break;
			default:
				assert(0==1);
				break;
		}	
		row = row->next;
	}	
	
	
	/* register our update handler on changes of the desired country */
	/* desired country is part of the dot11ext smt table */
	hnd = handler_add(H_SET, desired_country_updated);
	handler_connect(hnd, ".1.3.6.1.4.1.937.2.2.2.1.1");
	
	return R_NOERR;
}

static int set(void) {
	
	struct row_t* row;
	struct schent_t* sch;
	struct object_t* obj;
	struct index_t* idx;
	int ifnum, regdom;
	char *reg_id;
	struct country_domain_t* cd;
	
	sch = schema_get(".1.2.840.10036.4.1.1");
	assert(sch);
	
	row = sch->rowlist;
	while(row) {
		
		/* get the interfacenr for this row out of the index */
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx,0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(obj);
		index_done(idx);
		
		switch(ifnum) {
			case 3: /* FULL MAC INTERFACE(S) */
				obj = row_get_obj(row, 1); /* Current Regdomain */
				if(object_flag_isset(obj, OF_DIRTY)) {

					/* get the regdomain_id */
					assert(obj->flags & OF_DATAPRESENT);
					regdom = object_get_int(obj);
			
					if (regdom != 0x00) {
			
						reg_id = regulatory_value2domainid(regdom);
						assert(reg_id);

						/* get the corresponding country_domain struct and
			   	   		   set it to 'current' */
						cd = regulatory_get_domain(reg_id);
						assert(cd);
						regulatory_set_current_country_domain(ifnum, cd);

						/* update the Country IE for this interface */
						country_information_set_ie(ifnum);

						/* update the outputpower table for this interface */
						calibration_update_outputpowertable(ifnum);

						object_flag_clear(obj, OF_DIRTY);
					}
				}
				break;
			case 5: /* SOFTMAC INTERFACES */
				/* Regdomain is ignored for SoftMac interfaces,
				 * because we cannot retrieve the default country
				 * for a domain (yet).
				 */
				break;
			default:
				assert(0==1);
				break;
		}
		
		row = row->next;
	}
	
	
		
	return R_NOERR;
}

static int check(void)
{
	struct row_t* row, *domsup_row;
	struct schent_t* sch, *domsup_sch;
	struct object_t *obj;
	int regdom;
	int ifindex, domsup_ifindex;
	int regdom_ok;
	struct index_t* idx;
	
	
	/* check if the regdomains in the current rows are valid */
	
	sch = schema_get(".1.2.840.10036.4.1.1");
	assert(sch);
	domsup_sch = schema_get(".1.2.840.10036.4.7.1");
	assert(domsup_sch);
	
	row = sch->rowlist;
	while(row) {

		obj = row_get_obj(row, 1);
		assert(obj);
		
		/* only check 'dirty' objects */
		if(object_flag_isset(obj, OF_DIRTY)) {
			
			assert(obj->flags & OF_DATAPRESENT);
			regdom = object_get_int(obj);

			/* get interface index for this row */
			idx = index_get(row);
			assert(idx);
			obj = index_get_obj(idx, 0);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			ifindex = object_get_int(obj);
			index_done(idx);

			/* try to find regdom in supported domsup table */
			domsup_row = domsup_sch->rowlist;
			regdom_ok = 0; /*false*/
			while(domsup_row && !regdom_ok) {
				/* get iface index for this row */
				idx = index_get(domsup_row);
				assert(idx);
				obj = index_get_obj(idx,0);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				domsup_ifindex = object_get_int(obj);
				index_done(idx);
				
				if(domsup_ifindex == ifindex) {
					obj = row_get_obj(domsup_row,1);
					assert(obj);
					assert(obj->flags & OF_DATAPRESENT);
					if(object_get_int(obj) == regdom) {
						/* OK, regdom is supported */
						regdom_ok = 1;
					}
				}

				domsup_row = domsup_row->next;
			}

			if(!regdom_ok) {
				return R_CHECKFAILED;
			}
		}
		
		row = row->next;
	}
	
	return R_NOERR;
}

static int post_init(void) {
	/* call our update_handler(s) for initialisation */
	desired_country_updated();
	
	return R_NOERR;
}

struct moddescr_t mod_dot11phy_op = {

	name:		__FILE__,
	fn_chk:		check,
	fn_get:		get,
	fn_set:		set,
	fn_ini:		init,
	fn_pin:		post_init,
	oidlist:	".1.2.840.10036.4.1.1"

};




/*

The following schema entries are provided by this module :

O .1.2.840.10036.4.1.1
L dot11PhyOperationEntry
T TABLE

	O .1.2.840.10036.4.1.1.1
	L dot11PHYType
	T INTEGER
	A R
	R 1,2,3,4,5

	O .1.2.840.10036.4.1.1.2
	L dot11CurrentRegDomain
	T INTEGER32
	A RW

	O .1.2.840.10036.4.1.1.3
	L dot11TempType
	T INTEGER
	A R
	R 1,2


*/
