/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>

#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "objflag.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"
#include "calibration.h"

static int init(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct object_t  *power_obj;
	
	/* Until Beta 2, update 2 FW, the txpower table was not yet implemented.
	 * All the 8 powerlevels in the configuration were 0 mW. These 0 mW could
	 * still be in the userdata, which leads to an AP sending with almost no
	 * output power. To prevent this, the values are checked and if the 
	 * powerlevel 1 is 0 mW, the list is updated. This peace of code can be
	 * removed after a few releases.
	 */
			
	sch = schema_get(".1.2.840.10036.4.3.1");
	assert(sch);

	row = sch->rowlist;
	
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		power_obj = row_get_obj(row, 1);
		assert(power_obj);
		assert(power_obj->flags & OF_DATAPRESENT);
		
		if(object_get_int(power_obj) == 0) {
			object_set_int(power_obj, 100);
			object_set_int(row_get_obj(row, 2), 65);
			object_set_int(row_get_obj(row, 3), 40);
			object_set_int(row_get_obj(row, 4), 25);					
			object_set_int(row_get_obj(row, 5), 16);
			object_set_int(row_get_obj(row, 6), 10);
			object_set_int(row_get_obj(row, 7), 6);
			object_set_int(row_get_obj(row, 8), 4);
		}
	
		row = row->next;
	}

	return(0);

}


static int set(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;
	char *ifname;
	
	struct object_t *obj, *power_obj;
	int curtxlevel;
	int ifnum = 0;
			
	sch = schema_get(".1.2.840.10036.4.3.1");
	assert(sch);

	row = sch->rowlist;
	
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
			
		ifname = ifindex2ifname(ifnum);

		index_done(index);

		// Put the ifname in the global...
	
		prism_set_ifname(ifname);
		
		obj = row_get_obj(row, 9);
		if(object_flag_isset(obj, OF_DIRTY)) {
			curtxlevel = object_get_int(obj);
			
			assert((curtxlevel >= 1) && (curtxlevel <= 8));
			
			/* get the object corresponding to this level */
			power_obj = row_get_obj(row, curtxlevel);
			
			calibration_set_tx_powerlevel(ifnum, object_get_int(power_obj));
			
			object_flag_clear(obj, OF_DIRTY);
		}
	
		row = row->next;
	}

	return(0);

}


struct moddescr_t mod_dot11phy_txpower = {

	name:		__FILE__,
	fn_ini:		init,
	fn_set:		set,
	oidlist:	".1.2.840.10036.4.3.1"

};


/*

The following schema entries are provided by this module :

O .1.2.840.10036.4.3.1
L dot11PhyTxPowerEntry
T TABLE

0	O .1.2.840.10036.4.3.1.1
	L dot11NumberSupportedPowerLevels
	T INTEGER
	A R
	R 1/8

1	O .1.2.840.10036.4.3.1.2
	L dot11TxPowerLevel1
	T INTEGER
	A R
	R 0/10000

2	O .1.2.840.10036.4.3.1.3
	L dot11TxPowerLevel2
	T INTEGER
	A R
	R 0/10000

3	O .1.2.840.10036.4.3.1.4
	L dot11TxPowerLevel3
	T INTEGER
	A R
	R 0/10000

4	O .1.2.840.10036.4.3.1.5
	L dot11TxPowerLevel4
	T INTEGER
	A R
	R 0/10000

5	O .1.2.840.10036.4.3.1.6
	L dot11TxPowerLevel5
	T INTEGER
	A R
	R 0/10000

6	O .1.2.840.10036.4.3.1.7
	L dot11TxPowerLevel6
	T INTEGER
	A R
	R 0/10000

7	O .1.2.840.10036.4.3.1.8
	L dot11TxPowerLevel7
	T INTEGER
	A R
	R 0/10000

8	O .1.2.840.10036.4.3.1.9
	L dot11TxPowerLevel8
	T INTEGER
	A R
	R 0/10000

9	O .1.2.840.10036.4.3.1.10
	L dot11CurrentTxPowerLevel
	T INTEGER
	A RW
	R 1/8

*/


