/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>
#include <aplog.h>
#include "types.h"
#include "table.h"
#include "misc.h"
#include "object.h"
#include "objdata.h"
#include "objflag.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"
#define MAX_RATES 6

//#define DEBUG

/* Sets the Operational Rates & Extended Rates in the MVC */
void mod_dot11smt_sc_SetRates(int ifnum) {
	
	u_char MVC_rates[126];
	char oidbuf[MAX_OID_LENGTH];
	int i, j;
	u_char rate;
	
	struct object_t* rate_obj, *ext_rate_obj;
	struct string_t *rate_str, *ext_rate_str, *new_rate_str;
	
	int no_MVC_rates = 0;

	assert((ifnum == 3) || (ifnum == 5));
	
	prism_set_ifname(ifindex2ifname(ifnum));
	
	snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.1.1.11.%u", ifnum);
	rate_obj = object_get(oidbuf);
	assert(rate_obj);
	
	snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.22.%u", ifnum);
	ext_rate_obj = object_get(oidbuf);
	assert(ext_rate_obj);
	
	/* TODO check this. This prevents crashes during init, but shouldn't be here */
	if((!(rate_obj->flags & OF_DATAPRESENT)) || (!(ext_rate_obj->flags & OF_DATAPRESENT))) {
		return;		
	}

	rate_str = object_get_string(rate_obj);
	ext_rate_str = object_get_string(ext_rate_obj);
	
	for(i=0; i < rate_str->len; i++) {
		rate = rate_str->data[i];
		
		for(j=0; j < ext_rate_str->len; j++) {
			if(rate == ext_rate_str->data[j]) {
				break;
			}
		}
		if(j >= ext_rate_str->len) {
			/* didn't find it in ext_rates */
			MVC_rates[no_MVC_rates++] = rate;
		}
	}	

#ifdef DEBUG
	printf("aRates = ");
	for(i=0; i < no_MVC_rates; i++) {
		printf("%02X ", MVC_rates[i]);
	}
	printf("\n");
	printf("aExtendedRates = ");
	for(i=0; i < ext_rate_str->len; i++) {
		printf("%02X ", ext_rate_str->data[i]);
	}
	printf("\n");

	
#endif
	new_rate_str = mkstring(MVC_rates, no_MVC_rates);
	prism_set_string(DOT11_OID_RATES, new_rate_str);
	free(new_rate_str);
	prism_set_string(DOT11_OID_EXTENDEDRATES, ext_rate_str);


}


static int get(void)
{
	struct schent_t *sch;
	struct schent_t *ifsch;
	struct row_t *row;
	struct row_t *ifrow;
	struct index_t *index;
	struct index_t *ifindex;

	char *ifname;
	struct string_t *stationid;
	

	int ifnum = 0;
	int cfpoll;
	int privimpl;
	int disreason;
	struct string_t *disstation;
	int deauthreason;
	struct string_t *deauthstation;
	int authfailstatus;
	struct string_t *authfailstation;
			
	sch = schema_get(".1.2.840.10036.1.1.1");
	ifsch = schema_get(".1.3.6.1.2.1.2.2.1");
	assert(sch);
	assert(ifsch);

	row = sch->rowlist;
	
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
			
		// Now get the ifname for this ifnum from the iftable

		ifindex = index_new(ifsch);
		object_set_int(index_get_obj(ifindex, 0), ifnum);
		ifrow = row_get_by_index(ifindex);
		ifname = ifindex2ifname(ifnum);

		index_done(index);
		index_done(ifindex);

		// Put the ifname in the global...
	
		prism_set_ifname(ifname);
	
		// Station ID is a bit strange. It should default to the stations MAC addres,
		// but should be freely settable later. If the address is 00:00:00:00:00:00 
		// (factory default), the mac address is read from the iftable and put in here

		stationid = object_get_string(row_get_obj(row, 0));
		if(memcmp(stationid->data, "\x00\x00\x00\x00\x00\x00", 6) == 0) { 
			stationid = object_get_string(row_get_obj(ifrow, 5));
			object_set_string(row_get_obj(row, 0), stationid->data, stationid->len);
		}
	
		
		// CFPollable
		cfpoll = prism_get_long(DOT11_OID_MPDUTXSUCCESSFUL);
		object_set_int(row_get_obj(row, 2), cfpoll);
		
		// PrivacyOption
		privimpl= prism_get_long(GEN_OID_MSDUTXMULTICAST);
		object_set_int(row_get_obj(row, 6), privimpl);

		// PSM (always return 'active' 
		object_set_int(row_get_obj(row, 6), 1);
		
		// DisassociateReason
		disreason=  prism_get_mlme_code(DOT11_OID_DISASSOCIATE);
		object_set_int(row_get_obj(row, 14), disreason);
		
		//DisassociateStation
		disstation=  prism_get_mlme_addr(DOT11_OID_DISASSOCIATE);
		object_set_string(row_get_obj(row, 15), disstation->data, disstation->len);
		free(disstation);
		
		//DeauthenticateReason
		deauthreason=  prism_get_mlme_code(DOT11_OID_DEAUTHENTICATE);
		object_set_int(row_get_obj(row, 16), disreason);
		
		//DeauthenticateStation
		deauthstation=  prism_get_mlme_addr(DOT11_OID_DEAUTHENTICATE);
		object_set_string(row_get_obj(row, 17), deauthstation->data,deauthstation->len);
		free(deauthstation);
		
		//AuthenticateFailStatus
                authfailstatus=  prism_get_mlme_code(DOT11_OID_AUTHENTICATE);
		object_set_int(row_get_obj(row, 18), authfailstatus);
		
		//AuthenticateFailStation
                //Show only "really" failed authentications. 
                if (authfailstatus != DOT11_SC_SUCCESSFUL) 
                {
                    authfailstation=  prism_get_mlme_addr(DOT11_OID_AUTHENTICATE);
                    object_set_string(row_get_obj(row, 19), authfailstation->data,authfailstation->len);
                    free(authfailstation);		
                }
                else
                    object_set_string(row_get_obj(row, 19), "\x00\x00\x00\x00\x00\x00", 6);
		
                row = row->next;
	}

	return(10);
}


static int set(void)
{
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;
	char oidbuf[MAX_OID_LENGTH];
	char *ifname;

	unsigned int	medocclimit;
	unsigned int	cfpperiod;
	unsigned int	cfpmaxduration;
	unsigned int 	authresptimeout;
	unsigned int	pwrmgtmode;
	struct string_t *ssid;
	unsigned int	bsstype;
	unsigned int	beaconperiod;
	unsigned int	dtimperiod;
	unsigned int	assresptimeout;

	struct object_t *obj;
	int ShroudedSSID;
	int ifnum = 0;
			
	sch = schema_get(".1.2.840.10036.1.1.1");
	assert(sch);

	row = sch->rowlist;
	
	while(row) {

		// Get the indices of this row. ifnum is a the numeric 
		// index from the iftable
	
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
		ifname = ifindex2ifname(ifnum);
		index_done(index);

		// Put the ifname in the global...
	
		prism_set_ifname(ifname);

		obj = row_get_obj(row, 1);
		if(object_flag_isset(obj, OF_DIRTY)) {
			medocclimit = object_get_int(obj);
			prism_set_long(DOT11_OID_MEDIUMLIMIT, medocclimit);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 3);
		if(object_flag_isset(obj, OF_DIRTY)) {
			cfpperiod = object_get_int(obj);
			prism_set_long(DOT11_OID_CFPPERIOD, cfpperiod);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 4);
		if(object_flag_isset(obj, OF_DIRTY)) {
			cfpmaxduration = object_get_int(obj);
			prism_set_long(DOT11_OID_CFPDURATION, cfpmaxduration);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 5);
		if(object_flag_isset(obj, OF_DIRTY)) {
			authresptimeout = object_get_uint(obj);
			prism_set_long(DOT11_OID_AUTHRESPTIMEOUT, authresptimeout);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 7);
		if(object_flag_isset(obj, OF_DIRTY)) {
			pwrmgtmode = object_get_int(obj);
			//prism_set_long(DOT11_OID_PSM, pwrmgtmode);
			object_flag_clear(obj, OF_DIRTY);
		}
                
		obj = row_get_obj(row, 9);
		if(object_flag_isset(obj, OF_DIRTY)) {
			bsstype = object_get_int(obj);
			prism_set_long(DOT11_OID_BSSTYPE, bsstype);
			object_flag_clear(obj, OF_DIRTY);
		}
	
		obj = row_get_obj(row, 10);
		if(object_flag_isset(obj, OF_DIRTY)) {
			mod_dot11smt_sc_SetRates(ifnum);
			object_flag_clear(obj, OF_DIRTY);
		}
		obj = row_get_obj(row, 11);
		if(object_flag_isset(obj, OF_DIRTY)) {
			beaconperiod = object_get_int(obj);
			prism_set_long(DOT11_OID_BEACONPERIOD, beaconperiod);
			object_flag_clear(obj, OF_DIRTY);
		}

		obj = row_get_obj(row, 12);
		if(object_flag_isset(obj, OF_DIRTY)) {
			dtimperiod = object_get_int(obj);
			prism_set_long(DOT11_OID_DTIMPERIOD, dtimperiod);
			object_flag_clear(obj, OF_DIRTY);
		}
		obj = row_get_obj(row, 13);
		if(object_flag_isset(obj, OF_DIRTY)) {
			assresptimeout = object_get_uint(obj);
			prism_set_long(DOT11_OID_ASSOCRESPTIMEOUT, assresptimeout);
			object_flag_clear(obj, OF_DIRTY);
		}
		
                /*  
                    Don't check the dirty flag for this object.
                    Some OIDs of this table are "latched" and become active
                    by setting the SSID. Not checking the dirty flag results
                    in setting the SSID whenever an object in this table is changed.
                */
                obj = row_get_obj(row, 8);
                ssid = object_get_string(obj);
		prism_set_ssid(DOT11_OID_SSID, ssid);
		object_flag_clear(obj, OF_DIRTY);

		/* Setting of DOT11_OID_SSID overwrites the DOT11_OID_SSIDOVERRIDE 
		 * Therefore, if this OID was set before, it should be reset */
		snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.4.1.937.2.2.2.1.1.24.%d", ifnum);
		obj = object_get(oidbuf);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ShroudedSSID = object_get_int(obj);
		if(ShroudedSSID == 1) {
			struct string_t *ssid;
                        char null_ssid[] = "\00";
                        ssid = mkstring(null_ssid, 1);
                        assert(ssid);
                        prism_set_ssid(DOT11_OID_SSIDOVERRIDE,ssid);
		}
		
		row = row->next;
	}
	 /*Stardate 10307.21 by Max Yin
         * Originally, SIGALARM will generated by respawnd or apdpd(in the future) itself.
         * But, latest test found that, OID:10000003 read operation will crash WiSoC on a-little-high
         * load. Therefore, SIGALRM will be generated here, and respawnd/apdpd will complete the
         * eth1 reset state machine.
         */
	/*
	 * eth1 mon is move to sysd
	 */
        sleep(1);
        system("/bin/killall -ALRM sysd");
	return(0);

}

static int check() {
	struct schent_t* sch;
	struct row_t* row;
	struct object_t* obj;
	int i;
	struct string_t *rate_str;
	u_char rate;
	
	sch = schema_get(".1.2.840.10036.1.1.1");
	assert(sch);
	
	row = sch->rowlist;
	while(row) {
		
		/*************************************/
		/* OPERATIONAL RATES                 */
		/*************************************/
		
		obj = row_get_obj(row, 10);
		assert(obj);
		/* check only dirty objects */
		if(object_flag_isset(obj, OF_DIRTY)) {
			
			assert(obj->flags & OF_DATAPRESENT);

			/* At least one rate should be marked as Basic Rate */
			rate_str = object_get_string(obj);

			for(i=0; i < rate_str->len; i++) {
				rate = rate_str->data[i];

				if(rate & DOT11_BASICRATE) {
					break;
				}
			}

			if(i >= rate_str->len) {
				/* no match */
				aplog(EV_SNMP_CHECKFAILED, "At least one of the rates should have the Basic Rate bit set", NULL);
				return R_CHECKFAILED;
			}
		}		

		
				
		row = row->next;
	}
	
	return R_NOERR;
}

static int post_init() {
	
	struct schent_t *sch;
	struct row_t *row;
	struct index_t *index;
	int ifnum = 0;
			
	sch = schema_get(".1.2.840.10036.1.1.1");
	assert(sch);
	row = sch->rowlist;
	while(row) {
		// Get the indices of this row. ifnum is a the numeric 
		// index
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
		index_done(index);
		mod_dot11smt_sc_SetRates(ifnum);		
		row = row->next;
	}
	
	return R_NOERR;

}

struct moddescr_t mod_dot11smt_sc = {

	name:		__FILE__,
	fn_get:		get,
	fn_set:		set,
	fn_chk:		check,
	fn_pin:		post_init,
	oidlist:	".1.2.840.10036.1.1.1"

};





/*

The following schema entries are provided by this module :

O .1.2.840.10036.1.1.1
L dot11StationConfigEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1

0	O .1.2.840.10036.1.1.1.1
	L dot11StationID
	T OCTETSTR
	A RW
	R 6

1	O .1.2.840.10036.1.1.1.2
	L dot11MediumOccupancyLimit
	T INTEGER
	A RW
	R 0/1000

2	O .1.2.840.10036.1.1.1.3
	L dot11CFPollable
	T INTEGER
	A R
	R 1,2

3	O .1.2.840.10036.1.1.1.4
	L dot11CFPPeriod
	T INTEGER
	A RW
	R 0/255

4	O .1.2.840.10036.1.1.1.5
	L dot11CFPMaxDuration
	T INTEGER
	A RW
	R 0/65535

5	O .1.2.840.10036.1.1.1.6
	L dot11AuthenticationResponseTimeOut
	T UNSIGNED32
	A RW
	R 1/4294967295

6	O .1.2.840.10036.1.1.1.7
	L dot11PrivacyOptionImplemented
	T INTEGER
	A R
	R 1,2

7	O .1.2.840.10036.1.1.1.8
	L dot11PowerManagementMode
	T INTEGER
	A RW
	R 1,2

8	O .1.2.840.10036.1.1.1.9
	L dot11DesiredSSID
	T OCTETSTR
	A RW
	R 0/32

9	O .1.2.840.10036.1.1.1.10
	L dot11DesiredBSSType
	T INTEGER
	A RW
	R 1,2,3

10	O .1.2.840.10036.1.1.1.11
	L dot11OperationalRateSet
	T OCTETSTR
	A RW
	R 1/126

11	O .1.2.840.10036.1.1.1.12
	L dot11BeaconPeriod
	T INTEGER
	A RW
	R 1/65535

12	O .1.2.840.10036.1.1.1.13
	L dot11DTIMPeriod
	T INTEGER
	A RW
	R 1/255

13	O .1.2.840.10036.1.1.1.14
	L dot11AssociationResponseTimeOut
	T UNSIGNED32
	A RW
	R 1/4294967295

14	O .1.2.840.10036.1.1.1.15
	L dot11DisassociateReason
	T INTEGER
	A R
	R 0/65535

15	O .1.2.840.10036.1.1.1.16
	L dot11DisassociateStation
	T OCTETSTR
	A R
	R 6

16	O .1.2.840.10036.1.1.1.17
	L dot11DeauthenticateReason
	T INTEGER
	A R
	R 0/65535

17	O .1.2.840.10036.1.1.1.18
	L dot11DeauthenticateStation
	T OCTETSTR
	A R
	R 6

18	O .1.2.840.10036.1.1.1.19
	L dot11AuthenticateFailStatus
	T INTEGER
	A R
	R 0/65535

19	O .1.2.840.10036.1.1.1.20
	L dot11AuthenticateFailStation
	T OCTETSTR
	A R
	R 6

20	O .1.2.840.10036.1.1.1.21
	L dot11MultiDomainCapabilityImplemented
	T INTEGER
	A RW
	R 1,2

21	O .1.2.840.10036.1.1.1.22
	L dot11MultiDomainCapabilityEnabled
	T INTEGER
	A RW
	R 1,2

22	O .1.2.840.10036.1.1.1.23
	L dot11CountryString
	T OCTETSTR
	A R
	R 3
*/
