/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "schema.h"
#include "objdata.h"
#include "object.h"
#include "table.h"

#include "cfg_file_lib.h"

static int init(void)
{
	struct object_t* obj;
	
	/* reset Event control to Active. Normaly it will always be 
	   active, but it could happen that the state 'resetLog' was
	   written to the configfile */
	obj = object_get(".1.3.6.1.4.1.937.2.3.3.1.0");
	assert(obj);
	object_set_int(obj, 2);	
	return 0;
}

static int set(void)
{
	struct object_t* obj;
	struct schent_t* sch;
	struct row_t* row, *nextrow;
	
	obj = object_get(".1.3.6.1.4.1.937.2.3.3.1.0");
	assert(obj);
	
	/* EventControl */
	if (object_get_int(obj) == 1) {
		/* reset eventlog */
		object_set_int(obj, 2);
		
		system("/etc/init.d/aplogd restart");
		
		/* Remove all rows from the datastore */
                sch = schema_get(".1.3.6.1.4.1.937.2.3.3.4.1");
		assert(sch);
		row = sch->rowlist;
		
		while(row) {
			nextrow = row->next;
			row_delete(row);
			row = nextrow;
		}
	}
	
	write_config_aplog();
	
	return 0;
}

static int check(void) {
	
	struct object_t* obj;
	unsigned int max_length;
	
	/* check "our" range for MaxTableLength */
	obj = object_get(".1.3.6.1.4.1.937.2.3.3.3.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	
	max_length = object_get_uint(obj);
	
	if(max_length < 10 || max_length > 250) {
		return R_CHECKFAILED;
	}
	
	return R_NOERR;
	
}

struct moddescr_t mod_eventcontrol = {

	name:		__FILE__,
        fn_ini:         init,			
	fn_set:		set,
	fn_chk:		check,
	oidlist:	".1.3.6.1.4.1.937.2.3.3.1 .1.3.6.1.4.1.937.2.3.3.2.1 .1.3.6.1.4.1.937.2.3.3.3"

};


/*

O .1.3.6.1.4.1.937.2.3.3.1
L eventControl
T INTEGER
A RW
R 1,2

O .1.3.6.1.4.1.937.2.3.3.2.1
L eventControlEntry
T TABLE
I .1.3.6.1.4.1.937.2.3.3.2.1.1

	O .1.3.6.1.4.1.937.2.3.3.2.1.1
	L eventPriority
	T INTEGER
	A -
	R 1,2,3,4,5,6,7,8

	O .1.3.6.1.4.1.937.2.3.3.2.1.2
	L eventReporting
	T BITS
	A RW
	R 0,1,2



*/
