/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <aplog_defs.h>


#include "schema.h"
#include "objdata.h"
#include "table.h"

static int get(void)
{	
	struct row_t *row;
	struct row_t *nextrow;
	struct schent_t *sch;

	static FILE *logfd;
	static struct aplog_header_t loghdr;
	static struct aplog_entry_t logent;
	int first, last;
	int rownum;
	int slotnum;
	int lastrownuminstore;
	long fptr;

	struct tm *timestamp;
	int year;
	char t1[8], t2[8];

	struct index_t *idx;

	logfd = fopen(APLOGFILE, "r");

	fread(&loghdr, sizeof(loghdr), 1, logfd);

	last  = loghdr.current_number;
	first = last - loghdr.num_slots + 1;
	if(first<1) first=1;
	

	sch = schema_get(".1.3.6.1.4.1.937.2.3.3.4.1");
	assert(sch);
	row = sch->rowlist;


	// First delete rows that are no longer in the event log

	lastrownuminstore=0;

	while(row) {
		nextrow = row->next;
	
		rownum = object_get_uint(row_get_obj(row, 0));
		if(rownum < first) row_delete(row);

		if(rownum > lastrownuminstore) lastrownuminstore = rownum;

		row = nextrow;
		
	}

	if(lastrownuminstore < first) lastrownuminstore = first-1;


	// And now add the new rows if there are...

	for(rownum=lastrownuminstore+1; rownum<=last; rownum++) {

		slotnum = (rownum % loghdr.num_slots);

		fptr = sizeof(loghdr) + slotnum * sizeof(logent);
	
		fseek(logfd, fptr, 0);
		fread(&logent, sizeof(logent), 1, logfd);
	
		if(logent.firsttime != 0) {
			timestamp = localtime(&logent.firsttime);
		        year = timestamp->tm_year + 1900;
		        t1[0] = (u_char) ((year & 0xFF00)>>8);
			t1[1] = (u_char) (year & 0x00FF);
			t1[2] = (u_char) timestamp->tm_mon + 1;
			t1[3] = (u_char) timestamp->tm_mday;
			t1[4] = (u_char) timestamp->tm_hour;
			t1[5] = (u_char) timestamp->tm_min;
			t1[6] = (u_char) timestamp->tm_sec;
			t1[7] = 0; /* we don't have miliseconds */

			timestamp = localtime(&logent.lasttime);
		        year = timestamp->tm_year + 1900;
		        t2[0] = (u_char) ((year & 0xFF00)>>8);
			t2[1] = (u_char) (year & 0x00FF);
			t2[2] = (u_char) timestamp->tm_mon + 1;
			t2[3] = (u_char) timestamp->tm_mday;
			t2[4] = (u_char) timestamp->tm_hour;
			t2[5] = (u_char) timestamp->tm_min;
			t2[6] = (u_char) timestamp->tm_sec;
			t2[7] = 0; /* we don't have miliseconds */

			// Create the new row and fill in it's data
	
			idx = index_new(sch);
			assert(idx);
	
			object_set_uint(index_get_obj(idx, 0), rownum);
	
			row = row_create_by_index(idx);
			assert(row);
	
			object_set_uint(   row_get_obj(row, 0), rownum);
			object_set_uint(   row_get_obj(row, 1), logent.count);
			object_set_string( row_get_obj(row, 2), t1, 8);
			object_set_string( row_get_obj(row, 3), t2, 8);
			object_set_int(	   row_get_obj(row, 4), (logent.eventid & 0xFF000000) >> 24);
			object_set_int(    row_get_obj(row, 5), (logent.eventid & 0x00FF0000) >> 16);
			object_set_uint(   row_get_obj(row, 6), (logent.eventid & 0x0000FFFF));
			/* check for "\n" */
			if(logent.msg[strlen(logent.msg) - 1] == 0x0D) {
				object_set_string( row_get_obj(row, 7), logent.msg, strlen(logent.msg) - 1);
			}
			else {
				object_set_string( row_get_obj(row, 7), logent.msg, strlen(logent.msg));
			}

	
			index_done(idx);

		}

	}

	fclose(logfd);

	return(3);
}

struct moddescr_t mod_eventlog = {

	name:		__FILE__,
	fn_get:		get,
	oidlist:	".1.3.6.1.4.1.937.2.3.3.4.1"
};


/*

O .1.3.6.1.4.1.937.2.3.3.4.1
L eventReportEntry
T TABLE
I .1.3.6.1.4.1.937.2.3.3.4.1.1

0	O .1.3.6.1.4.1.937.2.3.3.4.1.1
	L eventReportIndex
	T UNSIGNED32
	A -
	R 1/2147483647

1	O .1.3.6.1.4.1.937.2.3.3.4.1.2
	L eventReportCount
	T UNSIGNED32
	A R

2	O .1.3.6.1.4.1.937.2.3.3.4.1.3
	L eventReportFirstTime
	T OCTETSTR
	A R
	R 8,11

3	O .1.3.6.1.4.1.937.2.3.3.4.1.4
	L eventReportLastTime
	T OCTETSTR
	A R
	R 8,11

4	O .1.3.6.1.4.1.937.2.3.3.4.1.5
	L eventReportLevel
	T INTEGER
	A R
	R 1,2,3,4,5,6,7,8

5	O .1.3.6.1.4.1.937.2.3.3.4.1.6
	L eventReportFacility
	T INTEGER
	A R
	R 1,2,5,10,14,15,17,18,19,20

6	O .1.3.6.1.4.1.937.2.3.3.4.1.7
	L eventReportId
	T UNSIGNED32
	A R
	R 0/65535

7	O .1.3.6.1.4.1.937.2.3.3.4.1.8
	L eventReportText
	T OCTETSTR
	A R
	R 0/255

*/
