/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <linux/route.h>
#include <aplog.h>
#include "misc.h"
#include "table.h"
#include "schema.h"
#include "objdata.h"
#include "mod_mibII_interfaces.h"


static char buf[160];

static int get(void) {
	struct schent_t *sch;
	struct row_t *row, *nextrow;
	struct index_t *idx;
	FILE* fp;
	char *part[11];
	int parts;
	struct in_addr addr;
	struct object_t* obj;
	unsigned int flags;
	int type;
	
	/* first grep the table and delete it */
	sch = schema_get(".1.3.6.1.2.1.4.24.4.1");
	assert(sch);
	row = sch->rowlist;
	while(row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}
	
	fp = fopen("/proc/net/route", "r");
	assert(fp);	
	
	/* read header line */
	if (fgets(buf, sizeof(buf), fp) == 0){
		aplog(EV_SNMP_GENERAL_ERROR, "Error reading /proc/net/route",NULL);
		return(0);
	}
	
	while(fgets(buf, sizeof(buf), fp)) {
		
		/* chop */
		buf[strlen(buf)-1] = '\0';
		
		parts = splitws(buf, part, 11);
		
		idx = index_new(sch);
		assert(idx);
	
		/* ipCidrrouteDest */
		obj = index_get_obj(idx, 0);
		assert(obj);
		sscanf(part[1], "%8X", &addr.s_addr);
		object_set_ipaddr(obj, addr);
				
		/* ipCidrRouteMask */
		obj = index_get_obj(idx, 1);
		assert(obj);
		sscanf(part[7], "%8X", &addr.s_addr);
		object_set_ipaddr(obj, addr);
		
		/* ipCidrRouteTos */
		obj = index_get_obj(idx, 2);
		/* TOS is 0000 (normal operation) */
		object_set_int(obj, 0);
		
		/* ipCidrRouteNextHop */
		obj = index_get_obj(idx, 3);
		assert(obj);
		sscanf(part[2], "%8X", &addr.s_addr);
		object_set_ipaddr(obj, addr);
		
		/* check for duplicate */
		row = row_get_by_index(idx);
		if(row) {
			/* duplicate line, free index and continue to next line */
			index_done(idx);
			continue;
		}
		
		row = row_create_by_index(idx);
		assert(row);
		
		/* ipCidrRouteIfIndex */
		obj = row_get_obj(row, 4);
		assert(obj);
		if (strcmp(part[0], "*") == 0) {
			object_set_int(obj, 0); /* no interface */			
		}
		else {
			object_set_int(obj, ifname2ifindex(part[0]));
		}
		
		/* ipCidrRouteType */
		sscanf(part[3], "%4X", &flags);

		type = 1; /*other */
		if (flags & RTF_HOST) {
			type = 3; /* local */
		}
		if (flags & RTF_GATEWAY) {
			type = 4; /* remote */
		}
		if (flags & RTF_REJECT) {
			type = 2; /* reject */
		}
		obj = row_get_obj(row, 5);
		assert(obj);
		object_set_int(obj, type);
		
		/* ipCidrRouteMetric1 */
		obj = row_get_obj(row,10);
		assert(obj);
		object_set_int(obj, atoi(part[6]));
		
		/* The following objects are not implemented 
		 * The information is not available 
		 * Only OID's with NO default value need attention */
		
		/* ipCidrRouteProto */
		obj = row_get_obj(row, 6);
		assert(obj);
		object_set_int(obj, 1); /* other */

		/* ipCidrRouteInfo */
		obj = row_get_obj(row, 8);
		assert(obj);
		object_set_objectid(obj, "0.0");
		
		/* ipCidrRouteStatus */
		obj = row_get_obj(row, 15);
		assert(obj);
		object_set_rowstat(obj, 1); /* active */
		
	}	
		
	
	return(3); /* cache is valid for 3 seconds */
	
}

struct moddescr_t mod_ipCidrRoute = {

	name:		__FILE__,
	fn_get:		get,
	oidlist:	".1.3.6.1.2.1.4.24.3 .1.3.6.1.2.1.4.24.4.1"

};


/*

O .1.3.6.1.2.1.4.24.3
L ipCidrRouteNumber
T GAUGE
A R
R 0/4294967295

O .1.3.6.1.2.1.4.24.4.1
L ipCidrRouteEntry
T TABLE
I .1.3.6.1.2.1.4.24.4.1.1 .1.3.6.1.2.1.4.24.4.1.2 .1.3.6.1.2.1.4.24.4.1.3 .1.3.6.1.2.1.4.24.4.1.4

	O .1.3.6.1.2.1.4.24.4.1.1
	L ipCidrRouteDest
	T IPADDR
	A R
	R 4

	O .1.3.6.1.2.1.4.24.4.1.2
	L ipCidrRouteMask
	T IPADDR
	A R
	R 4

	O .1.3.6.1.2.1.4.24.4.1.3
	L ipCidrRouteTos
	T INTEGER
	A R

	O .1.3.6.1.2.1.4.24.4.1.4
	L ipCidrRouteNextHop
	T IPADDR
	A R
	R 4

	O .1.3.6.1.2.1.4.24.4.1.5
	L ipCidrRouteIfIndex
	T INTEGER
	A C
	D 0

	O .1.3.6.1.2.1.4.24.4.1.6
	L ipCidrRouteType
	T INTEGER
	A C
	R 1,2,3,4

	O .1.3.6.1.2.1.4.24.4.1.7
	L ipCidrRouteProto
	T INTEGER
	A R
	R 1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16

	O .1.3.6.1.2.1.4.24.4.1.8
	L ipCidrRouteAge
	T INTEGER
	A R
	D 0

	O .1.3.6.1.2.1.4.24.4.1.9
	L ipCidrRouteInfo
	T OBJECTID
	A C

	O .1.3.6.1.2.1.4.24.4.1.10
	L ipCidrRouteNextHopAS
	T INTEGER
	A C
	D 0

	O .1.3.6.1.2.1.4.24.4.1.11
	L ipCidrRouteMetric1
	T INTEGER
	A C
	D -1

	O .1.3.6.1.2.1.4.24.4.1.12
	L ipCidrRouteMetric2
	T INTEGER
	A C
	D -1

	O .1.3.6.1.2.1.4.24.4.1.13
	L ipCidrRouteMetric3
	T INTEGER
	A C
	D -1

	O .1.3.6.1.2.1.4.24.4.1.14
	L ipCidrRouteMetric4
	T INTEGER
	A C
	D -1

	O .1.3.6.1.2.1.4.24.4.1.15
	L ipCidrRouteMetric5
	T INTEGER
	A C
	D -1

	O .1.3.6.1.2.1.4.24.4.1.16
	L ipCidrRouteStatus
	T ROWSTAT
	A C
	R 1,2,3,4,5,6



*/
