/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <aplog.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>

#include "table.h"
#include "objdata.h"
#include "schema.h"
#include "misc.h"
#include "objflag.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"
#include "regulatory.h"
#include "debug.h"

static int get(void)
{
	struct row_t *row, *nextrow;
	struct schent_t *sch;
	struct obj_bss *bss;
	struct index_t *idx;
	struct object_t *obj;
	unsigned char capinfo;
	int wlan_if[2] = {3, 0}; /* FIX ME */
	char *ifname, ssid[34];
	unsigned char rates[12] = {0x02, 0x04, 0x0b, 0x16, 0xc, 0x12,  
				   0x18, 0x24, 0x30, 0x48, 0x60, 0x6c};
	unsigned char sup_rates[16];
	int i, j=0, interface;
	long max_bss_entries, length, nr_bss, channel;
	long *data = NULL;
	
    	sch = schema_get(".1.3.6.1.4.1.937.2.2.3.11.1");
	assert(sch);

	row = sch->rowlist;

	/* erase existing lines */
	while (row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}
	
	for(interface=0; wlan_if[interface] != 0; interface++)
	{
	    ifname = ifindex2ifname(wlan_if[interface]);
	    prism_set_ifname(ifname);
	
	    //printf("bssTable: neighbouring bss's for interface %s\n", ifname);
	    
	    max_bss_entries = prism_get_long(DOT11_OID_BSSS);

	    length = max_bss_entries * sizeof(struct obj_bss) + 4;

	    data = (long*)malloc(length);
	    
	    if( get_prismoid( ifname, DOT11_OID_BSSLIST, data, length) != 0 )
	    {
		dbg(DBG_MODULE, "get prism oid for DOT11_OID_BSSLIST faild!\n");
		return 10;
	    }

	    nr_bss = *data; /* number of found BSS's */

	    //printf("bssTable: there are %d neighbouring bss's found\n", nr_bss);

	    bss = (struct obj_bss*)(data+1);

	    /* loop until 1. Entry 0 is the AP itself */
	    while(--nr_bss >= 0)
	    {
		//printf("bssTable: bss nr: %d\n", nr_bss);
		
		idx = index_new(sch);
		assert(idx);
		obj = index_get_obj(idx, 0);	
		assert(obj);
		object_set_int(obj, wlan_if[interface]);
		obj = index_get_obj(idx, 1);	
		assert(obj);
		object_set_int(obj,nr_bss+1);
		
		row = row_create_by_index(idx);
		assert(row);
		index_done(idx);
		
		object_set_string(row_get_obj(row, 1), bss[nr_bss].address, 6);
		/* The SSID is prefixed with the length of the SSID. Should be struct obj_ssid?! */
		object_set_string(row_get_obj(row, 2), bss[nr_bss].ssid+1, bss[nr_bss].ssid[0]);
		
		/* Convert channel frequency to channel number */
		channel = regulatory_freq2channel(bss[nr_bss].channel);
		
		object_set_int(row_get_obj(row, 3), channel);
		capinfo = bss[nr_bss].capinfo;
		for(i=0; i<=8; i++)
		{
		    object_set_bit(row_get_obj(row, 4), i, (capinfo>>i & 0x01));
		}
		
		/* Convert rate bitmap to per rate octes */ 
		j=0;	
		for(i=0; i<12; i++)
		{
		    if((bss[nr_bss].rates >> i) & 0x0001)
			sup_rates[j] = rates[i];
		    else
			continue;
			    
		    sup_rates[j] |= bss[nr_bss].basic_rates >> i ? 0x80 : 0x00;
		    j++;
		        
		}
		object_set_string(row_get_obj(row, 5), sup_rates, j); 
		object_set_gauge(row_get_obj(row, 7), bss[nr_bss].age);
		object_set_int(row_get_obj(row, 8), bss[nr_bss].rssi);

	    }

	    free(data);
	}

	return 10;
}


struct moddescr_t mod_isl_dot11ext2_bssTable= {

	name:		__FILE__,
	fn_get:		get, 
	oidlist:	".1.3.6.1.4.1.937.2.2.3.11.1"

};

 
/*
 
O .1.3.6.1.4.1.937.2.2.3.11.1
L bssEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1 .1.3.6.1.4.1.937.2.2.3.11.1.1

	O .1.3.6.1.4.1.937.2.2.3.11.1.1
	L bssIndex
	T INTEGER
	A -
	R 1/2147483647

	O .1.3.6.1.4.1.937.2.2.3.11.1.2
	L bssBSSID
	T OCTETSTR
	A R
	R 6

	O .1.3.6.1.4.1.937.2.2.3.11.1.3
	L bssSSID
	T OCTETSTR
	A R
	R 0/32

	O .1.3.6.1.4.1.937.2.2.3.11.1.4
	L bssChannelNumber
	T INTEGER
	A R

	O .1.3.6.1.4.1.937.2.2.3.11.1.5
	L bssCapability
	T BITS
	A R
	R 0,1,2,3,4,5,6,7

	O .1.3.6.1.4.1.937.2.2.3.11.1.6
	L bssSupportedRates
	T OCTETSTR
	A R
	R 0/255

	O .1.3.6.1.4.1.937.2.2.3.11.1.7
	L bssCountryCode
	T OCTETSTR
	A R
	R 3

	O .1.3.6.1.4.1.937.2.2.3.11.1.8
	L bssAge
	T GAUGE
	A R
	R 0/4294967295

	O .1.3.6.1.4.1.937.2.2.3.11.1.10
	L bssRSSI
	T INTEGER
	A R

*/
