/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <aplog.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>

#include "table.h"
#include "objdata.h"
#include "schema.h"
#include "misc.h"
#include "objflag.h"
#include "easyprism.h"
#include "mod_mibII_interfaces.h"
#include "debug.h"
#include "object.h"

//#define DEBUG

#define MIN(X, Y)	( ((X) < (Y)) ? (X) : (Y) )

static const int wlan_if[] = { 

    3, 		// 2.4 GHz
    5, 		// 5.0 GHz - Uncomment when Poldhu is no longer used
    0		// End of the line
};


/*
 * MVC client state -> MIB client state translation table
 * MIB states 2 and 4 are not allowed, apparantly. 
 */
static const int states[] = {

	1, 	// Disconnected
	1, 	// Authenticating
	3, 	// Authenticated
	3, 	// Associating
	3, 	// Probably reassociating -> Associating (?)
	8, 	// Associated
	6, 	// IBSS -> ulAuthenticated (?)
	7,	// WDS
};

static int get( void )
{
    struct row_t *row, *nextrow;
    struct schent_t *sch;
    struct index_t *idx;
    struct object_t *obj;
    unsigned int max_clients_ds, max_clients_mvc, max_clients;
    int ife;

    // Retrieve maximum number of clients allowed in clientTable

    obj = object_get(".1.3.6.1.4.1.937.2.2.3.1.0");
    assert(obj);
    max_clients_ds = object_get_uint(obj);

#ifdef DEBUG
    fprintf(stderr, "Maximum number of clients allowed in clientTable: %d\n", max_clients);
#endif

    sch = schema_get(".1.3.6.1.4.1.937.2.2.3.2.1");
    assert(sch);

    // Erase existing rows

    row = sch->rowlist;
    while (row)
    {
	nextrow = row->next;
	row_delete(row);
	row = nextrow;
    }
    ife = 0;
    while (wlan_if[ife] != 0)
    {
        char *ifname;
        int cur;
        struct obj_sta sta;
	int no_found;

        // Open Easyprism interface

        ifname = ifindex2ifname(wlan_if[ife]);
	prism_set_ifname(ifname);

        max_clients_mvc = prism_get_long(DOT11_OID_CLIENTS);

    	max_clients = MIN(max_clients_ds, max_clients_mvc);

#ifdef DEBUG
	fprintf(stderr, "Number of clients limited to %d\n", max_clients);
#endif
        cur = 0;
	no_found = 0;
        while (cur < max_clients)
        {
	    int client = DOT11_OID_CLIENT1 + cur;

	    if (get_prismoid(ifname, client, &sta, sizeof(sta)) == 0)
	    {
#ifdef DEBUG
		fprintf(stderr, "MVC returned sta %02X:%02X:%02X:%02X:%02X:%02X\n", 
				sta.address[0], sta.address[1],
				sta.address[2], sta.address[3],
				sta.address[4], sta.address[5]);
		fprintf(stderr, "sta.node = %i\n", sta.node);
#endif
		    
		    
		// Only create if this node is a client
		if (sta.node == DOT11_NODE_CLIENT ) 
		{
		    // Now create new row in the table

		    no_found++;
		    idx = index_new(sch);
		    assert(idx);

		    // Double index - first: interface...

		    obj = index_get_obj(idx, 0);
		    assert(obj);
		    object_set_int(obj, wlan_if[ife]);

		    // ...second: client

		    obj = index_get_obj(idx, 1);
		    assert(obj);
		    object_set_int(obj, no_found);

		    row = row_create_by_index(idx);
		    assert(row);
		    index_done(idx);

		    object_set_string(row_get_obj(row, 1), sta.address, 6);
		    object_set_int(row_get_obj(row, 2), sta.age);
		    object_set_int(row_get_obj(row, 3), sta.rssi);
		    object_set_int(row_get_obj(row, 4), sta.rate);
		    object_set_int(row_get_obj(row, 5), states[(int) sta.state]);
		}
	    }
            ++cur;
        }

    	// Next interface

        ++ife;	
    }

    return (5);
}



struct moddescr_t mod_isl_dot11ext2_clientTable = {

    name:	__FILE__,
    fn_get:	get, 
    oidlist:	".1.3.6.1.4.1.937.2.2.3.2.1"
};


/*

The following schema entries are provided by this module :

O .1.3.6.1.4.1.937.2.2.3.2.1
L clientEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1 .1.3.6.1.4.1.937.2.2.3.2.1.1

0	O .1.3.6.1.4.1.937.2.2.3.2.1.1
	L clientIndex
	T INTEGER
	A -
	R 1/2007

1	O .1.3.6.1.4.1.937.2.2.3.2.1.2
	L clientAddress
	T OCTETSTR
	A R
	R 6


E	O .1.3.6.1.4.1.937.2.2.3.2.1.3
	L clientQuality
	T INTEGER
	A R

2	O .1.3.6.1.4.1.937.2.2.3.2.1.4
	L clientAge
	T INTEGER
	A R

3	O .1.3.6.1.4.1.937.2.2.3.2.1.5
	L clientRSSI
	T INTEGER
	A R

4	O .1.3.6.1.4.1.937.2.2.3.2.1.6
	L clientRate
	T INTEGER
	A R

5	O .1.3.6.1.4.1.937.2.2.3.2.1.7
	L clientState
	T INTEGER
	A R
	R 1,2,3,4,5,6,7,8,9


*/
