/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
 
#include <assert.h>

#include "misc.h"
#include "object.h"
#include "objdata.h"
#include "objflag.h"
#include "handler.h"

static void startstop_iappd(struct object_t *obj) {

	int iapp_enabled;
	char* cmd[3];
	
	cmd[0] = "/etc/init.d/iappd";
	cmd[2] = NULL;
	
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	
	iapp_enabled = object_get_int(obj);
	
	if (iapp_enabled == 1) {
		cmd[1] = "restart";
	}
	else {
		cmd[1] = "stop";
	}
	
	exec(cmd);
}

static int radius_updated(void) {
	struct object_t* obj;
	
	/* Radius servers where updated. Restart iapp */
	obj = object_get(".1.3.6.1.4.1.937.2.2.3.12.1.0");
	assert(obj);

	startstop_iappd(obj);
	
	return R_NOERR;
}

static int init(void) {
	struct handler_t* hnd;
	
	/* register an update handler on the radiusservers OID */
	hnd = handler_add(H_SET, radius_updated);
	handler_connect(hnd, ".1.3.6.1.4.1.937.2.1.4.2.1");
	
	return R_NOERR;
}

static int set(void)
{
	struct object_t *obj;
	
		
	obj = object_get(".1.3.6.1.4.1.937.2.2.3.12.1.0");
	assert(obj);
	
	if(object_flag_isset(obj, OF_DIRTY)) 
	{
		startstop_iappd(obj);		
		object_flag_clear(obj, OF_DIRTY);
	}
	
	return R_NOERR;

}


struct moddescr_t mod_isl_dot11ext2_iapp = {

	name:		__FILE__,
	fn_set:		set,
	fn_ini:		init,
	oidlist:	".1.3.6.1.4.1.937.2.2.3.12.1 .1.3.6.1.4.1.937.2.2.3.12.2 .1.3.6.1.4.1.937.2.2.3.12.3"

};


/*

O .1.3.6.1.4.1.937.2.2.3.12.1
L iappEnabled
T INTEGER
A RW
R 1,2

O .1.3.6.1.4.1.937.2.2.3.12.2
L iappAddTimeout
T UNSIGNED32
A RW
R 1/4294967295

O .1.3.6.1.4.1.937.2.2.3.12.3
L iappMoveTimeout
T UNSIGNED32
A RW
R 1/4294967295

*/
