/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <assert.h>

#include "table.h"
#include "misc.h"
#include "object.h"
#include "objdata.h"

#include "pda.h"
#include "identification/interfaces.h"

#include "mod_mibII_interfaces.h"

#define PHYTYPE_OFDM (0x01 << 0)
#define PHYTYPE_ERP  (0x01 << 1)


static unsigned int get_phytype_from_PDR(struct Pdr_Struct* pdrPtr) {
	
	u_char* p; 
	int i;
	unsigned short temp;
	unsigned int phytype = 0;
	
	p = &(pdrPtr->PdrData[0]);
	
	for(i=0; i < (pdrPtr->PdrLength/5); i++) {
		p = p + 2; /* skip role */
		
		temp = (*p);
		p++;
		temp |= (*p << 8);
		if(temp == BR_IF_ID_ISL39300A) {
			/* found the HW interface */
			p++; /* goto variant */
			
			temp = (*p);
			p++;
			temp |= (*p << 8);
			
			if((temp & BR_IF_VAR_ISL39000_24_GHZ_MASK) != BR_IF_VAR_ISL39000_24_GHZ_DISABLED) {
				phytype |= PHYTYPE_ERP;
			}
			
			if((temp & BR_IF_VAR_ISL39000_5_GHZ_MASK) != BR_IF_VAR_ISL39000_5_GHZ_DISABLED) {
				phytype |= PHYTYPE_OFDM;
			}
			
			return (phytype);
		}
		else {
			p = p + 7; /* goto next interface */
		}
	}
	
	fprintf(stderr, "No HW interface for the ISL39300A found in PDR INTERFACE_LIST !\n");
	return(phytype);
}


static unsigned int get_phytype(int ifnum) {
	struct Pdr_Struct* pdr = NULL;
	
	switch(ifnum) {
		
		case 3: /* internal interface */
			pdr = pda_Find_Pdr(PDR_INTERFACE_LIST, NULL);
			break;
		case 5:
			/* TODO */
			break;
		default:
			break;
	}
	
	if(pdr) {
		return (get_phytype_from_PDR(pdr));
	}
	else {
		fprintf(stderr, "PDR_INTERFACE_LIST not present for interface %i, assuming ERP & OFDM\n", ifnum);
		return (PHYTYPE_OFDM | PHYTYPE_ERP);
	}
}


static int init(void) {
	struct schent_t* sch, *sc_sch;
	struct row_t* row, *sc_row;
	struct object_t* obj, *if_obj;
	struct index_t* idx;
	int ifnum, index;
	unsigned int phytype;
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.5.1");
	assert(sch);
	
	/* create rows in table. Supported phyType for second radio is
	 * depending on type of mini-pci card */
	
	/* Create rows, depending on the number of rows in the
	   dot11StationConfigTable */
	sc_sch = schema_get(".1.2.840.10036.1.1.1"); /* dot11 station config table */
	assert(sc_sch);
	
	/* loop all entries in station config table */
	sc_row = sc_sch->rowlist;
	while(sc_row) {
		
		index = 1;
		/* get the interface number of this row */
		idx = index_get(sc_row);
		assert(idx);
		if_obj = index_get_obj(idx, 0);
		assert(if_obj);
		assert(if_obj->flags & OF_DATAPRESENT);
		ifnum = object_get_int(if_obj);
		index_done(idx);
	
		if(!interface_present(ifnum)) {
			goto next;
		}
			
		phytype = get_phytype(ifnum);
		
		if(phytype & PHYTYPE_OFDM) {
			idx = index_new(sch);
			assert(idx);			
			
			obj = index_get_obj(idx, 0);
			assert(obj);
			object_set_int(obj, ifnum);
			
			obj = index_get_obj(idx, 1);
			assert(obj);
			object_set_int(obj, index);
		
			row = row_create_by_index(idx);
			assert(row);
			
			index_done(idx);
			
			obj = row_get_obj(row, 0);
			assert(obj);
			object_set_int(obj, index); 
			
			obj = row_get_obj(row, 1);
			assert(obj);
			object_set_int(obj, 4); /* ODFM */
			
			index++;
		}
		
		
		if(phytype & PHYTYPE_ERP) {
			idx = index_new(sch);
			assert(idx);			
			
			obj = index_get_obj(idx, 0);
			assert(obj);
			object_set_int(obj, ifnum);
			
			obj = index_get_obj(idx, 1);
			assert(obj);
			object_set_int(obj, index);
			
			row = row_create_by_index(idx);
			assert(row);
			
			index_done(idx);
			
			obj = row_get_obj(row, 0);
			assert(obj);
			object_set_int(obj, index); 
			
			obj = row_get_obj(row, 1);
			assert(obj);
			object_set_int(obj, 6); /* ERP */

			index++;
		}
next:		
		sc_row = sc_row->next;
		
	}
		


	return R_NOERR;
}

struct moddescr_t mod_dot11ext2_phySupportTable = {
	name:		__FILE__,
	fn_ini:		init,
	oidlist:	".1.3.6.1.4.1.937.2.2.2.5.1"
};


/*

The following schema entries are provided by this module :

O .1.3.6.1.4.1.937.2.2.2.5.1
L phySupportEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1 .1.3.6.1.4.1.937.2.2.2.5.1.1

	O .1.3.6.1.4.1.937.2.2.2.5.1.1
	L phySupportIndex
	T INTEGER
	A -
	R 1/2147483647

	O .1.3.6.1.4.1.937.2.2.2.5.1.2
	L phySupportType
	T INTEGER
	A R
	R 1,2,3,4,5,6

*/
