/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "pda.h"

#include "table.h"
#include "misc.h"
#include "object.h"
#include "objdata.h"
#include "misc.h"
#include "regulatory.h"
#include "prismoids.h"

// #define DEBUG 

/* list of rows */
struct rowlist_t {
	struct row_t* row;
	struct rowlist_t* next;
};

/* tmp struct to hold the Regulatory Group <-> ChannelSet relation */
struct rg_cs_relation_t {
	struct regulatory_group_t *rg;
	struct rowlist_t* rowlist;
	struct rg_cs_relation_t* next;
};

static struct rg_cs_relation_t* rgcs_list = NULL;

static struct rg_cs_relation_t* find_relation(struct regulatory_group_t* rg) {
	struct rg_cs_relation_t* rgcs = NULL;
	
	rgcs = rgcs_list;
	
	while(rgcs) {
		if(rg == rgcs->rg) {
			return (rgcs);
		}
		rgcs = rgcs->next;
	}
	
	return NULL;
		
}

static struct rg_cs_relation_t* relation_create(struct regulatory_group_t* rg) {
	struct rg_cs_relation_t* new_rgcs = NULL;
	
	assert( find_relation(rg) == 0);
	
	new_rgcs = (struct rg_cs_relation_t*) calloc(sizeof(struct rg_cs_relation_t), 1);
	
	new_rgcs->rg = rg;
	new_rgcs->rowlist = NULL;

	/* Store it in the linked list */
	new_rgcs->next = rgcs_list;
	rgcs_list = new_rgcs;
	
	return (new_rgcs);
}

static void relation_create_row(struct schent_t* sch, struct rg_cs_relation_t* rgcs, int start_freq, int no_freqs, int freq_gap) {
	
	int start_channel, channel_gap, tmp_channel;
	struct index_t* idx;
	struct row_t* row;
	struct object_t* obj;
	static int index = 0;
	struct rowlist_t* nw_rowlist;
	
	start_channel = regulatory_freq2channel(start_freq);
	tmp_channel = regulatory_freq2channel(start_freq + freq_gap);
	channel_gap = (tmp_channel - start_channel);

#ifdef DEBUG		
	fprintf(stderr, "start_channel:%i, channel_gap:%i, no_channels:%i\n", start_channel, channel_gap, no_freqs);
#endif
	
	assert(sch);
	idx = index_new(sch);
	assert(idx);
	obj = index_get_obj(idx, 0);
	assert(obj);
	object_set_int(obj, 3); /* TODO */
	obj = index_get_obj(idx, 1); 
	index++;
	object_set_int(obj, index);
	
	row = row_create_by_index(idx);
	assert(row);
	index_done(idx);
	
	/* smtChanSetIndex */
	obj = row_get_obj(row, 0);
	assert(obj);
	object_set_int(obj, index);
	
	/* smtChanSetPhyType */
	obj = row_get_obj(row, 1); 
	assert(obj);
	object_set_int(obj, regulatory_freq2band(start_freq) == REGULATORY_BAND_2DOT4 ? 6 : 4);
	
	/* smtChanSetFirstChannelNumber */
	obj = row_get_obj(row, 2);
	assert(obj);
	object_set_int(obj, start_channel);
	
	/* smtChanSetNumberOfChannels */
	obj = row_get_obj(row, 3);
	assert(obj);
	object_set_int(obj, no_freqs);
	
	/* smtChanSetChannelInterval */
	obj = row_get_obj(row, 4);
	assert(obj);
	object_set_int(obj, channel_gap);
	
	/* Store this row */
	nw_rowlist = (struct rowlist_t*) calloc(sizeof(struct rowlist_t), 1);
	nw_rowlist->row = row;
	nw_rowlist->next = rgcs->rowlist;
	rgcs->rowlist = nw_rowlist;
	
	return;	
}

static int find_country_index(char* country_code, int ifnum) {
	struct schent_t *sch;
	struct row_t* row;
	struct object_t* obj;
	char ccode[5];
	struct index_t* idx;
	int ifindex;
	struct string_t* acc;
	
	/* pad code with spaces */
	snprintf(ccode, sizeof(ccode), "%s   ", country_code);
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.6.1");
	assert(sch);
	
	row = sch->rowlist;
	while(row) {
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifindex = object_get_int(obj);
		index_done(idx);
		
		if(ifindex == ifnum) {
			obj = row_get_obj(row, 1);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			
			acc = object_get_string(obj);
			
			if(strncmp(acc->data, ccode, 3) == 0) {
				obj = row_get_obj(row, 0);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				return(object_get_int(obj));
			}
				
		}
		
		row = row->next;
	}
	
	return -1;
}

static void free_temp() {
	struct rg_cs_relation_t* rgcs;
	struct rowlist_t* rowlist, *rowlist_next;
	
	while(rgcs_list) {
		rgcs = rgcs_list;
		rowlist = rgcs->rowlist;
		while(rowlist) {
			rowlist_next = rowlist->next;
			free(rowlist);
			rowlist = rowlist_next;
		}
		rgcs_list = rgcs->next;
		free(rgcs);
	}
}

static int init(void)
{
	struct schent_t *sch;
	struct object_t *obj;
	struct row_t *row, *next_row;
	struct country_domain_t* cd;
	struct regulatory_group_list_t *rgl;
	struct power_group_t *pg;
	int start_freq, freq_count, freq_gap, freq, i, last_freq;
	struct rg_cs_relation_t* rgcs = NULL;
	static char buf[128];

	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.8.1");
	
	/* loop all country_domains to get all regulatory_groups */
	cd = regulatory_get_country_domains();
	while(cd) {
		
		rgl = cd->regul_groups;
		while(rgl) {
			rgcs = find_relation(rgl->regul_grp);
			if(!rgcs) {
				/* New Regulatory Group, process it */
				
				/* First create an new relation struct */
				rgcs = relation_create(rgl->regul_grp);
				
				
				pg = rgl->regul_grp->power_groups;
				while(pg) {
					start_freq = 0;
					freq_gap = 0;
					freq_count = 0;
					last_freq = 0;
					for(i = 0; i < pg->no_freqs; i++) {
						freq = pg->frequencies[i];
#ifdef DEBUG						
						fprintf(stderr, "*******************\n");
						fprintf(stderr, "cd->id         = %s\n", cd->id);
						fprintf(stderr, "start_freq     = %i\n", start_freq);
						fprintf(stderr, "freq_gap       = %i\n", freq_gap);
						fprintf(stderr, "last_freq      = %i\n", last_freq);
						fprintf(stderr, "freq           = %i\n", freq);
						fprintf(stderr, "freq_count     = %i\n", freq_count);
						fprintf(stderr, "*******************\n");
#endif						
						switch(freq_count) {
							case 0:
								start_freq = freq;
								break;
							case 1:
								freq_gap = freq - start_freq;
								break;
							default:	
								/* check gap */
								if((freq - last_freq) != freq_gap) {
									/* start a new group */
									relation_create_row(sch, rgcs, start_freq, freq_count, freq_gap);
									start_freq = freq;								
									freq_count = 0;
									freq_gap = 0;
								}
								break;
						}
						freq_count++;
						last_freq = freq;
					}
					
					if(freq_count > 0) {
						relation_create_row(sch, rgcs, start_freq, freq_count, freq_gap);
					}	
				
					pg = pg->next;
				}
				
			}
			
			/* Store all the counties in this cd in the corresponding rows */
			for(i = 0; i < cd->no_countries; i++) {
				int ci;
				int ifnum = 3; /* TODO */
				struct rowlist_t* rowlist;
				struct string_t *oldlist;
				int oldlength;
				
				ci = find_country_index(cd->countries[i], ifnum);
				
				/* if the country was not found, this country is not in
				 * the PDA -> skip it */
				if (ci != -1) {
								
					/* loop al channelset rows for this regul group */
					rowlist = rgcs->rowlist;
					while(rowlist) {
						row = rowlist->row;
						assert(row);
						obj = row_get_obj(row, 5);
						assert(obj);
						if(obj->flags & OF_DATAPRESENT) {
							oldlist = object_get_string(obj);
							memcpy(buf, oldlist->data, oldlist->len);
							oldlength = oldlist->len;
						}
						else {
							oldlength = 0;
						}
						buf[oldlength] = ci;
						object_set_string(obj, buf, oldlength+1);
						rowlist = rowlist->next;					
					}
				}
			}
			
			rgl = rgl->next;
		}				
		cd = cd->next;
	}
		
	/* Cleanup all the rows that ended up with no connected countries */
	row = sch->rowlist;
	while(row) {
		next_row = row->next;
		
		obj = row_get_obj(row, 5);
		if(!(obj->flags & OF_DATAPRESENT)) {
			/* delete row */
			row_delete(row);
		}
		row = next_row;
	}	
	
	/* Clean up the internal temporary malloced structs */
	free_temp();
	
	return R_NOERR;
}

#define DOT11_OID_COUNTRY               0x1f000002
struct obj_frequency
{
    uint16_t mhz;
    uint16_t reserved;
};

struct obj_country
{
    uint8_t string[4];
    struct obj_frequency frequencies[1];
};


static void store_softmac_row(char* country, int start_freq, int freq_count, int freq_gap, int interface) {

	struct schent_t* sch;
	struct row_t* row, *found_row;
	struct index_t* idx;
	struct object_t *obj;
	int phy_type;
	int start_channel;
	int tmp_channel;
	int channel_gap;
	int ifindex;
	int max_index = 0;
	int current_index;
	int ci;
	char buf[128];
	struct string_t* oldlist;
	int oldlength;
	
	phy_type = (regulatory_freq2band(start_freq) == REGULATORY_BAND_2DOT4) ? 6 : 4;
	
	start_channel = regulatory_freq2channel(start_freq);
	tmp_channel = regulatory_freq2channel(start_freq + freq_gap);
	channel_gap = (tmp_channel - start_channel);

#ifdef DEBUG		
	fprintf(stderr, "store sm row %s, sc:%i, cg:%i, fc:%i, pt:%i\n", country, start_channel, channel_gap, freq_count, phy_type);
#endif	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.8.1");
	assert(sch);
	found_row = NULL;
	row = sch->rowlist;
	/* walk current rows for this interface and try to find a matching row */
	while(row && (!found_row)) {
				
		idx = index_get(row);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifindex = object_get_int(obj);
		
		obj = index_get_obj(idx, 1);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		current_index = object_get_int(obj);
		
		index_done(idx);
		
		/* check if this is the correct interface */
		if(ifindex != interface) {
			goto next;
		}
		
		/* check if this index is the highest one we have seen so far */
		if(current_index > max_index) {
			max_index = current_index;
		}
		
		/* check if this is a row we are looking for */
		if( (object_get_int(row_get_obj(row, 1)) == phy_type)
				&& (object_get_int(row_get_obj(row, 2)) == start_channel)
				&& (object_get_int(row_get_obj(row, 3)) == freq_count)
				&& (object_get_int(row_get_obj(row, 4)) == channel_gap)
				) {
			found_row = row;
		}
next:		
		row = row->next;
	}
	
	/* if row does not exist -> create one */
	if(!found_row) {
		idx = index_new(sch);
		assert(idx);
		obj = index_get_obj(idx, 0);
		assert(obj);
		object_set_int(obj, interface);

		obj = index_get_obj(idx, 1); 
		object_set_int(obj, max_index + 1);
	
		row = row_create_by_index(idx);
		assert(row);
		index_done(idx);
	
		/* smtChanSetIndex */
		obj = row_get_obj(row, 0);
		assert(obj);
		object_set_int(obj, max_index + 1);
	
		/* smtChanSetPhyType */
		obj = row_get_obj(row, 1); 
		assert(obj);
		object_set_int(obj, phy_type);
	
		/* smtChanSetFirstChannelNumber */
		obj = row_get_obj(row, 2);
		assert(obj);
		object_set_int(obj, start_channel);
	
		/* smtChanSetNumberOfChannels */
		obj = row_get_obj(row, 3);
		assert(obj);
		object_set_int(obj, freq_count);
	
		/* smtChanSetChannelInterval */
		obj = row_get_obj(row, 4);
		assert(obj);
		object_set_int(obj, channel_gap);
		
		found_row = row;
		
	}
	assert(found_row);
		
	/* find the corresponding country index */
	ci = find_country_index(country, interface);
	
	/* add this country to the found/created row */
	obj = row_get_obj(found_row, 5);
	assert(obj);
	if(obj->flags & OF_DATAPRESENT) {
		oldlist = object_get_string(obj);
		memcpy(buf, oldlist->data, oldlist->len);
		oldlength = oldlist->len;
	}
	else {
		oldlength = 0;
	}
	buf[oldlength] = ci;
	object_set_string(obj, buf, oldlength+1);
	
	
}

static int post_init(void) {
	
	struct schent_t *sch, *cs_sch;
	struct row_t *cs_row;
	struct index_t* idx;
	struct object_t * obj;
	int ifindex;
	struct string_t* str;
	char buf[1024];
	struct obj_country* country;
	
	cs_sch = schema_get(".1.3.6.1.4.1.937.2.2.2.6.1");
	assert(cs_sch);
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.8.1");
	assert(sch);
	
	cs_row = cs_sch->rowlist;
	while(cs_row) {
		
		/* retrieve ifindex */
		idx = index_get(cs_row);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifindex = object_get_int(obj);
		index_done(idx);
		
		
		if(ifindex == 5) { 
			int start_freq;
			int freq_gap;
			int last_freq;
			int freq_count;
			int freq;
			int i = 0;
			
			/* get the country code */
			str = object_get_string(row_get_obj(cs_row, 1));
			assert(str);
			
			strncpy(buf, str->data, str->len);
			
			get_prismoid("eth3", DOT11_OID_COUNTRY, buf, sizeof(buf));
			
			country = (struct obj_country*) buf;
#ifdef DEBUG			
			fprintf(stderr, "COUNTRY = [%s]\n", country->string);
#endif			
			start_freq = 0;
			freq_gap = 0;
			freq_count = 0;
			last_freq = 0;
			i = 0;
			while(country->frequencies[i].mhz > 0) {
			
				freq = country->frequencies[i].mhz;
#ifdef DEBUG				
				fprintf(stderr, "freq = %i\n", freq);
#endif				
				switch(freq_count) {
					case 0:
						start_freq = freq;
						break;
					case 1:
						freq_gap = freq - start_freq;
						break;
					default:	
						/* check gap */
						if((freq - last_freq) != freq_gap) {
							/* start a new group */
#ifdef DEBUG						
							fprintf(stderr, "start: %i, nof: %i, gap:%i\n", start_freq, freq_count, freq_gap);
#endif							
							store_softmac_row(country->string, start_freq, freq_count, freq_gap, ifindex);
							start_freq = freq;								
							freq_count = 0;
							freq_gap = 0;
						}
						break;
				}
				
				freq_count++;
				last_freq = freq;
				i++;
			}
			
			if(freq_count > 0) {
				store_softmac_row(country->string, start_freq, freq_count, freq_gap, ifindex);
			}
		}
		cs_row = cs_row->next;
	}
	
	return R_NOERR;
}

struct moddescr_t mod_dot11ext2_smtChannelSetTable = {

	name:		__FILE__,
	fn_ini:		init,
	fn_pin:		post_init,
	oidlist:	".1.3.6.1.4.1.937.2.2.2.8.1"

};


/*

The following schema entries are provided by this module :

O .1.3.6.1.4.1.937.2.2.2.8.1
L smtChannelSetEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1 .1.3.6.1.4.1.937.2.2.2.8.1.1

0	O .1.3.6.1.4.1.937.2.2.2.8.1.1
	L smtChanSetIndex
	T INTEGER
	A -
	R 1/2147483647

1	O .1.3.6.1.4.1.937.2.2.2.8.1.2
	L smtChanSetPhyType
	T INTEGER
	A R
	R 1,2,3,4,5,6

2	O .1.3.6.1.4.1.937.2.2.2.8.1.3
	L smtChanSetFirstChannelNumber
	T INTEGER
	A R

3	O .1.3.6.1.4.1.937.2.2.2.8.1.4
	L smtChanSetNumberOfChannels
	T INTEGER
	A R

4	O .1.3.6.1.4.1.937.2.2.2.8.1.5
	L smtChanSetChannelInterval
	T INTEGER
	A R

5	O .1.3.6.1.4.1.937.2.2.2.8.1.6
	L smtChanSetCountryList
	T OCTETSTR
	A R
	R 1/255

*/
