/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "pda.h"

#include "table.h"
#include "misc.h"
#include "object.h"
#include "objdata.h"
#include "misc.h"
#include "regulatory.h"
#include "prismoids.h"
#include "mod_mibII_interfaces.h"

static struct Pdr_Struct *get_country_list_PDR(ifindex) {
	
	switch(ifindex) {
		case 3:
			 return(pda_Find_Pdr(PDR_COUNTRY_LIST, NULL));
			 break;
		case 5:
			/* TODO */
			return (NULL);
			break;
		default:
			return (NULL);
			break;
	}
	
	return NULL;
}

static int country_domain_in_PDR(struct Pdr_Struct* pdr, struct country_domain_t* cd) {
	
	int no_countries, i;
	unsigned char* p = NULL;
	assert(cd);
	
	if(!pdr) {
		/* No PDR country list, 'cd' is always supported */
		return B_TRUE;
	}
	
	no_countries = (pdr->PdrLength - 1) / 2;
	
	if(no_countries == 0) {
		fprintf(stderr, "Error in PDA, countrylist is empty !!!\n");
		
		/* TEMP FIX, TODO REMOVE THIS */
		if(strcmp(cd->id, "FCC") == 0) {
			return B_TRUE;
		}
		else {
			return B_FALSE;
		}
	}
	
	p = pdr->PdrData;
	
	for(i = 0; i < no_countries; i++) {
		if((p[0] != 0x00) && (cd->type == REGULATORY_TYPE_DOMAIN)) {
			/* regdomain */
			if(p[0] == regulatory_domainid2value(cd->id)) {
				return B_TRUE;
			}			
		}
		else if ((p[0] == 0x00) && (cd->type == REGULATORY_TYPE_COUNTRY)) {
			/* country code */
			if(strncmp(&(p[1]), cd->id, 3) == 0) {
				return B_TRUE;
			}
		}
		
		p += 4;
	}
	
	return B_FALSE;
}

static void generate_fullmac_rows(int iface) {

	struct schent_t *sch;
	struct index_t *idx;
	struct row_t *row;
	int index,i;
	struct country_domain_t* cd;
	char country_code[4]; 
	struct regulatory_group_list_t *rgl;
	struct Pdr_Struct *country_list_pdrPtr = NULL;
	char found_countries[256];
	char * fc_p;
	int regdom;
	index = 1;
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.6.1");
	assert(sch);

	fc_p = &(found_countries[0]);
	*fc_p = '\0';
		
	country_list_pdrPtr = get_country_list_PDR(iface);
			
	cd = regulatory_get_country_domains();
		
	while(cd) {
		/* check the current phytype is supported by this cd and if
		 * the current cd is supported by the PDA */
		rgl = cd->regul_groups;
		while(rgl) {
			if (country_domain_in_PDR(country_list_pdrPtr, cd)) {
				/* add countries in this cd */
				for(i=0; i < cd->no_countries; i++) {
					snprintf(country_code, sizeof(country_code), "%s   ", cd->countries[i]);
					if (strstr(found_countries, country_code) != 0) {
						break; /* country already present */
					}
					else {
						sprintf(fc_p, "%s", country_code);
						fc_p = fc_p + 3;
						assert(fc_p < found_countries + sizeof(found_countries));
					}
					
					idx = index_new(sch);
					assert(idx);
						
					object_set_int(index_get_obj(idx, 0), iface);
					object_set_int(index_get_obj(idx, 1), index);
					row = row_create_by_index(idx);
					assert(row);
					index_done(idx);
			
					object_set_int(row_get_obj(row,0), index);
						
					object_set_string(row_get_obj(row,1), country_code, 3);
						
					regdom = regulatory_domainid2value(cd->id);
					if (regdom == 0xFF) {
						regdom = 0x00;
					}
					
					object_set_int(row_get_obj(row,2), regdom);
					index++;
				}
										
			}
			rgl = rgl->next;
		}						
			
		cd = cd->next;
	}
	
	
}

#define DOT11_OID_COUNTRYLIST           0x1f000001
#define MAX_COUNTRYLIST_SIZE 1024
static void generate_softmac_rows(int iface) {

	struct schent_t *sch;
	char* buf = calloc(sizeof(char), MAX_COUNTRYLIST_SIZE);
	char *p;
	struct index_t* idx;
	struct row_t* row;
	int index = 1;
	int regdom;
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.2.6.1");
	assert(sch);

	if(0 == get_prismoid(ifindex2ifname(iface), DOT11_OID_COUNTRYLIST, buf, MAX_COUNTRYLIST_SIZE)) {
		p = buf;
		while((*p)) {
			
			idx = index_new(sch);
			assert(idx);
						
			object_set_int(index_get_obj(idx, 0), iface);
			object_set_int(index_get_obj(idx, 1), index);
			row = row_create_by_index(idx);
			assert(row);
			index_done(idx);
			
			object_set_int(row_get_obj(row,0), index);
						
			object_set_string(row_get_obj(row,1), p, 3);
						
			regdom = 0x00; /* TODO, lookup regdom for country */
				
			object_set_int(row_get_obj(row,2), regdom);
			index++;
			
			p+=4;
		}
	}
	
	
	free(buf);	

}
		
static int init(void)
{
	struct schent_t *sc_sch;
	struct index_t *idx;
	struct object_t *obj;
	struct row_t *sc_row;
	int ifindex;
			
	sc_sch = schema_get(".1.2.840.10036.1.1.1");
	assert(sc_sch);
	
	/* loop all entries in station config tables */
	sc_row = sc_sch->rowlist;
	while(sc_row) {
		
		
		/* retrieve ifindex */
		idx = index_get(sc_row);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifindex = object_get_int(obj);
		index_done(idx);
		
		/* FullMac/SoftMac interfaces are handled different */
		switch(ifindex) {
			case 3:
				generate_fullmac_rows(ifindex);
				break;
			case 5:
				// Temp disabled, done in post_init now
				//generate_softmac_rows(ifindex);
				break;
			default:
				assert(1==0);
				break;
		}
		
		sc_row = sc_row->next;
	}
	
	return R_NOERR;
}

static int post_init(void) {
	
	/* temp solutions to prevent boot problems. (uAp is not reachable for countries in init-phase) */
	generate_softmac_rows(5);
	return R_NOERR;
}

struct moddescr_t mod_dot11ext2_smtCountrySupport = {

	name:		__FILE__,
	fn_ini:		init,
	fn_pin:		post_init,
	oidlist:	".1.3.6.1.4.1.937.2.2.2.6.1"

};


/*

The following schema entries are provided by this module :

O .1.3.6.1.4.1.937.2.2.2.6.1
L smtCountrySupportEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1 .1.3.6.1.4.1.937.2.2.2.6.1.1

	O .1.3.6.1.4.1.937.2.2.2.6.1.1
	L smtCountrySupportIndex
	T INTEGER
	A -
	R 1/255

	O .1.3.6.1.4.1.937.2.2.2.6.1.2
	L smtCountrySupportCode
	T OCTETSTR
	A R
	R 3

	O .1.3.6.1.4.1.937.2.2.2.6.1.3
	L smtCountrySupportRegDomain
	T INTEGER
	A R
	R 0,16,32,48,49,50,64

*/
