/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <string.h>
#include <aplog.h>
#include <stdlib.h>
#include <flash_layout.h>
#include "types.h"
#include "schema.h"
#include "table.h"
#include "objdata.h"
#include "mod_mibII_interfaces.h"


#include "identification/bootrec.h"
#include "identification/interfaces.h"
#include "identification/components.h"
#include "image.h"


#define FMT_HW_AP_COMPONENTS 		"1.3.6.1.4.1.937.1.26.2.%d"
#define FMT_SOFTWARE_COMPONENTS 	"1.3.6.1.4.1.937.1.27.%d"
#define FMT_INTERFACES 			"1.3.6.1.4.1.937.1.28.%d"


/* Role */
#define IDENT_MIB_ROLE_PRODUCER		0
#define IDENT_MIB_ROLE_CONSUMER		1


/*
 * Hardware Access Point Component IDs. The component ID in Flash must
 * be mapped to an object identifier that is specified in the ISL Reg MIB
 */

#define ISLREG_HW_AP_COMP_DUALBAND	2
#define ISLREG_HW_AP_COMP_DUALBAND_GW	3
#define ISLREG_HW_AP_COMP_3893_RECOVERY	4
#define ISLREG_HW_AP_COMP_PRISMGT	5
#define ISLREG_HW_AP_COMP_PRISMGT_GW	6


/*
 * Software Component IDs.
 */
 
#define ISLREG_SWCOMP_BOOTLOADER	1
#define ISLREG_SWCOMP_MVC		2
#define ISLREG_SWCOMP_UCLINUX_SYSTEM	3
#define ISLREG_SWCOMP_RESCUE_SYSTEM	4
#define ISLREG_SWCOMP_MTFW		5
#define ISLREG_SWCOMP_3893_BOOTLOADER	6
#define ISLREG_SWCOMP_DBAP_FIRMWARE	7
#define ISLREG_SWCOMP_INDIGO_MICRO_AP	8
#define ISLREG_SWCOMP_DBAP_FILESET	9
#define ISLREG_SWCOMP_PRISMGT_MICRO_AP	10
#define ISLREG_SWCOMP_DUETTE_MICRO_AP	11

/*
 * Interface IDs. The interface ID in Flash must be mapped to an
 * object identifier that is specified in the ISL Reg MIB
 */
 
#define ISLREG_IF_MAC_CTRL_IF		1
#define ISLREG_IF_RF_MODEM_IF		2
#define ISLREG_IF_MAC_COPR_IF		3
#define ISLREG_IF_SNMP_IF		4
#define ISLREG_IF_SNWNMP_IF		5
#define ISLREG_IF_HTTP_IF		6
#define ISLREG_IF_MANUF_TST_IF		7
#define ISLREG_IF_DEBUG_IF		8
#define ISLREG_IF_PRODUCT_IF		9
#define ISLREG_IF_OEM_IF		10
#define ISLREG_IF_HOST_IF               11
#define ISLREG_IF_39300_HW_IF           12
#define ISLREG_IF_37700M_HW_IF          13
#define ISLREG_IF_FILESET_IF            14
#define ISLREG_IF_BOOT_IF               15
#define ISLREG_IF_FIRMWARE_TFTP_IF      16
#define ISLREG_IF_PRIMARY_PROD_IF       17
#define ISLREG_IF_PRIMARY_OEM_IF        18
#define ISLREG_IF_39000M_HW_IF          19
#define ISLREG_IF_37700M_AP_IF          20
#define ISLREG_IF_39000M_AP_IF          21


static int componentId_to_oid(long compId, char *dst, int len);
static int interfaceId_to_oid(long interfaceId, char *dst, int len);
static int role_br2mib(int bootrec_role);



static int nof_components = 0;
static int nof_exp_if = 0;
static int nof_dep_if = 0;
static int nof_dep_alt_if = 0;



static int init(void)
{
	struct row_t *ct_row;
	struct index_t *ct_idx;
	struct schent_t *ct_sch;

	struct row_t *et_row;
	struct index_t *et_idx;
	struct schent_t *et_sch;

	struct row_t *dt_row;
	struct index_t *dt_idx;
	struct schent_t *dt_sch;

	struct row_t *at_row;
	struct index_t *at_idx;
	struct schent_t *at_sch;

	int last_record;
	int compIdx, expIfIndex, depIfIndex, depAltIfIndex;
	int count, first;
	struct s_bootrecord *record;
	struct s_bootrec_dep_if *depIf;
	struct s_bootrec_exp_if *expIf;
	long *current_bootrecs[10];
	const char *no_version_info = "No version info";
	const char *search_devs[] = {
		MTD_BOOTLOADER,
		MTD_APFIRMWARE,
		FILE_WEBFILESET,
		FILE_UAP_FIRMWARE,
		NULL};

	char buf[64];

	nof_components = get_current_bootrecords(search_devs, current_bootrecs, NULL);

	if(nof_components == 0) {
		aplog(EV_SNMP_GENERAL_ERROR,"Objectidentifier: No components found for Identification MIB",NULL);
		return(0);
	}

	/* If we received a record for the dummy Micro AP component remove it from
	 * the found bootrecords. The dummy Micro AP component should not be shown
	 * on the web interface.
	 */
	compIdx = 0;
	while ( compIdx < nof_components ) {
		record = (struct s_bootrecord *)current_bootrecs[compIdx];
		if((!record) || (valid_bootrecord(record) < 0)) {
			continue;
		}
		if ( ( record->hdr.code == BR_CODE_COMPONENT_ID ) && 
		     ( (struct s_bootrec_comp_id *)record)->id == BR_COMP_ID_DUMM ) {
			/* We found the dummy value for the microAP component.
			 * This entry should not be shown in the component list
			 * so delete it from the current_bootrecs array.
			 */
			for ( count = compIdx+1; count < nof_components; count++) 
				current_bootrecs[count-1] = current_bootrecs[count];
			nof_components--;
			break;
		}
		compIdx++;
	}

	ct_sch = schema_get(".1.3.6.1.4.1.937.2.3.2.1.1");
	et_sch = schema_get(".1.3.6.1.4.1.937.2.3.2.2.1");
	dt_sch = schema_get(".1.3.6.1.4.1.937.2.3.2.3.1");
	at_sch = schema_get(".1.3.6.1.4.1.937.2.3.2.4.1");

	for(compIdx = 0; compIdx < nof_components; compIdx++) {

		record = (struct s_bootrecord *)current_bootrecs[compIdx];

		ct_idx = index_new(ct_sch);
		object_set_uint(index_get_obj(ct_idx, 0), compIdx+1 );
		ct_row = row_create_by_index(ct_idx);
		index_done(ct_idx);	

		object_set_uint(row_get_obj(ct_row, 0), compIdx+1 );

		/* fill dummy data in the row. If the info is found in 
		 * the BR this dummy data will be overwritten. This 
		 * dummy data is needed to have 'complete' rows in
		 * SNMP. */
		object_set_objectid(row_get_obj(ct_row, 1), "0.0");
		object_set_string(row_get_obj(ct_row, 2), "", 0);
		
		depIfIndex = 0;
		last_record = 0;

		do {
			if((!record) || (valid_bootrecord(record) < 0)) {
				aplog(EV_SNMP_GENERAL_ERROR, "Find bootrec: No valid bootrecord", NULL);
				continue;
			}
			
			switch(record->hdr.code) {

				case BR_CODE_COMPONENT_ID:

					if(componentId_to_oid(((struct s_bootrec_comp_id *)record)->id, buf, sizeof(buf))) {
						object_set_objectid(row_get_obj(ct_row, 1), buf);
					} else {
						object_set_objectid(row_get_obj(ct_row, 1), "0.0");
					}
					break;
		
				case BR_CODE_COMPONENT_VERSION:
					if ( strlen(((struct s_bootrec_comp_ver *)record)->ver_string) == 0) {
						strncpy(buf, no_version_info, sizeof(buf));
					}
					else
					{
						strncpy(buf, ((struct s_bootrec_comp_ver *)record)->ver_string, sizeof(buf));
					}
					object_set_string(row_get_obj(ct_row, 2), buf, strlen(buf));
					break;

				case BR_CODE_DEPENDENT_IF:
		
					count = nof_dep_interfaces(record);
		
					if(count <= 0) {
						char tmp1[256];
						snprintf(tmp1, sizeof(tmp1), "Invalid dependent interface record %d %d", compIdx, count);
						aplog(EV_SNMP_GENERAL_ERROR,tmp1,NULL);
						break;
					}
		
					first = 1;
					depAltIfIndex = 0;
					depIf = (struct s_bootrec_dep_if *)&(record->start_of_data);
		
					while(count--) {
						interfaceId_to_oid(depIf->if_id, buf, sizeof(buf));
			
						if(first) {
			
							dt_idx = index_new(dt_sch);
							object_set_uint(index_get_obj(dt_idx, 0), compIdx+1 );
							object_set_uint(index_get_obj(dt_idx, 1), depIfIndex+1 );
							dt_row = row_create_by_index(dt_idx);
							index_done(dt_idx);
					
							object_set_uint    (row_get_obj(dt_row, 0), depIfIndex+1 );
							object_set_objectid(row_get_obj(dt_row, 1), buf );
							object_set_int     (row_get_obj(dt_row, 2), role_br2mib(depIf->role) );
							object_set_uint    (row_get_obj(dt_row, 3), depIf->variant );
			
							nof_dep_if++;
							first = 0;
			
						} else {
			
							at_idx = index_new(at_sch);
							object_set_uint(index_get_obj(at_idx, 0), compIdx+1 );
							object_set_uint(index_get_obj(at_idx, 1), depIfIndex+1 );
							object_set_uint(index_get_obj(at_idx, 2), depAltIfIndex+1 );
							at_row = row_create_by_index(at_idx);
							index_done(at_idx);
					
							object_set_uint    (row_get_obj(at_row, 0), depAltIfIndex+1 );
							object_set_objectid(row_get_obj(at_row, 1), buf );
							object_set_int     (row_get_obj(at_row, 2), role_br2mib(depIf->role) );
							object_set_uint    (row_get_obj(at_row, 3), depIf->variant );
			
							nof_dep_alt_if++;
							depAltIfIndex++;
						}
						depIf++;
					}
			
					depIfIndex++;
					break;
			
				case BR_CODE_EXPOSED_IF:
		
					count = nof_exp_interfaces(record);
		
					if(count <= 0 ) {
						char tmp1[256];
						snprintf(tmp1, sizeof(tmp1), "Invalid exposed interface record %d %d", compIdx, count);
						aplog(EV_SNMP_GENERAL_ERROR,tmp1,NULL);

						break;
					}
		
					expIf = (struct s_bootrec_exp_if *)&(record->start_of_data);
				       	
					for(expIfIndex = 0; expIfIndex < count; expIfIndex++, nof_exp_if++, expIf++) {
						et_idx = index_new(et_sch);
						object_set_uint(index_get_obj(et_idx, 0), compIdx+1 );
						object_set_uint(index_get_obj(et_idx, 1), expIfIndex+1 );
						et_row = row_create_by_index(et_idx);
						index_done(et_idx);
			
						interfaceId_to_oid(expIf->if_id, buf, sizeof(buf));

						object_set_uint    (row_get_obj(et_row, 0), expIfIndex+1 );
						object_set_objectid(row_get_obj(et_row, 1), buf );
						object_set_int     (row_get_obj(et_row, 2), role_br2mib(expIf->role) );
						object_set_uint    (row_get_obj(et_row, 3), expIf->variant );
						object_set_uint    (row_get_obj(et_row, 4), expIf->btm_compat);
						object_set_uint    (row_get_obj(et_row, 5), expIf->top_compat);

					}
					break;
		
				case BR_CODE_END_OF_BRA:
					last_record = 1;
					break;
		
				default:
					break;
		
			}
	
			record = next_bootrecord(record);
	
		} while(!last_record);
	
		/* Free the memory */
	
		free(current_bootrecs[compIdx]);
	
	} /* for all components */

	return(0);
}


struct moddescr_t mod_isl_prism_identification = {

	name:		__FILE__,
	fn_ini:		init,
	oidlist: 	""
};


////////////////////////////////////////////////////////////////////////////////////
// Some helper functions


static int componentId_to_oid(long compId, char *dst, int len)
{
	int n,l;
	char *format = FMT_SOFTWARE_COMPONENTS;

	switch(compId) {
		/* Software Component IDs */
		case BR_COMP_ID_SYST:	n = ISLREG_SWCOMP_UCLINUX_SYSTEM;	break;
		case BR_COMP_ID_RESC:	n = ISLREG_SWCOMP_RESCUE_SYSTEM;	break;
		case BR_COMP_ID_MVC1:	n = ISLREG_SWCOMP_MVC;			break;
		case BR_COMP_ID_BOOT:	n = ISLREG_SWCOMP_BOOTLOADER;		break;
		case BR_COMP_ID_LDV1:	n = ISLREG_SWCOMP_MTFW;			break;
		case BR_COMP_ID_DBO1:	n = ISLREG_SWCOMP_3893_BOOTLOADER;	break;
		case BR_COMP_ID_DAP1:	n = ISLREG_SWCOMP_DBAP_FIRMWARE;  	break;
		case BR_COMP_ID_DMT1:   n = ISLREG_SWCOMP_MTFW;           	break;
		case BR_COMP_ID_DFS1:	n = ISLREG_SWCOMP_DBAP_FILESET;   	break;
		case BR_COMP_ID_ITM1:	n = ISLREG_SWCOMP_INDIGO_MICRO_AP;	break;
		case BR_COMP_ID_LTM1:	n = ISLREG_SWCOMP_DUETTE_MICRO_AP;	break;
		/* case BR_COMP_ID_???:	n = ISLREG_SWCOMP_PRISMGT_MICRO_AP	break; */
		/* Hardware Component IDs */
		case BR_COMP_ID_ISL39300A:
			format = FMT_HW_AP_COMPONENTS;
			if(waneth_present())
				n = ISLREG_HW_AP_COMP_DUALBAND_GW;
			else
				n = ISLREG_HW_AP_COMP_DUALBAND;
			break;
		default:		return(0);
	}

	l =  snprintf(dst, len, format, n);

	return(l);
}



static int interfaceId_to_oid(long interfaceId, char *dst, int len)
{
	int n,l;
	switch(interfaceId) {
		case BR_IF_ID_ISL36356A:	n = ISLREG_IF_MAC_CTRL_IF;	break;
		case BR_IF_ID_MVC:		n = ISLREG_IF_MAC_COPR_IF;	break;
		case BR_IF_ID_SNMP_IF:		n = ISLREG_IF_SNMP_IF;		break;
		case BR_IF_ID_SNWNMP_IF:	n = ISLREG_IF_SNWNMP_IF;	break;
		case BR_IF_ID_HTTP_IF:		n = ISLREG_IF_HTTP_IF;		break;
		case BR_IF_ID_MANUF_TST_IF:	n = ISLREG_IF_MANUF_TST_IF;	break;
		case BR_IF_ID_DEBUG_IF:		n = ISLREG_IF_DEBUG_IF;		break;
		case BR_IF_ID_PRODUCT_IF:	n = ISLREG_IF_PRODUCT_IF;	break;
		case BR_IF_ID_OEM_IF:		n = ISLREG_IF_OEM_IF;		break;
		case BR_IF_ID_PCIHOST_IF:	n = ISLREG_IF_HOST_IF;	        break;
		case BR_IF_ID_ISL37704C:	n = ISLREG_IF_37700M_HW_IF;     break;
		case BR_IF_ID_TFTP:		n = ISLREG_IF_FIRMWARE_TFTP_IF;	break;
		case BR_IF_ID_ISL39000:		n = ISLREG_IF_39000M_HW_IF;     break; 
		case BR_IF_ID_ISL39300A:	n = ISLREG_IF_39300_HW_IF;	break;
		case BR_IF_ID_BOOT_IF:		n = ISLREG_IF_BOOT_IF;		break;
		case BR_IF_ID_FS_IF:		n = ISLREG_IF_FILESET_IF;	break;
		case BR_IF_ID_ISL37700_UAP_IF:	n = ISLREG_IF_37700M_AP_IF;	break;
		case BR_IF_ID_ISL39000_UAP_IF:	n = ISLREG_IF_39000M_AP_IF;	break;
		case BR_IF_ID_PRIMARY_PRODUCT_IF: n = ISLREG_IF_PRIMARY_PROD_IF;	break;
		case BR_IF_ID_PRIMARY_OEM_IF:	n = ISLREG_IF_PRIMARY_OEM_IF;	break;
		default:			return(0);
	}


	l = snprintf(dst, len, FMT_INTERFACES, n);

	return(l);
}


static int role_br2mib(int bootrec_role)
{
	char tmp1[256];
	switch(bootrec_role) {
	case BR_IF_ROLE_SERVER:
		return IDENT_MIB_ROLE_PRODUCER;
	case BR_IF_ROLE_CLIENT:
		return IDENT_MIB_ROLE_CONSUMER;
	default:
		snprintf(tmp1, sizeof(tmp1), "Illegal role in bootrecord %d\n", bootrec_role);
		aplog(EV_SNMP_GENERAL_ERROR,tmp1,NULL);

		return IDENT_MIB_ROLE_PRODUCER;
	}
}
                                                                 
                                                                 
                                                                 


