/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>

#include "object.h"
#include "objdata.h"
#include "table.h"
#include "schema.h"
#include "objflag.h"
#include "configfile.h"
#include "alarm.h"
#include "types.h"
#include "aplog.h"
#include "handler.h"

#include "mod_mibII_interfaces.h"
#include "lib_calls.h"

#define TEST_TIME 15
#define FILTERSTATUS_ENABLED  1
#define FILTERSTATUS_DISABLED 2
#define FILTERSTATUS_TEST     3

#define iptables_cmd "/bin/iptables-restore /tmp/iptables-input"

static void write_standard_rules(FILE *fp) {

	/* write static rules to accept packets on local interfaces */
    	fprintf(fp, "-A INPUT -i lo -j ACCEPT\n");
    	fprintf(fp, "-A OUTPUT -o lo -j ACCEPT\n");
	
	/* adjust MSS */
	fprintf(fp, "-A FORWARD -p tcp -m tcp --tcp-flags SYN,RST SYN -j TCPMSS --clamp-mss-to-pmtu\n");
}

static void write_tcp_flags(FILE *fp, struct object_t *obj) {
    int first = B_TRUE;
    
    if (object_get_bit(obj, 0)) {
	fprintf(fp,"SYN");
	first = B_FALSE;
    }

    if (object_get_bit(obj, 1)) {
	if(!first) {
	    fprintf(fp,",");
	}
	else {
	    first = B_FALSE;
	}
	fprintf(fp,"ACK");
    }

    if (object_get_bit(obj, 2)) {
    	if(!first) {
    	    fprintf(fp,",");
    	}
    	else {
    	    first = B_FALSE;
    	}
    	fprintf(fp,"FIN");
    }

    if (object_get_bit(obj, 3)) {
    	if(!first) {
    	    fprintf(fp,",");
    	}
    	else {
    	    first = B_FALSE;
    	}
    	fprintf(fp,"RST");
    }

    if (object_get_bit(obj, 4)) {
    	if(!first) {
    	    fprintf(fp,",");
    	}
    	else {
    	    first = B_FALSE;
    	}
    	fprintf(fp,"URG");
    }

    if (object_get_bit(obj, 5)) {
    	if(!first) {
    	    fprintf(fp,",");
    	}
    	else {
    	    first = B_FALSE;
    	}
    	fprintf(fp,"PSH");
    }

    if (first == B_TRUE) {
	/* no flags, write NONE */
	fprintf(fp,"NONE");
    }
    
    fprintf(fp, " ");
}



static void generate_iptables_rules(FILE* fp) {
	struct object_t* obj, *rs_obj;
	struct row_t* row = NULL;
	int pol, contr, prot;
	int i, rs;
	struct in_addr *ip;
	struct string_t* str;
	char buf[32];
	struct schent_t* sch = NULL;
	char oidbuf[MAX_OID_LENGTH];
	
	/* DEFAULT INPUT POLICY */
	obj = object_get(".1.3.6.1.4.1.937.2.1.9.2.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	pol = object_get_int(obj);
	if(pol == 1) {
		fprintf(fp, ":INPUT DROP\n");
	}
	else {
		fprintf(fp, ":INPUT ACCEPT\n");
	}
	
	/* DEFAULT OUTPUT POLICY */
	obj = object_get(".1.3.6.1.4.1.937.2.1.9.3.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	pol = object_get_int(obj);
	if(pol == 1) {
		fprintf(fp, ":OUTPUT DROP\n");
	}
	else {
		fprintf(fp, ":OUTPUT ACCEPT\n");
	}
	
	/* DEFAULT FORWARD POLICY */
	obj = object_get(".1.3.6.1.4.1.937.2.1.9.4.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	pol = object_get_int(obj);
	if(pol == 1) {
		fprintf(fp, ":FORWARD DROP\n");
	}
	else {
		fprintf(fp, ":FORWARD ACCEPT\n");
	}

	/* write standard rules */
	write_standard_rules(fp);
	
	
	/* The order of generation is relevant, order by OID */
	rs_obj = object_get_next(".1.3.6.1.4.1.937.2.1.9.5.1.99");
	if(rs_obj && (strcmp(rs_obj->sch->oid, ".1.3.6.1.4.1.937.2.1.9.5.1.99") == 0)) {
		row = rs_obj->row;
		sch = rs_obj->sch;		
	}
	while(row) {

		/* RowStatus */
		assert(rs_obj->flags & OF_DATAPRESENT);
		rs = object_get_rowstat(rs_obj);
		if (rs == RS_ACTIVE) {  
			
			/* Chain */
			obj = row_get_obj(row, 0);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			i = object_get_int(obj);
			switch(i) {
				case 1:
					fprintf(fp, "-A INPUT ");
					break;
				case 2:
					fprintf(fp, "-A OUTPUT ");
					break;
				case 3:
					fprintf(fp, "-A FORWARD ");
					break;
				default:
					assert(1==0);
					break;
			}
			
			/* Index */
			obj = row_get_obj(row, 1);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			i = object_get_int(obj);
		        fprintf(fp, "-m userindex --userindex %i ", i);	
			
			/* Control */
			obj = row_get_obj(row, 2);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			contr = object_get_int(obj);
			switch(contr) {
				case 1:
					fprintf(fp, "-j ACCEPT ");
					break;
				case 2:
					fprintf(fp, "-j DROP ");
					break;
				case 3:
					fprintf(fp, "-j REJECT ");
					break;
				case 4:
					fprintf(fp, "-j LOG ");
					break;
				case 5:
					/* no target */					
					break;
				default:
					assert(1==0);
					break;
			}
			
			/* Saddr */
			obj = row_get_obj(row, 3);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			ip = object_get_ipaddr(obj);
			fprintf(fp, "-s %s/", inet_ntoa(*ip));
			
			/* Smask */
			obj = row_get_obj(row, 4);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			ip = object_get_ipaddr(obj);
			fprintf(fp, "%s ", inet_ntoa(*ip));
			
			/* Daddr */
			obj = row_get_obj(row, 5);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			ip = object_get_ipaddr(obj);
			fprintf(fp, "-d %s/", inet_ntoa(*ip));
			
			/* Dmask */
			obj = row_get_obj(row, 6);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			ip = object_get_ipaddr(obj);
			fprintf(fp, "%s ", inet_ntoa(*ip));
												
			/* Protocol */
			obj = row_get_obj(row, 7);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			prot = object_get_int(obj);
			if (prot != 256) {
				fprintf(fp, "-p %i ", prot);
			}
			
			/* source/dest port only for TCP/UDP */
			if ((prot == IPPROTO_TCP) || (prot == IPPROTO_UDP)) {
				switch(prot) {
					case IPPROTO_TCP:
						fprintf(fp, "-m tcp ");
						break;
					case IPPROTO_UDP:
						fprintf(fp, "-m udp ");						
						break;
					default:
						assert(1==0);
						break;
				}
				
				/* Sport low */
				obj = row_get_obj(row, 8);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				i = object_get_int(obj);
				fprintf(fp, "--sport %i:", i);
				
				/* Sport high */
				obj = row_get_obj(row, 9);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				i = object_get_int(obj);
				fprintf(fp, "%i ", i);					

				/* Dport low */
				obj = row_get_obj(row, 10);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				i = object_get_int(obj);
				fprintf(fp, "--dport %i:", i);
				
				/* Dport high */
				obj = row_get_obj(row, 11);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				i = object_get_int(obj);
				fprintf(fp, "%i ", i);					
			}


			/* TCP related options */
			if (prot == IPPROTO_TCP) {
				
			 	fprintf(fp, "--tcp-flags ");
				
				/* mask */
				obj = row_get_obj(row, 13);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				write_tcp_flags(fp, obj);
				
				/* flags */
				obj = row_get_obj(row, 12);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				write_tcp_flags(fp, obj);
				
				/* Tcp Option */
				obj = row_get_obj(row, 14);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				i = object_get_int(obj);
				if(i != 256) {
					fprintf(fp, "--tcp-option %i ", i);
				}
				
			}
			
			/* Interface IN */
			obj = row_get_obj(row,15);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			i = object_get_int(obj);
			if (i != 0) {
				fprintf(fp, "-i %s ", ifindex2ifname(i));
			}
			
			/* Interface OUT */
			obj = row_get_obj(row,16);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			i = object_get_int(obj);
			if (i != 0) {
				fprintf(fp, "-o %s ", ifindex2ifname(i));
			}
			
			/* LOG control specific */
			if(contr == 4) {
				/* Log Level */
				obj = row_get_obj(row, 18);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				i = object_get_int(obj);
				fprintf(fp, "--log-level %i ", i);
				
				/* Log Prefix */
				obj = row_get_obj(row, 19);
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				str = object_get_string(obj);
				
				/* Convert STR to NULL-terminated string */
				strncpy(buf, str->data, str->len);
				buf[str->len] = '\0';
				
				fprintf(fp, "--log-prefix \"%s\" ", buf);
				
				/* Static limit for logging */
				fprintf(fp, "-m limit --limit 3/hour --limit-burst 5 ");
				
			}			
			
						
			/* end of line */
			fprintf(fp, "\n");			
			
		}
		
		row = NULL;
		object_get_oid(oidbuf, sizeof(oidbuf), rs_obj);
		rs_obj = object_get_next(oidbuf);
		if(rs_obj) {
			/* check if this is the same column */
			if(rs_obj->sch == sch) {
				row = rs_obj->row;
			}
		}
	}
	
}



static void switch_firewall(int fs) {
	FILE* fp;
	struct object_t* obj;
	
	fp = fopen("/tmp/iptables-input", "w");
	assert(fp);

	fprintf(fp, "# iptables-save. Generated by datastore\n");
	fprintf(fp, "*filter\n");
			
	switch(fs) {
		case FILTERSTATUS_ENABLED:
		case FILTERSTATUS_TEST:
			aplog(EV_NET_FIREWALL_DFLT, "disabled", NULL);
			aplog(EV_NET_FIREWALL_USER, "enabled", NULL);
			generate_iptables_rules(fp);
			break;

		case FILTERSTATUS_DISABLED:
			aplog(EV_NET_FIREWALL_USER, "disabled", NULL);
			aplog(EV_NET_FIREWALL_DFLT, "enabled", NULL);
			fprintf(fp, ":INPUT ACCEPT\n");
			fprintf(fp, ":FORWARD ACCEPT\n");
			fprintf(fp, ":OUTPUT ACCEPT\n");

			write_standard_rules(fp); 
						
			/* get the WanConnectionType object. If the WANtype
			 * is PPTP, we must prevent blocking the static IP
			 * address on the WAN ethernet port. */
			obj = object_get(".1.3.6.1.4.1.937.2.1.3.2.0");
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			if(object_get_int(obj) == 5 /* pptp */) {
				/* get the static WAN ip address */
				obj = object_get(".1.3.6.1.4.1.937.2.1.3.3.0");
				assert(obj);
				assert(obj->flags & OF_DATAPRESENT);
				
				/* add accept rule for this ip-address */
				fprintf(fp, "-A INPUT -j ACCEPT -i eth2 -d %s/255.255.255.255 \n", inet_ntoa(*(object_get_ipaddr(obj))));
			}

			/* block WAN port(s) */
			fprintf(fp, "-A INPUT -j DROP -i eth2\n");
			fprintf(fp, "-A INPUT -j DROP -i ppp+\n");
		
			break;
		default:
			assert(0==1);
			break;
	}

	fprintf(fp, "COMMIT\n");	
	fprintf(fp, "# Completed \n");
	fclose(fp);
	
	system(iptables_cmd);
	
//	unlink("/tmp/iptables-input");
	
}

static void callback(void) {
	int fs;
	struct object_t *obj;
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.9.1.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	fs = object_get_int(obj);
	
	if (fs == FILTERSTATUS_TEST) {
		object_set_int(obj, FILTERSTATUS_DISABLED);
		switch_firewall(FILTERSTATUS_DISABLED);
		/* schedule a write for datastore.dat */
		configfile_schedule_write();
	}
	
}

static int set(void){
	struct object_t *obj;
	int fs;
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.9.1.0");
	assert(obj);
	if (object_flag_isset(obj, OF_DIRTY)) {
		assert(obj->flags & OF_DATAPRESENT);
		fs = object_get_int(obj);
		
		if(fs == FILTERSTATUS_TEST) {
			/* test */
			alarm_register(callback,TEST_TIME);
		}				
		object_flag_clear(obj, OF_DIRTY);
		switch_firewall(fs);
	}

	return(0);

}

static int check(void){
	struct object_t *obj;
	int fs;
	struct schent_t *sch;
	struct row_t* row;
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.9.1.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	fs = object_get_int(obj);
	
	if( (fs != 2) && !(object_flag_isset(obj, OF_DIRTY))) {
		/* Filter Status is not Disabled and also not dirty */
		
		/* An object has been set and this is not allowed because 
		 * the FilterStatus is not 'Disabled' */
		return (R_CHECKFAILED);
	}

	/* HACK to fill in a default value for count OID */
	/* Be sure the WFS can always retrieve an object 
         * This should be done by adding a MIBException mechanism
	 * to the toolchain (smidump)
	 */
	sch = schema_get(".1.3.6.1.4.1.937.2.1.9.5.1");
	assert(sch);
	row = sch->rowlist;
	while(row) {
		obj = row_get_obj(row, 17);
		assert(obj);
		if(!(obj->flags & OF_DATAPRESENT)) {
			object_set_counter(obj, 0);
		}
		row = row->next;
	}
		
	
	return (R_NOERR);
}


static int get(void){
	int fd;
	FILE* fp;
	char* cmd[3];
	char buf[160];
	char* sp;
	int ui, count, chain;
	struct object_t* obj;
	
	cmd[0] = "/bin/iptables-save";
	cmd[1] = "--counters";
	cmd[2] = NULL;
	
	fd = RunAndRead(cmd);
	
	fp = fdopen(fd, "r");
	if(fp) {
		while(fgets(buf, sizeof(buf), fp)) {
			if(buf[0] != '[') {
				continue;
			}			
			
			sp = strstr(buf, "--userindex");
			if (sp) {
				sp = sp + strlen("--userindex ");
				ui = atoi(sp); /* user index */
				
				sp = strstr(buf, "-A");
				sp = sp + strlen("-A ");
				
				if(strncmp(sp, "INP", 3) == 0){
					chain = 1;
				}
				else if(strncmp(sp, "OUT", 3) == 0) {
					chain = 2;
				}
				else {
					chain = 3;
				}
				
				sp = buf+1;
				
				count = atoi(sp);
				
				/* try to find this oid */
				snprintf(buf, sizeof(buf), ".1.3.6.1.4.1.937.2.1.9.5.1.21.%i.%i", chain, ui);
				obj = object_get(buf);
				
				if(obj) {
					object_set_counter(obj, count);					
				}
				else {
					aplog(EV_SNMP_GENERAL_ERROR, "Error fetching counter object", NULL);
				}
				
				
			}
			
						
		}
		
	}
	else {
		aplog(EV_SNMP_GENERAL_ERROR, "Error fetching firewall counters", NULL);
	}
		
	
	
	
	return(5);
}

static int wanconnection_updated(void) {
	struct object_t *obj;
	int fs;
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.9.1.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	
	/* only update if the FW is switched off */
	fs = object_get_int(obj);
	if(fs == FILTERSTATUS_DISABLED) {
		switch_firewall(fs);
	}

	return R_NOERR;	
}

static int init(void) {
	
	static struct handler_t* hnd;
	
	/* register an update handler on the WAN connection type and 
	 * WAN ipaddress. The default firewall depends on these settings */
	hnd = handler_add(H_SET, wanconnection_updated);

	/* update on change of WanConnectionType */
	handler_connect(hnd, ".1.3.6.1.4.1.937.2.1.3.2");
	
	/* and update on change of Wan Ipaddress */
	handler_connect(hnd, ".1.3.6.1.4.1.937.2.1.3.3");
	
	return R_NOERR;
}


struct moddescr_t mod_isl_standardext_generic_firewall = {
	name:		__FILE__,
	fn_set:		set,
	fn_chk:		check,
	fn_get:		get,
	fn_ini:		init,
	oidlist:	".1.3.6.1.4.1.937.2.1.9.1 .1.3.6.1.4.1.937.2.1.9.2 .1.3.6.1.4.1.937.2.1.9.3 .1.3.6.1.4.1.937.2.1.9.4 .1.3.6.1.4.1.937.2.1.9.5.1"
};


/*

O .1.3.6.1.4.1.937.2.1.9.1
L fwIpFilterStatus
T INTEGER
A RW
R 1,2,3
D 2

O .1.3.6.1.4.1.937.2.1.9.2
L fwIpFilterRuleDefaultInputPolicy
T INTEGER
A RW
R 1,2
D 2

O .1.3.6.1.4.1.937.2.1.9.3
L fwIpFilterRuleDefaultOutputPolicy
T INTEGER
A RW
R 1,2
D 2

O .1.3.6.1.4.1.937.2.1.9.4
L fwIpFilterRuleDefaultForwardPolicy
T INTEGER
A RW
R 1,2
D 1

O .1.3.6.1.4.1.937.2.1.9.5.1
L fwIpFilterRuleEntry
T TABLE
I .1.3.6.1.4.1.937.2.1.9.5.1.1 .1.3.6.1.4.1.937.2.1.9.5.1.2

0	O .1.3.6.1.4.1.937.2.1.9.5.1.1
	L fwIpFilterRuleChain
	T INTEGER
	A -
	R 1,2,3

1	O .1.3.6.1.4.1.937.2.1.9.5.1.2
	L fwIpFilterRuleIndex
	T INTEGER
	A -
	R 1/2147483647

2	O .1.3.6.1.4.1.937.2.1.9.5.1.4
	L fwIpFilterRuleControl
	T INTEGER
	A C
	R 1,2,3,4,5
	D 1

3	O .1.3.6.1.4.1.937.2.1.9.5.1.6
	L fwIpFilterRuleSaddr
	T IPADDR
	A C
	R 4
	D HEX:00000000

4	O .1.3.6.1.4.1.937.2.1.9.5.1.7
	L fwIpFilterRuleSmask
	T IPADDR
	A C
	R 4
	D HEX:00000000

5	O .1.3.6.1.4.1.937.2.1.9.5.1.8
	L fwIpFilterRuleDaddr
	T IPADDR
	A C
	R 4
	D HEX:00000000

6	O .1.3.6.1.4.1.937.2.1.9.5.1.9
	L fwIpFilterRuleDmask
	T IPADDR
	A C
	R 4
	D HEX:00000000

7	O .1.3.6.1.4.1.937.2.1.9.5.1.10
	L fwIpFilterRuleProtocol
	T INTEGER
	A C
	R 0/256
	D 256

8	O .1.3.6.1.4.1.937.2.1.9.5.1.11
	L fwIpFilterRuleSourcePortLow
	T INTEGER
	A C
	R 0/65535
	D 0

9	O .1.3.6.1.4.1.937.2.1.9.5.1.12
	L fwIpFilterRuleSourcePortHigh
	T INTEGER
	A C
	R 0/65535
	D 65535

10	O .1.3.6.1.4.1.937.2.1.9.5.1.13
	L fwIpFilterRuleDestPortLow
	T INTEGER
	A C
	R 0/65535
	D 0

11	O .1.3.6.1.4.1.937.2.1.9.5.1.14
	L fwIpFilterRuleDestPortHigh
	T INTEGER
	A C
	R 0/65535
	D 65535

12	O .1.3.6.1.4.1.937.2.1.9.5.1.16
	L fwIpFilterRuleTcpFlags
	T BITS
	A C
	R 0,1,2,3,4,5

13	O .1.3.6.1.4.1.937.2.1.9.5.1.17
	L fwIpFilterRuleTcpFlagsMask
	T BITS
	A C
	R 0,1,2,3,4,5

14	O .1.3.6.1.4.1.937.2.1.9.5.1.18
	L fwIpFilterTcpOption
	T INTEGER
	A C
	R 0/256
	D 256

15	O .1.3.6.1.4.1.937.2.1.9.5.1.19
	L fwIpFilterRuleIfInIndex
	T INTEGER
	A C
	R 0/2147483647

16	O .1.3.6.1.4.1.937.2.1.9.5.1.20
	L fwIpFilterRuleIfOutIndex
	T INTEGER
	A C
	R 0/2147483647

17	O .1.3.6.1.4.1.937.2.1.9.5.1.21
	L fwIpFilterRuleMatches
	T COUNTER
	A R
	R 0/4294967295

18	O .1.3.6.1.4.1.937.2.1.9.5.1.22
	L fwIpFilterRuleLogLevel
	T INTEGER
	A C
	R 0,1,2,3,4,5,6,7

19	O .1.3.6.1.4.1.937.2.1.9.5.1.23
	L fwIpFilterRuleLogPrefix
	T OCTETSTR
	A C
	R 0/29

20	O .1.3.6.1.4.1.937.2.1.9.5.1.99
	L fwIpFilterRuleRowStatus
	T ROWSTAT
	A C
	R 1,2,3,4,5,6


*/
