/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>

#include "object.h"
#include "objdata.h"
#include "serialize.h"
#include "table.h"
#include "objflag.h"
#include "schema.h"

#include "cfg_file_lib.h"

#define MAX_RADIUS_SECRET_LENGTH 32
#define MAX_STRING_LENGTH 255

#define CONFIGFILE "radius"

static int set(void)
{
	struct object_t *obj;
	struct object_t *objentry;
	struct string_t *secret;
	struct in_addr *ip;
	int port;
	int rowstat;
	char oidbuf[MAX_OID_LENGTH];
	FILE* file = NULL;

	file = create_cfg_file(CONFIGFILE);
	
	objentry = object_get_next(".1.3.6.1.4.1.937.2.1.4.2.1.3");		
	assert(objentry);
	while(strcmp(objentry->sch->oid,".1.3.6.1.4.1.937.2.1.4.2.1.3")==0) 
	{
		int i = 1;
		rowstat = object_get_rowstat(row_get_obj(objentry->row,1));
		
		if(rowstat == 1)
		{
			fprintf(file, "server %i ", i++);
			
			obj = row_get_obj(objentry->row, 2);
			ip = object_get_ipaddr(obj);	
			fprintf(file,"%s ",inet_ntoa(*ip));
			
			obj = row_get_obj(objentry->row, 4);
			secret = object_get_string(obj);

			/* Convert to HEX */
			fprintf(file, "HEX:");
			for(i=0; i < secret->len; i++) {
				fprintf(file, "%02X", secret->data[i]);		
			}
			fprintf(file, " ");
			
			obj = row_get_obj(objentry->row, 3);
			port = object_get_int(obj);
			fprintf(file, "%d \n", port);
		}
		object_get_oid(oidbuf, sizeof(oidbuf), objentry);
		objentry = object_get_next(oidbuf);
		
	}
	
	close_cfg_file(file);
	
	return(0);
}

// Check whether the new entry can be added to the table.
// The table may not contain more than a maximum number of entries
static int check(void)
{
	struct schent_t *sch;
	struct row_t *row;
	int num = 0;
	long maxtableentries=0;
	
	sch = schema_get(".1.3.6.1.4.1.937.2.1.4.2.1");
	assert(sch);

	row = sch->rowlist;
	
	num = table_count(sch);

	maxtableentries = object_get_int(object_get(".1.3.6.1.4.1.937.2.1.4.1.0"));

	if(num>maxtableentries)
	{
		return R_CHECKFAILED;
	}
	else
	{
		return R_NOERR;
	}
}

struct moddescr_t mod_isl_standardext_generic_radius = {

	name:		__FILE__,
	fn_set:		set,
	fn_chk:		check,
	oidlist:	".1.3.6.1.4.1.937.2.1.4.1 .1.3.6.1.4.1.937.2.1.4.2.1"

};




