/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <aplog.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <signal.h>

#include "object.h"
#include "objflag.h"
#include "objdata.h"
#include "mod_web.h"

#include "aplock.h"
#include "lib_calls.h"

const char pass_cfg[] = "/var/etc/passwd.conf";

static struct string_t *get_lan_mac_addr(void)
{
	struct object_t *ifobj;
	struct string_t *ifmac;

	/* get MAC from IF-table from interface 3 */
	ifobj = object_get(".1.3.6.1.2.1.2.2.1.6.3");
	assert(ifobj);
	assert(ifobj->flags & OF_DATAPRESENT);
	ifmac = object_get_string(ifobj);

	return ifmac;
}

/* PostInit because the MAC address is only available in post_init */
static int post_init(void)
{
	struct object_t *chgip_obj;
	struct string_t *chgipaddr, *ifmac;

	/* Set the first 6 octets of snmpChangeIPAddress to our own MAC address */
	chgip_obj = object_get(".1.3.6.1.4.1.937.2.1.2.6.0");
	assert(chgip_obj);
	assert(chgip_obj->flags & OF_DATAPRESENT);
	chgipaddr = object_get_string(chgip_obj);

	ifmac = get_lan_mac_addr();
	
	memcpy(chgipaddr->data, ifmac->data, 6);
	
	return R_NOERR;
}

static int check(void)
{
	struct object_t *chip_obj, *usedhcp_obj;
	struct string_t *chgipaddr, *ifmac;
        int usedhcp;
	
	/* This chip_obj should contain our MAC address. If not, a
	 * SIDED packet for an other AP has been recieved and we should
	 * reject this set */
	chip_obj = object_get(".1.3.6.1.4.1.937.2.1.2.6.0");
	assert(chip_obj);
	
	if(object_flag_isset(chip_obj, OF_DIRTY)) {
		assert(chip_obj->flags & OF_DATAPRESENT);
		chgipaddr = object_get_string(chip_obj);
		
		ifmac = get_lan_mac_addr();
				
		if(memcmp(chgipaddr->data, ifmac->data, 6) != 0) {
			aplog(EV_SNMP_SIDED,"rejected",NULL);
			return (R_CHECKFAILED);
		}
                
		aplog(EV_SNMP_SIDED, "accepted", NULL);
                
	}

        /* snmpUseDHCP can only be set to never (3) when we have a valid 
           ip-address for the snmpChangeIpAddress object. */
        usedhcp_obj = object_get(".1.3.6.1.4.1.937.2.1.2.7.0");
	assert(usedhcp_obj);
        
        if (object_flag_isset(usedhcp_obj, OF_DIRTY)) {
            assert(usedhcp_obj->flags & OF_DATAPRESENT);
	    usedhcp=object_get_int(usedhcp_obj);		
            
            if (usedhcp == 3) { /* never */
                assert(chip_obj->flags & OF_DATAPRESENT);
                chgipaddr = object_get_string(chip_obj);
              
                /* Check the ip address. 4 octets. */
                if (memcmp(&chgipaddr->data[6], "\x00\x00\x00\x00", 4) == 0) {
                    aplog(EV_SNMP_INFO, "Can't set snmpUseDHCP to never because no valid IP address to use is defined.", NULL);
                    return (R_CHECKFAILED);
                }
            }
        }
	
	return(R_NOERR);

}

void mod_isl_standardext_generic_snmp_kill_dhcpcd(char* ifname) {
	FILE* fp;
	static char buf[128];
	snprintf(buf, sizeof(buf), "/var/run/dhcpcd-%s.pid", ifname);
	fp = fopen(buf, "r");
	if(fp) {
		fgets(buf, sizeof(buf), fp);
		kill(atoi(buf), SIGHUP);
	}
	else {
		fprintf(stderr, "dhcpcd not running on interface %s\n", ifname);
	}
	sleep(1);
}

static int set(void)
{
	struct object_t *obj, *chip_obj, *usedhcp_obj, *hostname_obj;
	struct object_t *wangw_obj, *wantype_obj;
	int wangw_present = B_FALSE;
	int lockstate, altaddr;
	struct string_t *chgipaddr;
	struct string_t *wcomstr;
	struct string_t *host;
	struct in_addr *wangw;
        char *hostname;	
	int usedhcp;
	FILE* fp;
	char sip[16]; /* static ip */	
	char snm[16]; /* static netmask */
	char sgw[16]; /* static gateway */
	char passwd[65];
	char sysbuf[80];
	
	memset(passwd,0,65);

	obj = object_get(".1.3.6.1.4.1.937.2.1.2.5.0");
	assert(obj);
	if(object_flag_isset(obj, OF_DIRTY)) 
	{
		lockstate=object_get_int(obj);
		if(lockstate == 1) {
			aplock();
		} else {
			aplock_nolock();
		}
		object_flag_clear(obj, OF_DIRTY);
	}	
	
	chip_obj = object_get(".1.3.6.1.4.1.937.2.1.2.6.0");
	assert(chip_obj);
	usedhcp_obj = object_get(".1.3.6.1.4.1.937.2.1.2.7.0");	
	assert(usedhcp_obj);
        hostname_obj = object_get(".1.3.6.1.4.1.937.2.1.2.8.0");	
        assert(hostname_obj);
	
	if( object_flag_isset(chip_obj, OF_DIRTY) || 
            object_flag_isset(usedhcp_obj, OF_DIRTY) ||
            object_flag_isset(hostname_obj, OF_DIRTY))
	{
		aplog(EV_SYS_SUBSYS_STOP,"LAN network",NULL);
		
		/* Check if the WAN connection is enabled and if a 
		 * gateway was set on the WAN connection */
		wantype_obj = object_get(".1.3.6.1.4.1.937.2.1.3.2.0");
		assert(wantype_obj);
		assert(wantype_obj->flags & OF_DATAPRESENT);
		
		wangw_obj = object_get(".1.3.6.1.4.1.937.2.1.3.5.0");
		assert(wangw_obj);
		assert(wangw_obj->flags & OF_DATAPRESENT);
		wangw = object_get_ipaddr(wangw_obj);
		
		wangw_present = ((object_get_int(wantype_obj) != 1 /* disabled */)
				 && (wangw->s_addr != 0));
		
		/* remove default route on the LAN */
		system("/bin/route del default br0");
			
	/* Stardate 10405.4 Max Yin
		The above code cannot really remove default route
	*/

		system("/bin/ifconfig br0 down");
		system("/bin/ifconfig br0 up");
	
		/* Stop dhcpcd for this interface */
 		mod_isl_standardext_generic_snmp_kill_dhcpcd("br0");
		
		assert(chip_obj->flags & OF_DATAPRESENT);
		chgipaddr=object_get_string(chip_obj);
		assert(usedhcp_obj->flags & OF_DATAPRESENT);
		usedhcp=object_get_int(usedhcp_obj);		
                
                sprintf(sip,"%d.%d.%d.%d",chgipaddr->data[6],chgipaddr->data[7],chgipaddr->data[8],chgipaddr->data[9]);
		sprintf(snm,"%d.%d.%d.%d",chgipaddr->data[10],chgipaddr->data[11],chgipaddr->data[12],chgipaddr->data[13]);
		sprintf(sgw,"%d.%d.%d.%d",chgipaddr->data[14],chgipaddr->data[15],chgipaddr->data[16],chgipaddr->data[17]);		

                if(usedhcp==3) /* never */
		{
			/* Set default gateway only if on the WAN connection no
			 * gateway is defined and the LAN gateway != 0.0.0.0
			 */
			if(!wangw_present && (strcmp("0.0.0.0", sgw) != 0)) {
				snprintf(sysbuf, sizeof(sysbuf), "/bin/ifup br0:0 static %s %s %s", sip, snm, sgw);
			}
			else {
				snprintf(sysbuf, sizeof(sysbuf), "/bin/ifup br0:0 static %s %s", sip, snm);
			}
			system(sysbuf);
			
			/* remove autoip address */
			Killall("autoip", NULL);
			system("/bin/ifup br0:1 none");

                        aplog(EV_NET_LAN_STATUS,"static",NULL);
		}
		else
		{
			Killall("autoip", NULL);
                        hostname = NULL;
                        if(hostname_obj->flags & OF_DATAPRESENT) {
				host = object_get_string(hostname_obj);
				if (host->len > 0) {
					hostname = host->data;
				}						
			}
			
			if(hostname) {
				snprintf(sysbuf, sizeof(sysbuf), "/bin/ifup br0:0 dhcp %s", hostname);
			}
			else {
				snprintf(sysbuf, sizeof(sysbuf), "/bin/ifup br0:0 dhcp");
			}
			system(sysbuf);
			
			/* switch on autoip */
			system("/bin/ifup br0:1 autoip");

			aplog(EV_NET_LAN_STATUS,"dynamic",NULL);
                        
		}

		/* Add multicast route on LAN */
		system("/bin/route add -net 224.0.0.0 netmask 240.0.0.0 br0");
		
		aplog(EV_SYS_SUBSYS_START,"LAN network",NULL);	
		
		object_flag_clear(chip_obj, OF_DIRTY);
		object_flag_clear(usedhcp_obj, OF_DIRTY);
                object_flag_clear(hostname_obj, OF_DIRTY);
	}	
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.2.2.0");
	assert(obj);
	if(object_flag_isset(obj, OF_DIRTY)) 
	{
		wcomstr = object_get_string(obj);
		strncpy(passwd,wcomstr->data,wcomstr->len);
		
		fp = fopen(pass_cfg,"w");
		fprintf(fp,"%s\n",passwd);
		fclose(fp);
		
		mod_web_reload_boa();
		
		object_flag_clear(obj, OF_DIRTY);
	}

	
	return R_NOERR;

}

struct moddescr_t mod_isl_standardext_generic_snmp = {

	name:		__FILE__,
	fn_pin:		post_init,
	fn_set:		set,
	fn_chk:		check,
	oidlist:	".1.3.6.1.4.1.937.2.1.2.2 .1.3.6.1.4.1.937.2.1.2.5 .1.3.6.1.4.1.937.2.1.2.6 .1.3.6.1.4.1.937.2.1.2.7 .1.3.6.1.4.1.937.2.1.2.8 .1.3.6.1.4.1.937.2.1.2.9"

};

/*
O .1.3.6.1.4.1.937.2.1.2.5
L snmpLockStatus
T INTEGER
A RW
R 1,2

O .1.3.6.1.4.1.937.2.1.2.6
L snmpChangeIPAddress
T OCTETSTR
A RW
R 18

O .1.3.6.1.4.1.937.2.1.2.7
L snmpUseDHCP
T INTEGER
A RW
R 1,2,3

O .1.3.6.1.4.1.937.2.1.2.8
L snmpHostName
T OCTETSTR
A RW
R 0/6

O .1.3.6.1.4.1.937.2.1.2.9
L snmpAlternateAddress (OBSOLETE)
T INTEGER
A RW
R 1,2

*/
