/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/sockios.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <net/ethernet.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <aplog_defs.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "objflag.h"
#include "object.h"
#include "easyprism.h"
#include "aplog.h"
#include "cfg_file_lib.h"
#include "lib_calls.h"
#include "mod_isl_standardext_generic_snmp.h"

#define WANIF  "eth2"
#define pptp_exe "/bin/pptp"

const char cfg_name_wan[]="wan";
const char cfg_name_pppoe[]="pppoe";
const char cfg_pap[]="/var/etc/pap-secrets";
const char cfg_chap[]="/var/etc/chap-secrets";
const char pppd_exe[]="/bin/pppd";

static void write_general_ppp_options(FILE* fp, struct string_t* username, struct in_addr localip) {

	fprintf(fp, "persist\n");
	fprintf(fp, "nodetach\n");
	fprintf(fp, "mru 1492\n");
	fprintf(fp, "mtu 1492\n");
	fprintf(fp, "defaultroute\n");
	fprintf(fp, "noipdefault\n");
	fprintf(fp, "usepeerdns\n");
	fprintf(fp, "lcp-echo-failure 3\n");
	fprintf(fp, "lcp-echo-interval 10\n");
	
	/* write user */
	if(username && (username->len > 0)) {
		fprintf(fp, "user \"%s\"\n", username->data);
	}

	if(localip.s_addr != 0) {
		fprintf(fp, "%s:0.0.0.0\n", inet_ntoa(localip));
	}

}


static int set(void)
{
	
	struct object_t *obj;
	int fd;
	int conntype;
	struct in_addr *wanip;
	struct in_addr *localip;	
	struct in_addr *wanmask;
	struct in_addr *wangateway;
	struct in_addr *pptpserver;	
	struct string_t * wanmac;
	struct string_t * username;
	struct string_t * secret;
	struct string_t * phone;	
	struct string_t * servicename;	
	struct string_t * host;
	char* hostname;
	FILE* ppp_fp;
		
	int wannapt;
	char wanipaddr[32];
	char wanmaskaddr[32];	
	char wangatewayaddr[32];
	
	char temp[1024];
	char sysbuf[80];
	
	unlink("/var/respawnd/non_fatal/pppd");
	unlink("/var/respawnd/non_fatal/pptp");
	
	system("/bin/killall -9 pptp_callmgr pppoe pppd pptp");
	
	mod_isl_standardext_generic_snmp_kill_dhcpcd("eth2");

	/* Flush NAT rules */
	system("/bin/iptables -t nat -F POSTROUTING");

        /* WAN MAC address */
	obj = object_get(".1.3.6.1.4.1.937.2.1.3.1.0");
	assert(obj);
	if (object_flag_isset(obj, OF_DIRTY) ) {
		struct ifreq ifr;
		unsigned char ifup = B_FALSE;

		wanmac = object_get_string(obj);

		if(memcmp(&wanmac->data[0], "\0\0\0\0\0\0", ETH_ALEN) == 0) {
			/* get the MAC address from the LAN port */
			obj = object_get(".1.3.6.1.2.1.2.2.1.6.2");
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			wanmac = object_get_string(obj);
		}
			
		/* The new MAC address can only be set if the interface 
		 * is down. 
		 */
			
		/* get the present state and save it */
		strncpy(ifr.ifr_name, WANIF, IF_NAMESIZE);
		IfIoctl(SIOCGIFFLAGS, &ifr);
		ifup = (ifr.ifr_flags & IFF_UP) ? B_TRUE: B_FALSE;
			
		/* bring interface down */
		ifr.ifr_flags &= ~IFF_UP;
		IfIoctl(SIOCSIFFLAGS, &ifr);

		/* Get the old MAC address struct */			
		IfIoctl(SIOCGIFHWADDR, &ifr);

		/* Set the new MAC address */						
		memcpy(&ifr.ifr_hwaddr.sa_data, &wanmac->data[0], ETH_ALEN);
		strncpy(ifr.ifr_name, WANIF, IF_NAMESIZE);
		IfIoctl(SIOCSIFHWADDR, &ifr);
						
		/* Return interface in old state */
		if(ifup) {
			ifr.ifr_flags |= IFF_UP;
			IfIoctl(SIOCSIFFLAGS, &ifr);			
		}
		
		object_flag_clear(obj, OF_DIRTY);
	}

	obj = object_get(".1.3.6.1.4.1.937.2.1.3.2.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	conntype = object_get_int(obj);

	obj = object_get(".1.3.6.1.4.1.937.2.1.3.3.0");
	assert(obj);
	wanip = object_get_ipaddr(obj);
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.3.4.0");
	assert(obj);
	wanmask = object_get_ipaddr(obj);
		
	obj = object_get(".1.3.6.1.4.1.937.2.1.3.5.0");
	assert(obj);
	wangateway = object_get_ipaddr(obj);
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.3.6.0");
	assert(obj);
	if(obj->flags & OF_DATAPRESENT) {
		username = object_get_string(obj);
	}
	else {
		username = NULL;
	}
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.3.7.0");
	assert(obj);
	if(obj->flags & OF_DATAPRESENT) {
		secret = object_get_string(obj);
	}
	else {
		secret = NULL;
	}
		
	obj = object_get(".1.3.6.1.4.1.937.2.1.3.8.0");
	assert(obj);
	localip = object_get_ipaddr(obj);

	obj = object_get(".1.3.6.1.4.1.937.2.1.3.9.0");
	assert(obj);
	if(obj->flags & OF_DATAPRESENT) {
		servicename = object_get_string(obj);
	}
	else {
		servicename = NULL;
	}
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.3.10.0");
	assert(obj);
	pptpserver = object_get_ipaddr(obj); 

	obj = object_get(".1.3.6.1.4.1.937.2.1.3.11.0");
	assert(obj);
	if(obj->flags & OF_DATAPRESENT) {
		phone = object_get_string(obj);
	}
	else {
		phone = NULL;
	}
	
	obj = object_get(".1.3.6.1.4.1.937.2.1.3.12.0");
	assert(obj);
	assert(obj->flags & OF_DATAPRESENT);
	wannapt = object_get_int(obj);
	
	snprintf(temp, sizeof(temp), "\"%s\" * \"%s\"\n", username->data, secret->data);
	
	fd = creat(cfg_pap,O_CREAT);
	write(fd,temp,strlen(temp));
	close(fd);
	
	fd = creat(cfg_chap,O_CREAT);
	write(fd,temp,strlen(temp));	
	close(fd);
	
	chmod(cfg_pap, S_IRUSR|S_IWUSR);
	chmod(cfg_chap,S_IRUSR|S_IWUSR);
	
	switch(conntype)
	{
	  case 1:  //DISABLED
	   		aplog(EV_NET_WAN_STATUS,"disabled",NULL);

			snprintf(sysbuf, sizeof(sysbuf), "/bin/ifup %s none", WANIF);						
			system(sysbuf);
		        break;
			
	  case 2:  //STATIC IP
		        aplog(EV_NET_WAN_STATUS,"static",NULL);
			break;
			
	  case 3:  //DYNAMIC IP
			aplog(EV_NET_WAN_STATUS, "dynamic",NULL);
			hostname = "";
			obj = object_get(".1.3.6.1.4.1.937.2.1.2.8.0");
			assert(obj);
			if(obj->flags & OF_DATAPRESENT) {
				host = object_get_string(obj);
				if (host->len > 0) {
					hostname = host->data;
				}						
			}
			snprintf(sysbuf, sizeof(sysbuf), "/bin/ifup %s dhcp %s", WANIF, hostname);
			system(sysbuf);		
	   	        break;
			
	  case 4:  //PPPOE
		  	ppp_fp = fopen("/var/etc/ppp_options", "w+");
			assert(ppp_fp);
			
			/* write static pppoe specific options */
			fprintf(ppp_fp, "plugin pppoe\n");
			fprintf(ppp_fp, "tty %s\n", WANIF);

			/* write general ppp options */
			write_general_ppp_options(ppp_fp, username, *localip);
			
			/* write servicename */
			if(servicename && (servicename->len > 0)) {
				fprintf(ppp_fp, "rp_pppoe_service %s\n", servicename->data);
			}			
					
			fclose(ppp_fp);
			
			/* create respawn file (first in /tmp to prevent respawnd from
			 * starting an unfinished file */
			ppp_fp = fopen("/tmp/pppd", "w+");
			assert(ppp_fp);
			fprintf(ppp_fp, "#!/bin/sh\n");
					
			fprintf(ppp_fp, "/bin/pppd file /var/etc/ppp_options");

			/* Be sure to run it in the background */
			fprintf(ppp_fp, " &\n");
			
			/* set read & execute bit on this file */
			fchmod(ppp_fp->fd,   S_IRUSR | S_IXUSR );
			
			fclose(ppp_fp);
			rename("/tmp/pppd", "/var/respawnd/non_fatal/pppd");
			
			/* pppoe will be started by respawnd automaticly. */
			
			aplog(EV_NET_WAN_STATUS,"pppoe",NULL);
			
			break;
		  
	  case 5 :  //PPTP
			ppp_fp = fopen("/var/etc/ppp_options", "w+");
			assert(ppp_fp);
			
			/* write general ppp options */
			write_general_ppp_options(ppp_fp, username, *localip);
			
			fclose(ppp_fp);
			
			/* create respawn file (first in /tmp to prevent respawnd from
			 * starting an unfinished file */
			ppp_fp = fopen("/tmp/pptp", "w+");
			assert(ppp_fp);
			fprintf(ppp_fp, "#!/bin/sh\n");

			fprintf(ppp_fp, "/bin/killall -9 pppd pptp pptp_callmgr\n");
			fprintf(ppp_fp, "sleep 2\n");

			fprintf(ppp_fp, "/bin/pptp %s ", inet_ntoa(*pptpserver));

			if(phone->len != 0 ) {
				fprintf(ppp_fp, "--phone %s ", phone->data);
			
			} 
			fprintf(ppp_fp, "file /var/etc/ppp_options");
			
			/* Be sure to run it in the background */
			fprintf(ppp_fp, " &\n");
			
			/* set read & execute bit on this file */
			fchmod(ppp_fp->fd,   S_IRUSR | S_IXUSR );
			
			fclose(ppp_fp);
			
			rename("/tmp/pptp", "/var/respawnd/non_fatal/pptp");
			
			/* pptp will be started by respawnd automaticly. */
							  	
			aplog(EV_NET_WAN_STATUS,"pptp",NULL);
			break;
	  default:
	  break;
	}

	/* set the static ip-address on the WAN if needed */
	if( (conntype == 2 /* static */)
			|| (conntype == 4 /* pppoe */)
			|| (conntype == 5 /* pptp */)
			) {

			strcpy(wanipaddr, inet_ntoa(*wanip));
			strcpy(wanmaskaddr, inet_ntoa(*wanmask));
			strcpy(wangatewayaddr, inet_ntoa(*wangateway));
			
			if(wangateway->s_addr == 0) {
				snprintf(sysbuf, sizeof(sysbuf), "/bin/ifup %s static %s %s", WANIF, wanipaddr, wanmaskaddr);
			}
			else {
				snprintf(sysbuf, sizeof(sysbuf), "/bin/ifup %s static %s %s %s", WANIF, wanipaddr, wanmaskaddr, wangatewayaddr);
			}	

			system(sysbuf);

	}
	
	/* check for NAPT */
	if(wannapt == 1 && conntype != 1)
	{
		aplog(EV_NET_NAPT_STATUS,"enabled",NULL);
		snprintf(sysbuf, sizeof(sysbuf), "/bin/iptables -t nat -A POSTROUTING -j MASQUERADE -o %s", WANIF);
		system(sysbuf);
		system("/bin/iptables -t nat -A POSTROUTING -j MASQUERADE -o ppp+");	
	}
	else
	{
		aplog(EV_NET_NAPT_STATUS,"disabled",NULL);
	}

	return R_NOERR;
}


static int init(void) {
	
	
	/* hmm. this should not be here */
	system("/bin/ifconfig lo 127.0.0.1 up");
//	IfConfig("lo", "127.0.0.1", "up", NULL);
	
	return R_NOERR;
}

struct moddescr_t mod_isl_standardext_generic_wan = {

	name:		__FILE__,
	fn_set:		set,
	fn_ini:		init,
	oidlist:	".1.3.6.1.4.1.937.2.1.3.1 .1.3.6.1.4.1.937.2.1.3.2 .1.3.6.1.4.1.937.2.1.3.3 .1.3.6.1.4.1.937.2.1.3.4 .1.3.6.1.4.1.937.2.1.3.5 .1.3.6.1.4.1.937.2.1.3.6 .1.3.6.1.4.1.937.2.1.3.7 .1.3.6.1.4.1.937.2.1.3.8 .1.3.6.1.4.1.937.2.1.3.9 .1.3.6.1.4.1.937.2.1.3.10 .1.3.6.1.4.1.937.2.1.3.11 .1.3.6.1.4.1.937.2.1.3.12 .1.3.6.1.4.1.937.2.1.3.13"

};

	
/*

*/
	


