/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <stdio.h>
#include <assert.h>

#include "mod_mibII_interfaces.h"
#include "table.h"
#include "schema.h"
#include "objdata.h"
#include "object.h"



static int get() {
	struct schent_t *sch;
	struct row_t *row, *next_row;
	struct object_t *obj, *ifopstat_obj;
	int ifopstat;
	struct index_t *idx;
	int i;
	char buf[MAX_OID_LENGTH+1];
	
	sch = schema_get(".1.3.6.1.2.1.31.1.2.1");
	assert(sch);

	/* destroy all existing rows */
	row = sch->rowlist;
	while(row) {
		next_row = row->next;
		row_delete(row);
		row = next_row;
	}
		
	/* Create entries for all operational interfaces */
	for(i = 1; i < NO_INTERFACES; i++) {
		
		/* Get operation Status from ifTable */
		snprintf(buf, sizeof(buf), ".1.3.6.1.2.1.2.2.1.8.%d", i);
		ifopstat_obj = object_get(buf);
		assert(ifopstat_obj);
		assert(ifopstat_obj->flags & OF_DATAPRESENT);
		ifopstat = object_get_int(ifopstat_obj);
		
		if (ifopstat > 5) {
			/* operational status is LowerLayerDown or NotPresent */
			continue;
		}
		
		switch(i) {
			case INTERFACE_LO:
			case INTERFACE_ETH2:
			case INTERFACE_PPP0:
				/* independant interfaces. Have no up- or down-
				 * dependancy */
				idx = index_new(sch);
				assert(idx);
				
				/* get higher layer */
				obj = index_get_obj(idx, 0);
				assert(obj);
				object_set_int(obj, i);
				
				/* get lower layer */
				obj = index_get_obj(idx, 1);
				assert(obj);
				object_set_int(obj, 0); /* NO LOWER LAYER */
				
				/* create row */
				row = row_create_by_index(idx);
				assert(row);
				
				/* Set rowstatus */
				obj = row_get_obj(row, 2);
				assert(obj);
				object_set_rowstat(obj, 1); /* active */
						
				index_done(idx);
				
				/* FALLTHRU */
			case INTERFACE_BR0:
				/* Bridge has no up-dependancy */
				idx = index_new(sch);
				assert(idx);
				
				/* get higher layer */
				obj = index_get_obj(idx, 0);
				assert(obj);
				object_set_int(obj, 0); /* NO UPPER LAYER */
				
				/* get lower layer */
				obj = index_get_obj(idx, 1);
				assert(obj);
				object_set_int(obj, i);
				
				/* create row */
				row = row_create_by_index(idx);
				assert(row);
				
				/* Set rowstatus */
				obj = row_get_obj(row, 2);
				assert(obj);
				object_set_rowstat(obj, 1); /* active */
				
				index_done(idx);
				
				break;
				
			default:
				/* interface under the bridge. Have no 
				 * down-dependancy, but have BR0 as 
				 * upper layer */
					
				idx = index_new(sch);
				assert(idx);
				
				/* get higher layer */
				obj = index_get_obj(idx, 0);
				assert(obj);
				object_set_int(obj, INTERFACE_BR0);
				
				/* get lower layer */
				obj = index_get_obj(idx, 1);
				assert(obj);
				object_set_int(obj, i);
				
				/* create row */
				row = row_create_by_index(idx);
				assert(row);
				
				/* Set rowstatus */
				obj = row_get_obj(row, 2);
				assert(obj);
				object_set_rowstat(obj, 1); /* active */
				
				index_done(idx);
				
				idx = index_new(sch);
				assert(idx);
				
				/* get higher layer */
				obj = index_get_obj(idx, 0);
				assert(obj);
				object_set_int(obj, i);
				
				/* get lower layer */
				obj = index_get_obj(idx, 1);
				assert(obj);
				object_set_int(obj, 0); /* NO LOWER LAYER */
				
				/* create row */
				row = row_create_by_index(idx);
				assert(row);
				
				/* Set rowstatus */
				obj = row_get_obj(row, 2);
				assert(obj);
				object_set_rowstat(obj, 1); /* active */
				
				index_done(idx);

				break;
			
					
		}				
	}
	
	
	return(10);
}



struct moddescr_t mod_mibII_ifMIB_ifStackTable = {

	name:		__FILE__,
	fn_get:		get,
	oidlist:	".1.3.6.1.2.1.31.1.2.1"

};


/*

O .1.3.6.1.2.1.31.1.2.1
L ifStackEntry
T TABLE
I .1.3.6.1.2.1.31.1.2.1.1 .1.3.6.1.2.1.31.1.2.1.2

	O .1.3.6.1.2.1.31.1.2.1.1
	L ifStackHigherLayer
	T INTEGER
	A -
	R 0/2147483647

	O .1.3.6.1.2.1.31.1.2.1.2
	L ifStackLowerLayer
	T INTEGER
	A -
	R 0/2147483647

	O .1.3.6.1.2.1.31.1.2.1.3
	L ifStackStatus
	T ROWSTAT
	A R
	R 1,2,3,4,5,6

*/
