/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>

#include "types.h"
#include "schema.h"
#include "table.h"
#include "object.h"
#include "objdata.h"
#include "serialize.h"
#include "aplog.h"
#include "mod_mibII_interfaces.h"


static int fd = 0;
static int set_called = 0; /* flags if the set_function has been called one or more times */
static int initialized = B_FALSE; /* flag indicating if where are complete initialized */

static struct
{
	char dev[16];
	int in_bytes;
	int in_packets;
	int in_errs;
	int in_drop;
	int in_fifo;
	int in_frame;
	int in_compressed;
	int in_multicast;
	int out_bytes;
	int out_packets;
	int out_errs;
	int out_drop;
	int out_fifo;
	int out_frame;
	int out_compressed;
	int out_multicast;
} dev_stats;

static char* interfaces[23] = { "lo", 
		             "eth0", "eth1", "eth2", "eth3", 
		             "br0", "ppp0",
			     "eth1_0", "eth1_1", "eth1_2", "eth1_3",
			     "eth1_4", "eth1_5", "eth1_6", "eth1_7",
			     "eth3_0", "eth3_1", "eth3_2", "eth3_3",
			     "eth3_4", "eth3_5", "eth3_6", "eth3_7" };


char* ifindex2ifname(int ifindex) {
	assert(ifindex > 0);
	assert(ifindex <= NO_INTERFACES);
	return(interfaces[ifindex-1]);
}

int ifname2ifindex(char* ifname)
{
	int ifindex;
	for(ifindex = 1 ; ifindex <= NO_INTERFACES; ifindex++)
	{
		if(strcmp(ifname,interfaces[ifindex-1])==0)
		{
			return ifindex;
		}			
	}
	return -1;
}		

#define IFINDEX_WAN	INTERFACE_ETH2

int interface_present(int ifnum)
{
	struct object_t *obj;
	char opstatus_oid[MAX_OID_LENGTH];

	/* Determine presence from Operational Status */
	snprintf(opstatus_oid, sizeof(opstatus_oid), ".1.3.6.1.2.1.2.2.1.8.%d", ifnum);
	obj = object_get(opstatus_oid);
	assert(obj);
	return(object_get_int(obj) != 6); /* 6 == not present */
}

int waneth_present(void)
{
	return(interface_present(IFINDEX_WAN));
}


static int init(void)
{

	fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	assert(fd != -1);	

	return(0);
}





void asptotab(char* buf)
{
	int i1,i2;
	char* buftmp;
	buftmp = (char*) malloc(strlen(buf));
	memset(buftmp,0,strlen(buf));
	
	i2=0;
	for(i1=0;i1<strlen(buf);i1++)
	{
		buftmp[i2]=buf[i1];
		i2++;

		if(buf[i1]==' ')
		{
			while(buf[i1]==' ')
			{
				i1++;
			}
			i1--;
		}
			
	}	
	strcpy(buf,buftmp);
	free(buftmp);
}

void remove_name(char* buf, char* name)
{
	int i1=0;
	int i2=0;
	int i3=0;
	char* buftmp;

	
	buftmp = (char*) malloc(strlen(buf));
	memset(buftmp,0,strlen(buf));
			
	while(buf[i1] != ':')
	{
	
	   if(buf[i1] == ' ')
	   	i3++;
		
	   i1++;		
	}
	memcpy(name,&buf[i3],i1-1);
	i1++;
	i1++;
	
	for(i1=i1;i1<strlen(buf);i1++)
	{
		buftmp[i2]=buf[i1];
		i2++;
			
	}	
	strcpy(buf,buftmp);
	free(buftmp);

	
}

#define PROC_NET_DEV "/proc/net/dev"

static int get(void)
{	
	struct row_t *row;
	struct schent_t *sch;
	struct ifreq ifr;
	FILE* devfp;
	int devnr = -1;
	int stop = 0;

	char nulstr[64];
	
	struct object_t* ifname;
	char buf[256];
	int ret;
	char devnam[32];

	memset(nulstr, 0, 64);
	memset(buf,0,256);
	memset(devnam,0,32);


	sch = schema_get(".1.3.6.1.2.1.2.2.1");
	assert(sch);
	row = sch->rowlist;

	while(row) {

		ifname = row_get_obj(row, 18);
		assert(ifname);
		serialize(ifname, ifr.ifr_name, IF_NAMESIZE);
		
		if (ioctl(fd, SIOCGIFMTU, &ifr) == 0) {
			/* mtu */
			object_set_int(row_get_obj(row, 3), ifr.ifr_mtu);
			
			/* speed */
			//object_set_gauge(row_get_obj(row, 4), ifr.ifr_mtu);
			
			/* Hardware address */
			if (ioctl(fd, SIOCGIFHWADDR, &ifr) == 0) {
				object_set_string(row_get_obj(row, 5), ifr.ifr_hwaddr.sa_data, 6);
			}
			else {
				fprintf(stderr, "%s", strerror(errno));
			}

			/* admin status, Operational Status, Promisc. mode, etc.*/
			/* Check for 'set_called' to prevent that the initial 
			   values are overwritten during startup */
			if ((ioctl(fd, SIOCGIFFLAGS, &ifr) == 0) && set_called) {
				
				/* Admin Status */
				object_set_int(row_get_obj(row, 6), (ifr.ifr_flags & IFF_UP)?1:2);
				
				/* Operational Status */
				object_set_int(row_get_obj(row, 7), (ifr.ifr_flags & IFF_UP)?1:2);
				
				/* Promiscuous mode */
				object_set_int(row_get_obj(row, 27), (ifr.ifr_flags & IFF_PROMISC)?1:2);
			}
			
		}
		else {
			/* device does not exist, reset all values */

			/* mtu */
			object_set_int(row_get_obj(row, 3), 0);

			/* speed */
			object_set_gauge(row_get_obj(row, 4), 0);
	
			/* Hardware address */
			object_set_string(row_get_obj(row, 5), nulstr , 6);
			
			/* admin status */
			object_set_int(row_get_obj(row, 6), 2);
			
			/* Operational Status (dependant on iftype) */
			switch(object_get_int(row_get_obj(row, 2))){
				case 23: /* PPP */
				case 209: /* bridge */
				case 157: /* WDS */
					object_set_int(row_get_obj(row, 7), 7);
					break;
				default:
					object_set_int(row_get_obj(row, 7), 6);					
					break;
			}
			
			/* Promiscuous mode */
			object_set_int(row_get_obj(row, 27), 2);
			
		}
		
		//////////////////////////
		// GET /proc/net/dev INFO
		//////////////////////////
		devfp = fopen(PROC_NET_DEV,"r");

                assert(devfp);
	
		fgets(buf,sizeof(buf),devfp); // rubbish
		fgets(buf,sizeof(buf),devfp); // rubbish
		
		devnr = -1;
		while(devnr != object_get_int(row_get_obj(row, 0)) && stop != 1)
		{
			//CLEAR STATS
			memset(&dev_stats,0,sizeof(dev_stats));
			
			if(fgets(buf,sizeof(buf),devfp)==NULL)
			{
				// REACH END OF 'FILE' STOP WHILE
				stop =1;
			}
			else
			{
				asptotab(buf);
			
				memset(devnam,0,sizeof(devnam));
				remove_name(buf,devnam);
			
				devnr = ifname2ifindex(devnam);
			}
			
		}
		stop = 0;
		
		ret = fclose(devfp);
		
		if(devnr == object_get_int(row_get_obj(row, 0)))
		{
			sscanf(buf,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
						&dev_stats.in_bytes,		&dev_stats.in_packets, 
						&dev_stats.in_errs,		&dev_stats.in_drop, 
						&dev_stats.in_fifo,		&dev_stats.in_frame, 
						&dev_stats.in_compressed, 	&dev_stats.in_multicast, 	
						&dev_stats.out_bytes,		&dev_stats.out_packets, 
						&dev_stats.out_errs,		&dev_stats.out_drop, 
						&dev_stats.out_fifo,		&dev_stats.out_frame, 
						&dev_stats.out_compressed, 	&dev_stats.out_multicast
				);	
		}
		
		// In Octets
		object_set_counter(row_get_obj(row, 9),dev_stats.in_bytes);
		
		// In Unicast Packets
		object_set_counter(row_get_obj(row, 10),dev_stats.in_packets-dev_stats.in_multicast);
		
		// In Discard
		object_set_counter(row_get_obj(row, 11),dev_stats.in_drop);		
		
		// In Errors
		object_set_counter(row_get_obj(row, 12),dev_stats.in_errs);

		// In Unknown Protos
		//object_set_counter(row_get_obj(row, 13),inunknown);
				
		// Out Octets
		object_set_counter(row_get_obj(row, 14),dev_stats.out_bytes);
				
		// Out Unicast Packts
		object_set_counter(row_get_obj(row, 15),dev_stats.out_packets-dev_stats.out_multicast);
				
		// Out Discards
		object_set_counter(row_get_obj(row, 16),dev_stats.out_drop);
				
		// Out Errors
		object_set_counter(row_get_obj(row, 17),dev_stats.out_errs);
		
		// IFXTABLE (augment)
		
		// ifInMulticastPkts & inInBroadcastPkts (linux counts them together)
		object_set_counter(row_get_obj(row, 19), dev_stats.in_multicast);
		object_set_counter(row_get_obj(row, 20), dev_stats.in_multicast);
		
		// ifOutMulticastPkts & ifOutBroadcastPkts (linux counts them together)
		object_set_counter(row_get_obj(row, 21), dev_stats.out_multicast);
		object_set_counter(row_get_obj(row, 22), dev_stats.out_multicast);
		
		// ifHCInOctets (linux counts only 32 bit, so we use the 32 bit value)
		object_set_counter64(row_get_obj(row, 23), (long long) dev_stats.in_bytes);
		
		// ifHCOutOctets (linux counts only 32 bit, so we use the 32 bit value)
		object_set_counter64(row_get_obj(row, 24), (long long) dev_stats.out_bytes);

		row = row->next;
		
	}
	
	
	
	return(5);
}



static int set(void)
{
	struct row_t *row;
	struct object_t *ifname;
	int adminstatus;
	struct schent_t *sch;
	struct ifreq ifr;

	/* flag flag */
	set_called = 1;
	
	sch = schema_get(".1.3.6.1.2.1.2.2.1");
	assert(sch);
	row = sch->rowlist;

	while(row) {

		ifname = row_get_obj(row, 18);
		assert(ifname);
		
		
		serialize(ifname, ifr.ifr_name, IF_NAMESIZE);

		adminstatus = object_get_int(row_get_obj(row,6));
		
		if ( initialized && (ioctl(fd, SIOCGIFFLAGS, &ifr) == 0)) {
			if(adminstatus == 1) {
				ifr.ifr_flags |= 1;
				aplog(EV_NET_IFACE_STATUS, ifr.ifr_name, "up", NULL);
			} else {
				ifr.ifr_flags &= ~1;
				aplog(EV_NET_IFACE_STATUS, ifr.ifr_name, "down" ,NULL);
			}
		
			ioctl(fd, SIOCSIFFLAGS, &ifr);
			
		}

		row = row->next;
	}

	return(0);
}	

static int post_init(void) {
	
	initialized = B_TRUE;
	
	/* Bring interfaces up. (the easiest way to do that is to call set ) */
	set();
	
	return R_NOERR;
}

struct moddescr_t mod_mibII_interfaces = {

	name:		__FILE__,
	fn_ini:		init,
	fn_get:		get, 
	fn_set:		set,
	fn_pin:		post_init,
	oidlist:	".1.3.6.1.2.1.2.2.1"
};

/*

The following schema entries are provided by this module :

O .1.3.6.1.2.1.2.2.1
L ifEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1 

0	O .1.3.6.1.2.1.2.2.1.1
	L ifIndex
	T INTEGER32
	A R
	R 1/2147483647

1	O .1.3.6.1.2.1.2.2.1.2
	L ifDescr
	T OCTETSTR
	A R
	R 0/255

2	O .1.3.6.1.2.1.2.2.1.3
	L ifType
	T INTEGER
	A R
	R 1,10,100,101,102,103,104,105,106,107,108,109,11,110,111,112,113,114,115,116,117,118,119,12,120,121,122,123,124,125,126,127,128,129,13,130,131,132,133,134,135,136,137,138,139,14,140,141,142,143,144,145,146,147,148,149,15,150,151,152,153,154,155,156,157,158,159,16,160,161,162,163,164,165,166,167,168,169,17,170,171,172,173,174,175,176,177,178,179,18,180,181,182,183,184,185,186,187,188,189,19,190,191,192,193,194,195,196,2,20,21,22,23,24,25,26,27,28,29,3,30,31,32,33,34,35,36,37,38,39,4,40,41,42,43,44,45,46,47,48,49,5,50,51,52,53,54,55,56,57,58,59,6,60,61,62,63,64,65,66,67,68,69,7,70,71,72,73,74,75,76,77,78,79,8,80,81,82,83,84,85,86,87,88,89,9,90,91,92,93,94,95,96,97,98,99

3	O .1.3.6.1.2.1.2.2.1.4
	L ifMtu
	T INTEGER32
	A R

4	O .1.3.6.1.2.1.2.2.1.5
	L ifSpeed
	T GAUGE
	A R

5	O .1.3.6.1.2.1.2.2.1.6
	L ifPhysAddress
	T OCTETSTR
	A R

6	O .1.3.6.1.2.1.2.2.1.7
	L ifAdminStatus
	T INTEGER
	A RW
	R 1,2,3

7	O .1.3.6.1.2.1.2.2.1.8
	L ifOperStatus
	T INTEGER
	A R
	R 1,2,3,4,5,6,7

8	O .1.3.6.1.2.1.2.2.1.9
	L ifLastChange
	T TICKS
	A R

9	O .1.3.6.1.2.1.2.2.1.10
	L ifInOctets
	T COUNTER
	A R

10	O .1.3.6.1.2.1.2.2.1.11
	L ifInUcastPkts
	T COUNTER
	A R

11	O .1.3.6.1.2.1.2.2.1.13
	L ifInDiscards
	T COUNTER
	A R

12	O .1.3.6.1.2.1.2.2.1.14
	L ifInErrors
	T COUNTER
	A R

13	O .1.3.6.1.2.1.2.2.1.15
	L ifInUnknownProtos
	T COUNTER
	A R

14	O .1.3.6.1.2.1.2.2.1.16
	L ifOutOctets
	T COUNTER
	A R

15	O .1.3.6.1.2.1.2.2.1.17
	L ifOutUcastPkts
	T COUNTER
	A R

16	O .1.3.6.1.2.1.2.2.1.19
	L ifOutDiscards
	T COUNTER
	A R

17	O .1.3.6.1.2.1.2.2.1.20
	L ifOutErrors
	T COUNTER
	A R

O .1.3.6.1.2.1.31.1.1.1
L ifXEntry
T AUGMENT
I .1.3.6.1.2.1.2.2.1

18	O .1.3.6.1.2.1.31.1.1.1.1
	L ifName
	T OCTETSTR
	A R
	R 0/255

19	O .1.3.6.1.2.1.31.1.1.1.2
	L ifInMulticastPkts
	T COUNTER
	A R
	R 0/4294967295

20	O .1.3.6.1.2.1.31.1.1.1.3
	L ifInBroadcastPkts
	T COUNTER
	A R
	R 0/4294967295

21	O .1.3.6.1.2.1.31.1.1.1.4
	L ifOutMulticastPkts
	T COUNTER
	A R
	R 0/4294967295

22	O .1.3.6.1.2.1.31.1.1.1.5
	L ifOutBroadcastPkts
	T COUNTER
	A R
	R 0/4294967295

23	O .1.3.6.1.2.1.31.1.1.1.6
	L ifHCInOctets
	T COUNTER64
	A R
	R 0/18446744073709551615

24	O .1.3.6.1.2.1.31.1.1.1.10
	L ifHCOutOctets
	T COUNTER64
	A R
	R 0/18446744073709551615

25	O .1.3.6.1.2.1.31.1.1.1.14
	L ifLinkUpDownTrapEnable
	T INTEGER
	A R
	R 1,2

26	O .1.3.6.1.2.1.31.1.1.1.15
	L ifHighSpeed
	T GAUGE
	A R
	R 0/4294967295

27	O .1.3.6.1.2.1.31.1.1.1.16
	L ifPromiscuousMode
	T INTEGER
	A R
	R 1,2

28	O .1.3.6.1.2.1.31.1.1.1.17
	L ifConnectorPresent
	T INTEGER
	A R
	R 1,2

29	O .1.3.6.1.2.1.31.1.1.1.18
	L ifAlias
	T OCTETSTR
	A R
	R 0/64

30	O .1.3.6.1.2.1.31.1.1.1.19
	L ifCounterDiscontinuityTime
	T UNSIGNED32
	A R


*/
