/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <net/if.h>
#include <errno.h>
#include <fnmatch.h>
#include "aplog.h"
#include <aplog_defs.h>
#include "types.h"
#include "schema.h"
#include "table.h"
#include "objdata.h"
#include "object.h"
#include "objflag.h"

#include "mod_mibII_interfaces.h"

#define DEBUG 0

const char proc_arp[] = "/proc/net/arp";
const char proc_snmp[] = "/proc/net/snmp";
const char ipv4_path[] = "/proc/sys/net/ipv4/";

#define MAX_IPADDR_STR 16
#define MAX_MACADDR_STR 18
#define MAX_IFNAME_STR 8

static struct
{
	char   ifname[MAX_IFNAME_STR];
	struct sockaddr_in ipaddr;
	int    index;
	struct sockaddr_in netmask;
	struct sockaddr_in broadcast;
	int    reasm;		
} ip_info;

static struct
{
	int  ifindex;
	int  used;
	int  confirmed;
	int  updated;
	int  ref;
	int  staticarp;
	int  noarp;
	char ipaddress[MAX_IPADDR_STR];
	char macaddress[MAX_MACADDR_STR];
} arp_info;

static struct
{
	int fwd;
	int ttl;
	int inrecv;
	int inhdrerr;
	int inaddrerr;
	int fwddg;
	int inunkprot;
	int indisc;
	int indel;
	int outreq;
	int outdisc;
	int outnoroute;
	int reasmto;
	int reasmreq;
	int reasmok;
	int reasmfails;
	int fraqok;
	int fragfail;
	int fragcreat;
} ip_stats;

void create_entry_arp_table()
{
	struct row_t *row;
	struct schent_t *sch;
	struct index_t *idx;
	struct object_t *obj;
	int tmpmac[32];
	char macvalue[32];
	struct in_addr ip;

	sch = schema_get(".1.3.6.1.2.1.4.22.1");
	assert(sch);
	
	idx= index_new(sch);
	assert(idx);
	obj = index_get_obj(idx, 0);	
	assert(obj);
	
	object_set_int(obj, arp_info.ifindex);	
	obj = index_get_obj(idx, 1);	
	assert(obj);
	
	inet_aton(arp_info.ipaddress, &ip);
	object_set_ipaddr(obj, ip);	

	row = row_create_by_index(idx);
	assert(row);
	index_done(idx);
	
	sscanf(arp_info.macaddress,"%x:%x:%x:%x:%x:%x",&tmpmac[0],&tmpmac[1],&tmpmac[2],&tmpmac[3],&tmpmac[4],&tmpmac[5]);
	sprintf(macvalue,"%c%c%c%c%c%c",tmpmac[0],tmpmac[1],tmpmac[2],tmpmac[3],tmpmac[4],tmpmac[5]);

	
	// SET ALL VALUES
	object_set_int(	  row_get_obj(row, 0),arp_info.ifindex);
	object_set_string(row_get_obj(row, 1),macvalue,6);
	object_set_ipaddr(row_get_obj(row, 2),ip);
	object_set_int(   row_get_obj(row, 3), (arp_info.staticarp) ? 4 : 3);
	
}

static int get_arp_info()
{
	FILE* fp;
	char buf[1024];
	char d1[4];
	char flag[8];
	int  flagint;
	char mac[MAX_MACADDR_STR];
	char dev[MAX_IFNAME_STR];
	char ip[MAX_IPADDR_STR];
	char mask[MAX_IPADDR_STR];
	char* ptr;
	int ret;
	
	fp = fopen(proc_arp,"r");
	fgets(buf,sizeof(buf),fp);
	while(fgets(buf,sizeof(buf),fp))
	{
		sscanf(buf,"%s %s %s %s %s %s",ip,d1,flag,mac,mask,dev);
	
		strcpy(arp_info.ipaddress,ip);
		strcpy(arp_info.macaddress,mac);
		arp_info.ifindex=ifname2ifindex(dev);
		flagint = strtol(flag,&ptr,MAX_IPADDR_STR);
		flagint = (flagint <<5)>>7;
		arp_info.staticarp = flagint;
		create_entry_arp_table();	
	}

	ret = fclose(fp);
	return ret;
	
}


static int get_ip_stats()
{
	FILE* fp;
	char buf[1024];
	int ret;

	fp = fopen(proc_snmp,"r");
	while(fgets(buf,sizeof(buf),fp))
	{
		if(strncmp(buf,"Ip:",3)==0)
		{
			fgets(buf,sizeof(buf),fp);
			sscanf(buf,"Ip: %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
				&ip_stats.fwd,
				&ip_stats.ttl,
				&ip_stats.inrecv,
				&ip_stats.inhdrerr,
				&ip_stats.inaddrerr,
				&ip_stats.fwddg,
				&ip_stats.inunkprot,
				&ip_stats.indisc,
				&ip_stats.indel,
				&ip_stats.outreq,
				&ip_stats.outdisc,
				&ip_stats.outnoroute,
				&ip_stats.reasmto,
				&ip_stats.reasmreq,
				&ip_stats.reasmok,
				&ip_stats.reasmfails,
				&ip_stats.fraqok,
				&ip_stats.fragfail,
				&ip_stats.fragcreat);
		}
	}
	ret = fclose(fp);
	return ret;
}


/*
 * Function: create_entry_ip_table
 *
 * This function creates a single row in the ipAddrTable pertaining to a single interface
 */
void create_entry_ip_table()
{
	struct row_t *row;	
	struct schent_t *sch;
	struct index_t *idx;
	struct object_t *obj;
	struct in_addr ip, net, bc;

        unsigned int bcast;

	sch = schema_get(".1.3.6.1.2.1.4.20.1");
	assert(sch);

	idx = index_new(sch);
	assert(idx);

	// Index of the row is actually its IP address

	obj = index_get_obj(idx, 0);	
	assert(obj);
	memcpy(&ip, &ip_info.ipaddr.sin_addr, sizeof(ip));
	object_set_ipaddr(obj, ip);	

	row = row_create_by_index(idx);
	assert(row);
	index_done(idx);

	// NETMASK
	memcpy(&net, &ip_info.netmask.sin_addr, sizeof(net));

	// BROADCAST
	memcpy(&bc, &ip_info.broadcast.sin_addr, sizeof(bc));
	bcast = ntohl(bc.s_addr) & 0x1;

	// SET ALL VALUES
	object_set_ipaddr( row_get_obj(row, 0), ip);
	object_set_int(	   row_get_obj(row, 1), ip_info.index);
	object_set_ipaddr( row_get_obj(row, 2), net);
	object_set_int(	   row_get_obj(row, 3), bcast);
	object_set_int(	   row_get_obj(row, 4), ip_info.reasm);
}



/*
 * This function strips the ":0", etc. from the interface name - 'name' MUST be 0-terminated!
 */
char *strip_name(char *name)
{
    static char sname[MAX_IPADDR_STR] = { '\0' };
    char *r = name;
    char *p = sname;

    while (*r != ':' && *r != '\0')
    {
        *p++ = *r++;
    }
    *p = '\0';

    return sname;
}


/*
 * Function: get_ifr_info
 *
 * This function retrieves information for a single interface that is 
 * refered to by 'name'.
 */
int get_ifr_info(int sock, char *name)
{
    struct ifreq ifr;
    int err = -15;

    ip_info.index = ifname2ifindex(strip_name(name));
    ifr.ifr_addr.sa_family = AF_INET;

    // IP Address

    strncpy(ifr.ifr_name, name, MAX_IPADDR_STR);
    if (ioctl(sock, SIOCGIFADDR, &ifr) >= 0)
    {
        memcpy(&ip_info.ipaddr, &ifr.ifr_addr, sizeof(struct sockaddr_in));
        err += 8;
    }
    else
        fprintf(stderr, "Couldn't get IP address for i/f %s: %s\n", name, strerror(errno));

    // Broadcast Address
        
    if (ioctl(sock, SIOCGIFBRDADDR, &ifr) >= 0)
    {
        memcpy(&ip_info.broadcast, &ifr.ifr_broadaddr, sizeof(struct sockaddr_in));
        err += 4;
    }
    else
        fprintf(stderr, "Couldn't get BC address for i/f %s: %s\n", name, strerror(errno));

    // Interface Netmask

    if (ioctl(sock, SIOCGIFNETMASK, &ifr) >= 0)
    {
        memcpy(&ip_info.netmask, &ifr.ifr_netmask, sizeof(struct sockaddr_in));
        err += 2;
    }
    else
        fprintf(stderr, "Couldn't get netmask for i/f %s: %s\n", name, strerror(errno));

    // Maximum Transmission Unit

    if (ioctl(sock, SIOCGIFMTU, &ifr) >= 0)
    {
    	ip_info.reasm = ifr.ifr_mtu;
    	++err;
    }
    else
        fprintf(stderr, "Couldn't get MTU for i/f %s: %s\n", name, strerror(errno));

    return err;
}


/*
 * Function: get_ip_info
 *
 * This function retrieves information for the various interfaces
 */
void get_ip_info( void )
{
    int num = 30;
    int err;
    int fd;
    int n;
    struct ifconf ifc;
    struct ifreq *ifr;

    fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd < 0)
    {
    	fprintf(stderr, "Couldn't open a socket: %s\n", strerror(errno));
        return;
    }

    ifc.ifc_buf = NULL;
    for (;;) {
	ifc.ifc_len = sizeof(struct ifreq) * num;
	ifc.ifc_buf = realloc(ifc.ifc_buf, ifc.ifc_len);
	if (ifc.ifc_buf == NULL)
	{
	    fprintf(stderr, "Couldn't realloc: %s\n", strerror(errno));
	    return;
	}
	if (ioctl(fd, SIOCGIFCONF, &ifc) < 0) 
	{
	    fprintf(stderr, "Couldn't ioctl SIOCGIFCONF: %s\n", strerror(errno));
	    goto end;
	}
	if (ifc.ifc_len == sizeof(struct ifreq) * num) 
	{
	    num += 10;
	    continue;
	}
	break;
    }
    ifr = ifc.ifc_req;
    for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) 
    {
	if ((err = get_ifr_info(fd, ifr->ifr_name)) == 0)
	{
	    create_entry_ip_table();
	}
	else
	{
	    fprintf(stderr, "Error getting info for %s: %d\n", ifr->ifr_name, err);
	}
	ifr++;
    }

end:
    close(fd);
    free(ifc.ifc_buf);
    return;
}


static int get(void)
{	
	struct row_t *row, *nextrow;
	struct schent_t *sch;
	struct object_t *obj;

	sch = schema_get(".1.3.6.1.2.1.4.20.1");
	assert(sch);
	row = sch->rowlist;

	while (row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}

	get_ip_info();
	get_ip_stats();
				
	obj=object_get(".1.3.6.1.2.1.4.1.0");
	assert(obj);
	object_set_int(obj,ip_stats.fwd);					
	
	obj=object_get(".1.3.6.1.2.1.4.2.0");
	assert(obj);
	object_set_int(obj,ip_stats.ttl);					

	obj=object_get(".1.3.6.1.2.1.4.3.0");
	assert(obj);
	object_set_counter(obj,ip_stats.inrecv);	

	obj=object_get(".1.3.6.1.2.1.4.4.0");
	assert(obj);
	object_set_counter(obj,ip_stats.inhdrerr);					

	obj=object_get(".1.3.6.1.2.1.4.5.0");
	assert(obj);
	object_set_counter(obj,ip_stats.inaddrerr);					

	obj=object_get(".1.3.6.1.2.1.4.6.0");
	assert(obj);
	object_set_counter(obj,ip_stats.fwddg);					

	obj=object_get(".1.3.6.1.2.1.4.7.0");
	assert(obj);
	object_set_counter(obj,ip_stats.inunkprot);					

	obj=object_get(".1.3.6.1.2.1.4.8.0");
	assert(obj);
	object_set_counter(obj,ip_stats.indisc);					

	obj=object_get(".1.3.6.1.2.1.4.9.0");
	assert(obj);
	object_set_counter(obj,ip_stats.indel)	;				

	obj=object_get(".1.3.6.1.2.1.4.10.0");
	assert(obj);
	object_set_counter(obj,ip_stats.outreq);					

	obj=object_get(".1.3.6.1.2.1.4.11.0");
	assert(obj);
	object_set_counter(obj,ip_stats.outdisc);					

	obj=object_get(".1.3.6.1.2.1.4.12.0");
	assert(obj);
	object_set_counter(obj,ip_stats.outnoroute);;					

	obj=object_get(".1.3.6.1.2.1.4.13.0");
	assert(obj);
	object_set_int(obj,ip_stats.reasmto);					

	obj=object_get(".1.3.6.1.2.1.4.14.0");
	assert(obj);
	object_set_counter(obj,ip_stats.reasmreq);					

	obj=object_get(".1.3.6.1.2.1.4.15.0");
	assert(obj);
	object_set_counter(obj,ip_stats.reasmok);					

	obj=object_get(".1.3.6.1.2.1.4.16.0");
	assert(obj);
	object_set_counter(obj,ip_stats.reasmfails);					

	obj=object_get(".1.3.6.1.2.1.4.17.0");
	assert(obj);
	object_set_counter(obj,ip_stats.fraqok);					

	obj=object_get(".1.3.6.1.2.1.4.18.0");
	assert(obj);
	object_set_counter(obj,ip_stats.fragfail);					

	obj=object_get(".1.3.6.1.2.1.4.19.0");
	assert(obj);
	object_set_counter(obj,ip_stats.fragcreat);			
	
	
	sch = schema_get(".1.3.6.1.2.1.4.22.1");
	assert(sch);
	row = sch->rowlist;


	while (row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}


	get_arp_info();		
	
	
	obj=object_get(".1.3.6.1.2.1.4.23.0");
	assert(obj);
	object_set_counter(obj,0);			

	
	return(5);
}

void write_int(char* name, int value)
{
	FILE *fp;
	char fn[256];
	strcpy(fn,ipv4_path);
	strcat(fn,name);
	fp = fopen(fn,"w");
	fprintf(fp,"%d",value);
	fclose(fp);
}

static int set(void)
{
	struct object_t *obj;
	int ttl;
	int ipfwd;

	obj = object_get(".1.3.6.1.2.1.4.1.0");
	assert(obj);
	if(object_flag_isset(obj, OF_DIRTY)) 
	{
	 	ipfwd = object_get_int(obj);
		write_int("ip_forward",ipfwd);
 	 	object_flag_clear(obj, OF_DIRTY);
	}	
	
	
	obj = object_get(".1.3.6.1.2.1.4.2.0");
	assert(obj);
	if(object_flag_isset(obj, OF_DIRTY)) 
	{
		ttl= object_get_int(obj);
		write_int("ip_default_ttl",ttl);		
		object_flag_clear(obj, OF_DIRTY);
	}	
	return (5);
}


struct moddescr_t mod_mibII_ip = {

	name:		__FILE__,
	fn_get: 	get,
	fn_set: 	set,
	oidlist: 	".1.3.6.1.2.1.4.1 .1.3.6.1.2.1.4.2 .1.3.6.1.2.1.4.3 .1.3.6.1.2.1.4.4 .1.3.6.1.2.1.4.5 .1.3.6.1.2.1.4.6 .1.3.6.1.2.1.4.7 .1.3.6.1.2.1.4.8 .1.3.6.1.2.1.4.9 .1.3.6.1.2.1.4.10 .1.3.6.1.2.1.4.11 .1.3.6.1.2.1.4.12 .1.3.6.1.2.1.4.13 .1.3.6.1.2.1.4.14 .1.3.6.1.2.1.4.15 .1.3.6.1.2.1.4.16 .1.3.6.1.2.1.4.17 .1.3.6.1.2.1.4.18 .1.3.6.1.2.1.4.19 .1.3.6.1.2.1.4.20.1 .1.3.6.1.2.1.4.22.1 .1.3.6.1.2.1.4.23"
};


/*

The following schema entries are provided by this module :

O .1.3.6.1.2.1.4.1
L ipForwarding
T INTEGER
A RW
R 1,2

O .1.3.6.1.2.1.4.2
L ipDefaultTTL
T INTEGER
A RW
R 1/255

O .1.3.6.1.2.1.4.3
L ipInReceives
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.4
L ipInHdrErrors
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.5
L ipInAddrErrors
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.6
L ipForwDatagrams
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.7
L ipInUnknownProtos
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.8
L ipInDiscards
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.9
L ipInDelivers
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.10
L ipOutRequests
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.11
L ipOutDiscards
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.12
L ipOutNoRoutes
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.13
L ipReasmTimeout
T INTEGER
A R

O .1.3.6.1.2.1.4.14
L ipReasmReqds
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.15
L ipReasmOKs
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.16
L ipReasmFails
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.17
L ipFragOKs
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.18
L ipFragFails
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.19
L ipFragCreates
T COUNTER
A R
R 0/4294967295

O .1.3.6.1.2.1.4.20.1
L ipAddrEntry
T TABLE
I .1.3.6.1.2.1.4.20.1.1

	O .1.3.6.1.2.1.4.20.1.1
	L ipAdEntAddr
	T IPADDR
	A R
	R 4

	O .1.3.6.1.2.1.4.20.1.2
	L ipAdEntIfIndex
	T INTEGER
	A R
	R 1/2147483647

	O .1.3.6.1.2.1.4.20.1.3
	L ipAdEntNetMask
	T IPADDR
	A R
	R 4

	O .1.3.6.1.2.1.4.20.1.4
	L ipAdEntBcastAddr
	T INTEGER
	A R
	R 0/1

	O .1.3.6.1.2.1.4.20.1.5
	L ipAdEntReasmMaxSize
	T INTEGER
	A R
	R 0/65535

O .1.3.6.1.2.1.4.22.1
L ipNetToMediaEntry
T TABLE
I .1.3.6.1.2.1.4.22.1.1 .1.3.6.1.2.1.4.22.1.3

	O .1.3.6.1.2.1.4.22.1.1
	L ipNetToMediaIfIndex
	T INTEGER
	A C
	R 1/2147483647

	O .1.3.6.1.2.1.4.22.1.2
	L ipNetToMediaPhysAddress
	T OCTETSTR
	A C

	O .1.3.6.1.2.1.4.22.1.3
	L ipNetToMediaNetAddress
	T IPADDR
	A C
	R 4

	O .1.3.6.1.2.1.4.22.1.4
	L ipNetToMediaType
	T INTEGER
	A C
	R 1,2,3,4

O .1.3.6.1.2.1.4.23
L ipRoutingDiscards
T COUNTER
A R
R 0/4294967295


*/
