/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>
#include <asm/types.h>
#include <linux/netlink.h>
#include <linux/tcp.h>
#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "easyprism.h"
#include "object.h"

#define MAX_RATES 6

#define SNMP_TCPSTATE_CLOSED      1
#define SNMP_TCPSTATE_LISTEN      2
#define SNMP_TCPSTATE_SYNSENT     3
#define SNMP_TCPSTATE_SYNRECEIVED 4
#define SNMP_TCPSTATE_ESTABLISHED 5
#define SNMP_TCPSTATE_FINWAIT1    6
#define SNMP_TCPSTATE_FINWAIT2    7
#define SNMP_TCPSTATE_CLOSEWAIT   8
#define SNMP_TCPSTATE_LASTACK     9
#define SNMP_TCPSTATE_CLOSING    10
#define SNMP_TCPSTATE_TIMEWAIT   11
#define SNMP_TCPSTATE_deleteTCB  12

static struct
{
	int rtoalg;
	int rtomin;
	int rtomax;
	int maxconn;
	int actopens;
	int passopens;
	int attfails;
	int estrst;
	int currest;
	int inseq;
	int outseq;
	int retrns;
	int inerr;
	int outrst;

} tcp_stats;

static int get_tcp_stats()
{
	FILE* fp;
	char buf[1024];
	const char proc[]="/proc/net/snmp";
	int ret;
	
	fp = fopen(proc,"r");
	while(fgets(buf,sizeof(buf),fp))
	{
		if(strncmp(buf,"Tcp:",4)==0)
		{
			fgets(buf,sizeof(buf),fp);
			sscanf(buf,"Tcp: %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
				&tcp_stats.rtoalg,
				&tcp_stats.rtomin,
				&tcp_stats.rtomax,
				&tcp_stats.maxconn,
				&tcp_stats.actopens,
				&tcp_stats.passopens,
				&tcp_stats.attfails,
				&tcp_stats.estrst,
				&tcp_stats.currest,
				&tcp_stats.inseq,
				&tcp_stats.outseq,
				&tcp_stats.retrns,
				&tcp_stats.inerr,
				&tcp_stats.outrst
				);
		}
	}
	ret = fclose(fp);
	return ret;
}


void create_tcp_table()
{
	struct row_t *row;//, *nextrow;
	struct schent_t *sch;
	struct index_t *idx;
	struct object_t *obj;

	FILE* fp;
	char buf[1024];
	const char proc[]="/proc/net/tcp";

	char localip[16];
	struct in_addr localip_addr;
	char remoteip[16];
	struct in_addr remoteip_addr;
	int  state,dummy,l1,l2,l3,l4,lp,r1,r2,r3,r4,rp;
	int snmp_state;

	sch = schema_get(".1.3.6.1.2.1.6.13.1");
	assert(sch);
	


	fp = fopen(proc,"r");
	fgets(buf,sizeof(buf),fp);
	while(fgets(buf,sizeof(buf),fp))
	{
	   
		sscanf(buf,"%d: %2x%2x%2x%2x:%4x %2x%2x%2x%2x:%4x %x",&dummy,&l1,&l2,&l3,&l4,&lp,&r1,&r2,&r3,&r4,&rp,&state);
		sprintf(localip,"%d.%d.%d.%d",l4,l3,l2,l1);
		sprintf(remoteip,"%d.%d.%d.%d",r4,r3,r2,r1);
		
		inet_aton(localip, &localip_addr);
		inet_aton(remoteip, &remoteip_addr);
		
		idx= index_new(sch);
		assert(idx);
	
		obj = index_get_obj(idx, 0);	
		assert(obj);
		object_set_ipaddr(obj, localip_addr);	
	
		obj = index_get_obj(idx, 1);	
		assert(obj);
		object_set_int(obj, lp);	

		obj = index_get_obj(idx, 2);	
		assert(obj);
		object_set_ipaddr(obj, remoteip_addr);	
	
		obj = index_get_obj(idx, 3);	
		assert(obj);
		object_set_int(obj, rp);	
		
		/* check if a row with this index already exists. If so,
		   just get it and overwrite it */
		row = row_get_by_index(idx);
		if(!row) {
			row = row_create_by_index(idx);
		}
		assert(row);
		index_done(idx);
	
		switch(state) {
			case TCP_ESTABLISHED:
				snmp_state = SNMP_TCPSTATE_ESTABLISHED;
				break;
			case TCP_SYN_SENT:
				snmp_state = SNMP_TCPSTATE_SYNSENT;
				break;
			case TCP_SYN_RECV:
				snmp_state = SNMP_TCPSTATE_SYNRECEIVED;
				break;
			case TCP_FIN_WAIT1:
				snmp_state = SNMP_TCPSTATE_FINWAIT1;
				break;
			case TCP_FIN_WAIT2:
				snmp_state = SNMP_TCPSTATE_FINWAIT2;
				break;
			case TCP_TIME_WAIT:
				snmp_state = SNMP_TCPSTATE_TIMEWAIT;
				break;
			case TCP_CLOSE:
				/* unknown state for SNMP. */
				snmp_state = SNMP_TCPSTATE_CLOSEWAIT;
				break;
			case TCP_CLOSE_WAIT:
				snmp_state = SNMP_TCPSTATE_CLOSEWAIT;
				break;
			case TCP_LAST_ACK:
				snmp_state = SNMP_TCPSTATE_LASTACK;
				break;
			case TCP_LISTEN:
				snmp_state = SNMP_TCPSTATE_LISTEN;
				break;
			case TCP_CLOSING:
				snmp_state = SNMP_TCPSTATE_CLOSING;
				break;
			default:
				printf("Unknown TCP state \n");
				snmp_state = 0;
				break;
		}	
		object_set_ipaddr(row_get_obj(row, 1),localip_addr);		
		object_set_int(row_get_obj(row, 2),lp);
		object_set_ipaddr(row_get_obj(row, 3),remoteip_addr);		
		object_set_int(row_get_obj(row,4),rp);
		object_set_int(row_get_obj(row,0),snmp_state);

	}
	 fclose(fp);
}

static int get(void)
{
	struct row_t *row, *nextrow;
	struct schent_t *sch;
//	struct index_t *idx;
	struct object_t *obj;


	get_tcp_stats();

	obj=object_get(".1.3.6.1.2.1.6.1.0");
	assert(obj);
	object_set_int(obj,tcp_stats.rtoalg);
	obj=object_get(".1.3.6.1.2.1.6.2.0");
	assert(obj);
	object_set_int(obj,tcp_stats.rtomin);					
	obj=object_get(".1.3.6.1.2.1.6.3.0");
	assert(obj);
	object_set_int(obj,tcp_stats.rtomax);					
	obj=object_get(".1.3.6.1.2.1.6.4.0");
	assert(obj);
	object_set_int(obj,tcp_stats.maxconn);					
	obj=object_get(".1.3.6.1.2.1.6.5.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.actopens);					
	obj=object_get(".1.3.6.1.2.1.6.6.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.passopens);					
	obj=object_get(".1.3.6.1.2.1.6.7.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.attfails);					
	obj=object_get(".1.3.6.1.2.1.6.8.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.estrst);					
	obj=object_get(".1.3.6.1.2.1.6.9.0");
	assert(obj);
	object_set_gauge(obj,tcp_stats.currest);					
	obj=object_get(".1.3.6.1.2.1.6.10.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.inseq);					
	obj=object_get(".1.3.6.1.2.1.6.11.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.outseq);					
	obj=object_get(".1.3.6.1.2.1.6.12.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.retrns);					
	obj=object_get(".1.3.6.1.2.1.6.14.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.inerr);					
	obj=object_get(".1.3.6.1.2.1.6.15.0");
	assert(obj);
	object_set_counter(obj,tcp_stats.outrst);					

		
	sch = schema_get(".1.3.6.1.2.1.6.13.1");
	assert(sch);
	row = sch->rowlist;

	while (row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}
	
	create_tcp_table();

	return(10);
}


struct moddescr_t mod_mibII_tcp = {

	name:		__FILE__,
	fn_get:		get,
	oidlist:	".1.3.6.1.2.1.6.1 .1.3.6.1.2.1.6.2 .1.3.6.1.2.1.6.3 .1.3.6.1.2.1.6.4 .1.3.6.1.2.1.6.5 .1.3.6.1.2.1.6.6 .1.3.6.1.2.1.6.7 .1.3.6.1.2.1.6.8 .1.3.6.1.2.1.6.9 .1.3.6.1.2.1.6.10 .1.3.6.1.2.1.6.11 .1.3.6.1.2.1.6.12 .1.3.6.1.2.1.6.13.1 .1.3.6.1.2.1.6.14 .1.3.6.1.2.1.6.15"

};






