/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>
#include <asm/types.h>
#include <linux/netlink.h>
#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "easyprism.h"
#include "object.h"

#define MAX_RATES 6

#ifndef MAX_IPADDR_STR
#define MAX_IPADDR_STR 16
#endif

static struct
{
	int indg;
	int noports;
	int inerr;
	int outdg;
} udp_stats;

static int get_udp_stats()
{
	FILE* fp;
	int ret;
	char buf[1024];
	const char proc_snmp[]="/proc/net/snmp";

	fp = fopen(proc_snmp,"r");
	while(fgets(buf,sizeof(buf),fp))
	{
		if(strncmp(buf,"Udp:",4)==0)
		{
			fgets(buf,sizeof(buf),fp);
			sscanf(buf,"Udp: %d %d %d %d",
				&udp_stats.indg,
				&udp_stats.noports,
				&udp_stats.inerr,
				&udp_stats.outdg
				);
		}
	}
	ret =fclose(fp);
	return ret;
}

void create_udp_table()
{
	struct row_t *row;//, *nextrow;
	struct schent_t *sch;
	struct index_t *idx;
	struct object_t *obj;
	const char proc_udp[]="/proc/net/udp";
	
	FILE* fp;
	char buf[1024];

	char localip[MAX_IPADDR_STR];
	struct in_addr localip_addr;
	int  dummy,l1,l2,l3,l4,lp;

	sch = schema_get(".1.3.6.1.2.1.7.5.1");
	assert(sch);

	fp = fopen(proc_udp,"r");
	fgets(buf,sizeof(buf),fp);
	while(fgets(buf,sizeof(buf),fp))
	{
	   
		sscanf(buf,"%d: %2x%2x%2x%2x:%4x",&dummy,&l1,&l2,&l3,&l4,&lp);
		sprintf(localip,"%d.%d.%d.%d",l1,l2,l3,l4);
		inet_aton(localip, &localip_addr);

		idx= index_new(sch);
		assert(idx);
	
		obj = index_get_obj(idx, 0);	
		assert(obj);
		object_set_ipaddr(obj, localip_addr);	
	
		obj = index_get_obj(idx, 1);	
		assert(obj);
		object_set_int(obj, lp);	

		/* Only create the row if it doesn't exist. Linux can have more
		 * than one UDP port/ip address combination in its udp table
		 */
		if((row = row_get_by_index(idx)) == NULL) {
			row = row_create_by_index(idx);
			assert(row);
		}
		index_done(idx);

	
		object_set_ipaddr(row_get_obj(row, 0),localip_addr);		
		object_set_int(row_get_obj(row, 1),lp);
		
		
	}
	 fclose(fp);
}


static int get(void)
{
	struct row_t *row, *nextrow;
	struct schent_t *sch;
	struct object_t *obj;

	get_udp_stats();

	obj=object_get(".1.3.6.1.2.1.7.1.0");
	assert(obj);
	object_set_counter(obj,udp_stats.indg);
	obj=object_get(".1.3.6.1.2.1.7.2.0");
	assert(obj);
	object_set_counter(obj,udp_stats.noports);					
	obj=object_get(".1.3.6.1.2.1.7.3.0");
	assert(obj);
	object_set_counter(obj,udp_stats.inerr);					
	obj=object_get(".1.3.6.1.2.1.7.4.0");
	assert(obj);
	object_set_counter(obj,udp_stats.outdg);

	
	sch = schema_get(".1.3.6.1.2.1.7.5.1");
	assert(sch);
	row = sch->rowlist;

	while (row) {
		nextrow = row->next;
		row_delete(row);
		row = nextrow;
	}

	create_udp_table();

	return(10);
}


struct moddescr_t mod_mibII_udp = {

	name:		__FILE__,
	fn_get:		get,
	oidlist:	".1.3.6.1.2.1.7.1 .1.3.6.1.2.1.7.2 .1.3.6.1.2.1.7.3 .1.3.6.1.2.1.7.4 .1.3.6.1.2.1.7.5.1"

};





