/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>
#include <aplog.h>
#include "mod_mibII_interfaces.h"
#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "easyprism.h"
#include "hardconfig.h"
#include "cfg_file_lib.h"
#include "lib_calls.h"
#include "debug.h"
#include "handler.h"

#define CFGNAME_IFACE "wss"
#define CFGNAME_RADIUS "radius"
#define PAED_PID_FILE "/var/run/paed.pid"
#define MAX_FP 10


static int set_wss(void)
{
	struct row_t* row_ul_auth;
	struct row_t* row_dot1x;
	struct row_t* row_smt;
	struct row_t* row_priv;
	struct row_t* row_key;
	struct schent_t* sch_ul_auth;
	struct schent_t* sch_dot1x;
	struct schent_t* sch_smt;
	struct schent_t* sch_priv;
	struct schent_t* sch_key;
	struct object_t *obj;
	/* In dot1x table */
	unsigned int reauthtime;
	unsigned int rekeying;
	unsigned int rekeyingtime;
	unsigned int rekeyingpackets;
	unsigned int grouprekeying;
	unsigned int grouprekeyingtime;
	unsigned int grouprekeyingpackets;
	unsigned int groupkeystrictupdate;
	unsigned int usepreshared;
	struct string_t* pskvaluep;
	struct string_t* passphrasepskvaluep;
	/* In ul_auth table */
	unsigned int ACLEnabled;
	unsigned int Dot1XEnabled;
	unsigned int WPAEnabled;
	unsigned int WEPEnabled;
	int keyid;
	unsigned int WEPKeySize[MAX_FP][4];
	int KeySize;
	unsigned int Dot1XWEPKeySize;
	/* In smt table */

	struct string_t* dot11DesiredSSIDp;
	/* In priv table */
	unsigned int dot11PrivacyInvoked;
	unsigned int dot11ExcludeUnencrypted;
	unsigned int dot11WEPDefaultKeyID[MAX_FP];
	/* In key table */
	int i;
	struct index_t* index;
	struct string_t* keyp;

	int interfaceindex;
	char interfacename[256];
	char filename[256];
	FILE* fp[MAX_FP];

	dbg(DBG_MODULE, "Start module mod_paed_cfg");

	for(i=0;i<MAX_FP;i++)
	{
	    	WEPKeySize[i][0] = 0;
	    	WEPKeySize[i][1] = 0;
	    	WEPKeySize[i][2] = 0;
	    	WEPKeySize[i][3] = 0;
	    	dot11WEPDefaultKeyID[i] = 0;
		fp[i] = NULL;
    	}

	/* Initialise all schemas */
	sch_ul_auth = schema_get(".1.3.6.1.4.1.937.2.2.3.8.1");
	assert(sch_ul_auth);
	sch_dot1x   = schema_get(".1.3.6.1.4.1.937.2.2.3.9.1");
	assert(sch_dot1x);
	sch_smt	    = schema_get(".1.2.840.10036.1.1.1");
	assert(sch_smt);
	sch_priv    = schema_get(".1.2.840.10036.1.5.1");
	assert(sch_priv);
	sch_key	    = schema_get(".1.2.840.10036.1.3.1");
	assert(sch_key);

	/* Initialise all rows */
	row_ul_auth = sch_ul_auth->rowlist;
	row_dot1x   = sch_dot1x->rowlist;
	row_smt     = sch_smt->rowlist;
	row_priv    = sch_priv->rowlist;
	row_key     = sch_key->rowlist;
	/* Create for each enabled interface a config file */

	while(row_smt)
	{
		index = index_get(row_smt);
		interfaceindex = object_get_int(index_get_obj(index, 0));
		if (interfaceindex >= MAX_FP)
		{
		    dbg(DBG_MODULE, "Found interfce index: %d not so nice!", interfaceindex);
   		    index_done(index);
		    row_smt = row_smt->next;
		    continue;
		}
		index_done(index);

		strncpy(interfacename,ifindex2ifname(interfaceindex),256);
	    	sprintf(filename, "%s.%s", CFGNAME_IFACE, interfacename);
	    	fp[interfaceindex] = create_cfg_file(filename);
		dbg(DBG_MODULE, "Opened config file: %s for interface: %s with index: %d\n", filename, interfacename, interfaceindex);

		row_smt = row_smt->next;
	}

	/* Parse entries from dot11_smt table */

	row_smt     = sch_smt->rowlist;
	while(row_smt)
	{

		index = index_get(row_smt);
		interfaceindex = object_get_int(index_get_obj(index, 0));
		if (interfaceindex >= MAX_FP || fp[interfaceindex] == NULL)
		{
			dbg(DBG_MODULE, "Found interfce index: %d not so nice!", interfaceindex);
			index_done(index);
		        row_smt = row_smt->next;
		        continue;
		}
		index_done(index);

		obj = row_get_obj(row_smt, 8);
		dot11DesiredSSIDp = object_get_string(obj);
		write_cfg_line_str(fp[interfaceindex], "dot11DesiredSSID", dot11DesiredSSIDp->data);

		row_smt = row_smt->next;
    	}

	/* Parse entries from dot11_priv table */
	while(row_priv)
	{
		index = index_get(row_priv);
		interfaceindex = object_get_int(index_get_obj(index, 0));
		if (interfaceindex >= MAX_FP || fp[interfaceindex] == NULL)
		{
			dbg(DBG_MODULE, "Found interfce index: %d not so nice!", interfaceindex);
			index_done(index);
		        row_priv = row_priv->next;
		        continue;
		}
		index_done(index);

		obj = row_get_obj(row_priv, 0);
		dot11PrivacyInvoked = object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "dot11PrivacyInvoked", dot11PrivacyInvoked);
		obj = row_get_obj(row_priv, 1);
		dot11WEPDefaultKeyID[interfaceindex] = object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "dot11WEPDefaultKeyID", dot11WEPDefaultKeyID[interfaceindex] );
		obj = row_get_obj(row_priv, 3);
		dot11ExcludeUnencrypted = object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "dot11ExcludeUnencrypted", dot11ExcludeUnencrypted);
		row_priv = row_priv->next;
    	}

	/* Parse entries from dot11_priv table */
	while(row_key)
	{
		index = index_get(row_key);
		interfaceindex = object_get_int(index_get_obj(index, 0));
		if (interfaceindex >= MAX_FP || fp[interfaceindex] == NULL)
		{
			dbg(DBG_MODULE, "Found interfce index: %d not so nice!", interfaceindex);
			index_done(index);
			row_key = row_key->next;
			continue;
		}
		keyid = object_get_int(index_get_obj(index, 1));
		keyid -= 1;
		index_done(index);

		obj = row_get_obj(row_key, 1);
		keyp = object_get_string(obj);

		/* Copy keylen in array of wepkeys for legacy wep key size calculations */
		WEPKeySize[interfaceindex][keyid] = keyp->len;

		switch (keyid)
		{
		    case 0:
			write_cfg_line_hex(fp[interfaceindex], "dot11WEPDefaultKey1", keyp);
		    break;
		    case 1:
			write_cfg_line_hex(fp[interfaceindex], "dot11WEPDefaultKey2", keyp);
		    break;
		    case 2:
			write_cfg_line_hex(fp[interfaceindex], "dot11WEPDefaultKey3", keyp);
		    break;
		    case 3:
			write_cfg_line_hex(fp[interfaceindex], "dot11WEPDefaultKey4", keyp);
		    break;
		}

		row_key = row_key->next;
    	}

	/* Parse entries from dot1x table */
	while(row_dot1x)
	{
		obj = row_get_obj(row_dot1x, 0);
		assert(obj);
		interfaceindex = object_get_int(obj);

		if (interfaceindex >= MAX_FP || fp[interfaceindex] == NULL)
		{
			dbg(DBG_MODULE, "Found interfce index: %d not so nice!", interfaceindex);
		    	row_dot1x = row_dot1x->next;
			continue;
		}

		obj = row_get_obj(row_dot1x, 1);
		reauthtime= object_get_uint(obj);
		write_cfg_line_int(fp[interfaceindex], "ReauthenticationTime", reauthtime);

		obj = row_get_obj(row_dot1x, 2);
		rekeying= object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "Dot1XRekeying",rekeying );

		obj = row_get_obj(row_dot1x, 3);
		rekeyingtime= object_get_uint(obj);
		write_cfg_line_int(fp[interfaceindex], "Dot1XRekeyingTime",rekeyingtime );

		obj = row_get_obj(row_dot1x, 4);
		rekeyingpackets= object_get_uint(obj);
		write_cfg_line_int(fp[interfaceindex], "Dot1XRekeyingPackets",rekeyingpackets );

		obj = row_get_obj(row_dot1x, 5);
		grouprekeying= object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "WPAGroupRekeying", grouprekeying);

		obj = row_get_obj(row_dot1x, 6);
		grouprekeyingtime= object_get_uint(obj);
		write_cfg_line_int(fp[interfaceindex], "WPAGroupRekeyingTime",grouprekeyingtime );

		obj = row_get_obj(row_dot1x, 7);
		grouprekeyingpackets= object_get_uint(obj);
		write_cfg_line_int(fp[interfaceindex], "WPAGroupRekeyingPackets", grouprekeyingpackets);

		obj = row_get_obj(row_dot1x, 8);
		groupkeystrictupdate= object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "WPAGroupKeyStrictUpdate", groupkeystrictupdate);

		obj = row_get_obj(row_dot1x, 9);
		usepreshared= object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "WPAUsePreSharedKey", usepreshared);

		obj = row_get_obj(row_dot1x, 10);
		pskvaluep = object_get_string(obj);
		write_cfg_line_hex(fp[interfaceindex], "WPAConfigPSKValue", pskvaluep);

		obj = row_get_obj(row_dot1x, 11);
		passphrasepskvaluep = object_get_string(obj);
		write_cfg_line_str(fp[interfaceindex], "WPAConfigPSKPassPhrase", passphrasepskvaluep->data);

		row_dot1x = row_dot1x->next;
	}

	/* Parse entries from dot1x table */
	while(row_ul_auth)
	{
		obj = row_get_obj(row_ul_auth, 0);
		assert(obj);
		interfaceindex = object_get_int(obj);
		if (interfaceindex >= MAX_FP || fp[interfaceindex] == NULL)
		{
			dbg(DBG_MODULE, "Found interfce index: %d not so nice!", interfaceindex);
		    	row_ul_auth = row_ul_auth->next;
			continue;
	        }

		obj = row_get_obj(row_ul_auth, 1);
		ACLEnabled = object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "ACLEnabled", ACLEnabled);

		obj = row_get_obj(row_ul_auth, 2);
		Dot1XEnabled = object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "Dot1XEnabled", Dot1XEnabled);

		obj = row_get_obj(row_ul_auth, 3);
		WPAEnabled = object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "WPAEnabled", WPAEnabled);

		obj = row_get_obj(row_ul_auth, 4);
		WEPEnabled = object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "WEPEnabled", WEPEnabled);

		/* Calculate the size ourselfs. The oid for this could still be out of date */
		KeySize = WEPKeySize[interfaceindex][dot11WEPDefaultKeyID[interfaceindex]];
		write_cfg_line_int(fp[interfaceindex], "WEPKeySize", KeySize);

		obj = row_get_obj(row_ul_auth, 6);
		Dot1XWEPKeySize = object_get_int(obj);
		write_cfg_line_int(fp[interfaceindex], "Dot1XWEPKeySize", Dot1XWEPKeySize);

		row_ul_auth = row_ul_auth->next;
	}

	dbg(DBG_MODULE, "Going to close all open config files");

	/* Close all open files */
	for (i = 0 ; i < MAX_FP; i++)
	{
		if(fp[i] != NULL) {
		    	dbg(DBG_MODULE, "Close file with index %d", i);

			close_cfg_file(fp[i]);
		}
    	}

	return 0;

}

int restart_paed (void)
{
//	StartAppl("/etc/init.d/paed", "stop", NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	StartAppl("/etc/init.d/wep", "eth1", NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	return R_NOERR;
}

static int init(void) {
	struct handler_t* hnd;

	/* register an update handler on the radiusservers OID */
	hnd = handler_add(H_SET, restart_paed);
	handler_connect(hnd, ".1.3.6.1.4.1.937.2.1.4.2.1");

	return R_NOERR;
}

static int set(void)
{
	/* Set wireless security settings */
	dbg(DBG_MODULE, "calling set_wss()");
	set_wss();

	/* ReStart paed */
	restart_paed();
	return R_NOERR;
}


struct moddescr_t mod_paed_cfg = {
	name:		__FILE__,
	fn_set:		set,
	fn_ini:		init,
	oidlist:	".1.3.6.1.4.1.937.2.2.3.8.1 .1.3.6.1.4.1.937.2.2.3.9.1 .1.2.840.10036.1.1.1 .1.2.840.10036.1.5.1 .1.2.840.10036.1.3.1"

};




