/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <aplog_defs.h>
#include "mod_mibII_interfaces.h"
#include "table.h"
#include "misc.h"
#include "cfg_file_lib.h"
#include "objdata.h"
#include "easyprism.h"
#include "hardconfig.h"
#include "object.h"

const char ulauth_name[]="/var/etc/ulauth.conf";

static int get(void)
{
	struct row_t* row;
	struct schent_t* sch;
	struct index_t *index;
	int ifnum = 0;
	char oidbuf[MAX_OID_LENGTH];
	int keyid;
	struct string_t* wepkey;
	struct object_t* obj;
	
	
	/* check if regdomain is supported */
	sch = schema_get(".1.3.6.1.4.1.937.2.2.3.8.1");
	assert(sch);

	row = sch->rowlist;
	while(row) 
	{
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
		index_done(index);
		
		/****************************************/
		/* Determine the ulAuthLegacyWEPKeySize */
		/****************************************/
		
		/* get the dot11WEPDefaultKeyID for this interface */
		snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.5.1.2.%d", ifnum);
		obj = object_get(oidbuf);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		keyid = object_get_int(obj);
		
		/* Get the corresponding WEP key object */
		snprintf(oidbuf, sizeof(oidbuf), ".1.2.840.10036.1.3.1.2.%d.%d", ifnum, keyid + 1);
		obj = object_get(oidbuf);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		wepkey = object_get_string(obj);
		
		/* get the ulAuthLegacyWEPKeySize object */
		obj = row_get_obj(row, 5);
		assert(obj);
		object_set_int(obj, wepkey->len);
		
		row = row->next;
	}
	
	return 1;
	
}

static int set(void)
{
	struct row_t* row;
	struct schent_t* sch;
	struct object_t *obj;
	int aclen;
	int dot1xen;
	int ssnen;
	int legacy;
	FILE* fd;

	sch = schema_get(".1.3.6.1.4.1.937.2.2.3.8.1");
	assert(sch);
	row = sch->rowlist;
	assert(row);

	fd = fopen(ulauth_name,"w");
	fprintf(fd,"#ulauth\n");
	while(row) 
	{
		obj = row_get_obj(row, 1);
		aclen= object_get_int(obj);
		fprintf(fd,"dot11Ext2AuthACLEnabled %d\n",aclen);
		
		obj = row_get_obj(row, 2);
		dot1xen= object_get_int(obj);
		fprintf(fd,"dot11Ext2AuthDot1XEnabled %d\n", dot1xen);
		
		obj = row_get_obj(row, 3);
		ssnen= object_get_int(obj);
		fprintf(fd,"dot11Ext2AuthSSNEnabled %d\n", ssnen);

		obj = row_get_obj(row, 4);
		legacy= object_get_int(obj);
		fprintf(fd,"dot11Ext2AuthLegacyWEPEnabled %d\n", legacy);

		row = row->next;
	}
	
	fclose(fd);

	return 0;
	
}

struct moddescr_t mod_ulauth = {
	name:		__FILE__,
	fn_get:		get,
	fn_set:		set,
	oidlist:	".1.3.6.1.4.1.937.2.2.3.8.1"

};


/* 
O .1.3.6.1.4.1.937.2.2.3.8.1
L ulAuthEntry
T TABLE
I .1.3.6.1.4.1.937.2.2.3.8.1.1

0	O .1.3.6.1.4.1.937.2.2.3.8.1.1
	L ulAuthIfIndex
	T INTEGER
	A -
	R 0/2147483647

1	O .1.3.6.1.4.1.937.2.2.3.8.1.2
	L ulAuthACLEnabled
	T INTEGER
	A RW
	R 1,2
	D 2

2	O .1.3.6.1.4.1.937.2.2.3.8.1.3
	L ulAuthDot1XEnabled
	T INTEGER
	A RW
	R 1,2
	D 2

3	O .1.3.6.1.4.1.937.2.2.3.8.1.4
	L ulAuthSSNEnabled
	T INTEGER
	A RW
	R 1,2
	D 2

4	O .1.3.6.1.4.1.937.2.2.3.8.1.5
	L ulAuthLegacyWEPEnabled
	T INTEGER
	A RW
	R 1,2
	D 2

5	O .1.3.6.1.4.1.937.2.2.3.8.1.6
	L ulAuthLegacyWEPKeySize
	T INTEGER
	A R

6	O .1.3.6.1.4.1.937.2.2.3.8.1.7
	L ulAuthDot1XWEPKeySize
	T INTEGER
	A RW
	R 5,13
	
*/

