/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */

#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/un.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include <aplog_defs.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "table.h"
#include "misc.h"
#include "objdata.h"
#include "objflag.h"
#include "easyprism.h"
#include "cfg_file_lib.h"
#include "linux_mvc_oids.h"
#include <prismoids.h>
#include <device_config.h>
#include <stdio.h>
#include <unistd.h>
#include "mod_mibII_interfaces.h"
#include <aplog.h>
#include "object.h"

#include <net/if.h>
#include <sys/signal.h>

// #define DEBUG_MAC_TABLE 

#define log(a,args...) log_entry(a, "[" ## __FUNCTION__ ## "] " args);

struct wds_link_rec {
    int  stat;
    char mac[6];
    char dev_name[32];
    char brif_name[32];
};

struct wds_link_msg {
    unsigned char mac[6];
    char name[32];
} __attribute__ ((packed));

static struct wds_link_rec wds_link_cache[MAX_WDS_LINKS];
static unsigned int wds_free_index = 0;

static void wds_cache_announce( char *ifname, struct string_t *mac )
{
    unsigned int i;
    struct wds_link_rec *ip; 

#ifdef DEBUG_MAC_TABLE
    fprintf (stderr, "wds_cache_announce(%s, %2x.%2x.%2x.%2x.%2x.%2x)\n", 
             ifname, mac->data[0], mac->data[1], mac->data[2], mac->data[3], mac->data[4], mac->data[5] ); 
#endif
    for ( i = 0, ip = wds_link_cache; (i < wds_free_index) && (i < MAX_WDS_LINKS); i++, ip++ ) {
        if ( ( strncmp ( ip->dev_name, ifname, 32 ) == 0 ) && 
             ( memcmp  ( ip->mac, mac->data, 6 ) == 0 ) ) {
            break;
        }
    }

    if ( i < wds_free_index ) {
        ip->stat = 0;
        return;
    }
    
    if ( (i == wds_free_index) && (i < MAX_WDS_LINKS) ) {
        ip->stat = 1;
        strncpy( ip->dev_name, ifname, 32);
        memcpy ( ip->mac, mac->data, 6 );
        wds_free_index++;
    }

    return;
}

static void set_ifindex(struct wds_link_rec* ip) {
	
	struct schent_t* sch;
	struct row_t* row;
	struct object_t* obj;
	struct index_t* index;
	int ifnum;
	char* ifname;
	struct string_t* mac;
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.3.7.1");
	
	row = sch->rowlist;
	
	while(row) {
		
		/* only process rows with a rowstatus of 1 */
		obj = row_get_obj(row, 3) ;
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);

		if(object_get_rowstat(obj) == 1) {
			
			index = index_get(row);
			ifnum = object_get_int(index_get_obj(index, 0));
			ifname= ifindex2ifname(ifnum);
			index_done(index);

			obj = row_get_obj(row, 1);
			assert(obj);
			assert(obj->flags & OF_DATAPRESENT);
			mac = object_get_string(obj);
			
			if((memcmp(mac->data, ip->mac, 6) == 0) &&
					(strncmp(ifname, ip->dev_name, 32) == 0)) {
				obj = row_get_obj(row, 2);
				assert(obj);
				if(ip->brif_name[0] != 0) {
					object_set_int(obj, ifname2ifindex(ip->brif_name));
				}
				else {
					/* Use a known interface */
					object_set_int(obj, 3);
				}
			}
		}
		
		row = row->next;
	}
}

static void wds_cache_execute( void )
{
    unsigned int i, j;
    struct wds_link_rec *ip, *jp; 
    struct wds_link_msg wlm;
    char cmd[64];
    char oidbuf[MAX_OID_LENGTH];
    int r;
    struct object_t* obj;

#ifdef DEBUG_MAC_TABLE
    fprintf (stderr, "wds_cache_execute()\n"); 
#endif
    
    if ( wds_free_index > MAX_WDS_LINKS) {   /* consistency check */
        wds_free_index = MAX_WDS_LINKS;
    }

    /* send update to driver */
    for ( i = 0, ip = wds_link_cache; (i < wds_free_index) && (i < MAX_WDS_LINKS); i++, ip++ ) {
        switch ( ip->stat ) {
        case 1:
            memcpy( wlm.mac, ip->mac, 6 );
            wlm.name[0] = 0;
            r = set_prismoid(ip->dev_name, DOT11_OID_WDSLINKADD, &wlm, sizeof (struct wds_link_msg)); 
            if ( r == 0 ) {
                strncpy(ip->brif_name, wlm.name, 32 );
                snprintf( cmd, 64, "brctl addif br0 %s", ip->brif_name );
                system( cmd ); 
        	    /* update the corresponding ifindex in the wdsMac table */
        	    set_ifindex(ip);
        	    /* switch the AdminState in the ifTable for this interface to 'up' */
        	    snprintf(oidbuf, sizeof(oidbuf), ".1.3.6.1.2.1.2.2.1.7.%i", ifname2ifindex(ip->brif_name));
        	    obj = object_get(oidbuf);
        	    assert(obj);
        	    assert(obj->flags & OF_DATAPRESENT);
        	    object_set_int(obj, 1);
            }
	    
#ifdef DEBUG_MAC_TABLE
            fprintf (stderr, "set_prismoid(DOT11_OID_WDSLINKADD, %2x.%2x.%2x.%2x.%2x.%2x, %s, %d)\n",
                     ip->mac[0], ip->mac[1], ip->mac[2], ip->mac[3], ip->mac[4], ip->mac[5], ip->brif_name, r ); 
#endif
            break;
        default:
        case 0:
            /* do nothing */
            break;
        case -1:
            snprintf( cmd, 64, "brctl delif br0 %s", ip->brif_name );
            system( cmd ); 

#ifdef DEBUG_MAC_TABLE
            fprintf (stderr, "Remove WDS link: %s\n", cmd );
#endif
            
            memcpy(wlm.mac, ip->mac, 6);
            set_prismoid(ip->dev_name, DOT11_OID_WDSLINKREMOVE, &wlm, sizeof (struct wds_link_msg)); 

#ifdef DEBUG_MAC_TABLE
            fprintf (stderr, "set_prismoid(DOT11_OID_WDSLINKREMOVE, %2x.%2x.%2x.%2x.%2x.%2x, %s)\n",
                     ip->mac[0], ip->mac[1], ip->mac[2], ip->mac[3], ip->mac[4], ip->mac[5], ip->brif_name ); 
#endif
            break;
        }
    }

    /* purge cache */
    for ( j = wds_free_index, jp = &(wds_link_cache[wds_free_index-1]); (jp->stat == -1) && (j > 0); j--, jp-- ) {
        memset ( jp, 0, sizeof(struct wds_link_rec) );
        wds_free_index--;
    }

    for ( i = 0, ip = wds_link_cache; (i < wds_free_index) && (i < MAX_WDS_LINKS); i++, ip++ ) {
        if (ip->stat == -1) {
            if ( i < ( wds_free_index-1 ) ) {
                memcpy ( ip, &(wds_link_cache[wds_free_index-1]), sizeof(struct wds_link_rec) );
            }
            memset ( &(wds_link_cache[wds_free_index-1]), 0, sizeof(struct wds_link_rec) );
            wds_free_index--;

            /* remove tail records */
            for ( j = wds_free_index, jp = &(wds_link_cache[wds_free_index-1]); (jp->stat == -1) && (j > 0); j--, jp-- ) {
                memset ( jp, 0, sizeof(struct wds_link_rec) );
                wds_free_index--;
            }
        }
    }

    /* prepare next cycle */
    for ( i = 0, ip = wds_link_cache; (i < wds_free_index) && (i < MAX_WDS_LINKS); i++, ip++ ) {
        ip->stat = -1;
    }
}


static int set(void)
{
	struct object_t *obj;
	struct schent_t *sch;
	struct row_t *row;
	int rowstat,ifnum;
	struct index_t *index;
	char *ifname;
	struct string_t *wds_mac;
	
	/* The reservation table */
	sch = schema_get(".1.3.6.1.4.1.937.2.2.3.7.1");
	assert(sch);

	row = sch->rowlist;
	while(row) 
	{
		index = index_get(row);
		ifnum = object_get_int(index_get_obj(index, 0));
		ifname= ifindex2ifname(ifnum);
		index_done(index);
		
		rowstat = object_get_rowstat(row_get_obj(row, 3));
		switch(rowstat)	{
			case 1:
				obj = row_get_obj(row, 1);
				wds_mac=object_get_string(obj);
				assert(wds_mac);
				wds_cache_announce(ifname, wds_mac);
				if(object_flag_isset(obj,OF_DIRTY))
				{
					object_flag_clear(obj, OF_DIRTY);
				}
				
				break;
			case 2:
			case 6:
				obj = row_get_obj(row, 1);
				wds_mac=object_get_string(obj);	
				assert(wds_mac);
				if(object_flag_isset(obj,OF_DIRTY))
				{
					object_flag_clear(obj, OF_DIRTY);
				}
				break;
			default:
				break;			
		}

		row = row->next;
	}

    wds_cache_execute();

	return(0);
}

static int check(void) {
	
	struct schent_t* sch;
	struct row_t* row;
	struct index_t* idx;
	struct object_t* obj;
	int ifindex;
	
	int count3 = 0;
	int count5 = 0;
	
	sch = schema_get(".1.3.6.1.4.1.937.2.2.3.7.1");
	assert(sch);
	
		
	/* Count the number off WDS link per interface */
	row = sch->rowlist;
	while(row) {
		idx = index_get(row);
		obj = index_get_obj(idx, 0);
		assert(obj);
		assert(obj->flags & OF_DATAPRESENT);
		ifindex = object_get_int(obj);
		index_done(idx);
		
		switch(ifindex) {
			case 3:
				count3++;
				if(count3 > 8) 
					return R_CHECKFAILED;
				break;
			case 5:
				count5++;
				if(count5 > 8)
					return R_CHECKFAILED;
				break;
			default:
				/* wrong ifindex */
				return R_CHECKFAILED;
				break;
		}			
		
		/* TODO REMOVE THIS UGLY HACK */
		
		/* Be sure the WFS can always retrieve an object 
		 * This should be done by adding a MIBException mechanism
		 * to the toolchain (smidump)
		 */
		obj = row_get_obj(row, 2);
		if(!(obj->flags & OF_DATAPRESENT)) {
			object_set_int(obj, 3);
		}
		
		row = row->next;
	}
	
	return R_NOERR;
}


struct moddescr_t mod_wds_MACTable = {

	name:		__FILE__,
	fn_set:		set,
	fn_chk:		check,
	oidlist:	".1.3.6.1.4.1.937.2.2.3.7.1"

};


 /*
 O .1.3.6.1.4.1.937.2.2.3.7.1
L wdsMACEntry
T TABLE
I .1.3.6.1.2.1.2.2.1.1 .1.3.6.1.4.1.937.2.2.3.7.1.1

0	O .1.3.6.1.4.1.937.2.2.3.7.1.1
	L wdsMACIndex
	T INTEGER
	A -
	R 1/256

1	O .1.3.6.1.4.1.937.2.2.3.7.1.2
	L wdsMACAddress
	T OCTETSTR
	A C
	R 6

2	O .1.3.6.1.4.1.937.2.2.3.7.1.3
	L wdsIfIndex
	T INTEGER
	A R
	R 1/2147483647

3	O .1.3.6.1.4.1.937.2.2.3.7.1.4
	L wdsMACRowStatus
	T ROWSTAT
	A C
	R 1,2,3,4,5,6
	
*/

