/*
 *      $Header$
 *
 *      IMPORTANT--READ CAREFULLY:
 *      Right, including without limitation, use, reproduction, modification,
 *      and creation of derivative works, pertaining to this Source Code
 *      ("Document") are governed by a legal Agreement ("Agreement") between
 *      you (either an individual or a single entity) and Intersil Americas
 *      Inc ("Intersil") and/or its licensee (both hereinafter called
 *      Licensor).
 *      By reproducing, modifying, creating derivative works, accepting,
 *      installing, viewing, accessing, transforming or otherwise using this
 *      Document, you agree to be bound by the terms of the Agreement. If you
 *      do not agree to the terms of the Agreement, Licensor is unwilling to
 *      license rights to this Document to you. In such event, you may not
 *      use, reproduce, modify, create derivative works, install, view,
 *      access, transport, transform or circulate this Document, and you
 *      should promptly contact Licensor from which you acquired this Document
 *      for instructions on returning or destroying this Document.
 *
 *      CONFIDENTIAL, PROPRIETARY AND TRADE SECRET INFORMATION:
 *      This Document contains confidential, proprietary and trade secret
 *      information of Intersil Americas Inc. and should be treated in a
 *      confidential manner.
 *      Furthermore, this Document is protected by national copyright laws and
 *      international treaties.
 *      Unauthorized use, reproduction, modification, creation of derivative
 *      works or circulation is strictly prohibited.
 *
 *      Intersil  is a registered trademark of Intersil Americas Inc.
 *
 *      Copyright  2002 Intersil Americas Inc. All Rights Reserved.
 */
#ifndef _DATASTORE_REGULATORY_H_
#define _DATASTORE_REGULATORY_H_

#define REGULATORY_TYPE_COUNTRY              1 
#define REGULATORY_TYPE_DOMAIN               2

#define REGULATORY_BAND_2DOT4		     1
#define REGULATORY_BAND_5		     2

/* maxima */
#define REGULATORY_MAX_COUNTRIES_IN_DOMAIN  25
#define REGULATORY_MAX_FREQS_IN_GROUP      100

#define REGULATORY_CONF_FILE "/etc/regulatory.conf"

struct power_group_t {
	unsigned int max_power;
	unsigned int no_freqs;
	unsigned int* frequencies;
	struct power_group_t* next;
};

struct regulatory_group_t {
	unsigned int default_freq2; /* default frequency for 2.4Ghz band */
	unsigned int default_freq5; /* default frequence for 5Ghz band */
	unsigned char eirp; 
	struct power_group_t* power_groups;
};

struct regulatory_group_list_t {
	struct regulatory_group_t* regul_grp;
	struct regulatory_group_list_t* next;
};

struct country_domain_t {
	char* id;
	unsigned char type; /* reg_domain = 1, country = 2; */
	struct regulatory_group_list_t* regul_groups;
	int no_countries;
	char** countries;
	char* default_country;
	struct country_domain_t* next;
};

int regulatory_freq2band(int freq);
struct country_domain_t *regulatory_get_country(char *id);
struct country_domain_t *regulatory_get_domain(char *id);
struct country_domain_t *regulatory_get_country_domain_by_country_code(char* id);
struct country_domain_t *regulatory_get_country_domains(void);
struct country_domain_t *regulatory_get_default_country_domain(void);
struct country_domain_t* regulatory_get_current_country_domain(int ifnum);
void regulatory_set_current_country_domain(int ifnum, struct country_domain_t* cd);

void regulatory_init(void);
void regulatory_dump(void);

int regulatory_domainid2value(char* id);
char* regulatory_value2domainid(int value);

int          regulatory_freq2channel(unsigned int freq);
unsigned int regulatory_channel2freq(int channel, int band);

int	      regulatory_country_in_domain(struct country_domain_t* cd, char* country);
double mW2dBm(int milliwats);

#endif
