/*
 *  apdpmod_mt.c
 *  sysd
 *
 *  Created by Max Yin on Tue Sep 16 2003.
 *  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
 *
 */

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <net/if.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "prismoids.h"
#include "datastore_if.h"
#include "apdpmod_mt.h"
#define DEBUG

int IfIoctl(int ioctl_num, struct ifreq *ifr);

static int mt_check(struct BMP_buf *buf)
{
    struct mt *mtreq = (struct mt*)buf;

    // check signature
    if(strncmp(mtreq->mt_signature, MT_SIGNATURE, strlen(MT_SIGNATURE))){
#ifdef DEBUG
	printf("%s(): MT signature received: %s, expect: %s\n",
	__PRETTY_FUNCTION__,
	mtreq->mt_signature, MT_SIGNATURE);
#endif
	return -1;
    }
    if(do_checksum(&mtreq->opcode, (sizeof(struct mt)-4)/2)){
#ifdef DEBUG
	printf("%s(): checksum not match\n", __PRETTY_FUNCTION__);
#endif
//	return -1;
    }
    return 0;
}

static void *mt_run(const struct BMP_buf *buf, int *ret_size)
{
    struct mt_resp *resp = (struct mt_resp*)bmps_ret_buf;
    struct mt *mtreq = (struct mt *)buf;
    int data_i;
    struct ifreq ifr;
    char cmd[128];

    strcpy(ifr.ifr_name, BMPS_WLANIF);
    
    resp->mtmagicno = MT_MAGICNO;
    memcpy(resp->mt_signature, MT_SIGNATURE, strlen(MT_SIGNATURE));
    resp->s_no = mtreq->s_no;
#ifdef DEBUG
   printf("%s(): opcode -> 0x%02x\n", __PRETTY_FUNCTION__, mtreq->opcode); 
#endif 
   switch(mtreq->opcode){
	case CMD_SET:
#ifdef DEBUG
		printf("%s(): CMD_SET: ssid: %s, ch: %d, tx_rate: %d\n",
				__PRETTY_FUNCTION__,
				mtreq->ssid, mtreq->ch, mtreq->tx_rate);
#endif
	    data_i = mtreq->ch;
	    snprintf(cmd, 128, "/bin/setoid %s 17000007 long %d",
	      BMPS_WLANIF, data_i);
	    printf("Cmd: %s\n", cmd);
	    system(cmd);
	    /*
	    if(set_prismoid(BMPS_WLANIF, DOT11_OID_CHANNEL, &data_i, 4)<0){
		printf("Cannot set channel\n");
	    }
	     */
	    
	    snprintf(cmd, 128, "/bin/setoid %s 10000002 ssid %s",
	      BMPS_WLANIF, mtreq->ssid );
	    printf("Cmd: %s(%d)\n", cmd, strlen(mtreq->ssid));
	    system(cmd);
	    /*
	    if(set_prismoid(BMPS_WLANIF, DOT11_OID_SSID, mtreq->ssid,
		     strlen(mtreq->ssid))<0){
		printf("Cannot set SSID\n");
	    }
	    */
		data_i = (int)(mtreq->tx_rate==255)?-1:mtreq->tx_rate;
	    printf("tx rate will be %d\n", data_i);
	    snprintf(cmd, 128, "/bin/setoid %s 1d000004 long %d",
	      BMPS_WLANIF, data_i);
	    printf("Cmd: %s\n", cmd);
	    system(cmd);
	    /*
	    if(set_prismoid(BMPS_WLANIF, DOT11_OID_ALOFT_FIXEDRATE, &data_i, 4)<0){
		printf("Cannot set fixed rate\n");
	    }
	    */
	    // Activate parameters
	    IfIoctl(SIOCGIFFLAGS, &ifr);
	    ifr.ifr_flags &= (~IFF_UP) ;
	    IfIoctl(SIOCSIFFLAGS, &ifr);
	    usleep(500000);
	    ifr.ifr_flags |= IFF_UP ;
	    IfIoctl(SIOCSIFFLAGS, &ifr);
	    
	    break;
	case CMD_RESET:
	{
#if 0
	    int mtd0fd = open("/dev/mtdblock0", O_RDWR);
	    char blank[4] = {'\xff', '\xff', '\xff', '\xff'};
	    if(mtd0fd==-1){
		perror("Cannot open /dev/mtdblock0");
		break;
	    }
	    write(mtd0fd, blank, 4);
	    fsync(mtd0fd);
	    close(mtd0fd);
#endif
	    system("/bin/reboot");
	}
	    break;
	case CMD_RESET2DEF:
	    break;
	case CMD_ALLOW_FW_UPLOAD:
	    break;
	case CMD_TEST_PASSWD:
	    if(bmp_readSNMPCommunity(community, MAX_DATA_LENGTH)>0){
		if(strncmp(community, mtreq->ssid, 12)!=0)
		    goto MT_FAIL;
	    }
	    goto MT_FAIL;
	    break;
    }

   ret_size[0] = sizeof(struct mt_resp);
    resp->rc = MT_RESPONSE;
    return resp;
    
MT_FAIL:
    resp->rc = MT_RESPONSE_PASSWD_FAIL;
    return resp;
}


struct BMP_mod apdpmod_mt={
pid:	0,
check:	mt_check,
errHandler: NULL,
run:	mt_run
};
