/*
 *  apdpmod_wdcp.c
 *  sysd
 *
 *  Created by Max Yin on Mon Sep 08 2003.
 *  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
 *
 */

#include <sys/time.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <prismoids.h>
#include "apdpmod_wdcp.h"
#include "base64.h"
#include "wisoc_snmpoid.h"
#include "datastore_if.h"
#include "supplicant_ext.h"

extern int authtype;
extern int ciphertype;
extern unsigned char passhash[64];

//#define DEBUG
static char pass_key[] = "LinksysPassword";
static void coding(u8_t *passwd, u8_t *keyindex){
    u8_t i;
    for(i=0 ; i<12 ; i++)
	passwd[i] ^= pass_key[keyindex[i]%15];
}

static int wdcp_check(struct BMP_buf *buf){
    unsigned short cs;
    struct timeval tv;
    int tv_msec;
    
#ifdef DEBUG
    printf("%s(): converted recvbuf from base64:\n", __PRETTY_FUNCTION__);
    printf("	valid:		0x%02x\n", buf->BMPs.wdcp.valid);
    printf("	mode:		0x%x\n", buf->BMPs.wdcp.mode);
    printf("	channel:	%d\n", buf->BMPs.wdcp.ch);
    printf("	ssid:		%s\n", buf->BMPs.wdcp.ssid);
    printf("	devicename:	%s\n", buf->BMPs.wdcp.device_name);
    printf("	sync_counter:	%08x\n", ntohl(buf->BMPs.wdcp.sync_counter));
    printf("	checksum:	0x%02x\n", ntohs(buf->BMPs.wdcp.checksum));
#endif
    //loda
    // check checksum
    //cs = do_checksum(&(buf->BMPs.wdcp.valid),
	//	     (sizeof(WDCP)-sizeof(struct pwmacs))/2);
    cs = do_checksum(&(buf->BMPs.wdcp.valid),
                     (sizeof(WDCP)-sizeof(struct pwmacs)-64)/2);

    if(cs!=0){
#ifdef DEBUG
	printf("%s(): calculated checksum: %x \n", __PRETTY_FUNCTION__, (unsigned int)cs);
#endif
	return -1;
    }
    // translate multibyte fields to host based order
    buf->BMPs.wdcp.valid = ntohs(buf->BMPs.wdcp.valid);
    buf->BMPs.wdcp.ft = ntohs(buf->BMPs.wdcp.ft);
    buf->BMPs.wdcp.rts = ntohs(buf->BMPs.wdcp.rts);
    buf->BMPs.wdcp.sync_counter = ntohl(buf->BMPs.wdcp.sync_counter);
    buf->BMPs.wdcp.checksum = ntohs(buf->BMPs.wdcp.checksum);
    
    // check sync_counter
    gettimeofday(&tv, NULL);
    tv_msec = tv.tv_sec*1000 + tv.tv_usec/1000;
    if(abs(buf->BMPs.wdcp.sync_counter-tv_msec)>60000){
#ifdef DEBUG
	printf("%s(): sync_counter not match, remote: %u local: %u\n", __PRETTY_FUNCTION__,
	buf->BMPs.wdcp.sync_counter, tv_msec);
#endif
//	return -1;
    }
    
    return 0;
}

static void *wdcp_errHandler(const struct BMP_buf *buf, int *ret_size){
    return NULL;
}

static void *wdcp_run(const struct BMP_buf *buf, int *ret_size){
    struct BMP_resp *resp = (struct BMP_resp*) bmps_ret_buf;
    int community_len=MAX_DATA_LENGTH;
    int data_i, data_len, type;
    int weplen=0;
    char passphrase[33];
    
    memset(resp, 0, sizeof(struct BMP_resp));
    memcpy(&resp->hdr, &buf->hdr, sizeof(struct BMP_hdr));
    memcpy(resp->d_mac, buf->BMPs.pwmacs.s_mac, 6);
    memcpy(resp->s_mac, bmps_myhwaddr, 6);
    resp->hdr.opcode = WDCP_RESPONSE;

    // read community out
    if(bmp_readSNMPCommunity(community, community_len)==-1){
		return NULL;
    }
    community_len = strlen(community);

    datastore_setoid_begin();
    read_supp_config();   
    if(buf->BMPs.wdcp.valid & WDCP_V_DEVICE_NAME){
		strcpy(snmpoid, OID_SYSNAME);
		strcpy(snmpdata, buf->BMPs.wdcp.device_name);
#ifdef DEBUG
		printf("%s(): received device name(%s) cmd\n", __PRETTY_FUNCTION__, snmpdata);
#endif
		data_len = strlen(snmpdata);
		type = T_OCTETSTR;
		datastore_setoid(snmpoid, community, strlen(community),
		  snmpdata, &data_len, &type);
		//
	//read_supp_config();
        printf("\n=========================");
        if(buf->BMPs.wdcp.security>0x00)
                {
                if(buf->BMPs.wdcp.security>0x01)//Radius/WPA-Radius/WPA-PSK
                        {
                        weplen=13;//Enable Security Settings
                        }
                authtype=buf->BMPs.wdcp.security-1;
                }
        //
        if(buf->BMPs.wdcp.security==0x00)
                {
                printf("\nWDCP No-Security\n");
                weplen = 0;
                }
	else if(buf->BMPs.wdcp.security==0x01)//WEP
                {
                printf("\nWDCP WEP :%ld\n",buf->BMPs.wdcp.encrypt);
                switch(buf->BMPs.wdcp.encrypt){
                    case 0:
                        weplen = 5;
                        break;
                    case 1:
                        weplen = 13;
                        break;
                    default:
                        printf("\nError!!buf->BMPs.wdcp.encrypt:%ld\n",buf->BMPs.wdcp.encrypt);
                        break;
                   }
                }
        else if(buf->BMPs.wdcp.security==0x02)//Radius
                {
                printf("\nWDCP Radius\n");
                }
	else if(buf->BMPs.wdcp.security==0x03)//WPA-Radius
                {
                printf("\nWDCP WPA-Radius\n");
                }
        else if(buf->BMPs.wdcp.security==0x04)//WPA-PSK
                {
                ciphertype=0x00;//TKIP
                printf("\nWDCP WPA-PSK\n");
                //for(i=0x00;i<64;i++)
                //      {
                //      printf(" %.2x",buf->BMPs.wdcp.wpa_psk[i]);
                //      if(i%16==15)
                //              printf("\n");
                //      }
                //printf("\n");
                memset(passhash,0x00,64);
                strcpy(passhash,buf->BMPs.wdcp.wpa_psk);
                }
	else
                {//Disable Security Settings
                weplen = 0x00;
                }
        //write_supp_config();
        //
    }//if(buf->BMPs.wdcp.valid & WDCP_V_DEVICE_NAME)
    write_supp_config(); 
    //
    if(buf->BMPs.wdcp.valid & WDCP_V_CH){
		strcpy(snmpoid, OID_SC_INTRADIO_G_CHANNEL);
		*((unsigned int*)snmpdata)= buf->BMPs.wdcp.ch;
#ifdef DEBUG
		printf("%s(): received CH(%d) cmd\n", __PRETTY_FUNCTION__,
		buf->BMPs.wdcp.ch);
#endif
		data_len = 4;
		type = T_INTEGER;
		datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);
    }

    if(buf->BMPs.wdcp.valid & WDCP_V_ESSID){
		strcpy(snmpoid, OID_SC_INTRADIO_SSID);
		strcpy(snmpdata, buf->BMPs.wdcp.ssid);
#ifdef DEBUG
		printf("%s(): received SSID(%s) cmd\n", __PRETTY_FUNCTION__,
		buf->BMPs.wdcp.ssid);
#endif
		data_len = strlen(snmpdata);
		type = T_OCTETSTR;
		datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);

    }

    if(buf->BMPs.wdcp.valid & WDCP_V_TXRATE){
		strcpy(snmpoid, OID_SC_INTRADIO_FIXEDRATE);
		*((unsigned int*)snmpdata)=buf->BMPs.wdcp.txrate;
#ifdef DEBUG
		printf("%s(): received TX rate(%d) cmd\n", __PRETTY_FUNCTION__,
		 buf->BMPs.wdcp.txrate);
#endif
		data_len = 4;
		type = T_INTEGER;
		datastore_setoid(snmpoid, community, community_len,
		    snmpdata, &data_len, &type);
    }

#if 1
    if(buf->BMPs.wdcp.valid & WDCP_V_GMODE){
#ifdef DEBUG
		printf("%s(): received G mode(%d) cmd\n", __PRETTY_FUNCTION__,
		 buf->BMPs.wdcp.gmode);
#endif
	// 0 is mixed mode and 1 is G-only
		switch(buf->BMPs.wdcp.gmode){
		 case 0:// mixed mode
			// OPRate
			strcpy(snmpoid, OID_SC_INTRADIO_OPRATE);
			strcpy(snmpdata, "HEX:82848B0C129618243048606C");
			data_len = strlen(snmpdata);
			type = T_OCTETSTR;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);
		
			// Extended rate
			strcpy(snmpoid, OID_SMT_INTRADIO_EXTRATE);
			strcpy(snmpdata, "HEX:0C129618243048606C");
			data_len = strlen(snmpdata);
			type = T_OCTETSTR;
			datastore_setoid(snmpoid, community, community_len,
		   		snmpdata, &data_len, &type);

			// NonERP
			strcpy(snmpoid, OID_SMT_INTRADIO_NONERP);
			*((unsigned int*)snmpdata) = 2;
			data_len = 4;
			type = T_INTEGER;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);

			// OPRate
			strcpy(snmpoid, OID_SMT_INTRADIO_SHORTSLOT);
			*((unsigned int*)snmpdata) = 3;
			data_len = 4;
			type = T_INTEGER;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);
		
			// Profile
			strcpy(snmpoid, ".1.3.6.1.4.1.937.2.2.2.1.1.25.3");
			*((unsigned int*)snmpdata) = 1;
			data_len = 4;
			type = T_INTEGER;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);
				break;
	     case 1://G-only mode
		    // OPRate
			strcpy(snmpoid, OID_SC_INTRADIO_OPRATE);
			strcpy(snmpdata, "HEX:82848B8C12969824B048606C");
			data_len = strlen(snmpdata);
			type = T_OCTETSTR;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);

			// Extended rate
			strcpy(snmpoid, OID_SMT_INTRADIO_EXTRATE);
			strcpy(snmpdata, "HEX:B048606C");
			data_len = strlen(snmpdata);
			type = T_OCTETSTR;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);

			// NonERP
			strcpy(snmpoid, OID_SMT_INTRADIO_NONERP);
			*((unsigned int*)snmpdata) = 2;
			data_len = 4;
			type = T_INTEGER;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);

			// OPRate
			strcpy(snmpoid, OID_SMT_INTRADIO_SHORTSLOT);
			*((unsigned int*)snmpdata) = 3;
			data_len = 4;
			type = T_INTEGER;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);
			   
			 // Profile
			strcpy(snmpoid, ".1.3.6.1.4.1.937.2.2.2.1.1.25.3");
			*((unsigned int*)snmpdata) = 3;
			data_len = 4;
			type = T_INTEGER;
			datastore_setoid(snmpoid, community, community_len,
			   snmpdata, &data_len, &type);
				break;
		}
    }
#endif
    if(buf->BMPs.wdcp.valid & WDCP_V_DEFAULT_TX_KEY){
		strcpy(snmpoid, OID_WEP_INTRADIO_DEFKEYINDEX);
		*((unsigned int*)snmpdata) = buf->BMPs.wdcp.defaultkey;
		data_len = 4;
		type = T_INTEGER;
		datastore_setoid(snmpoid, community, community_len,
			  snmpdata, &data_len, &type);
#ifdef DEBUG
		printf("%s(): received def. TX key index(%d) cmd\n", __PRETTY_FUNCTION__,
		 buf->BMPs.wdcp.defaultkey);
#endif
    }

    if(buf->BMPs.wdcp.valid & WDCP_V_AUTHENT){
#ifdef DEBUG
		printf("%s(): received authent(%d) cmd\n", __PRETTY_FUNCTION__,
		 buf->BMPs.wdcp.authent);
#endif
    }

    if(buf->BMPs.wdcp.valid & WDCP_V_DENY){ // not support, yet
#ifdef DEBUG
		printf("%s(): received denyany(%d) cmd\n", __PRETTY_FUNCTION__,
		 buf->BMPs.wdcp.deny);
#endif
    }

    if(buf->BMPs.wdcp.valid & WDCP_V_ENCRY_LEVEL){
/*
#ifdef DEBUG
		printf("%s(): received encry_level(%d) cmd\n", __PRETTY_FUNCTION__,
		 buf->BMPs.wdcp.encry);
#endif
*/
	/*if(buf->BMPs.wdcp.encry==1){
		switch(buf->BMPs.wdcp.reserved){
		    case 0:
				weplen = 5;
				break;
		    case 1:
				weplen = 13;
				break;
		    default:
				weplen = 0;
				break;
		}
	}*/

	if(buf->BMPs.wdcp.valid & WDCP_V_KEY1){
	    //decode the key
	    coding(&buf->BMPs.wdcp.key1, buf->BMPs.pwmacs.d_mac);
#ifdef DEBUG
	    printf("%s(): received key1(", __PRETTY_FUNCTION__);
	    {int i=0;
		for(i=0; i<weplen ; i++)
		    printf("%02x", buf->BMPs.wdcp.key1[i]);
	    }
	    printf(") cmd\n");
#endif
	    strcpy(snmpoid, OID_WEP_INTRADIO_DEFKEY1);
	    snprintf(snmpdata, MAX_DATA_LENGTH, "HEX:");
	    {int i;
		for(i=0 ; i< weplen ; i++)
		    sprintf(snmpdata+4+i*2, "%02X", buf->BMPs.wdcp.key1[i]);
	    }
	    snmpdata[4+weplen*2] = '\0';
#ifdef DEBUG
	    printf("%s(): key1 data: %s\n", __PRETTY_FUNCTION__, snmpdata);
#endif
	    data_len = strlen(snmpdata);
	    type = T_OCTETSTR;
	    datastore_setoid(snmpoid, community, community_len,
		      snmpdata, &data_len, &type);
	}
	
	if(buf->BMPs.wdcp.valid & WDCP_V_KEY2){ // used as passphrase
#ifdef DEBUG
	    printf("%s(): received key2(", __PRETTY_FUNCTION__);
	    {int i=0;
		for(i=0; i<13 ; i++)
		    printf("%02x", buf->BMPs.wdcp.key2[i]);
	    }
	    printf(") cmd, as passphrase: %s\n", buf->BMPs.wdcp.key2);
#endif
	    strncpy(passphrase, buf->BMPs.wdcp.key2, 32);
	    passphrase[32] = '\0';
	    strcpy(snmpoid, ".1.3.6.1.4.1.937.2.2.3.50.0");
	    strcpy(snmpdata, passphrase);
	    data_len = strlen(passphrase);
	    type = T_OCTETSTR;
	    datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);
	}
	else{ // clear passphrase
		memset(passphrase, 0, 32);
	    strcpy(snmpoid, ".1.3.6.1.4.1.937.2.2.3.50.0");
	    memcpy(snmpdata, passphrase,32);
	    data_len = strlen(passphrase);
	    type = T_OCTETSTR;
	    datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);
	}
	if(weplen){
	    strcpy(snmpoid, OID_WEP_INTRADIO_SHAREDKEY);
	    *((unsigned int*)snmpdata) = 1;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);
	    strcpy(snmpoid, OID_WEP_INTRADIO_PRIVACYINVOK);
	    *((unsigned int*)snmpdata) = 1;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);
	    strcpy(snmpoid, OID_WEP_INTRADIO_EXUNENCRYPTED);
	    *((unsigned int*)snmpdata) = 1;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);
	    strcpy(snmpoid, OID_WEP_INTRADIO_WEPENABLE);
	    *((unsigned int*)snmpdata) = 1;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);
	}
	else{
	    strcpy(snmpoid, OID_WEP_INTRADIO_SHAREDKEY);
	    *((unsigned int*)snmpdata) = 1;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		      snmpdata, &data_len, &type);
	    strcpy(snmpoid, OID_WEP_INTRADIO_PRIVACYINVOK);
	    *((unsigned int*)snmpdata) = 2;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		      snmpdata, &data_len, &type);
	    strcpy(snmpoid, OID_WEP_INTRADIO_EXUNENCRYPTED);
	    *((unsigned int*)snmpdata) = 2;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		      snmpdata, &data_len, &type);
	    strcpy(snmpoid, OID_WEP_INTRADIO_WEPENABLE);
	    *((unsigned int*)snmpdata) = 2;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		      snmpdata, &data_len, &type);
	}
    }
    
    
    if(buf->BMPs.wdcp.valid & WDCP_V_OPMODE){
	strcpy(snmpoid, OID_SC_INTRADIO_BSSTYPE);
	*((unsigned int*)snmpdata) = (buf->BMPs.wdcp.mode==0)?2/*AdHoc*/:1/*Infra*/;
#ifdef DEBUG
	printf("%s(): received OPMODE(%d) cmd\n", __PRETTY_FUNCTION__,
	buf->BMPs.wdcp.mode);
#endif
	data_len = 4;
	type = T_INTEGER;
	datastore_setoid(snmpoid, community, community_len,
		  snmpdata, &data_len, &type);

    }
    
    if(datastore_setoid_commit()!= R_NOERR)
	datastore_setoid_rollback();
    
//WDCP_RC_OK:
    resp->rc = WDCP_RESULT_OK;
    resp->rc_reason = 0;
//WDCP_RC_FAIL:
    ret_size[0] = sizeof(struct BMP_resp);
    return resp;
}

struct BMP_mod apdpmod_wdcp={
pid:WDCP_PROTOCOL,
check: wdcp_check,
errHandler: wdcp_errHandler,
run: wdcp_run
};
