/*
led.c
	LED control procesures for WiSoC product
*/

/*
Stardate 10309.14 Max Yin
	Currently, only initialize the startup status.
	In the future, may need to control all LED pins
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>

#define ADDR_GPIO3DR	0xC0000480
#define ADDR_GPIO3DDR	0xC0000488

static FILE *FILE_netdev;

//#define DEBUG

int IfIoctl(int ioctl_num, struct ifreq *ifr);

static int fd_kmem = -1;

static void led_outdw(unsigned int addr, unsigned int data)
{
	if(lseek(fd_kmem, addr, SEEK_SET)==-1){
		printf("%s(): cannot seek to 0x%08x", __PRETTY_FUNCTION__,
			addr);
		perror("");
	}
#ifdef DEBUG1
	printf("%s() write 0x%08x to addr: 0x%08x\n", __PRETTY_FUNCTION__,
		data, addr);
#endif
	if(write(fd_kmem, &data, 4)==-1){
		perror("Cannot write to /dev/mem");
	}
}

static unsigned int led_indw(unsigned int addr)
{
 unsigned int input=0;

 	if(lseek(fd_kmem, addr, SEEK_SET)==-1){
		printf("%s(): cannot seek to 0x%08x", __PRETTY_FUNCTION__,
			addr);
		perror("");
	}
 	if(read(fd_kmem, &input, 4)==-1){
 		perror("Cannot read from /dev/mem");
 	}
 #ifdef DEBUG1
 	printf("%s(): reg. addr:0x%08x : 0x%08x\n" , __PRETTY_FUNCTION__,
 		addr, input);
 #endif
 	return input;
}

int led_G_on_t=0x00;     
void led_init()
{
	fd_kmem = open("/dev/kmem", O_RDWR);
	if(fd_kmem==-1){
		printf("%s(): Cannot open /dev/kmem for read/write!\n", __PRETTY_FUNCTION__);
	}
	led_G_on_t=0x00;
	// Setup GP3_3 and GP3_2
	led_outdw(ADDR_GPIO3DDR, led_indw(ADDR_GPIO3DDR)|0x000C000C);

	// Open /proc/net/dev
	FILE_netdev = fopen("/proc/net/dev", "r");
}

void led_G_on()
{
    time_t ltime;  
    //led_outdw(ADDR_GPIO3DR, led_indw(ADDR_GPIO3DR)|0x00040004);
    //printf("\nled_G_on\n");    
    time( &ltime ); 
    if(ltime>=30)
    	{//Delay WLAN LED
    	led_outdw(ADDR_GPIO3DR, 0x00040004);
    	}
    else if(led_indw(ADDR_GPIO3DR)&0x00080008)//Check if LAN is ON
	{
	if(led_G_on_t==0x00)
		{
		sleep(1);
		}
	led_G_on_t=0x01;
	led_outdw(ADDR_GPIO3DR, 0x00040004); 
	}
	
}

void led_G_off()
{
//	led_outdw(ADDR_GPIO3DDR, led_indw(ADDR_GPIO3DDR)|0x00040004);
    //printf("\nled_G_off\n");  
    led_outdw(ADDR_GPIO3DR, 0x00040000);
		//(led_indw(ADDR_GPIO3DR)&0xFFFBFFFB)|0x00040000);
}

void led_LAN_on()
{
 //   led_outdw(ADDR_GPIO3DDR, led_indw(ADDR_GPIO3DDR)|0x00080008);
//    led_outdw(ADDR_GPIO3DR, led_indw(ADDR_GPIO3DR)|0x00080008);
    //printf("\nled_LAN_on\n");  
    //led_outdw(ADDR_GPIO3DR, 0x00080008);
}

void led_LAN_off()
{
//    led_outdw(ADDR_GPIO3DDR, led_indw(ADDR_GPIO3DDR)|0x00080008);
    //printf("\nled_LAN_off\n");  
    //led_outdw(ADDR_GPIO3DR, 0x00080000);
//	      (led_indw(ADDR_GPIO3DR)&0xFFF7FFF7)|0x00080000);
}
void led_PW_on()
{
    //printf("\nled_PW_on\n");
    led_outdw(ADDR_GPIO3DR, 0x00100010);
}
                                                                                
void led_PW_off()
{
    //printf("\nled_PW_off\n");
    led_outdw(ADDR_GPIO3DR, 0x00100000);
}

struct ethact{
    char *ifname;
    int testRunning; // wlan does not have RUNNING flag
    void (*led_on)();
    void (*led_off)();
    int txrx;
    int led_state;
};

static struct ethact eth0={
ifname: "eth0",
testRunning: 1,
led_on: led_LAN_on,
led_off: led_LAN_off,
txrx: 0,
led_state: 1
};

static struct ethact eth1={
ifname: "eth1",
testRunning: 0,
led_on: led_G_on,
led_off: led_G_off,
txrx: 0,
led_state: 1
};

static struct ethact *eths[] = {
    &eth0, &eth1, NULL
};

static int testETHact(const struct ethact *eth)
{
    char buf[128], *p, *dev;
    unsigned int tx, rx, txrx=0, tmp;
    int ifFound = 0;
    struct ifreq ifr;

    
    if(FILE_netdev == NULL)
	FILE_netdev = fopen("/proc/net/dev", "r");

    strcpy(ifr.ifr_name, eth->ifname);
    IfIoctl(SIOCGIFFLAGS, &ifr);
    if(eth->testRunning && !(ifr.ifr_flags & IFF_RUNNING) ){
	eth->led_off();
#ifdef DEBUG
	printf("%s not running\n", eth->ifname);
#endif
	return -1; // interface is not running
    }
    while(fgets(buf, 128, FILE_netdev)){
	if((p=strchr(buf,':'))){
	    dev = strtok(buf, " :");
	    if(dev==NULL||strcmp(dev, eth->ifname)!=0)
		continue;
	    ifFound = 1;
	    p+=2;
	    sscanf(p, "%u %u %u %u %u %u %u %u %u %u",
	    &tmp, &rx, &tmp, &tmp, &tmp, &tmp, &tmp, &tmp,
	    &tmp, &tx);
	    txrx = tx + rx;
	    //printf("tmp: %u\n", tmp);
	    break;
	}
    }
    rewind(FILE_netdev);

    if(!ifFound){
	// Hey, where is interface ?
	eth->led_off();
#ifdef DEBUG
	printf("%s not found\n", eth->ifname);
#endif	
	return -1;
    }
    else{
	if(eth->txrx == txrx){
	    eth->led_on();
	}
	else{ // toggle
	    if(eth->led_state){
		eth->led_off();
		eth->led_state = 0;
	    }
	    else{
		eth->led_on();
		eth->led_state = 1;
	    }
	}
	eth->txrx = txrx;
	return txrx;
    }
}

void leds()
{
    int status;
    struct ethact *eth;
    int i=0;
    int abnormal=0;
    while(eths[i]){
	status = testETHact(eths[i]);
	if(status==-1)
	    abnormal++;
#if 0
	if(status == -1){
	    eths[i]->led_off();
	    eths[i]->led_state = 0;
	}
	else if(status != eths[i]->txrx){ // toggle
	    if(eths[i]->led_state){
		eths[i]->led_off();
		eths[i]->led_state = 0;
	    }
	    else{
		eths[i]->led_on();
		eths[i]->led_state = 1;
	    }
	    eths[i]->txrx = status;
	}
	else if(status == eths[i]->txrx){
	    eths[i]->led_on();
	    eths[i]->led_state = 1;
	}
#endif
	i++;
    }
    if(abnormal>1){
	printf("Both eth0 and eth1 go to -1\n");
    }

}
