#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ether.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>


#include "arpping.h"
#define GW_MON_BR_IF	"br0:0"
#define GW_MON_LAN_IF	"eth0"
#define GW_MON_WLAN_IF	"eth1"
#define GW_MON_WAN_IF	"eth2"
#define GW_MON_EWLAN_IF	"eth3"
#define GW_MON_STATUS_PATH	"/var/run/status-defgw"

static int parse_proc_route(unsigned int *gwip, char *ifname);
static void writeToStatus(const unsigned int gwip);

int StartAppl(char* opt0, char* opt1, char* opt2,char* opt3,char* opt4,char* opt5, char* opt6,char* opt7, char* opt8);


#define GW_STATE_OK		1
#define GW_STATE_NONE	0
static int gw_state=GW_STATE_NONE;


unsigned int gw_mon_gwip=0;

void sysd_gw_mon()
{
  unsigned char gwmac[6];
 char ifname[64];
 char mac_str[20];

 memset(gwmac, 0, 6);
 
	// Parse /proc/net/route to find out if the default route is there
	if(parse_proc_route(&gw_mon_gwip, ifname)==-1){
		// Default router not found, well, maybe next time
		gw_state = GW_STATE_NONE;
		return;
	}
	else if(gw_state == GW_STATE_NONE){ 
	// OK, found default router, and ebtable is not yet configured
	 unsigned int myip;
	 unsigned char mymac[6];
	 int sockfd;
	 struct ifreq ifr;
#ifdef GW_MON_TEST
         printf("OK found router location: %08x\n", gw_mon_gwip);
#endif
		// Send ARP to this IP address to retrieve MAC address
		//	First, find out my own IP address
		sockfd = socket(AF_INET, SOCK_DGRAM, 0);
		strcpy(ifr.ifr_name, GW_MON_BR_IF);
                if(ioctl(sockfd, SIOCGIFADDR, &ifr)==-1){
                    perror("cannot get my IP");
                    myip = 0;
                }
                else
                    myip = ((struct sockaddr_in*)&ifr.ifr_addr)->sin_addr.s_addr;
                if(ioctl(sockfd, SIOCGIFHWADDR, &ifr)==-1){
                    perror("cannot get my MAC");
                    close(sockfd);
			return;
                }
                close(sockfd);
		memcpy(mymac, ifr.ifr_hwaddr.sa_data, 6);
#ifdef GW_MON_TEST
	{
		struct in_addr addr;
		addr.s_addr = myip;
		printf("My IP is: %s\n", 
		 inet_ntoa(addr));
		printf("My MAC is: %02x:%02x:%02x:%02x:%02x:%02x\n",
		 mymac[0], mymac[1], mymac[2], mymac[3], mymac[4], mymac[5]);
	}
#endif
		// OK, send ARP
		if(arpping(gw_mon_gwip, gwmac, myip, mymac, GW_MON_LAN_IF)==0){
			// OK, found router's MAC address, configure ebtable
		    char sysbuf[128];
#ifdef GW_MON_TEST
		{
			printf("router MAC address is:  %02x:%02x:%02x:%02x:%02x:%02x\n",
			gwmac[0], gwmac[1], gwmac[2], 
			gwmac[3], gwmac[4], gwmac[5]);
		}
#endif
                    sprintf(sysbuf, "/bin/ebtables -t nat -A PREROUTING -i eth0 --rec-ipmac-pair %02x:%02x:%02x:%02x:%02x:%02x",
                           gwmac[0], gwmac[1], gwmac[2],
                           gwmac[3], gwmac[4], gwmac[5]);
		    system("/bin/ebtables -t nat -D PREROUTING 2");
		/*
                    StartAppl("/bin/ebtables", "-t", "nat", "-D", "PREROUTING", "2",
                              NULL,NULL,NULL);
                    StartAppl("/bin/ebtables", "-t", "nat", "-A", "PREROUTING", "-i",
                              "eth0", "--rec-ipmac-pair", mac_str);
		     */
		    system(sysbuf);
		    gw_state = GW_STATE_OK;
		}
		else
			gw_state = GW_STATE_NONE;
	}
}

int parse_proc_route(unsigned int *gwip, char *ifname)
{
 FILE *Froute = fopen("/proc/net/route", "r");
 char buf[256];
 int found = 0;
 unsigned flag;
 char flag_str[]="0x0000", gwip_str[]="0x00000000";
 
 	if(Froute==NULL)	return -1;
	fgets(buf, 256, Froute); // remove the first line
	while(!found){
		if(fgets(buf, 256, Froute)==NULL)
			break;
		sscanf(buf, "%s %s %s %s %s %s %s %s %s %s %s",
			ifname, buf, gwip_str+2, flag_str+2, 
			buf, buf, buf, buf,
			buf, buf, buf);
		gwip[0] = strtoul(gwip_str, NULL, 16);
		flag = strtoul(flag_str, NULL, 16);
#ifdef GW_MON_TEST
	{
	 struct in_addr addr;
	 	addr.s_addr = gwip[0];
		printf("ifname: %s gwip: %s flag: %d\n",
			ifname, inet_ntoa(addr), flag);
	}
#endif
		if(flag & 0x02){
			// found gateway
			found = 1;
		}
	}
        fclose(Froute);
        if(found){
            writeToStatus(gwip[0]);
            return 0;
        }
	else{
	    writeToStatus(0);
            return -1;
	}
}



void writeToStatus(const unsigned int gwip)
{
    struct in_addr addr;
    FILE *Fstatus;
    
    addr.s_addr = gwip;
    Fstatus = fopen(GW_MON_STATUS_PATH, "w+");
    if(Fstatus == NULL)	return;
    fprintf(Fstatus, "%s", inet_ntoa(addr));
    fclose(Fstatus);
}

#ifdef GW_MON_TEST_1
int main()
{
	sysd_gw_mon();
	return 0;
}
#endif


