/*
 *  apdp.c
 *  sysd
 *
 *  Created by Max Yin on Mon Sep 01 2003.
 *  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <net/if.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include "apdp.h"
#include "apdp_mods.h"
#define LINKSYS_WET54GV2
#include "model_id.h"
#include "base64.h"
#include "datastore_if.h"
#include "apdpmod_mt.h"

#define DEBUG 1


char bmps_myhwaddr[6];
char bmps_ret_buf[BMPS_RET_BUF_SIZE];
char community[MAX_DATA_LENGTH];
char snmpdata[MAX_DATA_LENGTH];
char snmpoid[MAX_OID_LENGTH];
unsigned int bmps_mtmode = 0;


static char recvbuf[1024];
int IfIoctl(int ioctl_num, struct ifreq *ifr);

static int sockfd;

int bmp_initSocket()
{
    struct sockaddr_in	bindaddr;
    int one = 1;
    struct ifreq ifr;

    // Retrieve my own eth0's MAC address
    strcpy(ifr.ifr_name, "eth0");
    if(IfIoctl(SIOCGIFHWADDR, &ifr)==0){
	memcpy(bmps_myhwaddr, ifr.ifr_hwaddr.sa_data, 6);
#ifdef DEBUG
	printf("%s(): eth0 HWADDR: %02x:%02x:%02x:%02x:%02x:%02x\n",
	__PRETTY_FUNCTION__,
	bmps_myhwaddr[0],bmps_myhwaddr[1],bmps_myhwaddr[2],
	bmps_myhwaddr[3],bmps_myhwaddr[4],bmps_myhwaddr[5]);
#endif
    }
    sockfd = socket(AF_INET, SOCK_DGRAM, 0);
    if(sockfd==-1){
#ifdef DEBUG
        printf("Cannot create socket");
#endif
        return sockfd;
    }
#ifdef DEBUG
    printf("%s(): socket created: %d\n", __PRETTY_FUNCTION__, sockfd);
#endif
    
    bindaddr.sin_family = AF_INET;
    bindaddr.sin_port = htons(LISTEN_PORT);
    bindaddr.sin_addr.s_addr = INADDR_ANY;

    if(bind(sockfd, (struct sockaddr*)&bindaddr, sizeof(bindaddr))==-1){
#ifdef DEBUG
        perror("Cannot bind to port 4000");
#endif
        close(sockfd);
        return -1;
    }
    setsockopt(sockfd, SOL_SOCKET, SO_BROADCAST, &one, sizeof(one));

    return sockfd;
}

int bmp_run(const int fd)
{
    struct sockaddr_in fromaddr;
    int sz_fromaddr = sizeof(fromaddr);
    void *ret_buf;
    int ret_size;
    int chkResult;

    int recv_ret = recvfrom(fd, recvbuf, sizeof(recvbuf), 0,
			    (struct sockaddr*)&fromaddr, &sz_fromaddr);
#ifdef DEBUG
    printf("%s(): received %d byte(s) from port 4000\n", __PRETTY_FUNCTION__,
	   recv_ret);
#endif
    
    if(recv_ret>0 && (chkResult=bmp_checkRequest((const struct BMP_buf*)recvbuf))==0){
	ret_size = recv_ret;
	ret_buf = bmp_findAndRun((const struct BMP_buf*)recvbuf, &ret_size);
	if(ret_buf){
	    if(bmp_response(fd, ret_buf, ret_size, &fromaddr)==-1)
		perror("sendto failed");
	    return 0;
	}
    }
    else if(recv_ret>0 && chkResult == -2){//password error
	struct BMP_buf *req=(struct BMP_buf*)recvbuf;
	struct BMP_resp *resp=( struct BMP_resp*) bmps_ret_buf;
	memset(bmps_ret_buf, 0, BMPS_RET_BUF_SIZE);
	memcpy(&resp->hdr, &req->hdr, sizeof(struct BMP_hdr));
	memcpy(resp->d_mac, req->BMPs.pwmacs.s_mac, 6);
	memcpy(resp->s_mac, bmps_myhwaddr, 6);
	resp->rc = RESULT_FAIL;
	resp->rc_reason = INV_PASSWORD;
	ret_size = sizeof(struct BMP_resp);
#ifdef DEBUG
	printf("%s(): password error, send INV_PASSWD back\n", __PRETTY_FUNCTION__);
	printf("	hdr: magic: %x\n", resp->hdr.magicno);
	printf("	hdr: pid: 0x%x\n", resp->hdr.pid);
	printf("	hdr: opcode: 0x%x\n", resp->hdr.opcode);
	printf("	hdr: sequence: 0x%x\n", resp->hdr.sequence);
	printf("	d_mac: %02x:%02x:%02x:%02x:%02x:%02x\n",
	resp->d_mac[0], resp->d_mac[1], resp->d_mac[2],
	resp->d_mac[3], resp->d_mac[4], resp->d_mac[5]);
	printf("	s_mac: %02x:%02x:%02x:%02x:%02x:%02x\n",
	resp->s_mac[0], resp->s_mac[1], resp->s_mac[2],
	resp->s_mac[3], resp->s_mac[4], resp->s_mac[5]);
	printf("	rc: 0x%x\n", resp->rc);
	printf("	rc_result: 0x%x\n", resp->rc_reason);
#endif
	if(bmp_response(fd, resp, ret_size, &fromaddr)==-1)
	    perror("sendto failed");
	return 0;
	
    }
    else if(recv_ret>0 && chkResult == 1){
	// manufactueing mode
	bmps_mtmode = 1;
	if(apdpmod_mt.check((const struct BMP_buf*)recvbuf)==-1)
	    return -1;
	ret_buf = apdpmod_mt.run((const struct BMP_buf*)recvbuf, &ret_size);
	if(ret_buf){
	    if(bmp_response(fd, ret_buf, ret_size, &fromaddr)==-1)
		perror("sendto failed");
	    bmps_mtmode = 0;
	    return 0;
	}
	bmps_mtmode = 0;
    }
    return -1;
}


static int bmp_chkPasswd(char *passwd, const int pwlen, char *keyindex)
{
    char community[64];
    int community_len = sizeof(community);
    char *passkey = MODELID_PASSWD_KEY;
    int i;
    bmp_readSNMPCommunity(community, community_len);

    for(i=0 ; i<pwlen ; i++){
	passwd[i] ^= passkey[keyindex[i]%MODELID_PASSWD_KEYLEN];
    }
#ifdef DEBUG
    printf("%s(): received passwd: %s\n", __PRETTY_FUNCTION__, passwd);
#endif
    if(strcmp(passwd, community)==0)
	return 0;
    return -1;
}

#define DMAC_BROADCAST	0
#define DMAC_UNICAST2ME	1
#define DMAC_OTHERS	2

int bmp_checkRequest(const struct BMP_buf *buf)
{
    int dmac_dest=DMAC_OTHERS;
    const struct BMP_hdr *hdr = (const struct BMP_hdr*)buf;
    // first check magic number
    if(hdr->magicno != htonl(MAGIC_NO)){
#ifdef DEBUG
	printf("%s(): wrong magic number, got 0x%08x, expect 0x%08x\n",
	__PRETTY_FUNCTION__, ntohl(hdr->magicno), MAGIC_NO);
#endif
	if(hdr->magicno == htonl(MT_MAGICNO)){
#ifdef DEBUG
	    printf("%s(): Received manufacturing request\n", __PRETTY_FUNCTION__);
#endif
	    return 1; // Manufacturing mode
	}
	return -1;
    }
#ifdef DEBUG
    printf("%s(): magic number passed\n", __PRETTY_FUNCTION__);
#endif

    // check destination MAC, to see if it's broadcast or unicast to me
#ifdef DEBUG
    printf("%s(): dmac: %02x:%02x:%02x:%02x:%02x:%02x\n", __PRETTY_FUNCTION__,
	   buf->BMPs.pwmacs.d_mac[0],buf->BMPs.pwmacs.d_mac[1],buf->BMPs.pwmacs.d_mac[2],
	   buf->BMPs.pwmacs.d_mac[3],buf->BMPs.pwmacs.d_mac[4],buf->BMPs.pwmacs.d_mac[5]);
    printf("%s(): mymac: %02x:%02x:%02x:%02x:%02x:%02x\n", __PRETTY_FUNCTION__,
	   bmps_myhwaddr[0],bmps_myhwaddr[1],bmps_myhwaddr[2],
	   bmps_myhwaddr[3],bmps_myhwaddr[4],bmps_myhwaddr[5]);
#endif
    if(memcmp(buf->BMPs.pwmacs.d_mac, "\xff\xff\xff\xff\xff\xff", 6)==0)
	dmac_dest = DMAC_BROADCAST;
    else if(memcmp(buf->BMPs.pwmacs.d_mac, bmps_myhwaddr, 6)==0)
	dmac_dest = DMAC_UNICAST2ME;
    else
	dmac_dest = DMAC_OTHERS;
#ifdef DEBUG
    {
	char *dmac_str[3]={"broadcast", "unicast to me", "unicast to others" };
	printf("%s(): destination MAC is to %s\n",
	__PRETTY_FUNCTION__, dmac_str[dmac_dest]);
    }
#endif

    // Now, check protocl ID
    switch(hdr->pid){
	case APDP_PROTOCOL:
#ifdef DEBUG
	    printf("%s(): received APDP packet\n", __PRETTY_FUNCTION__);
#endif
	    return 0;
	    break;
	case IPSP_PROTOCOL:
	case WDCP_PROTOCOL:
	    if(dmac_dest == DMAC_UNICAST2ME){
#ifdef DEBUG
		printf("%s(): received IPSP or WDCP packet\n", __PRETTY_FUNCTION__);
#endif
		// check password
		if(bmp_chkPasswd(buf->BMPs.pwmacs.passwd, 12, buf->BMPs.pwmacs.d_mac)==0){
		    return 0;
		}
		return -2; // password error
	    }
	    return -1;  // unicast to others
	    break;
	default:
	    break;
    }
    
    return -1;
}

void *bmp_findAndRun(const struct BMP_buf *buf, int *ret_size)
{
    struct BMP_mod *mod_ptr;
    const struct BMP_hdr *hdr = (const struct BMP_hdr*)buf;
    void *base64ptr = &(buf->BMPs.wdcp.valid) ;
    int i;
    void *retptr = NULL;

    convertFromBase64(base64ptr,
		      ret_size[0]-sizeof(struct BMP_hdr)-sizeof(struct pwmacs), base64ptr);
#ifdef DEBUG
    printf("%s(): convert %d bytes of data from base64\n",
	   __PRETTY_FUNCTION__, ret_size[0]-sizeof(struct BMP_hdr)-sizeof(struct pwmacs));
#endif

    for(i=0 ; bmp_modules[i]!=NULL ; i++){
	mod_ptr = bmp_modules[i];
	if(hdr->pid == mod_ptr->pid){
	    if(mod_ptr->check(buf)==0){
		retptr = mod_ptr->run(buf, ret_size);
		break;
	    }
	    else if(mod_ptr->errHandler){
		retptr = mod_ptr->errHandler(buf, ret_size);
		break;
	    }
	}
    }

    return retptr;
}

int bmp_response(const int fd, void *ret_buf, const int ret_size, struct sockaddr_in *from)
{
    struct BMP_hdr *hdr = (struct BMP_hdr *)ret_buf;
    unsigned short port_tmp;
    int ret;

    if(!bmps_mtmode)
	hdr->magicno = htonl(MAGIC_NO);
    
    from->sin_addr.s_addr = INADDR_BROADCAST;
	usleep(100000);
    ret = sendto(fd, ret_buf, ret_size, 0, (struct sockaddr*)from, sizeof(struct sockaddr_in));
    if(ret == -1){
	/*
	 FIXME:
	 Here is hard coding to add a routing to broadcasting address via br0 interface
	 */
	system("/bin/route add 255.255.255.255 br0");
	ret = sendto(fd, ret_buf, ret_size, 0, (struct sockaddr*)from,
	      sizeof(struct sockaddr_in));
    }
    return ret;
}

int bmp_readSNMPCommunity(char *community, int community_len)
{
    struct stat stat_buf;
    int comm_fd;
    int ret;
#if 0
    if(stat(BMPS_SNMPCOMMUNITY_FILE, &stat_buf) == -1){
#ifdef DEBUG
	printf("%s(): file %s not exist\n", __PRETTY_FUNCTION__, BMPS_SNMPCOMMUNITY_FILE);
#endif
       return -1;
    }
#endif
    // OK, file exists
    comm_fd = open(BMPS_SNMPCOMMUNITY_FILE, O_RDONLY);
    if(comm_fd==-1)	return -1;
    memset(community, 0, community_len);
    ret = read(comm_fd, community, community_len);
    if(ret!=-1){
	// remove "\n"
	char *p = community;
	int i;
	for(i=0 ; i<ret ; i++){
	    if( *p == '\n')
		*p = 0;
	    p ++;
	}
#ifdef DEBUG
	printf("%s(): community: %s(%d)\n", __PRETTY_FUNCTION__, community, ret);
#endif
    }
    close(comm_fd);
    return ret;
}

/* Other supporting function */
u16_t do_checksum(u16_t* buf, int len)
{
    int i;
    u32_t checksum = 0;
//    printf("%s(): len: %d\n", __PRETTY_FUNCTION__, len);
    for(i=0 ; i<len ; i++){
	checksum = (ntohs(buf[i]) + checksum)&0xFFFF;
    }
//    printf("checksum: 0x%8x\n", checksum);
    return (u16_t)checksum ;
}
