/*
 *  apdp.h
 *  sysd
 *
 *  Created by Max Yin on Mon Sep 01 2003.
 *  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
 *
 *  APDP ver 4
 *
 */

#ifndef _APDP_H_
#define _APDP_H_
#include <sys/types.h>
#include <netinet/in.h>

#include "int.h"

#define MAGIC_NO 	0x87654321
#define MT_MAGICNO	0xaa5555aa
#define LISTEN_PORT 	4000

// Original Protocol ID
#define OAPDP_PROTOCOL 	0x10
#define OIPSP_PROTOCOL 	0x20
#define OWDMP_PROTOCOL	0x30
#define OWDCP_PROTOCOL	0x40
#define OSRRP_PROTOCOL	0x50

// Protocol ID
#define APDP_PROTOCOL 	0x14
#define IPSP_PROTOCOL 	0x24
#define WDMP_PROTOCOL	0x31
#define WDCP_PROTOCOL	0x44
#define SRRP_PROTOCOL	0x50
#define OTHER_PROTOCOL  0xA0
#define FW_UPGRADE_PROTOCOL		0xB0

// define OP Code ;
#define APDP_REQUEST 	0x00
#define APDP_RESPONSE 	0x10
#define RESULT_FAIL	0x0001

#define IPSP_RESPONSE	0x00
#define IPSP_STATIC_IP 	0x10
#define IPSP_DHCP	0x11
#define IPSP_RESULT_OK	0x0000
#define IPSP_RESULT_FAIL	0x0001

#define WDCP_REQUEST 	0x10
#define WDCP_RESPONSE 	0x00
#define WDCP_RESULT_OK	0x0000
#define WDCP_RESULT_FAIL	0x0001

#define WDMP_NREQUEST	0x10
#define WDMP_UREQUEST	0x40
#define WDMP_RESPONSE	0x00
#define WDMP_RENEW	0x20
#define WDMP_NOTII	0x30
#define WDMP_RESULT_OK	0x0000
#define WDMP_RESULT_INP	0x0001
#define WDMP_RESULT_NRE	0x0002
#define WDMP_RESULT_IID	0x0003

#define SRRP_REQUEST	0x00
#define SRRP_RESPONSE	0x10
#define SRRP_RESULT_OK	0x0000
#define SRRP_RESULT_FAIL	0x0001

#define	FW_UPGRADE_RESPONSE_PDA	0x0A
#define	FW_UPGRADE_RESPONSE 	0x00
#define FW_UPGRADE_READPDA	0x10
#define FW_UPGRADE_WRITE	0x20

#define RESET_DEVICE	0x99
#define INV_PASSWORD	0x00ff

#define WDCP_V_OPMODE 		0x0001
#define WDCP_V_CH 		0x0002
#define WDCP_V_ESSID 		0x0004
#define WDCP_V_TXRATE 		0x0008
#define WDCP_V_PREMBLE 		0x0010
#define WDCP_V_DEVICE_NAME	0x0020
#define WDCP_V_GMODE 		0x0040
#define WDCP_V_RTS 		0x0080
#define WDCP_V_ENCRY_LEVEL	0x0100
#define WDCP_V_DEFAULT_TX_KEY	0x0200
#define WDCP_V_AUTHENT 		0x0400
#define WDCP_V_DENY 		0x0800
#define WDCP_V_KEY1 		0x1000
#define WDCP_V_KEY2 		0x2000
#define WDCP_V_KEY3 		0x4000
#define WDCP_V_KEY4 		0x8000


struct BMP_hdr{
    u32_t 	magicno;
    u8_t	pid;
    u8_t 	opcode;
    u16_t	sequence;
}__attribute__((packed));

typedef struct model_id{
    u8_t 	type;
    u8_t 	name[5];
    u8_t	hardware_rev;
    u8_t	msize;
}MODEL_ID  __attribute__((packed));

typedef struct apdp{
    u8_t d_mac[6];
    u8_t s_mac[6];
    MODEL_ID modelid;
    u8_t ssid[32];
    u8_t padding;
    u8_t ch;
    u8_t wep;
    u8_t namelen;
    u32_t ip;
    u8_t device_name[33];
    u32_t netmask;
    u32_t gateway;
    u32_t sync_counter;
    u8_t passphrase[32];
}__attribute__((packed)) APDP;

typedef struct ipsp{
    u8_t password[12];
    u8_t d_mac[6];
    u8_t s_mac[6];
    u32_t ip;
    u32_t mask;
    u32_t gateway;
    u32_t sync_counter;
    u16_t checksum;
} __attribute__((packed)) IPSP;

typedef struct scan_result{	//scan resault
    u8_t signal;		//signal level
    u8_t ch;		//channel no
    u8_t wep;		//wep
    u8_t bsstype;		//BSS type 0 = Ad 1 = Infr
    u8_t ssid[32];
} __attribute__((packed)) SR;

typedef struct srrp{
    u8_t d_mac[6];
    u8_t s_mac[6];
    u16_t listno;		//number of list entries
    u16_t result;
    SR sr[8];		//scan resault
} __attribute__((packed)) SRRP;

typedef struct wdcp{
    u8_t password[12];
    u8_t d_mac[6];
    u8_t s_mac[6];
    u16_t valid;
    u8_t mode;
    u8_t ch;
    u8_t ssid[32];
    u8_t txrate;
    u8_t preamble;
    u8_t gmode;
    u8_t reserved;
    u16_t ft;
    u16_t rts;
    u8_t encry;
    u8_t defaultkey;
    u8_t authent;
    u8_t deny;
    u8_t key1[13];
    u8_t key2[13];
    u8_t key3[13];
    u8_t key4[13];
//    u8_t pend[52-1];
    //u8_t name_len;
    u8_t device_name[32];
    u32_t sync_counter;
    u16_t checksum;
} __attribute__((packed))WDCP;

typedef struct wdmpr{
    u8_t linkstatus;
    u8_t mode;
    u8_t ch;
    u8_t txrate;
    u16_t rssi;
    u16_t lqual;
    u8_t bssid[6];
    u8_t essid[32];
}  __attribute__((packed)) WDMPR;

typedef struct wdmp{
    u8_t password[12];
    u8_t d_mac[6];
    u8_t s_mac[6];
    u16_t interval;
    u16_t expired;
    u16_t subscribe;
    u16_t result;
} __attribute__((packed)) WDMP;

typedef struct pwmacs{
    u8_t passwd[12];
    u8_t d_mac[6];
    u8_t s_mac[6];
} __attribute__((packed)) PWMACS;

struct BMP_resp{
    struct BMP_hdr hdr;
    u8_t d_mac[6];
    u8_t s_mac[6];
    u16_t rc; // return code
    u16_t rc_reason;    
}__attribute__((packed));

struct BMP_buf{
    struct BMP_hdr	hdr;
    union{
	PWMACS	pwmacs;
        APDP	apdp;
        IPSP	ipsp;
        SRRP	srrp;
        WDCP	wdcp;
        WDMP	wdmp;
    } BMPs;
} __attribute__((packed));

struct BMP_mod{
    int pid;
    int (*check)(struct BMP_buf *buf);
    void *(*errHandler)(const struct BMP_buf *buf, int *ret_size);
    void *(*run)(const struct BMP_buf *buf, int *ret_size);
};

int bmp_initSocket();
int bmp_checkRequest(const struct BMP_buf *buf);
int bmp_run(const int sockfd);
void *bmp_findAndRun(const struct BMP_buf *buf, int *ret_size);
int bmp_response(const int sockfd, void *ret_buf, const int ret_size, struct sockaddr_in *from);
int bmp_readSNMPCommunity(char *community, int community_len);
u16_t do_checksum(u16_t* buf, int len);

extern char bmps_ret_buf[];
#define BMPS_RET_BUF_SIZE	1024
extern char bmps_myhwaddr[];
extern char community[];
extern char snmpdata[];
extern char snmpoid[];

#define BMPS_LANIF	"eth0"
#define BMPS_WLANIF	"eth1"
#define BMPS_WANIF	"eth2"
#define BMPS_EXTWLANIF	"eth3"
#define BMPS_BRIF	"br0"
#define BMPS_IPIF	"br0:0"
#define BMPS_AUTOIF	"br0:1"
#define BMPS_SNMPCOMMUNITY_FILE	"/var/etc/passwd.conf"

#endif

