/*
 *  apdpmod_ipsp.c
 *  sysd
 *
 *  Created by Max Yin on Wed Sep 10 2003.
 *  Copyright (c) 2003 __MyCompanyName__. All rights reserved.
 *
 */

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "apdpmod_ipsp.h"
#include "datastore_if.h"
#include "wisoc_snmpoid.h"

#define DEBUG

static int ipsp_check(struct BMP_buf *buf){
    unsigned short cs;
    struct timeval tv;
    int tv_msec;

    if(!(buf->hdr.opcode == IPSP_STATIC_IP ||
       buf->hdr.opcode == IPSP_DHCP))
	return -1;
    
#ifdef DEBUG
    printf("%s(): converted recvbuf from base64:\n", __PRETTY_FUNCTION__);
    printf("	ip:		0x%08x\n", buf->BMPs.ipsp.ip);
    printf("	mask:		0x%08x\n", buf->BMPs.ipsp.mask);
    printf("	gateway:	0x%08x\n", buf->BMPs.ipsp.gateway);
    printf("	sync_counter:	%08x\n", ntohl(buf->BMPs.ipsp.sync_counter));
    printf("	checksum:	0x%02x\n", ntohs(buf->BMPs.ipsp.checksum));
#endif
    // check checksum
    cs = do_checksum(&(buf->BMPs.ipsp.ip),
		     (sizeof(IPSP)-sizeof(struct pwmacs))/2);
    if(cs!=0){
#ifdef DEBUG
	printf("%s(): calculated checksum: %x \n", __PRETTY_FUNCTION__, (unsigned int)cs);
#endif
	return -1;
    }
    // translate multibyte fields to host based order
    buf->BMPs.ipsp.sync_counter = ntohl(buf->BMPs.ipsp.sync_counter);
    buf->BMPs.ipsp.checksum = ntohs(buf->BMPs.ipsp.checksum);

    // check sync_counter
    gettimeofday(&tv, NULL);
    tv_msec = tv.tv_sec*1000 + tv.tv_usec/1000;
    if(abs(buf->BMPs.ipsp.sync_counter-tv_msec)>60000){
#ifdef DEBUG
	printf("%s(): sync_counter not match, remote: %u local: %u\n", __PRETTY_FUNCTION__,
	buf->BMPs.ipsp.sync_counter, tv_msec);
#endif
	//	return -1;
    }

    return 0;
}


static void * ipsp_run(const struct BMP_buf *buf, int *ret_size){
    struct BMP_resp *resp = (struct BMP_resp*) bmps_ret_buf;
    int community_len=MAX_DATA_LENGTH;
    int data_len, type;

    memset(resp, 0, sizeof(struct BMP_resp));
    memcpy(&resp->hdr, &buf->hdr, sizeof(struct BMP_hdr));
    memcpy(resp->d_mac, buf->BMPs.pwmacs.s_mac, 6);
    memcpy(resp->s_mac, bmps_myhwaddr, 6);
    resp->hdr.opcode = IPSP_RESPONSE;

    // read community out
    if(bmp_readSNMPCommunity(community, community_len)==-1){
	return NULL;
    }
    community_len = strlen(community);
    
    datastore_setoid_begin();
    switch(buf->hdr.opcode){
	case IPSP_DHCP:
	    // Enabled DHCP client
	    strcpy(snmpoid, OID_IP_DHCPMODE);
	    *((unsigned int*)snmpdata) = 1;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		      snmpdata, &data_len, &type);
	    break;
	case IPSP_STATIC_IP:
	    // Disable DHCP client
	    strcpy(snmpoid, OID_IP_DHCPMODE);
	    *((unsigned int*)snmpdata) = 3;
	    data_len = 4;
	    type = T_INTEGER;
	    datastore_setoid(snmpoid, community, community_len,
		      snmpdata, &data_len, &type);
	    // Configure IP network
	    strcpy(snmpoid, OID_IP_LANSETTINGS);
	    snprintf(snmpdata, MAX_DATA_LENGTH, "HEX:%02X%02X%02X%02X%02X%02X%08X%08X%08X",
	      bmps_myhwaddr[0],bmps_myhwaddr[1],bmps_myhwaddr[2],bmps_myhwaddr[3],
	      bmps_myhwaddr[4],bmps_myhwaddr[5],
	      htonl(buf->BMPs.ipsp.ip),
	      htonl(buf->BMPs.ipsp.mask),
	      htonl(buf->BMPs.ipsp.gateway));
#ifdef DEBUG
	    printf("%s(): LAN settings: %s\n", __PRETTY_FUNCTION__, snmpdata);
#endif
	    data_len = strlen(snmpdata);
	    type = T_OCTETSTR;
	    datastore_setoid(snmpoid, community, community_len,
		      snmpdata, &data_len, &type);
	    break;
    }

    datastore_setoid_commit();
    
    resp->rc = IPSP_RESULT_OK;
    resp->rc_reason = 0x00;
    ret_size[0] = sizeof(struct BMP_resp);
    //usleep(100000);
    return resp;
}

struct BMP_mod apdpmod_ipsp = {
pid: IPSP_PROTOCOL,
check: ipsp_check,
errHandler: NULL,
run: ipsp_run
};





