#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <signal.h>
#include "apdp.h"
#include "datastore_if.h"
#include "led.h"

//#define DEBUG

void sysd_gw_mon();
void sysd_mac_autoclone();
void eth1mon(int signo);

static void sigusr2_handler(int signo);

void sigusr2_handler(int signo){
#ifdef DEBUG
	printf("sysd: received SIGUSR2, prepare to reboot system\n");
#endif
	sleep(2);
	system("/bin/reboot");
}



void main(int argc, char **argv)
{
    fd_set fds;
    int maxfd=-1;
    int bmp_sockfd;
    struct timeval t;
    unsigned int counter = 0;

    // Initialize LED status
    led_init();
    
    // Initialization of file descriptors that needed
    bmp_sockfd = bmp_initSocket();
    if(maxfd < bmp_sockfd){
	maxfd = bmp_sockfd;
    }
    
    // Setup signal handler for USR2: reboot system
    signal(SIGUSR2, sigusr2_handler);
    // Setup SIGALRM handler to monitor eth1
    signal(SIGALRM, eth1mon);
    
    while(1){
	FD_ZERO(&fds);
	FD_SET(bmp_sockfd, &fds);
	t.tv_sec=0;
	t.tv_usec=50000;

	// Handle job(s) that is(are) releated to a file descriptor
#ifdef DEBUG
	printf("%s(): Handle file descriptor related job\n", __PRETTY_FUNCTION__);
#endif
	if(select(maxfd+1, &fds, NULL, NULL, &t)>0){
	    if(FD_ISSET(bmp_sockfd, &fds)){ // Test APDP protocol set
#ifdef DEBUG
		printf("%s(): received APDP request\n", __PRETTY_FUNCTION__);
#endif
		bmp_run(bmp_sockfd);
	    }
	    counter ++;
	    continue;
	}
	counter ++;

	// now, blink LEDs
	leds();
	
	if((counter % 100) == 0){ //5 secs
	    // Now, handle gateway detection
#ifdef DEBUG
	    printf("%s(): process sysd_gw_mon\n", __PRETTY_FUNCTION__);
#endif
	    sysd_gw_mon();

	    // Check auto clone
	    sysd_mac_autoclone();
	}
    }
}

// One supporting function
int IfIoctl(int ioctl_num, struct ifreq *ifr)
{
    int fd, ret;

    fd = socket(PF_INET, SOCK_DGRAM, 0);
    assert(fd > 0);

    ret = ioctl(fd, ioctl_num, ifr);
    close(fd);

    return ret;
}


