#ifdef __KERNEL__
#ifdef MODULE
#include <linux/module.h>
#endif
#include <linux/config.h>
#include <asm/param.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/string.h>
extern unsigned long volatile jiffies;
#else
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/signal.h>
#endif

#include "generic.h"
#include "ledstatus.h"
#include "ledtimer.h"
#include "ledproc.h"
#include "netiftimer.h"

// This is for WET54GV2 and WET55AG
#define NR_LEDS 5
struct LED_STATUS leds[NR_LEDS];
// 0->DIAG(ORANGE), 1->POWER(GREEN), 2->WLAN G, 3-> WLAN A 4->WLAN search
// LAN LED is controled by MVC

#ifdef __KERNEL__
static struct timer_list ledtimer;
static void ledktask_postinit();
static void ledktask_preexit();
static int ledktask_init()
{ int i;
	// Generic init function
	led_status_init(leds, NR_LEDS);
	init_timer(&ledtimer);
	ledtimer.function = led_timer;
	ledtimer.expires = jiffies + 1; // next jiffies
	ledtimer.data = (unsigned long)(&ledtimer);
	add_timer(&ledtimer);

	for(i=0 ; i<NR_LEDS ; i++){
		led_enable(leds[i]);
	}
	
	ledktask_postinit();
	
	ledproc_init();
	
	return 0;
}

static void ledktask_exit()
{
	ledktask_preexit();
	del_timer_sync(&ledtimer);
	ledproc_exit();
	
}

module_init(ledktask_init);
module_exit(ledktask_exit);

/* Specific LED init
 */
static struct timer_list netiftimer;
#define LAN "eth0"
#define WLANA "eth1"
#define WLANG "eth1"
#define WAN "eth2"
static struct led_netif netifs[2]; // both monitor eth1 but one for A and one for G

static void ledktask_postinit()
{
	leds[1].blink_period = HZ/2; // Green LED blinks when boots
	led_disable(leds[2]); // disable G led
	led_disable(leds[3]); // disable A led
	
	// init. net interface monitor timer
	strcpy(netifs[0].ifname, WLANG);
	netifs[0].ledindex = 2;
	netifs[0].blink_period = 3; // 3 ticks
	strcpy(netifs[1].ifname, WLANA);
	netifs[1].ledindex = 3;
	netifs[1].blink_period = 3; // 3 ticks
	netif_timer_init(netifs, 2);
	init_timer(&netiftimer);
	
	netiftimer.function = netif_timer;
	netiftimer.expires = jiffies + 1;
	netiftimer.data = (unsigned long)(&netiftimer);
	add_timer(&netiftimer);
}	

static void ledktask_preexit()
{
	del_timer_sync(&netiftimer);
}

#else
int main (int argc, const char * argv[]) {
	int i;
	struct itimerval t;
    // insert code here...
    printf("Hello, World!\n");
	oem_hw_led.init();
	
	for(i=0 ; i<8 ; i++){
		oem_hw_led.on(i);
		sleep(1);
		oem_hw_led.off(i);
	}
	
	led_status_init(leds, 8);
	led_off(leds[7]);
	led_on(leds[0]);
	for(i=0 ; i<5 ; i++){
		led_toggle(leds[1]);
		sleep(1);
	}
	led_off(leds[1]);
	signal(SIGALRM, led_timer);
	t.it_interval.tv_sec = 0;
	t.it_interval.tv_usec = 10000;
	t.it_value.tv_sec = 0;
	t.it_value.tv_usec = 10000;
	setitimer(ITIMER_REAL, &t, NULL);
	leds[2].blink_period = 100;
	leds[3].blink_period = 10;
	leds[4].blink_period = 20;
	leds[5].blink_period = 30;
	leds[6].blink_period = 40;
	leds[7].blink_period = 50;
	while(1){
		pause();
	}
	printf("\n");
    return 0;
}

#endif

