/*
 *  ledproc.c
 *  LEDKTask
 *
 *  Created by Max Yin on Tue Apr 20 2004.
 *  Copyright (c) 2004 __MyCompanyName__. All rights reserved.
 *
 */
#ifdef __KERNEL__

#include <asm/param.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#ifdef CONFIG_PROC_FS

#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include "ledproc.h"
#include "ledstatus.h"

static ssize_t ledcontrol_proc_write(struct file *file, const char *buf, size_t count, loff_t *ppos);
static struct proc_dir_entry *proc_ledcontrol;
static struct file_operations ledcontrol_fops={
write: ledcontrol_proc_write
};

static int index=0, cmd=0, blink_jiffies=0;
static char *ledindex, *ledcmd, *arg;
static char cmds[64];
/* ledcmd supports: disable(0), enable(1), on(2), off(3), toggle(4) and blink(5). Blink needs further
   argument, jiffies
 */


static int ledcontrol_parse_args(char *u_cmd)
{
	char *p = u_cmd;
	//printk("%s(): cmd: %s\n", __PRETTY_FUNCTION__, u_cmd);
	// first, ledindex
	ledindex = p;
	while(!(*p==' ' || *p=='\0'))   p++;
	*p++ = '\0';
	index = simple_strtoul(ledindex, NULL, 10);
	//printk("LED index: %d\n", index);
	if(index<0 || index>nr_leds)	return -1;
	// now parse cmd
	ledcmd = p;
	while(!(*p==' ' || *p=='\0')) p++;
	*p++ = '\0';
	//printk("LED cmd: %s\n", ledcmd);
	if(strncmp(ledcmd, "disable", 7)==0)
		cmd = 0;
	else if(strncmp(ledcmd, "enable", 6)==0)
		cmd = 1;
	else if(strncmp(ledcmd, "on", 2)==0)
		cmd = 2;
	else if(strncmp(ledcmd, "off", 3)==0)
		cmd = 3;
	else if(strncmp(ledcmd, "toggle", 6)==0)
		cmd = 4;
	else if(strncmp(ledcmd, "blink", 5)==0){
		cmd = 5;
		// OK, if cmd==5, blink, get blink_jiffies
		arg = p;
		blink_jiffies = simple_strtoul(arg, NULL, 10);
	}
	else{
		printk("%s(): unknown command: %s\n", __PRETTY_FUNCTION__, ledcmd);
		return -2;
	}
#if 0
	printk("%s(): led[%d], cmd: %s ", __PRETTY_FUNCTION__,
		   index, ledcmd);
	if(cmd==5)
		printk("blink_jiffies: %d\n", blink_jiffies);
#endif
	return 0;
}

ssize_t ledcontrol_proc_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	memset(cmds, 0, 64);
	if(copy_from_user(cmds, buf, (count>60)?60:count)<0)
		return -EFAULT;
	if(ledcontrol_parse_args(cmds)<0)
		return -EFAULT;
	switch(cmd){
		case 0:// disable
			led_disable(leds_status[index]);
			break;
		case 1:// enable
			led_enable(leds_status[index]);
			break;
		case 2:// on
			led_on(leds_status[index]);
			break;
		case 3:// off
			led_off(leds_status[index]);
			break;
		case 4://toggle
			led_toggle(leds_status[index]);
			break;
		case 5://blink
			leds_status[index].blink_period = blink_jiffies;
			break;
	}
	return count;
}

void ledproc_init()
{
 	// Create /proc/ledcontrol directory
	proc_ledcontrol = create_proc_entry("ledcontrol", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH, &proc_root);
 	proc_ledcontrol->proc_fops = & ledcontrol_fops;
}

void ledproc_exit()
{
	remove_proc_entry("ledcontrol", &proc_root);
}

#else  // CONFIG_PROC_FS
void ledproc_init()
{
}
void ledproc_exit()
{
}
#endif // CONFIG_PROC_FS

#endif // __KERNEL__

