/*  $Header$
 *
 *  linux/arch/arm/mach-poldhu/arch.c
 *
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  Architecture specific fixups.  This is where any
 *  parameters in the params struct are fixed up, or
 *  any additional architecture specific information
 *  is pulled from the params struct.
 */
#include <linux/tty.h>
#include <linux/delay.h>
#include <linux/pm.h>
#include <linux/init.h>

#include <asm/elf.h>
#include <asm/setup.h>
#include <asm/uaccess.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <asm/arch/blobcalls.h>

extern void genarch_init_irq(void);
extern void init_ram_mpu_area(unsigned long , unsigned long );

/* These are initialized in head-armv.S
 * the boot_param_block is not supported on Poldhu
 */
struct boot_struct *bootstruct = NULL;
long *boot_param_block;

struct param_struct common_params = { { {
	page_size:		4096,
	nr_pages:		0,
	flags:    		FLAG_RDLOAD,
	rootdev:		0
} } };


static void __init
fixup_poldhu(struct machine_desc *desc, struct param_struct *params,
             char **cmdline, struct meminfo *mi)
{
	if(!bootstruct || bootstruct->boot_magic != BOOTMAGIC_OKAY) {
		printk("Invalid bootstruct %p (magic %lx)\n", bootstruct, bootstruct->boot_magic);
		panic("Invalid Bootstruct");
	}
	/* Not supported on Poldhu */
	boot_param_block = NULL;

	params->u1.s.nr_pages = bootstruct->boot_ramsize >> PAGE_SHIFT;
	/* The bank should start at the physical start of RAM, while our
	 * boot_rambase is the start of RAM that is available after MVC init.
	 */
	mi->nr_banks      = 1;
	mi->bank[0].start = PHYS_OFFSET;
	mi->bank[0].size  = bootstruct->boot_rambase + bootstruct->boot_ramsize - PHYS_OFFSET;
	mi->bank[0].node  = 0;

	/* Initialize an MPU area for the Entire RAM */
	init_ram_mpu_area(mi->bank[0].start, mi->bank[0].size);
	/* Switch to KERNEL_DS. We switch to USER_DS when the first app is started */
	set_fs(get_ds());
}

MACHINE_START(POLDHU, "POLDHU")
	BOOT_MEM(0x12345678, 0x12345678, 0x12345678)
	MAINTAINER("TBD")
	INITIRQ(genarch_init_irq)
	BOOT_PARAMS((unsigned int)&common_params)
	FIXUP(fixup_poldhu)
MACHINE_END


