/* $Header$
 *
 * arch/arm/mm/mm-poldhu.c
 *
 * Extra MM routines for the Poldhu architecture
 *
 * Copyright (C) 1998 Russell King
 * Copyright (C) 2002 Intersil Americas Inc.
 */


#include <linux/config.h>
#include <asm/mmu.h>
#include <asm/uaccess.h>

#ifdef CONFIG_ARM9_MPU
void print_mpu_areas( void )
{
	int i;
	unsigned long mpu_dregs[8];
	unsigned long mpu_iregs[8];
	unsigned long mpu_dperm;
	unsigned long mpu_iperm;

	/* Read MPU area definitions */
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c0, 0 " : "=r" (mpu_dregs[0]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c0, 1 " : "=r" (mpu_iregs[0]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c1, 0 " : "=r" (mpu_dregs[1]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c1, 1 " : "=r" (mpu_iregs[1]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c2, 0 " : "=r" (mpu_dregs[2]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c2, 1 " : "=r" (mpu_iregs[2]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c3, 0 " : "=r" (mpu_dregs[3]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c3, 1 " : "=r" (mpu_iregs[3]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c4, 0 " : "=r" (mpu_dregs[4]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c4, 1 " : "=r" (mpu_iregs[4]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c5, 0 " : "=r" (mpu_dregs[5]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c5, 1 " : "=r" (mpu_iregs[5]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c6, 0 " : "=r" (mpu_dregs[6]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c6, 1 " : "=r" (mpu_iregs[6]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c7, 0 " : "=r" (mpu_dregs[7]) :);
	__asm__ __volatile__ ( "mrc p15, 0, %0, c6, c7, 1 " : "=r" (mpu_iregs[7]) :);

	/* Read MPU Access Permissions */
	__asm__ __volatile__ ( "mrc p15, 0, %0, c5, c0, 0 " : "=r" (mpu_dperm) :); // Data
	__asm__ __volatile__ ( "mrc p15, 0, %0, c5, c0, 1 " : "=r" (mpu_iperm) :); // Instr

	printk("MPU register contents:\n");
	for (i = 0; i < 8; i++) {
		printk("%d: Data  (0x%.8lx-0x%.8lx) P=%ld\n", i,
			ARM_MPU_AREA_START(mpu_dregs[i]),
			ARM_MPU_AREA_END(mpu_dregs[i]),
			(mpu_dperm >> (i<<1)) & 0x3);
		printk("   Instr (0x%.8lx-0x%.8lx) P=%ld\n",
			ARM_MPU_AREA_START(mpu_iregs[i]),
			ARM_MPU_AREA_END(mpu_iregs[i]),
			(mpu_iperm >> (i<<1)) & 0x3);
	}
}

#endif /* CONFIG_ARM9_MPU */
