/*
 *  linux/arch/armnommu/mach-poldhu/mpu.S
 *
 */

#include <linux/config.h>
#include <linux/linkage.h>
#include <asm/assembler.h>
#include <asm/memory.h>

No_access   = 0                         @ no access
Privileged  = 1                         @ privileged only
Read_only   = 2                         @ privileged rw/user ro
Full_access = 3                         @ user rw access

@ Memory access permissions
@

Area_disable = 0
Area_enable = 1

Size_4KB    = (0x0b << 1)
Size_8KB    = (0x0c << 1)
Size_16KB   = (0x0d << 1)
Size_32KB   = (0x0e << 1)
Size_64KB   = (0x0f << 1)
Size_128KB  = (0x10 << 1)
Size_256KB  = (0x11 << 1)
Size_512KB  = (0x12 << 1)
Size_1MB    = (0x13 << 1)
Size_2MB    = (0x14 << 1)
Size_4MB    = (0x15 << 1)
Size_8MB    = (0x16 << 1)
Size_16MB   = (0x17 << 1)
Size_32MB   = (0x18 << 1)
Size_64MB   = (0x19 << 1)
Size_128MB  = (0x1a << 1)
Size_256MB  = (0x1b << 1)
Size_512MB  = (0x1c << 1)
Size_1GB    = (0x1d << 1)
Size_2GB    = (0x1e << 1)
Size_4GB    = (0x1f << 1)


.section ".text.init",#alloc,#execinstr
		.type	init_mpu_sections, #function 

/*
 * Configure memory protection unit and the caches
 *
 * Area 4, 5, 6 and 7 are used for memory protection.
 * If memory protection is disabled, we use area 4 to
 * give full access to the entire RAM area.
 *
 * Trashes r0 and r1
 */
ENTRY(init_mpu_sections)
	mrc		p15, 0, r0, c5, c0, 0	@ read data access permissions
#ifdef CONFIG_ARM9_MPU
									@ Area 4 (read only) is used for text mapping
	ldr		r1, =(Read_only << 8) + (Full_access << 10) + (Full_access << 12) + (Full_access << 14 )
#else /* CONFIG_ARM9_MPU */
	ldr		r1, =(Full_access << 8)	@ Full access for area 4
#endif
	orr		r0, r0, r1
	mcr		p15, 0, r0, c5, c0, 0	@ set data access permissions


	mrc		p15, 0, r0, c5, c0, 1	@ read code access permissions
#ifdef CONFIG_ARM9_MPU
									@ Full access for areas 4, 5, 6, 7
	ldr		r1, =(Full_access << 8) + (Full_access << 10) + (Full_access << 12) + (Full_access << 14)
#else
	ldr		r1, =(Full_access << 8)	@ Full access for area 4
#endif
	orr		r0, r0, r1
	mcr		p15, 0, r0, c5, c0, 1	@ set code access permissions


@ define area 4 5 6 7

#ifdef CONFIG_ARM9_MPU

@ Disable all areas. They are set when they are needed
@ for memory protection.
	ldr	r0, =0x00000000 + Size_4GB + Area_disable

	mcr     p15, 0, r0, c6, c4, 0	@ data area 4
	mcr     p15, 0, r0, c6, c5, 0	@ data area 5
	mcr     p15, 0, r0, c6, c6, 0	@ data area 6
	mcr     p15, 0, r0, c6, c7, 0	@ data area 7

	mcr     p15, 0, r0, c6, c4, 1	@ instruction area 4
	mcr     p15, 0, r0, c6, c5, 1	@ instruction area 5
	mcr     p15, 0, r0, c6, c6, 1	@ instruction area 6
	mcr     p15, 0, r0, c6, c7, 1	@ instruction area 7

#else

@ Area 4 is the entire RAM
	mov	r0, #PHYS_OFFSET
	add	r0, r0, #Size_16MB + Area_enable

	mcr     p15, 0, r0, c6, c4, 0	@ data area 4
	mcr     p15, 0, r0, c6, c4, 1	@ instruction area 4

@ Disable other areas
	mov	r0, #0x00000000
	add	r0, r0, #Size_4GB + Area_disable

	mcr     p15, 0, r0, c6, c5, 0	@ data area 5
	mcr     p15, 0, r0, c6, c6, 0	@ data area 6
	mcr     p15, 0, r0, c6, c7, 0	@ data area 7

	mcr     p15, 0, r0, c6, c5, 1	@ instruction area 5
	mcr     p15, 0, r0, c6, c6, 1	@ instruction area 6
	mcr     p15, 0, r0, c6, c7, 1	@ instruction area 7
#endif

/* Note: due to a bug in the Poldhu, one cannot use cached and buffered
 * at the same time when using SDRAM.
 */
#define DATA_CACHED
#ifdef DATA_CACHED
/*
 * define cacheable data areas
 * 
 * Note that if data is cached, we need to set these areas to NOT bufferable
 */

	mrc		p15, 0, r0, c2, c0, 0	                @ read cacheable data settings
	orr		r0, r0, #(1 << 4) + (1 << 5) + (1 << 6) + (1 << 7)	@ enable for area 4, 5, 6, 7
	mcr		p15, 0, r0, c2, c0, 0	                @ cacheable data

#else

/*
 * define bufferable data areas
 *
 */
	mrc		p15, 0, r0, c3, c0, 0                   @ read bufferable data settings
	orr		r0, r0, #(1 << 4) + (1 << 5) + (1 << 6) + (1 << 7)	@ enable for area 5, 6, 7
	mcr		p15, 0, r0, c3, c0, 0                   @ bufferable data
 
#endif

/*
 * define cacheable code areas
 */
	mrc		p15, 0, r0, c2, c0, 1	                @ read cacheable code settings
	orr		r0, r0, #(1 << 4) + (1 << 5) + (1 << 6) + (1 << 7)	@ enable for area 4, 5, 6, 7
	mcr		p15, 0, r0, c2, c0, 1	                @ cacheable code 

	mov		pc, lr


