/*  $Header$
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/init.h>

#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/mach/pci.h>
#include <asm/arch/pci.h>

#ifdef CONFIG_PCI

#define PCI_TEST
#ifdef PCI_TEST
#   define IRQ_PCIINT 1
   static char pci_dev1[0x1000];
#endif /* PCI_TEST */

#define PCI_BASE 	    0xc0000980 
#define PCI_MEM 	    PCI_BASE + 0x1000 
#define PCI_MEM_SIZE	0x1000
#define PCI_MEM_MASK    0xfff

#define PCI_CONFIG_READ	 0xfa
#define PCI_CONFIG_WRITE 0xfb

#define PCIBIOS_NOT_SUCCESSFUL 0x80

static int __init isl38xx_map_irq(struct pci_dev *dev, u8 slot, u8 pin)
{
	return IRQ_PCIINT;
}

static void __init pci_setup_resources(struct resource **resource);

static void isl38xx_init(void *);

struct hw_pci isl38xx_pci __initdata = {
	setup_resources:	pci_setup_resources,
    init:			    isl38xx_init,
	mem_offset:		    0x00000000,
	swizzle:		    no_swizzle,
	map_irq:		    isl38xx_map_irq
};

static spinlock_t isl_lock = SPIN_LOCK_UNLOCKED;

int master_wait( int size );
void master_setup( int addr, int size, int mode );
int pci_read( int offset);
int pci_write( int offset, int value );

int master_wait( int size )
{
    int timeout = 0x80*size;  

    while( !(uPCI->ARMIntReg & PCIARMIntRegMASTERCOMPLETE) )
    {
        if( --timeout <= 0 )
	    return( -1 );
    }
    return( 0 );
}

void master_setup( int addr, int size, int mode )
{
   uPCI->ARMIntAckReg = PCIARMIntAckRegMASTERCOMPLETE;
   uPCI->MasHostMemAddr = addr;
   uPCI->MasLenReg = size;
   uPCI->MasControlReg = mode;		// ignore length
}

#ifdef PCI_TEST

int pci_read( int offset)
{
    u32 val;
    memcpy( &val, &(pci_dev1[offset]), 4);
    return ( val );
}

int pci_write( int offset, int value )
{
    memcpy( &(pci_dev1[offset]), &value, 4);
    return( 0 );
}

#else /* PCI_TEST */

int pci_read( int offset)
{
    master_setup( offset, 1 , PCIMasControlRegENABLE 
    			    | PCIMasControlRegUSEDATAREG );
    master_wait( 1 ); 
    return( uPCI->MasDataReg );
}

int pci_write( int offset, int value )
{
    uPCI->MasDataReg = value; 
    master_setup( offset, 1, PCIMasControlRegENABLE 
    			   | PCIMasControlRegUSEDATAREG
                           | PCIMasControlRegTOHOST );
    return( master_wait( 1 ) );
}

#endif /* PCI_TEST */

int pci_read_config( int offset )
{
    int value;

    uPCI->MasOptReg = PCIMasOptRegOPTIONENABLE | PCI_CONFIG_READ; 
    value = pci_read( offset ); 
    uPCI->MasOptReg = 0;
    return( value );
}

int pci_write_config( int offset, int value )
{
    int error;

    uPCI->MasOptReg = PCIMasOptRegOPTIONENABLE | PCI_CONFIG_WRITE;
    error = pci_write( offset, value ); 
    uPCI->MasOptReg = 0;
    return( error );
}

int isl_read_config_byte(struct pci_dev *dev, int where, u8 *val)
{
	unsigned long flags;
	unsigned int  offset;
    unsigned int  align; 
	u32 v;

    if ( dev->devfn != 0 ) { return PCIBIOS_NOT_SUCCESSFUL; }

    offset = where & (~0x3);
    align  = where & 0x3;
    
	spin_lock_irqsave(&isl_lock, flags);
    
    v = pci_read_config( offset );
    
    spin_unlock_irqrestore(&isl_lock, flags);

    *val = (u8)(v >> (8*align)) & 0xFF ;
							
#ifdef PCI_TEST
	printk( KERN_INFO "read_config_byte: %08x %08x\n", where, *val );
#endif
    return PCIBIOS_SUCCESSFUL;
}

int isl_read_config_word(struct pci_dev *dev, int where, u16 *val)
{
	unsigned long offset;
	unsigned long flags;
    unsigned int  align; 
	u32 v;

    if ( dev->devfn != 0 ) { return PCIBIOS_NOT_SUCCESSFUL; }
    
    offset = where & (~0x3);
    align  = where & 0x3;
    // assert ( align == 0x00 || align == 0x10 ); 

	spin_lock_irqsave(&isl_lock, flags);
	
    v = pci_read_config( offset );
    
    spin_unlock_irqrestore(&isl_lock, flags);

    *val = (u16)(v >> (8*align)) & 0xFFFF ;
							
#ifdef PCI_TEST
	printk( KERN_INFO "read_config_word: %08x %08x\n", where, *val );
#endif    
    return PCIBIOS_SUCCESSFUL;
}

int isl_read_config_dword(struct pci_dev *dev, int where, u32 *val)
{
	unsigned long offset;
	unsigned long flags;
    unsigned int  align; 
	u32 v;

    if ( dev->devfn != 0 ) { return PCIBIOS_NOT_SUCCESSFUL; }
    
    offset = where & (~0x3);
    align  = where & 0x3;
    // assert ( align == 0x00 );              
    
	spin_lock_irqsave(&isl_lock, flags);
	
    v = pci_read_config( offset );
    
    spin_unlock_irqrestore(&isl_lock, flags);

    *val = v;
							
#ifdef PCI_TEST
	printk( KERN_INFO "read_config_dword: %08x %08x\n", where, *val );
#endif    
    return PCIBIOS_SUCCESSFUL;
}

int isl_write_config_byte(struct pci_dev *dev, int where, u8 val)
{
	unsigned long flags;
	unsigned int  offset;
    unsigned int  align; 
	u32 v;

    if ( dev->devfn != 0 ) { return PCIBIOS_NOT_SUCCESSFUL; }
    
    offset = where & (~0x3);
    align  = where & 0x3;

	spin_lock_irqsave(&isl_lock, flags);
	
    v  = pci_read_config( offset );

    v &= ~(0xFF << (8*align));
    v |= val << (8*align);
    
    pci_write_config( offset, v );
    
    spin_unlock_irqrestore(&isl_lock, flags);

#ifdef PCI_TEST
	printk( KERN_INFO "write_config_byte: %08x %08x\n", where, val );
#endif	
    return PCIBIOS_SUCCESSFUL;
}

int isl_write_config_word(struct pci_dev *dev, int where, u16 val)
{
	unsigned long flags;
	unsigned int  offset;
    unsigned int  align; 
	u32 v;

    if ( dev->devfn != 0 ) { return PCIBIOS_NOT_SUCCESSFUL; }
    
    offset = where & (~0x3);
    align  = where & 0x3;
    // assert ( align == 0x00 || align == 0x10 ); 

	spin_lock_irqsave(&isl_lock, flags);
	
    v  = pci_read_config( offset );

    v &= ~(0xFFFF << (8*align));
    v |= val << (8*align);
    
    pci_write_config( offset, v );
    
    spin_unlock_irqrestore(&isl_lock, flags);

#ifdef PCI_TEST
	printk( KERN_INFO "write_config_word: %08x %08x\n", where, val );
#endif	
    return PCIBIOS_SUCCESSFUL;
}

int isl_write_config_dword(struct pci_dev *dev, int where, u32 val)
{
	unsigned long flags;
	unsigned int  offset;
    unsigned int  align; 

    if ( dev->devfn != 0 ) { return PCIBIOS_NOT_SUCCESSFUL; }
    
    offset = where & (~0x3);
    align  = where & 0x3;
    // assert ( align == 0x00 ); 

	spin_lock_irqsave(&isl_lock, flags);
	
    pci_write_config( offset, val );
    
    spin_unlock_irqrestore(&isl_lock, flags);

#ifdef PCI_TEST
	printk( KERN_INFO "write_config_dword: %08x %08x\n", where, val );
#endif	
    return PCIBIOS_SUCCESSFUL;
}

static struct resource isl38xx_iomem = {
	name:	"PCI isl38xx_iomem",
	start:	PCI_MEM,
	end:	PCI_MEM + PCI_MEM_SIZE,
	flags:	IORESOURCE_MEM,
};

static void __init pci_setup_resources(struct resource **resource)
{
	if (request_resource(&iomem_resource, &isl38xx_iomem))
		printk("PCI: unable to allocate non-prefetchable memory region\n");

	resource[0] = &ioport_resource;         /* the IO resource for this bus */
    resource[1] = &iomem_resource;          /* the mem resource for this bus */
	resource[1] = &isl38xx_iomem;
	resource[2] = NULL;                     /* the prefetch mem resource for this bus */
}

static struct pci_ops isl38xx_ops = {
	read_byte:	 isl_read_config_byte,
	read_word:	 isl_read_config_word,
	read_dword:	 isl_read_config_dword,
	write_byte:	 isl_write_config_byte,
	write_word:	 isl_write_config_word,
	write_dword: isl_write_config_dword,
};

void __init isl38xx_init(void *sysdata)     /* struct pci_sys_data sysdata  */
{
	unsigned long flags;

    spin_lock_irqsave(&isl_lock, flags);

    pci_write_config( 4, pci_read_config(4) | 6 );	

    /* set Base Address Register 0 to PCI_BASE */
    pci_write_config( 0x10, PCI_BASE ) ;
	
	spin_unlock_irqrestore(&isl_lock, flags);

#ifdef PCI_TEST
    isl_write_config_byte(  NULL, PCI_HEADER_TYPE, PCI_HEADER_TYPE_NORMAL  );
    isl_write_config_dword( NULL, PCI_VENDOR_ID, PCI_VENDOR_ID_INTERSIL );
    isl_write_config_dword( NULL, PCI_DEVICE_ID, PCI_DEVICE_ID_INTERSIL_3893 );
#endif	
    
    pci_scan_bus(0, &isl38xx_ops, sysdata);
}

#endif