/*  $Header$
 *  
 *  prism.h: Header file for the Intersil prism driver.
 *
 *  Copyright (C) <2002> Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __prism_h
#define __prism_h

/* So we can test parts of this driver on a poldhu based board. */
#include <linux/autoconf.h>
#ifdef CONFIG_POLDHU_TEST_DB_DRIVER
#define TEST_ON_POLDHU 
#else
#undef TEST_ON_POLDHU
#endif

/* INCLUDES */
#include <linux/isil_netlink.h>
#include <linux/if_ether.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <device_config.h>

/* DEFINES */
#define TRUE  1
#define FALSE 0

/* Return the device type, located in the first 16 bits */
#define GET_DEV_TYPE(var) ((var & 0xFFFF0000) >> 16)

/* Debug flag to get debug information for specific sections */
#define DBG_TX	        0x00000001
#define DBG_RX	        0x00000002
#define DBG_IRQ	        0x00000004
#define DBG_IOCTL       0x00000008
#define DBG_OPEN        0x00000010
#define DBG_CLOSE       0x00000020
#define DBG_ERROR       0x00000040
#define DBG_SKB	        0x00000080
#define DBG_WDS	        0x00000100
#define DBG_MGMT        0x00000200
#define DBG_PROBE       0x00000400
#define DBG_PIMFOR      0x00000800
#define DBG_FRAME       0x00001000
#define DBG_NETLINK     0x00002000
#define DBG_TRAP        0x00004000
#define	DBG_FUNC_CALL   0x00008000 

#define DBG_NONE	0x00000000
#define DBG_ALL	        0xFFFFFFFF

#ifndef NET_DEBUG
#define NET_DEBUG   (DBG_ERROR) 
#endif
static unsigned int net_debug = NET_DEBUG;

/* macros to simplify debug checking */
#define DEBUGLVL(x) if (net_debug & (x))
#ifdef GCC_3X
#define DEBUG(x,args,...) \
	do{DEBUGLVL(x) printk((args), ## __VA_ARGS__);}while(0)
#else
#define DEBUG(x,args...) do{DEBUGLVL(x) printk(## args);}while(0)
#endif
#define _DEBUG(x,args,...)

/* Structure definitions */

/* Holds information about associated clients */
struct client_information
{
    struct client_information *next_client;
    struct client_information **pprev_client;
    char                      mac_addr[ETH_ALEN];
};

/* Information that need to be kept for each device. */
struct net_local
{
    int (*send_mvc_msg) (struct net_device *dev, int operation,
                         int oid, long *data, unsigned long data_len);
    
    struct net_device_stats	    stats;	/* tx/rx statistics */
    struct bis_dev_descr_element    *dev_desc;	/* MVC device desriptor */
    int				    maxhead;	/* maximum head, tail and mtu */
    int				    maxtail;	/* sizes for all MVC devices */
    int				    maxmtu;	/* (so we can bridge w/h copy)*/
    struct trap_info		    *traps;	/* registrated traps */
    unsigned char		    *acl;
    int				    acl_nr_items;
    unsigned int		    tx_misaligned;
    unsigned int		    rx_misaligned;
    int				    maxnrframes;
    struct msg_frame		    *frame_pool;
    struct msg_frame		    *frame_freelist;
    unsigned int		    watchdog_timeout;  /* 0 = Off */
    unsigned int		    watchdog_event;
    struct net_device		    *wds_dev[MAX_WDS_LINKS];
    unsigned int 		    current_nof_wds_links;
    struct client_information       *assoc_table[MAX_ASSOCIATIONS];
};


/* Function prototypes. */

/* data */
int send_data(struct sk_buff *skb, struct net_device *dev);
int recv_data(struct sk_buff *skb, struct net_device *dev);

/* mgmt */
void mgmt_dev_associate (struct net_device *dev, char *mac_addr);
void mgmt_dev_disassociate (struct net_device *dev, char *mac_addr);
void mgmt_handle_trap(struct net_device *dev,  unsigned long oid, char *data, unsigned long data_len);
int  mgmt_send_mvc_msg(struct net_device *dev, int operation, int oid, long *data, unsigned long data_len);

/* netlink */
int  netlink_tx(u32 to_pid, int seq, int operation, int oid, char *data, unsigned long data_len, __u32 trap_flags, struct net_device *dev);
void netlink_open_socket(void);
void netlink_rx(struct sock *sk, int len);

/* PIMFOR */
int pimfor_encode_header(int operation, unsigned long oid, int device_id, int flags, int length, struct pimfor_hdr *header);
int pimfor_decode_header(struct pimfor_hdr *header, int *version, int *operation, unsigned long *oid,int *device_id, 
                         int *flags, unsigned long *length);
int pimfor_send_trap(struct net_device *dev, unsigned long oid, long *data, unsigned long data_len, __u32 trap_flags);

/* Debug functions prototpyes*/
#if (NET_DEBUG & DBG_SKB)
static void dbg_dump_hex_assci (unsigned char *buf, int size);
#endif
#if (NET_DEBUG & DBG_SKB)
static int dbg_check_area (unsigned char *p, int size, char *area_name);
#endif

void print_mac(unsigned char macaddress[ETH_ALEN]);

#endif /* !defined(__prism_h) */
