#include <linux/config.h>
#include <linux/version.h>
#include <linux/tty.h>
#include <linux/init.h>
#include <linux/module.h>
#include <asm/arch/blobcalls.h>
#ifdef CONFIG_MVC_DEBUG_CONSOLE
#include <linux/console.h>
#endif
#include "prism.h"
#include "prism_mgt.h"

extern unsigned int src_seq;

/**********************************************************************
 *  handle_blob_reqs
 *
 *  DESCRIPTION: Handle requests for device blob. These requests are
 *		 handles within the Ethernet device. Therefore, traps
 *		 for the blob device should be added to the Ethernet device.
 *
 *  PARAMETERS:	 Device struct of the Ethernet device
 *
 *  RETURN:	 NONE
 *
 *  NOTE:	 Called from the interrupt service routine.
 *
 **********************************************************************/
void handle_blob_reqs(struct net_device *dev)
{
	int requests;

	requests = blob_reqs();
        
        if (!requests)
	return;

	if (requests & 0xFFFFEFE0) {
		printk (" called with unknown requests: %d!!\n", requests);
		return;
	}

	if (requests & REQTRAP)
	{
		struct msg_conf *trapmsg = blob_trap( );

		if( trapmsg != NULL ) {
#ifdef TEST_ON_POLDHU
			/* Since there is no size parameter in MVC v1 we use 100 for the length
			 * of the trap msg. This should be big enough to hold all trap msgs.
			 */
                    mgt_response( 0, src_seq++, DEV_NETWORK, dev->ifindex, 
                       PIMFOR_OP_TRAP, trapmsg->oid, trapmsg->data, 100 ); 
#else
         	    mgt_response( 0, src_seq++, DEV_NETWORK, dev->ifindex, 
                       PIMFOR_OP_TRAP, trapmsg->oid, trapmsg->data, trapmsg->size ); 
#endif
		}
	}

	if (requests & REQWATCHDOG)
	{
		/* Lets trigger the blob watchdog service, for the BLOB device */
		blob_watchdog();
	}
}



/*
 * ------------------------------------------------------------
 * console driver
 * ------------------------------------------------------------
 */
static int debugdev_enabled = 0;

#ifdef CONFIG_MVC_DEBUG_CONSOLE

static struct console mvc_dbg_cons;

/*
 *	Print a string to the serial port trying not to disturb
 *	any possible real use of the port...
 *
 *	The console_lock must be held when we get here.
 */
static void mvc_debug_console_write(struct console *co, const char *s,
				unsigned count)
{
	struct msg_data debug_msg;

	debug_msg.data = "<printk> ";
	debug_msg.length = strlen(debug_msg.data);
	debug_msg.max = 0;
	dev_write(DEVICE_DEBUG, &debug_msg);

	debug_msg.data = (unsigned char *)s;
	debug_msg.length = count;
	dev_write(DEVICE_DEBUG, &debug_msg);
}

static kdev_t mvc_debug_console_device(struct console *c)
{
	return MKDEV(TTY_MAJOR, 64 + c->index);
}

/*
 *	Setup initial baud/bits/parity/flow control. We do two things here:
 *	- construct a cflag setting for the first rs_open()
 *	- initialize the serial port
 *	Return non-zero if we didn't find a serial port.
 */
static int __init mvc_debug_console_setup(struct console *co, char *options)
{
	int ret = -1;

	if(!debugdev_enabled) {
		/* start the MVC debug interface */
		if(dev_start(DEVICE_DEBUG, 0) < BER_NONE)
			goto out;

		if(dev_run(DEVICE_DEBUG) < BER_NONE) {
			dev_stop(DEVICE_DEBUG);
			goto out;
		}
	}

	co->cflag = CREAD | HUPCL | CLOCAL | B115200;
	debugdev_enabled = 1;
	ret = 0;
out:
	return ret;
}

static struct console mvc_dbg_cons = {
	name:		"mvcdbg",
	write:		mvc_debug_console_write,
	device:		mvc_debug_console_device,
	setup:		mvc_debug_console_setup,
	flags:		CON_PRINTBUFFER,
	index:		-1,
};

/*
 *	Register console.
 */
void __init mvc_debug_console_init(void)
{
	register_console(&mvc_dbg_cons);
}
#endif

/*
 * Startup and exit functions
 */
static int __init prism_blobdrv_init(void)
{
        int ret = -ENODEV;
#ifdef CONFIG_MVC_DEBUG
	/* start the MVC debug interface */
	if(dev_start(DEVICE_DEBUG, 0) < BER_NONE)
		goto out;

	if(dev_run(DEVICE_DEBUG) < BER_NONE) {
		dev_stop(DEVICE_DEBUG);
		goto out;
	}

	debugdev_enabled = 1;
	ret = 0;
#endif
out:
        return ret;
}

static void __exit prism_blobdrv_exit(void)
{
#ifdef CONFIG_MVC_DEBUG
	debugdev_enabled = 0;

	dev_halt(DEVICE_DEBUG);
	dev_stop(DEVICE_DEBUG);
#endif
}


EXPORT_NO_SYMBOLS;

module_init(prism_blobdrv_init);
module_exit(prism_blobdrv_exit);
