/* $Header$
*  prism_generic.c: Prism Ethernet driver, generic functions.
*
*  Partly based on previous prism ethernet driver by Martijn Schipper 2001.
*
*  Written by Bas Muns, Intersil 2002.
*
*  Copyright (C) 2001 Intersil Americas Inc.
*  Copyright (C) 2002 Intersil Americas Inc.
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/


/* Begin of include files */
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/net.h>
#include <linux/netdevice.h>
#include <net/sock.h>
#include <linux/netlink.h>

#include <asm/arch/blobcalls.h>
#include "prism.h"

//FIXME, temp here.
#include "prism_wds.h"
/* End of include files */

/* Begin of DEFINES */

#define MAC_HASH_BITS 8
#define MAC_HASH_SIZE (1 << MAC_HASH_BITS)

#define MAX_ASSOC_IDS   MAX_ASSOCIATIONS+1 /* Association id can be -1. We add 1, so -1 maps on 0. */ 
#define MAX_IE_ELEMENTS 20

/* End of DEFINES */

/* Static variable for the trap sequence counter */
__u32 trap_seq = 0; 

struct frame_attachment *frame_attach[MAX_ASSOC_IDS];

struct information_element
{
    struct information_element *next;   /* Pointer to next information element. */
    unsigned char elem_id;              /* Identifier of the element. */
    short size;                         /* Size of information element. */
    char  *data;                        /* Data of information element (including elem_id and size) */
};

struct frame_attachment
{
    struct frame_attachment *next;       /* Pointer to next frame_attachment. */
    short  frame_type;                   /* Frame type of this attachment. */
    struct information_element *ie_root; /* Pointer to first information element for this attachment. */
    char   *all_ie_data;                 /* Pointer to (temporary) buffer holding all information elements. */
};

/* pointer to the netlink socket */
struct sock *nl_sock = NULL;

/****************************************************************
 *								*
 *    			Begin of management functions		*
 *								*
 ***************************************************************/


/**********************************************************************
 *  mac_hash
 *
 *  DESCRIPTION: Computes hash value for a given mac-address. 
 *               Taken from bridging code by Lennert Buytenhek.
 *
 *  PARAMETERS:	 Pointer to mac address.
 *
 *  RETURN:      Hash value.
 *
 *  NOTE:        Inline function.	 
 *
 **********************************************************************/
static inline int mac_hash(unsigned char *mac)
{
    unsigned long x;

    //FIXME, use faster hash func.
    
    x = mac[0];
    x = (x << 2) ^ mac[1];
    x = (x << 2) ^ mac[2];
    x = (x << 2) ^ mac[3];
    x = (x << 2) ^ mac[4];
    x = (x << 2) ^ mac[5];

    x ^= x >> 8;

    return x & (MAC_HASH_SIZE - 1);
}


/*
 * Inline function that links client into the association table.
 */
static inline void client_link(struct net_local *lp, struct client_information *cip, int hash)
{
	cip->next_client = lp->assoc_table[hash];

        if(cip->next_client != NULL)
            cip->next_client->pprev_client = &cip->next_client; //FIXME, check neccesity...
        lp->assoc_table[hash] = cip;
        cip->pprev_client = &lp->assoc_table[hash];
}


/*
 * Inline function that removes client from the association table.
 */
static inline void client_unlink(struct client_information *cip)
{
	*(cip->pprev_client) = cip->next_client;
	if (cip->next_client != NULL)
		cip->next_client->pprev_client = cip->pprev_client;
	cip->next_client = NULL;
	cip->pprev_client= NULL;
}


/**********************************************************************
 *  mgmt_find_client
 *
 *  DESCRIPTION: Searches the association table of the device for the
 *               given client mac-address.
 *
 *  PARAMETERS:	 - pointer to device that holds the association table.
 *               - pointer to mac address we're searching for.
 *
 *  RETURN:      Pointer to client_information struct or
 *               NULL when client is not found.
 *
 *  NOTE:        Inline function.	 
 *
 **********************************************************************/
static inline struct client_information *mgmt_find_client(struct net_device *dev, char *address)
{
    struct net_local *lp = dev->priv;
    int hash = mac_hash(address);
    struct client_information *cip = lp->assoc_table[hash];

    while(cip != NULL) 
    {
        if (!memcmp(cip->mac_addr, address, ETH_ALEN))
            return (cip);

        cip = cip->next_client;
    }
    
    _DEBUG(DBG_MGMT, "Address %2.2x.%2.2x.%2.2x.%2.2x.%2.2x.%2.2x not found in assoc table... \n",
                    address[0], address[1], address[2], address[3], address[4], address[5]);

    return NULL;
}


static int mgmt_add_client(struct net_device *dev, char *address)
{
    struct client_information *cip;
    struct net_local *lp = dev->priv;
    int hash;

    hash = mac_hash(address);

    cip = lp->assoc_table[hash];
    
    DEBUG(DBG_MGMT, "mgmt_add_client, hash = %i, addr = %2.2x.%2.2x.%2.2x.%2.2x.%2.2x.%2.2x \n",
                    hash, address[0], address[1], address[2], address[3], address[4], address[5]);
    
    while(cip != NULL) 
    {
        if (!memcmp(cip->mac_addr, address, ETH_ALEN))
        {
            DEBUG(DBG_MGMT, "address %2.2x.%2.2x.%2.2x.%2.2x.%2.2x.%2.2x already in assoc_table...\n",
                            address[0], address[1], address[2], address[3], address[4], address[5]);
            return -EINVAL;
        }

        cip = cip->next_client;
    }

    cip = kmalloc(sizeof(*cip), GFP_ATOMIC);
    if (cip == NULL)
    {
        printk("Unable to allocate memory for client \n");
        return -ENOMEM;
    }

    DEBUG(DBG_MGMT, "Address %2.2x.%2.2x.%2.2x.%2.2x.%2.2x.%2.2x not yet in table, going to add it...",
                    address[0], address[1], address[2], address[3], address[4], address[5]);
    
    /* Fill in the client information fields... */
    memcpy(cip->mac_addr, address, ETH_ALEN);
    
    /* Link the client into the assoc table... */
    client_link(lp, cip, hash);

    return 0;

}


/* 
 * mgmt_del_client
 *
 * Removes a client from the assoc table of the device.
 *
 * RETURN: 0 when client is removed from assocation table.
 *        -1 when client is not removed from assocation table.
 */
int mgmt_del_client(struct net_device *dev, char *address)
{
    struct client_information *cip;
    struct net_local *lp = dev->priv;
    int hash;
    
    hash = mac_hash(address);

    cip = lp->assoc_table[hash];

    while(cip != NULL) 
    {
        if (!memcmp(cip->mac_addr, address, ETH_ALEN))
        {
            DEBUG(DBG_MGMT, "found address %2.2x.%2.2x.%2.2x.%2.2x.%2.2x.%2.2x to remove from assoc_table...\n",
                   address[0], address[1], address[2], address[3], address[4], address[5]);
            
            client_unlink(cip);
            kfree(cip); 
            return 0;
        }

        cip = cip->next_client;
    }

    DEBUG(DBG_MGMT, "Address %2.2x.%2.2x.%2.2x.%2.2x.%2.2x.%2.2x not found in assoc table... \n",
                    address[0], address[1], address[2], address[3], address[4], address[5]);

    return -1;
}


/**********************************************************************
 *  mgmt_dev_associate
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
void mgmt_dev_associate (struct net_device *dev, char* mac_addr)
{   
    
    //FIXME, check return van paed app. Als dit successfull is moet client pas ge-ad worden.
    
    //if (obj->code == DOT11_SC_SUCCESSFUL)
        /* Add client to our assoc table... */
        //mgmt_add_client(dev, obj->address);
}


/**********************************************************************
 *  mgmt_dev_disassociate
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
void mgmt_dev_disassociate (struct net_device *dev, char* mac_addr)
{
    _DEBUG(DBG_TRAP, "mgmt_dev_disassociate: STA ");
    print_mac(mac_addr);
        
    /* Remove client from our assoc table... */
    mgmt_del_client(dev, mac_addr);
}


/**********************************************************************
 *  mgmt_make_obj_attachment
 *
 *  DESCRIPTION: Calculates the size of all cached information elements.
 *               Mallocs a new obj attachment with this size.
 *               Copies all information elements into the data of this 
 *               obj attachment.
 *
 *  PARAMETERS:	 Pointer to frame_attachment struct. 
 *               Pointer to old object attachment for oa type and id.
 *
 *  RETURN:	 Pointer to new object attachment when succesfully malloced
 *               and data succesfully copied.
 *               NULL when information element == NULL or kmalloc failed.
 *
 **********************************************************************/
static struct obj_attachment *mgmt_make_obj_attachment(struct frame_attachment *fa, struct obj_attachment *old_oa)
{
    int i;
    
    int totalsize = 0;
    int offset = 0;
    struct information_element *ie = fa->ie_root;
    struct obj_attachment *oa = NULL;

    DEBUG(DBG_FUNC_CALL | DBG_MGMT, "mgmt_make_ie_buffer \n");
    
    if (ie == NULL)
    {
        printk(KERN_WARNING "No information elements for this attachment found. \n");
        return NULL;
    }

    /* We need to determine the total size of the information elements first. */
    for (i=0; (i <= MAX_IE_ELEMENTS && ie != NULL); i++, ie=ie->next)
        totalsize += (short)ie->size; 
    
    /* Set the pointer back to it's roots... */
    ie = fa->ie_root;
    
    /* Malloc a buffer for the obj_attachment + the data. */ 
    if ( (oa = kmalloc( (sizeof(struct obj_attachment) + totalsize), GFP_KERNEL)) == NULL)
        return NULL;
    
    oa->type = old_oa->type;
    oa->id   = old_oa->id;
    oa->size = totalsize;

    /* Copy the information elements into the buffer. */
    for (i=0; (i <= MAX_IE_ELEMENTS && ie != NULL); i++, ie=ie->next)
     {
        if (ie->data == NULL || ie->size == 0)
            continue;
        memcpy((oa->data+offset), ie->data, ie->size);
        offset += ie->size;
    }
    
    return(oa);
}

/**********************************************************************
 *  mgmt_add_information_element
 *
 *  DESCRIPTION: Checks if we have an information element for this element id.
 *               If so, it replaces the current element data for the new one,
 *               or removes the element when size == 0.
 *               If no previous information element for this id was found,
 *               a new information element is malloced, and put into the list.
 *
 *  PARAMETERS:	 Pointer to frame_attachment for this frame type.
 *               Pointer to the information element data. Contains element id.
 *               ie_size, which is the size of ie_data.   
 *
 *  RETURN:	 0 when succesfully added or replaced.
 *               1 when succesfully removed.
 *               -1 when element was not found, and MAX_IE_ELEMENTS was reached.
 *               -ENOMEM when we couldn't allocate memory for the information
 *               element or elements data.
 *
 **********************************************************************/
static int mgmt_add_information_element(struct frame_attachment *fa, char *ie_data, short ie_size)
{
    int i;
    unsigned char elem_id;
    
    int found_element                = FALSE;
    struct information_element *ie   = fa->ie_root;
    struct information_element *prev = ie;

    DEBUG(DBG_MGMT, "mgmt_add_information_element. \n");
    
    /* The element id should be set as first parameter in the information element data. */
    elem_id  = *ie_data;
    
    /* Check if we have an id for this element */
    for (i=0; (ie != NULL && i <= MAX_IE_ELEMENTS); i++)
    {   
        if (ie->elem_id == elem_id)
        {
            found_element = TRUE;

            if (ie->data == NULL)
                DEBUG(DBG_MGMT, "Found information element, but it contains no data. \n");
            
            DEBUG(DBG_MGMT, "Found information element %d with size %d\n", ie->elem_id, ie->size);
            
            if (ie_size == 0) {
                DEBUG(DBG_MGMT, "Going to remove information element %d \n", elem_id);
                
                if (ie == fa->ie_root)
                    fa->ie_root = ie->next;
                else 
                    prev->next = ie->next;
                kfree(ie); 
                return 1; 
            }
            else {
                /* If size is the same or less as the previous data size and we have a buffer, copy the data */
                if (ie->size <= ie_size && ie->data != NULL) {
                    ie->size = ie_size;
                    memcpy(ie->data, ie_data, ie->size);
                    return 0; 
                }
                else {
                    /* We need to malloc a new area for the data. */
                    kfree(ie->data); /* Free the old area first! */
                    
                    if ( (ie->data = kmalloc(ie_size, GFP_KERNEL)) == NULL) {
                        printk(KERN_ERR "kmalloc of data for information element %d failed. \n", ie->elem_id);
                        ie->size = 0;
                        return -ENOMEM;
                    }
                    ie->size = ie_size;
                    memcpy(ie->data, ie_data, ie->size);
                    return 0;
                }
            }
        }
        else {
            prev = ie;
            ie   = ie->next;
        }
    }
    
    if (i == MAX_IE_ELEMENTS && !found_element) {
        printk(KERN_WARNING "Element not found. Maximum information elements (%d) reached. \n", MAX_IE_ELEMENTS);
        return -1;
    }

    if (!found_element) {
        DEBUG(DBG_MGMT, "We didn't find an element... \n");
        
        if (ie_size == 0) {
            /* If ie_size == 0, we're asked to remove this element.
               But since we didn't find an element in our list we can't do that. */
            DEBUG(DBG_MGMT, "Asked to remove information element (%d), but no previous element found. \n", ie->elem_id);
            return -1;
        }
        
        /* Malloc a new information element. */
        if ( (ie = kmalloc(sizeof(struct information_element), GFP_KERNEL)) == NULL) {
            printk("Couldn't malloc memory for information element. \n");
            return -ENOMEM;
        }
        
        ie->elem_id = elem_id;
        ie->size    = 0;
        ie->data    = NULL;
        ie->next    = NULL;
        
        if ( (ie->data = kmalloc(ie_size, GFP_KERNEL)) == NULL) {
            printk(KERN_ERR "kmalloc of data for information element failed. \n");
            kfree(ie);
            ie->size = 0;
            return -ENOMEM;
        }
        ie->size = ie_size;
        memcpy(ie->data, ie_data, ie_size);
        
        /* Add the information element to the list. */
        ie->next    = fa->ie_root;
        fa->ie_root = ie;
        
        return 0; 
    }
    
    /* We shouldn't come here... */
    DEBUG(DBG_MGMT, "mgmt_add_information_element, shouldn't come here... \n");
    return -1;
}


/*
 * Finds frame attachment for frame type and association id. 
 *
 * - Returns pointer to frame attachment when found.
 * - Returns new frame_attachment for this type when no frame attachment
 *   was found and malloc of new frame attachment succeeded. 
 * - Returns NULL when malloc failed.
 */
static struct frame_attachment *find_frame_attach(short type, short id)
{
    int i;
    
    struct frame_attachment *fa   = frame_attach[id+1]; /* id+1, as id can be -1. We map -1 to 0 this way. */
    struct frame_attachment *prev = NULL;
    prev->next = NULL;

    /* Find if we have an attachment for this type already. */
    for (i=0; (fa->frame_type != type && fa != NULL); fa=fa->next)
        prev = fa;
    
    if (fa == NULL) /* Not found */
    {
        /* No previous frame_attachment for this type and id. Malloc one. */
        if ( (fa = kmalloc(sizeof(struct frame_attachment), GFP_KERNEL)) == NULL)
        {
            printk(KERN_ERR "Couldn't malloc space for frame_attachment. \n");
            return NULL;
        }
        fa->next        = prev->next;
        fa->frame_type  = type;
        fa->ie_root     = NULL;
        fa->all_ie_data = NULL;
        
        frame_attach[id+1] = fa;
        return(fa);
    }
    else /* Found it */
        return(fa);
}

/**********************************************************************
 *  mgmt_cache_frame_attachment
 *
 *  DESCRIPTION: Information elements in frame attachments must be cached
 *               in the driver, as the MVC won't do that for us.
 *
 *  PARAMETERS:	 data       - pointer to the struct obj_attachment.
 *               data_len   - length of the
 *               dev        - pointer to device we need to send information
 *                            elements to.
 *
 *  RETURN:	 0  when caching and sending the information elements to
 *                  the MVC succeeded.
 *               -1 when caching and sending information elements to the
 *                  MVC failed.
 *
 **********************************************************************/
static int mgmt_cache_frame_attachment(struct net_device *dev, char *data)
{
    int error;
    
    struct frame_attachment *fa     = NULL;
    struct obj_attachment   *oa     = NULL;
    struct obj_attachment   *old_oa = (struct obj_attachment *)data; 

    /* Find the corresponding frame_attachment for this type and id. If none exists, one is malloced for us.*/
    if ( (fa = find_frame_attach(old_oa->type, old_oa->id)) == NULL) {
        printk(KERN_ERR "Couldn't malloc space for frame attachment. \n");
        return -ENOMEM;
    }
    
    /* Add information element to this frame attachment. */
    if ( (error = mgmt_add_information_element(fa, old_oa->data, old_oa->size)) < 0) {
        printk(KERN_WARNING "Error %d. Couldn't add information element to frame attachment (type %d, id %d). \n", error, old_oa->type, old_oa->id);
        return -1;
    }
    
    /* Okay, we have a valid list of information elements now. Place them in one obj attachment. */
    oa = mgmt_make_obj_attachment(fa, old_oa);
    
    if (oa == NULL) {
        printk(KERN_ERR "Couldn't make new object attachment for information elements. \n");
        return -ENOMEM;
    }
    
    error = mgmt_send_mvc_msg(dev, OPSET, DOT11_OID_ATTACHMENT, (long*)oa, oa->size);
    
    kfree(oa); /* No need to save it, as all data is available in the information elements list. */
    
    if (error < 0) {
        DEBUG(DBG_MGMT, "Error %d setting DOT11_OID_ATTACHMENT. \n", error);
        return -1;
    }

    return 0;
}


/**********************************************************************
 *  mgmt_set_oid_attachment
 *
 *  DESCRIPTION: Information elements in frame attachments must be cached
 *               in the driver, as the MVC won't do that for us.
 *
 *  PARAMETERS:	 data       - pointer to the struct obj_attachment.
 *               dev        - pointer to device we need to send information
 *                            elements to.
 *
 *  RETURN:	 0  when caching and sending the information elements to
 *                  the MVC succeeded.
 *               -1 when caching and sending information elements to the
 *                  MVC failed.
 *
 **********************************************************************/
static int mgmt_set_oid_attachment(char *data, struct net_device *dev)
{
    int err;
    
    DEBUG(DBG_FUNC_CALL | DBG_MGMT, "mgmt_set_oid_attachment \n");
    
    if ( (err = mgmt_cache_frame_attachment(dev, data)) != 0)
        return -1;
    
    return 0;
}


/**********************************************************************
 *  mgmt_send_mvc_msg
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
int mgmt_send_mvc_msg(struct net_device *dev, int operation, int oid, long *data, unsigned long data_len)
{
    int error;
    
    struct net_local *lp = dev->priv;
    
    DEBUG(DBG_FUNC_CALL | DBG_MGMT, "mgmt_send_mvc_msg, dev->base_addr = %lx \n", dev->base_addr);
    
    /* We use the send_mvc_msg that is defined for this device. */
    if (lp->send_mvc_msg)
    {
        if ((error = lp->send_mvc_msg(dev, operation, oid, data, data_len)) < 0)
            printk(KERN_WARNING "send_mvc_msg returned error (%d) \n", error);
    }
    else
    {
        printk(KERN_ERR "Callback function for send_mvc_msg not defined for dev %s \n", dev->name);
        return -ENODEV;
    }
    
    return error;
}


/**********************************************************************
 *  mgmt_handle_trap
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
void mgmt_handle_trap(struct net_device *dev, unsigned long oid, char *data, unsigned long data_len)
{   
    __u32  trap_flags = 0; 
    int type = GET_DEV_TYPE(dev->base_addr); //FIXME, temp here. Will change.
    
    DEBUG(DBG_FUNC_CALL | DBG_MGMT, "mgmt_handle_trap \n");
    
    /* Determine trap_flags, depending on device type trap is received on */
    if (type == BIS_DT_802DOT3)           
        trap_flags |= TRAPGRP_ETH_TRAPS;          /* Traps from Ethernet interfaces */
    else if (type == BIS_DT_802DOT11)   
        trap_flags |= TRAPGRP_WLAN_TRAPS;         /* Traps from Wireless interfaces */
    else 
        trap_flags |= TRAPGRP_BLOBDEV_TRAPS;      /* gTraps from MVC Device */

    /* There are some special OIDs we need to do extra things for */
    switch (oid)
    {
        case DOT11_OID_ASSOCIATE:
            mgmt_dev_associate (dev, data);
            break;
        case DOT11_OID_DISASSOCIATE:
            mgmt_dev_disassociate (dev, data);
            break;
        default:
            break;
    }
    
    /* Now we need to tell the applications. */
    pimfor_send_trap(dev, oid, (long*)data, data_len, trap_flags);
}


/****************************************************************
 *								*
 *    			End of management functions		*
 *								*
 ***************************************************************/


/****************************************************************
 *								*
 *    			Begin of PIMFOR functions		*
 *								*
 ***************************************************************/

/*
 * PIMFOR (Proprietary Intersil Mechanism For Object Relay). PIMFOR is a simple
 * request-response protocol used to query and set items of management information
 * residing in the MVC. It is used to provide an object interface to device firmware.
 *
 * PIMFOR is also used for communication between applications and the driver.
 * For the uAP, PIMFOR is used for communication between the driver and the MVC.
 *
 */


/**********************************************************************
 *  pimfor_encode_header
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
int pimfor_encode_header(int operation, unsigned long oid, int device_id, 
                                int flags, int length, struct pimfor_hdr *header)
{
    DEBUG(DBG_FUNC_CALL | DBG_PIMFOR, "pimfor_encode_header \n");

    if(!header)
            return -1;

    memset(header, 0, sizeof(struct pimfor_hdr));

    // byte oriented members
    header->version = PIMFOR_VERSION_1;
    header->operation = operation;
    header->device_id = device_id;
    header->flags = flags;

    // word oriented members with byte order depending on the flags
    if (flags & PIMFOR_FLAG_LITTLE_ENDIAN) {
            // use little endian coding
            header->oid = cpu_to_le32(oid);
            header->length = cpu_to_le32(length);
    } else {
            // use big endian coding
            header->oid = cpu_to_be32(oid);
            header->length = cpu_to_be32(length);
    }
    return 0;
}


/**********************************************************************
 *  pimfor_decode_header
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
int pimfor_decode_header(struct pimfor_hdr *header, int *version,
                                int *operation, unsigned long *oid,
                                int *device_id, int *flags, unsigned long *length)
{
    DEBUG(DBG_FUNC_CALL | DBG_PIMFOR, "pimfor_decode_header \n");

    if(!header)
            return -1;

    // byte oriented members
    *version   = header->version;
    *operation = header->operation;
    *device_id = header->device_id;
    *flags = header->flags;

    // word oriented members with byte order depending on the flags
    if (*flags & PIMFOR_FLAG_LITTLE_ENDIAN) {
            // use little endian coding
            *oid = le32_to_cpu(header->oid);
            *length = le32_to_cpu(header->length);
    } else {
            // use big endian coding
            *oid = be32_to_cpu(header->oid);
            *length = be32_to_cpu(header->length);
    }
    
    return 0;
}


/*
 * pimfor_get_oid
 *
 * FIXME: descr.
 *
 * We're asked to get the value for a certain OID.
 * We know some OIDs in the driver, so we should not bother to ask the MVC for those OIDs...
 *
 * data points to data in PIMFOR packet.
 */
int pimfor_get_oid(unsigned long oid, char *data, unsigned long data_len, struct net_device *dev, u32 pid, int seq)
{
    int err = 0;
    
    DEBUG((DBG_FUNC_CALL | DBG_PIMFOR), "pimfor_get_oid \n");
    DEBUG(DBG_PIMFOR, "We're asked to get OID nr %lx. \n", oid);
    
    switch (oid)
    {   
        /*
        case XXX_OID_YYY:
           break;
        */

        default:    /* We don't need to handle the OID in the driver, so send it to the MVC */
            err = mgmt_send_mvc_msg(dev, PIMFOR_OP_GET, oid, (long*)data, data_len);
            break;
    }

    return err;
}


/* 
 * pimfor_set_oid
 *
 * FIXME: descr.
 *
 * We're asked to set a value for a certain OID.
 * Some OIDs also need things set in the driver. Check...
 *
 * data points to data in PIMFOR packet
 *
 */
int pimfor_set_oid(unsigned long oid, char *data, unsigned long data_len, struct net_device *dev)
{
    int err;
    
    DEBUG((DBG_FUNC_CALL | DBG_PIMFOR), "pimfor_set_oid, OID: %lx \n", oid);
    
    switch (oid)
    {
        case DOT11_OID_WDSLINKADD:
            if((err = add_wds_link(data, dev)) != 0)
                return -EINVAL;
            if ((err = mgmt_send_mvc_msg(dev, PIMFOR_OP_SET, oid, (long*)data, data_len)) < 0)
            {
                /* Setting the WDSLink in the MVC went wrong, so delete the WDS device... */
                del_wds_link(data, dev);
                return -EINVAL;
            }
            break;

        case DOT11_OID_WDSLINKREMOVE:
            if ((err = mgmt_send_mvc_msg(dev, PIMFOR_OP_SET, oid, (long*)data, data_len)) < 0)
            {
                printk("Couldn't remove wdslink from MVC.\n");
                print_mac(data); 
                return -EINVAL;
            }
            if((err = del_wds_link(data, dev)) != 0)
                return -EINVAL;
            
            break;

        case DOT11_OID_ATTACHMENT:
            if ((err =  mgmt_set_oid_attachment(data, dev)) < 0)
            {
                printk("Couldn't set oid_attachment. \n");
                return -EINVAL;
            }
            break;
            

        default: /* We don't need to do something in the driver for this OID, so send OID to the MVC */
            if ((err = mgmt_send_mvc_msg(dev, PIMFOR_OP_SET, oid, (long*)data, data_len)) < 0)
                return -EINVAL;
            break;
    }
    
    return err;
}


/* 
 * pimfor_send_trap
 *
 * FIXME: descr.
 *
 */
int pimfor_send_trap(struct net_device *dev, unsigned long oid, long *data, unsigned long data_len, __u32 trap_flags)
{ 
    DEBUG(DBG_FUNC_CALL | DBG_PIMFOR | DBG_MGMT | DBG_TRAP, 
          "pimfor_send_trap: Trap from dev %s, oid %lx, trap_flags = %x\n", dev->name, oid, trap_flags);
    
    netlink_tx(0 /*to_pid*/, ++trap_seq, PIMFOR_OP_TRAP, oid, (char*)data, data_len, trap_flags, dev);
    
    return 0;
}


/*
 * Receives pointer to received PIMFOR type NETLINK message
 *
 * FIXME: descr.
 *
 * In this function we check the action to perform,
 * for which device it's meant
 * and what else...
 *
 */
static int pimfor_decode_packet(struct sk_buff *skb, u32 pid, __u32 seq)
{   
    int version;
    int operation;
    unsigned long oid;
    int dev_id;
    int flags;
    unsigned long length;
    
    struct pimfor_hdr *header = (struct pimfor_hdr*)(skb->data+(sizeof(struct nlmsghdr)));
    char *data = PIMFOR_DATA(header);
    int err = 0;
    int need_reply = FALSE;
    struct net_device *dev = NULL;
    
    DEBUG((DBG_FUNC_CALL | DBG_PIMFOR), "pimfor_decode_packet \n");
    
    /* Decode PIMFOR header */
    pimfor_decode_header(header, &version, &operation, &oid, &dev_id, &flags, &length);
    
    /* Check PIMFOR version. We support version 1. */
    if (version != PIMFOR_VERSION_1)
    {
        printk("Received PIMFOR packet with unsupported version (version %i). \n", version);
        return -ENOTSUPP;
    }
    
    /* Check if the PIMFOR packet is encapsulated into a tunnel OID */
    /*
    if (oid == TUNNEL_OID)
    {   
        DEBUG(DBG_PIMFOR, "PIMFOR packet is tunneled... \n");
        
        //Check what to do for tunneled PIMFOR packets...
        //Note: header size is double when tunneled... 
    }
    */

    /* Find dev via the ifindex given in the PIMFOR packet */
    dev = dev_get_by_index(dev_id);
    if (dev == NULL)
    {
        printk("Couldn't find device for ifindex %i \n", dev_id);
        return -1;
    }
    
    /* Determine the PIMFOR operation */
    switch (operation)
    {
        case PIMFOR_OP_GET:
                    DEBUG(DBG_PIMFOR, "PIMFOR GET received...\n");
                    /* Check what we need to do with the OID */
                    err = pimfor_get_oid(oid, data, length, dev, pid, seq);                    
                    need_reply = TRUE;
                    break;
    
        case PIMFOR_OP_SET:
                    DEBUG(DBG_PIMFOR, "PIMFOR SET received...\n");
                    /* Check what we need to do with the OID */
                    err = pimfor_set_oid(oid, data, length, dev);
                    need_reply = TRUE;
                    break;
    
        case PIMFOR_OP_RESPONSE:
                    DEBUG(DBG_PIMFOR, "PIMFOR RESPONSE received...\n");
                    err = -ENOTSUPP;
                    break;
            
        case PIMFOR_OP_ERROR:
                    DEBUG(DBG_PIMFOR, "PIMFOR ERROR received...\n");
                    err = -ENOTSUPP;
                    break;
    
        case PIMFOR_OP_TRAP:
                    DEBUG(DBG_PIMFOR, "PIMFOR TRAP received...\n");
                    err = -ENOTSUPP;
                    break;
        
        default:    /* Unsupported operation */
                    printk("Unknown PIMFOR Operation (%i) received. \n", operation);
                    err = -ENOTSUPP;
    }
    
    /* The operation is handled, send a reply back if we need to... */
    if (need_reply == TRUE)
    {
        if (err < 0) /* If there was an error (g/s) etting the data, return the error */
        {
            DEBUG(DBG_PIMFOR, "There was an error (g/s)etting the data. Error nr = %i \n", err);
            netlink_tx(pid, seq, PIMFOR_OP_ERROR, oid, data, length, 0 /*trap_flags*/, dev);
        }else /* Okay, we've got the data and need to send a response */
        {
            DEBUG(DBG_PIMFOR, "GETted/SETted data. \n");
            netlink_tx(pid, seq, PIMFOR_OP_RESPONSE, oid, data, length, 0 /*trap_flags*/, dev);
        }
    }

    return 0; 
}


/****************************************************************
 *								*
 *    			End of PIMFOR functions 		*
 *								*
 ***************************************************************/


/****************************************************************
 *								*
 *    			Begin of NETLINK functions		*
 *								*
 ***************************************************************/


/**********************************************************************
 *  netlink_open_socket
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
void netlink_open_socket(void)
{  
    /* Open the netlink socket only once */
    if (nl_sock != NULL)
        return;
    
    nl_sock = netlink_kernel_create(NETLINK_ISIL, netlink_rx);
    if (nl_sock == NULL)
		printk(KERN_WARNING "netlink_kernel_create failed\n");
}
   

/**********************************************************************
 *  netlink_encode_packet
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static struct sk_buff *netlink_encode_packet(int pid, int seq, int operation, int oid, char *data, unsigned long data_len, struct net_device *dev)
{
    unsigned char *old_tail;
    struct sk_buff *skb;
    struct nlmsghdr *nlh;
    struct pimfor_hdr *pimfor_hdr;
    size_t size;

    size = NLMSG_LENGTH(data_len + sizeof(struct pimfor_hdr));
    
    DEBUG((DBG_FUNC_CALL | DBG_NETLINK), "netlink_encode_packet \n");

    skb = alloc_skb(size, GFP_ATOMIC);
    if (!skb)
        goto nlmsg_failure;

    old_tail = skb->tail;
    nlh = NLMSG_PUT(skb, pid, seq, NETLINK_TYPE_PIMFOR, size - NLMSG_ALIGN(sizeof(struct nlmsghdr)));
    pimfor_hdr = (struct pimfor_hdr *)NLMSG_DATA(nlh);
    
    pimfor_encode_header(operation, oid, dev->ifindex, 0 /*flags*/, data_len, pimfor_hdr);
    /* Add data after the PIMFOR header */
    memcpy(PIMFOR_DATA(pimfor_hdr), data, data_len);
    
    nlh->nlmsg_len = skb->tail - old_tail;
    
    return skb;

nlmsg_failure:
    if (skb)
        kfree(skb);
    printk(KERN_ERR "error creating netlink message\n");
    return NULL;


}


/**********************************************************************
 *  netlink_tx
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
int netlink_tx(u32 to_pid, int seq, int operation, int oid, char *data, unsigned long data_len, __u32 trap_flags, struct net_device *dev)
{
    struct sk_buff *skb;
    int offset;
    int err;

    DEBUG((DBG_FUNC_CALL | DBG_NETLINK), "netlink_tx \n");

    skb = netlink_encode_packet(to_pid, seq, operation, oid, data, data_len, dev);

    if (!skb)
        return -ENOMEM;
    
    DEBUG(DBG_NETLINK, "SENDING skb: \n");
    for (offset = 0; offset < skb->len; offset++)
            DEBUG(DBG_NETLINK, "%02X ", skb->data[offset]);
    DEBUG(DBG_NETLINK, "\n");

    if (skb && (trap_flags == 0)) /* unicast message */
    {
        DEBUG(DBG_NETLINK, "netlink unicast to pid %d\n", to_pid);
        err = netlink_unicast(nl_sock, skb, to_pid, MSG_DONTWAIT);
        DEBUG(DBG_NETLINK, "netlink_tx: netlink_unicast return value = %i \n", err);
    }
    else if (skb && (trap_flags != 0)) /* multicast message */
    {
        DEBUG(DBG_NETLINK, "netlink multicast, seq = %i, trap_flags = %x \n", seq, trap_flags);
        netlink_broadcast(nl_sock, skb, 0, trap_flags ,GFP_ATOMIC); 
        err = 0; /* netlink_broadcast does not return a value. */ 
    }
    else /* error */
    {
        printk("netlink_tx, couldn't send netlink packet\n");
        err = -1;
    }
    
    return err;
}

#define NL_RCV_SKB_FAIL(err) do { netlink_ack(skb, nl_header, (err)); return; } while (0)

/**********************************************************************
 *  netlink_rx_skb
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 *  NOTE:        Inline function.
 *
 **********************************************************************/
static inline void netlink_rx_skb(struct sk_buff *skb)
{
    int offset; //For debug print
    int status;
    
    u32 to_pid = NETLINK_CB(skb).pid;
    struct nlmsghdr *nl_header = (struct nlmsghdr*) skb->data; 
    __u32 seq = nl_header->nlmsg_seq;

    DEBUG((DBG_FUNC_CALL | DBG_NETLINK), "netlink_rx_skb \n");
    DEBUG(DBG_NETLINK, "netlink_rx_skb: skb->data %p, skb %p, flags %x\n", skb->data, skb, nl_header->nlmsg_flags);
    
    for (offset = 0; offset < skb->len; offset++) //Debug
    DEBUG(DBG_NETLINK, "%02X ", skb->data[offset]);
    DEBUG(DBG_NETLINK, "\n");
    
    /* We've received a NETLINK message. Check the NETLINK type */
    if (nl_header->nlmsg_type == NETLINK_TYPE_PIMFOR)    /* Currently, we only support PIMFOR type messages... */
    {   
        DEBUG((DBG_NETLINK | DBG_PIMFOR), "Received a PIMFOR type netlink packet... \n");
        /* It's a PIMFOR type packet, decode it */
        status = pimfor_decode_packet(skb, to_pid, seq);
    }
    else /* Unsupported NETLINK type */
    {
        printk(KERN_WARNING "netlink_rx_skb: Received NETLINK message with unsupported type (%d)... \n", nl_header->nlmsg_type);
        status = -ENOTSUPP;
    }
    
    if (status < 0)
        /* Send a netlink error back. */
        NL_RCV_SKB_FAIL(status);
    
    if (nl_header->nlmsg_flags & NLM_F_ACK)
        netlink_ack(skb, nl_header, 0);
    
    return;
}


/**********************************************************************
 *  netlink_rx
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
void netlink_rx(struct sock *sk, int len)
{
	struct sk_buff *skb;
        
        DEBUG((DBG_FUNC_CALL | DBG_NETLINK), "netlink_rx \n");
        DEBUG(DBG_NETLINK, "netlink_rx(socket: %p, len: %d)\n", sk, len);

	if (nl_sock == NULL) 
        {
            printk("netlink_rx: nl_sock is NULL\n");
            return;
	}

	do 
        {
            if (rtnl_shlock_nowait())
                    return;
            while ((skb = skb_dequeue(&sk->receive_queue)) != NULL) 
            {
                DEBUG(DBG_NETLINK, "netlink_rx: got skb, len=%d\n", skb->len);
                DEBUG(DBG_NETLINK, "pid %d grps %d dstpid %d dstgrps %d\n", NETLINK_CB(skb).pid, 
                                    NETLINK_CB(skb).groups, NETLINK_CB(skb).dst_pid, NETLINK_CB(skb).dst_groups);
                netlink_rx_skb(skb); 
                
                DEBUG(DBG_NETLINK, "Freeing skbuff %p\n", skb);
                kfree_skb(skb);
            }
            up(&rtnl_sem);

	} while (nl_sock && nl_sock->receive_queue.qlen);
}


/****************************************************************
 *								*
 *    			End of NETLINK functions		*
 *								*
 ***************************************************************/


/***************************************************************
*								*
*    		Start of generic debugging functions		*
*								*
***************************************************************/


#if (NET_DEBUG & DBG_SKB)
/**********************************************************************
 *  dbg_dump_hex_assci
 *
 *  DESCRIPTION: Dump the content of buffer 'buf' to the screen. The
 *		 data is presented in hexadecimal and ASSCI form, 16
 *		 bytes on each line.
 *
 *  PARAMETERS:	 buf  - pointer to buffer to display
 *		 size - size of the buffer
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
static void dbg_dump_hex_assci (unsigned char *buf, int size)
{
    int linesize, i;
    unsigned char *p;

    while (size)
    {
	linesize = size > 16 ? 16 : size;
	size -= linesize;
	for (p = buf, i = 0; i < linesize; i++, p++)
	    printk (" %02X", *p);
	for (; i < 17; i++)
	    printk ("	");
	for (i = 0; i < linesize; i++, buf++)
	{
	    if (*buf > 31 && *buf < 128)
		printk ("%c", *buf);
	    else
		printk (".");
	}
	printk ("\n");
    }
}


/**********************************************************************
 *  dbg_check_area
 *
 *  DESCRIPTION: Check if a memory area contains all zero's. If not,
 *		 print a warning, the address that isn't zero and dump
 *		 the content of the rest of the area on the screen.
 *
 *  PARAMETERS:	 p	   - start address of the area to check
 *		 size	   - size of the area
 *		 area_name - name of the area that is displayed in the
 *			     warning message
 *
 *  RETURN:	 0 if the whole area contains zero's, otherwise 1.
 *
 **********************************************************************/
static int dbg_check_area (unsigned char *p, int size, char *area_name)
{
    while (size)
    {
	if (*p)
	{
	    printk ("WARNING: %s changed at %p!\n", area_name, p);
	    dbg_dump_hex_assci (p, size);
	    return 1;
	}
	p++;
	size--;
    }
    return 0;
}
#endif /* (NET_DEBUG & DBG_SKB) */


void do_hard_print(char *string)
{   

#ifdef CONFIG_POLDEBUG_PRINTK

       struct msg_data debug_msg; 

       debug_msg.data = string;
       debug_msg.length = strlen(string) + 1;
       debug_msg.max = 0;

       dev_write(DEVICE_DEBUG, &debug_msg);
      
#endif /* CONFIG_POLDEBUG_PRINTK */

}


/*
 * print_mac
 *
 * This function prints the given MAC address out.
 *
 */
void print_mac(unsigned char macaddress[ETH_ALEN])
{
        unsigned int i;
        
        printk(" MAC address: %02X", macaddress[0]);
        for (i = 1; i < ETH_ALEN; i++)
                printk(":%02X", macaddress[i]);
        printk(" \n");
}

/****************************************************************
 *								*
 *  		End of generic debugging functions              *
 *								*
 ***************************************************************/



