/* $Header$
*  prism_local_mvc.c: Intersil prism Ethernet driver for local MVC v2.
*
*  Based on skeleton ethernet driver by Donald Becker.
*  Based on previous prism ethernet driver by Martijn Schipper 2001.
*
*  Extended by Bas Muns, Intersil 2002.
*
*  Copyright 1993 United States Government as represented by the
*  Director, National Security Agency.
*  Copyright (C) 2001 Intersil Americas Inc.
*  Copyright (C) 2002 Intersil Americas Inc.
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/

/* Include files */
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <asm/irq.h>
#include <asm/arch/intctl.h> /* for uIRQ */

/* Intersil include files */
#include <linux_mvc_oids.h>

#include <asm/arch/blobcalls.h>
#include "prism.h"
#include "prism_mgt.h"
#include "prism_wds.h"

#ifdef TEST_ON_POLDHU
#include <asm/arch/serial.h>
#include <asm/arch/gpio.h>
#include "../../../lib/Inl/include/hardconfig.h"
#endif /* TEST_ON_POLDHU */

/* End of include files */

/* DEFINES */

#define QUICK_HACK

unsigned int src_seq = 0;

/* Comment this out to use HardConfig. */
#define USE_PDA

#define REQ_UNKNOWN (REQFRAMERX & REQFRAMERETURN & REQTRAP & REQFRAMEADD & REQSERVICE)

/* STATES */
#define PRISM_ETH_STOP  0
#define PRISM_ETH_HALT  1
#define PRISM_ETH_START 2
#define PRISM_ETH_RUN   3

/*
 * MACROS to get and set the MVC device nr and MVC nr
 */
/* Add the device type, located in the first 16 bits, to var */
#define ADD_DEV_TYPE(var, type) var = ((var & 0x0000FFFF) + (type << 16))
/* Add the device number, located in the last 16 bits, to var */
#define ADD_DEV_NR(var, dev_nr) var = ((var & 0xFFFF0000) + dev_nr)
/* Return the MVC device number, located in the last 16 bits */
#define GET_DEV_NR(var) (var & 0x0000FFFF)

/* defines for cache mask */
#ifdef TEST_ON_POLDHU
#define PRISM_CACHE_MASK 0x04000000
#else
#define PRISM_CACHE_MASK 0x04000000 /* Subject to change... */
#endif /* TEST_ON_POLDHU */

/* DEFINES */
/* Number of TX retries */
#define TX_NR_RETRIES	10

/* Offset in RAM between cache-able and non-cache-able RAM
   point to the same physical RAM location. */
#define MAP_CACHED(data)   (data |  PRISM_CACHE_MASK)
#define MAP_UNCACHED(data) (data &=  ~PRISM_CACHE_MASK)

static const char *version = "version 0.1.00";
/* The name of the card. Is used for messages and in the requests for irqs */
static const char* cardname = "Prism Embedded MVC v2 packet IF";

/* End of defines */

#ifdef TEST_ON_POLDHU
/* prototype that is only needed when testing this driver on the poldhu platform... */
static int set_output_powertable(struct net_device *dev);
#endif

/* Probe function is called from Space.c */
int prism_mvc_probe(struct net_device *dev);

extern const unsigned char *blobtool_find_element(int id, int skip);

/* Start of local mvc prototypes */
static int local_net_open(struct net_device *dev);
static void local_net_tx_timeout(struct net_device *dev);
static int local_net_send_packet(struct sk_buff *skb, struct net_device *dev);
static int local_send_mvc_msg(struct net_device *dev, int operation, int oid, long* data, unsigned long data_len);
static int local_dev_associate_hndl( unsigned int dev_type, unsigned int dev_seq,
                                     unsigned int src_id,   unsigned int src_seq,
                                     unsigned int op,       unsigned int oid, void *data, unsigned long len);
static int local_dev_disassociate_hndl( unsigned int dev_type, unsigned int dev_seq,
                                        unsigned int src_id,   unsigned int src_seq,
                                        unsigned int op,       unsigned int oid, void *data, unsigned long len);
static int local_send_mvc_msg_hndl( unsigned int dev_type, unsigned int dev_seq,
                                    unsigned int src_id,   unsigned int src_seq,
                                    unsigned int op,       unsigned int oid, void *data, unsigned long len);
static int local_dev_authenticate_hndl( unsigned int src_id, unsigned int src_seq,
                                     unsigned int dev_id,   unsigned int dev_seq,
                                     unsigned int op,       unsigned int oid, void *data, unsigned long len);
static int local_dev_associate_hndl( unsigned int src_id, unsigned int src_seq,
                                     unsigned int dev_id,   unsigned int dev_seq,
                                     unsigned int op,       unsigned int oid, void *data, unsigned long len);
static void local_net_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static int local_net_close(struct net_device *dev);
static int local_set_mac_addr(struct net_device *dev, void *addr);
static int local_get_mvc_msg(struct net_device *dev, int oid, long *data, long size);
static int local_set_mvc_msg(struct net_device *dev, int oid, long *data, long data_len);
static struct net_device_stats *local_net_get_stats(struct net_device *dev);
static int prism_eth_init(struct net_device *dev, int startp);
static int prism_eth_start ( struct net_device *dev );
static int prism_eth_run ( struct net_device *dev );
/* End of local_mvc prototypes */

/* Begin of frame prototypes */
static int  alloc_frame_pool(struct net_local *lp);
static inline void req_frame_rx(struct net_device *dev);
static inline void req_frame_return(struct net_device *dev);
static inline void req_trap_handling(struct net_device *dev);
static inline void req_frame_add(struct net_device *dev);
static inline void req_service(struct net_device *dev);
static void local_dev_handle_reqs(struct net_device *dev, int requests);
static struct msg_frame *skb_to_frame (const struct sk_buff *skb);
static struct sk_buff *frame_to_skb (struct msg_frame *framep);
static struct sk_buff *dev_copy_expand_skb (struct net_device *dev, const struct sk_buff *skb);
static struct msg_frame *dev_alloc_frame (struct net_device *dev, int size);
static void dev_free_frame (struct net_device *dev, struct msg_frame *frame);
/* End of frame prototypes */

#ifdef USE_PDA
static unsigned char * pda_get_mac_address(void);
#endif

/* Begin of debug function prototypes */
#if (NET_DEBUG & DBG_SKB)
static int dbg_frame_filter_check (struct net_device *dev, struct msg_frame *framep);
#endif
#if (NET_DEBUG & (DBG_TX | DBG_RX | DBG_ERROR))
static void dbg_print_frame (struct msg_frame *framep);
#endif
/* End of debug function prototypes */

/* FIXME, temp here for character driver... */
extern void handle_blob_reqs(struct net_device *dev);

/****************************************************************
 *								*
 *    		    Begin of functions		                *
 *								*
 ***************************************************************/

#ifdef TEST_ON_POLDHU
/**********************************************************************
 *  set_output_powertable
 *
 *  Sets output power table to the MVC. Only here to be compatible when testing
 *  this driver on POLDHU platform.
 *
 **********************************************************************/
static int set_output_powertable(struct net_device *dev)
{
    struct obj_outputpowertable outpwrtable;
    long MVC_OutputPower;
    int error = BER_NONE;

    /* Set the output power structure */
    outpwrtable.outputpowertable = (long *)&HardConfig->CalibrateRows;
    if(HardConfig->DefaultRegDomain == 0x40)
        outpwrtable.index = DOT11_OUTPWRTABLE_MKK;
    else
        outpwrtable.index = DOT11_OUTPWRTABLE_FCC;

    if ((error = local_set_mvc_msg(dev, DOT11_OID_OUTPUTPOWERTABLE, (long*)&outpwrtable, sizeof(outpwrtable))) < BER_NONE)
        return error;

    DEBUG(DBG_OPEN, "OUTPUTPOWERTABLE set..");

    /* Set the desired output power to the MVC */
    if (HardConfig->OutputPower == 0xffffffff)
        MVC_OutputPower = 5000;
    else
        MVC_OutputPower = (HardConfig->OutputPower / 20);

    if ((error = local_set_mvc_msg(dev, DOT11_OID_OUTPUTPOWER, &MVC_OutputPower, sizeof(int))) < BER_NONE)
        return error;

    return error;

}
#endif /* TEST_ON_POLDHU */
/* End of POLDHU testing specific functions... */


/****************************************************************
 *								*
 *    		Begin of driver configurator functions		*
 *								*
 ***************************************************************/


/**********************************************************************
 *  prism_mvc_probe
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 *  NOTE:        Called from Space.c
 *
 **********************************************************************/
int prism_mvc_probe(struct net_device *dev)
{
    struct bis_dev_element *dev_element;
    int    nr_mvc_devices;  /* Nr of devices present on the MVC. */
    unsigned char *mac_address;
    char type;
    int i;

    struct bis_dev_descr_element *dev_desc = NULL;
    struct net_local *lp = NULL;
    int dev_found = FALSE;
    int dev_first_802_3_found = FALSE;

    DEBUG((DBG_FUNC_CALL | DBG_OPEN), "prism_mvc_probe \n");

    /* Allocate a new 'dev' if needed. */
    if (dev == NULL)
    {
	/*
	 * Don't allocate the private data here, it is done later
	 * This makes it easier to free the memory when this driver
	 * is used as a module.
	 */
	dev = init_etherdev(0,0);
	if (dev == NULL)
	    return -ENOMEM;
    }

    /* Get the number of devices available in the MVC. */
    dev_element = (struct bis_dev_element *) blobtool_find_element(BIS_DEVS, 0);

#ifdef TEST_ON_POLDHU
    nr_mvc_devices = dev_element->dev;  /* MVC v1 */
#else
    nr_mvc_devices = dev_element->devs; /* MVC v2 */
#endif /* TEST_ON_POLDHU */

    DEBUG(DBG_OPEN, "Nr of MVC devices = %i \n", nr_mvc_devices);

    /*
       We have the following (static) mapping for the devices. This mapping ensures
       that the devices get the same device name, even when another device isn't available.

       eth0 - LAN ethernet interface. (1)
       eth1 - WLAN interface. (2)
       eth2 - WAN ethernet interface. This interface is a hardware option. (3)

       We know for which device this probe function is called because the dev->base_addr
       is filled in Space.c with the corresponding number.
    */
    if (dev->base_addr == 1)      /* Probe for the LAN interface */
        type = BIS_DT_802DOT3;
    else if (dev->base_addr == 2) /* Probe for the WLAN interface */
        type = BIS_DT_802DOT11;
    else if (dev->base_addr == 3) /* Probe for the WAN interface */
        type = BIS_DT_802DOT3;
    else
    {
        printk(KERN_WARNING "We're asked to probe for a device (nr %lx) we don't know.\n", dev->base_addr);
        return -ENODEV;
    }

    /* Loop through the list and find the types of devices available... */
    for (i = 0; i < nr_mvc_devices && !dev_found; i++)
    {
        dev_desc = (struct bis_dev_descr_element *) blobtool_find_element (BIS_DEVDESCR, i);

        if (!dev_desc)
        {
            printk(KERN_ALERT "No dev_desc found!!!");
            return -ENODEV;
        }

        /* Let's find out if the MVC accomodates this type... */
        if (dev_desc->type == type)
        {
            /* For the WAN interface, we need the second device, so skip this device when we find the first... */
            if (!dev_first_802_3_found && dev->base_addr == 3)
            {
                dev_first_802_3_found = TRUE;
                continue;
            }

            /* The MVC has a device with the right type for us... */
            DEBUG(DBG_OPEN, "Found MVC device with type nr %i to use for %s \n", i, dev->name);

            /* We (ab) use the base_addr field for storing the device number and device type for later reference. */

            ADD_DEV_TYPE(dev->base_addr, type);
            ADD_DEV_NR(dev->base_addr, i);

            /* Initialize the device structure. */
            if (dev->priv == NULL)
            {
        	dev->priv = kmalloc(sizeof(struct net_local), GFP_KERNEL);
        	if (dev->priv == NULL)
        	    return -ENOMEM;
            }
            memset(dev->priv, 0, sizeof(struct net_local));
            lp = (struct net_local *)dev->priv;

            /* Set max mtu, head and tail sizes for this MVC packet device */
            lp->maxhead  = (dev_desc->header + 3) & ~3;
            lp->maxtail  = dev_desc->trailer;
            lp->maxmtu   = dev_desc->mtu;

            lp->maxnrframes = 100; /* FIXME, this should be part of MVC device descr */
            lp->dev_desc = dev_desc;

            DEBUG(DBG_OPEN, "dev_descr: nr: %d, type: %d, mtu: %d, head: %d, tail: %d\n",
		            lp->dev_desc->nr, lp->dev_desc->type, lp->dev_desc->mtu,
		            lp->dev_desc->header, lp->dev_desc->trailer);

            dev_found = TRUE;
        }
    }

    if (!dev_found)
    {
        printk("Couldn't find an MVC device type (%ld) for %s \n", dev->base_addr, dev->name);
        return -ENODEV;
    }
    else
        printk(KERN_INFO "%s: %s %s found,", dev->name, cardname, version );

    /* For both ethernet interfaces (LAN & WAN) and the onboard radio the same MAC address is used.
       The MAC address of the WAN interface can be changed later to a different MAC address
       to support MAC cloning for the WAN interface. */

#ifdef USE_PDA
    /* Get the MAC address from the PDA. Function returns pointer to MACAddress. */
    mac_address = pda_get_mac_address();
#else /* Use hardconfig */
    mac_address = HardConfig->MACAddress;
#endif /* USE_PDA */

    if (mac_address == NULL) /* If we couldn't get the MAC address from the PDA, exit... */
    {
        printk(KERN_ERR "Couldn't get MAC address for %s from PDA... aborting... \n", dev->name);
        kfree(lp);
        return -ENODEV;
    }

    memcpy(dev->dev_addr, mac_address, ETH_ALEN);
    print_mac(mac_address);

    dev->open		    = local_net_open;
    dev->stop		    = local_net_close;
    dev->get_stats	    = local_net_get_stats;
    dev->hard_start_xmit    = local_net_send_packet;
    dev->irq                = IRQ_SOFINT;
#ifdef HAVE_TX_TIMEOUT
    dev->tx_timeout	    = local_net_tx_timeout;
    dev->watchdog_timeo	    = HZ;
#endif
#ifdef HAVE_SET_MAC_ADDR
    dev->set_mac_address    = local_set_mac_addr;
#endif

    //FIXME, test for callback func! Will be replaced for other mechanism.
    lp->send_mvc_msg        = local_send_mvc_msg;

    /* Allocate the frame pool */
    if (alloc_frame_pool(lp) < 0)
    {
        kfree(lp);
        return -ENOMEM;
    }

    /* Fill in the fields of the device structure with ethernet values. */
    ether_setup(dev);

    mgt_init();
    mgt_indication_handler ( DEV_NETWORK, dev->ifindex, 0, local_send_mvc_msg_hndl );
    printk(KERN_ERR "mgt_indication_handler: %d\n", dev->ifindex );
    /*
    mgt_indication_handler ( DEV_NETWORK, dev->ifindex, DOT11_OID_ASSOCIATE, local_dev_associate_hndl );
    mgt_indication_handler ( DEV_NETWORK, dev->ifindex, DOT11_OID_DISASSOCIATE, local_dev_disassociate_hndl );
    */
    /* FIXME, temp only. */
#ifdef QUICK_HACK
    mgt_response_handler ( 0, 0, DOT11_OID_AUTHENTICATE, local_dev_authenticate_hndl );
    mgt_response_handler ( 0, 0, DOT11_OID_ASSOCIATE, local_dev_associate_hndl );
#endif

    /* Let's bring the device in the READY state */
    if (prism_eth_init( dev, PRISM_ETH_START) < 0 )
    {
        kfree(lp);
        printk(KERN_ERR "Couldn't start device %s \n", dev->name);
        return -EBUSY;
    }

    return 0;
}


/**********************************************************************
 *  prism_eth_init
 *
 *  DESCRIPTION: Initialise the MVC device to a specific state.
 *
 *  PARAMETERS:	 Device struct of the Ethernet device
 * 		 The operation to perform on the MVC
 *
 * RETURN:	 Possible error code
 *
 *  NOTE:	 Depending on the current state some values should
 * 		 possibly be reinitialised in the MVC.
 *
 **********************************************************************/
static int prism_eth_init( struct net_device *dev, int startp )
{
    long state;

    long dev_nr = GET_DEV_NR(dev->base_addr);
    int error  = 0;

    DEBUG(DBG_FUNC_CALL | DBG_OPEN, "prism_eth_init: base_addr: %ld \n", dev_nr);

    /* first make sure the device is in ready state, we don't know in what
       state we come here
     */

    state = dev_state (dev_nr);
    DEBUG(DBG_OPEN, "dev_state: %ld...", state);
    switch (startp)
    {
	case PRISM_ETH_STOP:  /* MVC State will end up being STSTOPPED */
	    /*
	     * In this case state doesn't really matter,
	     * if not already stopped stop now!
	     */
	    if(state != STSTOPPED)
	    {
		DEBUG(DBG_OPEN, "dev_stop...");
		error = dev_stop (dev_nr);
	    }
	break;
	case PRISM_ETH_HALT:  /* MVC State will end up being STREADY */
	    /*
	     * Man can only HALT a device if it's RUNNING.
	     */
	    if(state == STRUNNING)
	    {
		DEBUG(DBG_OPEN, "dev_halt...");
		error = dev_halt (dev_nr);
	    }
	    else
	    {
		error = -EBUSY;
	    }
	break;
	case PRISM_ETH_START: /* MVC State will end up being STREADY */
	    /*
	     * Man can only START a device if it's STOPPED.
	     */
	    if(state == STSTOPPED)
	    {
		DEBUG(DBG_OPEN, "dev_start...");

#ifdef TEST_ON_POLDHU
                error = dev_start(dev_nr, 0);
#else
                /* We can tell the MVC what device number we want to use. Use the one we figured out during the probe... */
                error = dev_start(dev_nr, dev_nr, 0);
#endif /* TEST_ON_POLDHU */

                if (error == BER_NONE)
   	            error = prism_eth_start ( dev );
	    }
	    else
	    {
		error = -EBUSY;
	    }
	break;
	case PRISM_ETH_RUN:   /* MVC State will end up being RUNNING */
	    /*
	     * Man can only RUN a device if it's READY to run.
	     */
	    if(state == STREADY)
	    {
		DEBUG(DBG_OPEN, "dev_run...");
		error = prism_eth_run ( dev );
		if (error == 0)
   	            error = dev_run (dev_nr);
	    }
	    else
	    {
		error = -EBUSY;
	    }
	break;
        default:
	    error = -EINVAL;
    }
    if (error != 0)
    {
        DEBUG((DBG_OPEN | DBG_ERROR), "error: %d\n", error);
        return -EINVAL;
    }

#if (NET_DEBUG & DBG_OPEN)
    else
    {
	state = dev_state (dev_nr);
	printk ("dev_state changed to: %ld...", state);
    }
#endif

    return 0;
}

/**********************************************************************
 *  prism_eth_start
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static int prism_eth_start ( struct net_device *dev )
{
    int requests;
    long mode;
    int error;

    DEBUG(DBG_FUNC_CALL, "prism_eth_start \n");

    /* set the MAC address */
    if (local_set_mvc_msg(dev, GEN_OID_MACADDRESS, (long*)dev->dev_addr, ETH_ALEN) < BER_NONE)
        return -EINVAL;

    DEBUG(DBG_OPEN, "MAC set..");

    /* for the WLAN device, set the MLME mode */
    //FIXME snmp or paed must set the MLME mode.
    if (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT11)
    {
        mode = DOT11_MLME_ADVANCED;
        if ((error = local_set_mvc_msg(dev, DOT11_OID_MLMEAUTOLEVEL, &mode, sizeof(long))) < 0)
            return error;

  	DEBUG(DBG_OPEN, "MLME mode set..");
    }

    /* handle all device requests (= add frames) */
    while ((requests = dev_reqs (GET_DEV_NR(dev->base_addr))) > BER_NONE)
    {
	local_dev_handle_reqs (dev, requests);
    }
    DEBUG(DBG_OPEN, "requests handled..");

    return 0;
}


/**********************************************************************
 *  prism_eth_run
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static int prism_eth_run ( struct net_device *dev )
{
    DEBUG(DBG_FUNC_CALL, "prism_eth_run \n");

#ifdef TEST_ON_POLDHU
    /* Output power for MVC v2 is set via datastore... */
    if (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT11)
        set_output_powertable(dev);
#endif

#if (defined(TEST_ON_POLDHU) && !defined(CONFIG_FREE_WLAN_PINS))
    /* if we are starting the WLAN device, then... */
    if (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT11)
    {
	/* don't disable the UART completely, this crashes the system ?! */
	uUART->CR &= ~0x78;	/* disable UART interrupts */
	/* make the UART pins free for the radio */
	uGPIO->FSEL = (GPIOFSELUARTTx | GPIOFSELUARTRx) << 16;
    }
#endif /* TEST_ON_POLDHU && CONFIG_FREE_WLAN_PINS */

    return 0;
}


/**********************************************************************
 *  local_net_open
 *
 *  DESCRIPTION:  Open/initialize the board. This is called (in the current kernel)
 *                sometime after booting when the 'ifconfig' program is run.
 *                This routine should set everything up anew at each open, even
 *                registers that "should" only need to be set once at boot, so that
 *                there is non-reboot way to recover if something goes wrong.
 *
 *  PARAMETERS:	 pointer dev to net_device
 *
 *  RETURN:	 -EBUSY when unable to get IRQ, 0 on success.
 *
 **********************************************************************/
static int local_net_open(struct net_device *dev)
{
    int retval;
    long linkstate;

    DEBUG(DBG_FUNC_CALL, "local_net_open \n");

    /* Reset the hardware and set the card MAC address. */
    if (prism_eth_init(dev, PRISM_ETH_RUN) < 0)
    {
        printk("Couldn't reset hardware \n");
        return -EBUSY;
    }

    netif_start_queue (dev);

    retval = request_irq(dev->irq, &local_net_interrupt, SA_SHIRQ | SA_INTERRUPT,
		    dev->name, dev);
    if (retval)
    {
	printk("%s: unable to get IRQ %d (retval=%d).\n",
	       dev->name, dev->irq, retval);
	netif_stop_queue (dev);
        return -EBUSY;
    }

    /* determine initial linkstate for Ethernet interface (LAN or WAN) */
    if (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT3)
    {
	local_get_mvc_msg(dev, GEN_OID_LINKSTATE, &linkstate, sizeof(long));
        if (linkstate == 0)
            netif_carrier_off(dev);
    }

    /* If the current device is host device for WDS devices, bring the WDS devices up and running. */
    if (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT11)
        open_wds_links(dev);

    return 0;
}


/**********************************************************************
 *  local_net_tx_timeout
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
#ifdef HAVE_TX_TIMEOUT
static void local_net_tx_timeout(struct net_device *dev)
{
    long linkstate;

    struct net_local *lp = (struct net_local *)dev->priv;

    DEBUG(DBG_FUNC_CALL, "local_net_tx_timeout \n");

    local_get_mvc_msg(dev, GEN_OID_LINKSTATE, &linkstate, sizeof(long));

    if (linkstate == 0)
    {
	printk(KERN_WARNING "%s: link down, network cable problem?\n", dev->name);

	netif_carrier_off (dev);
    }
    else
    {
	printk(KERN_WARNING "%s: reset interface\n", dev->name);

	/* Try to restart the device. */
        prism_eth_init(dev, PRISM_ETH_STOP);
        prism_eth_init(dev, PRISM_ETH_START);
        prism_eth_init(dev, PRISM_ETH_RUN);
	lp->stats.tx_errors++;

        netif_wake_queue(dev);
    }
}
#endif /* HAVE_TX_TIMEOUT */


/**********************************************************************
 *  local_net_send_packet
 *
 *  DESCRIPTION: This function puts an skbuf into an MVC frame,
 *               updates the statistics and sends the frame to the local MVC.
 *
 *  PARAMETERS:	 skb - pointer to a linux packet buffer.
 *               dev - pointer to a linux net device.
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static int local_net_send_packet(struct sk_buff *skb, struct net_device *dev)
{
    //FIXME, still needed? Put in dev->priv.
    static int tries = 1000000; /* give the first packet a lot of retries */

    unsigned char *src;
    struct sk_buff *skb2;
    struct msg_frame *framep;
    int error;
    int WDSLink;
    long linkstate;
    unsigned char WDS_Mac[ETH_ALEN];

    struct net_local *lp = (struct net_local *)dev->priv;
    long dev_nr = GET_DEV_NR(dev->base_addr);

    DEBUG((DBG_FUNC_CALL), "local_net_send_packet \n");

    /* If this skb came from a WDS device, the
       skb->cb field will be filled with a WDSLink MACaddress.
       Since the skb->cb field won't be cloned, we check it here. */
    WDSLink = check_skb_for_wds(skb, WDS_Mac);

    /* if the head or tail room doesn't comply with the device or
       the skb is cloned .. */
    if (skb_headroom (skb) < lp->dev_desc->header ||
	skb_tailroom (skb) < lp->dev_desc->trailer ||
	skb_cloned (skb))
    {
#if (NET_DEBUG & DBG_TX)
        if (!skb_cloned (skb))
	{
	    DEBUG(DBG_TX, "no head/tail room, head:(%d,%d),tail(%d,%d)..",
		skb_headroom (skb), lp->dev_desc->header,
		skb_tailroom (skb), lp->dev_desc->trailer);
	}
#endif
	/* see if we can shift the data in the buffer */
	if (!skb_cloned (skb) &&
	    skb_headroom (skb) + skb_tailroom (skb) >=
	    lp->dev_desc->header + lp->dev_desc->trailer)
	{
	    _DEBUG(DBG_TX, "shift data: %d..",
		    lp->dev_desc->header - skb_headroom (skb));

	    src = skb->data;
	    skb_reserve (skb, lp->dev_desc->header - skb_headroom (skb));
	    memmove (skb->data, src, skb->len);
            lp->tx_misaligned++;
	}
	else	/* copy to a new buffer with enough head and tail room */
	{
#if (NET_DEBUG & DBG_TX)
	    if (skb_cloned (skb))
		printk ("cloned sk_buff.. ");
	    printk ("copy data.. ");
#endif
	    skb2 = dev_copy_expand_skb (dev, skb);
	    dev_kfree_skb (skb);
	    if (skb2 == NULL)
            {
                lp->stats.tx_dropped++;
		return 0;
            }
	    skb = skb2;
	}
    }

    if ((framep = skb_to_frame (skb)) == NULL)
    {
	DEBUG((DBG_TX | DBG_ERROR), "NO frame!!!\n");
	dev_kfree_skb (skb);
	lp->stats.tx_dropped++;
        return 0;
    }

    /* send the frame to the local MVC */

    _DEBUG(DBG_TX, "send: size: %d, headroom: %d..", framep->size,
            skb_headroom (framep->handle));

    /* Copy the found WDS MAC address in the msg_frame */
    if (WDSLink == 1) copy_wds_to_frame(framep, WDS_Mac);

    //FIXME, CHECK the next statement for MVC v2!!!
    do
    {
        error = dev_frame_tx (dev_nr, framep);
        if (error < BER_NONE)
            dev_service (dev_nr); /* Give the device some time to process */
        tries--;
    }
    while (error < BER_NONE &&
           error != BER_DSTUNKN &&
           framep->status != FSTXREJECTED && /* If the framep->status == FSTXREJECTED the data could
                                                have been altered/shifted in the MVC. We don't want to
                                                send the data again when this happened! */
           tries);
    tries = TX_NR_RETRIES;

    if (error < BER_NONE)
    {
        if (error != BER_DSTUNKN)
        {
#if !defined(CONFIG_FREE_WLAN_PINS) && defined(TEST_ON_POLDHU)
                DEBUG((DBG_TX | DBG_ERROR),
			KERN_ALERT "TX dev: %s, NOT accepted!!\n", dev->name);
#endif
            /* test if the link is down */
            local_get_mvc_msg(dev, GEN_OID_LINKSTATE, &linkstate, sizeof(long));
            if (linkstate == 0)
                netif_carrier_off (dev);
            else
            /* if it's not accepted, then stop queing packets */
                netif_stop_queue (dev);

        lp->stats.tx_dropped++;
        }

    /* abort the frame */
    dev_free_frame (dev, framep);
    }
    else
        _DEBUG((DBG_TX | DBG_WDS), "OK!!\n");

    dev->trans_start = jiffies;

    return 0;
}


/**********************************************************************
 *  local_net_interrupt
 *
 *  DESCRIPTION: The typical workload of the driver:
 *               Handle the network interface interrupts
 *               XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static void local_net_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
    int			requests;

    struct net_device	*dev = NULL;

    uIRQ->Soft = 0;	    /* remove soft irq */

    DEBUG((DBG_FUNC_CALL | DBG_IRQ), "local_net_interrupt \n");

    if (dev_id == NULL)
    {
	printk(KERN_WARNING "%s: irq %d for unknown device.\n", cardname, irq);
	return;
    }

    dev = (struct net_device *) dev_id;

    /* We handle blob requests within the Ethernet device */
    if (GET_DEV_TYPE(dev-> base_addr) == BIS_DT_802DOT3)
	handle_blob_reqs(dev);

    if ((requests = dev_reqs (GET_DEV_NR(dev->base_addr))) > BER_NONE)
	local_dev_handle_reqs (dev, requests);

    return;
}


/**********************************************************************
 *  local_net_close
 *
 *  DESCRIPTION: The inverse routine to local_net_open().
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static int local_net_close(struct net_device *dev)
{
    struct msg_frame *framep;

    long dev_nr = GET_DEV_NR(dev->base_addr);

    DEBUG(DBG_FUNC_CALL, "local_net_close \n");

    /* If the current device is host device for WDS devices,
       bring the associated WDS devices DOWN. */
    if (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT11)
        close_wds_links(dev);

    netif_stop_queue(dev);

    if (dev_halt(dev_nr) >= BER_NONE)
    {
        while ((framep = dev_frame_return(dev_nr)) != NULL)
            dev_free_frame(dev, framep);
    }

    free_irq(dev->irq, dev);

    return 0;
}


/****************************************************************
 *								*
 *    		Begin of frame handling block			*
 *              (driver configurator module)			*
 *								*
 ***************************************************************/


/**********************************************************************
 *  req_frame_rx
 *
 *  DESCRIPTION: Handle the frame rx requests.
 *
 *  PARAMETERS:	 XXX.
 *
 *  RETURN:	 Nothing.
 *
 *  NOTE:	 Inline function.
 *
 **********************************************************************/
static inline void req_frame_rx(struct net_device *dev)
{
    struct msg_frame        *framep;
    struct net_device       *wds_dev;
    struct wds_net_local    *wds_lp;
    struct sk_buff          *skb;

    struct net_local        *lp = (struct net_local *) dev->priv;

    if ((framep = dev_frame_rx (GET_DEV_NR(dev->base_addr))) != NULL)
    {
        if (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT11)
        {
            /* Check if this frame contains one of the WDS addresses */
            wds_dev = wds_find_device((msg_frame*)framep, dev);
        }
        else
            wds_dev = NULL;

        dev->last_rx = jiffies;
        if (framep->size < 1569 && framep->status < 10)
        {

            _DEBUG(DBG_IRQ, "..RX");
#if (NET_DEBUG & DBG_RX)
            dbg_print_frame (framep);
#endif
#if (NET_DEBUG & DBG_SKB)
            if (dbg_frame_filter_check (dev, framep))
#endif
            {
                skb = frame_to_skb (framep);
                skb->protocol = eth_type_trans( skb, dev );
                lp->stats.rx_packets++;
                lp->stats.rx_bytes += skb->len;

                /* This will change in the new WDS implementation. */
                if (wds_dev != NULL)
                {
                    wds_lp = (struct wds_net_local *) wds_dev->priv;
                    wds_lp->stats.rx_packets++;
                    wds_lp->stats.rx_bytes += skb->len;
                    skb->dev = wds_dev;
                }
                /* Give the data to the Linux kernel for further processing. */
                netif_rx(skb);
            }
        }
        else
            printk (KERN_ERR " RX, size %d, status %d!!\n",
                    framep->size, framep->status);
    }

}


/**********************************************************************
 *  req_frame_return
 *
 *  DESCRIPTION: Handle the frame return requests.
 *
 *  PARAMETERS:	 XXX.
 *
 *  RETURN:	 Nothing.
 *
 *  NOTE:	 Inline function.
 *
 **********************************************************************/
static inline void req_frame_return(struct net_device *dev)
{
    struct msg_frame    *framep;

    struct net_local        *lp = (struct net_local *) dev->priv;

    if ((framep = dev_frame_return (GET_DEV_NR(dev->base_addr))) != NULL)
    {
        _DEBUG(DBG_IRQ, "..RET");
        if (framep->status <= FSRETURNED) /* Check these defines carefully with frame state defines.  */
        {
            if (framep->status <= FSTXSUCCESSFUL)
            {
                if (framep->status == FSTXSUCCESSFUL)
                {
                    lp->stats.tx_packets++;
                    lp->stats.tx_bytes += framep->size;
                }
                else
                {
                    printk (KERN_ERR " TX error!!. status %d\n",
                            framep->status);
                    lp->stats.tx_errors++;
                }
                netif_wake_queue (dev);
            }
            dev_free_frame (dev, framep);
        }
        else
            printk (KERN_ERR " RET, size %d, status %d!!\n",
                    framep->size, framep->status);
    }
}


/**********************************************************************
 *  req_trap_handling
 *
 *  DESCRIPTION: Handle traps from the MVC.
 *
 *  PARAMETERS:	 XXX.
 *
 *  RETURN:	 Nothing.
 *
 *  NOTE:	 Inline function.
 *
 **********************************************************************/
static inline void req_trap_handling(struct net_device *dev)
{
    struct msg_conf     *trapmsg;

    _DEBUG(DBG_IRQ, "..TRAP");

    /* Get the trap from the MVC... */
    if ((trapmsg = dev_trap (GET_DEV_NR(dev->base_addr))) != NULL)
    {
        /* This is an embedded MVC specific trap, so we handle it partly here... */
        if (trapmsg->oid == GEN_OID_LINKSTATE && (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT3))
        {

#ifdef TEST_ON_POLDHU
            /* Set/Clear the Ethernet LED */
            long mask = 1;
            struct msg_conf config;

            config.operation = OPSET;
            config.data = &mask;

            if( *trapmsg->data ) /* LINK State is UP... */
            {
                config.oid = BLOB_OID_LEDSET;
                printk ("%s: link up\n", dev->name);
                netif_carrier_on (dev);
            }
            else                 /* LINK State is DOWN... */
            {
                config.oid = BLOB_OID_LEDCLEAR;
                printk ("%s: link down\n", dev->name);
                netif_carrier_off (dev);
            }

            blob_conf( &config ); /* For setting/clearing the led */

#else
            {
                //FIXME, GPIO banks are used instead of LED/BUTTON defines. Check!
            }

#endif /* TEST_ON_POLDHU */

        }

        /* We handle all other traps in the generic part, and also send the GEN_OID_LINKSTATE up */
#ifdef TEST_ON_POLDHU

        mgt_response( 0, src_seq++, DEV_NETWORK, dev->ifindex,
           PIMFOR_OP_TRAP, trapmsg->oid, trapmsg->data, 100 );

#else

        mgt_response( 0, src_seq++, DEV_NETWORK, dev->ifindex,
           PIMFOR_OP_TRAP, trapmsg->oid, trapmsg->data, trapmsg->size);

#endif
    }
}


/**********************************************************************
 *  req_frame_add
 *
 *  DESCRIPTION: Add frames to the device.
 *
 *  PARAMETERS:	 XXX.
 *
 *  RETURN:	 Nothing.
 *
 *  NOTE:	 Inline function.
 *
 **********************************************************************/
static inline void req_frame_add(struct net_device *dev)
{
    struct msg_frame    *framep;

    _DEBUG((DBG_IRQ | DBG_FRAME), "..ADD");
	if ((framep = dev_alloc_frame (dev, 0)) != NULL)
	{
	    _DEBUG(DBG_IRQ, "dev_alloc_frame returned %p pointer..", framep);
	    dev_frame_add (GET_DEV_NR(dev->base_addr), framep);
	}
	else
	    printk(KERN_WARNING "dev_alloc_frame returned NULL pointer.\n");
}

/**********************************************************************
 *  req_service
 *
 *  DESCRIPTION: Service the device.
 *
 *  PARAMETERS:	 XXX.
 *
 *  RETURN:	 Nothing.
 *
 *  NOTE:	 Inline function.
 *
 **********************************************************************/
static inline void req_service(struct net_device *dev)
{
   _DEBUG(DBG_IRQ, "..service");
	dev_service (GET_DEV_NR(dev->base_addr));
	dev_service (DEVICE_DEBUG); /* We service the debug device here. */
}


/**********************************************************************
 *  local_dev_handle_reqs
 *
 *  DESCRIPTION: Handle requests 'requests' for device 'dev'.
 *
 *  PARAMETERS:	 dev	  - device to handle requests for
 *		 requests - requests to handle
 *
 *  RETURN:	 Nothing.
 *
 *  NOTE:	 Also called from the interrupt service routine.
 *
 **********************************************************************/
static void local_dev_handle_reqs(struct net_device *dev, int requests)
{
    _DEBUG(DBG_IRQ, "dev_handle_reqs: dev %ld\n", GET_DEV_NR(dev->base_addr));

    if (requests & REQ_UNKNOWN)
    {
	printk (" called with unknown requests: %d!!\n", requests);
        return;
    }
    if (requests & REQFRAMERX)
    {
        req_frame_rx(dev);
    }
    if (requests & REQFRAMERETURN)
    {
	req_frame_return(dev);
    }
    if (requests & REQTRAP)
    {
        req_trap_handling(dev);
    }
    if (requests & REQFRAMEADD)
    {
        req_frame_add(dev);
    }
    if (requests & REQSERVICE)
    {
        req_service(dev);
    }
}


/**********************************************************************
 *  alloc_frame_pool
 *
 *  DESCRIPTION: Allocates a frame pool for the local MVC.
 *               The frame pool is allocated and the free frame list
 *               is allocated with these frames.
 *
 *  PARAMETERS:	 Pointer to the net_local structure lp.
 *
 *  RETURN:	 Zero on success, or -ENOMEM when malloc failed.
 *
 **********************************************************************/
static int alloc_frame_pool(struct net_local *lp)
{
    int i;

    DEBUG((DBG_FUNC_CALL | DBG_FRAME), "alloc_frame_pool \n");

    /* allocate frame pool */
    lp->frame_pool = kmalloc(sizeof(struct msg_frame) * lp->maxnrframes,
 			    GFP_KERNEL);
    if (lp->frame_pool == NULL)
	return -ENOMEM;

    /* add all frames to the free frame list */
    lp->frame_freelist = NULL;
    for (i = 0; i < lp->maxnrframes; i++)
    {
	lp->frame_pool[i].handle = lp->frame_freelist;
	lp->frame_freelist = &lp->frame_pool[i];
    }

    return 0;
}

/**********************************************************************
 *  frame_alloc
 *
 *  DESCRIPTION: Allocates a frame from the free frames list.
 *               The wds info in the frame is zeroed.
 *
 *  PARAMETERS:	 Pointer to net_device structure dev.
 *
 *  RETURN:	 framep pointer to allocated frame or NULL when alloc failed.
 *
 **********************************************************************/
static struct msg_frame *frame_alloc (struct net_device *dev)
{
    struct msg_frame *framep;
    unsigned long flags;
    msg_frame *wdsframep;

    DEBUG((DBG_FUNC_CALL | DBG_FRAME), "frame_alloc \n");

    /* We could change this for a spinlock mechanism */
    save_flags_cli (flags);

    framep = ((struct net_local *)dev->priv)->frame_freelist;
    if (framep)
	((struct net_local *)dev->priv)->frame_freelist = framep->handle;

    restore_flags(flags);

    /* Make sure the wds info in the frame is zeroed */
    if (framep)
    {
        wdsframep = (msg_frame*)framep;
#ifdef TEST_ON_POLDHU
        memset(wdsframep->wds_link, 0, ETH_ALEN);
#else
        memset(wdsframep->wds, 0, ETH_ALEN);
#endif /* TEST_ON_POLDHU */
    }

    return framep;
}


/**********************************************************************
 *  frame_free
 *
 *  DESCRIPTION: Frees the frame pointed to by the framep.
 *               It returns the framep to the free frames list.
 *
 *  PARAMETERS:	 Pointer to net_device structure dev.
 *               Pointer to msg_frame struct framep.
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
static void frame_free (struct net_device *dev, struct msg_frame *framep)
{
    unsigned long flags;

    _DEBUG((DBG_FUNC_CALL | DBG_FRAME), "frame_free \n");

    if (!framep)
        return;

    save_flags_cli (flags);

    framep->handle = ((struct net_local *)dev->priv)->frame_freelist;
    ((struct net_local *)dev->priv)->frame_freelist = framep;

    restore_flags(flags);
}


/**********************************************************************
 *  skb_to_frame
 *
 *  DESCRIPTION: Attach a prism frame structure to linux sk_buff 'skb'
 *		 and initialize all fields of the frame structure with
 *		 the current information of 'skb'.
 *
 *  PARAMETERS:	 skb - pointer to linux packet buffer.
 *
 *  RETURN:	 Pointer to the attached frame, or NULL if there wasn't
 *		 enough memory.
 *
 **********************************************************************/
static struct msg_frame *skb_to_frame (const struct sk_buff *skb)
{
    struct msg_frame *framep;

    DEBUG((DBG_FUNC_CALL | DBG_FRAME), "skb_to_frame \n");

    /* first malloc memory for msg_frame */
    framep = frame_alloc (skb->dev);
    if (framep != NULL)
    {
	framep->handle	 = (void *) skb;
	framep->mpdu	 = (unsigned char *) MAP_CACHED((unsigned)skb->data);
        framep->size	 = skb->len;
	framep->priority = -1; /* Priority isn't implemented yet.
                                  We should map the skb->priority to the frame priority here. */
    }

#if (NET_DEBUG & DBG_SKB)
    memset (skb->head, 0, skb_headroom(skb));
    memset (skb->tail, 0, skb_tailroom(skb));
#endif
     return framep;
}


/**********************************************************************
 *  frame_to_skb
 *
 *  DESCRIPTION: Update the fields of the linux sk_buff assigned to
 *		 frame 'framep' just received from the MVC.
  *		 The frame is freed.
 *
 *  PARAMETERS:	 framep - pointer to frame received from MVC.
 *
 *  RETURN:	 Pointer to the updated sk_buff that was assigned to the
 *		 frame.
 *
 **********************************************************************/
static struct sk_buff *frame_to_skb (struct msg_frame *framep)
{
     struct sk_buff *skb;
    int offset;
    unsigned char *src;

    DEBUG((DBG_FUNC_CALL | DBG_FRAME), "frame_to_skb \n");

    skb = framep->handle;

    /* first map pointer back in uncached area */
    MAP_UNCACHED((unsigned)framep->mpdu);

    /* determine shift in framep->mpdu between add and rx */
     offset = framep->mpdu - skb->data;

    skb_reserve (skb, offset);
    skb_put (skb, framep->size);

    /* determine offset between IP header and a word boundary */
    offset = (int) (skb->data + 6 + 6 + 2) & 0x03;

    /* if there is any, shift the data */
    if (offset)
    {
	struct net_local *lp = (struct net_local *)skb->dev->priv;
	lp->rx_misaligned++;

	if (skb_headroom (skb) < offset)
	{
	    if (skb_tailroom (skb) < (4 - offset))
		printk(KERN_NOTICE "Unaligned skb\n");
	    else
		offset = 4 - offset;
	}
	else
	    offset = -offset;
#if (NET_DEBUG & DBG_RX)
	printk ("frame_to_skb: need to shift data %d\n", offset);
#endif
	src = skb->data;
	skb_reserve (skb, offset);
	memmove (skb->data, src, skb->len);
    }
    skb->priority = -1; /* Priority isn't implemented yet. Add mapping between skb and frame priority here. */

    frame_free (skb->dev, framep);

    return skb;
}


/**********************************************************************
 *  dev_alloc_frame
 *
 *  DESCRIPTION: Allocate a frame and a sk_buff with MTU size 'size' for
 *		 device 'dev'. Set all fields of the frame and the
 *		 sk_buff. If size equals 0 then the maximum MTU size is
 *		 used. The head and the tail room for the frame (and
 *		 sk_buff) is equal to the maximum head and maximum tail
 *		 room for all devices.
 *
 *  PARAMETERS:	 dev  - device to allocate a frame for
 *		 size - the MTU size for the frame.
 *
 *  RETURN:	 Pointer to the allocated frame or NULL on error.
 *
 **********************************************************************/
static struct msg_frame *dev_alloc_frame (struct net_device *dev, int size)
{
    struct sk_buff *skb;
    struct msg_frame *framep;
    int bufsize;
    struct net_local *lp;

    DEBUG((DBG_FUNC_CALL | DBG_FRAME), "dev_alloc_frame \n");

    /* first allocate a sk_buff with enough head and tail room */
    lp = (struct net_local *)dev->priv;
    bufsize = lp->maxhead + 2 +
		    (size ? size : lp->maxmtu) + lp->maxtail;
    if ((skb = dev_alloc_skb (bufsize)) == NULL)
        return NULL;

    /* fill in some sk_buff fields */
    skb->dev = dev;
    skb_reserve (skb, lp->maxhead + 2); /* word align IP header */

    if ((framep = skb_to_frame(skb)) == NULL)
        dev_kfree_skb(skb);

    return framep;
}


/**********************************************************************
 *  dev_free_frame
 *
 *  DESCRIPTION: Free MVC frame 'frame' and the attached sk_buff.
 *
 *  PARAMETERS:	 frame - address of frame to free
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
static void dev_free_frame (struct net_device *dev, struct msg_frame *frame)
{
    DEBUG((DBG_FUNC_CALL | DBG_FRAME), "dev_free_frame \n");

    if (frame->handle)
	dev_kfree_skb_any (frame->handle);
    frame_free (dev, frame);
}


/**********************************************************************
 *  dev_copy_expand_skb
 *
 *  DESCRIPTION: Allocate a new sk_buff and copy the data of sk_buff
 *		 'skb'. The new sk_buff will have the required head and
 *		 tail room for device 'dev'.
 *
 *  PARAMETERS:	 dev - device to allocate a frame for
 *		 skb - the sk_buff to copy the data from
 *
 *  RETURN:	 Pointer to the new allocated sk_buff or NULL on error.
 *
 **********************************************************************/
static struct sk_buff *dev_copy_expand_skb (struct net_device *dev,
				const struct sk_buff *skb)
{
    int bufsize;
    struct sk_buff *newskb;
    struct net_local *lp = (struct net_local *)dev->priv;

    DEBUG((DBG_FUNC_CALL | DBG_FRAME), "dev_copy_expand_skb \n");

    bufsize = skb->len + lp->dev_desc->header + lp->dev_desc->trailer;
    if ((newskb = dev_alloc_skb (bufsize)) == NULL)
	return NULL;

    /* reserve head room */
    skb_reserve (newskb, lp->dev_desc->header);

    /* set the tail pointer and length */
    skb_put(newskb, skb->len);

    /* copy the data */
    memcpy(newskb->data, skb->data, skb->len);
    newskb->priority = skb->priority;

    newskb->dev = dev;
    return newskb;
}


/****************************************************************
*								*
*    		End of frame handling block                     *
*               (driver configurator module)			*
*								*
***************************************************************/


/****************************************************************
 *								*
 *    		End of driver configurator functions		*
 *								*
 ***************************************************************/


/****************************************************************
 *								*
 *    		Begin of management handling functions		*
 *								*
 ***************************************************************/


/**********************************************************************
 *  local_set_mac_addr
 *
 *  DESCRIPTION: This sets/changes the mac address on the interface.
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
#ifdef HAVE_SET_MAC_ADDR
static int local_set_mac_addr(struct net_device *dev, void *addr)
{
    unsigned char mac_addr[ETH_ALEN];
    unsigned char mac_null_addr[ETH_ALEN];

    memcpy(mac_addr, addr, ETH_ALEN);

    memset(mac_null_addr, 0, ETH_ALEN);

    /* A MAC address containing all zeroes is used to reset the MAC address to it's default. */
    if (memcmp(mac_addr, mac_null_addr, ETH_ALEN) == 0)
    {

#ifdef USE_PDA
    /* Get the MAC address from the PDA. Function returns pointer to MACAddress. */
    memcpy(mac_addr, pda_get_mac_address(), ETH_ALEN);
#else
    mac_addr[0] = HardConfig->MACAddress;
#endif /* USE_PDA */

        if (mac_addr == NULL)
        {
            printk(KERN_ERR "Couldn't get MAC address for %s from PDA...\n", dev->name);
            return -EINVAL;
        }
    }

    /* We don't allow multicast/broadcast MAC addresses. */
    if (mac_addr[0] != 0)
        return -EINVAL;

    /* Send a message to the MVC that we want to use this MAC address... */
    if (mgmt_send_mvc_msg(dev, OPSET, GEN_OID_MACADDRESS, (long*)mac_addr, ETH_ALEN) < 0);
    {
        printk("local_set_mac_addr: Couldn't change MAC address for dev %s ", dev->name);
        print_mac(mac_addr);
        return -EINVAL;
    }

    /* Okay, we succeeded, so use this MAC address from now on... */
    memcpy(dev->dev_addr, mac_addr, ETH_ALEN);
    DEBUG(DBG_OPEN | DBG_MGMT, "MAC address changed. ");
    print_mac(mac_addr);

    return 0;
}
#endif /* HAVE_SET_MAC_ADDR */


/**********************************************************************
 *  local_set_mvc_msg
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static int local_set_mvc_msg(struct net_device *dev, int oid, long *data, long data_len)
{
    struct msg_conf setmsg;

    setmsg.operation = OPSET;
    setmsg.oid       = oid;
    setmsg.data      = data;
#ifndef TEST_ON_POLDHU
    setmsg.size      = data_len; /* MVC v2 only */
#endif

    if (dev_conf(GET_DEV_NR(dev->base_addr), &setmsg) < BER_NONE)
        return -EINVAL;
printk("local_set_mvc_msg: SET OID %x\n", oid);
    return 0;
}


/**********************************************************************
 *  local_get_mvc_msg
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static int local_get_mvc_msg(struct net_device *dev, int oid, long *data, long size)
{
    struct msg_conf getmsg;

    getmsg.operation = OPGET;
    getmsg.oid       = oid;
#ifndef TEST_ON_POLDHU
    getmsg.size      = size; /* MVC v2 only */
#endif

    if (dev_conf (GET_DEV_NR(dev->base_addr), &getmsg) == EOVERFLOW)
    {
#ifdef TEST_ON_POLDHU
        printk("Size passed to MVC (%ld) not correct.\n", size);
#else
        printk("Size passed to MVC (%ld) not correct. MVC needs %ld\n", size, getmsg.size);
#endif
        return -1;
    }
printk("local_get_mvc_msg: GOT OID %x\n", oid);
    memcpy(data, getmsg.data, size);

    return 0;
}


/**********************************************************************
 *  local_net_get_stats
 *
 *  DESCRIPTION: Get the current statistics.
 *               This may be called with the card open or closed.
 *
 *  PARAMETERS:	 pointer to net_device dev
 *
 *  RETURN:	 struct net_device stats pointer
 *
 **********************************************************************/
static struct net_device_stats *local_net_get_stats(struct net_device *dev)
{
    struct net_local *lp = (struct net_local *)dev->priv;
    struct net_device_stats *stats = &lp->stats;
    long mvc_errors;

    DEBUG(DBG_FUNC_CALL, "net_get_stats \n");

    /*
        Most statistics are updated in the driver code.
        However, we can't update error counters for received packets that
        were dropped or erroneous, since we don't get those packets from the MVC.
        So, we ask the MVC for those errors...
    */
    if (!local_get_mvc_msg(dev, GEN_OID_MSDURXFAILED, &mvc_errors, sizeof(long)))
        stats->rx_errors += mvc_errors;

    if (!local_get_mvc_msg(dev, GEN_OID_MSDURXDROPPED, &mvc_errors, sizeof(long)))
        stats->rx_dropped += mvc_errors;

    return stats;
}


/**********************************************************************
 *  local_dev_authenticate
 *
 *  FIXME: temp here as long as authentication app isn't functioning yet.
 *
 **********************************************************************/
void local_dev_authenticate (struct net_device *dev, char *data)
{
   struct obj_mlme *obj = (struct obj_mlme*) data;

   DEBUG(DBG_MGMT, "local_dev_authenticate. \n");
   print_mac(obj->address);

   /* FIXME: since the authentication application is not ready yet, succeed always for now!
      FIXME: Remove this code when the authentication app is ready!!! */
   obj->code  = 0; // = DOT11_SC_SUCCESFUL;
   mgmt_send_mvc_msg(dev, OPSET, DOT11_OID_AUTHENTICATE, (long *) obj, sizeof(struct obj_mlme));

}


/* DOT11_OID_AUTHENTICATE */
static int local_dev_authenticate_hndl( unsigned int src_id, unsigned int src_seq,
                                     unsigned int dev_id,   unsigned int dev_seq,
                                     unsigned int op,       unsigned int oid, void *data, unsigned long len)
{
    struct net_device *dev;

    printk("local_dev_authenticate_hndl \n");

    if ( ! (dev = dev_get_by_index(dev_seq)) )
        return -1;

    local_dev_authenticate (dev, (char*) data);

    return 0;
}


/**********************************************************************
 *  local_dev_associate
 *
 *  FIXME: temp here as long as authentication app isn't functioning yet.
 *
 **********************************************************************/
void local_dev_associate (struct net_device *dev, char *data)
{
   struct obj_mlme *obj = (struct obj_mlme*) data;

   DEBUG(DBG_MGMT, "local_dev_associate. \n");
   print_mac(obj->address);

   /* FIXME: since the authentication application is not ready yet, succeed always for now!
      FIXME: Remove this code when the authentication app is ready!!! */
   if (obj->state == DOT11_STATE_ASSOCING)
   {
       obj->code  = 0; // = DOT11_SC_SUCCESFUL;
       mgmt_send_mvc_msg(dev, OPSET, DOT11_OID_ASSOCIATE, (long *) obj, sizeof(struct obj_mlme));
   }
}


/* DOT11_OID_ASSOCIATE */
static int local_dev_associate_hndl( unsigned int src_id, unsigned int src_seq,
                                     unsigned int dev_id,   unsigned int dev_seq,
                                     unsigned int op,       unsigned int oid, void *data, unsigned long len)
{
    struct net_device *dev;

    printk("local_dev_associate_hndl \n");

    if ( ! (dev = dev_get_by_index(dev_seq)) )
        return -1;

    local_dev_associate (dev, (char*) data);

    return 0;
}



/* All OIDs */

#define PRISMOP2DEVOP(op) (op == PIMFOR_OP_GET ) ? OPGET : ( (op == PIMFOR_OP_SET ) ? OPSET : -1 )

static int local_send_mvc_msg_hndl( unsigned int dev_type, unsigned int dev_seq,
                                    unsigned int src_id,   unsigned int src_seq,
                                    unsigned int prismop,  unsigned int oid, void *data, unsigned long len)
{
    struct net_device *dev;
    unsigned int devop = PRISMOP2DEVOP(prismop);

    if ( devop < 0 )
        return -1;

    if ( ! (dev = dev_get_by_index(dev_seq)) )
        return -1;

    /* src_id and src_seq sunk */

    if ( local_send_mvc_msg( dev, devop, oid, data, len ) < 0 ) {
        return -1;
    }

    if ( (mgt_response( src_id, src_seq, dev_type, dev_seq, PIMFOR_OP_RESPONSE, oid, data, len)) < 0 ) {
        return -1;
    }

    return 0;
}


/**********************************************************************
 *  local_send_mvc_msg
 *
 *  DESCRIPTION: XXX
 *
 *  PARAMETERS:	 XXX
 *
 *  RETURN:	 XXX
 *
 **********************************************************************/
static int local_send_mvc_msg(struct net_device *dev, int operation, int oid, long* data, unsigned long data_len)
{
    int error = 0;
    struct msg_conf msg;

    DEBUG(DBG_FUNC_CALL | DBG_MGMT, "local_send_mvc_msg \n");

    msg.operation = operation;
    msg.oid       = oid;
    msg.data      = data;
#ifndef TEST_ON_POLDHU
    msg.size      = data_len; /* MVC v2 only */
#endif

#ifdef TEST_ON_POLDHU
    /*
       OIDs for the blob device need some special handling. Don't know if MVC v2 still needs a
       construction like this?
    */

    if ( (oid & 0xFF000000) == 0x80000000) /* This is an OID for the blob device */
    {
        DEBUG(DBG_MGMT, "Msg for the BLOB device...\n");
        /* This is an OID for the BLOB device. We only allow access to the blob device via eth0  */
        if (GET_DEV_TYPE(dev->base_addr) == BIS_DT_802DOT3)
        {
            /* Check if we're manipulating the WLAN LED.
             * If so, we should halt the WLAN device.
             */
            error = BER_NONE;
            switch (oid)
            {
                case BLOB_OID_LEDSET:
                case BLOB_OID_LEDCLEAR:
                if ( (*data & 2) &&
                         (dev_state(1 /* DEVICE_WLAN*/ ) == STRUNNING) )
                {
                    error = dev_halt(1/* DEVICE_WLAN*/);
                }
                break;
                case BLOB_OID_LED_RESET:
                if ( (*data & 2) &&
                         (dev_state(1 /* DEVICE_WLAN*/) == STREADY) )
                {
                    error = dev_run(1 /* DEVICE_WLAN*/);
                }
                return (error < BER_NONE) ? -EINVAL : 0;
                default:
                /* Not one of these */
                break;
            }
            if (error < BER_NONE)
                return -EINVAL;

            if ((error = blob_conf(&msg)) < BER_NONE)
                return -EINVAL;
        }
        else
        {
            return -EOPNOTSUPP;
        }
    }
    else
#endif /* TEST_ON_POLDHU */

    {
printk("Datapointer: %p copied to %p\n", data, msg.data);

        /* Send message to MVC */
        DEBUG(DBG_MGMT, "local_send_mvc_msg: Msg for MVC_LOCAL\n");
        error = dev_conf(GET_DEV_NR(dev->base_addr), &msg);
        printk("local_send_mvc_msg: dev_conf called for oid %x\n", oid);
    }

    /* Copy the message data for retrieval by the calling function. */
    memcpy(data, msg.data, data_len);

    return(error);
}


/****************************************************************
 *								*
 *    		End of management handling functions		*
 *								*
 ***************************************************************/


/****************************************************************
 *								*
 *    	Functions to retrieve the MAC address from the PDA      *
 *      (Production Data Area)                  	        *
 *								*
 ***************************************************************/
#ifdef USE_PDA

#define PDR_END                                 0x0000
#define PDR_MAC_ADDRESS                         0x0101

#define PTR_PDA_AREA_START 0x00000208 /* Contents of address defines the start of the PDA area. */
#ifdef CONFIG_MTD_PHYSMAP_START
static unsigned long *Pda_area_start = (unsigned long*) (CONFIG_MTD_PHYSMAP_START + PTR_PDA_AREA_START);
#else
// configuration not set, assume it is 0.
static unsigned long *Pda_area_start = (unsigned long*) PTR_PDA_AREA_START;
#endif

struct Pdr_Struct {
   unsigned short   PdrLength;
   unsigned short   PdrCode;
   unsigned char    PdrData[2];
};

/**********************************************************************
 *  pda_find_pdr
 *
 *  DESCRIPTION: Finds a Production Data Record in the PDA. (Production Data Area)
 *
 *  RETURN:	 Pointer to found PDR data, or NULL when not found.
 *
 **********************************************************************/
static struct Pdr_Struct *pda_find_pdr (unsigned short pdrCode)
{
   unsigned short   pdaOffset = 0;
   unsigned short    *pdaPtr;    /* Used as index through the PDA pointing to start of records. */
   struct Pdr_Struct *pdrPtr;    /* Used as pointer to the contents of the record. */

   pdaPtr = (unsigned short *) *Pda_area_start;
   pdrPtr = (struct Pdr_Struct *) pdaPtr;

   /* Walk through the cached PDA and check each PDR for it's PDRcode to match the one we are looking for. */
   while (pdrPtr->PdrCode != PDR_END) {
      if (pdrPtr->PdrCode == pdrCode) {
         return pdrPtr;
      } else {
         pdaOffset = pdrPtr->PdrLength + 1;
         /* If we get past 8k or we have an invalid PdrLength, return NULL
            since this indicates an invalid PDA. */
         if ((pdaOffset > 8192) || (pdrPtr->PdrLength == 65535) || (pdrPtr->PdrLength == 0)) {
             return (struct Pdr_Struct *)0;
         }
        /* point to next record */
         pdaPtr = (unsigned short *)pdaPtr + pdaOffset;
         pdrPtr = (struct Pdr_Struct *)pdaPtr;
      }
   };

   if (pdrCode == PDR_END) {
      /* If we were looking for PDR_END, we found it */
       return pdrPtr;
   } else {
      /* We were looking for another PDR and did not find it */
     return (struct Pdr_Struct *)0;
   }
}


/**********************************************************************
 *  pda_get_mac_address
 *
 *  DESCRIPTION: Gets the MAC address from the PDA. (Production Data Area)
 *
 *  RETURN:	 Pointer to MAC address, or NULL when not found.
 *
 **********************************************************************/
static unsigned char * pda_get_mac_address(void)
{
   struct Pdr_Struct *pdrStructPtr;

   pdrStructPtr = pda_find_pdr (PDR_MAC_ADDRESS);
   if (pdrStructPtr) {
       return (unsigned char *)(pdrStructPtr->PdrData);
   } else {
       return (unsigned char *)0;
   }
}
#endif /* USE_PDA */

/****************************************************************
 *								*
 *      	End of PDA MAC address retrieve functions       *
 *								*
 ***************************************************************/


/***************************************************************
*								*
*    		Start of local debugging functions  		*
*								*
***************************************************************/

#if (NET_DEBUG & (DBG_TX | DBG_RX | DBG_ERROR))
/**********************************************************************
 *  dbg_print_frame
 *
 *  DESCRIPTION: Print the fields of prism frame 'framep' and the data.
 *
 *  PARAMETERS:	 framep - pointer to a prism frame
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
static void dbg_print_frame (struct msg_frame *framep)
{
    int offset;
    struct sk_buff *skb;

    skb = framep->handle;
    printk ("framep: size %d, status %d, mpdu %p, skb->data %p, skb %p, skb->len %d \n",
	    framep->size, framep->status, framep->mpdu, skb->data, skb, skb->len);
    printk ("pr_frame: skb->head %p, skb->tail %p, skb->dev->name %s \n",
            skb->head, skb->tail, skb->dev->name);
    for (offset = 0; offset < framep->size; offset++)
	printk ("%02X ", framep->mpdu[offset]);
    printk ("\n");
}
#endif /* (NET_DEBUG & (DBG_TX | DBG_RX | DBG_ERROR)) */


#if (NET_DEBUG & DBG_SKB)
/**********************************************************************
 *  dbg_frame_filter_check
 *
 *  DESCRIPTION: Filter frames not of interest for the Linux IP stack.
 *		 (unicasts not containing our MAC address and Intersil
 *		 debug frames). Perform some checks on the frame and
 *		 print warnings if something suspious is found.
 *
 *  PARAMETERS:	 dev	- device that received the frame
 *		 framep - pointer to the received prism frame
 *
 *  RETURN:	 0 if the frame is absorbed by this function, otherwise 1.
 *
 **********************************************************************/
static int dbg_frame_filter_check (struct net_device *dev, struct msg_frame *framep)
{
    struct sk_buff *skb;
    int offset;
    unsigned char *src;
    int mtu;

    /* filer out unicast frames not directed to our MAC address */
    if ((framep->mpdu[0] & 0x01) == 0)
    {
	for (offset = 0; offset < ETH_ALEN; offset++)
	{
	    if (framep->mpdu[offset] != dev->dev_addr[offset])
	    {
		dev_free_frame (dev, framep);
		return 0;
	    }
	}
    }

    /* filer out Intersil debug frames */
    if ((framep->mpdu[12] == 0x88) && (framep->mpdu[13] == 0x27))
    {
	dev_free_frame (dev, framep);
	return 0;
    }

    skb = framep->handle;

    /* first map pointer back in 0x08000000 area */
    ((unsigned)framep->mpdu) &= ~PRISM_CACHE_MASK;

    /* determine shift in framep->mpdu between add and rx */
    offset = framep->mpdu - skb->data;

    if (offset)
    {
	printk ("frame_to_skb: blob changed mpdu: %d!!!\n", offset);
	dbg_print_frame (framep);
    }
    if (skb_tailroom (skb) < framep->size)
	printk ("frame_to_skb: NOT enough room!! need %d, tailroom %d, len %d\n",
		framep->size, skb_tailroom (skb), skb->len);

    mtu = ((struct net_local *)dev->priv)->dev_desc->mtu;
    offset = dbg_check_area (skb->head, framep->mpdu - skb->head, "header");
    for (src = skb->end-1; (*src == 0) && (src > framep->mpdu+mtu); src--)
	;
    if (src > framep->mpdu+mtu)
    {
	dbg_check_area (framep->mpdu+mtu, (src - framep->mpdu+mtu) + 1, "trailer");
	offset = 1;
    }
    if (offset)
	dbg_print_frame (framep);
    return 1;
}
#endif /* (NET_DEBUG & DBG_SKB) */


/***************************************************************
*								*
*    		End of local debugging functions  		*
*								*
***************************************************************/

