/* $Header$
*  prism_wds.c: Prism Wireless Distribution Set (WDS) driver.
*
*  Partly based on previous prism ethernet driver by Martijn Schipper 2001.
*
*  Extended/rewritten by Bas Muns, Intersil 2002.
*
*  Copyright (C) 2001 Intersil Americas Inc.
*  Copyright (C) 2002 Intersil Americas Inc.
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*/

/*
 * FIXME: add description what WDS is.
 */

//FIXME, check if the header files are still all necessary...

#include <linux/types.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/isil_netlink.h>
#include <linux/netdevice.h>

#include <blob.h>
#include "prism.h"  
#include "prism_wds.h"

/* End of include files */

static struct net_device_stats *wds_net_get_stats(struct net_device *dev);


/****************************************************************
 *								*
 *    			Begin of WDS functions			*
 *								*
 ***************************************************************/


/**********************************************************************
 *  add_wds_link
 *
 *  DESCRIPTION: Add a WDSlink device to the system.
 * 		  On adding a WDSLINK to the MVC the following should be done:
 * 		  - Add a new device to the WDSlink devices list.
 * 		  - Create a device for this WDSlink (dev_alloc wds%d)
 * 		  - Register the just created dummy device (register_netdevice)
 *               - Bring the just created device up and running by calling dev_open.
 * 		  - Somehow the brigde needs to be told to accept the new interface.
 *
 *  PARAMETERS:  - MACAddress, MACAddress to add.
 *               - dev points to device WDSLink need to be added to.
 *
 *
 *  RETURN:	  0 if the device created OK, otherwise -1.
 *
 **********************************************************************/
int add_wds_link(unsigned char *MACAddress, struct net_device *dev)
{
    int err = 0;
    int i=0, j=0;

    struct net_local *lp = (struct net_local *)dev->priv;
    struct net_device *wds_dev;

    DEBUG(DBG_FUNC_CALL, "add_wds_link \n");
    
    /* If the parent device isn't bridged, the WDS will not work */
    if (dev->br_port == NULL)
    {
	DEBUG(DBG_WDS | DBG_ERROR, "add_wds_link: Error parent device (%s) is not bridged, can't add WDS link.\n",dev->name);
        return -1;
    }
    /* First check if MAX_WDS_LINKS is reached */
    if (lp->current_nof_wds_links == MAX_WDS_LINKS)
    {
	/* Max number of links reached */
	DEBUG(DBG_WDS | DBG_ERROR,"add_wds_link: Max number of WDS links (%d) reached!\n", MAX_WDS_LINKS);
	return -1;
    }

    /* If this MAC address is our own mac address */
    if(memcmp(dev->dev_addr, MACAddress, 6) == 0)
    {
        DEBUG(DBG_WDS | DBG_ERROR,"add_wds_link: Cannot add WDS link for own MAC address: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x\n",
                  MACAddress[0], MACAddress[1], MACAddress[2], MACAddress[3], MACAddress[4], MACAddress[5]);
        return -1;
    }
    
    /* Check if this MAC addres is already in the list. */
    for (i=0 ; i< MAX_WDS_LINKS; i++)
    {
	/* If this MAC address is already linked don't add new WDS device */
	if(memcmp(((struct wds_net_local*)(lp->wds_dev[i]->priv))->wds_mac, MACAddress, 6) == 0)
	{
	    DEBUG(DBG_WDS | DBG_ERROR,"add_wds_link: WDS link for MAC: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x already active!\n",
		      MACAddress[0], MACAddress[1], MACAddress[2], MACAddress[3], MACAddress[4], MACAddress[5]);
	    return -1;
	}
	
    }
    /* Find an empty field to add WDS device */
    for (i=0 ; i< MAX_WDS_LINKS; i++)
    {
	if (lp->wds_dev[i] == NULL)
	{
	    /* We have found an empty structure
	     * So we can actualy create a new device for this link
	     */

	    if (!(lp->wds_dev[i] = dev_alloc("wds%d", &err)))
	    {
		DEBUG(DBG_WDS | DBG_ERROR,"add_wds_link: dev_alloc() for WDS failed!\n");
	        return err;
	    }

	    /* Get a pointer to this WDS device */
	    wds_dev = (struct net_device *)lp->wds_dev[i];

	    /* May be overridden by piggyback drivers */
	    ether_setup(wds_dev);

	    /* Initialize the device priv structure. */
	    if (wds_dev->priv == NULL)
	    {
		wds_dev->priv = kmalloc(sizeof(struct wds_net_local), GFP_KERNEL);
		if (wds_dev->priv == NULL)
		    return -ENOMEM;
	    }
	    memset(wds_dev->priv, 0, sizeof(struct wds_net_local));
	    /* Set the parent device for this WDS link */
	    ((struct wds_net_local*)(wds_dev->priv))->parent_dev = dev;
	    DEBUG(DBG_WDS, "add_wds_link: Added parent_dev->name %s\n",
	    	   ((struct wds_net_local*)(wds_dev->priv))->parent_dev->name);
	    /* Set WDS MAC in priv structure */
	    for (j = 0; j < 6; j++)
		printk(" %2.2x", ((struct wds_net_local*)(wds_dev->priv))->wds_mac[j] = MACAddress[j] );
	    printk("\n");

	    printk("add_wds_link: for mac address:");
	    /* Retrieve and print the ethernet address. */
	    for (j = 0; j < 6; j++)
		printk(" %2.2x", wds_dev->dev_addr[j] = ((struct wds_net_local*)(wds_dev->priv))->parent_dev->dev_addr[j] );
	    printk("\n");
	    wds_dev->addr_len = 6;

	    /* Override the network functions we need to use */
	    wds_dev->open		= wds_net_open;
	    wds_dev->stop		= wds_net_close;
	    wds_dev->get_stats		= wds_net_get_stats;
	    wds_dev->hard_start_xmit	= wds_net_send_packet;
	    wds_dev->do_ioctl		= wds_net_ioctl;

	    /* Make ifconfig display some details */
	    wds_dev->base_addr = 0;

	    /* No need to lock cause the net_ioctl is already locked */
	    err = register_netdevice(wds_dev);
	    if (err)
	    {
		DEBUG(DBG_WDS | DBG_ERROR,"add_wds_link: register_netdev() for WDS failed!\n");
		return err;
	    }
	    DEBUG(DBG_WDS, "add_wds_link: registered device name: %s\n", wds_dev->name);

	    /* Bring the device up, when the root device is UP... */
	    if (dev->flags & IFF_UP)
	    {
		err = dev_open(wds_dev);
		if (err)
		{
		    DEBUG(DBG_WDS | DBG_ERROR,"add_wds_link: dev_open() for WDS failed!\n");
		    return err;
		}
		DEBUG(DBG_WDS,"add_wds_link: bringing up device name: %s\n", wds_dev->name);
	    }
	    else
	    {
		DEBUG(DBG_WDS,"add_wds_link: NOT bringing up device name: %s, cause root device isn't up yet...\n", wds_dev->name);
	    }
	    lp->current_nof_wds_links++;
	    return err;
	}
    }
    /* This should not happen*/
    DEBUG(DBG_WDS | DBG_ERROR,"add_wds_link: WDS link administration screwed\n");
    return -1;
}

/**********************************************************************
 *  del_wds_link
 *
 *  DESCRIPTION: Del a WDSlink device to the system.
 * 		On deleting a WDSLINK from the MVC the following should be done:
 *      - Bring the device down by calling dev_close.
 * 		- Unregister the asociated dummy device (unregister_netdevice)
 * 		- Free the allocated space for this device
 * 		- Remove this device from the WDSlink devices list.
 *
 *  PARAMETERS:	 - MACAddress, MACAddress to remove.
 *               - dev points to device WDSLink need to be removed from.
 *
 *
 *  RETURN:	 0 if the device created OK, otherwise 1.
 *
 **********************************************************************/
int del_wds_link(unsigned char *MACAddress, struct net_device *dev)
{
    int err = 0;
    int i=0;

    struct net_local *lp = (struct net_local *)dev->priv;
    struct net_device *wds_dev;

    DEBUG(DBG_FUNC_CALL, "del_wds_link \n");
    
    /* Check if this MAC addres is in the WDS link list. */
    for (i=0 ; i< MAX_WDS_LINKS; i++)
    {
	if(memcmp(((struct wds_net_local*)(lp->wds_dev[i]->priv))->wds_mac, MACAddress, 6) == 0)
	{

	    wds_dev = (struct net_device *)lp->wds_dev[i];
            /* Bring the device down */
	    err = dev_close(wds_dev);
	    if (err)
	    {
	    	DEBUG(DBG_WDS | DBG_ERROR,"del_wds_link: dev_close() for WDS failed!\n");
	    	return err;
	    }
	    /* No need to lock cause the net_ioctl is already locked */
	    err = unregister_netdevice(wds_dev);
	    if (err != 0)
	    {
		DEBUG(DBG_WDS | DBG_ERROR,"del_wds_link: MAC: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x could not be unregistered!\n",
			MACAddress[0], MACAddress[1], MACAddress[2], MACAddress[3], MACAddress[4], MACAddress[5]);
		return err;
	    }
	    kfree(wds_dev->priv);
	    kfree(wds_dev);
	    lp->current_nof_wds_links--;
	    lp->wds_dev[i] = NULL;

	    DEBUG(DBG_WDS, "del_wds_link: WDSLink removed succesfully\n");
	    return err;
	}
    }
    DEBUG(DBG_WDS | DBG_ERROR,"del_wds_link: MAC: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x is not a WDS link\n",
	      MACAddress[0], MACAddress[1], MACAddress[2], MACAddress[3], MACAddress[4], MACAddress[5]);
    return -1;
}

/**********************************************************************
 *  open_wds_links
 *
 *  DESCRIPTION: Brings all associated WDS devices up and running.
 *
 *  PARAMETERS:	 dev points to device WDS links are put under.
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
void open_wds_links(struct net_device *dev)
{
    struct net_local *lp = (struct net_local *)dev->priv;
    int i;
    
    DEBUG(DBG_FUNC_CALL, "open_wds_links \n");
    
    /* Bring the associated WDS devices up and RUNNING. */
    DEBUG(DBG_WDS, "Open all associated WDS links: ");
    
    for (i = 0; i < lp->current_nof_wds_links; i++) 
    {
        DEBUG(DBG_WDS, "WDS%d...",i);
        dev_open(lp->wds_dev[i]);
    }
    DEBUG(DBG_WDS, "\n");              
}

/**********************************************************************
 *  close_wds_links
 *
 *  DESCRIPTION: Closes all associated WDS links for the device.
 *
 *  PARAMETERS:	 dev points to device WDS links are put under.
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
void close_wds_links(struct net_device *dev)
{   
    struct net_local *lp = (struct net_local *)dev->priv;
    int i;

    DEBUG(DBG_FUNC_CALL, "close_wds_links \n");
    
    DEBUG(DBG_WDS, "Close associated WDS links: ");
    for (i = 0; i < lp->current_nof_wds_links; i++) 
    {
        DEBUG(DBG_WDS, "WDS%d...",i);
        dev_close(lp->wds_dev[i]);
    }
    DEBUG(DBG_WDS, "\n");
}


/**********************************************************************
 *  check_skb_for_wds
 *
 *  DESCRIPTION: Checks the skb->cb field for the WDS Link magic.
 *               It copies the MAC address into WDS_MAC.
 *
 *  PARAMETERS:	 pointer to sk_buff skb.
 *               WDS_Mac, is updated with the found WDS Link MAC address. 
 *
 *  RETURN:	 Return 1 if a WDS link is found, 0 otherwise
 *
 *  NOTE:        inline fucntion.
 *
 **********************************************************************/
inline int check_skb_for_wds(struct sk_buff *skb, unsigned char *WDS_Mac)
{
    /* If this skb came from a WDS device, the
     * skb->cb field will be filled with a WDSLink MACaddress.
     * In this case the MAC address must be filled in the msg_frame.
     * Due to the reason the cb field is free to use, the MAC address
     * will be packed in the following keys "WDSL"<MAC ADDRESS>".
     */

    if(*(long*)&skb->cb[0] == 0x4c534457) /* WDSL */
    {
	memcpy(WDS_Mac, skb->cb+4, 6);       
#if (NET_DEBUG & DBG_WDS)
	printk("net_send_packet: found wdsmac in skb: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x \n",
		WDS_Mac[0],WDS_Mac[1],WDS_Mac[2],WDS_Mac[3],WDS_Mac[4],WDS_Mac[5]);
#endif
 	/* Make the WDS data invalid */
	*(long*)&skb->cb[0] = 0;

        return(1); //WDS Link is available */
    }                                        
    return(0); /* WDS Link is unavailable */ 
}


/**********************************************************************
 *  copy_wds_to_frame
 *
 *  DESCRIPTION: This copies the WDS_Mac address into the msg_frame structure.
 *
 *  PARAMETERS:	 frame pointer to msg_frame structure
 *               WDS MAC address to put into the frame.        
 *
 *  RETURN:	 Nothing
 *
 *  NOTE:        inline function.
 *
 **********************************************************************/
#ifdef BLOB_V2
inline void copy_wds_to_frame(msg_frame *framep, unsigned char *WDS_Mac)
{
    /* Copy the found MAC addres in the msg_frame. */
    memcpy(framep->wds, WDS_Mac, 6); /* MVC v1 */
}
#else
inline void copy_wds_to_frame(msg_frame *framep, unsigned char *WDS_Mac)
{
    memcpy(framep->wds_link, WDS_Mac, 6); /* MVC v2 */
}
#endif

/**********************************************************************
 *  wds_find_device
 *
 *  DESCRIPTION: Find the WDS device to use for receiving this frame
 *
 *  PARAMETERS:	 struct msg_frame *framep - The frame received
 *  		 struct net_device    *dev    - The device the frame was
 *                                              received on.
 *
 *  RETURN:	 *net_dev of the device found, otherwise NULL.
 *
 **********************************************************************/
#ifdef BLOB_V2

struct net_device *wds_find_device( msg_frame *framep, struct net_device *dev )
{
    int i=0;
    struct net_local *lp;

    DEBUG(DBG_FUNC_CALL | DBG_WDS, "wds_find_device \n");
    
    if (framep->wds[0] == 0 && framep->wds[1] == 0 &&           /* MVC v2 */
    	framep->wds[2] == 0 && framep->wds[3] == 0 &&
    	framep->wds[4] == 0 && framep->wds[5] == 0)
    {
	DEBUG(DBG_WDS, "wds_find_device: framep->wds no MAC found, returning NULL\n");
	return NULL;
    }
    lp = (struct net_local *)dev->priv;
    for (i=0; i<MAX_WDS_LINKS; i++)
    {
	if (lp->wds_dev[i] == NULL)
	    continue;
        if(memcmp(((struct wds_net_local*)(lp->wds_dev[i]->priv))->wds_mac, framep->wds, 6) == 0)
	{
	    DEBUG(DBG_WDS, "wds_find_device: framep->wds MAC found, returning dev->name: %s\n",
	    	   lp->wds_dev[i]->name);
	    return ((struct net_device*)lp->wds_dev[i]);
	}
    }
    DEBUG(DBG_WDS, "wds_find_device: WDS from mac %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x not in current config, returning NULL\n",
    	   framep->wds[0],framep->wds[1],framep->wds[2],framep->wds[3],framep->wds[4],framep->wds[5]);
    return NULL;
}

#else

struct net_device *wds_find_device( msg_frame *framep, struct net_device *dev )
{
    int i=0;
    struct net_local *lp;

    DEBUG(DBG_FUNC_CALL | DBG_WDS, "wds_find_device \n");
    
    /* If no wds_link MAC address supplied */
    if  (framep->wds_link[0] == 0 && framep->wds_link[1] == 0 && /* MVC v1 */
    	 framep->wds_link[2] == 0 && framep->wds_link[3] == 0 &&
    	 framep->wds_link[4] == 0 && framep->wds_link[5] == 0)
    {
	DEBUG(DBG_WDS, "wds_find_device: framep->wds no MAC found, returning NULL\n");
	return NULL;
    }
    lp = (struct net_local *)dev->priv;
    for (i=0; i<MAX_WDS_LINKS; i++)
    {
	if (lp->wds_dev[i] == NULL)
	    continue;
        if(memcmp(((struct wds_net_local*)(lp->wds_dev[i]->priv))->wds_mac, framep->wds_link, 6) == 0)
	{
	    DEBUG(DBG_WDS, "wds_find_device: framep->wds MAC found, returning dev->name: %s\n",
	    	   lp->wds_dev[i]->name);
	    return ((struct net_device*)lp->wds_dev[i]);
	}
    }
    DEBUG(DBG_WDS, "wds_find_device: WDS_link from mac %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x not in current config, returning NULL\n",
    	   framep->wds_link[0],framep->wds_link[1],framep->wds_link[2],framep->wds_link[3],framep->wds_link[4],framep->wds_link[5]);
    return NULL;
}

#endif

/**********************************************************************
 *  wds_net_open
 *
 *  DESCRIPTION: Bring the WDS dev up and running.
 * 		 The WDS device needs to be added to the bridge,
 * 		 where the parent dev is also bridged.
 *
 *  PARAMETERS:	 struct net_device    *dev    - The WDS device
 *
 *  RETURN:	 zero on succes, otherwise error (<0).
 *
 **********************************************************************/
int wds_net_open(struct net_device *dev)
{
    int err = 0;
    struct wds_net_local   *lp = (struct wds_net_local *)dev->priv;
    struct ifreq	    ifr;
    
    DEBUG(DBG_FUNC_CALL | DBG_WDS, "wds_net_open \n");
    
    netif_start_queue(dev);

    /* Check if the parent device is bridged */
    if (lp->parent_dev->br_port == NULL)
    {
	DEBUG(DBG_WDS | DBG_ERROR, "wds_net_open: Error parent device (%s) is not bridged, can't add WDS link.\n",lp->parent_dev->name);
        return -1;
    }

    /* Put the PROMISC flag on */
    ifr.ifr_flags = dev->flags | IFF_PROMISC;
    dev_ioctl(SIOCSIFFLAGS, &ifr);

    return err;
}

/**********************************************************************
 *  wds_net_send_packet
 *
 *  DESCRIPTION: Sends a WDS packet via the WDS' parent device.
 *               This function puts the WDS MACaddress in the skb->cb
 *               private field, alters the device to send from into the
 *               parent device for this WDS link and puts the skb into
 *               the transmit device queue.
 *
 *  PARAMETERS:	 struct sk_buff    *skb -  skbuff to send.
 *               struct net_device *dev -  wds device. 
 *
 *  RETURN:	 zero.
 *
 **********************************************************************/
int wds_net_send_packet(struct sk_buff *skb, struct net_device *dev)
{
    struct wds_net_local *lp = (struct wds_net_local *)dev->priv;

    DEBUG(DBG_FUNC_CALL | DBG_WDS, "wds_net_send_packet \n");
    
    if (lp == NULL)
    {
	printk("wds_net_send_packet: NULL pointer detected\n");
	return 0;
    }
    if ((lp->parent_dev->flags & IFF_UP) && (dev->flags & IFF_UP))
    {
        /* OK lets copy the WDSlink MACAddres in the skb->cb fields */
        memcpy(skb->cb, "WDSL", 4);
        memcpy(skb->cb+4, ((struct wds_net_local*)(dev->priv))->wds_mac, dev->addr_len);

        /* Set the "transmit" dev in this skb */
        skb->dev = lp->parent_dev;

        /* Move this skb to the "transmit" devices queue */
        if(dev_queue_xmit(skb) < 0)
        {
            printk("wds_net_send_packet: dev_queue_xmit failed\n");
        }
        /* Update stats for this device */
        lp->stats.tx_packets++;
        lp->stats.tx_bytes += skb->len;
    }
    return 0;
}

/**********************************************************************
 *  wds_net_close
 *
 *  DESCRIPTION: Deletes the WDS device from the bridge.
 *
 *  PARAMETERS:	 struct net_device *dev - WDS device to remove.
 *
 *  RETURN:	 0 on success and -1 when parent device isn't bridged.
 *
 **********************************************************************/
int wds_net_close(struct net_device *dev)
{
    int err = 0;
    struct wds_net_local   *lp = (struct wds_net_local *)dev->priv;

    DEBUG(DBG_FUNC_CALL | DBG_WDS, "wds_net_close \n");
    
    netif_stop_queue(dev);

    /* Check if the parent device is bridged */
    if (lp->parent_dev->br_port == NULL)
    {
	DEBUG(DBG_WDS | DBG_ERROR, "wds_net_close: Error parent device (%s) is not bridged, can't close WDS link.\n",lp->parent_dev->name);
        return -1;
    }

    return err;
}


/**********************************************************************
 *  wds_net_get_stats
 *
 *  DESCRIPTION: Gets the statistics for the WDS device. 
 *
 *  PARAMETERS:	 struct net_device *dev - WDS device to get statistics from.
 *
 *  RETURN:	 pointer to net_device_stats struct.
 *
 **********************************************************************/
static struct net_device_stats *
wds_net_get_stats(struct net_device *dev)
{
    struct wds_net_local *lp = (struct wds_net_local *)dev->priv;
    struct net_device_stats *stats = &lp->stats;

    DEBUG(DBG_FUNC_CALL | DBG_WDS, "wds_net_get_stats \n");

    return stats;
}

/**********************************************************************
 *  wds_net_ioctl
 *
 *  DESCRIPTION: Ioctl for WDS devices are NOT supported. Tell them.
 *
 *  PARAMETERS:	 struct net_device *dev -
 *               struct ifreq      *ifr -
 *               int    cmd             -
 *
 *  RETURN:	 -EINVAL. 
 *
 **********************************************************************/
int wds_net_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    printk("wds_net_ioctl: Ioctls for WDS devices are NOT supported in this driver.\n");
    return -EINVAL;
}


/***************************************************************
*								*
*    			End of WDS functions			*
*								*
***************************************************************/

