/* $Header$
 * prismeth.c: Prism Ethernet driver.
 *
 * Based on skeleton ethernet driver by Donald Becker.
 *
 * Copyright 1993 United States Government as represented by the
 * Director, National Security Agency.
 * Copyright (C) 2001 Intersil Americas Inc.
 *
 * Modified by Martijn Schipper 2001.
 *
 */
#define WDSLINK_SUPPORT

static const char *version = "PRISM packet driver: version 0.2.0\n";

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/capability.h>
#include <linux/netdevice.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/serial.h>
#include <asm/arch/gpio.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <asm/arch/blobcalls.h>

#ifdef WDSLINK_SUPPORT
#include "../../net/bridge/br_private.h"
#endif

#include "../../../lib/Inl/include/hardconfig.h"

#include "prismeth.h"
#include "../../../include/linux_mvc_oids.h"

#ifdef WDSLINK_SUPPORT

#define MAX_WDS_LINKS 8
struct wds_net_local
{
    struct net_device_stats	stats;		/* tx/rx statistics */
    struct net_device 	       *parent_dev;
    unsigned char 		wds_mac[6];
};

/* Function prototypes. */
static int AddWDSLink(struct prism_ioctl req, struct net_device *dev);
static int DelWDSLink(struct prism_ioctl req, struct net_device *dev);
struct net_device *wds_find_device(struct msg_frame_wds *framep, struct net_device *dev);

static int wds_net_open(struct net_device *dev);
static int wds_net_send_packet(struct sk_buff *skb, struct net_device *dev);
static int wds_net_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);
static int wds_net_close(struct net_device *dev);
static struct net_device_stats *wds_net_get_stats(struct net_device *dev);
extern int br_add_if(struct net_bridge *br, struct net_device *dev);

#endif //WDSLINK_SUPPORT

/* Maximum number of clients that can authenticate and associate */
/* FIXME: these should be configuration items */
#define MAX_NR_CLIENTS_AUTH	2048
#define MAX_NR_CLIENTS_ASSOC	2007

/*
 * The name of the card. Is used for messages and in the requests for irqs
 */
static const char* cardname = "Prism packet IF";

/* Debug flag to get debug information for specific sections */
#define DBG_TX	    0x00000001
#define DBG_RX	    0x00000010
#define DBG_IRQ	    0x00000100
#define DBG_TRAP    0x00000200
#define DBG_IOCTL   0x00001000
#define DBG_OPEN    0x00010000
#define DBG_CLOSE   0x00100000
#define DBG_ERROR   0x01000000
#define DBG_SKB	    0x10000000
#define DBG_WDS	    0x00000010
#define DBG_ALL	    0xFFFFFFFF

#ifndef NET_DEBUG
#define NET_DEBUG   (DBG_OPEN | DBG_ERROR)
#endif
static unsigned int net_debug = NET_DEBUG;

/* macros to simplify debug checking */
#define BUGLVL(x) if (net_debug & (x))
#ifdef GCC_3X
#define BUGMSG(x,args,...) \
	do{BUGLVL(x) printk((args), ## __VA_ARGS__);}while(0)
#else
#define BUGMSG(x,args...) do{BUGLVL(x) printk(## args);}while(0)
#endif
#define _BUGMSG(x,args,...)

/* A MAC address of ff-ff-ff-ff-ff-ff in the ACL means everybode has access */
static const char cAclFullAccess[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

/* Information to convert a trap to a signal for a process */
struct trap_info
{
    struct trap_info *next;
    int oid;
    int pid;
    int sig;
};


/* Information that need to be kept for each device. */
struct net_local
{
    struct net_device_stats	    stats;	/* tx/rx statistics */
    struct bis_dev_descr_element    *dev_desc;	/* MVC device desriptor */
    int				    maxhead;	/* maximum head, tail and mtu */
    int				    maxtail;	/* sizes for all MVC devices */
    int				    maxmtu;	/* (so we can bridge w/h copy)*/
    struct trap_info		    *traps;	/* registrated traps */
    unsigned char		    *acl;
    int				    acl_nr_items;
    unsigned int		    tx_misaligned;
    unsigned int		    rx_misaligned;
    int				    maxnrframes;
    struct msg_frame		    *frame_pool;
    struct msg_frame		    *frame_freelist;
    unsigned int		    watchdog_timeout;  /* 0 = Off */
    unsigned int		    watchdog_event;
#ifdef WDSLINK_SUPPORT
    struct net_device		    *wds_dev[MAX_WDS_LINKS];
    unsigned int 		    current_nof_links;
#endif
};

/* Information about the WDS bridge devices available */



/* Function prototypes. */
extern const unsigned char *blobtool_find_element(int id, int skip);
extern int prism_ether_probe(struct net_device *dev);

static int prism_probe1(struct net_device *dev, int ioaddr);
static int net_open(struct net_device *dev);
static void net_tx_timeout(struct net_device *dev);
static int net_send_packet(struct sk_buff *skb, struct net_device *dev);
static void net_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static int net_close(struct net_device *dev);
static int net_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);
static struct net_device_stats *net_get_stats(struct net_device *dev);
static int prism_eth_init(struct net_device *dev, int startp);
static int prism_eth_start ( struct net_device *dev );
static int prism_eth_run ( struct net_device *dev );

#if (NET_DEBUG & (DBG_TX | DBG_RX | DBG_ERROR))
static void print_frame (struct msg_frame *framep);
#endif
#if (NET_DEBUG & DBG_SKB)
static void dump_hex_assci (unsigned char *buf, int size);
#endif
#if (NET_DEBUG & DBG_SKB)
static int check_area (unsigned char *p, int size, char *area_name);
#endif
#if (NET_DEBUG & DBG_SKB)
static int frame_filter_check (struct net_device *dev, struct msg_frame *framep);
#endif
static struct msg_frame *skb_to_frame (const struct sk_buff *skb);
static struct sk_buff *frame_to_skb (struct msg_frame *framep);
static struct msg_frame *dev_alloc_frame (struct net_device *dev, int size);
static void dev_free_frame (struct net_device *dev, struct msg_frame *frame);
static struct sk_buff *dev_copy_expand_skb (struct net_device *dev,
				const struct sk_buff *skb);
static void dev_handle_reqs (struct net_device *dev, int requests);
static void handle_blob_reqs(struct net_device *dev);
static struct trap_info *trap_prev_find (struct net_device *dev,
					struct trap_info *trap);
static struct trap_info *trap_oid_find (struct trap_info *root, int oid);
static struct trap_info *trap_find (struct net_device *dev, int oid, int pid,
				    int sig);
static int trap_add (struct net_device *dev, int oid, int pid, int sig);
static int trap_del (struct net_device *dev, int oid, int pid, int sig);
static int convert_mvc_trap_to_oid(int oid, int data);
static void dev_authenticate (struct net_device *dev, struct obj_mlme *obj);
static void dev_deauthenticate (struct net_device *dev, struct obj_mlme *obj);
static void dev_associate (struct net_device *dev, struct obj_mlme *obj);
static void dev_reassociate (struct net_device *dev, struct obj_mlme *obj);
static void dev_disassociate (struct net_device *dev, struct obj_mlme *obj);
static long dev_getval (struct net_device *dev, int oid);

static struct obj_mlme mlme;

/*
 * Check for a network adaptor of this type, and return '0' iff one exists.
 * If dev->base_addr == 0, probe all likely locations.
 * If dev->base_addr == 1, always return failure.
 * If dev->base_addr == 2, allocate space for the device and return success
 * (detachable devices only).
 */
int
/* prism_ether_probe */
poldhu_ether_probe(struct net_device *dev)
{
    return prism_probe1( dev, DEVICE_HOSTDS );
}


/*
 * This is the real probe routine.
 */
static int prism_probe1(struct net_device *dev, int ioaddr)
{
    static unsigned version_printed = 0;
    int i;
    struct net_local *lp;
    struct bis_dev_descr_element *dev_desc;

    /* Allocate a new 'dev' if needed. */
    if (dev == NULL)
    {
	/*
	 * Don't allocate the private data here, it is done later
	 * This makes it easier to free the memory when this driver
	 * is used as a module.
	 */
	dev = init_etherdev(0, 0);
	if (dev == NULL)
	    return -ENOMEM;
    }

    if (net_debug  &&  version_printed++ == 0)
    {
	printk(KERN_DEBUG "%s", version);

#ifdef CONFIG_POLDEBUG_ENABLE

	/* start the MVC debug interface */
	dev_start (DEVICE_DEBUG, 0);
	dev_run (DEVICE_DEBUG);

#endif

    }

    printk(KERN_INFO "%s: %s found,", dev->name, cardname );

    /* Fill in the 'dev' fields. */
    if (dev->base_addr == 0)
	dev->base_addr = ioaddr;

    /* Retrieve and print the ethernet address. */
    for (i = 0; i < 6; i++)
	printk(" %2.2x", dev->dev_addr[i] = HardConfig->MACAddress[i] );

    dev->irq = IRQ_SOFINT;

    /* Initialize the device structure. */
    if (dev->priv == NULL)
    {
	dev->priv = kmalloc(sizeof(struct net_local), GFP_KERNEL);
	if (dev->priv == NULL)
	    return -ENOMEM;
    }

    memset(dev->priv, 0, sizeof(struct net_local));
    lp = (struct net_local *)dev->priv;

    dev->open		    = net_open;
    dev->stop		    = net_close;
    dev->get_stats	    = net_get_stats;
    dev->hard_start_xmit    = net_send_packet;
    dev->do_ioctl	    = net_ioctl;
#ifdef HAVE_TX_TIMEOUT
    dev->tx_timeout	    = net_tx_timeout;
    dev->watchdog_timeo	    = HZ;
#endif
    /* determine max mtu, head and tail sizes for all MVC packet devices */
    for (i = DEVICE_WLAN; i <= DEVICE_HOSTDS; i++)
    {
	dev_desc = (struct bis_dev_descr_element *)
				blobtool_find_element (BIS_DEVDESCR, i);
	if (dev_desc)
	{
	    if (dev_desc->header > lp->maxhead)
		lp->maxhead = (dev_desc->header + 3) & ~3;
	    if (dev_desc->trailer > lp->maxtail)
		lp->maxtail = dev_desc->trailer;
	    if (dev_desc->mtu > lp->maxmtu)
		lp->maxmtu = dev_desc->mtu;
	}
	if (i == dev->base_addr)
	    lp->dev_desc = dev_desc;
    }
    if (lp->dev_desc)
    {
	printk ("\ndev_descr: nr: %d, type: %d, mtu: %d, head: %d, tail: %d\n",
		lp->dev_desc->nr, lp->dev_desc->type, lp->dev_desc->mtu,
		lp->dev_desc->header, lp->dev_desc->trailer);
    }
    else
	printk (KERN_ALERT "NO dev_desc found!!!!\n");

    lp->maxnrframes = 100; /* this should be part of MVC device descr */

    /* allocate frame pool */
    lp->frame_pool = kmalloc(sizeof(struct msg_frame) * lp->maxnrframes,
			    GFP_KERNEL);
    if (lp->frame_pool == NULL)
	return -ENOMEM;

    /* allocate all frames to the free frame list */
    lp->frame_freelist = NULL;
    for (i = 0; i < lp->maxnrframes; i++)
    {
	lp->frame_pool[i].handle = lp->frame_freelist;
	lp->frame_freelist = &lp->frame_pool[i];
    }

    /* Fill the watchdog settings */
    lp->watchdog_timeout = DEFAULT_WATCHDOG_TIME;
    lp->watchdog_event = WATCHDOG_NOT_TRIGGERED;

    /* Fill in the fields of the device structure with ethernet values. */
    ether_setup(dev);

    SET_MODULE_OWNER(dev);

    /* Let's bring the device in the READY state */
    prism_eth_init( dev, PRISM_ETH_START );

    return 0;

}


struct msg_frame *frame_alloc (struct net_device *dev)
{
    struct msg_frame *framep;
    unsigned long flags;
    struct msg_frame_wds *wdsframep;

    save_flags_cli (flags);

    framep = ((struct net_local *)dev->priv)->frame_freelist;
    if (framep)
	((struct net_local *)dev->priv)->frame_freelist = framep->handle;

    restore_flags(flags);
    /* Make sure the wds info in the frame is zeroed */
    wdsframep = (struct msg_frame_wds*)framep;
    memset(wdsframep->wds_link, 0, 6);
    return framep;
}


void frame_free (struct net_device *dev, struct msg_frame *framep)
{
    unsigned long flags;
    save_flags_cli (flags);

    framep->handle = ((struct net_local *)dev->priv)->frame_freelist;
    ((struct net_local *)dev->priv)->frame_freelist = framep;

    restore_flags(flags);
}


/**********************************************************************
 *  skb_to_frame
 *
 *  DESCRIPTION: Attach a prism frame structure to linux sk_buff 'skb'
 *		 and initialize all fields of the frame structure with
 *		 the current information of 'skb'.
 *
 *  PARAMETERS:	 skb - pointer to linux packet buffer.
 *
 *  RETURN:	 Pointer to the attached frame, or NULL if there wasn't
 *		 enough memory.
 *
 **********************************************************************/
static struct msg_frame *skb_to_frame (const struct sk_buff *skb)
{
    struct msg_frame *framep;

    /* first malloc memory for msg_frame */
    framep = frame_alloc (skb->dev);
    if (framep != NULL)
    {
	framep->handle	 = (void *) skb;
	framep->mpdu	 = (unsigned char *)
				(((unsigned)skb->data) | 0x04000000);
	framep->size	 = skb->len;
	framep->priority = -1;
    }

#if (NET_DEBUG & DBG_SKB)
    memset (skb->head, 0, skb_headroom(skb));
    memset (skb->tail, 0, skb_tailroom(skb));
#endif
    return framep;
}


/**********************************************************************
 *  frame_to_skb
 *
 *  DESCRIPTION: Update the fields of the linux sk_buff assigned to
 *		 frame 'framep' just received from the MVC.
 *		 The frame is freed.
 *
 *  PARAMETERS:	 framep - pointer to frame received from MVC.
 *
 *  RETURN:	 Pointer to the updated sk_buff that was assigned to the
 *		 frame.
 *
 **********************************************************************/
static struct sk_buff *frame_to_skb (struct msg_frame *framep)
{
    struct sk_buff *skb;
    int offset;
    unsigned char *src;

    skb = framep->handle;

    /* first map pointer back in 0x08000000 area */
    ((unsigned)framep->mpdu) &= ~0x04000000;

    /* determine shift in framep->mpdu between add and rx */
    offset = framep->mpdu - skb->data;

    skb_reserve (skb, offset);
    skb_put (skb, framep->size);

    /* determine offset between IP header and a word boundary */
    offset = (int) (skb->data + 6 + 6 + 2) & 0x03;

    /* if there is any, shift the data */
    if (offset)
    {
	struct net_local *lp = (struct net_local *)skb->dev->priv;
	lp->rx_misaligned++;

	if (skb_headroom (skb) < offset)
	{
	    if (skb_tailroom (skb) < (4 - offset))
		printk(KERN_NOTICE "Unaligned skb\n");
	    else
		offset = 4 - offset;
	}
	else
	    offset = -offset;
#if (NET_DEBUG & DBG_RX)
	printk ("frame_to_skb: need to shift data %d\n", offset);
#endif
	src = skb->data;
	skb_reserve (skb, offset);
	memmove (skb->data, src, skb->len);
    }
    frame_free (skb->dev, framep);

    return skb;
}


/**********************************************************************
 *  dev_alloc_frame
 *
 *  DESCRIPTION: Allocate a frame and a sk_buff with MTU size 'size' for
 *		 device 'dev'. Set all fields of the frame and the
 *		 sk_buff. If size equals 0 then the maximum MTU size is
 *		 used. The head and the tail room for the frame (and
 *		 sk_buff) is equal to the maximum head and maximum tail
 *		 room for all devices.
 *
 *  PARAMETERS:	 dev  - device to allocate a frame for
 *		 size - the MTU size for the frame.
 *
 *  RETURN:	 Pointer to the allocated frame or NULL on error.
 *
 **********************************************************************/
static struct msg_frame *dev_alloc_frame (struct net_device *dev, int size)
{
    struct sk_buff *skb;
    int bufsize;
    struct net_local *lp;

    /* first allocate a sk_buff with enough head and tail room */
    lp = (struct net_local *)dev->priv;
    bufsize = lp->maxhead + 2 +
		    (size ? size : lp->maxmtu) + lp->maxtail;
    if ((skb = dev_alloc_skb (bufsize)) == NULL)
	return NULL;

    /* fill in some sk_buff fields */
    skb->dev = dev;
    skb_reserve (skb, lp->maxhead + 2); /* word align IP header */

    return skb_to_frame (skb);
}


/**********************************************************************
 *  dev_free_frame
 *
 *  DESCRIPTION: Free MVC frame 'frame' and the attached sk_buff.
 *
 *  PARAMETERS:	 frame - address of frame to free
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
static void dev_free_frame (struct net_device *dev, struct msg_frame *frame)
{
    if (frame->handle)
	dev_kfree_skb_any (frame->handle);
    frame_free (dev, frame);
}


/**********************************************************************
 *  dev_copy_expand_skb
 *
 *  DESCRIPTION: Allocate a new sk_buff and copy the data of sk_buff
 *		 'skb'. The new sk_buff will have the required head and
 *		 tail room for device 'dev'.
 *
 *  PARAMETERS:	 dev - device to allocate a frame for
 *		 skb - the sk_buff to copy the data from
 *
 *  RETURN:	 Pointer to the new allocated sk_buff or NULL on error.
 *
 **********************************************************************/
static struct sk_buff *dev_copy_expand_skb (struct net_device *dev,
				const struct sk_buff *skb)
{
    int bufsize;
    struct sk_buff *newskb;
    struct net_local *lp = (struct net_local *)dev->priv;

    bufsize = skb->len + lp->dev_desc->header + lp->dev_desc->trailer;
    if ((newskb = dev_alloc_skb (bufsize)) == NULL)
	return NULL;

    /* reserve head room */
    skb_reserve (newskb, lp->dev_desc->header);

    /* set the tail pointer and length */
    skb_put(newskb, skb->len);

    /* copy the data */
    memcpy(newskb->data, skb->data, skb->len);

    newskb->dev = dev;
    return newskb;
}


/**********************************************************************
 *  dev_handle_reqs
 *
 *  DESCRIPTION: Handle requests 'requests' for device 'dev'.
 *
 *  PARAMETERS:	 dev	  - device to handle requests for
 *		 requests - requests to handle
 *
 *  RETURN:	 Nothing.
 *
 *  NOTE:	 Also called from the interrupt service routine.
 *
 **********************************************************************/
static void dev_handle_reqs (struct net_device *dev, int requests)
{
    struct net_local		*lp;
    struct sk_buff		*skb;
    struct msg_conf		*trapmsg;
    struct trap_info		*trap;
    struct trap_info		*root;
    struct msg_frame		*framep;
#ifdef WDSLINK_SUPPORT
    struct net_device   	*wds_dev;
    struct wds_net_local	*wds_lp;
#endif //WDSLINK_SUPPORT
    lp = (struct net_local *) dev->priv;

    _BUGMSG(DBG_IRQ, "dev_handle_reqs: dev %ld", dev->base_addr);

    if (requests & 0xFFFFEFE0)
	printk (" called with unknown requests: %d!!\n", requests);

    if (requests & REQFRAMERX)
    {
	if ((framep = dev_frame_rx (dev->base_addr)) != NULL)
	{
#ifdef WDSLINK_SUPPORT
	    if (dev->base_addr == DEVICE_WLAN)
	    {
		/* Check if this frame contains one of the WDS addresses */
		wds_dev = wds_find_device((struct msg_frame_wds*)framep, dev);
	    }
	    else
	    	wds_dev = NULL;
#endif //WDSLINK_SUPPORT
	    dev->last_rx = jiffies;
	    if (framep->size < 1569 && framep->status < 10)
	    {

		_BUGMSG(DBG_IRQ, "..RX");
#if (NET_DEBUG & DBG_RX)
		print_frame (framep);
#endif
#if (NET_DEBUG & DBG_SKB)
		if (frame_filter_check (dev, framep))
#endif
		{
		    skb = frame_to_skb (framep);
		    skb->protocol = eth_type_trans( skb, dev );
		    lp->stats.rx_packets++;
		    lp->stats.rx_bytes += skb->len;
#ifdef WDSLINK_SUPPORT
		    if (wds_dev != NULL)
		    {
			wds_lp = (struct wds_net_local *) wds_dev->priv;
			wds_lp->stats.rx_packets++;
			wds_lp->stats.rx_bytes += skb->len;
			skb->dev = wds_dev;
		    }
#endif //WDSLINK_SUPPORT
		    netif_rx (skb);
		}
	    }
	    else
		printk (KERN_ERR " RX, size %d, status %d!!\n",
			framep->size, framep->status);
	}
    }
    if (requests & REQFRAMERETURN)
    {
	if ((framep = dev_frame_return (dev->base_addr)) != NULL)
	{
	    _BUGMSG(DBG_IRQ, "..RET");
	    if (framep->status < 10)
	    {
		if (framep->status <= FSTXSUCCESSFUL)
		{
		    if (framep->status == FSTXSUCCESSFUL)
		    {
			lp->stats.tx_packets++;
			lp->stats.tx_bytes += framep->size;
		    }
		    else
		    {
			printk (KERN_ERR " TX error!!. status %d\n",
				framep->status);
			lp->stats.tx_errors++;
		    }
		    netif_wake_queue (dev);
		}
		dev_free_frame (dev, framep);
	    }
	    else
		printk (KERN_ERR " RET, size %d, status %d!!\n",
			framep->size, framep->status);
	}
    }

    if (requests & REQTRAP)
    {
	_BUGMSG(DBG_IRQ, "..TRAP");
	if ((trapmsg = dev_trap (dev->base_addr)) != NULL)
	{
	    switch (trapmsg->oid)
	    {
		case DOT11_OID_AUTHENTICATE:
		    dev_authenticate (dev, (struct obj_mlme *) trapmsg->data);
		    break;
		case DOT11_OID_DEAUTHENTICATE:
		    dev_deauthenticate (dev, (struct obj_mlme *) trapmsg->data);
		    break;
	        case DOT11_OID_REASSOCIATE:
	            dev_reassociate(dev, (struct obj_mlme *)trapmsg->data);
	            break;
		case DOT11_OID_ASSOCIATE:
		    dev_associate (dev, (struct obj_mlme *) trapmsg->data);
		    break;
		case DOT11_OID_DISASSOCIATE:
		    dev_disassociate (dev, (struct obj_mlme *) trapmsg->data);
		    break;
		case GEN_OID_LINKSTATE:
		    if (dev->base_addr == DEVICE_HOSTDS)
		    {
			/* Set/Clear the Ethernet LED */
			long mask = 1;
			struct msg_conf config;

			config.operation = OPSET;
			config.data = &mask;

			if( *trapmsg->data )
			{
			    config.oid = BLOB_OID_LEDSET;
			    printk ("%s: link up\n", dev->name);
			    netif_carrier_on (dev);
			}
			else
			{
			    config.oid = BLOB_OID_LEDCLEAR;
			    printk ("%s: link down\n", dev->name);
			    netif_carrier_off (dev);
			}

			blob_conf( &config );
		    }
	    }
	    root = lp->traps;

	    while ((trap = trap_oid_find (root, trapmsg->oid)) != NULL)
	    {
		_BUGMSG(DBG_TRAP, " sending signal %d to pid: %d\n", trap->sig,
			    trap->pid);
		root = trap->next;
		if (kill_proc (trap->pid, trap->sig, 1) != 0)
		    trap_del (dev, trap->oid, trap->pid, trap->sig);
	    }
	}
    }
    if (requests & REQFRAMEADD)
    {
	_BUGMSG(DBG_IRQ, "..ADD");
	if ((framep = dev_alloc_frame (dev, 0)) != NULL)
	{
	    _BUGMSG(DBG_IRQ, "dev_alloc_frame returned %p pointer..", framep);
	    dev_frame_add (dev->base_addr, framep);
	}
	else
	    _BUGMSG(DBG_IRQ, "dev_alloc_frame returned NULL pointer..\n");
    }
    if (requests & REQSERVICE)
    {
	_BUGMSG(DBG_IRQ, "..service");
	dev_service (dev->base_addr);
	dev_service (DEVICE_DEBUG);
    }
    if (requests & REQWATCHDOG)
    {
    	_BUGMSG(DBG_IRQ, "..watchdog");
    	if (lp->watchdog_event == WATCHDOG_TRIGGERED &&
    	    lp->watchdog_timeout > 0)
    	{
    	    /*
    	     * Lets trigger the watchdog service, for the current
    	     * device.
    	     */
    	    dev_watchdog(dev->base_addr);
    	    lp->watchdog_event = WATCHDOG_NOT_TRIGGERED;
    	    _BUGMSG(DBG_IRQ, "..set for dev: %s",dev->name);
    	}
    }

    _BUGMSG(DBG_IRQ, "\n");
}

/**********************************************************************
 *  convert_mvc_trap_to_oid
 *
 *  DESCRIPTION: Convert a trap that is generated by the MVC and the
 *		 data value of the trapped OID to one new OID. We need
 *		 one OID because apps can subscribe to traps and receive
 *		 a signal when the trap occurs. It is not possible to send
 *		 data with it, so we encode the data in the new OID.
 *
 *  PARAMETERS:	 trap OID and trap data.
 *
 *  RETURN:	 A new OID that represents the trap and the data.
 *
 *  NOTE:	 The generated OIDs are defined in linux_mvc_oids.h.
 *
 **********************************************************************/
static int convert_mvc_trap_to_oid(int oid, int data)
{
    int retval;

    switch(oid)
    {
    case BLOB_OID_BUTTONSTATE:
	if (data == 0) {
	    retval = BLOB_OID_BUTTONSTATE_RELEASED;
	} else {
	    retval = BLOB_OID_BUTTONSTATE_PRESSED;
	}
	break;
    default:
	printk("Unknown oid 0x%lx to convert\n", (long unsigned int) oid);
	retval = 0;
    }

    return retval;
}

/**********************************************************************
 *  handle_blob_reqs
 *
 *  DESCRIPTION: Handle requests for device blob. These requests are
 *		 handles within the Ethernet device. Therefore, traps
 *		 for the blob device should be added to the Ethernet device.
 *
 *  PARAMETERS:	 Device struct of the Ethernet device
 *
 *  RETURN:	 NONE
 *
 *  NOTE:	 Called from the interrupt service routine.
 *
 **********************************************************************/
static void handle_blob_reqs(struct net_device *dev)
{
    int requests;

    requests = blob_reqs( );

    if (!requests)
	return;

    if (requests & 0xFFFFEFE0) {
	printk (" called with unknown requests: %d!!\n",
		requests);
	return;
    }

    if (requests & REQTRAP)
    {
	struct msg_conf *trapmsg = blob_trap( );
	if( trapmsg != NULL )
	{
	    struct trap_info	*trap;
	    struct trap_info	*root;
	    struct net_local	*lp;
	    int			trap_oid;

	    lp = (struct net_local *) dev->priv;

	    root = lp->traps;

	    trap_oid = convert_mvc_trap_to_oid(trapmsg->oid, *trapmsg->data);
	    while ((trap = trap_oid_find (root, trap_oid)) != NULL)
	    {
		BUGMSG(DBG_TRAP, " sending signal %d to pid: %d\n", trap->sig,
			    trap->pid);
		root = trap->next;
		if (kill_proc (trap->pid, trap->sig, 1) != 0)
		    trap_del (dev, trap_oid, trap->pid, trap->sig);
	    }
	}
    }
    if (requests & REQWATCHDOG)
    {
        /*
         * Lets trigger the blob watchdog service, for the BLOB device
         */
        blob_watchdog();
        _BUGMSG(DBG_IRQ, "..set for dev: DEVICE_BLOB");
    }
}


/**********************************************************************
 *  prism_eth_init
 *
 *  DESCRIPTION: Initialise the MVC to a specific state.
 *
 *  PARAMETERS:	 Device struct of the Ethernet device
 * 		 The operation to perform on the MVC
 *
 * RETURN:	 Possible error code
 *
 *  NOTE:	 Depending on the current state some values should
 * 		 possibly be reinitialised in the MVC.
 *
 **********************************************************************/
static int prism_eth_init( struct net_device *dev, int startp )
{
    long state;
    int error = BER_NONE;

    /* first make sure the device is in ready state, we don't know in what
       state we come here
     */
    BUGMSG(DBG_OPEN, "prism_eth_init: base_addr: %ld...", dev->base_addr);

    state = dev_state (dev->base_addr);
    BUGMSG(DBG_OPEN, "dev_state: %ld...", state);
    switch (startp)
    {
	case PRISM_ETH_STOP:  /* MVC State will end up being STSTOPPED */
	    /*
	     * In this case state doesn't realy mater,
	     * if not already stopped stop now!
	     */
	    if(state != STSTOPPED)
	    {
		BUGMSG(DBG_OPEN, "dev_stop...");
		error = dev_stop (dev->base_addr);
	    }
	break;
	case PRISM_ETH_HALT:  /* MVC State will end up being STREADY */
	    /*
	     * Man can only HALT a device if it's RUNNING.
	     */
	    if(state == STRUNNING)
	    {
		BUGMSG(DBG_OPEN, "dev_halt...");
		error = dev_halt (dev->base_addr);
	    }
	    else
	    {
		error = -EPERM;
	    }
	break;
	case PRISM_ETH_START: /* MVC State will end up being STREADY */
	    /*
	     * Man can only START a device if it's STOPPED.
	     */
	    if(state == STSTOPPED)
	    {
		BUGMSG(DBG_OPEN, "dev_start...");
   	        error = dev_start (dev->base_addr, 0);
   	        if (error == BER_NONE)
   	            error = prism_eth_start ( dev );
	    }
	    else
	    {
		error = -EPERM;
	    }
	break;
	case PRISM_ETH_RUN:   /* MVC State will end up being RUNNING */
	    /*
	     * Man can only RUN a device if it's READY to run.
	     */
	    if(state == STREADY)
	    {
		BUGMSG(DBG_OPEN, "dev_run...");
		prism_eth_run ( dev );
		if (error == BER_NONE)
   	            error = dev_run (dev->base_addr);
	    }
	    else
	    {
		error = -EPERM;
	    }
	break;
        default:
	    error = BER_UNSUP;
    }
    if (error != BER_NONE)
        BUGMSG(DBG_OPEN | DBG_ERROR, "error: %d\n", error);
#if (NET_DEBUG & BG_OPEN)
    else
    {
	state = dev_state (dev->base_addr);
	printk ("dev_state changed to: %ld...", state);
    }
#endif
    return error;
}

static int prism_eth_start ( struct net_device *dev )
{
    int error;
    int requests = 0;
    struct msg_conf setmsg;
    long mode;

    /* set the MAC address */
    setmsg.operation = OPSET;
    setmsg.oid = GEN_OID_MACADDRESS;
    setmsg.data = (long *) dev->dev_addr;
    if ((error = dev_conf (dev->base_addr, &setmsg)) < BER_NONE)
    return error;

    BUGMSG(DBG_OPEN, "MAC set..");

    /* for the WLAN device, set the MLME mode */
    if (dev->base_addr == DEVICE_WLAN)
    {
	mode = DOT11_MLME_SIMPLE;
	setmsg.oid = DOT11_OID_MLMEAUTOLEVEL;
	setmsg.data = &mode;
	if ((error = dev_conf (dev->base_addr, &setmsg)) < BER_NONE)
	    return error;

	BUGMSG(DBG_OPEN, "MLME mode set..");
    }

    /* handle all device requests (= add frames) */
    while ((requests = dev_reqs (dev->base_addr)) > BER_NONE)
    {
	dev_handle_reqs (dev, requests);
    }
    BUGMSG(DBG_OPEN, "requests handled..");

    return error;
}

static int prism_eth_run ( struct net_device *dev )
{
    struct msg_conf setmsg;
    struct obj_outputpowertable outpwrtable;
    int MVC_OutputPower;
    int error = BER_NONE;

    /* if we are starting the WLAN device, then... */
    if (dev->base_addr == DEVICE_WLAN)
    {
	/* set the OUTPUTPOWERTABLE address */
	setmsg.operation = OPSET;
	setmsg.oid = DOT11_OID_OUTPUTPOWERTABLE;

	/* Set the output power structure */
	outpwrtable.outputpowertable = (long *)&HardConfig->CalibrateRows;
	if(HardConfig->DefaultRegDomain == 0x40)
	    outpwrtable.index = DOT11_OUTPWRTABLE_MKK;
	else
	    outpwrtable.index = DOT11_OUTPWRTABLE_FCC;
	setmsg.data = (long*)&outpwrtable;
	if ((error = dev_conf (dev->base_addr, &setmsg)) < BER_NONE)
	    return error;

	BUGMSG(DBG_OPEN, "OUTPUTPOWERTABLE set..");

	/* Set the desired output power to the MVC */
	if (HardConfig->OutputPower == 0xffffffff)
	    MVC_OutputPower = 5000;
	else
	    MVC_OutputPower = (HardConfig->OutputPower / 20);
	setmsg.operation = OPSET;
	setmsg.oid = DOT11_OID_OUTPUTPOWER;
	setmsg.data = (long *)&MVC_OutputPower;
	if ((error = dev_conf (dev->base_addr, &setmsg)) < BER_NONE)
	    return error;

	BUGMSG(DBG_OPEN, "OUTPUTPOWER set..\n");

#ifndef CONFIG_FREE_WLAN_PINS

	/* don't disable the UART completely, this crashes the system ?! */
	uUART->CR &= ~0x78;	/* disable UART interrupts */

	/* make the UART pins free for the radio */
	uGPIO->FSEL = (GPIOFSELUARTTx | GPIOFSELUARTRx) << 16;

#endif

    }

    return error;
}

/*
 * Open/initialize the board. This is called (in the current kernel)
 * sometime after booting when the 'ifconfig' program is run.
 *
 * This routine should set everything up anew at each open, even
 * registers that "should" only need to be set once at boot, so that
 * there is non-reboot way to recover if something goes wrong.
 */
static int
net_open(struct net_device *dev)
{
    struct net_local *lp = (struct net_local *)dev->priv;
    int irqval;
    int i;

    /* Reset the hardware and set the card MAC address. */
    prism_eth_init( dev, PRISM_ETH_RUN );

    netif_start_queue (dev);

    irqval = request_irq(dev->irq, &net_interrupt, SA_SHIRQ | SA_INTERRUPT,
		    dev->name, dev);
    if (irqval)
    {
	printk("%s: unable to get IRQ %d (irqval=%d).\n",
	       dev->name, dev->irq, irqval);
	return -EBUSY;
    }

    /* determine initial linkstate for Ethernet interface */
    if (dev->base_addr == DEVICE_HOSTDS)
    {
	if (dev_getval (dev, GEN_OID_LINKSTATE) == 0)
	    netif_carrier_off (dev);
    }

    /*
     * If the current device, is host device for WDS devices:
     * bring the associated WDS devices up and RUNNING.
     */
    if (dev->base_addr == DEVICE_WLAN && lp->current_nof_links > 0)
    {
	printk("Open all associated WDS links: ");
	for (i = 0; i < lp->current_nof_links; i++)
	{
	    printk("WDS%d...",i);
	    dev_open(lp->wds_dev[i]);
	}
	printk("\n");
    }
    return 0;
}


void do_hard_print(char *string)
{
#ifdef CONFIG_POLDEBUG_PRINTK
       struct msg_data debug_msg;

       debug_msg.data = string;
       debug_msg.length = strlen(string) + 1;
       debug_msg.max = 0;

       dev_write(DEVICE_DEBUG, &debug_msg);
#endif
}



static void net_tx_timeout(struct net_device *dev)
{
    struct net_local *np = (struct net_local *)dev->priv;

    if (dev_getval (dev, GEN_OID_LINKSTATE) == 0)
    {
	printk(KERN_WARNING "%s: link down, network cable problem?\n",
		dev->name);
	netif_carrier_off (dev);
    }
    else
    {
	printk(KERN_WARNING "%s: reset interface\n", dev->name);

	/* Try to restart the adaptor. */
	prism_eth_init(dev, PRISM_ETH_STOP);
	prism_eth_init(dev, PRISM_ETH_START);
	prism_eth_init(dev, PRISM_ETH_RUN);
	np->stats.tx_errors++;

	netif_wake_queue(dev);
    }
}

static int
net_send_packet(struct sk_buff *skb, struct net_device *dev)
{
    static int tries = 1000000; /* give the first packet a lot of retries */

    unsigned char *src;
    struct sk_buff *skb2;
    struct msg_frame *framep;
    int error;
#ifdef WDSLINK_SUPPORT
    int WDSLink = 0;
    unsigned char WDS_Mac[6];
#endif

    struct net_local *lp = (struct net_local *)dev->priv;

    _BUGMSG(DBG_TX, "net_send_packet:..");

#ifdef WDSLINK_SUPPORT
    /* If this skb came from a WDS device, the
     * skb->cb field will be filled with a WDSLink MACaddress.
     * In this case the MAC address must be filled in the msg_frame.
     * Due to the reason the cb field is free to use, the MAC address
     * will be packed in the following keys "WDS"<MAC ADDRESS>"Link".
     */
    if(*(long*)&skb->cb[0] == 0x4c534457) /* WDSL */
    {
	memcpy(WDS_Mac, skb->cb+4, 6);
	WDSLink = 1;
#if (NET_DEBUG & DBG_WDS)
	printk("net_send_packet: found wdsmac in skb: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x \n",
		WDS_Mac[0],WDS_Mac[1],WDS_Mac[2],WDS_Mac[3],WDS_Mac[4],WDS_Mac[5]);
#endif
 	/* Make the WDS data invalid */
	*(long*)&skb->cb[0] = 0;
    }
#endif //WDSLINK_SUPPORT

    /* if the head or tail room doesn't comply with the device or
       the skb is cloned .. */
    if (skb_headroom (skb) < lp->dev_desc->header ||
	skb_tailroom (skb) < lp->dev_desc->trailer ||
	skb_cloned (skb))
    {
	if (!skb_cloned (skb))
	{
	    _BUGMSG(DBG_TX, "no head/tail room, head:(%d,%d),tail(%d,%d)..",
		skb_headroom (skb), lp->dev_desc->header,
		skb_tailroom (skb), lp->dev_desc->trailer);
	    lp->tx_misaligned++;
	}
	/* see if we can shift the data in the buffer */
	if (!skb_cloned (skb) &&
	    skb_headroom (skb) + skb_tailroom (skb) >=
	    lp->dev_desc->header + lp->dev_desc->trailer)
	{
	    _BUGMSG(DBG_TX, "shift data: %d..",
		    lp->dev_desc->header - skb_headroom (skb));

	    src = skb->data;
	    skb_reserve (skb, lp->dev_desc->header - skb_headroom (skb));
	    memmove (skb->data, src, skb->len);
	}
	else	/* copy to a new buffer with enough head and tail room */
	{
#if (NET_DEBUG & DBG_TX)
	    if (skb_cloned (skb))
		printk ("cloned sk_buff.. ");
	    printk ("copy data.. ");
#endif
	    skb2 = dev_copy_expand_skb (dev, skb);
	    dev_kfree_skb (skb);
	    if (skb2 == NULL)
		return 0;
	    skb = skb2;
	}
    }

    if ((framep = skb_to_frame (skb)) == NULL)
    {
	BUGMSG(DBG_TX | DBG_ERROR, "NO frame!!!\n");
	dev_kfree_skb (skb);
	lp->stats.tx_dropped++;
    }
    else
    {
	/* send the frame to prism */

	_BUGMSG(DBG_TX, "send: size: %d, headroom: %d..", framep->size,
		skb_headroom (framep->handle));

#ifdef WDSLINK_SUPPORT
	if (WDSLink == 1)
	{
	    /* Copy the found MAC addres in the msg_frame. */
	    memcpy(((struct msg_frame_wds*)framep)->wds_link, WDS_Mac, 6);

#if (NET_DEBUG & DBG_WDS)
	    {
		struct msg_frame_wds *wdsframep;
		wdsframep = (struct msg_frame_wds*)framep;
		printk("net_send_packet: copied wdsmac in framep %p: \
		    %02x-%02x-%02x-%02x-%02x-%02x\n", framep,
		    wdsframep->wds_link[0],wdsframep->wds_link[1],
		    wdsframep->wds_link[2],wdsframep->wds_link[3],
		    wdsframep->wds_link[4],wdsframep->wds_link[5]);

	    }
#endif
	}
#endif //WDSLINK_SUPPORT

	do
	{
	    error = dev_frame_tx (dev->base_addr, framep);
	    if (error < BER_NONE)
		dev_service (dev->base_addr);
	    tries--;
	}
	while (error < BER_NONE &&
	       error != BER_DSTUNKN &&
	       framep->status != FSTXREJECTED &&
	       tries);
	tries = TX_NR_RETRIES;

        if (error < BER_NONE)
	{
	    if (error != BER_DSTUNKN)
	    {
#ifndef CONFIG_FREE_WLAN_PINS
		BUGMSG(DBG_TX | DBG_ERROR,
			KERN_ALERT "TX dev: %s, NOT accepted!!\n", dev->name);
#endif
		/* test if the link is down */
		if (dev_getval (dev, GEN_OID_LINKSTATE) == 0)
		    netif_carrier_off (dev);
		else
		/* if it's not accepted, then stop queing packets */
		    netif_stop_queue (dev);

		lp->stats.tx_dropped++;
	    }

	    /* abort the frame */
	    dev_free_frame (dev, framep);
	}
	else
	    _BUGMSG(DBG_TX | DBG_WDS, "OK!!\n");
    }

    dev->trans_start = jiffies;

    return 0;
}


/*
 * The typical workload of the driver:
 *   Handle the network interface interrupts.
 */
static void
net_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{
    struct net_device	    *dev;
    int			requests;

    uIRQ->Soft = 0;	    /* remove soft irq */

#if (NET_DEBUG & DBG_IRQ)
    {
	static int irq_count;
	static int starttime;
	int diff;

	irq_count++;

	if ((diff = jiffies - starttime) >= HZ)
	{
	    printk ("irq_count: %d\n", (irq_count * diff) / HZ);
	    irq_count = 0;
	    starttime = jiffies;
	}
    }
#endif

    if (dev_id == NULL)
    {
	printk(KERN_WARNING "%s: irq %d for unknown device.\n", cardname, irq);
	return;
    }

    dev = (struct net_device *) dev_id;

    // We handle blob requests withing the Ethernet device
    if (dev->base_addr == DEVICE_HOSTDS)
	handle_blob_reqs(dev);

    if ((requests = dev_reqs (dev->base_addr)) > BER_NONE)
	dev_handle_reqs (dev, requests);

    return;
}


/* The inverse routine to net_open(). */
static int
net_close(struct net_device *dev)
{
    struct net_local *lp = (struct net_local *)dev->priv;
    struct msg_frame *framep;
    int i;

    /*
     * If the current device, is host device for WDS devices:
     * bring the associated WDS devices DOWN.
     */
    if (dev->base_addr == DEVICE_WLAN && lp->current_nof_links > 0)
    {
	printk("Close associated WDS links: ");
	for (i = 0; i < lp->current_nof_links; i++)
	{
	    printk("WDS%d...",i);
	    dev_close(lp->wds_dev[i]);
	}
	printk("\n");
    }

    netif_stop_queue (dev);

    if (dev_halt (dev->base_addr) >= BER_NONE)
    {
	do
	{
	    if ((framep = dev_frame_return (dev->base_addr)) != NULL)
	    {
		dev_free_frame (dev, framep);
	    }
	}
	while (framep != NULL);
    }

    free_irq(dev->irq, dev);

    return 0;
}


static struct trap_info *trap_prev_find (struct net_device *dev,
					struct trap_info *trap)
{
    struct trap_info *prev;
    struct trap_info *root;

    prev = NULL;
    root = ((struct net_local *)dev->priv)->traps;

    while (root)
    {
	if (root == trap)
	    break;
	prev = root;
	root = root->next;
    }
    return prev;
}


static struct trap_info *trap_oid_find (struct trap_info *root, int oid)
{
    BUGMSG (DBG_TRAP, "trap_oid_find: oid: %08X, ", oid);

    while (root)
    {
	if (root->oid == oid)
	    break;
	root = root->next;
    }

    BUGMSG (DBG_TRAP, root == NULL ? "not found\n" : "pid: %d, sig: %d\n",
	root->pid, root->sig);

    return root;
}


static struct trap_info *trap_find (struct net_device *dev, int oid, int pid,
				    int sig)
{
    struct trap_info *trap;

    trap = ((struct net_local *)dev->priv)->traps;
    while (trap)
    {
	if (trap->oid == oid && trap->pid == pid && trap->sig == sig)
	    break;
	trap = trap->next;
    }
    return trap;
}


static int trap_add (struct net_device *dev, int oid, int pid, int sig)
{
    struct trap_info *trap;

    /* don't add a trap for the second time to the device */
    if (trap_find (dev, oid, pid, sig))
	return 0;

    if ((trap = kmalloc(sizeof(struct trap_info), GFP_KERNEL)) == NULL)
	return -ENOMEM;

    BUGMSG (DBG_TRAP, "trap_add: dev: %s, oid: %08X, pid: %d, sig: %d\n",
	    dev->name, oid, pid, sig);

    trap->oid = oid;
    trap->pid = pid;
    trap->sig = sig;
    trap->next = ((struct net_local *)dev->priv)->traps;
    ((struct net_local *)dev->priv)->traps = trap;
    return 0;
}


static int trap_del (struct net_device *dev, int oid, int pid, int sig)
{
    struct trap_info *trap;
    struct trap_info *prev;

    BUGMSG (DBG_TRAP, "trap_del: dev: %s, oid: %08X, pid: %d, sig: %d\n",
	    dev->name, oid, pid, sig);

    if ((trap = trap_find (dev, oid, pid, sig)) == NULL)
	return 0;
    prev = trap_prev_find (dev, trap);

    if (prev)
	prev->next = trap->next;
    else
	((struct net_local *)dev->priv)->traps = trap->next;
    kfree (trap);
    return 0;
}


static void dev_authenticate (struct net_device *dev, struct obj_mlme *obj)
{
    int i;
    struct msg_conf prism_msg;
    struct net_local *lp;

    _BUGMSG(DBG_TRAP, "dev_authenticate: STA %02x-%02x-%02x-%02x-%02x-%02x\n",
	    obj->address[0], obj->address[1], obj->address[2],
	    obj->address[3], obj->address[4], obj->address[5]);

    lp = (struct net_local *)dev->priv;

    if ((lp->acl == NULL) || (lp->acl_nr_items == 0))
    {
	obj->code = DOT11_SC_UNSPECIFIED;
    }
    else
    {
	for (i = 0; i < lp->acl_nr_items; i++)
	{
	    if ( (memcmp (&lp->acl[i*6], cAclFullAccess, 6) == 0) ||
		 (memcmp (&lp->acl[i*6], obj->address, 6) == 0) )
		break;
	}
	if (i < lp->acl_nr_items) {
	    obj->code = DOT11_SC_SUCCESSFUL;
	}
	else {
	    obj->code = DOT11_SC_UNSPECIFIED;
	}
    }
    prism_msg.operation = OPSET;
    prism_msg.oid = DOT11_OID_AUTHENTICATE;
    prism_msg.data = (long *) obj;
    dev_conf (dev->base_addr, &prism_msg);

    // Copy mlme data to global mlme for later retrieval by GETMLME ioctl
    memcpy(&mlme, obj, sizeof(struct obj_mlme));

}


static void dev_deauthenticate (struct net_device *dev, struct obj_mlme *obj)
{
    _BUGMSG(DBG_TRAP, "dev_deauthenticate: STA %02x-%02x-%02x-%02x-%02x-%02x\n",
	    obj->address[0], obj->address[1], obj->address[2],
	    obj->address[3], obj->address[4], obj->address[5]);
}

/**********************************************************************
 *  iapp_send_update_frame
 *
 *  DESCRIPTION: sends an IAPP update frame when a wireless station
 *               associates (refer to 802.11f, sec. 6.3).
 *               The "802.2 Type 1 Logical Link Control Exchange Identifier
 *               update response frame" is sent using the MAC address of the
 *               wireless station that has associated.
 *               So, any layer 2 devices can update their forwarding tables
 *               with the correct port to reach the new location.
 *
 *  PARAMETERS:  dev - device to handle requests for (since this is called when
 *                     an XXXassociate trap is received, this is always the WLAN device.
 *               obj - Contains the mlme object trap data.
 *
 *  RETURN:      0 for success, -1 for  failure
 *
 **********************************************************************/
static int iapp_send_update_frame(struct net_device *dev, const struct obj_mlme *obj)
{
    struct sk_buff     *skb;
    unsigned char       buf[20];

    memset(buf, 0xff, 6);		/* Broadcast Destination address */
    memcpy(buf + 6, obj->address, 6);	/* Source address of associated client */

    buf[12] = 0;			/* Length of payload = 6 octets */
    buf[13] = 6;
    buf[14] = 0;			/* DSAP */
    buf[15] = 0;			/* SSAP */
    buf[16] = 175;			/* Control field */

    buf[17] = 129;			/* XID = 129.1.0 */
    buf[18] = 1;
    buf[19] = 0;

    skb = dev_alloc_skb(20+2);

    if (skb)
    {
	memcpy(skb_put(skb, 20), buf, 20);

	skb->dev = dev;
	skb->protocol = eth_type_trans(skb, dev);
	skb->ip_summed = CHECKSUM_UNNECESSARY;
	netif_rx(skb);
	return 0;
    }

    return -1;
}

static void dev_associate (struct net_device *dev, struct obj_mlme *obj)
{
    struct msg_conf prism_msg;
    int error;
    long nr_clients_assoc;

    _BUGMSG(DBG_TRAP, "dev_associate: STA %02x-%02x-%02x-%02x-%02x-%02x\n",
	    obj->address[0], obj->address[1], obj->address[2],
	    obj->address[3], obj->address[4], obj->address[5]);

    prism_msg.operation = OPGET;
    prism_msg.oid = DOT11_OID_CLIENTSASSOCIATED;
    error = dev_conf (dev->base_addr, &prism_msg);
    nr_clients_assoc = *prism_msg.data;

    if (error < BER_NONE || nr_clients_assoc >= MAX_NR_CLIENTS_ASSOC) {
	obj->code = DOT11_SC_CANNOT_HANDLE_ASSOCS;
    }
    else {
	obj->code = DOT11_SC_SUCCESSFUL;

	if(iapp_send_update_frame(dev, obj) < 0)
	    printk(KERN_WARNING "Failed to send IAPP update frame\n");
    }

    prism_msg.operation = OPSET;
    prism_msg.oid = DOT11_OID_ASSOCIATE;
    prism_msg.data = (long *) obj;
    dev_conf (dev->base_addr, &prism_msg);

    // Copy mlme data to global mlme for later retrieval by GETMLME ioctl
    memcpy(&mlme, obj, sizeof(struct obj_mlme));
}

static void dev_reassociate (struct net_device *dev, struct obj_mlme *obj)
{
    struct msg_conf prism_msg;
    struct obj_sta *sta;
    struct obj_mlme mobj;
    int error;
    int err;
    long nr_clients_assoc;

    _BUGMSG(DBG_TRAP, "dev_reassociate: AP  %02x-%02x-%02x-%02x-%02x-%02x\n",
	    obj->address[0], obj->address[1], obj->address[2],
	    obj->address[3], obj->address[4], obj->address[5]);

    prism_msg.operation = OPGET;
    prism_msg.oid = DOT11_OID_CLIENTSASSOCIATED;
    error = dev_conf (dev->base_addr, &prism_msg);
    nr_clients_assoc = *prism_msg.data;

    // We have a reassociate - get client MAC address

    prism_msg.operation = OPGET;
    prism_msg.oid = DOT11_OID_CLIENT1 + obj->id;
    err = dev_conf(dev->base_addr, &prism_msg);
    sta = prism_msg.data;

    _BUGMSG(DBG_TRAP, "dev_reassociate: STA %02x-%02x-%02x-%02x-%02x-%02x\n",
	    sta->address[0], sta->address[1], sta->address[2],
	    sta->address[3], sta->address[4], sta->address[5]);

    memcpy(&mobj.address, &sta->address, 6);

    if (error < BER_NONE || nr_clients_assoc >= MAX_NR_CLIENTS_ASSOC) {
	obj->code = DOT11_SC_CANNOT_HANDLE_ASSOCS;
    }
    else {
	obj->code = DOT11_SC_SUCCESSFUL;

	if(iapp_send_update_frame(dev, &mobj) < 0)
	    printk(KERN_WARNING "Failed to send IAPP update frame\n");
    }

    prism_msg.operation = OPSET;
    prism_msg.oid = DOT11_OID_REASSOCIATE;
    prism_msg.data = (long *) obj;
    dev_conf (dev->base_addr, &prism_msg);

    // Copy mlme data to global mlme for later retrieval by GETMLME ioctl
    memcpy(&mlme, obj, sizeof(struct obj_mlme));
}


static void dev_disassociate (struct net_device *dev, struct obj_mlme *obj)
{
    _BUGMSG(DBG_TRAP, "dev_disassociate: STA %02x-%02x-%02x-%02x-%02x-%02x\n",
	    obj->address[0], obj->address[1], obj->address[2],
	    obj->address[3], obj->address[4], obj->address[5]);

    // Copy mlme data to global mlme for later retrieval by GETMLME ioctl
    memcpy(&mlme, obj, sizeof(struct obj_mlme));
}


static int net_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    struct prism_ioctl req;
    struct msg_conf prism_msg;
    int err;
    unsigned int nr_items;
    struct net_local *lp;
    struct prism_ioctl *ioc;

    if (!capable (CAP_NET_ADMIN))
	return -EPERM;

    if (dev == NULL)		/* woops! */
	return -ENODEV;

    if (cmd != PRISMIOCTL)
	return -ENOTTY;		/* Inapropriate ioctl for device */

    if (copy_from_user (&req, ifr->ifr_data, sizeof(struct prism_ioctl)) != 0)
    	return -EFAULT;

    switch (req.cmd)
    {
	case OPGET:
	    break;
	case OPSET:
	    if (!access_ok(VERIFY_READ, req.data, req.len))
		return -EFAULT;
	    break;
	case OPTRAPADD:
	    err = trap_add (dev, req.oid, current->pid, (int) req.data);
	    return err;
	case OPTRAPDEL:
	    err = trap_del (dev, req.oid, current->pid, (int) req.data);
	    return err;
	case OPRESET:
	    err = blob_reset ();
	    return err;
	case OPGETMLME:
	    ioc = (struct prism_ioctl *)(ifr->ifr_data);
	    err = copy_to_user(ioc->data, &mlme, sizeof(struct obj_mlme));
	    return (err == 0) ? 0 : -EPERM;
	default:
	    return -EINVAL;
    }

    lp = (struct net_local *)dev->priv;

    prism_msg.operation = req.cmd;
    prism_msg.oid = req.oid;
    prism_msg.data = (long *) req.data;

    if (req.oid == DOT11_OID_ACL)
    {
	if (req.cmd == OPSET)
	{
	    if (req.len % 6)
	    return -EINVAL;
	    nr_items = req.len / 6;
	    if (nr_items > MAX_NR_CLIENTS_AUTH)
	    return -EINVAL;
	    if (lp->acl != NULL)
	    {
	    kfree (lp->acl);
	    lp->acl_nr_items = 0;
	    }
	    if ((lp->acl = kmalloc (req.len, GFP_KERNEL)) == NULL)
	    return -ENOMEM;
	    if (copy_from_user (lp->acl, req.data, req.len) != 0)
		    return -EFAULT;
	    lp->acl_nr_items = nr_items;
	}

	if (req.cmd == OPGET)
	{
	    prism_msg.data = (long *) lp->acl;
	    req.len = lp->acl_nr_items * 6;
	}
    }
    else if (req.oid == DOT11_OID_WDSLINKADD && req.cmd == OPSET)
    {
	if((err = AddWDSLink(req, dev)) != BER_NONE)
	    return err;
	if ((err = dev_conf (dev->base_addr, &prism_msg)) < BER_NONE)
	{
	    DelWDSLink(req,dev);
	    return -EINVAL;
	}
	return err;
    }
    else if (req.oid == DOT11_OID_WDSLINKREMOVE && req.cmd == OPSET)
    {
	if((err = DelWDSLink(req, dev)) != BER_NONE)
	    return err;
	if ((err = dev_conf (dev->base_addr, &prism_msg)) < BER_NONE)
	    return -EINVAL;
	return err;
    }
    else if (req.oid == GEN_OID_TRIGGER_WATCHDOG)
    {
	/* Watchdog trigger event, this trigger is stored untill
	 * a watchdog request is received.
	 */
	BUGMSG(DBG_IOCTL, "net_ioctl: GEN_OID_TRIGGER_WATCHDOG for dev: %s\n",
	       dev->name);
	lp->watchdog_event = WATCHDOG_TRIGGERED;
    }
    else if (req.oid == GEN_OID_WATCHDOG && req.cmd == OPSET)
    {
	/* Set WATCHDOG timeout on dev */
	if ((err = dev_conf (dev->base_addr, &prism_msg)) < BER_NONE)
	    return -EINVAL;
	/* Store the watchdog timeout */
	lp->watchdog_timeout = *prism_msg.data;
	BUGMSG(DBG_IOCTL, "net_ioctl: GEN_OID_WATCHDOG for dev: %s set to %ld\n",
		dev->name, *prism_msg.data);
    }
    else
    {

	if ( (req.oid & 0xFF000000) == 0x80000000)
	{
	    /* This is an OID for the BLOB device. We only allow
	     * access to the blob device via eth0
	     */
	    if (dev->base_addr == DEVICE_HOSTDS)
	    {
		/* Check if we're manipulating the WLAN LED.
		 * If so, we should halt the WLAN device.
		 */
		err = BER_NONE;
		switch (req.oid)
		{
		    case BLOB_OID_LEDSET:
		    case BLOB_OID_LEDCLEAR:
		    if ( (*req.data & 2) &&
			     (dev_state(DEVICE_WLAN) == STRUNNING) )
		    {
			err = dev_halt(DEVICE_WLAN);
		    }
		    break;
		    case BLOB_OID_LED_RESET:
		    if ( (*req.data & 2) &&
			     (dev_state(DEVICE_WLAN) == STREADY) )
		    {
			err = dev_run(DEVICE_WLAN);
		    }
		    return (err < BER_NONE) ? -EINVAL : 0;
		    default:
		    /* Not one of these */
		    break;
		}
		if (err < BER_NONE)
		    return -EINVAL;

		if ((err = blob_conf(&prism_msg)) < BER_NONE)
		    return -EINVAL;
	    }
	    else
	    {
		return -EOPNOTSUPP;
            }
	}
	else
	{
	    if ((err = dev_conf (dev->base_addr, &prism_msg)) < BER_NONE)
		return -EINVAL;
	}
    }

    if (req.cmd == OPGET)
    {
       if (copy_to_user (req.data, prism_msg.data, req.len) != 0)
	   return -EFAULT;
    }
    return 0;
}


static long dev_getval (struct net_device *dev, int oid)
{
    struct msg_conf getmsg;

    getmsg.operation = OPGET;
    getmsg.oid = oid;
    dev_conf (dev->base_addr, &getmsg);
    return *(getmsg.data);
}


/*
 * Get the current statistics.
 * This may be called with the card open or closed.
 */
static struct net_device_stats *
net_get_stats(struct net_device *dev)
{
    int multicasts;
    struct net_local *lp = (struct net_local *)dev->priv;
    struct net_device_stats *stats = &lp->stats;

    /* get the statistics from the device registers. */
    multicasts = dev_getval (dev, GEN_OID_MSDUTXMULTICAST);
    stats->tx_packets = dev_getval (dev, GEN_OID_MSDUTXSUCCESSFUL)
				    + multicasts;
    stats->tx_errors = dev_getval (dev, GEN_OID_MSDUTXFAILED);
    stats->tx_dropped = dev_getval (dev, GEN_OID_MSDUTXDROPPED);

    stats->multicast = dev_getval (dev, GEN_OID_MSDURXMULTICAST);
    stats->rx_packets = dev_getval (dev, GEN_OID_MSDURXSUCCESSFUL)
				    + stats->multicast;
    stats->rx_errors = dev_getval (dev, GEN_OID_MSDURXFAILED);
    stats->rx_dropped = dev_getval (dev, GEN_OID_MSDURXDROPPED);

    /* only WLAN device has rx CRC error counter */
    if (dev->base_addr == DEVICE_WLAN)
	stats->rx_crc_errors = dev_getval (dev, DOT11_OID_FRAMEERRORS);

    return stats;
}


#ifdef MODULE

static struct net_device prism_devs[PRISM_NR_OF_DEVS];

static int __init prism_init_module(void)
{
    struct net_device *dev;
    int err;

    for (i = 0; i < PRISM_NR_OF_DEVS; i++)
    {
	dev = &prism_devs[i];
	dev->init = prism_init;

	/* Find a name for this unit */
	if ((err = dev_alloc_name(dev,"eth%d")) < 0)
	    return err;

	/* Register the device */
	if ((err = register_netdev(dev)) < 0)
	    return err;
    }
    return 0;
}

static void __exit prism_cleanup_module(void)
{
    int i;

    for (i = 0; i < PRISM_NR_OF_DEVS; i++)
    {
	dev = &prism_devs[i];
	unregister_netdev(dev);
	if (dev->priv)
	    kfree(dev->priv);
//	memset(dev, 0, sizeof(*dev));
//	dev->init = prism_init;
    }
}

module_init(prism_init_module);
module_exit(prism_cleanup_module);

#endif


/*
 *  Some debugging functions
 */

#if (NET_DEBUG & (DBG_TX | DBG_RX | DBG_ERROR))
/**********************************************************************
 *  print_frame
 *
 *  DESCRIPTION: Print the fields of prism frame 'framep' and the data.
 *
 *  PARAMETERS:	 framep - pointer to a prism frame
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
static void print_frame (struct msg_frame *framep)
{
    int offset;
    struct sk_buff *skb;

    skb = framep->handle;
    printk ("pr_frame: size %d, status %d, mpdu %p, skb->data %p, skb %p\n",
	    framep->size, framep->status, framep->mpdu, skb->data, skb);
    for (offset = 0; offset < framep->size; offset++)
	printk ("%02X ", framep->mpdu[offset]);
    printk ("\n");
}
#endif


#if (NET_DEBUG & DBG_SKB)

/**********************************************************************
 *  dump_hex_assci
 *
 *  DESCRIPTION: Dump the content of buffer 'buf' to the screen. The
 *		 data is presented in hexadecimal and ASSCI form, 16
 *		 bytes on each line.
 *
 *  PARAMETERS:	 buf  - pointer to buffer to display
 *		 size - size of the buffer
 *
 *  RETURN:	 Nothing.
 *
 **********************************************************************/
static void dump_hex_assci (unsigned char *buf, int size)
{
    int linesize, i;
    unsigned char *p;

    while (size)
    {
	linesize = size > 16 ? 16 : size;
	size -= linesize;
	for (p = buf, i = 0; i < linesize; i++, p++)
	    printk (" %02X", *p);
	for (; i < 17; i++)
	    printk ("	");
	for (i = 0; i < linesize; i++, buf++)
	{
	    if (*buf > 31 && *buf < 128)
		printk ("%c", *buf);
	    else
		printk (".");
	}
	printk ("\n");
    }
}


/**********************************************************************
 *  check_area
 *
 *  DESCRIPTION: Check if a memory area contains all zero's. If not,
 *		 print a warning, the address that isn't zero and dump
 *		 the content of the rest of the area on the screen.
 *
 *  PARAMETERS:	 p	   - start address of the area to check
 *		 size	   - size of the area
 *		 area_name - name of the area that is displayed in the
 *			     warning message
 *
 *  RETURN:	 0 if the whole area contains zero's, otherwise 1.
 *
 **********************************************************************/
static int check_area (unsigned char *p, int size, char *area_name)
{
    while (size)
    {
	if (*p)
	{
	    printk ("WARNING: %s changed at %p!\n", area_name, p);
	    dump_hex_assci (p, size);
	    return 1;
	}
	p++;
	size--;
    }
    return 0;
}
#endif

#if (NET_DEBUG & DBG_SKB)

/**********************************************************************
 *  frame_filter_check
 *
 *  DESCRIPTION: Filter frames not of interest for the Linux IP stack.
 *		 (unicasts not containing our MAC address and Intersil
 *		 debug frames). Perform some checks on the frame and
 *		 print warnings if something suspious is found.
 *
 *  PARAMETERS:	 dev	- device that received the frame
 *		 framep - pointer to the received prism frame
 *
 *  RETURN:	 0 if the frame is absorbed by this function, otherwise 1.
 *
 **********************************************************************/
static int frame_filter_check (struct net_device *dev, struct msg_frame *framep)
{
    struct sk_buff *skb;
    int offset;
    unsigned char *src;
    int mtu;

    /* filer out unicast frames not directed to our MAC address */
    if ((framep->mpdu[0] & 0x01) == 0)
    {
	for (offset = 0; offset < 6; offset++)
	{
	    if (framep->mpdu[offset] != dev->dev_addr[offset])
	    {
		dev_free_frame (dev, framep);
		return 0;
	    }
	}
    }

    /* filer out Intersil debug frames */
    if ((framep->mpdu[12] == 0x88) && (framep->mpdu[13] == 0x27))
    {
	dev_free_frame (dev, framep);
	return 0;
    }

    skb = framep->handle;

    /* first map pointer back in 0x08000000 area */
    ((unsigned)framep->mpdu) &= ~0x04000000;

    /* determine shift in framep->mpdu between add and rx */
    offset = framep->mpdu - skb->data;

    if (offset)
    {
	printk ("frame_to_skb: blob changed mpdu: %d!!!\n", offset);
	print_frame (framep);
    }
    if (skb_tailroom (skb) < framep->size)
	printk ("frame_to_skb: NOT enough room!! need %d, tailroom %d, len %d\n",
		framep->size, skb_tailroom (skb), skb->len);

    mtu = ((struct net_local *)dev->priv)->dev_desc->mtu;
    offset = check_area (skb->head, framep->mpdu - skb->head, "header");
    for (src = skb->end-1; (*src == 0) && (src > framep->mpdu+mtu); src--)
	;
    if (src > framep->mpdu+mtu)
    {
	check_area (framep->mpdu+mtu, (src - framep->mpdu+mtu) + 1, "trailer");
	offset = 1;
    }
    if (offset)
	print_frame (framep);
    return 1;
}
#endif

#ifdef WDSLINK_SUPPORT

#define ERROR(args...)   printk(KERN_ERR args)
/**********************************************************************
 *  AddWDSLink
 *
 *  DESCRIPTION: Add a WDSlink device to the system.
 * 		On adding a WDSLINK to the MVC the following should be done:
 * 		- Add a new device to the WDSlink devices list.
 * 		- Create a device for this WDSlink (dev_alloc wds%d)
 * 		- Register the just created dummy device (register_netdevice)
 *              - Bring the just created device up and running by calling dev_open.
 * 		- Somehow the brigde needs to be told to accept the new interface.
 *
 *  PARAMETERS:	 struct prism_ioctl req	- The request and data received
 *
 *
 *  RETURN:	 0 if the device created OK, otherwise 1.
 *
 **********************************************************************/
static int AddWDSLink(struct prism_ioctl req, struct net_device *dev)
{
    int err = BER_NONE;
    int i=0, j=0;

    struct net_local *lp = (struct net_local *)dev->priv;
    struct net_device *wds_dev;

    /* If the parent device isn't bridged, the WDS will not work */
    if (dev->br_port == NULL)
    {
	BUGMSG(DBG_WDS | DBG_ERROR, "AddWDSLink: Error parent device (%s) is not bridged, can't add WDS link.\n",dev->name);
        return -1;
    }
    /* First check if MAX_WDS_LINKS is reaced */
    if (lp->current_nof_links == MAX_WDS_LINKS)
    {
	/* Max number of links reaced */
	BUGMSG(DBG_WDS | DBG_ERROR,"AddWDSLink: Max number of WDS links (%d) reaced!\n", MAX_WDS_LINKS);
	return -1;
    }

    /* Check if this MAC addres is already in the list. */
    for (i=0 ; i< MAX_WDS_LINKS; i++)
    {
	/* If this MAC address is already linked don't add new WDS device */
	if(memcmp(((struct wds_net_local*)(lp->wds_dev[i]->priv))->wds_mac, req.data, 6) == 0)
	{
	    BUGMSG(DBG_WDS | DBG_ERROR,"AddWDSLink: WDS link for MAC: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x already active!\n",
		      req.data[0], req.data[1], req.data[2], req.data[3], req.data[4], req.data[5]);
	    return -1;
	}
	/* If this MAC address is our own mac address */
	if(memcmp(dev->dev_addr, req.data, 6) == 0)
	{
	    BUGMSG(DBG_WDS | DBG_ERROR,"AddWDSLink: Cannot add WDS link for own MAC address: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x\n",
		      req.data[0], req.data[1], req.data[2], req.data[3], req.data[4], req.data[5]);
	    return -1;
	}
    }
    /* Find an empty field to add WDS device */
    for (i=0 ; i< MAX_WDS_LINKS; i++)
    {
	if (lp->wds_dev[i] == NULL)
	{
	    /* We have found an empty structure
	     * So we can actualy create a new device for this link
	     */

	    if (!(lp->wds_dev[i] = dev_alloc("wds%d", &err)))
	    {
		BUGMSG(DBG_WDS | DBG_ERROR,"AddWDSLink: dev_alloc() for WDS failed!\n");
	        return err;
	    }

	    /* Get a pointer to this WDS device */
	    wds_dev = (struct net_device *)lp->wds_dev[i];

	    /* May be overridden by piggyback drivers */
	    ether_setup(wds_dev);

	    /* Initialize the device priv structure. */
	    if (wds_dev->priv == NULL)
	    {
		wds_dev->priv = kmalloc(sizeof(struct wds_net_local), GFP_KERNEL);
		if (wds_dev->priv == NULL)
		    return -ENOMEM;
	    }
	    memset(wds_dev->priv, 0, sizeof(struct wds_net_local));
	    /* Set the parent device for this WDS link */
	    ((struct wds_net_local*)(wds_dev->priv))->parent_dev = dev;
	    BUGMSG(DBG_WDS, "AddWDSLink: Added parent_dev->name %s\n",
	    	   ((struct wds_net_local*)(wds_dev->priv))->parent_dev->name);
	    /* Set WDS MAC in priv structure */
	    for (j = 0; j < 6; j++)
		printk(" %2.2x", ((struct wds_net_local*)(wds_dev->priv))->wds_mac[j] = req.data[j] );
	    printk("\n");

	    printk("AddWDSLink: for mac address:");
	    /* Retrieve and print the ethernet address. */
	    for (j = 0; j < 6; j++)
		printk(" %2.2x", wds_dev->dev_addr[j] = ((struct wds_net_local*)(wds_dev->priv))->parent_dev->dev_addr[j] );
	    printk("\n");
	    wds_dev->addr_len = 6;

	    /* Override the network functions we need to use */
	    wds_dev->open		= wds_net_open;
	    wds_dev->stop		= wds_net_close;
	    wds_dev->get_stats		= wds_net_get_stats;
	    wds_dev->hard_start_xmit	= wds_net_send_packet;
	    wds_dev->do_ioctl		= wds_net_ioctl;

	    /* Make ifconfig display some details */
	    wds_dev->base_addr = 0;

	    /* No need to lock cause the net_ioctl is already locked */
	    err = register_netdevice(wds_dev);
	    if (err)
	    {
		BUGMSG(DBG_WDS | DBG_ERROR,"AddWDSLink: register_netdev() for WDS failed!\n");
		return err;
	    }
	    BUGMSG(DBG_WDS, "AddWDSLink: registered device name: %s\n", wds_dev->name);

	    /* Bring the device up, when the root device is UP... */
	    if (dev->flags & IFF_UP)
	    {
		err = dev_open(wds_dev);
		if (err)
		{
		    BUGMSG(DBG_WDS | DBG_ERROR,"AddWDSLink: dev_open() for WDS failed!\n");
		    return err;
		}
		BUGMSG(DBG_WDS,"AddWDSLink: bringing up device name: %s\n", wds_dev->name);
	    }
	    else
	    {
		BUGMSG(DBG_WDS,"AddWDSLink: NOT bringing up device name: %s, cause root device isn't up yet...\n", wds_dev->name);
	    }
	    lp->current_nof_links++;
	    return err;
	}
    }
    /* This should not happen*/
    BUGMSG(DBG_WDS | DBG_ERROR,"AddWDSLink: WDS link administration screwed\n");
    return -1;
}

/**********************************************************************
 *  DelWDSLink
 *
 *  DESCRIPTION: Del a WDSlink device to the system.
 * 		On deleting a WDSLINK from the MVC the following should be done:
 *      - Bring the device down by calling dev_close.
 * 		- Unregister the asociated dummy device (unregister_netdevice)
 * 		- Free the allocated space for this device
 * 		- Remove this device from the WDSlink devices list.
 *
 *  PARAMETERS:	 struct prism_ioctl req	- The request and data received
 *
 *
 *  RETURN:	 0 if the device created OK, otherwise 1.
 *
 **********************************************************************/
static int DelWDSLink(struct prism_ioctl req, struct net_device *dev)
{
    int err = BER_NONE;
    int i=0;

    struct net_local *lp = (struct net_local *)dev->priv;
    struct net_device *wds_dev;

    /* Check if this MAC addres is in the WDS link list. */
    for (i=0 ; i< MAX_WDS_LINKS; i++)
    {
	if(memcmp(((struct wds_net_local*)(lp->wds_dev[i]->priv))->wds_mac, req.data, 6) == 0)
	{

	    wds_dev = (struct net_device *)lp->wds_dev[i];
            /* Bring the device down */
	    err = dev_close(wds_dev);
	    if (err)
	    {
	    	BUGMSG(DBG_WDS | DBG_ERROR,"DelWDSLink: dev_close() for WDS failed!\n");
	    	return err;
	    }
	    /* No need to lock cause the net_ioctl is already locked */
	    err = unregister_netdevice(wds_dev);
	    if (err != BER_NONE)
	    {
		BUGMSG(DBG_WDS | DBG_ERROR,"DelWDSLink: MAC: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x could not be unregistered!\n",
			req.data[0], req.data[1], req.data[2], req.data[3], req.data[4], req.data[5]);
		return err;
	    }
	    kfree(wds_dev->priv);
	    kfree(wds_dev);
	    lp->current_nof_links--;
	    lp->wds_dev[i] = NULL;

	    BUGMSG(DBG_WDS, "DelWDSLink: WDSLink removed succesfully\n");
	    return err;
	}
    }
    BUGMSG(DBG_WDS | DBG_ERROR,"DelWDSLink: MAC: %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x is not a WDS link\n",
	      req.data[0], req.data[1], req.data[2], req.data[3], req.data[4], req.data[5]);
    return -1;
}

/**********************************************************************
 *  wds_find_device
 *
 *  DESCRIPTION: Find the WDS device to use for receiving this frame
 *
 *  PARAMETERS:	 struct msg_frame_wds *framep - The frame received
 *  		 struct net_device    *dev    - The device where the
 *                                              frame was received.
 *
 *  RETURN:	 *net_dev of the device found, otherwise NULL.
 *
 **********************************************************************/
struct net_device *wds_find_device(struct msg_frame_wds *framep, struct net_device *dev)
{
    int i=0;
    struct net_local *lp;

    /* If no wds_link MAC address supplied */
    if (framep->wds_link[0] == 0 && framep->wds_link[1] == 0 &&
    	framep->wds_link[2] == 0 && framep->wds_link[3] == 0 &&
    	framep->wds_link[4] == 0 && framep->wds_link[5] == 0)
    {
	BUGMSG(DBG_WDS, "wds_find_device: framep->wds_link no MAC found, returning NULL\n");
	return NULL;
    }
    lp = (struct net_local *)dev->priv;
    for (i=0; i<MAX_WDS_LINKS; i++)
    {
	if (lp->wds_dev[i] == NULL)
	    continue;
	if(memcmp(((struct wds_net_local*)(lp->wds_dev[i]->priv))->wds_mac, framep->wds_link, 6) == 0)
	{
	    BUGMSG(DBG_WDS, "wds_find_device: framep->wds_link MAC found, returning dev->name: %s\n",
	    	   lp->wds_dev[i]->name);
	    return ((struct net_device*)lp->wds_dev[i]);
	}
    }
    BUGMSG(DBG_WDS, "wds_find_device: WDS_link from mac %2.2x-%2.2x-%2.2x-%2.2x-%2.2x-%2.2x not in current config, returning NULL\n",
    	   framep->wds_link[0],framep->wds_link[1],framep->wds_link[2],framep->wds_link[3],framep->wds_link[4],framep->wds_link[5]);
    return NULL;
}

/**********************************************************************
 *  wds_net_open
 *
 *  DESCRIPTION: Bring the WDS dev up and running.
 * 		 The WDS device needs to be added to the bridge,
 * 		 where the parent dev is also bridged.
 *
 *  PARAMETERS:	 struct net_device    *dev    - The WDS device
 *
 *  RETURN:	 zero in succes, otherwise error (<0).
 *
 **********************************************************************/
static int
wds_net_open(struct net_device *dev)
{
    int err = BER_NONE;
    struct wds_net_local   *lp = (struct wds_net_local *)dev->priv;
    struct net_bridge_port *br_port;
    struct net_bridge      *br;
    struct ifreq	    ifr;

    netif_start_queue(dev);

    /* Check if the parent device is bridged */
    if (lp->parent_dev->br_port == NULL)
    {
	BUGMSG(DBG_WDS | DBG_ERROR, "wds_net_open: Error parent device (%s) is not bridged, can't add WDS link.\n",lp->parent_dev->name);
        return -1;
    }

    /* Put the PROMISC flag on */
    ifr.ifr_flags = dev->flags | IFF_PROMISC;
    dev_ioctl(SIOCSIFFLAGS, &ifr);


    /* Find the bridge device for our parent */
    br_port = lp->parent_dev->br_port;
    br = (struct net_bridge*)br_port->br;

    /* Add the WDS device to the bridge */
    err = br_add_if(br, dev);

    /*
     * Set the bridge path cost to 90 (for wds decives)
     * Better than eth1, slower than eth0
     */
    br_stp_set_path_cost(dev->br_port, 100);
    return err;
}


/**********************************************************************
 *  wds_net_send_packet
 *
 *  DESCRIPTION:
 *
 *  PARAMETERS:	 struct sk_buff    *skb -
 *               struct net_device *dev -
 *
 *  RETURN:	 zero.
 *
 **********************************************************************/
static int
wds_net_send_packet(struct sk_buff *skb, struct net_device *dev)
{
    struct wds_net_local *lp = (struct wds_net_local *)dev->priv;

    if (lp == NULL)
    {
	printk("wds_net_send_packet: NULL pointer detected\n");
	return 0;
    }
    if ((lp->parent_dev->flags & IFF_UP) && (dev->flags & IFF_UP))
    {
        /* OK lets copy the WDSlink MACAddres in the skb->cb fields */
        memcpy(skb->cb, "WDSL", 4);
        memcpy(skb->cb+4, ((struct wds_net_local*)(dev->priv))->wds_mac, dev->addr_len);

        /* Set the "transmit" dev in this skb */
        skb->dev = lp->parent_dev;

        /* Move this skb to the "transmit" devices queue */
        if(dev_queue_xmit(skb) < BER_NONE)
        {
            printk("wds_net_send_packet: dev_queue_xmit failed\n");
        }
        /* Update stats for this device */
        lp->stats.tx_packets++;
        lp->stats.tx_bytes += skb->len;
    }
    return 0;
}

/**********************************************************************
 *  wds_net_close
 *
 *  DESCRIPTION:
 *
 *  PARAMETERS:	 struct net_device *dev -
 *
 *  RETURN:	 zero.
 *
 **********************************************************************/
static int
wds_net_close(struct net_device *dev)
{
    int err = BER_NONE;
    struct wds_net_local   *lp = (struct wds_net_local *)dev->priv;
    struct net_bridge_port *br_port;
    struct net_bridge      *br;

    netif_stop_queue(dev);

    /* Check if the parent device is bridged */
    if (lp->parent_dev->br_port == NULL)
    {
	BUGMSG(DBG_WDS | DBG_ERROR, "wds_net_close: Error parent device (%s) is not bridged, can't add WDS link.\n",lp->parent_dev->name);
        return -1;
    }

    /* Find the bridge device for our parent */
    br_port = lp->parent_dev->br_port;
    br = (struct net_bridge*)br_port->br;

    /* Del the WDS device to the bridge */
    err = br_del_if(br, dev);

    return err;
}

/**********************************************************************
 *  wds_net_ioctl
 *
 *  DESCRIPTION:
 *
 *  PARAMETERS:	 struct net_device *dev -
 *               struct ifreq      *ifr -
 *               int    cmd             -
 *
 *  RETURN:	 zero.
 *
 **********************************************************************/
static int wds_net_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    printk("wds_net_ioctl: Called don't know what to do\n");
    return 0;
}

/**********************************************************************
 *  wds_net_get_stats
 *
 *  DESCRIPTION:
 *
 *  PARAMETERS:	 struct net_device *dev -
 *
 *  RETURN:	 pointer to net_device_stats sruct.
 *
 **********************************************************************/
static struct net_device_stats *
wds_net_get_stats(struct net_device *dev)
{
    struct wds_net_local *lp = (struct wds_net_local *)dev->priv;
    struct net_device_stats *stats = &lp->stats;

    return stats;
}

#endif //WDSLINK_SUPPORT
