/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef __prismeth_h
#define __prismeth_h

/*
 *	ioctl's for the Prism Ethernet/802.11 driver
 *
 *	Typical use:
 *
 *	struct ifreq ifr;			// declare an interface request structure
 *	struct prism_ioctl ioc;
 *	char	macaddr[6];
 *  int     sock;
 *
 *	ioc.cmd = OPGET;
 *	ioc.oid = BLOB_OID_GEN_MACADDRESS;
 *	ioc.len = sizeof(macaddr);
 *	ioc.data = (caddr_t) macaddr;
 *
 *	strcpy(ifr.ifr_name, "eth0");                     // interface name
 *	ifr.ifr_data = (caddr_t) &ioc;
 *  sock = socket(AF_INET, SOCK_STREAM, 0);           // get a socket
 *  rc = ioctl(sock, PRISMIOCTL, &ifr);               // do ioctl on socket
 *
 *	printf("rc=%d mac=%x-%x-%x-%x-%x-%x\n", rc, macaddr[0],..., macaddr[5]);
 *
 */
#include <linux/sockios.h>

#define PRISMIOCTL      SIOCDEVPRIVATE

struct prism_ioctl {
	unsigned int cmd;		/* Command to run. See blob.h */
	unsigned int oid;		/* Prism Object ID. See bloboid.h */
	unsigned int len;		/* Length of the data buffer */
	unsigned char *data;	/* Pointer to the data buffer */
};

#define TX_NR_RETRIES	10

#define OPTRAPADD	10
#define OPTRAPDEL	11
#define OPRESET		12
#define OPGETMLME	13

#define PRISM_ETH_STOP  0
#define PRISM_ETH_HALT  1
#define PRISM_ETH_START 2
#define PRISM_ETH_RUN   3

#define WATCHDOG_NOT_TRIGGERED 0
#define WATCHDOG_TRIGGERED     1

#define DEFAULT_WATCHDOG_TIME  0

#endif /* !defined(__prismeth_h) */
