/*  $Header$
 *  
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef __ASM_ARCH_BLOBCALLS_H__
#define __ASM_ARCH_BLOBCALLS_H__

#define MVC1_3 /* Use MVC main. */

#include <linux/kernel.h>
#include <asm/page.h>		/* for BUG */
#include <blobv1.h>
#include <bloboid.h>
#include <blobstub.h>
#include <blobarch.h>

/* For the boot struct */
#include <bootdbg.h>

/*
 * This header file needs to be included for making calls to the MVC. It
 * will automatically include the correct MVC header files.
 *
 * The macros below specify the coprocessor interface of the MVC. There are
 * two types of calls: xx_to_mvc and xx_from_mvc. The xx can be _blobcall
 * for calls to the blobdevice or _devcall for calls to the regular devices
 * (WLAN, HOSTDS, DEBUG, etcetera).
 * There are different macros for functions with no arguments (..mvc0) and
 * one argument (..mvc1). There is also a special case for functions that
 * have no return value (.._noret).
 * The actual interface is defined in /arch/armnommu/mach-<machine>/blobcalls.c
 */

/* Some magic to let the preprocessor stringify the values of defines
 * and not the defines itself
 */
#define __s2(x) #x
#define __s1(x) __s2(x)

#define _blobcall_to_mvc0_noret(func, msg)						\
void func(void)										\
{											\
	__asm__ __volatile__ (								\
		"cdp	p11, " __s1(DEVICE_BLOB) ", c0, c0, c0, " __s1(msg)		\
		: :);									\
}

#define _blobcall_to_mvc0(type, func, msg)						\
type func(void)										\
{											\
	type __res;									\
	__asm__ __volatile__ (								\
		"mcr	p11, " __s1(DEVICE_BLOB) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
	return __res;									\
}

#define _blobcall_to_mvc1(type, func, msg, type1)					\
type func(type1 arg1)									\
{											\
	type __res;									\
	__asm__ __volatile__ (								\
		"mcr	p11, " __s1(DEVICE_BLOB) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
	return __res;									\
}

#define _blobcall_from_mvc0(type, func, msg)						\
type func(void)										\
{											\
	type __res;									\
	__asm__ __volatile__ (								\
		"mrc	p11, " __s1(DEVICE_BLOB) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
	return __res;									\
}

#define _blobcall_from_mvc1(type, func, msg, type1)					\
type func(type1 arg1)									\
{											\
	type __res;									\
	__asm__ __volatile__ (								\
		"mrc	p11, " __s1(DEVICE_BLOB) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
	return __res;									\
}

#define _devcall_to_mvc0_noret(func, msg)						\
void func(int device)									\
{											\
	switch(device) {								\
	case DEVICE_DEBUG:								\
		__asm__ __volatile__ (							\
		"cdp	p11, " __s1(DEVICE_DEBUG) ", c0, c0, c0, " __s1(msg)		\
		: :);									\
		break;									\
	case DEVICE_WLAN:								\
		__asm__ __volatile__ (							\
		"cdp	p11, " __s1(DEVICE_WLAN) ", c0, c0, c0, " __s1(msg)		\
		: :);									\
		break;									\
	case DEVICE_HOSTDS:								\
		__asm__ __volatile__ (							\
		"cdp	p11, " __s1(DEVICE_HOSTDS) ", c0, c0, c0, " __s1(msg)		\
		: :);									\
		break;									\
	case DEVICE_MISC:								\
		__asm__ __volatile__ (							\
		"cdp	p11, " __s1(DEVICE_MISC) ", c0, c0, c0, " __s1(msg)		\
		: :);									\
		break;									\
	default:									\
		BUG();									\
	}										\
}


#define _devcall_to_mvc0(type, func, msg)						\
type func(int device)									\
{											\
	type __res = (type)BER_UNSUP;							\
	switch(device) {								\
	case DEVICE_DEBUG:								\
		__asm__ __volatile__ (							\
		"mcr	p11, " __s1(DEVICE_DEBUG) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
		break;									\
	case DEVICE_WLAN:								\
		__asm__ __volatile__ (							\
		"mcr	p11, " __s1(DEVICE_WLAN) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
		break;									\
	case DEVICE_HOSTDS:								\
		__asm__ __volatile__ (							\
		"mcr	p11, " __s1(DEVICE_HOSTDS) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
		break;									\
	case DEVICE_MISC:								\
		__asm__ __volatile__ (							\
		"mcr	p11, " __s1(DEVICE_MISC) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
		break;									\
	default:									\
		BUG();									\
	}										\
	return __res;									\
}

#define _devcall_to_mvc1(type, func, msg, type1)					\
type func(int device, type1 arg1)							\
{											\
	type __res = (type)BER_UNSUP;							\
	switch(device) {								\
	case DEVICE_DEBUG:								\
		__asm__ __volatile__ (							\
		"mcr	p11, " __s1(DEVICE_DEBUG) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
		break;									\
	case DEVICE_WLAN:								\
		__asm__ __volatile__ (							\
		"mcr	p11, " __s1(DEVICE_WLAN) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
		break;									\
	case DEVICE_HOSTDS:								\
		__asm__ __volatile__ (							\
		"mcr	p11, " __s1(DEVICE_HOSTDS) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
		break;									\
	case DEVICE_MISC:								\
		__asm__ __volatile__ (							\
		"mcr	p11, " __s1(DEVICE_MISC) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
		break;									\
	default:									\
		BUG();									\
	}										\
	return __res;									\
}

#define _devcall_from_mvc0(type, func, msg)						\
type func(int device)									\
{											\
	type __res = (type)BER_UNSUP;							\
	switch(device) {								\
	case DEVICE_DEBUG:								\
		__asm__ __volatile__ (							\
		"mrc	p11, " __s1(DEVICE_DEBUG) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
		break;									\
	case DEVICE_WLAN:								\
		__asm__ __volatile__ (							\
		"mrc	p11, " __s1(DEVICE_WLAN) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
		break;									\
	case DEVICE_HOSTDS:								\
		__asm__ __volatile__ (							\
		"mrc	p11, " __s1(DEVICE_HOSTDS) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
		break;									\
	case DEVICE_MISC:								\
		__asm__ __volatile__ (							\
		"mrc	p11, " __s1(DEVICE_MISC) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res):);							\
		break;									\
	default:									\
		BUG();									\
	}										\
	return __res;									\
}

#define _devcall_from_mvc1(type, func, msg, type1)					\
type func(int device, type1 arg1)							\
{											\
	type __res = (type)BER_UNSUP;							\
	switch(device) {								\
	case DEVICE_DEBUG:								\
		__asm__ __volatile__ (							\
		"mrc	p11, " __s1(DEVICE_DEBUG) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
		break;									\
	case DEVICE_WLAN:								\
		__asm__ __volatile__ (							\
		"mrc	p11, " __s1(DEVICE_WLAN) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
		break;									\
	case DEVICE_HOSTDS:								\
		__asm__ __volatile__ (							\
		"mrc	p11, " __s1(DEVICE_HOSTDS) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
		break;									\
	case DEVICE_MISC:								\
		__asm__ __volatile__ (							\
		"mrc	p11, " __s1(DEVICE_MISC) ", %0, c0, c0, " __s1(msg)		\
		: "=r" (__res): "0" (arg1));						\
		break;									\
	default:									\
		BUG();									\
	}										\
	return __res;									\
}

#endif /* __ASM_ARCH_BLOBCALLS_H__ */
