/*
 * uclinux/include/asm-armnommu/arch-poldhu/serial.h
 */
/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef __ASM_ARCH_SERIAL_H
#define __ASM_ARCH_SERIAL_H
#include <linux/serial_reg.h>
#include <asm/arch/irqs.h>

/*
 * UART definitions
 */

/* - DR */
#define	UARTDR				(0x0)
#define	UARTDRMask			(0xff)

/* - RSRECR */
#define	UARTRSRECR			(0x4)
#define	UARTRSRECROE			(0x8)
#define	UARTRSRECRBE			(0x4)
#define	UARTRSRECRPE			(0x2)
#define	UARTRSRECRFE			(0x1)
#define	UARTRSRECRMask			(0xf)

/* - LCR_H */
#define	UARTLCR_H			(0x8)
#define	UARTLCR_HWLEN			(0x60)
#define	UARTLCR_HWLENShift		(0x5)
#define	UARTLCR_HSTP2			(0x8)
#define	UARTLCR_HEPS			(0x4)
#define	UARTLCR_HPEN			(0x2)
#define	UARTLCR_HBRK			(0x1)
#define	UARTLCR_HMask			(0x6f)

/* - LCR_M */
#define	UARTLCR_M			(0xc)
#define	UARTLCR_MMask			(0xff)

/* - LCR_L */
#define	UARTLCR_L			(0x10)
#define	UARTLCR_LMask			(0xff)

/* - CR */
#define	UARTCR				(0x14)
#define	UARTCRRTIE			(0x40)
#define	UARTCRTIE			(0x20)
#define	UARTCRRIE			(0x10)
#define	UARTCRMSIE			(0x8)
#define	UARTCRUARTEN			(0x1)
#define	UARTCRMask			(0x79)

/* - FR */
#define	UARTFR				(0x18)
#define	UARTFRTXE			(0x80)
#define	UARTFRRXF			(0x40)
#define	UARTFRTXF			(0x20)
#define	UARTFRRXE			(0x10)
#define	UARTFRBUSY			(0x8)
#define	UARTFRDCD			(0x4)
#define	UARTFRDSR			(0x2)
#define	UARTFRCTS			(0x1)
#define	UARTFRMask			(0xff)

/* - IIR */
#define	UARTIIR				(0x1c)
#define	UARTIIRRTIS			(0x8)
#define	UARTIIRTIS			(0x4)
#define	UARTIIRRIS			(0x2)
#define	UARTIIRMIS			(0x1)
#define	UARTIIRMask			(0xf)

/* - ICR */
#define	UARTICR				(0x20)
#define	UARTICRMIS			(0x1)
#define	UARTICRMask			(0x1)

/* Instance values */
#define	uUARTDR				(0xc0000a00)
#define	uUARTRSRECR			(0xc0000a04)
#define	uUARTLCR_H			(0xc0000a08)
#define	uUARTLCR_M			(0xc0000a0c)
#define	uUARTLCR_L			(0xc0000a10)
#define	uUARTCR				(0xc0000a14)
#define	uUARTFR				(0xc0000a18)
#define	uUARTIIR			(0xc0000a1c)
#define	uUARTICR			(0xc0000a20)

#ifndef __ASSEMBLER__

/* C struct view */
typedef struct s_UART {
 unsigned DR; /* @0 */
 unsigned RSRECR; /* @4 */
 unsigned LCR_H; /* @8 */
 unsigned LCR_M; /* @12 */
 unsigned LCR_L; /* @16 */
 unsigned CR; /* @20 */
 unsigned FR; /* @24 */
 unsigned IIR; /* @28 */
 unsigned ICR; /* @32 */
} s_UART;

#define uUART ((volatile struct s_UART *) 0xc0000a00)

/*
 * Definitions used in drivers/char/serial.c
 */
#define STD_COM_FLAGS (ASYNC_BOOT_AUTOCONF | ASYNC_SKIP_TEST)

#define BASE_BAUD ( 1843200 / 16 )

#define RS_TABLE_SIZE  1
#define STD_SERIAL_PORT_DEFNS \
	{  \
	magic: 0, \
	baud_base: BASE_BAUD, \
	port: (unsigned long)uUART, \
	irq: IRQ_UART, \
	flags: STD_COM_FLAGS \
	}       /* ttyS0 */
#define EXTRA_SERIAL_PORT_DEFNS


static inline void uart_initialize(void)
{
	/* enable UART. All interrupts disabled */
	uUART->CR = UARTCRUARTEN;

	/* Clear the interrupt registers. */
	uUART->ICR = UARTICRMask;

	/* empty the receive register & clear receive error flags. */
	(void) uUART->DR;
	uUART->RSRECR = 0;
}

static inline unsigned char uart_rx(void)
{
	return (unsigned char)uUART->DR;
}

static inline void uart_tx(unsigned char ch)
{
	uUART->DR = ch;
}

#define UART_IRQSTAT_RX(status)		(status & (UARTIIRRIS | UARTIIRRTIS | UARTIIRMIS))
#define UART_IRQSTAT_TX(status)		(status & UARTIIRTIS)
#define UART_IRQSTAT_ANY(status)	(status & UARTIIRMask)

static inline unsigned int uart_read_irqstat(void)
{
	return uUART->IIR;
}

static inline int uart_read_lsr(void)
{
	unsigned char uart_status;
	int status = 0;

	uart_status = uUART->RSRECR;
	if (uart_status & UARTRSRECROE)
		status |= UART_LSR_OE;
	if (uart_status & UARTRSRECRBE)
		status |= UART_LSR_BI;
	if (uart_status & UARTRSRECRPE)
		status |= UART_LSR_PE;
	if (uart_status & UARTRSRECRFE)
		status |= UART_LSR_FE;

	uart_status = uUART->FR;
	if (uart_status & UARTFRTXE) {
		status |= UART_LSR_TEMT;
		status |= UART_LSR_THRE;
	}
	if (uart_status & UARTFRRXF)
		status |= UART_LSR_DR;

	return status;
}

static inline void uart_clear_errors(void)
{
	uUART->RSRECR = 0;
}

static inline unsigned int uart_save_ier(void)
{
	return(uUART->CR & 0x78);
}

static inline void uart_restore_ier(unsigned int ier)
{
	uUART->CR |= (ier & 0x78);
}

static inline void uart_interrupt_enable(int interrupt)
{
	if(interrupt == UART_IER_THRI)		/* Transmitter holding register int. */
		uUART->CR |= UARTCRTIE;
	else if(interrupt == UART_IER_RDI)	/* Enable receiver data interrupt */
		uUART->CR |= UARTCRRIE;
	else
		printk("uart_interrupt_enable: Unknown Interrupt %d", interrupt);
}

static inline void uart_interrupt_disable(int interrupt)
{
	if(interrupt == UART_IER_THRI)		/* Transmitter holding register int. */
		uUART->CR &= ~UARTCRTIE;
	else if(interrupt == UART_IER_RDI)	/* Enable receiver data interrupt */
		uUART->CR &= ~UARTCRRIE;
	else
		printk("uart_interrupt_disable: Unknown Interrupt %d", interrupt);
}

#endif /* __ASSEMBLER__ */

#endif /* !defined(__uart_h) */
