#ifndef __LINUX_BRIDGE_EBT_MACNAT_H
#define __LINUX_BRIDGE_EBT_MACNAT_H

struct ebt_macnat_info
{
	char brifname[32];
	// EBT_ACCEPT, EBT_DROP, EBT_CONTINUE or EBT_RETURN
	int target;
	int alg_bootp_method;
        unsigned char gwmac[6];
};

#ifdef __KERNEL__
struct ethernet_arpbdy{
    unsigned char arp_sha[6];
    unsigned int  arp_sip;
    unsigned char arp_tha[6];
    unsigned int  arp_tip;
}__attribute__((packed));


struct bootp_hdr{
    unsigned char	op;
    unsigned char	htype;
    unsigned char	hlen;
    unsigned char	hops;
    unsigned int	xid;
    unsigned short	secs;
    unsigned short	flags;
    unsigned int	ciaddr;
    unsigned int	yiaddr;
    unsigned int	siaddr;
    unsigned int	giaddr;
    unsigned char	chaddr[16];
    char		sname[64];
    char		file[128];
    char		vend[64];
}  __attribute__((packed));


#define EBT_MACSNAT_TARGET "macsnat"
#define EBT_MACDNAT_TARGET "macdnat"

#define EBT_ALG_BOOTP_BROADCAST	1
#define EBT_ALG_BOOTP_RELAY	2


static unsigned int jiffies_saved = 0;
static unsigned int brip = 0;
static unsigned int brmask = 0;
#define BRIP_REFRESH_HZ	60*HZ

static inline unsigned int ebt_get_myipmask(const char *ifname, unsigned int* myip, unsigned int *mask)
{
    struct ifreq ifr;
    struct in_device *in_dev;
    struct net_device *dev;
    struct in_ifaddr **ifap=NULL, *ifa=NULL;
    int ret=0;
    char *colon;

    //strcpy(ifr.ifr_name, ifname);
    //dev_probe_lock();
    //rtnl_lock();
    if(((jiffies - jiffies_saved) > BRIP_REFRESH_HZ) || brip == 0 ){
        /*
         ret = devinet_ioctl(SIOCGIFADDR, (void*)&ifr);
         brip = ((struct sockaddr_in*)&ifr.ifr_addr)->sin_addr.s_addr;
         */
	dev_probe_lock();
        colon = strchr(ifname, ':');
        if(colon)
            *colon = 0;
        if((dev=__dev_get_by_name(ifname))==NULL){
            ret = -ENODEV;
            goto RETURN;
        }
        if(colon)
            *colon = ':';
        if((in_dev=__in_dev_get(dev))!=NULL){
            for(ifap=&in_dev->ifa_list; (ifa=*ifap)!=NULL; ifap=&ifa->ifa_next){
                if(strcmp(ifname, ifa->ifa_label)==0)
                    break;
            }
            if(ifa==NULL){
                ret = -EADDRNOTAVAIL;
                goto RETURN;
            }
            brip = ifa->ifa_local;
            brmask = ifa->ifa_mask;
        }
        jiffies_saved = jiffies;
RETURN:
    //    rtnl_unlock();
    	dev_probe_unlock();
    }
    if(ret == 0){
    	if(myip) *myip = brip;
    	if(mask) *mask = brmask;
        return brip;
    }
    else{
        brip = 0;
        brmask = 0;
#if 0
        printk("%s_%s(): cannot get my IP(%d)!!!\n",
               __FILE__, __PRETTY_FUNCTION__, ret);
#endif
        return (unsigned int)-1;
    }
}
#endif

#endif
