/*
 * ebt_macsnat:
 * 	MAC NAT snat module
 * */
#include <linux/netfilter_bridge/ebtables.h>
#include <linux/module.h>
#include <net/sock.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/string.h>
#include <net/checksum.h>
#include <net/arp.h>
#include <linux/netdevice.h>
#include <linux/netfilter_bridge/ebt_macnat.h>

static unsigned char eth0MAC[6];
//#define DEBUG

unsigned char* ipmac_search_entry(unsigned int ipaddr);
unsigned char* ipmac_gwmac(void);

static inline unsigned char* ebt_alg_arpdnat(struct sk_buff *pskb, unsigned myip){
 unsigned char *dmacaddr=NULL;
 struct arphdr *arph = pskb->nh.arph;
 struct ethernet_arpbdy *arpbdy = (struct ethernet_arpbdy*)&(arph[1]);
 struct arpreq arpr;
 int err;
#ifdef DEBUG
 printk("%s_%s(): translate ARP packet T:%u.%u.%u.%u S:%u.%u.%u.%u \n",
        __FILE__, __PRETTY_FUNCTION__,
        NIPQUAD(arpbdy->arp_tip), NIPQUAD(arpbdy->arp_sip));
#endif
 	if(ntohs(arph->ar_hrd) == ARPHRD_ETHER && /* ethernet ARP */ 
		ntohs(arph->ar_pro) == ETH_P_IP){ /* and IP protocol */
		if(arpbdy->arp_tip == myip){
#ifdef DEBUG
			printk("%s_%s(): ARP's target is myself, return my eth0MAC: %02x:%02x:%02x:%02x:%02x:%02x\n",
			__FILE__, __PRETTY_FUNCTION__,
			eth0MAC[0],eth0MAC[1],eth0MAC[2],eth0MAC[3],eth0MAC[4],eth0MAC[5]);
#endif
			return eth0MAC;
		}

#if 0
		if(arpbdy->arp_sip == myip){ // somehow, this packet is from me
#ifdef DEBUG
			printk("%s_%s(): My own ARP, but sent back to me?!\n", __FILE__,
				__PRETTY_FUNCTION__);
#endif
			arpbdy->arp_sip = 0;
			memset(arpbdy->arp_sha, 0, 6);
			return (unsigned char *)0xFFFFFFFF;
		}
#endif

		dmacaddr = ipmac_search_entry(arpbdy->arp_tip);
		if(dmacaddr){
			memcpy(arpbdy->arp_tha, dmacaddr, ETH_ALEN); // translate !
#ifdef DEBUG
			printk("%s_%s(): ARP dnat: %u.%u.%u.%u ==> ", __FILE__, __PRETTY_FUNCTION__,
				NIPQUAD(arpbdy->arp_tip));
                        printk("%02x:%02x:%02x:%02x:%02x:%02x\n",
                               dmacaddr[0], dmacaddr[1], dmacaddr[2],
                               dmacaddr[3], dmacaddr[4], dmacaddr[5]);
#endif
		}
	}
	
	return dmacaddr;
} 

static void ebt_alg_bootprelay(struct sk_buff *pskb, const char *brif, int bootp_method){
 struct ethhdr *ethernet = pskb->mac.ethernet;
 struct iphdr *iph = pskb->nh.iph;
 struct udphdr *uh = (struct udphdr*)&iph[1];
 struct bootp_hdr *bootphdr = (struct bootp_hdr*)&uh[1];
 
 unsigned int myip = ebt_get_myipmask(brif, NULL, NULL);
#ifdef DEBUG
 if(myip == 0xFFFFFFFF)
     printk("%s_%s(): cannot get my IP nad netamsk\n", __FILE__, __PRETTY_FUNCTION__);
#endif
#ifdef DEBUG
 printk("%s_%s(): translate bootp packet: CIP: %u.%u.%u.%u CHW: %02x:%02x:%02x:%02x:%02x:%02x\n",
        __FILE__, __PRETTY_FUNCTION__,
        NIPQUAD(bootphdr->yiaddr),
        bootphdr->chaddr[0], bootphdr->chaddr[1], bootphdr->chaddr[2],
        bootphdr->chaddr[3], bootphdr->chaddr[4], bootphdr->chaddr[5]);
#endif

 // change MAC layer's destination to bootp client
	if( (ntohs(bootphdr->flags)&0x8000)){
		// use broadcast MAC addr
		memset(ethernet->h_dest, 0xFF, ETH_ALEN);
		// use broadcast IP
		iph->daddr = 0xFFFFFFFF;
#ifdef DEBUG
		printk("%s_%s(): use broadcast\n", __FILE__, __PRETTY_FUNCTION__);
#endif
	}
	
	if((myip!=0 || myip!=0xFFFFFFFF) && (bootphdr->giaddr==myip)){
		// OK, this packet is relayed by me
		if(!(ntohs(bootphdr->flags)&0x8000)){ // not broadcast packet
		 	memcpy(ethernet->h_dest, bootphdr->chaddr, ETH_ALEN);
			// change IP destination IP
	 		iph->daddr = bootphdr->yiaddr;
#ifdef DEBUG
                        printk("%s_%s(): use unicast to %02x:%02x:%02x:%02x:%02x:%02x\n",
                               __FILE__, __PRETTY_FUNCTION__,
                               bootphdr->chaddr[0],bootphdr->chaddr[1],bootphdr->chaddr[2],
                               bootphdr->chaddr[3],bootphdr->chaddr[4],bootphdr->chaddr[5]);
#endif
		}
		// remove relay IP
		bootphdr->giaddr = 0;
		bootphdr->hops --;
		//change udp dest port to 68
		uh->dest = htons(68);
		uh->check = 0;
	}		

	// calculate IP checksum
	iph->check = 0;
	iph->check = ip_compute_csum((void *)iph, sizeof(struct iphdr));
}

static int ebt_target_macdnat(struct sk_buff **pskb, unsigned int hooknr,
   const struct net_device *in, const struct net_device *out,
   const void *data, unsigned int datalen)
{
 unsigned char *dmacaddr=NULL;
 struct ebt_macnat_info *info = (struct ebt_macnat_info *)data;
 unsigned char broadcast_macaddr[ETH_ALEN] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};
 struct udphdr *uh;
 unsigned char mcst_header[4];
#if 1
 	memcpy(mcst_header, ((**pskb).mac.ethernet)->h_dest, 4);
 	//printk("%s: mcst_header: %02x:%02x:%02x:%02x\n", __PRETTY_FUNCTION__,
 	//	mcst_header[0],mcst_header[1],mcst_header[2],mcst_header[3]);
	if( (mcst_header[0] & 0x01) /* && (mcst_header[1]== 0x00) &&
		(mcst_header[2] == 0x5E) && ((mcst_header[3]&0x80)==0x00)*/ ){
		/* This is multicast packet */
#ifdef DEBUG
            printk("%s_%s(): encounter multicast packet, let it through\n", __FILE__, __PRETTY_FUNCTION__);
#endif
		return info->target;
	}
#endif
// 	printk("%s(): \n", __PRETTY_FUNCTION__);
		
	switch(ntohs(((**pskb).mac.ethernet)->h_proto)){
	 case ETH_P_ARP: // ARP
	 	dmacaddr = ebt_alg_arpdnat(*pskb, ebt_get_myipmask(info->brifname, NULL, NULL));
	 	if(dmacaddr == (unsigned char *)0xFFFFFFFF)
	 		return EBT_DROP ;
		break;
	 case ETH_P_IP: // IPv4
	 	// only handle IPv4 packet
		uh = (struct udphdr *)&((**pskb).nh.iph[1]);
             
		if((((**pskb).nh.iph)->protocol==0x11) && (ntohs(uh->dest)==67)){
			/* UDP and bootps(this bridge acts as a bootp server for client) */
			ebt_alg_bootprelay(*pskb, info->brifname, info->alg_bootp_method);
		}
		else if(memcmp(((**pskb).mac.ethernet)->h_dest, 
				broadcast_macaddr, ETH_ALEN) !=0){
#if 0
			if(((**pskb).nh.iph)->saddr == ebt_get_myipmask(info->brifname, NULL, NULL)){
#ifdef DEBUG
				printk("%s_%s(): My packet sent back to me?!\n", __FILE__, 
					__PRETTY_FUNCTION__);
#endif
				((**pskb).nh.iph)->saddr = 0;
				memset(((**pskb).mac.ethernet)->h_source, 0,
				   		ETH_ALEN * sizeof(unsigned char));
				return EBT_DROP;
			}
			else 
#endif
			if(((**pskb).nh.iph)->daddr != ebt_get_myipmask(info->brifname, NULL, NULL)){
			 	dmacaddr = ipmac_search_entry(((**pskb).nh.iph)->daddr);
			 
				if(dmacaddr){
#if 0
#ifdef DEBUG
					printk("%s_%s(): %u.%u.%u.%u ==> ", __FILE__, __PRETTY_FUNCTION__,
						NIPQUAD(((**pskb).nh.iph)->daddr));
                    printk("%02x:%02x:%02x:%02x:%02x:%02x\n",
                                           dmacaddr[0],dmacaddr[1],dmacaddr[2],
                                           dmacaddr[3],dmacaddr[4],dmacaddr[5]);
#endif
#endif
				}
				else{
#ifdef DEBUG
                                    printk("%s_%s(): no entry for %u.%u.%u.%u, send to preset router\n", 
						__FILE__, __PRETTY_FUNCTION__,
						NIPQUAD(((**pskb).nh.iph)->daddr));
#endif
					dmacaddr = ipmac_gwmac();
#ifdef DEBUG
					printk("%s_%s(): I guess router is: %02x:%02x:%02x:%02x:%02x:%02x\n",
						__FILE__, __PRETTY_FUNCTION__,
            dmacaddr[0], dmacaddr[1], dmacaddr[2], dmacaddr[3], dmacaddr[4], dmacaddr[5]);
#endif
                                        if(dmacaddr[0] == 0xFF){
                                            // gateway not been set
#ifdef DEBUG
                                            printk("%s_%s(): gateway address has not been set, drop packet\n",
                                                   __FILE__, __PRETTY_FUNCTION__);
#endif
                                            return EBT_DROP;
                                        }
                                        
				}
			}

			else{
#ifdef DEBUG
				printk("%s_%s(): Received a packet to myself\n", __FILE__, __PRETTY_FUNCTION__);
#endif 
				dmacaddr = eth0MAC;
			}
		}
		break;
	 default:
	 	break;
	}
	
	if(dmacaddr){ // translate MAC layer's destination address
		memcpy(((**pskb).mac.ethernet)->h_dest, dmacaddr,
	   		ETH_ALEN * sizeof(unsigned char));
	}
	return info->target;
}

static int ebt_target_macdnat_check(const char *tablename, unsigned int hookmask,
   const struct ebt_entry *e, void *data, unsigned int datalen)
{
	struct ebt_macnat_info *info = (struct ebt_macnat_info *)data;
	struct net_device *dev;
	if (BASE_CHAIN && info->target == EBT_RETURN)
		return -EINVAL;
	CLEAR_BASE_CHAIN_BIT;
	if ( (strcmp(tablename, "nat") ||
	   (hookmask & ~((1 << NF_BR_PRE_ROUTING) | (1 << NF_BR_LOCAL_OUT)))) &&
	   (strcmp(tablename, "broute") || hookmask & ~(1 << NF_BR_BROUTING)) )
		return -EINVAL;
	if (datalen != sizeof(struct ebt_macnat_info))
		return -EINVAL;
	if (INVALID_TARGET)
		return -EINVAL;
	dev = __dev_get_by_name("eth0");
	if(dev==NULL){
		printk("%s_%s(): cannot find eth0 device\n", __FILE__, __PRETTY_FUNCTION__);
	}
	else
		memcpy(eth0MAC, dev->dev_addr, 6);
#ifdef DEBUG
	printk("%s_%s(): eth0MAC: %02x:%02x:%02x:%02x:%02x:%02x\n",
		__FILE__, __PRETTY_FUNCTION__,
		eth0MAC[0],eth0MAC[1],eth0MAC[2],eth0MAC[3],eth0MAC[4],eth0MAC[5]);
#endif	
	return 0;
}

static struct ebt_target macdnat =
{
	{NULL, NULL}, EBT_MACDNAT_TARGET, ebt_target_macdnat, ebt_target_macdnat_check,
	NULL, THIS_MODULE
};

static int __init init(void)
{
	return ebt_register_target(&macdnat);
}

static void __exit fini(void)
{
	ebt_unregister_target(&macdnat);
}

module_init(init);
module_exit(fini);
EXPORT_NO_SYMBOLS;

