/*
 * ebt_macsnat:
 * 	MAC NAT snat module
 * */
#include <linux/netfilter_bridge/ebtables.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/in.h>
#include <linux/if.h>
#include <linux/inetdevice.h>
#include <linux/string.h>
#include <linux/netfilter_bridge/ebt_macnat.h>

int macnat_get_cloned_macaddr(char *mac); // in ebt_natwatcher.c

//#define DEBUG

static int ebt_alg_arpsnat(struct sk_buff *pskb, const unsigned char *smac, 
								unsigned int myip)
{
 struct arphdr *arph = pskb->nh.arph;
 struct ethernet_arpbdy *arpbdy = (struct ethernet_arpbdy*)&(arph[1]);

#ifdef DEBUG
 printk("%s_%s(): translate ARP packet T:%u.%u.%u.%u S:%u.%u.%u.%u \n",
        __FILE__, __PRETTY_FUNCTION__,
        NIPQUAD(arpbdy->arp_tip), NIPQUAD(arpbdy->arp_sip));
#endif
 
 	if(arpbdy->arp_tip == myip){
#ifdef DEBUG
		printk("%s_%s(): encounter my ip: %u.%u.%u.%u.\n",
         __FILE__, __PRETTY_FUNCTION__,
	NIPQUAD(arpbdy->arp_tip));
#endif
		return 1;
	}
 	if(ntohs(arph->ar_hrd) == ARPHRD_ETHER && /* ethernet ARP */ 
			ntohs(arph->ar_pro) == ETH_P_IP){ /* and IP protocol */
		memcpy(arpbdy->arp_sha, smac, ETH_ALEN); // translate !
	}
	return 0;
}

static void ebt_alg_bootprelay(struct bootp_hdr *bootphdr, unsigned int myip,
	unsigned char* macaddr, int method)
{
 int cloneState;
 char clonedMAC[6];
#ifdef DEBUG
    printk("%s_%s(): translate bootp packet: CIP: %u.%u.%u.%u CHW: %02x:%02x:%02x:%02x:%02x:%02x\n",
           __FILE__, __PRETTY_FUNCTION__,
           NIPQUAD(bootphdr->ciaddr),
           bootphdr->chaddr[0], bootphdr->chaddr[1], bootphdr->chaddr[2],
           bootphdr->chaddr[3], bootphdr->chaddr[4], bootphdr->chaddr[5]);
#endif
    if(myip == 0 || myip == 0xFFFFFFFF){
        // my IP may not be right, change to use broadcast protocol to compensate
        method = 1;
#ifdef DEBUG
        printk("%s_%s():myip is not correct, change to use broadcast way...\n",
               __FILE__, __PRETTY_FUNCTION__);
#endif
    }
	switch(method){
	 case 1: // broadcast
	 	bootphdr->flags |= htons(0x8000);
#ifdef DEBUG
             printk("%s_%s(): Use broadcast protocol\n", __FILE__, __PRETTY_FUNCTION__);
#endif
	 	break;
	 case 2: // bootp relay
		if(bootphdr->giaddr==0){
			if(memcmp(macaddr, bootphdr->chaddr, bootphdr->hlen)){
				bootphdr->hops ++;
				bootphdr->giaddr = myip;
			}
			else{
				printk("%s_%s(): bootp request from myself\n",
					__FILE__, __PRETTY_FUNCTION__);
				// check MAC clone
				cloneState = macnat_get_cloned_macaddr(clonedMAC);
				if(cloneState){
					// now, my IP is not determined and cannot use cloned MAC
					//  try broadcast protocol
					bootphdr->flags |= htons(0x8000);
				}
			}
		}
#ifdef DEBUG
             printk("%s_%s(): Use bootp relay protocol\n", __FILE__, __PRETTY_FUNCTION__);
#endif
		break;
	}
}


static int ebt_target_macsnat(struct sk_buff **pskb, unsigned int hooknr,
   const struct net_device *in, const struct net_device *out,
   const void *data, unsigned int datalen)
{
 struct ebt_macnat_info *info = (struct ebt_macnat_info *) data;
 unsigned char mcst_header[4];
 unsigned int myip = ebt_get_myipmask(info->brifname, NULL, NULL);
 char smac[6];
 int useMACClone=0;

	useMACClone = macnat_get_cloned_macaddr(smac);
	
#ifdef DEBUG
 	if(myip == 0xFFFFFFFF)
    	printk("%s_%s(): cannot get my IP and netamsk\n", __FILE__, __PRETTY_FUNCTION__);
#endif
	if(out == NULL){
		return EBT_DROP;
	}
#if 0
	memcpy(mcst_header, ((**pskb).mac.ethernet)->h_dest, 4);
	if((mcst_header[0] & 0x01) && (mcst_header[1]== 0x00) &&
		(mcst_header[2] == 0x5E) && ((mcst_header[3]&0x80)==0x00)){
		// This is multicast packet
		return info->target;
	}
#endif
	/* Now we do some application layer gateway function */
	switch(ntohs(((**pskb).mac.ethernet)->h_proto)){
	 case ETH_P_ARP: // ARP

	 	if(ebt_alg_arpsnat(*pskb, (useMACClone==3||useMACClone==1)?smac:out->dev_addr, myip)==1){
			return EBT_DROP;
		}
	 	break;
	 case ETH_P_IP: // IPv4
#if 0
	 	if(((**pskb).nh.iph)->daddr == myip){
#ifdef DEBUG
			printk("%s_%s(): encounter my IP on IP layer\n", __FILE__, __PRETTY_FUNCTION__);
#endif
			return EBT_DROP;
	 	}
#endif
	 	if(((**pskb).nh.iph)->protocol == 0x11){ /*UDP*/
		 struct udphdr *uh = (struct udphdr *)&((**pskb).nh.iph[1]);
		 if(ntohs(uh->dest) == 67 /*bootps*/){
		 	//printk("%s(): receive packet to bootps\n", __PRETTY_FUNCTION__);
			/* do ebt_alg_bootprelay */
			ebt_alg_bootprelay((struct bootp_hdr*)&uh[1], myip, out->dev_addr,
				info->alg_bootp_method);
			uh->check = 0; /* to save some time, just disable checksum */
		 }
		}
	 	break;
	 default:
	 	break;
	}
	memcpy(((**pskb).mac.ethernet)->h_source, 
		(useMACClone==3||useMACClone==1)?smac:out->dev_addr, ETH_ALEN);
	return info->target;
}

static int ebt_target_macsnat_check(const char *tablename, unsigned int hookmask,
   const struct ebt_entry *e, void *data, unsigned int datalen)
{
	struct ebt_macnat_info *info = (struct ebt_macnat_info *) data;

	if (datalen != sizeof(struct ebt_macnat_info))
		return -EINVAL;
	if (BASE_CHAIN && info->target == EBT_RETURN)
		return -EINVAL;
	CLEAR_BASE_CHAIN_BIT;
	if (strcmp(tablename, "nat"))
		return -EINVAL;
	if (hookmask & ~(1 << NF_BR_POST_ROUTING))
		return -EINVAL;
	if (INVALID_TARGET)
		return -EINVAL;
	return 0;
}

static struct ebt_target macsnat =
{
	{NULL, NULL}, EBT_MACSNAT_TARGET, ebt_target_macsnat, ebt_target_macsnat_check,
	NULL, THIS_MODULE
};

static int __init init(void)
{
	return ebt_register_target(&macsnat);
}

static void __exit fini(void)
{
	ebt_unregister_target(&macsnat);
}

module_init(init);
module_exit(fini);
EXPORT_NO_SYMBOLS;

