/*
 * ebt_natwatcher:
 * 	MAC NAT watcher module
 * */
#include <linux/version.h>
#include <linux/config.h>
#include <linux/netfilter_bridge/ebtables.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/if_arp.h>
#include <linux/if_ether.h>
#include <linux/spinlock.h>
#include <linux/timer.h>
#include <linux/in.h>
#include <linux/if.h>
#include <linux/inetdevice.h>
#include <linux/netfilter_bridge/ebt_macnat.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/slab.h>
#include <linux/kmod.h>
#include <asm/uaccess.h>

//#define DEBUG


#define EBT_NATWATCHER	"natwatcher"
static spinlock_t ebt_natwatcher_lock = SPIN_LOCK_UNLOCKED;

static unsigned char ipmac_gw_mac[ETH_ALEN]={
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
struct ipmac_entry{
	unsigned char macaddr[ETH_ALEN];
	unsigned int ipaddr;
	unsigned short age;
};

#define MAX_IPMAC_ENTRIES	256
#define TIMER_DURATION		(5*100)
#define MAX_IPMAC_AGE		(60*5/5)

static struct ipmac_entry ipmac_table[MAX_IPMAC_ENTRIES];

int ipmac_add_entry(unsigned char *macaddr, unsigned int ipaddr)
{
 unsigned char hash_idx=0, search_idx;
 int i;

 	// Calculate hash index
	for(i=0 ; i<4 ; i++){
		hash_idx += (unsigned char)( (ipaddr>>(i*8))&0xFF );
	}
	//printk("%s(): hash_idx: %u, 0x%08x\n", __PRETTY_FUNCTION__, hash_idx, ipaddr);

	// Check if this host is already in table
	if(((memcmp(ipmac_table[hash_idx].macaddr, macaddr, 6)==0) &&
		(ipmac_table[hash_idx].ipaddr==ipaddr)) ||	// host in table
		ipmac_table[hash_idx].ipaddr==0 || 	// entry is empty
		ipmac_table[hash_idx].age>MAX_IPMAC_AGE){ // entry is too old
		memcpy(ipmac_table[hash_idx].macaddr, macaddr, 6);
		ipmac_table[hash_idx].ipaddr = ipaddr;
		ipmac_table[hash_idx].age = 0;
		return hash_idx;
	}
	else{ // OK, this hash index is not empty
#ifdef DEBUG
		printk("%s(): table[%u] occupied, searching...", __PRETTY_FUNCTION__,
			hash_idx);
		printk("By %u.%u.%u.%u(%02x:%02x:%02x:%02x:%02x:%02x)\n",
			NIPQUAD(ipmac_table[hash_idx].ipaddr),
			ipmac_table[hash_idx].macaddr[0],
			ipmac_table[hash_idx].macaddr[1],
			ipmac_table[hash_idx].macaddr[2],
			ipmac_table[hash_idx].macaddr[3],
			ipmac_table[hash_idx].macaddr[4],
			ipmac_table[hash_idx].macaddr[5]);
#endif
		search_idx = (hash_idx+1)%MAX_IPMAC_ENTRIES;
		while(search_idx != hash_idx){
			if(((memcmp(ipmac_table[search_idx].macaddr, macaddr, 6)==0) &&
				(ipmac_table[search_idx].ipaddr==ipaddr)) ||	// host in table
				ipmac_table[search_idx].ipaddr==0 || 	// entry is empty
				ipmac_table[search_idx].age>MAX_IPMAC_AGE){ // entry is too old
				memcpy(ipmac_table[search_idx].macaddr, macaddr, 6);
				ipmac_table[search_idx].ipaddr = ipaddr;
				ipmac_table[search_idx].age = 0;
#ifdef DEBUG
				printk("done, use table[%u]\n", search_idx);
#endif
				return search_idx;
			}
			search_idx ++;
			search_idx %= MAX_IPMAC_ENTRIES;
		}
	}

	return -1; // cannot add this host into IPMAC table, table fulfilled
}


unsigned char* ipmac_search_entry(unsigned int ipaddr){
 unsigned char hash_idx = 0, search_idx;
 unsigned char found = 0;
 int i;
 	// Calculate hash index
	for(i=0 ; i<4 ; i++){
		hash_idx += (unsigned char)( (ipaddr>>(i*8))&0xFF );
	}

	if(ipmac_table[hash_idx].ipaddr == ipaddr){
		ipmac_table[hash_idx].age = 0;
		return ipmac_table[hash_idx].macaddr;
	}

	// else, try next entry
	search_idx = (hash_idx +1)%MAX_IPMAC_ENTRIES;
	while(search_idx != hash_idx){
		if(ipmac_table[search_idx].ipaddr == ipaddr){
			ipmac_table[search_idx].age = 0;
			return ipmac_table[search_idx].macaddr;
			break;
		}
		search_idx ++;
		search_idx %= MAX_IPMAC_ENTRIES;
	}
	// still cannot found //,return router mac address
#ifdef DEBUG
	printk("%s: entry %u.%u.%u.%u not found, return router MAC address: %x:%x:%x:%x:%x:%x\n", __PRETTY_FUNCTION__, NIPQUAD(ipaddr),
	ipmac_gw_mac[0], ipmac_gw_mac[1], ipmac_gw_mac[2],
	ipmac_gw_mac[3], ipmac_gw_mac[4], ipmac_gw_mac[5]);
#endif
	return NULL;

}

unsigned char *ipmac_gwmac(void)
{
	return ipmac_gw_mac;
}

static struct timer_list ipmac_timer;
void ipmac_timer_func(unsigned long ptr)
{
 struct ipmac_entry *table = (struct ipmac_entry *)ptr;
 int i;

 	if(spin_trylock(&ebt_natwatcher_lock)){
		for(i=0 ; i<MAX_IPMAC_ENTRIES ; i++){
			if(table[i].ipaddr != 0){
#ifdef DEBUG1
			 int j;
			 unsigned char *p;
				printk("%s(): aging table[%u]: ", __PRETTY_FUNCTION__, i);
				p = table[i].macaddr;
				for (j = 0; j < ETH_ALEN; j++,p++)
					printk("%02x%c", *p, j == ETH_ALEN - 1 ? ' ':':');
				printk(" %u.%u.%u.%u age: %u\n", NIPQUAD(table[i].ipaddr),
					table[i].age);
#endif
				
				if((++table[i].age)> MAX_IPMAC_AGE){
					memset(&table[i], 0, sizeof(struct ipmac_entry));
				//	printk("too old, deleted\n");
				}
				
				++table[i].age; // only aging this entry, not remove it
			}
		}
 		spin_unlock(&ebt_natwatcher_lock);
	}
	//ipmac_timer.expires = jiffies + HZ;
	mod_timer(&ipmac_timer, jiffies + TIMER_DURATION); // modify timer for next time
}


/* Stardate 10309.28 Max Yin
	Here goes the PROC_FS stuff
*/
static int useMACClone=0; // 0->off, 1->static clone, 2->auto clone, 3->latched MAC clone
static char clonedMAC[ETH_ALEN]={0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};
static spinlock_t macclone_lock = SPIN_LOCK_UNLOCKED;

#ifdef CONFIG_PROC_FS
static ssize_t macnat_proc_enableMACClone_read(struct file *file , char *buf, size_t count, loff_t *ppos);
static ssize_t macnat_proc_enableMACClone_write(struct file *file, const char *buf, size_t count, loff_t *ppos);
static ssize_t macnat_proc_clonedMAC_read(struct file *file , char *buf, size_t count, loff_t *ppos);
static ssize_t macnat_proc_clonedMAC_write(struct file *file, const char *buf, size_t count, loff_t *ppos);


static struct proc_dir_entry *proc_macnat;
static struct file_operations enableMACClone_fops={
	read: macnat_proc_enableMACClone_read,
	write: macnat_proc_enableMACClone_write
};

static struct file_operations cloned_MAC_fops={
	read: macnat_proc_clonedMAC_read,
	write: macnat_proc_clonedMAC_write
};

static ssize_t macnat_proc_enableMACClone_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	char out[64];
	int len;
	
	if(ppos[0]) // not read from start
		return 0;
	spin_lock(&macclone_lock);
	snprintf(out,60, "%d\n", useMACClone);
	spin_unlock(&macclone_unlock);

	len = strlen(out)+1;
	if(copy_to_user(buf, out, len)){
		return -EFAULT;
	}
	file->f_pos = len;
	return len;
}
static ssize_t macnat_proc_enableMACClone_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	char in[64];
	int arg;

	if(copy_from_user(in, buf, (count>60)?60:count)){
		return -EFAULT;
	}
	arg = simple_strtol(in, NULL, 10);
//	printk("%s(): got %d\n", __PRETTY_FUNCTION__, arg);
	
	if(arg>2 || arg<0){
		return count; // ignore it
	}
	spin_lock(&macclone_lock);
	useMACClone = arg;
	spin_unlock(&macclone_lock);
	return count;
}

static ssize_t macnat_proc_clonedMAC_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	char out[128];
	int len;
	
	if(ppos[0]) // not read from start
		return 0;
	spin_lock(&macclone_lock);
	snprintf(out, 120, "%02x:%02x:%02x:%02x:%02x:%02x", 
		clonedMAC[0], clonedMAC[1], clonedMAC[2], 
		clonedMAC[3], clonedMAC[4], clonedMAC[5]);
	len = strlen(out)+1;
	if(copy_to_user(buf, out, len)){
		return -EFAULT;
	}
	file->f_pos = len;
	spin_unlock(&macclone_lock);
	return len;

}
static ssize_t macnat_proc_clonedMAC_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
 char in[128];
 unsigned int arg;
 int i;
 
 	if(useMACClone >= 2)
 		return count; // Auto-clone, cannot assign new MAC address
 	spin_lock(&macclone_lock);
 	if(copy_from_user(in, buf, (count>120)?120:count)){
 		spin_unlock(&macclone_lock);
 		return -EFAULT;
 	}
 	for(i=0 ; i<ETH_ALEN ; i++){
 		in[3*i+2]='\0';
 		arg = simple_strtoul(in+3*i, NULL, 16);
// 		printk("%s(): 0x%02x\n", __PRETTY_FUNCTION__, arg);
 		clonedMAC[i] = arg;
 	}
 	spin_unlock(&macclone_lock);
 	return count;
}


static ssize_t ipmac_proc_read(struct file *file , char *buf, size_t count, loff_t *ppos);

static struct file_operations ipmac_table_fops={
	read: ipmac_proc_read
};

/* For reference
#define MAX_IPMAC_ENTRIES	256
#define TIMER_DURATION		(5*HZ)
#define MAX_IPMAC_AGE		60/TIMER_DURATION
struct ipmac_entry{
	unsigned char macaddr[ETH_ALEN];
	unsigned int ipaddr;
	unsigned short age;
};
static struct ipmac_entry ipmac_table[MAX_IPMAC_ENTRIES];
*/
static int ipmac_proc_read_index=0;
static ssize_t ipmac_proc_read(struct file *file , char *buf, size_t count, loff_t *ppos)
{
 int len, ret_len;
 char out[256];

 	if(ppos[0]==0){
	 	snprintf(out, 255, "MAC NAT IP/MAC table\n");
	 	len = strlen(out);
	 	snprintf(out+len, 255-len, "index	IP address	MAC address		age\n");
	 	len = strlen(out);
	 	/*
	 	snprintf(out+len, 255-len, "-------------------------------------------\n");
	 	len = strlen(out);
	 	*/
	 	ppos[0] = len;
	 	ret_len = (count>len)?len+1:count;
	 	if(copy_to_user(buf, out, ret_len)){
	 		return -EFAULT;
	 	}
	 	ipmac_proc_read_index=0;
	 	return ret_len;
	}
	else if(ipmac_proc_read_index<MAX_IPMAC_ENTRIES){
		snprintf(out, 255,         "[%d]	%3u.%3u.%3u.%3u	%02x:%02x:%02x:%02x:%02x:%02x	%d\n",
			ipmac_proc_read_index, NIPQUAD(ipmac_table[ipmac_proc_read_index].ipaddr),
			ipmac_table[ipmac_proc_read_index].macaddr[0],
			ipmac_table[ipmac_proc_read_index].macaddr[1],
			ipmac_table[ipmac_proc_read_index].macaddr[2],
			ipmac_table[ipmac_proc_read_index].macaddr[3],
			ipmac_table[ipmac_proc_read_index].macaddr[4],
			ipmac_table[ipmac_proc_read_index].macaddr[5],
			ipmac_table[ipmac_proc_read_index].age);
		len = strlen(out);
		ret_len = (count>len)?len+1:count;
		ppos[0] += len;
		if(copy_to_user(buf, out, ret_len)){
	 		return -EFAULT;
	 	}
	 	ipmac_proc_read_index ++; 
	 	return ret_len;
	}
	else{
		return 0;
	}
}

static int macnat_proc_init(void)
{
 struct proc_dir_entry *p;
 	// Create /proc/net/macnat directory
 	proc_macnat = proc_mkdir("macnat", proc_net);
 	if(!proc_macnat){
 		return -ENOMEM;
 	}
 	
 	p = create_proc_entry("enable_MAC_clone", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH, proc_macnat);
 	p->proc_fops = & enableMACClone_fops;
 	p = create_proc_entry("cloned_MAC", S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH, proc_macnat);
 	p->proc_fops = & cloned_MAC_fops;
 	p = create_proc_entry("ipmac_table", S_IRUSR|S_IRGRP|S_IROTH, proc_macnat);
 	p->proc_fops = & ipmac_table_fops;
 	return 0;
}

static void macnat_proc_cleanup()
{
	remove_proc_entry("enable_MAC_clone", proc_macnat);
	remove_proc_entry("cloned_MAC", proc_macnat);
	remove_proc_entry("ipmac_table", proc_macnat);
	remove_proc_entry("macnat", proc_net);
}


#endif

int macnat_get_cloned_macaddr(char *mac)
{
int enabledMACClone;
	spin_lock(&macclone_lock);
	memcpy(mac, clonedMAC, ETH_ALEN);
	enabledMACClone = useMACClone;
	spin_unlock(&macclone_lock);
	return enabledMACClone;
}

static inline void ebt_natwatcher_clone_macaddr(const char *haddr)
{
	spin_lock(&macclone_lock);
	memcpy(clonedMAC, haddr, ETH_ALEN);
	useMACClone = 3;
	spin_unlock(&macclone_lock);
#if 0
{ // The following code is WiSoC MVC specific code
	char *argv[6]={"/bin/setoid", "eth1", "0", "long", "00:00:00:00:00:00", NULL};
	char *envp[3]={"HOME=/", "PATH=/sbin:/bin", NULL};
	sprintf(argv[4], "%02x:%02x:%02x:%02x:%02x:%02x", haddr[0], haddr[1], haddr[2],
		haddr[3], haddr[4], haddr[5]);
	printk("%s(): clone MAC address to MVC\n", __FILE__);
	call_usermodehelper(argv[0], argv, envp);
}
#endif
}

static int ebt_natwatcher_check(const char *tablename, unsigned int hookmask,
   const struct ebt_entry *e, void *data, unsigned int datalen)
{
    struct ebt_macnat_info *info = (struct ebt_macnat_info*)data;
    memcpy(ipmac_gw_mac, info->gwmac, 6);
	return 0;
}

static void ebt_natwatcher(const struct sk_buff *skb,
	const struct net_device *in, const struct net_device *out,
	const void *data, unsigned int datalen)
{
	//spin_lock_bh(&ebt_natwatcher_lock);
	spin_lock(&ebt_natwatcher_lock);

#ifdef DEBUG
	if (skb->dev->hard_header_len) {
		int i;
		unsigned char *p = (skb->mac.ethernet)->h_source;
		printk("MAC source = ");
		for (i = 0; i < ETH_ALEN; i++,p++)
			printk("%02x%c", *p, i == ETH_ALEN - 1 ? ' ':':');
		printk("MAC dest = ");
		p = (skb->mac.ethernet)->h_dest;
		for (i = 0; i < ETH_ALEN; i++,p++)
			printk("%02x%c", *p, i == ETH_ALEN - 1 ? ' ':':');
	}
#endif

#ifdef DEBUG
	printk("proto = 0x%04x\r\n", ntohs(((*skb).mac.ethernet)->h_proto));
#endif

	if(ntohs((skb->mac.ethernet)->h_proto) == 0x0800)//IP protocol
	{int idx;
	 unsigned char *p;
	 int i;
         struct iphdr *iph = skb->nh.iph;
#ifdef DEBUG
		printk(" IP SRC=%u.%u.%u.%u IP DST=%u.%u.%u.%u\r\n",
		   NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));
		printk(" IP tos=0x%02X, IP proto=%d\n", iph->tos, iph->protocol);
#endif
		if((idx=ipmac_add_entry((skb->mac.ethernet)->h_source,
			iph->saddr)) == -1){
			printk("%s(): cannot add host into table\n", __PRETTY_FUNCTION__);
		}
#ifdef DEBUG
		else{
			printk("table[%u].macaddr = ", idx);
			p = ipmac_table[idx].macaddr;
			for (i = 0; i < ETH_ALEN; i++,p++)
				printk("%02x%c", *p, i == ETH_ALEN - 1 ? ' ':':');
			printk("\ntable[%u].ipaddr = %u.%u.%u.%u\n", idx, 
				NIPQUAD(ipmac_table[idx].ipaddr));
			printk("table[%u].age = %u\n", idx, ipmac_table[idx].age);
			
		}
#endif
		if(useMACClone==2){
			ebt_natwatcher_clone_macaddr((skb->mac.ethernet)->h_source);
		}
	}/* IP protocol */
	else if ( /* Deal with ARP protocol */
	    ((skb->mac.ethernet->h_proto == __constant_htons(ETH_P_ARP)) ||
	    (skb->mac.ethernet->h_proto == __constant_htons(ETH_P_RARP)))) {
		struct arphdr * arph = skb->nh.arph;
		struct ethernet_arpbdy *arpbdy = (struct ethernet_arpbdy*)&arph[1];
		unsigned char *p;
		int i, idx;
#ifdef DEBUG
		printk(" ARP HTYPE=%d, PTYPE=0x%04x, OPCODE=%d \n",
		   ntohs(arph->ar_hrd), ntohs(arph->ar_pro),
		   ntohs(arph->ar_op));
		p = arpbdy->arp_sha;
		printk("MAC source = ");
		for (i = 0; i < ETH_ALEN; i++,p++)
			printk("%02x%c", *p, i == ETH_ALEN -1 ? ' ':':');
		p = arpbdy->arp_tha;
		printk("MAC dest = ");
		for (i = 0; i < ETH_ALEN; i++,p++)
			printk("%02x%c", *p, i == ETH_ALEN -1 ? ' ':':');
		printk(" IP SRC=%u.%u.%u.%u IP DST=%u.%u.%u.%u\n",
		   NIPQUAD(arpbdy->arp_sip), NIPQUAD(arpbdy->arp_tip));
#endif
		if((idx = ipmac_add_entry(arpbdy->arp_sha, 
			arpbdy->arp_sip)) == -1){
			printk("%s(): cannot add host into table\n", __PRETTY_FUNCTION__);
		}
#ifdef DEBUG
		else{
			printk("table[%u].macaddr = ", idx);
			p = ipmac_table[idx].macaddr;
			for (i = 0; i < ETH_ALEN; i++,p++)
				printk("%02x%c", *p, i == ETH_ALEN - 1 ? ' ':':');
			printk("\ntable[%u].ipaddr = %u.%u.%u.%u\n", idx, 
				NIPQUAD(ipmac_table[idx].ipaddr));
			printk("table[%u].age = %u\n", idx, ipmac_table[idx].age);
			
		}
#endif
		if(useMACClone==2){
			ebt_natwatcher_clone_macaddr(arpbdy->arp_sha);
		}

	}
	//spin_unlock_bh(&ebt_natwatcher_lock);
	spin_unlock(&ebt_natwatcher_lock);
}

static void ebt_natwatcher_destory(void *watcherdata, unsigned int datalen)
{
	memset(ipmac_table, 0,  sizeof(ipmac_table));
}

struct ebt_watcher natwatcher =
{
	{NULL, NULL}, EBT_NATWATCHER,
	ebt_natwatcher,
	ebt_natwatcher_check,
	ebt_natwatcher_destory,
	THIS_MODULE
};

static int __init init(void)
{
	memset(ipmac_table, 0,  sizeof(ipmac_table));
	init_timer(&ipmac_timer);
	ipmac_timer.expires = jiffies + HZ; // 1 second timeout
	ipmac_timer.function = ipmac_timer_func;
	ipmac_timer.data = (long)ipmac_table;
	add_timer(&ipmac_timer);
#ifdef CONFIG_PROC_FS
	macnat_proc_init();
#endif
	return ebt_register_watcher(&natwatcher);
}

static void __exit fini(void)
{
	ebt_unregister_watcher(&natwatcher);
	del_timer_sync(&ipmac_timer);
#ifdef CONFIG_PROC_FS
	macnat_proc_cleanup();
#endif
}

EXPORT_SYMBOL(ipmac_add_entry);
EXPORT_SYMBOL(ipmac_search_entry);
EXPORT_SYMBOL(ipmac_gwmac);
EXPORT_SYMBOL(macnat_get_cloned_macaddr);
module_init(init);
module_exit(fini);



