/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef __HARDCONFIG_H__
#define __HARDCONFIG_H__

#include "dispatch.h"

/* Generic HC Unknown */
#define HC_UNKNOWN		-1

/* ProductType */
#define HC_550_PCMCIA		0
#define HC_550_WBU		1
#define HC_550_AP		2
#define HC_550_SERIES		7

#define HC_1100_PCMCIA		8
#define HC_1100_WBU		9
#define HC_1100_AP		10
#define HC_1100_WBS		11
#define HC_1100_WBC		12
#define HC_1100_SERIES		15

#define HC_1150_PCMCIA		16
#define HC_1150_USB		17
#define HC_1150_PCI		18
#define HC_1150_WBU		19

#define HC_1150_AP_10BT		24	// 10 baseT
#define HC_1150_AP_100BT	25	// 10/100 baseT
#define HC_1150_AP_SB		25	// Small Business
#define HC_1150_AP_HOME		26	// Home
#define HC_1150_AP_ENT		27	// Enterprise

#define HC_1150_WB		32	// Generic Wireless Bridge

/* OEMID */
#define HC_OEM_UNKNOWN		HC_UNKNOWN
#define HC_OEM_WHITELABEL	4

/* Capabilities */
#define HC_CAP_WEP		0x00000001
#define HC_CAP_PCF		0x00000002
#define HC_CAP_AIRLOCK	0x00000004
#define HC_CAP_RC5		0x00000008
#define HC_CAP_PWSP		0x00000010
#define HC_CAP_ACL		0x00000020

#define HC_CALFLAG_MAXPOWER	0x01
#define MAX_SYSOBJID_LEN 88

struct HardConfigCalibrateData_struct
{
	u_char	Flags;			// MAX_POWER = 0x01
	char	LogPower;		// dBm       -16 - +30
	u_short	LinPower;		// 0.02 mW   1 - 50000

	u_char	Padd[4];

	u_char	FCCDetect[14];
	char	FCCControl[14];		// -128 - +126
	u_char	MKKDetect[14];
	char	MKKControl[14];		// -128 - +126
};

struct InterfaceDefinition_struct
{
	unsigned short role;
	unsigned short interface_id;
	unsigned short variant;
	unsigned short bottom_compatibility;
	unsigned short top_compatibility;
} __attribute__ ((packed));


struct HardConfig_struct
{
	u_long	Version;
	u_char	MACAddress[8];
	u_long	DefaultRegDomain;
	u_long	ProductType;
	u_char	OEMName[32];
	u_char	ProductName[32];
	u_long	OEMID;

	u_char	PrivateKey[32];
	u_char	PublicKey[128];
	u_char	PublicKey1[128];
	u_char	PublicKey2[128];
	u_char	PublicKey3[128];

	u_long	Reserved6[2];

	// Updateable firmware part

	u_long	Capabilities;
	u_long	RegDomains[8];
	u_long  BurnInSwitch;
	/* The sysboject ID,
	 * 1.3.6.1.4.1.<SysObjId>
	 */
	u_char	SysObjId[MAX_SYSOBJID_LEN];
	struct	InterfaceDefinition_struct InterfaceDefinitions[3];
	u_char	Reserved11[98];

	// Shared firmware / production part

	u_long	HardwareRev;
	u_long	OutputPower;		// In micro Watt

	u_long	CalibrateRSSIA;		// -70 dBm
	u_long	CalibrateRSSIB;		// -90 dBm
	u_long	CalibrateRows;		// Valid calibration rows
	struct HardConfigCalibrateData_struct Calibrate[8];

	// Production block

	u_char	Production[2040];
	u_char	BurnInData[8];
};



#define HARDCONFIG_DEFAULT( l, def ) ( (l==(u_long)-1)?def:l )

#define HardConfig ((struct HardConfig_struct *)(DISPATCH[2]))

#define HARDCONFIG_START (DISPATCH[2])
#define HARDCONFIG_END   (DISPATCH[3])

#define PRODUCTION_START (DISPATCH[4])
#define PRODUCTION_END   (DISPATCH[5])


int cache_hardconfig_parameters( char *mtddevice );

int mtd_read_ulong_array( int fd, int offset, u_long *value, int len );
int GetHCBootrecords(long *hc_boot_record);
unsigned long GetHcVersion(void);
unsigned char* GetHcMACAddress(void);
unsigned long GetHcDefaultRegDomain(void);
unsigned long GetHcProductType (void);
unsigned char* GetHcOEMName(void);
unsigned char* GetHcProductName(void);
unsigned long GetHcOEMID(void);

unsigned long GetHcCapabilities(void);
unsigned long* GetHcRegDomains(void);
unsigned char* GetHcSysObjId(void);

unsigned long GetHcHardwareRev(void);
unsigned long GetHcOutputPower(void);

unsigned long GetHcCalibrateRSSIA(void);
unsigned long GetHcCalibrateRSSIB(void);
unsigned long GetHcCalibrateRows(void);
unsigned long GetHcCalibrateRows_ptr(void);

struct HardConfig_struct * GetHcStruct(void);

#endif /* __HARDCONFIG_H__ */

