/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#ifndef __PDA_H__
#define __PDA_H__

#define PDA_SIZE        8192

/*---------------------------------------------------------------*
 |                        PDR CODE DEFINES                       |
 *---------------------------------------------------------------*/

#define PDR_END                                 0x0000
#define PDR_MF_PART_NUMBER                      0x0001
#define PDR_VERSION                             0x0002
#define PDR_NIC_SERIAL                          0x0003
#define PDR_NIC_RAM_SIZE                        0x0005
#define PDR_RFMODEM_SUP_RANGE                   0x0006
#define PDR_PRISM_MAC_SUP_RANGE                 0x0007
#define PDR_NIC_ID                              0x0008
#define PDR_MAC_ADDRESS                         0x0101
#define PDR_REGULATORY_DOMAIN_LIST              0x0103 /* obsolete */
#define PDR_ALLOWED_CHAN_SET                    0x0104
#define PDR_DEFAULT_CHAN                        0x0105
#define PDR_TEMPERATURE_TYPE                    0x0107
#define PDR_IFR_SETTING                         0x0200
#define PDR_RFR_SETTING                         0x0201
#define PDR_3861_BASELINE_REG_SETTINGS          0x0202
#define PDR_3861_SHADOW_REG_SETTINGS            0x0203
#define PDR_3861_IFRF_REG_SETTINGS              0x0204
#define PDR_3861_CHAN_CALIB_SET_POINTS          0x0300
#define PDR_3861_CHAN_CALIB_INTEGRATOR          0x0301
#define PDR_3842_PRISM_II_NIC_CONFIG            0x0400
#define PDR_PRISM_USB_ID                        0x0401
#define PDR_PRISM_PCI_ID                        0x0402
#define PDR_PRISM_PCI_IF_CONFIG                 0x0403
#define PDR_PRISM_PCI_PM_CONFIG                 0x0404
#define PDR_3861_MF_TEST_CHAN_SET_POINTS        0x0900
#define PDR_3861_MF_TEST_CHAN_INTEGRATORS       0x0901
#define PDR_COUNTRY_INFO                        0x1000 /* obsolete */
#define PDR_INTERFACE_LIST                      0x1001
#define PDR_HW_PLATFORM_COMPONENT_ID            0x1002
#define PDR_OEM_NAME                            0x1003
#define PDR_PRODUCT_NAME                        0x1004
#define PDR_UTF8_OEM_NAME                       0x1005
#define PDR_UTF8_PRODUCT_NAME                   0x1006
#define PDR_COUNTRY_LIST                        0x1007
#define PDR_DEFAULT_COUNTRY                     0x1008
#define PDR_ANTENNA_GAIN                        0x1100
#define PDR_PRISM_2_PA_CALIBRATION_DATA         0x1900
#define PDR_PRISM_INDIGO_PA_CALIBR_DATA         0x1901
#define PDR_RSSI_LINEAR_APPROXIMATION           0x1902
#define PDR_PRISM_PA_CAL_OUTPUT_POWER_LIMITS    0x1903
#define PDR_PRISM_PA_CAL_CURVE_DATA		0x1904
#define PDR_RSSI_LINEAR_APPROXIMATION_DB        0x1905
#define PDR_ZIF_TX_IQ_CALIBR_DATA		0x1906


/* Specific HW variant info. The should probably be moved to hardware_variants.h */
#define BR_IF_VAR_ISL39000_SYNTH_MASK		0x0007 
#define BR_IF_VAR_ISL39000_SYNTH_1		0x0001
#define BR_IF_VAR_ISL39000_SYNTH_2		0x0002
#define BR_IF_VAR_ISL39000_SYNTH_3		0x0003

#define BR_IF_VAR_ISL39000_IQ_CAL_MASK		0x0018
#define BR_IF_VAR_ISL39000_IQ_CAL_PA_DETECTOR   0x0000
#define BR_IF_VAR_ISL39000_IQ_CAL_DISABLED      0x0008
#define BR_IF_VAR_ISL39000_IQ_CAL_ZIF           0x0010

#define BR_IF_VAR_ISL39000_FAA_SWITCH_MASK	0x0020
#define BR_IF_VAR_ISL39000_FAA_SWITCH_ENABLED   0x0001

#define BR_IF_VAR_ISL39000_24_GHZ_MASK	        0x0040
#define BR_IF_VAR_ISL39000_24_GHZ_DISABLED      0x0040

#define BR_IF_VAR_ISL39000_5_GHZ_MASK	        0x0080
#define BR_IF_VAR_ISL39000_5_GHZ_DISABLED       0x0080

#define BR_IF_VAR_ISL39000_RX_DIV_MASK	        0x0100
#define BR_IF_VAR_ISL39000_RX_DIV_SUPPORTED     0x0100

#define BR_IF_VAR_ISL39000_TX_DIV_MASK	        0x0200
#define BR_IF_VAR_ISL39000_TX_DIV_SUPPORTED     0x0200

#define BR_IF_VAR_ISL39000_ASM_MASK		0x0400
#define BR_IF_VAR_ISL39000_ASM_XSWON		0x0400











/*-------------------------------------------------------*
 |                 Global Definitions                    |
 *-------------------------------------------------------*/
struct Pdr_Struct {
   unsigned short   PdrLength;
   unsigned short   PdrCode;
   unsigned char    PdrData[2];
} __attribute__ ((packed));


/*-------------------------------------------------------*
 |                 Function Prototypes                   |
 *-------------------------------------------------------*/

struct Pdr_Struct               * pda_Find_Pdr       (unsigned short pdrCode, struct Pdr_Struct *Start);
                           void   pda_Print          (void);
                  unsigned char * pdaGetMacAddress   (void);
struct s_bootloader_boot_record * pdaGetBootRecords  (void);
                           void * pda_Get_PdaPtr     (void);

#define PDR_START ((struct Pdr_Struct *)0)

int pda_Cache (char *mtddevice);

#endif /* __PDA_H__ */

