/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdlib.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <syslog.h>
#include <string.h>

#include "inl_debug.h"

#if DEBUG > DBL_NONE
  static void do_percentm (char *obuf, char *ibuf);

  static char mbuf [1024];
  static char fbuf [1024];
#endif /* DEBUG > 0 */

/* Log an error message and exit */
void error_and_die (char *fmt, ...)
{
#if DEBUG >= DBL_ERROR
    va_list list;
    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_ERR, mbuf);
    syslog (LOG_CRIT, "exiting.");
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_ERROR */
    exit(1);
}

/* Log an error message */
void error (char *fmt, ...)
{
#if DEBUG >= DBL_ERROR
    va_list list;
    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_ERR, mbuf);
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_ERROR */
}

/* Log an warning message */
void warn (char *fmt, ...)
{
#if DEBUG >= DBL_WARNING
    va_list list;
    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_WARNING, mbuf);
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_WARNING */
}

/* Log an notice message */
void note (char *fmt, ...)
{
#if DEBUG >= DBL_NOTICE
    va_list list;
    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_NOTICE, mbuf);
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_NOTICE */
}

/* Log an debug message */
void debug (char *fmt, ...)
{
#if DEBUG >= DBL_DEBUG
    va_list list;

    do_percentm (fbuf, fmt);

    va_start (list, fmt);
    vsnprintf (mbuf, sizeof mbuf, fbuf, list);
    va_end (list);

#ifdef DBL_LOCAL
    fprintf( stderr, "%s\n", mbuf );
#endif /* DBL_LOCAL */

#ifdef DBL_SYSLOG
    syslog (LOG_DEBUG, mbuf);
#endif /* DBL_SYSLOG */
#endif /* DEBUG >= DBL_DEBUG */
}

/* Find %m in the input string and substitute an error message string. */
#if DEBUG > DBL_NONE
static void do_percentm (char *obuf, char *ibuf)
{
    char *s = ibuf;
    char *p = obuf;
    int infmt = 0;
    char *m;

    while (*s)
    {
	if (infmt)
	{
	    if (*s == 'm')
	    {
		m = strerror (errno);
		if (!m)
		    m = "<unknown error>";
		strcpy (p - 1, m);
		p += strlen (p);
		++s;
	    }
	    else
		*p++ = *s++;
	    infmt = 0;
	}
	else
	{
	    if (*s == '%')
		infmt = 1;
	    *p++ = *s++;
	}
    }
    *p = 0;
}
#endif /* DEBUG > 0 */


