/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#if 0   //was #ifndef PDA

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>

#include "mtd.h"
#include "inl_debug.h"
#include "sockutil.h"
#include "hardconfig.h"
#include "identification/bootrec.h"
#include "identification/interfaces.h"
#include "identification/hardware_variants.h"
#include "identification/components.h"

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

struct s_bootloader_boot_record {
	struct s_bootrec_comp_id  component_id;
	struct s_bootrec_comp_ver component_version;
	struct s_bootrec_hdr hdr;
	struct s_bootrec_exp_if exposed_if[3];
	struct s_bootrec_end bootrecord_end;
} bootloader_boot_record =
{
	component_id:
	{
		{ BR_CODE_COMPONENT_ID, 1 },
		BR_COMP_ID_BOOT
	},
	component_version:
	{
		{ BR_CODE_COMPONENT_VERSION, VER_STRING_SIZE },
		/* Filled in later */
	},
	hdr:
	{ BR_CODE_EXPOSED_IF, 8 },
	/* Exposed interfaces will be copied from the Hardconfig */
	bootrecord_end:
	{
		{ BR_CODE_END_OF_BRA, 5 },	/* bootrec code, length */
		0, 0,				/* CRC-16, padding */
		{ 0, 0, 0, 0 }			/* MD5 */
	}
};

static struct HardConfig_struct hc;
unsigned long CalibrateRows_ptr = 0;

/*****************************************************************************
 * NAME:
 *   cache_hardconfig_parameters
 *
 * DESCRIPTION:
 * 	read the hardconfig parameters to be checked and cache them
 *
 * PARATETERS:
 *  char *	 			- IN - Pointer to MDT device to use.
 *
 * RETURN
 * success = 0
 * error = 1
 *****************************************************************************/
int
cache_hardconfig_parameters( char *mtddevice )
{
	int fd = -1, rc = 0, offset, i;
	u_long hcloc;				/* location of hardconfig */
	static char hardconfig_cached = 0;
	if ( hardconfig_cached )
		return 0;

	do
	{
		if ((fd = open( mtddevice, O_RDONLY )) < 0) {
			warn( "open %s failed: %m", mtddevice );
			rc = 1;
			break;
		}
		
		/* retrieve the bootloader version string located at offset 0x3E0 */
		offset = BOOTLOADER_VERSION_OFFSET;
		if ( (rc = mtd_read_uchar_array( fd, offset, (char *)bootloader_boot_record.component_version.ver_string, VER_STRING_SIZE * sizeof(long) )) == 1 )
			break;

		/* retrieve the location of hardconfig in flash */
		offset = DISPATCH_TABLE + DISPATCH_HC_START * sizeof(unsigned char *);
		if ( (rc = mtd_read_ulong( fd, offset, &hcloc )) == 1 )
			break;

		offset = offsetof(struct HardConfig_struct, Version);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.Version )) == 1 )
			break;

		offset = offsetof(struct HardConfig_struct, MACAddress);
		if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, hc.MACAddress, 8 )) == 1 )
			break;

		offset = offsetof(struct HardConfig_struct, DefaultRegDomain);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.DefaultRegDomain )) == 1 )
			break;

		offset = offsetof(struct HardConfig_struct, ProductType);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.ProductType )) == 1 )
			break;

		offset = offsetof(struct HardConfig_struct, OEMName);
		if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, hc.OEMName, 32 )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, ProductName);
		if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, hc.ProductName, 32 )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, OEMID);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.OEMID )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, Capabilities);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.Capabilities )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, RegDomains);
		if ( (rc = mtd_read_ulong_array( fd, offset + hcloc, hc.RegDomains, 8 )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, SysObjId);
		if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, hc.SysObjId, MAX_SYSOBJID_LEN )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, InterfaceDefinitions);
		if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, (char *)&hc.InterfaceDefinitions[0], sizeof(hc.InterfaceDefinitions))) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, HardwareRev);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.HardwareRev )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, OutputPower);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.OutputPower )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, CalibrateRSSIA);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.CalibrateRSSIA )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, CalibrateRSSIB);
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.CalibrateRSSIB )) == 1 )
			break;
		offset = offsetof(struct HardConfig_struct, CalibrateRows);
		CalibrateRows_ptr = offset + hcloc;
		if ( (rc = mtd_read_ulong( fd, offset + hcloc, &hc.CalibrateRows )) == 1 )
			break;
		if ( hc.CalibrateRows > 0 && hc.CalibrateRows <= 8 )
		{
			for ( i=0; i<hc.CalibrateRows; i++ )
			{
				offset = offsetof(struct HardConfig_struct, Calibrate[i].Flags);
				if ( (rc = mtd_read_uchar( fd, offset + hcloc, (u_long*)&hc.Calibrate[i].Flags )) == 1 )
					break;

				offset = offsetof(struct HardConfig_struct, Calibrate[i].LogPower);
				if ( (rc = mtd_read_uchar( fd, offset + hcloc, (u_long*)&hc.Calibrate[i].LogPower )) == 1 )
					break;

				offset = offsetof(struct HardConfig_struct, Calibrate[i].LinPower);
				if ( (rc = mtd_read_ushort( fd, offset + hcloc, (u_long*)&hc.Calibrate[i].LinPower )) == 1 )
					break;

				offset = offsetof(struct HardConfig_struct, Calibrate[i].FCCDetect);
				if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, hc.Calibrate[i].FCCDetect, 14 )) == 1 )
					break;

				offset = offsetof(struct HardConfig_struct, Calibrate[i].FCCControl);
				if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, hc.Calibrate[i].FCCControl, 14 )) == 1 )
					break;

				offset = offsetof(struct HardConfig_struct, Calibrate[i].MKKDetect);
				if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, hc.Calibrate[i].MKKDetect, 14 )) == 1 )
					break;

				offset = offsetof(struct HardConfig_struct, Calibrate[i].MKKControl);
				if ( (rc = mtd_read_uchar_array( fd, offset + hcloc, hc.Calibrate[i].MKKControl, 14 )) == 1 )
					break;

			}
		}
		hardconfig_cached = 1;		/* only need to cache it once */
	} while ( 0 );

	if ( fd >= 0 )
		close( fd );

	return rc;
}

int GetHCBootrecords(long *hc_boot_record)
{
	int i = 0, valid = 1;
	
	/* Check if the hardconfig is filled in with valid values for the
	 * exposed interfaces. If it is not, we assume it is an 'old' hardconfig
	 * and we fill in the exposed interfaces ourselves
	 */
	while((i < 3) && valid) {
		if((hc.InterfaceDefinitions[i].role != BR_IF_ROLE_SERVER &&
		    hc.InterfaceDefinitions[i].role != BR_IF_ROLE_CLIENT) ||
		   (hc.InterfaceDefinitions[i].interface_id != BR_IF_ID_ISL36356A &&
		    hc.InterfaceDefinitions[i].interface_id != BR_IF_ID_OEM_IF &&
		    hc.InterfaceDefinitions[i].interface_id != BR_IF_ID_PRODUCT_IF))
		   	valid = 0;
		i++;
	}

	if(valid) {
		memcpy((char *)bootloader_boot_record.exposed_if, hc.InterfaceDefinitions,
			sizeof(bootloader_boot_record.exposed_if));
	} else {
		bootloader_boot_record.exposed_if[0].role = BR_IF_ROLE_SERVER;
		bootloader_boot_record.exposed_if[0].if_id = BR_IF_ID_PRODUCT_IF;
		bootloader_boot_record.exposed_if[0].variant = hc.ProductType;
		bootloader_boot_record.exposed_if[0].btm_compat = hc.ProductType;
		bootloader_boot_record.exposed_if[0].top_compat = hc.ProductType;
		bootloader_boot_record.exposed_if[1].role = BR_IF_ROLE_SERVER;
		bootloader_boot_record.exposed_if[1].if_id = BR_IF_ID_OEM_IF;
		bootloader_boot_record.exposed_if[1].variant = hc.OEMID;
		bootloader_boot_record.exposed_if[1].btm_compat = hc.OEMID;
		bootloader_boot_record.exposed_if[1].top_compat = hc.OEMID;
		bootloader_boot_record.exposed_if[2].role = BR_IF_ROLE_SERVER;
		bootloader_boot_record.exposed_if[2].if_id = BR_IF_ID_ISL36356A;
		bootloader_boot_record.exposed_if[2].variant = BR_IF_VAR_ISL36356A_1;
		bootloader_boot_record.exposed_if[2].btm_compat = BR_IF_VAR_ISL36356A_1;
		bootloader_boot_record.exposed_if[2].top_compat = BR_IF_VAR_ISL36356A_1;
	}

	memcpy((char *)hc_boot_record,
	       (char *)&bootloader_boot_record, sizeof(bootloader_boot_record));
	return 0;
}

inline unsigned long GetHcVersion(void)
{
	return hc.Version;
}

inline unsigned char* GetHcMACAddress(void)
{
	return hc.MACAddress;
}

inline unsigned long GetHcDefaultRegDomain(void)
{
	return hc.DefaultRegDomain;
}

inline unsigned long GetHcProductType (void)
{
	return hc.ProductType;
}

inline unsigned char* GetHcOEMName(void)
{
	return hc.OEMName;
}

inline unsigned char* GetHcProductName(void)
{
	return hc.ProductName;
}

inline unsigned long GetHcOEMID(void)
{
	return hc.OEMID;
}

inline unsigned long GetHcCapabilities(void)
{
	return hc.Capabilities;
}

inline unsigned long* GetHcRegDomains(void)
{
	return hc.RegDomains;
}

inline unsigned char* GetHcSysObjId(void)
{
	return hc.SysObjId;
}

inline unsigned long GetHcHardwareRev(void)
{
	return hc.HardwareRev;
}

inline unsigned long GetHcOutputPower(void)
{
	return hc.OutputPower;
}

inline unsigned long GetHcCalibrateRSSIA(void)
{		// -70 dBm
	return hc.CalibrateRSSIA;
}

inline unsigned long GetHcCalibrateRSSIB(void)
{		// -90 dBm
	return hc.CalibrateRSSIB;
}

inline unsigned long GetHcCalibrateRows(void)
{
	return hc.CalibrateRows;
}

inline unsigned long GetHcCalibrateRows_ptr(void)
{
	return CalibrateRows_ptr;
}


struct HardConfig_struct * GetHcStruct(void)
{
	struct HardConfig_struct *hard_config;
	hard_config = &hc;
	return (hard_config);;
}

#endif /* #ifndef PDA */
