/*  $Header$
 *
 *  Copyright (C) 2002 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

const char passwd_cfg[]="/var/etc/passwd.conf";
const char aplock_cfg[]="/var/etc/aplock.conf";

/*
 * Name: snmp_get_wcs
 *
 * Description: Reads the WriteCommunitystring from the the /var/etc/passwd.conf file
 * returns -1 on error
 */
int snmp_get_wcs(char *password, int maxlen)
{
	FILE *conf_file;
	int ret = 0;

	if ( (conf_file = fopen( passwd_cfg, "r")) == NULL ) {
		memset(password, 0, maxlen);
		return -1;
	}

	if(fgets(password, maxlen, conf_file) == NULL) {
		memset(password, 0, maxlen);
		fclose (conf_file);
		return -1;
	}

	/* Terminate string at newline */
	password[strlen(password) - 1] = '\0';

	fclose (conf_file);
	return ret;
}

/*
 * Name: snmp_get_aplockstate
 *
 * Description: Reads the Lock state from the the /var/etc/aplock.conf file
 * returns:
 *	 1 if AP is locked
 *	 0 if AP is NOT locked
 *	-1 on error
 */
int snmp_get_aplockstate(void)
{
	FILE *conf_file;
	char lockstate[10];
	int ret = 0;

	if ( (conf_file = fopen( aplock_cfg, "r")) == NULL ) {
		return -1;
	}

	if(fgets(lockstate, sizeof(lockstate), conf_file) == NULL) {
		fclose (conf_file);
		return -1;
	}

	/* Terminate string at newline */
	lockstate[strlen(lockstate) - 1] = '\0';
	ret = atoi(lockstate);

	if(ret != 0 && ret != 1)
		ret = -1;

	fclose (conf_file);
	return ret;
}



/*
 * Name: inl_exec
 *
 * Description: Executes 'cmd' in the background using double fork to
 *              prevent 'zombie' processes
 */
void inl_exec(char *const cmd[])
{
  pid_t pid, pid2;

  if ( (pid = vfork()) == 0 )
  {
    if ( ( pid2 = vfork()) == 0)
    {
       execv(cmd[0], cmd);
       
       /* Should never reach this line */      
       exit(-1);
    }
    exit(0);
  }
  else
  {
    /* wait until child 1 ends */
    waitpid(pid, NULL, 0);
  }
}
