/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>

#include "mtd.h"
#include "inl_debug.h"
#include "sockutil.h"

/* return 0 on success, 1 on failure */
int
mtd_read(int fd, int offset, u_long *value, size_t size )
{
	int cnt;

	if (lseek(fd, offset, SEEK_SET) != offset) {
		return 1;
	}
	if ((cnt = readn(fd, value, size)) != size) {
		return 1;
	}
	return 0;
}

/* return 0 on success, 1 on failure */
int
mtd_read_uchar_array ( int fd, int offset, u_char *value, int len )
{
	int cnt;

	/* Huh?! MTD character device returns (end_of_mtd_device - 1) when
	 * an lseek past the end of the device is done iso. -1.
	 */
	if (lseek( fd, offset, SEEK_SET ) != offset) {
		return 1;
	}

	if (len == 0) {
		error("Illegal length for mtd_read_uchar_array!!");
		return 1;
	}

	if ((cnt = readn(fd, value, len)) != len) {
		return 1;
	}
	return 0;
}

/* return 0 on success, 1 on failure */
int
mtd_read_ushort_array( int fd, int offset, u_long *value, int len )
{
	int cnt;
	int i;
	if (lseek( fd, offset, SEEK_SET ) != offset) {
		return 1;
	}

	if (len == 0) {
		error("Illegal length for mtd_read_ushort_array!!");
		return 1;
	}

	for (i = 0 ; i < len; i++) {
		if ((cnt = readn(fd, (value + i), sizeof(u_short))) != sizeof(u_short)) {
			return 1;
		}
	}
	return 0;
}

/* return 0 on success, 1 on failure */
int
mtd_read_ulong_array( int fd, int offset, u_long *value, int len )
{
	int cnt;
	int i;
	if (lseek( fd, offset, SEEK_SET ) != offset) {
		return 1;
	}

	if (len == 0) {
		error("Illegal length for mtd_read_ulong_array!!");
		return 1;
	}

	for (i = 0 ; i < len; i++) {
		if ((cnt = readn(fd, (value + i), sizeof(u_long))) != sizeof(u_long)) {
			return 1;
		}
	}
	return 0;
}

