/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <hardconfig.h>
#include <regdomain.h>

static struct dsss_regulatory_struct
{
    u_char	domain;

    u_char	min_channel;
    u_char	max_channel;
    u_char	default_channel;

} dsss_regulatory_table[] =
{
    {REGDOMAIN_GLOBAL,			1, 13, 6},
    {REGDOMAIN_FCC,				1, 11, 6},
    {REGDOMAIN_IC,				1, 11, 6},
    {REGDOMAIN_ETSI,			1, 13, 6},
    {REGDOMAIN_SPAIN,			10, 11, 11},
    {REGDOMAIN_FRANCE,			10, 13, 11},
    {REGDOMAIN_MKK,				1, 14, 6},
	{(u_char)REGDOMAIN_UNKNOWN,	1, 14, 6},
    {0,							11, 11, 11}
};

/*
 * hwi_regdomain() looks up the regulatory domain structure
 * of the given regulatory domain. If the regulatory domain
 * cannot be found, a safe default is returned.
 *
 * in:	regulatory domain number.
 * ret: structure with regulatory information.
 */

static struct dsss_regulatory_struct *get_regdomain( int domain )
{
    int i = 0;
	for (i =0 ;i < CURRENT_KNOWN_REGDOMAINS; i++)
    {
	if( dsss_regulatory_table[i].domain == (u_char)domain )
	    break;
    }
    return &dsss_regulatory_table[i];
}


/*
 * hwi_default_channel() returns the default channel for
 * the given regulatory domain. If the channels parameter
 * is non-zero, the operational channels for the given
 * regulatory domain are also copied.
 *
 * in:	domain - regulatory domain number.
 *	channels - pointer to null-terminated channelset.
 * ret: default channel number.
 */

int default_channel( int domain )
{
    return get_regdomain( domain )->default_channel;
}

/*
 * hwi_channels() returns the minimum channel, maximum channel
 * and number of channels based on the regulatory domain. If
 * the regulary domain cannot be found, a safe default set
 * is returned.
 *
 * in:	domain - regulatory domain number.
 * out: min - pointer to lowest allowed channel.
 *	max - pointer to maximum allowed channel.
 * ret: number of channels between min and max.
 */

int channels( int domain, int* min, int* max )
{
    struct dsss_regulatory_struct *regdomain;

    regdomain = get_regdomain( domain );

    *min = regdomain->min_channel;
    *max = regdomain->max_channel;

    return( *max - *min + 1 );
}
