/*
 * sockutil.c - socket utility functions
 */
/*  $Header$
 *
 *  Copyright (C) 2001 Intersil Americas Inc
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License.
 * 
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <unistd.h>
#include <errno.h>


/* read n bytes from descriptor */
ssize_t
readn( int fd, void *buf, size_t count )
{
    size_t	nleft;
    ssize_t	nread;
    char	*ptr;

    ptr = buf;
    nleft = count;
    while( nleft > 0 ) {
	if( (nread = read(fd, ptr, nleft)) < 0) {
	    if( errno == EINTR )
		nread = 0;		/* call read() again */
	    else
		return( -1 );		/* error */
	} else if( nread == 0 )
	    break;			/* EOF */

	nleft -= nread;
	ptr += nread;
    }
    return( count - nleft );
}

/* write n bytes to descriptor */
ssize_t
writen( int fd, const void *buf, size_t count )
{
    size_t	nleft;
    ssize_t	nwritten;
    const char	*ptr;

    ptr = buf;
    nleft = count;
    while( nleft > 0 ) {
	if( (nwritten = write(fd, ptr, nleft)) <= 0) {
	    if( errno == EINTR )
		nwritten = 0;		/* call write() again */
	    else
		return( -1 );		/* error */
	}

	nleft -= nwritten;
	ptr += nwritten;
    }
    return( count );
}

