# $Header$
#
# Copyright (C) 2001 Intersil Americas Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

include $(ROCKBASE)/make_include/base_defs.mk

# APPBASE is the base directory of the application sources that is
# needed to put the object files in the APPBUILDDIR directory. If it is
# not defined, we just put the objects in the current directory

ifeq ($(strip $(APPBASE)),)

APPDIR		= $(CURDIR)
OBJDIRND	= $(APPDIR)
OBJDIRD		= $(APPDIR)
EXECDIR		= $(APPDIR)
EXECND		= $(EXECDIR)/$(EXEC)
EXECD		= $(EXECDIR)/$(EXEC)-debug


else

OBJDIRND	= $(subst $(APPBASE),$(APPBUILDDIR)/nodebug,$(CURDIR))
OBJDIRD		= $(subst $(APPBASE),$(APPBUILDDIR)/debug,$(CURDIR))
EXECND		= $(addprefix $(EXECDIR)/, $(EXEC))
EXECD		= $(addprefix $(OBJDIRD)/, $(EXEC))

endif #ifeq ($(APPBASE)),)

OBJECTS		= $(SOURCES:.c=.o)
OBJECTSND	= $(addprefix $(OBJDIRND)/, $(OBJECTS))
OBJECTSD	= $(addprefix $(OBJDIRD)/, $(OBJECTS))
USELIBSND	= $(addprefix $(OBJDIRND)/, $(USELIBS))
USELIBSD	= $(addprefix $(OBJDIRD)/, $(USELIBS))
DEPENDFILE	= ./.depend.$(EXEC)
DEPENDFILEND	= ./.depend.$(EXEC).nd
DEPENDFILED	= ./.depend.$(EXEC).d

ifneq ($(DEPENDFILEND),$(wildcard $(DEPENDFILEND)))
DEP = depend
else
ifneq ($(DEPENDFILED),$(wildcard $(DEPENDFILED)))
DEP = depend
endif
endif

usage : default_usage

default_usage :
	@echo "Usage:" >&2
	@echo " make all		# make $(EXEC)" >&2
	@echo " make clean		# remove all object files of $(EXEC)" >&2
	@echo " make depend		# create depend file" >&2
	@echo " make debug		# make debug version of $(EXEC)" >&2

.PHONY : all


ifeq ($(DEBUG),)
TARGET = $(EXECND)
else
TARGET = $(EXECD)
endif


all : objdir execdir $(TARGET)


$(EXECND) : $(OBJECTSND) $(USELIBSND) $(LIBS)
	$(LD) -r -T $(XARCH)/user.ld $(LDFLAGS) $^ -o $(OBJDIRND)/$(@F).$(OBJFMT)
	$(LD) -T $(XARCH)/user.ld $(LDFLAGS) $^ -o $(OBJDIRND)/$(@F).$(OBJFMTA)
	$(OBJFMT2FLT) $(FLTFLAGS) $(OBJDIRND)/$(@F).$(OBJFMT) -p $(OBJDIRND)/$(@F).$(OBJFMTA) -o $@

$(EXECD) : $(OBJECTSD) $(USELIBSD) $(LIBS)
	$(LD) -r -T $(XARCH)/debug.ld $(LDFLAGS) $^ -o $(OBJDIRD)/$(@F).$(OBJFMTD)
	$(LD) -T $(XARCH)/debug.ld $(LDFLAGS) $^ -o $(OBJDIRD)/$(@F).$(OBJFMTDA)
	$(OBJFMT2FLT) $(FLTFLAGS) $(OBJDIRD)/$(@F).$(OBJFMTD) -p $(OBJDIRD)/$(@F).$(OBJFMTDA) -o $@



depend : 
	@if [ ! -f "$(DEPENDFILEND)" ]; then touch $(DEPENDFILEND); fi
	@makedepend $(CFLAGS) $(INCS) -f $(DEPENDFILEND) -p $(OBJDIRND)/ $(SOURCES)
	@if [ ! -f "$(DEPENDFILED)" ]; then touch $(DEPENDFILED); fi
	@makedepend $(CFLAGS) $(INCS) -f $(DEPENDFILED) -p $(OBJDIRD)/ $(SOURCES)

.PHONY : objdir
objdir ::
	@if [ ! -d "$(OBJDIRND)" -o ! -d "$(OBJDIRD)" ]; then \
		mkdir -p $(OBJDIRND) $(OBJDIRD) ; \
		echo "Creating object directories for $(EXEC)" ; \
		echo "  $(OBJDIRND)" ; \
		echo "  $(OBJDIRD)" ; \
	fi

.PHONY : execdir
execdir ::
	@if [ ! -d "$(EXECDIR)" ]; then \
		mkdir -p $(EXECDIR) ; \
		echo "Creating exec directory for $(EXEC)" ; \
		echo "  $(EXECDIR)" ; \
	fi

.PHONY : clean
clean : default_clean

default_clean :
	@rm -f $(OBJDIRND)/*.$(OBJFMT) $(OBJDIRND)/*.$(OBJFMTA) $(OBJDIRD)/*.$(OBJFMTD) $(OBJDIRD)/*.$(OBJFMTDA)
	@rm -f $(EXECND) $(EXECD)
	@rm -f core $(EXECDIR)/core
	@rm -f $(OBJDIRND)/*.[oais] $(OBJDIRD)/*.[oais]
	@rm -f $(DEPENDFILEND) $(DEPENDFILEND).bak \
		$(DEPENDFILED) $(DEPENDFILED).bak

.SUFFIXES: .o
%.o : %.c
	$(CC) $(CFLAGS) $(DEBUG_CFLAG) $(OPTIMIZATION) -c $< -o $@

$(OBJDIRND)/%.o : %.c
	$(CC) $(CFLAGS) $(OPTIMIZATION) -c $< -o $@

$(OBJDIRD)/%.o : %.c
	$(CC) $(CFLAGS) $(DEBUG_CFLAG) -c $< -o $@

# explicit dependencies
$(OBJECTSND) : $(ROCKBASE)/make_include/base_defs.mk
$(OBJECTSND) : $(ROCKBASE)/make_include/app.mk
$(OBJECTSND) : Makefile
$(OBJECTSND) : $(ROCKBASE)/include/rockhopper.h
$(OBJECTSD) : $(ROCKBASE)/make_include/base_defs.mk
$(OBJECTSD) : $(ROCKBASE)/make_include/app.mk
$(OBJECTSD) : Makefile
$(OBJECTSD) : $(ROCKBASE)/include/rockhopper.h

# include dependfiles for debug and nodebug
ifeq ($(DEPENDFILEND),$(wildcard $(DEPENDFILEND)))
include $(DEPENDFILEND)
endif
ifeq ($(DEPENDFILED),$(wildcard $(DEPENDFILED)))
include $(DEPENDFILED)
endif
