# $Header$
#
# Copyright (C) 2001 Intersil Americas Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
include $(ROCKBASE)/make_include/base_defs.mk

# APPBASE is the base directory of the application sources that is
# needed to put the object files in the APPBUILDDIR directory. If it is
# not defined, we just put the objects in the current directory
ifeq ($(strip $(APPBASE)),)
APPDIR		= $(CURDIR)
OBJDIRND	= $(APPDIR)
OBJDIRD		= $(APPDIR)
TARGETND	= $(patsubst %,lib%.a,$(LIBNAME))
TARGETD		= $(patsubst %,lib%.a,$(LIBNAME))
else
OBJDIRND	= $(subst $(APPBASE),$(APPBUILDDIR)/nodebug,$(CURDIR))
OBJDIRD		= $(subst $(APPBASE),$(APPBUILDDIR)/debug,$(CURDIR))
TARGETND	= $(addprefix $(OBJDIRND)/, $(patsubst %,lib%.a,$(LIBNAME)))
TARGETD		= $(addprefix $(OBJDIRD)/, $(patsubst %,lib%.a,$(LIBNAME)))
endif #ifeq ($(APPBASE)),)

OBJECTS		= $(notdir $(SOURCES:.c=.o))
OBJECTSND	= $(addprefix $(OBJDIRND)/, $(OBJECTS))
OBJECTSD	= $(addprefix $(OBJDIRD)/, $(OBJECTS))
DEPENDFILE	= ./.depend
DEPENDFILEND    = ./.depend.$(LIBNAME).nd
DEPENDFILED     = ./.depend.$(LIBNAME).d

ifneq ($(DEPENDFILEND),$(wildcard $(DEPENDFILEND)))
DEP = depend
else
ifneq ($(DEPENDFILED),$(wildcard $(DEPENDFILED)))
DEP = depend
endif
endif

usage : default_usage

default_usage::
	@echo "Usage:" >&2
	@echo " make all		# make normal & debug of $(LIBNAME)" >&2
	@echo " make clean		# remove all object files of $(LIBNAME)" >&2
	@echo " make depend		# create depend file" >&2

ifeq ($(DEBUG),)
TARGET = nodebug
else
TARGET = debug
endif

.PHONY : all
#all : objdir $(DEP) $(TARGET)		# use this to do makedepend automatically
all : objdir $(TARGET)

.PHONY : debug
debug : $(TARGETD)

.PHONY : nodebug
nodebug : $(TARGETND)

$(TARGETND) : $(OBJECTSND)
	$(AR) $(ARFLAGS) $@ $^

$(OBJECTSND) : $(SOURCES)

$(TARGETD) : $(OBJECTSD)
	$(AR) $(ARFLAGS) $@ $^

$(OBJECTSD) : $(SOURCES)

depend : 
	@if [ ! -f "$(DEPENDFILEND)" ]; then touch $(DEPENDFILEND); fi
	@makedepend $(CFLAGS) $(INCS) -f $(DEPENDFILEND) -p $(OBJDIRND) $(SOURCES)
	@if [ ! -f "$(DEPENDFILED)" ]; then touch $(DEPENDFILED); fi
	@makedepend $(CFLAGS) $(INCS) -f $(DEPENDFILED) -p $(OBJDIRD) $(SOURCES)

.PHONY : objdir
objdir ::
	@if [ ! -d "$(OBJDIRND)" -o ! -d "$(OBJDIRD)" ]; then \
		mkdir -p $(OBJDIRND) $(OBJDIRD) ; \
		echo "Creating object directories for $(EXEC)" ; \
		echo "  $(OBJDIRND)" ; \
		echo "  $(OBJDIRD)" ; \
	fi

.PHONY : clean
clean:
	@rm -f $(OBJDIRND)/*.$(OBJFMT) $(OBJDIRD)/*.$(OBJFMTD)
	@rm -f $(TARGETND) $(TARGETD)
	@rm -f $(OBJDIRND)/*.[oais] $(OBJDIRD)/*.[oais]
	@rm -f $(DEPENDFILEND) $(DEPENDFILEND).bak \
		$(DEPENDFILED) $(DEPENDFILED).bak

.SUFFIXES: .o
%.o : %.c
	$(CC) $(CFLAGS) $(DEBUG_CFLAG) $(OPTIMIZATION) -c $< -o $@

$(OBJDIRND)/%o : %c
	$(CC) $(CFLAGS) $(OPTIMIZATION) -c $< -o $@

$(OBJDIRD)/%o : %c
	$(CC) $(CFLAGS) $(DEBUG_CFLAG) -c $< -o $@

# some explicit dependencies
$(OBJECTSND) : $(ROCKBASE)/make_include/base_defs.mk
$(OBJECTSND) : $(ROCKBASE)/make_include/lib.mk
$(OBJECTSND) : Makefile
$(OBJECTSD) : $(ROCKBASE)/make_include/base_defs.mk
$(OBJECTSD) : $(ROCKBASE)/make_include/lib.mk
$(OBJECTSD) : Makefile

# include dependfiles for debug and nodebug
ifeq ($(DEPENDFILEND),$(wildcard $(DEPENDFILEND)))
include $(DEPENDFILEND)
endif
ifeq ($(DEPENDFILED),$(wildcard $(DEPENDFILED)))
include $(DEPENDFILED)
endif
