/*  $Header$
 *  
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "image.h"
#include "identification/identification.h"
#include "identification/bootrec.h"


struct s_image_h {
    u_long magic1;			/* OKAY */
    u_long vector;			/* firmware entry point */
    u_long magic2;			/* MVC/RESC/SYST */
    u_long length;
    u_long status;			/* flash status */
};

/*
 * Exposed Interfaces
 */
#define BR_NOF_EXP_IF		2
#define BR_EXP_IF_SIZE		((sizeof(struct s_exposed_if_list) - \
				  sizeof(struct s_bootrec_hdr)) / 4)

struct s_exposed_if_list {
	struct s_bootrec_hdr hdr;
	struct s_bootrec_exp_if exposed_if[BR_NOF_EXP_IF];
};

/*
 * Dependent Interfaces. Each dependent interface record
 * specifies that at least one of the interfaces should be
 * present (logical OR).
 * If a component depends on more intefaces, this is specified
 * by defining a dependent interface record for each dependent
 * interface.
 */
#define BR_NOF_DEP_IF_1		1
#define BR_NOF_DEP_IF_2		2
#define BR_DEP_IF_SIZE(x)	((sizeof(struct s_dependent_if_list_##x) - \
				  sizeof(struct s_bootrec_hdr)) / 4)

struct s_dependent_if_list_1 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[BR_NOF_DEP_IF_1];
};

struct s_dependent_if_list_2 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[BR_NOF_DEP_IF_2];
};

/*
 * System Bootrecord Definition
 */
struct s_system_boot_record {
	struct s_image_h image_h;
	struct s_bootrec_comp_id  component_id;
	struct s_bootrec_comp_ver component_version;
	struct s_dependent_if_list_1 dependent_if_list_1;
	struct s_dependent_if_list_2 dependent_if_list_2;
	struct s_exposed_if_list exposed_if_list;
	struct s_bootrec_end  bootrecord_end;
} system_boot_record = {
	image_h:
	{
		MAGIC_OKAY,		/* "OKAY" to identify an image */
		0x00000000,
		BR_COMP_ID_SYST,
		0x00000001,		/* length = 1 */
		0x00000003,		/* flash status, prefered + valid */
	},
	component_id:
	{
		{ BR_CODE_COMPONENT_ID, 1 },
		BR_COMP_ID_SYST
	},
	component_version:
	{
		{ BR_CODE_COMPONENT_VERSION, VER_STRING_SIZE },
		"1.0.2.3 Demo"
	},
	dependent_if_list_1:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(1) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_PRODUCT_IF, 1234},
		}
	},
	dependent_if_list_2:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(2) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_MVC,  2},
			{ BR_IF_ROLE_SERVER, BR_IF_ID_MVC,  3},
		}
	},
	exposed_if_list:
	{
		{ BR_CODE_EXPOSED_IF, BR_EXP_IF_SIZE },
		{	/* role, interface ID, variant, bottom compat, top compat */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_ISL36356A, 0, 0, 0},
			{ BR_IF_ROLE_SERVER, BR_IF_ID_OEM_IF, 0, 0, 0},
		}
	},
	bootrecord_end:
	{
		{ BR_CODE_END_OF_BRA, 5 },	/* bootrec code, length */
		0x2345,				/* CRC-16 */
		0,				/* padding */
		{ 0, 0, 0, 0},			/* MD5 */
	}
};

/*
 * MVC Bootrecord Definition
 */

/*
 * Dependent Interfaces. Each dependent interface record
 * specifies that at least one of the interfaces should be
 * present (logical OR).
 * If a component depends on more intefaces, this is specified
 * by defining a dependent interface record for each dependent
 * interface.
 */
#define BR_MVC_DEP_IF_SIZE(x)	((sizeof(struct s_MVC_dependent_if_list_##x) - \
				  sizeof(struct s_bootrec_hdr)) / 4)

struct s_MVC_dependent_if_list_1 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_MVC_dependent_if_list_2 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

/*
 * Exposed Interfaces
 */
#define MVC_BR_NOF_EXP_IF		2
#define MVC_BR_EXP_IF_SIZE		((sizeof(struct s_MVC_exposed_if_list) - \
					  sizeof(struct s_bootrec_hdr)) / 4)

struct s_MVC_exposed_if_list {
	struct s_bootrec_hdr hdr;
	struct s_bootrec_exp_if exposed_if[MVC_BR_NOF_EXP_IF];
};

struct s_MVC_boot_record {
	struct s_image_h image_h;
	struct s_bootrec_comp_id  component_id;
	struct s_bootrec_comp_ver component_version;
	struct s_MVC_dependent_if_list_1 dependent_if_list_1;
	struct s_MVC_dependent_if_list_2 dependent_if_list_2;
	struct s_MVC_exposed_if_list exposed_if_list;
	struct s_bootrec_end  bootrecord_end;
} MVC_boot_record = {
	image_h:
	{
		MAGIC_OKAY,		/* "OKAY" to identify an image */
		0x00000000,
		BR_COMP_ID_MVC1,
		0x00000001,		/* length = 1 */
		0x00000003,		/* flash status, prefered + valid */
	},
	component_id:
	{
		{ BR_CODE_COMPONENT_ID, 1 },
		BR_COMP_ID_MVC1
	},
	component_version:
	{
		{ BR_CODE_COMPONENT_VERSION, VER_STRING_SIZE },
		"MVCtje 11"
	},
	dependent_if_list_1:
	{
		{ BR_CODE_DEPENDENT_IF, BR_MVC_DEP_IF_SIZE(1) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_OEM_IF,  9654},
		}
	},
	dependent_if_list_2:
	{
		{ BR_CODE_DEPENDENT_IF, BR_MVC_DEP_IF_SIZE(2) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_ISL36356A,  3333},
		}
	},
	exposed_if_list:
	{
		{ BR_CODE_EXPOSED_IF, MVC_BR_EXP_IF_SIZE },
		{	/* role, interface ID, variant, bottom compat, top compat */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_MVC, 0, 0, 0},
			{ BR_IF_ROLE_SERVER, BR_IF_ID_PRODUCT_IF, 0, 0, 0},
		}
	},
	bootrecord_end:
	{
		{ BR_CODE_END_OF_BRA, 5 },	/* bootrec code, length */
		0x9988,				/* CRC-16 */
		0,				/* padding */
		{ 0, 0, 0, 0},			/* MD5 */
	}
};


/*
 * Hardconfig Bootrecord Definition
 */

/*
 * Exposed Interfaces
 */
#define ILL_BR_NOF_EXP_IF		4
#define ILL_BR_EXP_IF_SIZE		((sizeof(struct s_ILL_exposed_if_list) - \
					  sizeof(struct s_bootrec_hdr)) / 4)

struct s_ILL_exposed_if_list {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_exp_if exposed_if[ILL_BR_NOF_EXP_IF];
};

struct s_ILL_boot_record {
	struct s_image_h image_h;
	struct s_bootrec_comp_id  component_id;
	struct s_bootrec_comp_ver component_version;
	struct s_ILL_exposed_if_list exposed_if_list;
	struct s_bootrec_end  bootrecord_end;
} ILLEGAL_boot_record = {
	image_h:
	{
		MAGIC_OKAY,		/* "OKAY" to identify an image */
		0x00000000,
		BR_COMP_ID_MVC1,
		0x00000001,		/* length = 1 */
		0x00000003,		/* flash status, prefered + valid */
	},
	component_id:
	{
		{ BR_CODE_COMPONENT_ID, 1 },
		BR_COMP_ID_BOOT
	},
	component_version:
	{
		{ 0x78295467, VER_STRING_SIZE },	/* ILLEGAL ENTRY */
		"Bootje demo"
	},
	exposed_if_list:
	{
		{ BR_CODE_EXPOSED_IF, ILL_BR_EXP_IF_SIZE },
		{	/* role, interface ID, variant, bottom compat, top compat */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_ISL36356A, 2, 2, 2},
			{ BR_IF_ROLE_SERVER, BR_IF_ID_PRODUCT_IF, 66, 66, 66},
			{ BR_IF_ROLE_SERVER, BR_IF_ID_OEM_IF, 77, 77, 77},
			{ BR_IF_ROLE_SERVER, BR_IF_ID_ISL36356A, 2, 3, 3},
		}
	},
	bootrecord_end:
	{
		{ BR_CODE_END_OF_BRA, 5 },	/* bootrec code, length */
		0x2345,				/* CRC-16 */
		0,				/* padding */
		{ 0, 0, 0, 0},			/* MD5 */
	}
};


static void print_bootrec(struct s_bootrecord *record)
{
	char *p;
	int i, nof_chars;

	printf("Bootrecord %p, size = %d\n", record, bootrecord_size(record));
	printf("Code:\t\t");
	switch (record->hdr.code) {
		case BR_CODE_COMPONENT_ID:
			printf("Component ID\n");
			break;
		case BR_CODE_COMPONENT_VERSION:
			printf("Component Version\n");
			break;
		case BR_CODE_DEPENDENT_IF:
			printf("Dependent IF\n");
			break;
		case BR_CODE_EXPOSED_IF:
			printf("Exposed IF\n");
			break;
		case BR_CODE_END_OF_BRA:
			printf("End of BRA\n");
			break;
		default:
			printf("Unknown Code: %ld\n", record->hdr.code);
			break;
	}
	printf("Size(longs):\t%ld\n", record->hdr.length);
	printf("Data:\n");
	nof_chars = record->hdr.length * sizeof(long);
	for(i = 0, p = (char *)&record->start_of_data; i < nof_chars; i++, p++)
		printf("%.2x ", *p);
	if(isprint(record->start_of_data)) {
		printf("  :  ");
		for(i = 0, p = (char *)&record->start_of_data; i < nof_chars; i++, p++)
			printf("%c", *p);
	}
	printf("\n");
}

void set_exp_OEM_if(unsigned short variant, unsigned short btm_compat,
		    unsigned short top_compat)
{
	system_boot_record.exposed_if_list.exposed_if[1].variant = variant;
	system_boot_record.exposed_if_list.exposed_if[1].btm_compat = btm_compat;
	system_boot_record.exposed_if_list.exposed_if[1].top_compat = top_compat;
}

void set_exp_PROD_if(unsigned short variant, unsigned short btm_compat,
		     unsigned short top_compat)
{
	MVC_boot_record.exposed_if_list.exposed_if[1].variant = variant;
	MVC_boot_record.exposed_if_list.exposed_if[1].btm_compat = btm_compat;
	MVC_boot_record.exposed_if_list.exposed_if[1].top_compat = top_compat;
}

void set_exp_MAC_COPR_if(unsigned short variant, unsigned short btm_compat,
			 unsigned short top_compat)
{
	MVC_boot_record.exposed_if_list.exposed_if[0].variant = variant;
	MVC_boot_record.exposed_if_list.exposed_if[0].btm_compat = btm_compat;
	MVC_boot_record.exposed_if_list.exposed_if[0].top_compat = top_compat;
}

void set_exp_MAC_CTRL_if(unsigned short variant, unsigned short btm_compat,
			 unsigned short top_compat)
{
	system_boot_record.exposed_if_list.exposed_if[0].variant = variant;
	system_boot_record.exposed_if_list.exposed_if[0].btm_compat = btm_compat;
	system_boot_record.exposed_if_list.exposed_if[0].top_compat = top_compat;
}

void set_ILL_BR_CODE(int code)
{
	ILLEGAL_boot_record.component_version.hdr.code = code;
}

void set_ILL_BR_LENGTH(int length)
{
	ILLEGAL_boot_record.component_version.hdr.length = length;

}


void res2string(int res)
{
	if(res == 0)
		printf("Image check Successfull");
	else {
		switch(res) {
		case 1:
			printf("upg_image_invalid");
			break;
		case 2:
			printf("upg_image_unsupported_product");
			break;
		case 3:
			printf("upg_image_unsupported_oem_id");
			break;
		case 4:
			printf("upg_image_unsupported_hardware_revs");
			break;
		case 5:
			printf("upg_image_unsupported_hardconfig");
			break;
		case 6:
			printf("upg_image_unsupported_mvc_version");
			break;
		default:
			printf("Image check failed, result = %d", res);
			break;
		}
	}
}

int print_result(int expected, int actual)
{
	if(expected == actual)
		printf("Success\n");
	else
		printf("Failed\n");
	printf("Expected: ");
	res2string(expected);
	printf("\nActual: ");
	res2string(actual);
	printf("\n");

	if(expected == actual)
		return 0;
	else
		return 1;
}

/* This program tests the image_check library function in the Inl library */
int main(int argc, char *argv[])
{
	long *cur_bootrecs[10];
	image_header_t *new_bootrecs[10];
	image_header_t *ILL_bootrecs[10];
	int res, last_record, i;
	struct s_bootrecord *record;
	int test_nr;
	int nof_test_failed;
	const char *search_devs[] = { "/dev/mtdr0", "/dev/mtdr1", "/dev/mtdr3" , NULL};

	new_bootrecs[0] = (image_header_t *)&MVC_boot_record;
	new_bootrecs[1] = (image_header_t *)&system_boot_record;
	new_bootrecs[2] = NULL;

	ILL_bootrecs[0] = (image_header_t *)&ILLEGAL_boot_record;
	ILL_bootrecs[1] = NULL;

	printf("BRTEST\n");

	res = get_current_bootrecords(search_devs, cur_bootrecs, NULL);

	printf("get current BRs returned %d\n", res);
	if(res >= 0) {
		for(i = 0; (record = (struct s_bootrecord *)cur_bootrecs[i]); i++) {
			last_record = 0;
			printf("Bootrec %d: %p\n", i, record);
			do {
				if((!record) || (valid_bootrecord(record) < 0)) {
					fprintf(stderr, "No valid bootrecord");
					return -1;
				}

				print_bootrec(record);
				if(record->hdr.code == BR_CODE_END_OF_BRA) {
					last_record = 1;
				}
				record = next_bootrecord(record);
			} while(!last_record);
		}
	}

	/*
	 * SYSTEM DEPENDENCIES:
	 * { BR_IF_ROLE_SERVER, BR_IF_ID_PRODUCT_IF, 1234} &&
	 * ( { BR_IF_ROLE_SERVER, BR_IF_ID_MVC,  2} ||
	 *   { BR_IF_ROLE_SERVER, BR_IF_ID_MVC,  3} )
	 *
	 * MVC DEPENDENCIES:
	 * { BR_IF_ROLE_SERVER, BR_IF_ID_OEM_IF, 9654}
	 * { BR_IF_ROLE_SERVER, BR_IF_ID_ISL36356A, 3333}
	 */

	test_nr = 1;
	nof_test_failed = 0;

	/* OEM Interface */
	printf("\n********\nTEST %d: Unsupported OEM (0,0,0)\n", test_nr);
	set_exp_PROD_if(1234, 1234, 1234);
	set_exp_MAC_COPR_if(2, 2, 2);
	set_exp_MAC_CTRL_if(3333, 3333, 3333);

	set_exp_OEM_if(0, 0, 0);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(3, res);
	test_nr++;

	printf("\n********\nTEST %d: Unsupported OEM - bmt compat > top compat (0,9655,9654)\n", test_nr);
	set_exp_OEM_if(0, 9655, 9654);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(3, res);
	test_nr++;

	printf("\n********\nTEST %d: OEM supported within compat (0,9653,9654)\n", test_nr);
	set_exp_OEM_if(0, 9653, 9654);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(0, res);
	test_nr++;

	printf("\n********\nTEST %d: OEM supported within compat (0,9654,9654)\n", test_nr);
	set_exp_OEM_if(0, 9654, 9654);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(0, res);
	test_nr++;

	printf("\n********\nTEST %d: OEM supported within compat (0,9654,9655)\n", test_nr);
	set_exp_OEM_if(0, 9654, 9655);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(0, res);
	test_nr++;

	printf("\n********\nTEST %d: OEM supported in variant, not within compat (9654,0,0)\n", test_nr);
	set_exp_OEM_if(9654, 0, 0);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(0, res);
	test_nr++;

	/* Product Interface*/
	printf("\n********\nTEST %d: Unsupported Product\n", test_nr);
	set_exp_OEM_if(9654, 9654, 9654);
	set_exp_PROD_if(0, 0, 0);
	set_exp_MAC_COPR_if(2, 2, 2);
	set_exp_MAC_CTRL_if(3333, 3333, 3333);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(2, res);
	test_nr++;

	/* MVC Interface */
	printf("\n********\nTEST %d: Unsupported MVC Version\n", test_nr);
	set_exp_OEM_if(9654, 9654, 9654);
	set_exp_PROD_if(1234, 1234, 1234);
	set_exp_MAC_COPR_if(0, 0, 0);
	set_exp_MAC_CTRL_if(3333, 3333, 3333);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(6, res);
	test_nr++;

	/* MAC Controller HW Interface */
	printf("\n********\nTEST %d: Unsupported Hardware\n", test_nr);
	set_exp_OEM_if(9654, 9654, 9654);
	set_exp_PROD_if(1234, 1234, 1234);
	set_exp_MAC_COPR_if(2, 2, 2);
	set_exp_MAC_CTRL_if(0, 0, 0);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(4, res);
	test_nr++;

	/* Succeed */
	printf("\n********\nTEST %d: Success\n", test_nr);
	set_exp_OEM_if(9654, 9654, 9654);
	set_exp_PROD_if(1234, 1234, 1234);
	set_exp_MAC_COPR_if(2, 2, 2);
	set_exp_MAC_CTRL_if(3333, 3333, 3333);
	res = image_file_check(new_bootrecs);
	nof_test_failed += print_result(0, res);
	test_nr++;


	/* Illegal Bootrecord Header */
	printf("\n********\nTEST %d: Illegal Bootrecord Header - Illegal CODE\n", test_nr);
	set_ILL_BR_CODE(0x75767675);
	set_ILL_BR_LENGTH(VER_STRING_SIZE);
	res = image_file_check(ILL_bootrecs);
	nof_test_failed += print_result(1, res);
	test_nr++;

	printf("\n********\nTEST %d: Illegal Bootrecord Header - Illegal LENGTH\n", test_nr);
	set_ILL_BR_CODE(BR_CODE_COMPONENT_VERSION);
	set_ILL_BR_LENGTH(513);
	res = image_file_check(ILL_bootrecs);
	nof_test_failed += print_result(1, res);
	test_nr++;

	if(nof_test_failed == 0) {
		printf("\n\nAll tests successfull\n");
		res = 0;
	}
	else {
		printf("\n\n%d Tests FAILED\n", nof_test_failed);
		res = 1;
	}

	return res;
}

