/*
changedhcp.c

Adjust some things in a dhcp data image
The image can then be extenden with an udp and an ip header using spak or
nemesis

usage:
changedhcp [-h] <options>

*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#define false (1==0)
#define true (0==0)

#ifdef DEBUG
	int traceon=false;
	#define TRACE( x ) {if (traceon) {printf("%s %d:", __FILE__, __LINE__); x;}}
#else
	#define TRACE( x ) {}
#endif

int findOption( int argc, const char* const * argv, const char * option, int reqPars )
{
	int myArgvItt;
	for (myArgvItt = 1; myArgvItt<argc-reqPars; myArgvItt++)
	{
		if (strcmp(argv[myArgvItt],option)==0)
		{
			return myArgvItt;
		}
	}
	return 0;
}

unsigned char* stomac( const unsigned char* s, unsigned char mac[6] )
{
	unsigned a,b,c,d,e,f;

	if (sscanf( s, "%x:%x:%x:%x:%x:%x", &a,&b,&c,&d,&e,&f )!=6)
	{
		fprintf(stderr,"Invald mac adres format %s\n", s);
		exit(-1);
	}
	mac[0]=a; mac[1]=b; mac[2]=c; mac[3]=d; mac[4]=e; mac[5]=f;

	return mac;
}

int main( int argc, const char * const *argv)
{
	unsigned char* myBuff =(unsigned char*) malloc(65536);
	int packetsize=65536;
	int i=0;
	int fd=0;

	if (findOption(argc,argv,"-h",0)!=0)
	{
		printf(
			"changedhcp - change parameters in a dhcp data image\n"
			"    and send the adjusted image to standard out\n"
			"usage: changedhcp [-v] -i <file> [-x <xid>] [-sm <mac addr>] [-rip <rip>]\n"
			"                  [-sip <sip>]\n"
			"	<file> 	input file. should be a valid dhcp image\n"
			"	<xid>   new transaction id\n"
			"	<mac>	source mac adress. Format=a:b:c:d:e:f\n"
			"   <sip>   server ip address\n"
			"	<rip>	request ip in options.\n"
			);
		exit(0);
	}
	if (findOption(argc,argv,"-i",1)==0)
	{
		fprintf(stderr,"ERROR: no input file specified. use -i <file>\n");
		exit(-1);
	} 

	fd = open( argv[findOption(argc,argv,"-i",1)+1], O_RDONLY );
	if (fd==-1)
	{
		fprintf( stderr, "Error: could not open file \"%s\", error=%d %s\n",
		        argv[findOption(argc,argv,"-i",1)+1], 
				errno, strerror(errno) );
		exit(-1);
	}
	packetsize= read( fd, myBuff, packetsize ); 
	TRACE( fprintf(stderr,"DEBUG: ActualSize = %d\n",packetsize) )
	myBuff = realloc( myBuff, packetsize );
	close( fd );

	for (i=1; i<argc; i++)
	{
		if (strcmp(argv[i],"-i")==0)
		{
			/* skip this item */
			i++;
		}
		else if (strcmp(argv[i],"-v")==0)
		{
			traceon=true;
		}
		else if (strcmp(argv[i],"-v-")==0)
		{
			traceon=false;
		}
		else if (strcmp(argv[i],"-sm")==0)
		{
			i++;
			if (i==argc)
			{
				fprintf(stderr,"ERROR: no parameter for option -sm\n");
				exit(-1);
			}

			/* then convert the mac adres and put it in the buffer */
			stomac( argv[i], myBuff+28 );
		}
		else if (strcmp(argv[i],"-sip")==0)
		{
			struct in_addr myaddr;
			unsigned myOptOffset;
			i++;
			if (i==argc)
			{
				fprintf(stderr,"ERROR: no parameter for option -sip\n");
				exit(-1);
			}
			
			if (inet_aton(argv[i],&myaddr)==0)
			{
				fprintf(stderr,"ERROR: -rip addres error\n");
				exit(-1);
			}
			
			/* fill in request ip in dhcp options */
			/* first find start of option */
			for (myOptOffset = 0xf0; 
				myOptOffset<packetsize; 
				myOptOffset+=(unsigned)(myBuff[myOptOffset+1])+2 )
			{
				TRACE(printf("cmd=%x, size=%x\n", myBuff[myOptOffset], myBuff[myOptOffset+1]))
				if (myBuff[myOptOffset]==0x36)
				{
					*((struct in_addr*)(&(myBuff[myOptOffset+2]))) = myaddr;
					break;
				}
			}

			if (myOptOffset>=packetsize)
			{
				fprintf(stderr,"ERROR: no server ip address option found in dhcp image file.");
				exit(-1);
			}
		}
		else if (strcmp(argv[i],"-x")==0)
		{
			int xid=0;
			i++;
			if (i==argc)
			{
				fprintf(stderr,"ERROR: no parameter for option -x\n");
				exit(-1);
			}

			if (sscanf( argv[i], "%i", &xid )!=1)
			{
				fprintf(stderr,"ERROR: no valid par for option -x\n");
				exit(-1);				
			}

			*((int*)(myBuff+4)) = htonl( xid );
		}
		else if (strcmp(argv[i],"-rip")==0)
		{
			struct in_addr myaddr;
			unsigned myOptOffset;
			i++;
			if (i==argc)
			{
				fprintf(stderr,"ERROR: no parameter for option -rip\n");
				exit(-1);
			}
			if (inet_aton(argv[i],&myaddr)==0)
			{
				fprintf(stderr,"ERROR: -rip addres error\n");
				exit(-1);
			}
			
			/* fill in request ip in dhcp options */
			/* first find start of option */
			for (myOptOffset = 0xf0; 
				myOptOffset<packetsize; 
				myOptOffset+=(unsigned)(myBuff[myOptOffset+1])+2 )
			{
				TRACE(printf("cmd=%x, size=%x\n", myBuff[myOptOffset], myBuff[myOptOffset+1]))
				if (myBuff[myOptOffset]==0x32)
				{
					*((struct in_addr*)(&(myBuff[myOptOffset+2]))) = myaddr;
					break;
				}
			}

			if (myOptOffset>=packetsize)
			{
				fprintf(stderr,"ERROR: no request address option found in dhcp image file.");
				exit(-1);
			}
		}
		else
		{
			fprintf(stderr,"ERROR: unknown option %s\n", argv[i]);
			exit(-1);
		}
	}

	write(fileno(stdout), myBuff,  packetsize);
	return 0;
}
