/***************************************************
* file:     dhcpclientemu.c
* simple dhcpc
*
* Simple single packet capture program
*****************************************************/
#pragma pack(1)
#include "packet.h"
#pragma pack()
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/if_ether.h> /* includes net/ethernet.h */
#include <unistd.h>

#define false (1==0)
#define true (0==0)
#ifdef DEBUG
unsigned traceon=false;
#define TRACE(x) \
	{if (traceon) {printf("%s %d:", __FILE__, __LINE__); x;}}
#else
#define TRACE(x) {}
#endif

unsigned char* stomac( const char *smac, unsigned char buff[6] )
{
	int a,b,c,d,e,f;

	if (sscanf( smac, "%x:%x:%x:%x:%x:%x", &a,&b,&c,&d,&e,&f )!=6)
	{
		fprintf(stderr, "ERROR: mac adress format should be a:b:c:d:e:f with a f inbetween 00 and ff\n" );
		exit(-1);
	}
	buff[0]=a; buff[1]=b; buff[2]=c; buff[3]=d; buff[4]=e; buff[5]=f;

	return buff;
}

/* next function is NOT thread safe */
const char* mactos( const unsigned char aMac[6], char* buff )
{
	int a,b,c,d,e,f;
	a=aMac[0]; b=aMac[1]; c=aMac[2]; d=aMac[3]; e=aMac[4]; f=aMac[5];
	sprintf( buff, "%x:%x:%x:%x:%x:%x", a,b,c,d,e,f );
	return buff;
}

int main(int argc, const char * const *argv)
{
	/* decode options */
	int filter_on_xid = false;
	unsigned valid_xid=0;
	int filter_on_dm = false;
	unsigned char valid_dm[6];
	int fd;
	unsigned found=false;
	int i;
	for (i=1; i<argc; i++)
	{
		if (strcmp(argv[i],"-h")==0)
		{
			printf(	"hatsji	- halt until selected dhcp client packet arrives\n"
				"then print info about that packet\n"
				"Usage:\n"
				"hatsji [-dm <mac>] [-x <xid>]\n"
				"	-x <xid>     only continue when dhcp xid is value\n"
				"	-dm <mac>	only continue when destination equals mac\n"
				"   -v          print debug info (only when compiled with -DDEBUG\n"
		  		);
			exit(0);
		}
		else if (strcmp(argv[i],"-v")==0)
		{
			traceon=true;
		}
		else if (strcmp(argv[i],"-x")==0)
		{
			i++;
			if (i==argc)
			{
				fprintf(stderr,"ERROR: commandline option -x expects a parameter\n");
				exit(-1);
			}
			filter_on_xid = true;
			sscanf( argv[i], "%d", &valid_xid );
		}
		else if (strcmp(argv[i],"-dm")==0)
		{
			i++;
			if (i==argc)
			{
				fprintf(stderr,"ERROR: commandline option -dm expects a parameter\n");
				exit(-1);
			}
			filter_on_dm = true;
			stomac( argv[i], valid_dm );
		}
		else
		{
			fprintf(stderr,"ERROR: unknown commandline option %s\n", argv[i]);
			exit(-1);
		}
	}

	/* sniff packets until the selected arrives */
	/* use a raw socket for this: */
	
	fd = socket( AF_INET, SOCK_PACKET, htons(ETH_P_ALL));
	if (fd == -1)
	{
		fprintf(stderr, "ERROR: could not open raw socket. Errno=%d: %s\n", errno, strerror(errno));
		exit (-1);
	}

	/* and try to sniff */
	while (!found)
	{
		unsigned char buff[60000];
		struct udp_dhcp_packet* ipudpdhcp=0;
		unsigned size=0;
		char wbuff1[30],wbuff2[30];

		size = recvfrom( fd, buff, sizeof(buff), 0, 0,0 );
		TRACE( printf("Packet recieved. size=%d, destmac=%s, srcmac=%s, ethertype=%x\n",
		               size, mactos(buff,wbuff1), mactos(buff+6,wbuff2),
		               ntohs(*((short*)(buff+12))) ));
		TRACE( 
			{
				int i;
				for ( i=0; i< size; i++)
				{
					if (i % 16 ==0) printf("\n");
					printf("%2x ", (unsigned)(buff[i]) );
				}
				printf("\n");
			} )
		ipudpdhcp = (struct udp_dhcp_packet*)(buff+14);
		if(ntohs(*((short*)(buff+12)))==0x800) /* ip */
		{
			/* ok, so it is IP */
			TRACE(printf("IP\n"))
			if (ipudpdhcp->ip.protocol==17) /* udp */
			{
				TRACE(printf("UDP: src=%x, dest=%x\n",
						(unsigned)ntohs(ipudpdhcp->udp.source), 
						(unsigned)ntohs(ipudpdhcp->udp.dest)))
				if (ntohs(ipudpdhcp->udp.dest)==68 && /* dhcp message */
		   		    ntohs(ipudpdhcp->udp.source)==67) /* dhcp message */
				{
					TRACE(printf("DHCP: xid=%x\n", ntohl(ipudpdhcp->data.xid)))
					if((filter_on_xid==false || 
					       ntohl(ipudpdhcp->data.xid)==valid_xid) &&
					   (filter_on_dm==false  || 
					      (valid_dm[0]==buff[0] && valid_dm[1]==buff[1] &&
					       valid_dm[2]==buff[2] && valid_dm[3]==buff[3] &&
					       valid_dm[4]==buff[4] && valid_dm[5]==buff[5])) 
					  )
					{
						struct in_addr myclientip;
						struct in_addr myserverip;
						myclientip.s_addr = ipudpdhcp->data.yiaddr;
						myserverip.s_addr = ipudpdhcp->ip.saddr;
						
						printf("servermac=%s ", mactos(buff+6,wbuff2));
						printf("clientip=%s ", inet_ntoa( myclientip ) );
						printf("serverip=%s ", inet_ntoa( myserverip ) );

						found=true;
					}
					else
					{
						TRACE(printf("dhcp Packet recieved but not selected\n"));
					}
				}
				else
				{
					TRACE(printf("udp packet is not dhcp\n"))
				}
			}
			else
			{
				TRACE(printf("ip packet is not udp"))
			}
		}
		else
		{
			TRACE(printf("not dhcp/udp/ip\n"));
		}
	}
	close(fd);
		
    return 0;
}
