/*  $Header$
 *
 *  Copyright (C) 2003 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * crc32.c -- calculate CRC32 and append to copy of file. Prints 
 * hexadecimal file length on stdout.
 * This is to be used for generating files for the Bootrom upload
 * command.
 */

#include <stdio.h>

unsigned long crc_update( unsigned long crc, unsigned long val )
{
   unsigned long t0, t1;

   val ^= crc;
   val ^= ( val << 6);
   t1   = val << 24;
   t0   = t1 ^ ( t1 >> 1);
   val  = t0 ^ ( t1 >> 2);
   val ^= ( val >> 10 );
   val ^= ( t0 >> 4 );
   val ^= ( t0 >> 7 );
   val ^= ( t0 >> 22 );
   val ^= ( t1 >> 16 );
   val ^= ( t1 >> 26 );
   return( val ^ (crc >> 8) );
}

int main( int argc, char **argv )
{
   if( argc == 3 )
   {
      FILE *in = fopen( argv[1], "rb" );
      FILE *out = fopen( argv[2], "wb" );
      int crc = ~0;
      int c;
      int len = 0;

      if( !in )
         perror( argv[1] );
      else if( !out )
         perror( argv[2] );
      else
      {
	 while( (c = getc(in)) != EOF )
	 {
	    putc( c, out );
	    crc = crc_update( crc, c );
	    len++;
	 }
	 crc = ~crc;
	 printf( "%x\n", len );
	 putc( crc, out );
	 putc( crc >> 8, out );
	 putc( crc >> 16, out );
	 putc( crc >> 24, out );
         return( 0 );
      }
   }
   else
   {
      fprintf( stderr, "usage: %s <infile> <outfile>\n", argv[0] );
      return( 1 );
   }
}
