#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

struct malloc_log_t {
	
	char* addr;
	char* function;
	char* file;
	long long bytes;
        struct malloc_log_t *next;	
	struct malloc_log_t *prev;
};



/*
 * Split a strign into 0-terminated parts and return them
 * Thanks to Ico Doornekamp
 */

int split (char *str, char sep, char **part, int maxparts)
{
	int parts = 0;
	char *p1 = str;
	char *p2;
	
	part[parts++] = p1;

	while ((p2 = strchr (p1, sep)) && (parts < maxparts)) {
		*p2 = 0;
		p1 = p2 + 1;
		part[parts++] = p1;
	}

	return parts;
}


struct malloc_log_t* list;


void dump_log(void) {
	struct malloc_log_t *ml;
	long long bytes = 0;

	ml = list;
	
	while(ml) {
		
		printf("Address %s malloced by %s @%s (%lli bytes)\n", 
				ml->addr,
				ml->function,
				ml->file,
				ml->bytes
				);
		
		bytes += ml->bytes;
		
		ml = ml-> next;
	}
	
	printf("\nTotal bytes still malloced: %lli\n", bytes);
}
	


struct malloc_log_t* find_addr(char* addr) {
	
	struct	malloc_log_t* ml = list;
		
	while(ml) {
		if (strcmp(ml->addr, addr) == 0) {
			return ml;
		}
		ml = ml->next;
	}
	
	return NULL;
}

void free_malloc(struct malloc_log_t* ml) {
	
	assert(ml);
	
	if(ml->next){
		ml->next->prev = ml->prev;
	}
	
	if(ml->prev){
		ml->prev->next = ml->next;
	}
	else {
		list = ml->next;
	}
	
	free(ml->addr);
	free(ml->function);
	free(ml->file);
	free(ml);
	
}

struct malloc_log_t* new_malloc(char* buf) {
	
	struct malloc_log_t* ml = malloc(sizeof(struct malloc_log_t));
	char* part[40];
	int parts;
	ml->prev = NULL;	
	
	if(!list) {
		ml->next = NULL;
		list = ml;
	}
	else {
		ml->next = list;
		ml->next->prev = ml;
		list = ml;
	}
	
	parts = split(buf, ' ' , part, 30);
	
	ml->bytes = atoi(part[2]);
	ml->addr = strdup(part[8]);
	ml->function = strdup(part[5]);
	ml->file = strdup(part[6]);
	
	
	return (ml);
}


int main (int argc, char**argv) {

	FILE* f;
	char buf[160];
	int line = 0;
	char* p;
	int no_skip = 0;
	struct malloc_log_t *ml;

	if (argc < 2 || argc > 3) {
		printf("Usage: %s filename [skip]\n\n", argv[0]);
		exit(-1);
	}	

	if (argc == 3) {
		no_skip = atoi(argv[2]);
	}

	list = NULL;

	f = fopen(argv[1], "r");
	if (!f) {
		printf("Couldn't open input file (%s)\n", argv[1]);
		exit (-1);
	}

	printf("Skipping first %i lines\n", no_skip);

	while(fgets(buf, 160, f)) {
		

		line++;
		
		/* skip first X lines */
		if (line <= no_skip) {
			continue;
		}
		
		/* chomp */
		buf[strlen(buf)-1] = 0;
		
		if ( (strncmp(buf, "malloc", 6) == 0)
				|| (strncmp(buf,"calloc", 6) == 0)
				|| (strncmp(buf,"strdup", 6) == 0)
				){
			p = strstr(buf, "0x");
			
			if (!p) {
				printf("Parse error on line %i\n", line);
				printf("Line %i: %s\n", line, buf);
			}
			assert(p);
				
			if(find_addr(p) != NULL) {
				printf("Malloc/Calloc returned already alloc-ed address (%s)\n", p);
				return -1;
			}
			new_malloc(buf);
			
		}
		else if (strncmp(buf, "free", 4) == 0) {
			char* p2;
			
			p = strdup(strstr(buf, "0x"));
			assert(p);
			
			p2 = strchr(p, ' ');
			assert(p2);
			*p2 = 0;
			
			ml = find_addr(p);
			free(p);
			if(ml) {
				free_malloc(ml);
			}
			else {
				printf("Freeing unknown address in: %s\n", buf);
			}
			
						
		}
		else {
			printf("unknown token in line %i: %s\n", line, buf);
		}
	}
	
	printf( "read %i lines\n", line);
	
	dump_log();
	
	return (0);
}
