/*  $Header$
 *  
 *  Copyright (C) 2001 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <prismoids.h>
#include <sys/time.h>

int data;
int *data_p;

#define MAX_MALLOCS	100
static char *mallocs[MAX_MALLOCS];

#define BUFFER_SIZE 5000

typedef void (*func) (void);

static void func_with_stack_overflow(void)
{
    int i;
    char stack_overflow_buffer[BUFFER_SIZE];

    for (i = 0; i < BUFFER_SIZE; i++) {
	stack_overflow_buffer[i] = 2;
    }
}

static void usage(char *name)
{
    printf("Usage: %s option\n", name);
    printf("\t1\t\t: write to memory area outside own user process\n");
    printf("\t2\t\t: generate a stack overflow\n");
    printf("\t3\t\t: read at memory area outside own user process\n");
    printf("\t4\t\t: generate a prefetch abort\n");
    printf("\t5\t\t: perform many mallocs to use <n> mpu areas\n");
    printf("\t6\t\t: test verify_area via an ioctl to get data to area outside process space\n");
    printf("\t7 <1-8>\t\t: test abort with different instructions (ldr/str/ldrh/strh/swp/ldrsb/ldrsw/ldm/stm\n");
    printf("\t8\t\t: test undefined instruction abort.\n");
    printf("\t9 <a> <l>\t: READ <l> bytes from address 0x<a>\n");
    printf("\t10 <a> <l> <c>\t: WRITE <l> bytes <c> from address 0x<a>\n");
    exit(0);
}

int main(int argc, char *argv[])
{
    int tmp = 0, tmp2 = 0, i, ret, len;
    unsigned long addr;
    int nof_mallocs;
    struct timeval tv1, tv2;
    long tdiff;
    volatile char *rwbuf;
    volatile char c;

    func null_pointer_function;

    if (argc < 2)
	usage(argv[0]);

    switch (atoi(argv[1])) {
    case 1:
	printf("writing to memory area outside own user process\n");
	data_p = &data;
	data_p += 0x10000;
	printf("Writing at address %p (original data = %p)\n", data_p,
	       &data);
	*data_p = 0x55;
	break;
    case 2:
	printf("Generating stack overflow\n");
	func_with_stack_overflow();
	break;
    case 3:
	printf("reading memory area outside own user process\n");
	data_p = &data;
	data_p += 0x10000;
	printf("Reading address %p\n", data_p);
	tmp = *data_p;
	printf("Data = %d", tmp);
	break;
    case 4:
	printf("Generating prefetch abort\n");
	null_pointer_function = NULL;
	null_pointer_function();
	break;
    case 5:
	nof_mallocs = 0;
	if (argv[2])
	    nof_mallocs = atoi(argv[2]);
	if (nof_mallocs < 1 || nof_mallocs > MAX_MALLOCS)
	    nof_mallocs = 10;
	printf("Using %d MPU areas\n", nof_mallocs);
	set_malloc_mmap_size(4000);
	for (i = 0; i < nof_mallocs; i++) {
	    mallocs[i] = (char *) malloc(4000);
	    if (!mallocs[i])
		break;
	    sprintf(mallocs[i], "malloc testje %d\n", i);
	    printf("%s", mallocs[i]);
	}
	nof_mallocs = i;
	printf("Generating 100000 random accesses on MPU regions\n");
	gettimeofday(&tv1, NULL);
	for (i = 0; i < 100000; i++) {
	    int index = rand() % nof_mallocs;
	    sprintf(mallocs[index], "malloc testje %d\n", i);
	}
	gettimeofday(&tv2, NULL);
	tdiff =
	    (tv2.tv_sec * 1000000 + tv2.tv_usec) - (tv1.tv_sec * 1000000 +
						    tv1.tv_usec);
	printf("Time: %ld.%ld\n", tdiff / 1000000, tdiff % 1000000);
	break;
    case 6:
	data_p = &data;
	data_p += 0x10000;


	printf("Doing an ioctl to get data at pointer outside our memory space: %p\n", data_p);

	ret =
	    get_prismoid("eth1", DOT11_OID_PRIVTXREJECTED, data_p,
			 sizeof(long));
	printf("ioctl to wireless interface driver returned %d\n", ret);
	perror("error returned");
	break;
    case 7:
	if (argc != 3)
	    usage(argv[0]);

	addr = 0x12345678;
	switch (atoi(argv[2])) {
	case 1:		/* ldrh */
	    printf("ldrh at addr 0x%lx\n", addr + 2);
	  __asm__ __volatile__("ldrh %0, [%1, #2]": "=r"(tmp):"r"(addr));
	    break;
	case 2:		/* strh */
	    printf("strh at addr 0x%lx\n", addr - 2);
	    __asm__ __volatile__("strh %0, [%1, #-2]"::"r"(tmp),
				 "r"(addr));
	    break;
	case 3:		/* swp */
	    printf("swp at addr 0x%lx\n", addr);
	  __asm__ __volatile__("swp %0, %1, [%2]": "=r"(tmp):"r"(tmp2), "r"(addr));
	    break;
	case 4:		/* swpb */
	    printf("swpb at addr 0x%lx\n", addr);
	  __asm__ __volatile__("swpb %0, %1, [%2]": "=r"(tmp):"r"(tmp2), "r"(addr));
	    break;
	case 5:		/* ldrsb */
	    printf("ldrsb at addr 0x%lx\n", addr + 3);
	  __asm__ __volatile__("ldrsb %0, [%1, #3]": "=r"(tmp):"r"(addr));
	    break;
	case 6:		/* ldrsh */
	    printf("ldrsh at addr 0x%lx\n", addr - 4);
	    __asm__ __volatile__("ldrsh %0, [%1, #-4]"::"r"(tmp),
				 "r"(addr));
	    break;
	case 7:		/* ldmia */
	    printf("ldmia of 2 registers at addr 0x%lx\n", addr + (1 * 4));
	    __asm__ __volatile__("ldmia %2, {%0, %1}":"=r"(tmp),
				 "=r"(tmp2):"r"(addr));
	    break;
	case 8:		/* stmdb */
	    printf("stmdb of 2 registers at addr 0x%lx\n", addr - (2 * 4));
	    __asm__ __volatile__("stmdb %2, {%0, %1}"::"r"(tmp),
				 "r"(tmp2), "r"(addr));
	    break;
	default:
	    printf("Unknown option %d\n", atoi(argv[2]));
	    break;
	}
	break;
    case 8:
	printf("Executing undefined instruction 0b0000.0111.1111.0000.0000.0000.1111.0000");
	__asm__ __volatile__(".long 0x07F000F0");
	break;
    case 9:
	if (argc != 4)
	    usage(argv[0]);

	rwbuf = (char *)strtoul(argv[2], NULL, 16);
	len = atoi(argv[3]);

        printf("READING %d bytes at address %p\n", len, rwbuf);
        for(i = 0; i < len; i++) {
	    c = *rwbuf;
	    printf("  %d\n", c);
	    rwbuf++;
	}
	break;
    case 10:
	if (argc != 5)
	    usage(argv[0]);

	rwbuf = (char *)strtoul(argv[2], NULL, 16);
	len = atoi(argv[3]);
        c = (char) atoi(argv[4]);

        printf("WRITING %d bytes (%d) at address %p\n", len, c, rwbuf);
	for(i = 0; i < len; i++) {
	    *rwbuf = c;
	    rwbuf++;
	}
	break;
    default:
	printf("Unknown option %d\n", atoi(argv[1]));
	break;
    }
    exit(0);
}
