#include <stdio.h>

#define u_short unsigned short

#include "pda.h"
#include "pdaTestData.h"

#define BIG_ENDIAN 1  /* 0 for LITTLE_ENDIAN */

/*=======================================================================================*
 | This Program generates a binary file from a PDA.                                      |
 | Currently "PdaTestData" is used from Software/firmware/OS/uClinux/test/pdaTestData.h" |
 | A CRC16 is calculated and appended to the binary file.                                |
 *---------------------------------------------------------------------------------------*/
int main(int argc, char **argv)
{
   FILE *fh;
   unsigned short  pdr_Size;
   unsigned short  cnt;
   unsigned short  pdaData;
   unsigned char   D1;
   unsigned char   D2;
   unsigned short  Crc = 0;
   
   if (argc != 2) {
      printf("Usage: pda2bin <OutputFilename>\n");
      printf("  pda2bin generates a binary file from a static pda-structure\n");
      return 0;
   }

   fh = fopen (argv[1], "wb");
   if (fh == 0) {
      printf("Error opening %s for write access\n", argv[1]);
      return 0;
   }

   /*-----------------------------------*
    | How many shorts long is the PDA ? |
    *-----------------------------------*/
   pdr_Size = sizeof(PdaTestData);
   printf("\nsizeof(PdaTestData) = %d bytes (excl. CRC)\n\n", pdr_Size*2);
  
   /*------------------------------------*
    | Walk through the shorts of the PDA |
    *------------------------------------*/
   for (cnt=0; cnt<pdr_Size; cnt++) {
      pdaData = PdaTestData[cnt];
      Crc += pdaData;
#if (BIG_ENDIAN == 1)
      D1 = (unsigned char)(pdaData >> 8);
      D2 = (unsigned char)pdaData;
#else
      D1 = (unsigned char)pdaData;
      D2 = (unsigned char)(pdaData >> 8);
#endif
      fwrite(&D1, (size_t)1, (size_t)1, fh);
      fwrite(&D2, (size_t)1, (size_t)1, fh);
   }

   /*---------------*
    | Write the CRC |
    *---------------*/
#if (BIG_ENDIAN == 1)
      D1 = (unsigned char)(Crc >> 8);
      D2 = (unsigned char)Crc;
#else
      D1 = (unsigned char)Crc;
      D2 = (unsigned char)(Crc >> 8);
#endif
   fwrite(&D1, (size_t)1, (size_t)1, fh);
   fwrite(&D2, (size_t)1, (size_t)1, fh);
   
   fclose (fh);

   return 0;
}
