/*  $Header$
 *
 *  Copyright (C) 2002 Intersil Americas Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "identification/identification.h"
#include "identification/bootrec.h"
#include "identification/image_header.h"
#include "include/component_interface.h"

/*
 * Exposed Interfaces
 */
#define BR_EXP_IF_SIZE		((sizeof(struct s_exposed_if_list) - \
				  sizeof(struct s_bootrec_hdr)) / 4)
/*
 * Dependent Interfaces. Each dependent interface record
 * specifies that at least one of the interfaces should be
 * present (logical OR).
 * If a component depends on more intefaces, this is specified
 * by defining a dependent interface record for each dependent
 * interface.
 */
#define BR_DEP_IF_SIZE(x)	((sizeof(struct s_dependent_if_list_##x) - \
				  sizeof(struct s_bootrec_hdr)) / 4)

#ifndef PRIVATE_BUILD
struct s_dependent_if_list_1 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_dependent_if_list_2 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_dependent_if_list_3 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};

struct s_dependent_if_list_4 {
	const struct s_bootrec_hdr hdr;
	const struct s_bootrec_dep_if dependent_if[1];
};
#endif /* PRIVATE_BUILD */

/*
 * System Bootrecord Definition
 */
const struct s_system_boot_record {
        struct image_header_st header;
        struct s_bootrec_comp_ver component_version;
#ifndef PRIVATE_BUILD
        struct s_dependent_if_list_1 dependent_if_list_1;
        struct s_dependent_if_list_1 dependent_if_list_2;
        struct s_dependent_if_list_1 dependent_if_list_3;
        struct s_dependent_if_list_1 dependent_if_list_4;
#endif /* PRIVATE_BUILD */
	struct s_bootrec_end  bootrecord_end;
} system_boot_record = {
        header:
        {
            MAGIC_OKAY,
            0x0000,
            0x0000,
            0x0000,
            0x0000,
            BR_CODE_COMPONENT_ID, 
            1,
            BR_COMP_ID_DFS1
        },
	component_version:
	{
		{ BR_CODE_COMPONENT_VERSION, VER_STRING_SIZE },
		WFS_VERSION_STRING
	},
#ifndef PRIVATE_BUILD	
        dependent_if_list_1:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(1) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_PRIMARY_OEM_IF, WFS_NEEDED_OEM_IF},
		}
	},
        dependent_if_list_2:
	{
	        { BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(1) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_PRIMARY_PRODUCT_IF, WFS_NEEDED_PRODUCT_IF},
		}    
            
	},
        dependent_if_list_3:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(1) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_SNMP_IF, WFS_NEEDED_SNMP_IF},
		}
	},
        dependent_if_list_4:
	{
		{ BR_CODE_DEPENDENT_IF, BR_DEP_IF_SIZE(1) },
		{	/* role, interface ID, variant */
			{ BR_IF_ROLE_SERVER, BR_IF_ID_DAP1_FS, WFS_NEEDED_FILESET_IF},
		}
	},
#endif /* PRIVATE_BUILD */	
        bootrecord_end:
	{
		{ BR_CODE_END_OF_BRA, 5 },	/* bootrec code, length */
		WEBFILESET_CRC16,			/* CRC-16 */
		0,				/* padding */
		WEBFILESET_MD5,			/* MD5 */
	}
};

